;-================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;- Read LICENSE file for additional copyright information (GNU GPL)
;-
;-  file  	: ansdoc.inc
;-  created	: 23-mai-00
;-  modified	: 17-jun-00
;-  version	; 0.24 (re "version.inc" for actual version no.)
;-  assembler	: nasm 0.98-3e
;-  description	: Forth kernel sharable dynamically loadable library.
;-		: ANS system documentation
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: "lib4th"/"calc4th"/"f8"
;-		: represent a minimal forth system.
;-  source	: me, f.i.g., many others and, the DPANS94 doc.
;-		:	http://home.snafu.de/phpr/
;-  requires	: compiling:	asmutils, nasm-0.98-3e
;-		: running:	nothing (w.r.t. the program)
;-		: programming:	thorough knowledge of Forth
;-		: system: 	i586-pc-linux 2.2.+
;-================================================================;
;-
;-	display/print this text w. at least 96 characters/line.
;-

;-	

;-
;-
;-		ANS system documentation
;-
;- { name(s) }
;-	(sequence of) Forth word(s)
;- "consequences are not predictable"
;-	used to describe certain error conditions means exactly that, i.e. ranges from silent 
;-	recovery till at worst, system crash and any kind of data or even hardware destruction.
;-
;-


;-
;-
;- From "dpans94":
;- 
;-================================================================;
;-
;-               4. Documentation requirements
;-
;- 4.1 System documentation
;-
;- 4.1.1 Implementation-defined options
;-
;- Linux signals:
;- single segfaults recover to {abort}; an immediately following other segfault most certainly 
;- will result in orderly program termination, w/ console reset to state on entry, as with {bye}.
;- arithmetic "faults" recover to the rsp overflow values.
;- "sigwinch" is used to re-fetch the main window's sizes.
;- CTR/C keyboard input triggers a "sigint" which will force program termination.
;- other signals send a text message or terminate the program, as appropriate.
;- all signal handlers may be modified, at runtime.
;-
;-	
;- CORE wordset
;-
;- The implementation-defined items in the following list represent characteristics and choices 
;- left to the discretion of the implementor, provided that the requirements of this Standard 
;- are met. A system shall document the values for, or behaviors of, each item.
;-
;-	{forth also ans}
;-	voc-stack setup required for ANS conformance.
;-
;- * aligned address requirements
;-   (3.1.3.3 Addresses)
;-	codespace requires no alignment.
;-	cell alignment required for data- and return-stacks.
;-	(double) cell alignment required for (optional) mmx op-s.
;-
;- * behavior of 6.1.1320 EMIT for non-graphic characters
;-	EMIT adjusts count in {out}, otherwise does not "behave"...
;-
;- * character editing of 6.1.0695 ACCEPT and 6.2.1390 EXPECT
;-	#preliminary# kbd input controls:
;-	<bel>	^G
;-		keybord signal (sound by PC speaker).
;-	<bs>	^H
;-		move cursor one char back, back-space.
;-	<esc>[D	cur.left
;-		ditto (vt102 type terminal).
;-	<tab>	^I
;-		non-overwriting space until eol, moves cursor right;
;-		if 1st keystroke initiates "last line recall" by cursor down.
;-	<esc>[C	cur.right
;-		non-overwriting space, moves cursor right - as <tab>.
;-	<esc>[A	cur.up
;-		cursor back to line beginning.
;-	<esc>[B	cur.dn
;-		after a <tab> expands till previous end of line.
;-	<ff>	^L
;-		clear console display.
;-	<del>	ascii 127, "<=" key
;-		backup and delete char under cursor.
;-	<nl>	ascii 10, <enter> key, ^J
;-		line terminator is <lf>.
;-	<cr>	^M
;-		carriage return, sends <cr/lf>.
;-	<lf>	^J
;-		line feed, exec <cr/lf>
;-
;-	<csi>	aequivalent to sequence of <esc> and "[".
;-	after input of any other ctrl-char <del>eting might require 
;-	one <del> more than chars visble because of leading <esc>.
;-	ctrl-chars (1..31,127) handling user definable via {cc4th}.
;-
;- * character set
;-   (3.1.2 Character types, 6.1.1320 EMIT, 6.1.1750 KEY)
;-	internal handling based on iso-latin-1, w/ ctrl codes 0...31. EMIT takes up to 4 chars 
;-	(bytes) per cell which will be sent, consecutively, l.s. byte first, until the remainder 
;-	is zero, in preparation for unicode which might be implemented, later.
;-
;- * character-aligned address requirements
;-   (3.1.3.3 Addresses)
;-	none
;-
;- * character-set-extensions matching characteristics
;-   (3.4.2 Finding definition names)
;-	words may not begin or end with blank spaces or control characters.
;-
;- * conditions under which control characters match a space delimiter
;-   (3.4.1.1 Delimiters)
;-	in {find} and {word} when parsing with {bl} as the delimiter.
;-
;- * format of the control-flow stack
;-   (3.2.3.2 Control-flow stack)
;-	control-flow stack is data-stack, unlimited within bounds of data-stack. items are 
;-	2 cells, ea. {ds-pick} provided for access to datastack while intermediately interpreting.
;-
;- * conversion of digits larger than thirty-five
;-   (3.2.1.2 Digit conversion)
;-	not defined, expect unpredictable results, base := 2...198, 36+ not tested, though 
;-	supposedly ok. permissible range: 1..256. w/ 256 no conversion at all.
;-
;- * display after input terminates in 6.1.0695 ACCEPT and 6.2.1390 EXPECT
;-	-?- (hp: I do not quite understand what this refers to)
;- 	-?- re below, * system prompt
;-		">>" while interpreting
;-		"==" while compiling
;-
;- * exception abort sequence
;-   (as in 6.1.0680 ABORT")
;-	restore base reference (i.e. 4th-dll kernel).
;-	restore RP from user vari R0.
;-	restore PS from user vari S0.
;-	check consistency of currently used user vari context,
;-		on failure generate signal SIGSTKFLT and terminate.
;-	force inter-active, interpreting state.
;-	push recursive entry into {quit} to RS.
;-	continue in standard {quit} loop, expecting any input.
;-
;- * input line terminator
;-   (3.2.4.1 User input device)
;-	code 10, ascii <lf>
;-
;- * maximum size of a counted string, in characters
;-   (3.1.3.4 Counted strings, 6.1.2450 WORD);
;-	255, no limit on <nul>-enclosed strings.
;-
;- * maximum size of a parsed string
;-   (3.4.1 Parsing)
;-	no limit
;-
;- * maximum size of a definition name, in characters
;-   (3.3.1.2 Definition names)
;-	255
;-
;- * maximum string length for 6.1.1345 ENVIRONMENT?, in characters
;-	255
;-
;- * method of selecting 3.2.4.1 User input device
;-	Default input channel is STDIN, which may be any readable device. Scripting works w. 
;-	STDIN from scriptfile, other chans as set on invocation, according to linux i/o channels 
;-	(STDOUT, STDERR) which default to the actual console device, if not re-directed. STDIN is 
;-	the respective file channel, also, while {includ}ing, etc. A word {name} can temporarily 
;-	receive keyboard input w. { KBD' name }.
;-	Get piped or re-directed channels by {in-chan}, {out-chan}, {err-chan}.
;-	Supplied parameters take precedence, after the evaluation of which current console will 
;-	become STDIN. the lib4th kernel provides some means to i/o redirection, by "channels".
;-
;- * method of selecting 3.2.4.2 User output device
;-	defaults to STDOUT
;-	storing a valid channel number to (stdout) selects output device. "channel number" is a 
;-	byte (8bit) sized index into lib4th channels redirection table providing 16 items for i/o 
;-	channels description. re "User input device" and, glossary for i/o channels redirection.
;-
;- * methods of dictionary compilation
;-   (3.3 The Forth dictionary)
;-	compiled code left in place, codespace ptr advanced accordingly, instead of immediately 
;-	being executed, which is "interpreting". word headers, ptrs, names cpl'd into data-space.
;-
;- * number of bits in one address unit
;-   (3.1.3.3 Addresses)
;-		8
;-
;- * number representation and arithmetic
;-   (3.2.1.1 Internal number representation)
;-	binary, w. signed numbers in 2-s complement representation, symmetric -ve/+ve range(s).
;-
;- * ranges for n, +n, u, d, +d, and ud,
;-   3.1.3 Single-cell types:
;-	n	                             -2147483647...                             2147483647
;-	+n	                                       0...                             2147483647
;-	u	                                       0...                             4294967295
;-	single integer notification by dpl@ = -1
;-   3.1.4 Cell-pair types:
;-	d	                    -9223372036854775807...                    9223372036854775807
;-	+d                                             0...                    9223372036854775807
;-	ud                                             0...                   18446744073709551615
;-	double integer notification by ve dpl@ = position of fractional mark before end of string.
;-   non-standard double cell pairs' "rational numbers" at absolute accuracy and, 
;-		"irrational numbers", displayable in floating point format at an accuracy 
;-		limited to the available digits:
;-	abs dv natural range and 0, -9223372036854775807...                    9223372036854775807
;-	abs dv < 1		            0,0000000004...                           0,9999999995
;-		and any proper fractions as fractions within the above ranges, accurately;
;-		improper fractions w/ a resolution of the rsp part accumulated to up to the 
;-		maximum of at the actual base displayable digits, less one diggit.
;-	real numbers notification by dpl@ > 0 and bit 23 set.
;-   non-standard "quad" q (ca. +/-17*10^36), uq, and multi-cell numbers:
;-	q	-170141183460469231731687303715884105727...170141183460469231731687303715884105727
;-	uq	                                       0...340282366920938463463374607431768211455
;-	quad integer notification by +ve dpl@ w. bit 15 set, count in bits 0..14.
;-   non-standard "bignum" nn, by counted cells groups:
;-	nn				up to +/- 10^2000 (tested, more may be possible)
;-	counted, multi cell integers, range limited by available memory
;-	bignum notification by -ve dpl@, $ff80nnnn, w/ dpl-count nnnn.
;-	maximal input is nnnn=255 due to limit imposed by {word} fetching a byte-counted string.
;-
;-   common to all arithmetic-numeric variants:
;-	Zero with only m.s.bit set ($80...0), is not strictly an integer in the representable 2-s 
;-	complement number range as, for instance, after negation its value would remain the same, 
;-	by either sequence of
;-		hex 80000000 invert 1+	or, hex 80000000 negate
;-	thus it cannot effectively be negated to result in a different, +ve value which, it has in 
;-	common with the value of Zero, therefore it will be regarded as the "negative zero" within 
;-	lib4th context, and not count to the valid range of two's complement integers.
;-	Further,
;-		{ ... DUP NEGATE 2DUP < ROT ROT > 2DUP . . = . }
;-	applied to
;-		{ -1 1 RSHIFT INVERT ... } - which is based on a silly "assumption" - 
;-	as well as to
;-		{ s" MAX-N" ENVIRONMENT? DROP ... } - which should always be correct -
;-	got to yield valid results, required by ANS standard. I.e. above sequence (from a certain 
;-	ANS compliance test) made to looking like standard, which it isn't, just plain nonsense...
;-	(die tiefe weisheit der protagonisten von c.l.f. deutet an, da man offenkundig bereit ist, 
;-	(diese ungereimtheit zu bersehen, um sich nicht mit der irregularitaet einer "negativen 
;-	(null" auseinandersetzen zu mssen - was an der tatsache als solcher leider nichts ndert.
;-
;- * read-only data-space regions
;-   (3.3.3 Data space)
;-	all pre-compiled code-space, which may be set to writable w/ the {mprotect} system call.
;-
;- * size of buffer at 6.1.2450 WORD
;-   (3.3.3.6 Other transient regions)
;-	working min. 160 bytes, default is {pad-size}, initially 4K bytes, at {here}. 
;-	pads and space for {word} modifiable at runtime w/ user-vari {(padsz)}.
;-
;- * size of one cell in address units
;-   (3.1.3 Single-cell types)
;-		4
;-
;- * size of one character in address units
;-   (3.1.2 Character types)
;-		1 (up to 4 may be used e.g. for unicode, prepared, not yet used)
;-
;- * size of the keyboard terminal input buffer
;-   (3.3.3.5 Input buffers)
;-	default is {page-size}, as of Linux PAGE_SIZE (4K),
;-	<d_mintib> = 256 bytes for local buffers e.g., by {included}.
;-
;- * size of the pictured numeric output string buffer
;-   (3.3.3.6 Other transient regions)
;-	min 160 bytes, {pad-size}, default is 4K, from (padsz), which may be modified at runtime. 
;-	other transient region at {sp@ cell +}, {unused}-1 free cells, used for intermediate 
;-	string storage, by {open-file} etc.
;-
;- * size of the scratch area whose address is returned by 6.2.2000 PAD
;-   (3.3.3.6 Other transient regions)
;-	{pad-size} min 160 bytes, max { 64 1024 cell */ depth - }, 
;-	initially 4K bytes, runtime-modifiable w/ user-vari {(padsz)}.
;-
;- * system case-sensitivity characteristics
;-   (3.4.2 Finding definition names)
;-	default is case independently, case-sensitivity may be explicitely set to case dependent 
;-	or case independent mode; upper to lower case mapping independently, by i/o sub-system.
;-
;- * system prompt
;-   (3.4 The Forth text interpreter, 6.1.2050 QUIT)
;-	">>" while interpreting
;-	"==" while compiling
;-	2nd char can be replaced by any action after storing the true 
;-	exec. address of a rsp. word to user variable {(ok)}, e.g.:
;-		{ ' .s idx>xec 4th>abs (ok) ! }
;-
;- * type of division rounding
;-   (3.2.2.1 Integer division, 6.1.0100 */, 6.1.0110 */MOD,
;-   (6.1.0230 /, 6.1.0240 /MOD, 6.1.1890 MOD
;-	round near zero
;-
;- * values of 6.1.2250 STATE when true
;-	except that <zero> state is always "interpreting", use of STATE is not fully compliant 
;-	to the standard:
;-		l.s. 16-bit carries no. of structures nesting levels,
;-		m.s. 16-bit is compile-flags, immediacy and others.
;-	
;-		.1.1	interpreting if at least l.s.word (16-bit)zero.
;-		.1.2	interpreting if flag <mimed> in state is set,
;-			i.e. between brackets [ ... ] when compiling.
;-		.1.3	any other non-zero value is compiling state.
;-	
;-	{cpl?} can be used to determining whether in permanently compiling state.
;-
;- * values returned after arithmetic overflow
;-   (3.2.2.2 Other integer operations)
;-	division: minimum or maximum, dependent on arguments
;-	addition/subtraction/multiplication:
;-		result truncated to expected size,
;-		in compiled code {flags} may be used to read processor 
;-		flags immediately after a rsp. words' execution.
;-	'ranum' operations return the 'overflow indicator' { 0,0 }.
;-
;- * whether the current definition can be found after 6.1.1250 DOES>
;-   (6.1.0450 :)
;-	No.
;-	
;-----------------------------------------------------------------;
;- BLOCK wordset
;-	
;-
;-    7.4.1.1 Implementation-defined options
;-    
;-     * the format used for display by 7.6.2.1770 LIST (if implemented);
;-		16 lines w/ 3 chars of a leading line no and 64 chars of screenfile content
;-
;-     * the length of a line affected by 7.6.2.2535 \ (if implemented).
;-		{c/l}, constant value of 64.
;-
;-	
;-----------------------------------------------------------------;
;- FILE wordset
;-
;-   
;- 11.3.2 Blocks in files
;-	implemented w/ BLOCK and BLOCK-EXT wordsets plus, {blkfile} vocabulary.
;-
;- * file access methods used by 11.6.1.0765 BIN, 11.6.1.1010 CREATE-FILE, 
;-   11.6.1.1970 OPEN-FILE, 11.6.1.2054 R/O, 11.6.1.2056 R/W,11.6.1.2425 W/O
;-	since <fam> is a cell item, newly created files' permissions set from
;-	user variable (FPERM), further subject to host's <umask> processing.
;-	For fully specified open call lib4th provides the additional word, 
;-	{mopen-file} to passing double cell of access mode and permissions.
;-	Linux specific flags and fd-s stored in 'channel definition table'.
;-		BIN		doesnt apply, a {noop}
;-
;- * file exceptions
;-	No 'exceptions', OS specific error code returned, re man (2) open & file "inc/errno.inc".
;-
;- * file line terminator (11.6.1.2090 READ-LINE);
;-	<lf>, preceding <cr> silently discarded.
;-
;- * file name format (11.3.1.4 File names);
;-	Linux/UNIX style, i.e. no conventions other than that <nul>s aren't permitted.
;-	Max size of path & filename is 4K bytes, often restricted to 1K by Linux kernel.
;-
;- * Information returned by 11.6.2.1524 FILE-STATUS
;-	-ve error or, ptr to description of file status
;-	required buffer size is linux kernel specific, re "linux/asm/stat.h"
;-	lib4th provides {stat-size} which returns the required buffer size.
;-	lx kernel 2.2. stat structure ("SHORT" is 16-bit, half CELL):
;-		00 	SHORT st_dev	; device
;-			SHORT pad1
;-		04	CELL  st_ino	; inode
;-		08	SHORT st_mode	; protection
;-		0a	SHORT st_nlink	; number of hard links
;-		0c	SHORT st_uid	; user ID of owner
;-		0e	SHORT st_gid	; group ID of owner
;-		10	SHORT st_rdev	; device type (if inode device)
;-			SHORT pad2
;-		14	CELL  st_size	; total size, in bytes
;-		18	CELL  st_blksize; blocksize for filesystem I/O
;-		1c	CELL  st_blocks	; number of blocks allocated
;-		20	CELL  st_atime	; time of last access
;-		24	CELL  unused1
;-		28	CELL  st_mtime	; time of last modification
;-		2c	CELL  unused2
;-		30	CELL  st_ctime	; time of last change
;-		34	CELL  unused3
;-		38	CELL  unused4
;-		3c	CELL  unused5
;-
;- * Input file state after an exception (11.6.1.1717 INCLUDE-FILE, 11.6.1.1718 INCLUDED)
;-	file remains open and accessible via {work}-channel, file pointer unchanged,
;-	stdin as set in next higher nesting level.
;-
;- * ior values and meaning (11.3.1.2 I/O results)
;-	Linux i/o error coding, re man (2) pages and file "inc/errno.inc".
;-
;- * maximum depth of file input nesting (11.3.4 Input source)
;-	limited by available stacks memory, default is 256 levels at empty stacks.
;-
;- * maximum size of input line (11.3.6 Parsing);
;-	255 bytes
;-
;- * methods for mapping block ranges to files (11.3.2 Blocks in files);
;-	a block is 1024 bytes, adding up to the rsp. file pointer.
;-
;- * number of string buffers provided (11.6.1.2165 S");
;-	one, at {pad2}, plus any no. of locally assigned buffers (re LOCALS wordset)
;-
;- * size of string buffer used by 11.6.1.2165 S"
;-	dft 4K bytes, also used by C", Z", ", INCLUDE".
;-
;-

;-
;-================================================================;
;-
;- 4.1.2 Ambiguous conditions
;-   
;- CORE wordset
;-
;- A system shall document the system action taken upon each of the
;- general or specific ambiguous conditions identified in this Standard.
;- See 3.4.4 Possible actions on an ambiguous condition.
;-
;- The following general ambiguous conditions could occur because of
;- a combination of factors:
;-
;- * a name is neither a valid definition name nor a valid number
;-   during text interpretation (3.4 The Forth text interpreter)
;-	return to line editor where the word concerned may be editted and re-entered, if input 
;-	source is terminal; message and return to {quit}, expecting input from next higher 
;-	nesting level, otherwise.
;-
;- * a definition name exceeded the maximum length allowed
;-   (3.3.1.2 Definition names)
;-	name length truncated to 255 bytes, consequentional results undefined.
;-
;- * addressing a region not listed in 3.3.3 Data Space
;-	self-recovering segfault after single attempt, segfault and terminating the program/job 
;-	concerned after repeated, un-recoverable segfaults. "recoverable" relates to e.g. false 
;-	addressing by { 0 >r } or { 0 0 ! }. a single <enter> or, return to {quit} may be required 
;-	to effectively recovering.
;-	segfaults-producing segfaults may not always be catched and, could run into an endless 
;-	loop. CTR/C can always be used to termination but, after a segfault may corrupt the 
;-	console i/o. - which can be restored w/ Linux' <reset> or, a new L4 job's {rs} word.
;-
;- * argument type incompatible with specified input parameter, e.g.
;-   passing a flag to a word expecting an n (3.1 Data types):
;-	stack underflow error may occur if number of cells not sufficient.
;-
;- * attempting to obtain the execution token, (e.g., with 6.1.0070 ',
;-   6.1.1550 FIND, etc.) of a definition with undefined interpretation semantics
;-	*obtaining* a token is independent of its semantics.
;-
;- * dividing by zero (6.1.0100 */, 6.1.0110 */MOD, 6.1.0230 /,
;-   6.1.0240 /MOD, 6.1.1561 FM/MOD, 6.1.1890 MOD, 6.1.2214 SM/REM,
;-   6.1.2370 UM/MOD, 8.6.1.1820 M*/)
;-	respective overflow value w. sign of dividend, 1 in the case of 0/0.
;-
;- * insufficient data-stack space or return-stack space
;-   (stack overflow);
;-	datastack:   entry into {quit} w. empty stack
;-	returnstack: data-stack top down overwriting, then segfault or other system error.
;-
;- * insufficient space for loop-control parameters
;-	return-stack error, probably
;-		data-stack overwriting, then segfault or, undeterminable system error.
;-
;- * insufficient space in the dictionary
;-	error message, ret to {quit} (v0.1.10: which was not yet tested).
;-
;- * interpreting a word with undefined interpretation semantics
;-	error message, ret to {quit}
;-
;- * modifying the contents of the input buffer or a string literal
;-   (3.3.3.4 Text-literal regions, 3.3.3.5 Input buffers)
;-	dependent on actual setting of memory protection mode; initially no specific action 
;-	by "lib4th" code, Linux kernel responds w. segfault to access to non-alterable memory 
;-	which will be responded to by "lib4th" w. (hopefully) orderly program/job termination.
;-	initial write protection applies to the calling program code and to lib4th. program, 
;-	lib4th and compilespace memory by dft marked readable, non-writeable and, executeable.
;-	{tib} is in r/w memory, {evaluate}d source may be non-writable, string literals are 
;-	non-modifiable in kernel or program space, and may be modifiable in compile-space.
;-
;- * overflow of a pictured numeric output string
;-	conversion is safe w. up to 4K single byte characters, i.e. overflow not possible if 
;-	the supplied words were used, up to 'quad' integers, which require at most 132 bytes. 
;-	else stg extends into data memory, below {here} w. most probably fatal consequences.
;-	when converting in local dataspace the overflow results depend on actual locals useage, 
;-	which may be overwriting previously stored variable values.
;-
;- * parsed string overflow
;-	string truncation to 255 bytes of length.
;-
;- * producing a result out of range, e.g., multiplication (using *) results in a value too big 
;-   to be represented by a single-cell integer 
;-   (6.1.0090 *, 6.1.0100 */, 6.1.0110 */MOD, 6.1.0570 >NUMBER, 6.1.1561 FM/MOD, 
;-    6.1.2214 SM/REM, 6.1.2370 UM/MOD, 6.2.0970 CONVERT, 8.6.1.1820 M*/);
;-	multiplication & string to number conversion will lead to truncated values,
;-	division results to respective overflow (min/max) values.
;-	number to string conversion always error free.
;-
;- * reading from an empty data stack or return stack (stack underflow)
;-	data-stack:	return to {quit} w. zero depth stack ptr.
;-	return-stack:	1st job would segfault,
;-		        later jobs extend into older jobs'
;-		        variables dataspace and may segfault.
;-
;- * unexpected end of input buffer, resulting in an attempt to use a zero-length string as a name.
;-	return to {quit}
;-
;-----------------------------------------------------------------;
;-
;- The following specific ambiguous conditions are noted in the
;- glossary entries of the relevant words:
;- 
;- * >IN greater than size of input buffer (3.4.1 Parsing)
;-	return to {quit}
;-
;- * 6.1.2120 RECURSE appears after 6.1.1250 DOES>
;-	compiling error message, ret to {quit} - yet to be solved!
;-
;- * argument input source different than current input source
;-   for 6.2.2148 RESTORE-INPUT
;-	any error might occur! most probably return to {quit} and loss of
;-	actual input source specification at time of attempt to restauration.
;-
;- * data space containing definitions is de-allocated
;-   (3.3.3.2 Contiguous regions)
;-	how? - neither the text interpreter nor the compiler would then be available.
;-
;- * data space read/write with incorrect alignment
;-   (3.3.3.1 Address alignment)
;-	does not apply to lib4th supplied system, i.e no alignment required.
;-
;- * data-space pointer not properly aligned
;-   (6.1.0150 ,, 6.1.0860 C,)
;-	does not apply to lib4th supplied system, i.e no alignment required.
;-
;- * less than u+2 stack items 
;-   (6.2.2030 PICK, 6.2.2150 ROLL);
;-	PICK might read erroneous values, ROLL could segfault, as with stack bounds underflow.
;-
;- * loop-control parameters not available ( 6.1.0140 +LOOP, 6.1.1680 I, 
;-   6.1.1730 J, 6.1.1760 LEAVE, 6.1.1800 LOOP, 6.1.2380 UNLOOP)
;-	Return to 3rd level caller but, consequences are not predictable.
;-
;- * most recent definition does not have a name (6.1.1710 IMMEDIATE)
;-	{immediate} applies to latest 'visible' word - which has a 4th header. compileflag words 
;-	as IMMEDIATE, UNSMUDGE, etc. should not be applied to headerless words, which are words 
;-	without a header and thus, no flags in a header... 
;-
;- * name not defined by 6.2.2405 VALUE used by 6.2.2295 TO
;-	TO stores to the address where >BODY would point to and, does (almost) no feasibilty 
;-	checking. It may also be applied to VARIABLEs and LOCALs. Expect any, unpredictable 
;-	consequences if applied to words for which useage of {to} is not explicitely permitted.
;-
;- * name not found
;-   (6.1.0070 ', 6.1.2033 POSTPONE, 6.1.2510 ['], 6.2.2530 [COMPILE])
;-	error message and return to {quit} via {abort}.
;-
;- * parameters are not of the same type
;-   (6.1.1240 DO, 6.2.0620 ?DO, 6.2.2440 WITHIN)
;-	excess cells remain on stack. errors may occur, accordingly.
;-
;- * 6.1.2033 POSTPONE or 6.2.2530 [COMPILE] applied to 6.2.2295 TO
;-	TO may be POSTPONEd or [COMPILE]d.
;-
;- * string longer than a counted string returned by 6.1.2450 WORD
;-	>IN advanced by actually received length, returned string size truncated to 255 bytes.
;-
;- * u greater than or equal to the number of bits in a cell
;-   (6.1.1805 LSHIFT, 6.1.2162 RSHIFT)
;-	shift words operate modulo 32(cell) or 64(double cell).
;-
;- * word not defined via 6.1.1000 CREATE (6.1.0550 >BODY, 6.1.1250 DOES>)
;-	Values returned by >BODY are meaningless or, at least, not reliably usable if applied to 
;-	words not built with CREATE. >BODY returns <pfa+4> of word header which is 1st by header
;-	creation unused address in data-memory. New headers follow at this address if no space 
;-	{allot}ted prior to new defns. Compiled code after DOES> receives this address on tos.
;-
;- * words improperly used outside 6.1.0490 <# and 6.1.0040 #>
;-   (6.1.0030 #, 6.1.0050 #S, 6.1.1670 HOLD, 6.1.2210 SIGN).
;-	"improper" results.., i.e. data between {here} and {pad}/{pad2} might be overwritten,
;-	which is of no consequences to other words.
;-
;-
;-----------------------------------------------------------------;
;- BLOCK wordset
;-
;- 7.4.1.2 Ambiguous conditions
;-    
;- * Correct block read was not possible;
;-	rsp. opr returns zero or doesn't operate.
;-
;- * I/O exception in block transfer;
;-	No such 'exception', rsp opr results in a noop or returns zero.
;-
;- * Invalid block number (7.6.1.0800 BLOCK, 7.6.1.0820 BUFFER, 7.6.1.1790 LOAD);
;-	rsp. opr returns zero or doesn't operate.
;-
;- * A program directly alters the contents of 7.6.1.0790 BLK;
;-	'normal' continuation w/ the changed blk
;-
;- * No current block buffer for 7.6.1.2400 UPDATE.
;-	an un-used, un-assigned (new) block of buffers would be constituted, noop otherwise.
;-
;-
;-----------------------------------------------------------------;
;- FILE wordset
;-
;-
;- * attempting to position a file outside its boundaries
;-   (11.6.1.2142 REPOSITION-FILE);
;-	file size extended accordingly, by host fss.
;-
;- * attempting to read from file positions not yet written
;-   (11.6.1.2080 READ-FILE, 11.6.1.2090 READ-LINE);
;-	file size extended accordingly, by host fss, <nul> bytes read.
;-
;- * fileid is invalid (11.6.1.1717 INCLUDE-FILE);
;-	operation specific error message and return to quit.
;-
;- * I/O exception reading or closing fileid 
;-   (11.6.1.1717 INCLUDE-FILE, 11.6.1.1718 INCLUDED)
;-	none
;-
;- * named file cannot be opened (11.6.1.1718 INCLUDED);
;-	message and {quit}
;-
;- * requesting an unmapped block number (11.3.2 Blocks in files);
;-	extends file to appropriate size, if write enabled,
;-	else returns last available block and <eof>.
;-
;- * using 11.6.1.2218 SOURCE-ID when 7.6.1.0790 BLK is not zero.
;-	SOURCE-ID is channel no. ORed w. max-chan, of rsp blockfile.
;-	
;-	
;-
;-----------------------------------------------------------------;
;- 
;- 4.1.3 Other system documentation
;-   
;- A system shall provide the following information:
;- 
;- * list of non-standard words using 6.2.2000 PAD
;-   (3.3.3.6 Other transient regions)
;-	path+file name composing word {[fname]}, and derived, store stg result to {pad}.
;-	use {fname"} w/ supplied buffer to protecting {pad}.
;-	stg fetching words store text to {pad2}, in interpreting mode.
;-
;- * operator's terminal facilities available
;-	same as supplied by OS (Linux console, etc), vt100/102 and ANS terminal sequences 
;-	maintained. 8-bit ascii, r/w.
;-
;- * program data space available, in address units
;-	256K bytes, configurable when lib4th compiling and/or with lib4th clones/jobs initiation. 
;-
;- * return stack space available, in cells
;-	32768.
;-
;- * stack space available, in cells
;-	32768, less acctually used LOCAL data-space.
;-
;- * system dictionary space required, in address units.
;-	4K bytes (one Linux memory page).
;-	
;-
;-----------------------------------------------------------------;
;- BLOCK wordset
;- 
;-	
;- 7.4.1.3 Other system documentation
;- 
;- * any restrictions a multiprogramming system places on the use of buffer addresses;
;-	None, buffer blocks assigned to the rsp 'owner's file descriptor.
;-	
;- * the number of blocks available for source text and data.
;-	Blocks constituted on demand, min. 4 buffers (4K) per screenfile-channel.
;-	additional blocks may be allocated up to available data-space, which is extendable 
;-	w/ {allocate} up to the limits set by host system.
;-	blocks removeable w/ {forget} of any word defined before the rsp block was constituted, 
;-	discarding any modified data and leaving the rsp. channels in their previous state.
;-
;-	
;-================================================================;
;-
;- Following applies to "calc4th", "4" and "f8" example applications:
;-	
;- 
;- 4.2 Program documentation
;- 
;- 4.2.1 Environmental dependencies
;-   
;- A program shall document the following environmental dependencies, where they apply, and 
;- should document other known environmental dependencies:
;-
;- * considering the pictured numeric output string buffer a fixed area with unchanging access 
;-   parameters (3.3.3.6 Other transient regions)
;-	No.
;-
;- * depending on the presence or absence of non-graphic characters in a received string 
;-   (6.1.0695 ACCEPT, 6.2.1390 EXPECT);
;-	re 4.1.1 Implementation-defined options,
;-	"character editing of 6.1.0695 ACCEPT and 6.2.1390 EXPECT"
;-
;- * relying on a particular rounding direction (3.2.2.1 Integer division)
;-	dft is round near zero
;-
;- * requiring a particular number representation and arithmetic 
;-   (3.2.1.1 Internal number representation)
;-	-ve numbers are regular 2-s complement values, which reversibly can be negated 
;-	("symmetric" range).
;-
;- * requiring non-standard words or techniques (3. Usage requirements)
;-	when compiling, re system documentation.
;-
;- * requiring the ability to send or receive control characters
;-   (3.1.2.2 Control characters, 6.1.1750 KEY)
;-	Linux Console.
;-
;- * using control characters to perform specific functions (6.1.1320 EMIT, 6.1.2310 TYPE)
;-	re 4.1.1, "character editing of 6.1.0695 ACCEPT and 6.2.1390 EXPECT"
;-
;- * using flags as arithmetic operands (3.1.3.1 Flags)
;-	No
;-
;- * using lower case for standard definition names or depending on the case sensitivity 
;-   of a system (3.3.1.2 Definition names)
;-	case-indepence expected, (most) words are lower case.
;-
;- * using the graphic character with a value of hex 24 (3.1.2.1 Graphic characters).
;-	"$" is expected code hex 24, but may depend on Linux Console.
;- 
;-
;-----------------------------------------------------------------;
;- BLOCK wordset
;-
;-  7.4.2 Program documentation
;-  
;-     * the number of blocks required by the program.
;-    
;-
;-----------------------------------------------------------------;
;- FILE wordset
;-
;-    
;-     * no additional requirements.
;-
;-----------------------------------------------------------------;
;- 
;- 4.2.2 Other program documentation
;- 
;- A program shall also document:
;- 
;- * minimum operator's terminal facilities required
;-	Linux Console (vt102 w. specific extensions)
;-
;- * whether a Standard System exists after the program is loaded.
;-	No.
;-
;- * non standardly operating standard words
;-	{leave}		execution continues until next {loop} - f.i.g. forth mode.
;-
;-
;-----------------------------------------------------------------;
;- BLOCK wordset
;-
;-     * no additional requirements.
;-
;-
;-----------------------------------------------------------------;
;- FILE wordset
;-
;-     * no additional requirements.

;-
;-
;-================================================================;
;- ansdoc.inc <eof>
