;-================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-
;-  file  	: debug.inc
;-  switch	: __DEBUG_INC
;-  created	: 15-jan-00
;-  modified	: 18-jan-01
;-  version	; 0.11
;-  assembler	: nasm 0.98-3e
;-  description	: debugging macros
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: no flags nor regs will be changed.
;-		: minimal inactive code cpl'd if switches unset,
;-		: debugging mode initially inactive.
;-		: macros's "_nolist" set to suppress macro expansion
;-  source	: -/-
;-  requires	: "asmutils" by K.Boldyshev,
;-		:	http://lightning.voshod.com:/asm
;-		: %xdefine required for several string operations,
;-		: supplied w. debugged & extended "nasm-098e"
;-		:	ftp://ftp.linuxgames.com/crystal/nasm/
;-================================================================;
;-
;-

%ifndef __DEBUG_INC
%define __DEBUG_INC
%define debuginc

%ifndef systeminc
%ifndef __SYSTEM_INC
%include "system.inc"
%define systeminc
%endif
%endif

%ifndef _nolist
%define _nolist
%endif


%ifndef jr ; lazy typing (relict from the z80) ("%ifndef" useless, maybe doubly defining)
%define jr jmp short
%endif

%ifndef regl
%xdefine regl
%endif

;================================================================
; debugging

;-
;- notation
;-	name 0-2
;-		macro <name> can take none or up to 2 parameters
;-	%1,..
;-		positional macro parameter
;-	%1(reg)
;-		shorter code if %1=reg
;-
;- most macros don't change any register and preserve processor
;- flags - re particular macro for particular specification.

;-
;-
;- : %assign-ed:

;-
;- i/o channels (file descriptors)
;-	stdout:=0 stdin:=1 stderr:=2
;-	[currently inactive! UNCOMMENT IF REQUIRED.
;-	[set in the forth asssembly to be read from a variable
;%ifndef stdin
;%assign stdout 0
;%assign stdin  1
;%assign stderr 2
;%endif

%ifndef STDIN
%assign STDOUT 0
%assign STDIN  1
%assign STDERR 2
%endif

%assign LF	__n
%assign TAB	__t
%assign BSP	__b
;-

;-
;- : debugging macros:
;-   don't compile anything if __DBG not %define-d

;-
;- PRSTG 1
;-	prints 4-bytes string %1 or @[ecx]
;-	regs & flags preserved
;-	requires macro <perrs>
%macro PRSTG 0-1_nolist [ecx]
%ifdef __DBG
%xdefine dregl regl
    push ecx
%ifnidni %1,ecx
    mov ecx,dword %{1}
%endif
    call prstg
    pop ecx
%xdefine regl dregl
%endif
%endmacro    

;-
;- PRNUM 1-3
;-	%1:num/val/ptr, %2:mode, %3:txt
;-	requires <perrn>, re below
;-	prints num %1(edx)
;-	%2 mode: l/1:8 dig's, b/2:byte w/3:word, s/else signed
;-	nasm macroes don't enable address mode detection, neither do 
;-	    those stupid "nidni"-s read the brackets around a register, 
;-	    thus reference into returnstack cannot be resolved other
;-	    than by passing an additional 4 bytes offset to this macro!
;-	    regs & flags preserved
%macro PRNUM 1-3_nolist l,':$'
%ifdef __DBG
    PRSTG %{3}
%xdefine dregl regl
    push edx
    mov edx,dword %{1}
%if %0 > 1
%define %%i %{2}
    %ifidni %2,b
%define %%i 2
    %elifidni %2,w
%define %%i 3
    %elifidni %2,s
%define %%i 0
    %else
%define %%i 1
    %endif
    %if %%i = 1
    call prerr_num	;8dig's
    %elif %%i = 2
    call prerr_n	;byte
    %elif %%i = 3
    call prerr_b	;word
    %else
    call prerr_val	;signed
    %endif
%else
    call prerr_num	;8dig's
%endif
    pop edx
%xdefine regl dregl
%endif
%endmacro

;-
;- DUMP 1-3
;-	dump 12 cells at %1(edx), optionally print text %2
;-	regs & flags preserved, ref to esp or RS taken into account
;-	requires macro mdump 0
%macro DUMP 1-3_nolist 'dmp:',b
; -- ;
%ifdef __DBG
%xdefine dregl regl
%if %0>1
    PRSTG %{2}
%endif
    push edx
%ifnidni %1,edx
    mov edx,%{1}
    %ifidni %{1},esp
    add edx,byte 4
    %endif
%endif;%1,edx
    call dumpm
    pop edx
%xdefine regl dregl
%endif;__DBG
%endmacro

;-
;- WKY 0-2
;-	wait until <enter> (code 10) to continue,
;-		optionally with DUMP if 2 parm-s passed
;-	regs & flags preserved
;-	locally using an auxilary channel to /dev/tty
;-	requires macro <perrs>
%macro WKY 0-2_nolist
%ifdef __DBG
%if %0 > 1
	DUMP %{1},%{2}
%endif
%xdefine dregl regl
	call wkey
%xdefine regl dregl
%endif
%endmacro
;-
;- PRTXT 1-2 dft: CRLF,2
;-	prints chars at [%1], till byte <nul> or 
;-		maxlen 130 or, optional length %2
;-	regs & flags preserved
;-	requires macro <perrs>
%macro PRTXT 0-2_nolist
%ifdef __DBG
%xdefine tregl regl
    push ecx
    push ebx
%ifnid %{2}
    mov ebx,130
%elifnidni %{2},ebx
    mov ebx,%{2}
%endif
%ifidni %{1},ebx
    mov ecx,[esp]
%elifnidni %{1},ecx
    lea ecx,[%{1}]
%endif
    call prerr_len
    pop ebx
    pop ecx
%xdefine regl tregl
%endif
%endmacro

;-
;- PRBIN 1 reg32/mem32
;-	inline code
;-	print %1 32 bits binary
;-	regs & flags preserved
;-	requires macro <pbits>
%macro PRBIN 1_nolist
%ifdef __DBG
%xdefine dregl regl
    push esi
%ifnidni %{1},esi
    mov esi,dword %{1}
%endif
    call bbits
    pop esi
%xdefine regl dregl
%endif
%endmacro

;-
;- PFBIN 0
;-	inline code
;-	print eflags 32 bits binary
;-	regs & flags preserved
;-	requires macro <pbits>
%macro PFBIN 0_nolist
%ifdef __DBG
%xdefine dregl regl
    call ebits
%xdefine regl dregl
%endif
%endmacro

;-
;- PFSYM 0
;-	inline code
;-	print eflags symbolic (if set)
;-	regs & flags preserved
;-	requires macro <psbits>
%macro PFSYM 0_nolist
%ifdef __DBG
%xdefine dregl regl
    call sbits
%xdefine regl dregl
%endif
%endmacro

;-
;-
;- : subroutines
;-   required for the above macros
;-   soubroutines compile just the label and <ret> if __DBG not %define-d
;-

;-
;- pbits 0, macro
;-	<ebits>, subroutine
;-		print eflags bits
;-		regs & flags preserved
;- 	<bbits> ,subroutine
;-		print esi binary
;-		regs & flags preserved
%macro pbits 0_nolist
    align 8
bbits:
%ifdef __DBG
    pushad
    pushfd
    jr ebits.b
    align 8
%endif
ebits:
%ifdef __DBG
    pushad
    pushfd
    mov esi,[esp]
.b:
    push byte 32
    pop edi
    push byte 1
    lea ecx,[esp]
    pop edx
%%l:
    push byte('0'/2)
    shl esi,1
    rcl byte[esp],1
%%w:
    sys_write STDOUT
    pop eax
    dec di
    jg %%l
    push byte' '
    jns %%w
    pop eax
    popfd
    popad
%endif
    ret
%endmacro

;-
;- psbits 0, macro
;- 	<sbits> ,subroutine
;-	print eflag bits symbols (if set)
;-	regs & flags preserved
%macro psbits 0_nolist
    align 8
sbits:
%ifdef __DBG
    pushad
    pushfd
    mov esi,[esp]
    push byte 32
    pop edi
    push byte 2
    pop edx
    push dword'    '
    lea ecx,[esp]
%%l:
    shl esi,1
    jnc %%n
    cmp di,byte 22
    ja %%n
    cmp di,byte 10
    seta al
    or dl,al
    mov ax,[sfchrs-2+edi*2]
    test al,al
    jz %%n
    mov word[esp],ax
    sys_write STDOUT
%%n:
    dec di
    jnz %%l
    pop edi
    popfd
    popad
    ret
sfchrs:
    db 'C ',0,0,'P ',0,0,'A ',0,0,'Z S T I D O IOioNT',0,0,'RFVMACVFVPID'
;	0    1   2    3   4    5   6 7 8 9 0 1 2 3 4    5   6 7 8 9 0 1
;	0				   1			    2
;	22 3 4 5 6 7 8 9 0 31
;    dw  0,0,0,0,0,0,0,0,0,0
%else 
    ret
%endif
%endmacro

;-
;- perrm 0
;-	subroutine, entry at <pmsg>
;-	print error message accordinding to -ve num in eax
;-	print +ve ernum and ovf-msg if out of range
;-	print +ve ernum if no message table presen ("errno.inc")
;-	noop if eax +ve, special msg for eax=0x80000000
;-	regs & flags preserved
;-	text output requires "errno.inc" included
%macro perrm 0_nolist
    align 8
pmsg:
%ifdef __DBG
	pushfd
	pushad
	neg eax
	jng %%p
%ifdef __I386_ERRNO
	mov ecx,dword[rm]
	cmp eax,ecx		; upper bound of errno-s
	jna %%a 		; within range
	mov edx,eax
	shl eax,1
	jz %%a
	mov eax,ecx
	call prerr_num		; print -ve ernum
%%a:	lea ecx,[rn]		; base of disp table
	movsx eax,word[byte ecx+eax*2-2]; msg-rn
	add ecx,eax
	movsx edx,word[ecx]
	inc ecx			; advance
	inc ecx			;   to leading <nl>
	test edx,edx
	jng %%p			;? zero or -ve length (which shouldn't happen)
	sys_write STDERR
%else
	PRNUM eax		; just +ve ernum if no message table present
%endif
%%p:	popad
	popfd
%endif
	ret
%endmacro

;-
;- perrs 0
;-	subroutines, macro <dbcons> required
;-	regs & flags preserved
;-	<prerr_len>	prints <nul>-term'd stg from (ecx)+, up to <ebx> chars
;-	<prerr_string>	prints <nul>-term'd stg from (ecx)+, up to 130 chars
;-	<wkey>		wait until <enter> - code 10 - to continue,
;-			independently of stdin or other regular input source.
;-	<gkey>		<wkey>, any key-code (if console permits), no " :> ".
;-	<prstg>		prints 4-bytes string = reg ecx
%macro perrs 0_nolist
    align 8
prstg:
%ifdef __DBG
    push byte 0
    push ecx
    mov ecx,esp
    call prerr_string
    pop ecx
    lea esp,[esp+4]	; flags!
    ret
    align 8
prerr_string:
    pushfd
    pushad
    mov ebx,130		;maxlen
    jr %%p

    align 8
prerr_len:
    pushfd
    pushad
%%p:
    xor edx,edx
    push ebx		; maxcount
    push edx		; aux buffer
    push ecx		; stringptr
%%l:
    dec dword[esp+8]
    js %%x
    mov eax,[esp]
    movzx eax,byte[eax] ;char
    test al,al
    jz %%x		;? ready
    mov al,[ctra+eax]	; table + char.disp
    lea ecx,[esp+4]
    inc dword [esp]
    mov [ecx],al
    sys_write STDERR,ecx,1
    jr %%l

    align 8
%%x:
    add esp,byte 12
    popad
    popfd
    ret
    align 8
gkey:
    pushfd
    pushad
    mov dx,-1		; flag 'any key'
    jr %%i
    align 8
wkey:
    pushfd
    pushad
    PRSTG ' :>'
    mov dx,10		; flag 'enter'
%%i:
    movsx edx,dx
    push edx
    sys_open DTTY,O_RDONLY; |O_NOCTTY ; aux cons channel ;
    pop edx
    mov ebx,eax
%%r:
    xor eax,eax
    push edx		;-> char.code
    push eax		;-> buffer
    mov ecx,esp		; addr
    lea edx,[eax+1]	; length
    sys_read
;; displaying the (eax) corresponding error message,
;; was used to debugging this routine
;; requires subroutine <pmsg> from macro perrm
;;	PRNUM eax
;;	call pmsg	`exec format error' if not an input channel
    pop ecx
    pop edx
    cmp cl,dl
    jz %%c
    test edx,edx
    jns %%r
    test eax,eax
    js %%r
%%c:
    sys_close		; close aux chan
    popad
    popfd
%else
prerr_len:
prerr_string:
gkey:
wkey:
%endif
    ret
%endmacro

;-
;- mdump 0-1
;-	subroutine
;-	<dumpm>	dump 9 or %{1} cells & chars at cell aligned addr edx
;-	regs & flags preserved
;-	(modified to bytewise display
%macro mdump 0-1_nolist 8
    align 8
dumpm:
%ifdef __DBG
%assign dumpnum %{1}
%xdefine mregl regl
    call prerr_num
    PRSTG 10<<16|'@:'
    pushfd
    pushad
    xor ecx,ecx
    lea ebx,[ecx-4]
    mov cl,dumpnum*4	;*4 (+)
    and ebx,edx
    mov eax,esp		;
    sub esp,dumpnum*4+4
    mov ebp,esp
    push eax		;
    mov byte[ebp],__n
%%d:
    test cl,3
    jnz %%s
    PRSTG' '
%%s:
    inc ebp		;
    movzx edx,byte[ebx]	;mov edx,[ebx]
    mov [ebp],dl	;edx
    call prerr_n	;_num
    inc ebx		;add ebx,byte 4
    dec cl
    jnz %%d
    lea ebp,[esp+4]	;
    mov dl,dumpnum*4+1	;*4 (+)
%%c:
    movzx ecx,byte[ebp]	;pop ecx
    inc ebp		;
    call prstg
    dec dl
    jnz %%c
    PRSTG __n
    mov esp,[esp]
    popad
    popfd
%xdefine regl mregl
%endif
    ret
%endmacro

;-
;- perrn 0
;-	subroutines, printing <edx> sedecimal to stderr
;-	<prerr_num>	8 digits, unsigned
;-	<prerr_b>	l.s. 4 digits,unsigned
;-	<prerr_n>	l.s. 2 digits,unsigned
;-	<prerr_val>	significant digits, signed
;-	regs & flags preserved
%macro perrn 0_nolist
    align 8
prerr_n:
%ifdef __DBG
    pushfd
    pushad
    xor eax,eax
    xor ecx,ecx
    mov cx,0x8001 	; a byte
    jr %%n
    align 8
prerr_b:
    pushfd
    pushad
    xor eax,eax
    xor ecx,ecx
    mov cx,0x8003 	; half cell
    jr %%n

    align 8
prerr_val:		; print significant digits
    pushfd
    pushad
    xor ecx,ecx
    xor eax,eax
    or eax,edx		; print signed, (signbit preserved for result)
    jns %%v		; +ve
    neg edx		; -(abs)
    jr %%v

    align 8
prerr_num:		; print 8 digits
    pushfd
    pushad
    xor eax,eax
    xor ecx,ecx
    mov ch,0x80
%%v:
    mov cl,7		; no. of digits (nibbles)
%%n:
    mov edi,esp
    push byte 1 	; count + trailing space
    lea esi,[esp]
    push byte 0
    lea ebx,[esp-1]
    sub esp,12
    mov byte[ebx],' '
%%p:
    inc dword[esi]
    shrd ax,dx,4
    shr ax,12
    inc ch
    shr edx,4
    jnz %%z		;? significant digit
    test ch,ch
    js %%z		;? fixed 8 digits size
    xor cl,cl		; insert "0" and exit
%%z:
    add al,0x90
    daa
    adc al,0x40
    daa
    dec ebx
    mov [ebx],al
    dec cl
    jns %%p
    mov ecx,[esi]
    test eax,eax
    jns %%s		; unsigned or +ve
    sub ebx,byte 2
    add ecx,byte 2
    mov word[ebx],'$-'
%%s:
    sys_write STDERR,ebx,ecx
    mov esp,edi		; drop loc. vari
    popad
    popfd
%else
prerr_b:
prerr_val:
prerr_num:
%endif
    ret
%endmacro

;-
;- p_num, subroutine
;-	print 64bit decimal number {edx:eax} to stdout
;-	all regs preserved
;-	<p_byt> l.s.byte, unsigned
;-	<p_num> qword, signed
%macro pnum 0_nolist
    align 8
%ifdef __DBG
p_byt:
    pushad
    pushfd
    movzx eax,al
    xor edx,edx
    jr %%b

    align 8
p_num:
    pushad
    pushfd
%%b:
    _mov ebx,10		; radix
    _mov ecx,1<<16|0	; trailing char count & char (words)
;entrypoint:
; eax:=num.lo.dw ebx:=radix ecx:hi.w=count,lo.w=trailer edx:num.hi.dw
%%n:
    test edx,-1
    jge %%p
    pushad
    push byte '-'
    lea ecx,[esp]
    _mov edx,1
    sys_write STDOUT
    popad
    neg eax
    sbb edx,0
    neg edx
%%p:
    mov ebp,esp
    sub esp,byte 38	; maxlen for binary, trailer, leading radix marker
    push ebp
    mov word[ebp-2],cx
    shr ecx,16
    push ecx		; preload length counter
    sub ebp,ecx
    mov ecx,ebp
%%l:
    dec ecx		; numbuf-
    inc dword[esp]
    xor edx,edx		; hi dword for div
    div ebx
    or dl,'0'
    mov [ecx],dl
    test eax,eax
    jnz %%l
    pop edx		; count
    sys_write STDOUT
    mov esp,[esp]
    popfd
    popad
%else
p_byt:
p_num:
%endif
    ret
%endm

;-
;- dbmrg 0
;-	subroutine, entry at <prmrg>
;-	prints left margin blanks
;-	regs & flags preserved
%macro dbmrg 0_nolist
    align 8
prmrg:
%ifdef __DBG
%ifdef __F8E
    pushfd
    pushad
    xor edx,edx
    or dl,[VMRG]
    jz %%m
    js %%a
    cmp dl,63
    jle %%c
%%a xor edx,edx
    mov [VMRG],edx
%%c:
    sys_write STDERR,SPCS
%%m:
    popad
    popfd
%endif
%endif
    ret
%endmacro
;-
;-
;- : data used in the debugging routines:
;-

;-
;- dbcons 0
;-	cons/vari for the debugging macroes
;-	__DBG:
;-		macro enabled only if switch '%define'd, else empty.
;-	__DBCONS:
;-		if defined declaring 'extern'al references, macro disabled;
;-		else compiling respective data and 'global' labels:
;-	__F8E undefined
;-		SPCS,BLSP,CRLF,BSPC,BSCB,BSBK,DTTY,ctrad,ctrat,ctra,spce
%ifndef __DBCONS
%define __DBCONS
%macro dbcons 0_nolist
%ifdef __DBG
%ifdef __MK_DBG
    global SPCS:data,BLSP:data,spce:data,CRLF:data,BSPC:data,BSCB:data,BSBK:data,DTTY:data,ctrad:data,ctrat:data,ctra:data
    align 4
%endif
SPCS:	times 63 db ' '	; one line of (64) blanks
BLSP:	db " "
spce:	db 0		; e.g. 13 blanks => <DBG spce-13>
CRLF:	db LF,0
BSPC:	db BSP,0
BSCB:	db BSP,"} ",0
BSBK:	db BSP,"] ",0
DTTY:	db '/dev/tty',0
     align 4
ctrad:	; code bytes debugging
ctrat:	; text & ctrl-s output
ctra:	; debugging simplified
	db 0,'.......',8,'.',10,'..',13,'..................'
	db 32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63
	db 64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95
	db 96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120
	db 121,122,123,124,125,126,127 ;;'.'
    times 32 dd '....'
;	db 128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151
;	db 152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175
;	db 176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199
;	db 200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223
;	db 224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247
;	db 248,249,250,251,252,253,254,255
    align 4
%endif;__DBG
%endm

%else;__DBCONS

%ifdef __DBG
%ifdef __MK_DBG
    extern SPCS,BLSP,spce,CRLF,BSPC,BSCB,BSBK,DTTY,ctrad,ctrat,ctra
%endif
%endif

%endif;__DBCONS
;
%endif	;__DEBUG_INC
;-
;-
;-================================================================
;- debug.inc <eof>
