;-================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;- Read ./doc/LEGAL file for full copyright information (GNU GPL)
;-
;-  file	: resource.h.inc
;-  switch	: _LINUX_RESOURCE_H, _I386_RESOURCE_H
;-  created	: 15-mai-2000
;-  modified	: 17-mai-2000
;-  version	: 0.02
;-  assembler	: nasm 0.98-3e
;-  description	: Resource limits/control/accounting
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: linux/include/:linux/resource.h, asm/resource.h
;-  requires	: types.inc from "asmutils" by K.Boldyshev,
;-			http://lightning.voshod.com:/asm
;-================================================================;
;-
;-
%ifndef _LINUX_RESOURCE_H
%define _LINUX_RESOURCE_H

%include "time.h.inc"
;-
;- Resource control/accounting header file for linux

;- Definition of struct rusage taken from BSD 4.3 Reno
;- We don't support all of these yet, but we might as well have them....
;- Otherwise, each time we add new items, programs which depend on this
;- structure will lose.  This reduces the chances of that happening.

%define	RUSAGE_SELF	0
%define	RUSAGE_CHILDREN	(-1)
%define RUSAGE_BOTH	(-2)	; sys_wait4() uses this 

	struc	rusage
.ru_utime:	ELF_ISTRUC timeval; user time used 
.ru_stime:	ELF_ISTRUC timeval; system time used 
.ru_maxrss:	long	; maximum resident set size 
.ru_ixrss:	long	; integral shared memory size 
.ru_idrss:	long	; integral unshared data size 
.ru_isrss:	long	; integral unshared stack size 
.ru_minflt:	long	; page reclaims 
.ru_majflt:	long	; page faults 
.ru_nswap:	long	; swaps 
.ru_inblock:	long	; block input operations 
.ru_oublock:	long	; block output operations 
.ru_msgsnd:	long	; messages sent 
.ru_msgrcv:	long	; messages received 
.ru_nsignals:	long	; signals received
.ru_nvcsw:	long	; voluntary context switches
.ru_nivcsw:	long	; involuntary "
	endstruc

	struc rlimit
.rlim_cur:	long
.rlim_max:	long
	endstruc

%define	PRIO_MIN	(-20)
%define	PRIO_MAX	20

%define	PRIO_PROCESS	0
%define	PRIO_PGRP	1
%define	PRIO_USER	2

;- Due to binary compatibility, the actual resource numbers
;- may be different for different linux versions..
;%include "asm+resource.h.inc"
%ifndef _I386_RESOURCE_H
%define _I386_RESOURCE_H

;- Resource limits

%define RLIMIT_CPU	0		; CPU time in ms
%define RLIMIT_FSIZE	1		; Maximum filesize
%define RLIMIT_DATA	2		; max data size
%define RLIMIT_STACK	3		; max stack size
%define RLIMIT_CORE	4		; max core file size
%define RLIMIT_RSS	5		; max resident set size
%define RLIMIT_NPROC	6		; max number of processes
%define RLIMIT_NOFILE	7		; max number of open files
%define RLIMIT_MEMLOCK	8		; max locked-in-memory address space
%define RLIMIT_AS	9		; address space limit
%define RLIMIT_LOCKS	10		; max file locks held

%define RLIM_NLIMITS	11

%define RLIM_INFINITY	(~0)

;%ifdef __KERNEL__
%define INIT_RLIMITS {{RLIM_INFINITY,RLIM_INFINITY},{RLIM_INFINITY,RLIM_INFINITY},{RLIM_INFINITY,RLIM_INFINITY},{_STK_LIM,RLIM_INFINITY},{0,RLIM_INFINITY},{RLIM_INFINITY,RLIM_INFINITY},{0,0},{INR_OPEN,INR_OPEN},{RLIM_INFINITY,RLIM_INFINITY},{RLIM_INFINITY,RLIM_INFINITY},{RLIM_INFINITY,RLIM_INFINITY},}
;%endif ; __KERNEL__

%endif; _I386_RESOURCE_H

%endif; _LINUX_RESOURCE_H
;-
;-================================================================;
;- resource.h.inc <eof>
