;-================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;- Read ./doc/LEGAL file for full copyright information (GNU GPL)
;-
;-  file	: time.h.inc
;-  switch	: _LINUX_TIME_H
;-  created	: 25-aug-2000
;-  modified	: 
;-  version	: 0.02
;-  assembler	: nasm 0.98-3e
;-  description	: definitions for linux terminal control
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	:
;-  source	: linux 2.2.14, linux/time.h
;-  requires	: types.inc from "asmutils" by K.Boldyshev
;-================================================================;
;-
;-
%ifndef _LINUX_TIME_H
%define _LINUX_TIME_H
%ifndef _nolist
%define _nolist
%endif

;%include <asm/param.h>
;%include <linux/types.h>

%ifndef _STRUCT_TIMESPEC
%define _STRUCT_TIMESPEC

%ifnid timespec
%define time_t	LONG 1	; posix_types.h
 struc timespec 
.tv_sec:	time_t	; seconds */
.tv_nsec:	LONG 1	; nanoseconds */
 endstruc
%endif
%endif ; _STRUCT_TIMESPEC */

; Change timeval to jiffies, trying to avoid overflows.
; Note that we don't want to return MAX_LONG, because for various timeout
; reasons we often end up having to wait "jiffies+1" in order to guarantee
; that we wait at _least_ "jiffies" - so "jiffies+1" had better still be positive.
%define MAX_JIFFY_OFFSET ((~0>>1)-1)

%ifnid timespec_to_jiffies
;static __inline__ unsigned long
%macro timespec_to_jiffies 1_nolist;(struct timespec *value)
    push ecx,edx
    mov eax,[%{1}-timespec+timespec.tv_nsec]
    mov ecx,1000000000/HZ
    add eax,ecx
    xor edx,edx
    dec eax
    div ecx
    add eax,[%{1}-timespec+timespec.tv_sec]
    adc edx,0
    jnz %%o
    cmp eax,MAX_JIFFY_OFFSET
    jc %%r
%%o:
    mov eax,MAX_JIFFY_OFFSET
%%r:
    pop ecx,edx
%endm
%endif

%ifnid jiffies_to_timespec
%macro jiffies_to_timespec 2_nolist;(unsigned long jiffies, struct timespec *value)
    push ecx,edx
    mov eax,%{1}
    mov ecx,HZ
    xor edx,edx
    div ecx
    mov [%{2}-timespec+timespec.tv_sec],eax; jiffies / HZ
    mov eax,edx
    mov ecx,1000000000/HZ
    xor edx,edx
    mul ecx
    mov [%{2}-timespec+timespec.tv_nsec],eax;(jiffies % HZ) * (1000000000L / HZ)
    pop ecx,edx
%endm
%endif
%ifnid timeval
%define suseconds_t	LONG 1
 struc timeval
.tv_sec:time_t		; seconds
.tvusec:suseconds_t	; microseconds
 endstruc
%endif
%ifnid timezone
 struc timezone
.tz_minuteswest:INT 1	; minutes west of Greenwich */
.tz_dsttime:	INT 1	; type of dst correction */
 endstruc
%endif

%define NFDBITS	__NFDBITS

%ifdef __KERNEL__
;extern void do_gettimeofday(struct timeval *tv);
;extern void do_settimeofday(struct timeval *tv);
;extern void get_fast_time(struct timeval *tv);
;extern void (*do_get_fast_time)(struct timeval *);
%endif

%define FD_SETSIZE		__FD_SETSIZE
%define FD_SET(fd,fdsetp)	__FD_SET(fd,fdsetp)
%define FD_CLR(fd,fdsetp)	__FD_CLR(fd,fdsetp)
%define FD_ISSET(fd,fdsetp)	__FD_ISSET(fd,fdsetp)
%define FD_ZERO(fdsetp)		__FD_ZERO(fdsetp)

; Names of the interval timers, and structure defining a timer setting.
%define	ITIMER_REAL	0
%define	ITIMER_VIRTUAL	1
%define	ITIMER_PROF	2

%ifnid itimerspec
 struc  itimerspec
.it_interval:			; timer period
.it_interval.tv_sec:	time_t	; seconds
.it_interval.tv_nsec:	LONG 1	; nanoseconds
.it_value:			; timer expiration
.it_value.tv_sec:	time_t	; seconds
.it_value.tv_nsec:	LONG 1	; nanoseconds
 endstruc
%endif
%ifnid itimerval
 struc	itimerval
.it_interval:			; timer interval
.it_interval.tv_sec:	time_t	; seconds
.it_interval.tv_nsec:	LONG 1	; nanoseconds
.it_value:			; current value
.it_value.tv_sec:	time_t	; seconds
.it_value.tv_nsec:	LONG 1	; nanoseconds
 endstruc
%endif
%endif
;-
;-=========================================================================;
;- time.h.inc <eof>
