;-================================================================;
;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;- Read ./doc/LEGAL file for full copyright information (GNU GPL)
;-
;-  file	: vt.inc
;-  switch	: _LINUX_VT_H
;-  created	: 15-jan-2000
;-  modified	: 17-jan-2000
;-  version	: 0.02
;-  assembler	: nasm 0.98-3e
;-  description	: definitions for linux terminal control
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	:
;-  source	: linux 2.2.14, linux/vt.h
;-  requires	: types.inc from "asmutils" by K.Boldyshev
;-================================================================;
;-
;-

%ifndef _LINUX_VT_H
%define _LINUX_VT_H
%define vtinc

;-
;- ioctl-s
;-	0x56 is 'V', to avoid collision with termios and kd
;- VT_OPENQRY	VT_GETMODE	VT_SETMODE	VT_AUTO		
;- VT_PROCESS	VT_ACKACQ	VT_GETSTATE	VT_SENDSIG	
;- VT_RELDISP	VT_ACTIVATE	VT_WAITACTIVE	VT_DISALLOCATE	
;- VT_RESIZE	VT_RESIZEX      VT_LOCKSWITCH   VT_UNLOCKSWITCH 
;-
%define VT_OPENQRY	0x5600	; find available vt
;-
;- struct-s:
;-	vt_mode, vt_stat, vt_sizes, vt_consize
%ifnid vt_mode
    struc vt_mode
.mode		CHAR	1	; vt mode
.waitv		SHORT	1	; if set, hang on writes if not active
.relsig		SHORT	1	; signal to raise on release req
.acqsig		SHORT	1	; signal to raise on acquisition
.frsig		SHORT	1	; unused (set to 0)
    endstruc
%endif
%define VT_GETMODE	0x5601	; get mode of active vt
%define VT_SETMODE	0x5602	; set mode of active vt
%define	VT_AUTO		0x00	; auto vt switching
%define	VT_PROCESS	0x01	; process controls switching
%define	VT_ACKACQ	0x02	; acknowledge switch
%ifnid vt_stat
    struc vt_stat
.v_active	SHORT	1	; active vt
.v_signal	SHORT	1	; signal to send
.v_state	SHORT	1	; vt bitmask
    endstruc
%endif
    
%define VT_GETSTATE	0x5603	; get global vt state info
%define VT_SENDSIG	0x5604	; signal to send to bitmask of vts

%define VT_RELDISP	0x5605	; release display

%define VT_ACTIVATE	0x5606	; make vt active
%define VT_WAITACTIVE	0x5607	; wait for vt active
%define VT_DISALLOCATE	0x5608  ; free memory associated to vt

%ifnid vt_sizes
    struc vt_sizes
.v_rows		SHORT	1	; number of rows
.v_cols		SHORT	1	; number of columns
.v_scrollsize	SHORT	1	; number of lines of scrollback
    endstruc
%endif

%define VT_RESIZE	0x5609	; set kernel's idea of screensize

%ifnid vt_consize
    struc vt_consize
.v_rows		SHORT	1	; number of rows
.v_cols		SHORT	1	; number of columns
.v_vlin		SHORT	1	; number of pixel rows on screen
.v_clin		SHORT	1	; number of pixel rows per character
.v_vcol		SHORT	1	; number of pixel columns on screen
.v_ccol		SHORT	1	; number of pixel columns per character
    endstruc
%endif

%define VT_RESIZEX      0x560A  ; set kernel's idea of screensize + more
%define VT_LOCKSWITCH   0x560B  ; disallow vt switching
%define VT_UNLOCKSWITCH 0x560C  ; allow vt switching

%endif ;_LINUX_VT_H
;- vt.inc <eof>
