;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4alogic.inc
;   created	: 04-aug-01
;   modified	: -> re "version.inc"
;   version	: 0.01(+)
;   switches	: __ALOGIC_I
;   assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;   description	: "lib4th" shared library Forth kernel
;   author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;   source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __ALOGIC_I
%define __ALOGIC_I
%define alogicinc

;-
;-================================================================
;---  arithmetic logic opr
;-	

;-
;- <>		( u1 u2 -- flg )		(a:cx)		'nequal'	ans
;-	tf if u1 not equal to u2
;-
;- =/=		( u1 u2 -- flg )		(m,11)		'nequal'	forth
;-	tf if u1 not equal to u2
nequal:GA M;:ok:
; c: -/-
	sub eax,TOS(1)
	TOSPULL(1)
	jz .r
	mov eax,-1
.r:
M nequal
	ret
;-
;- 0<>		( u1 -- flg )			(a)		'zneq'		ans
;-
;- 0=/=		( u1 -- flg )			(m,9)		'zneq'		forth
;-	tf if u1 non-zero
zneq:GA M;:ok:
; c: -/-
	test eax,eax
	jz .r
	mov eax,-1
.r:
M zneq
	ret
;slow! (above is twice as fast)
;    neg eax
;    sbb eax,eax
;    not eax

;-
;- -0=		( u1 -- flg )			(m,11)		'mzeq'		forth
;-	tf if only sign bit (m.s.bit) set of u1
mzeq:GA M;:ok:
	shl eax,1
	mov eax,0
	jnz .r
	sbb eax,eax
.r:
M mzeq
	ret
;-
;- true=  	( flg -- flg )			(m,9)		'tequ'		fig
;-	test f.i.g. true flag = 1
ftequ:GA M
	cmp eax,1
	jz .r
	mov eax,0
.r:
M tequ
	ret
;-
;- true=  	( flg -- flg )			(m,9)		'tequ'		forth
;-	test ans-4th true flag = -1
tequ:GA M
	cmp eax,-1
	jz .r
;-
;- 0and		( n -- 0 )			(m,5)		'dng'		hidden
;-	{ drop 0 }
dng:GA M
	mov eax,0
M dng
tequ.r:
M tequ
	ret
;-
;- =		( u1 u2 -- flg )		(m,12)		'equal'		forth
;-	tf if u1 = u2
equal:GA M;:ok:
	sub eax,TOS(1)
	TOSPULL(1)
	not eax
	jz .r
	mov eax,0
.r:
M equal
	ret
;-
;- 0=		( u1 -- flg )			(m,8)		'zeq'		forth
;-	tf if u1 =  0
zeq:GA M;:ok:
	cmp eax,byte 1
	sbb eax,eax
M zeq
	ret
;-
;- [0=]		( u1 -- flg )			(i)		'zeq'		ans
;-	"immediate"ly execute {0=}
izeq:equ zeq
;-
;- 0>		( u1 -- flg )			(m,5)		'zgreat'	forth
;-	tf if u1 +ve and non-zero
zgreat:GA M;:ok:
	neg eax
	sar eax,31
M zgreat
	ret
;-
;- 0<		( u1 -- flg )			(m,3)		'zless'		forth
;-	tf if u1 -ve
zless:GA M;:ok:
	sar eax,31
M zless
	ret
;-
;- <		( n1 n2 -- flg )		(m,15)		'less'		forth
;-	tf if signed n1 < n2
less:GA M;:ok:
	cmp eax,TOS(1)
	TOSPULL(1)
	mov eax,-1
	jg .r
	inc eax
.r:
M less
	ret
;-
;- within? 	( x1 x2 x3 -- x4 f )						forth
;-	x4 := x2 and tf if  x2 <= x1 < x3, whether signed or unsigned,
;-	if out of range ret x4 signed adjusted to the nearest bound and ff.
withinq:GA;:ok:
	call uwiq
	jz near enumn.f	;?; tf if in range
	call uwiq.v
	jg .r		;?; upper bound
.n:
	mov eax,edx	; lower bound
.r:
	mov TOS(1),eax
	_mov eax,0
	ret
;-
;- uwithin? 	( x1 x2 x3 -- x4 f )						forth
;-      x2 <= x1 < x3, signed or unsigned, x4:=x1 anf tf if in range
;-	if out of range ret x4 unsigned adjusted to the nearest bound.
uwithinq:GA;:ok:
	call uwiq
	jz near enumn.f	;?; tf
	call uwiq.v
	ja withinq.r	;?; lower bound
	jr withinq.n	; upper bound
;
uwiq:A
	mov edx,TOS(2)
	call within	; @PS:=x1-x2 ecx:=x3-x2
	inc eax
	mov eax,edx	; x1
	spush
	ret
; x1 = eax ; x2 = eax-[PS] ; x3 = [PS] ; ret eax:=x3 upper bound
.v:A
	mov edx,[PS]	; x2
	add ecx,edx	; x3
	cmp eax,edx	; <x1-x2>
	mov eax,ecx	; x3
	ret
;-
;- within 	( x1 x2 x3 -- flg )		(a:c)				forth
;-		( test low high -- flag )
;-	tf if x2 <= x1 < x3, signed or unsigned. - ANS-word
within:GA;:ok:
	mov ecx,eax
	spop
	sub ecx,eax	; n3-n2
	sub TOS(1),eax	; n1-n2
	mov eax,ecx	; n3-n2
;-
;- u<		( u1 u2 -- flg )		(m,13)		'uless'		forth
;-	tf if unsigned u1 < u2
uless:GA M;:ok:
; c: ecx
	cmp TOS(1),eax
	TOSPULL(1)
	mov eax,-1
	jc .r
	inc eax
.r:
M uless
	ret
;-
;- >		( n1 n2 -- flg )		(m,17)		'greater'	forth
;-	tf if signed n1 < n2
greater:GA M;:ok:
	cmp eax,TOS(1)
	TOSPULL(1)
	mov eax,-1
	jl .r
	inc eax
.r:
M greater
	ret
;-
;- u>		( u1 u2 -- flg )		(m,13)		'ugreater'	forth
;-	tf if unsigned u1 > u2
ugreater:GA M;:ok:
	cmp TOS(1),eax
	TOSPULL(1)
	mov eax,-1
	ja .r
	inc eax
.r:
M ugreater
	ret
;-
;- sgn		( n1 -- -1|0|1 )		(m,11)				forth
;-	signum of n1
sgn:GA M;:ok:
	test eax,eax
	jz .r
	sar eax,31
	lea eax,[eax*2+1]
.r:
M sgn
	ret
;-
;- parity  	( n1 -- flg )			(m,12)				forth
;-	ff if n1 w/ odd number of set bits
parity:GA M
	mov ecx,0
	or eax,ecx
	jnp .r
	dec ecx
.r:
	mov eax,ecx
M parity
	ret
;-
;- l-and  	( n1 n2 -- flg )				'land'		forth
;-	logic AND, tf only if n1 =/= 0 and n2 =/= 0.
land:GA M;:ok:
; c: -/-
	test eax,eax
	spop
	jz .z
	test eax,eax
.z:
	mov eax,0
	jz .r
	dec eax
.r:
M land
	ret
;-
;- l-or   	( n1 n2 -- flg )				'lor'		forth
;-	logic OR, ff only if n1 = 0 and n2 = 0.
lor:GA M;:ok:
; c: -/-
	or eax,TOS(1)
	mov eax,-1
	TOSPULL(1)
	jnz .r
	inc eax
.r:
M lor
	ret
;-
;--- 
;-

;-
;- ?udcmp	( d1 d2 -- d1 d2 f )		(e)		'qudcmp'	forth
;-	flag f from unsigned comparison, f=1 if d2 u< d1, f=-1 if d1 u< d2.
;-	{ d1 d2 -- ) 2over 2over du> if 1 else 2over 2over d= 0= endif ( -- d1 d2 f }
qudcmp:GA
	mov ecx,eax
	mov edx,TOS(1)
	sub ecx,TOS(4)	; m.s. sign
	spush
	mov eax,-1
	jnz .c
	sub edx,TOS(6)
	jz .p
	mov ecx,edx	; signed comparison
.c:
	jc .r
	inc eax
.p:
	inc eax
.r:
	ret
;-
;- dsgn		( d1 -- -1|0|1 )						forth
;-	signum of d1
dsgn:GA
	test eax,eax
	mov ecx,eax
	spop
	jnz .c
	mov ecx,eax
	test eax,eax
	jz .r
.c:
	sar ecx,31
	lea eax,[ecx*2+1]
.r:
	ret
;-
;- d>		( d1 d2 -- flg )		(m,17)(e)	'dgreat'	forth
;-	tf if signed d1 greater than d2
;-	(e: NZ*S if d1>d2, else Z*NS, flg is S from d2-d1
dgreat:GA M;:ok:
	mov edx,TOS(1)	;2l	; 2l
	mov ecx,TOS(2)	;1h	; 1h
	cmp edx,TOS(3)	;2l-1l	; 1l-2l
	sbb eax,ecx	;2h-1h	; 1h-2h
	TOSPULL(3)
	sar eax,31
M dgreat
	ret
;-
;- d<		( d1 d2 -- flg )		(a:d)(e)	'dless'		forth
;-	tf if signed d1 smaller than d2
;-	(e: NZ*S if d1<d2, else Z*NS, flg is S from d1-d2
dless:GA;:ok:
	mov ecx,TOS(3)	; 1l
	mov edx,TOS(2)	; 1h
	cmp ecx,TOS(1)	; 1l-2l
	sbb edx,eax	; 1h-2h
	TOSPULL(3)
	mov eax,edx
	sar eax,31
	ret
;-
;- du>		( d1 d2 -- flg )		(m,14)(e)	'dugreat'	forth
;-	tf if unsigned d1 greater than d2
;-	(e: NZ*S if d1 u> d2, else Z*NS, flg is C from d2-d1
dugreat:GA M;:ok:
	mov edx,TOS(1)	; 2l
	mov ecx,TOS(2)	; 1h
	cmp edx,TOS(3)	; 2l-1l
	sbb eax,ecx	; 2h-1h
	TOSPULL(3)
	sbb eax,eax
M dugreat
	ret
;-
;- du<		( d1 d2 -- flg )		(a:dx)(m,14)(e)	'duless'	forth
;-	tf if unsigned d1 smaller than d2
;-	(e: NZ*S if d1 u< d2, else Z*NS, flg is C from d1-d2
duless:GA M;:ok:
	mov ecx,TOS(3)	; 1l
	mov edx,TOS(2)	; 1h
	cmp ecx,TOS(1)	; 1l-2l
	sbb edx,eax	; 1h-2h
	TOSPULL(3)
	sbb eax,eax
M duless
	ret
;-
;- d0>		( d1 -- flg )			(e)		'dzgt'		forth
;-	tf if d1 non-zero +ve
;-	(e: eflags from <test m.s.cell>
dzgt:GA;:ok:
	test eax,eax
	mov eax,0
	js .r
	jnz .g
	cmp eax,TOS(1);
	jz .r
.g:
	not eax
.r:
	TOSPULL(1)
	ret	
;-
;- d0<		( d1 -- flg )			(a:d)(m,6)(e)	'dzless'	forth
;-	tf if d1 -ve
;-	(e: NZ*S if d1 < 0, else NS+Z
dzless:GA M;:ok:
	TOSPULL(1)
	sar eax,31
M dzless
	ret
;-
;- ?d0=		( d1 -- d1 flg )		(e)		'qdzeq'		forth
;-	leave d1 unchanged, ret tf if d1 equal to zero
;-	(e: Z if d1 = 0
qdzeq:GA
	test eax,eax
	mov edx,TOS(1)
	spush
	mov eax,0
	jnz .r
	test edx,edx
	jnz .r
	not eax
.r:
	ret
;-
;- d0=		( d1 -- flg )			(a:d)(m,14)(e)	'dzeq'		forth
;-	tf if d1 equal to 0
;-	(e: Z if d1 = 0
dzeq:GA M;:ok:
	or eax,TOS(1)
	TOSPULL(1)
	mov eax,-1
	jz .r
	not eax
.r:
M dzeq
	ret
;-
;- D0<>		( d1 -- flg )			(a)		'dzneq'		ans
;-
;- d0=/=  	( d1 -- flg )			(m,13)(e)	'dzneq'		forth
;-	tf if double, dx1 is non-zero
;-	(e: NZ if d1 =/= 0
dzneq:GA M;:ok:
	or eax,TOS(1)
	TOSPULL(1)
	jz .r
	mov eax,-1
.r:
M dzneq
	ret
;-
;- d=		( d1 d2 -- flg )		(79)(a:d)(e)	'deq'		forth
;-	tf if d1 equal to d2
;-	(e: Z if d1=d2
deq:GA;:ok:
	cmp eax,TOS(2);
	mov ecx,TOS(1)
	mov eax,0
	jnz .r
	cmp ecx,TOS(3);
	jnz .r
	not eax
.r:
	TOSPULL(3)
	ret
;-
;- D<>		( d1 d2 -- flg )				'dneq'		ans
;-	tf if d1 not equal to d2
;-
;- d=/=		( d1 d2 -- flg )				'dneq'		forth
;-	tf if d1 not equal to d2
dneq:GA;:ok:
	xor eax,TOS(2)
	jnz .n
	mov eax,TOS(1)
	xor eax,TOS(3)
	jz .r
.n:
	mov eax,-1
.r:
	TOSPULL(3)
	ret
;-
;- dparity	( d1 -- flg )							forth
;-	ff if d1 w/ odd number of set bits
dparity:GA M;:ok:-?-
	mov ecx,0
	xor eax,TOS(1)
	jnp .r
	dec ecx
.r:
	TOSPULL(1)
	mov eax,ecx
M dparity
	ret
;-
;-
;-================================================================
;--- binary logic opr
;-

;-
;- 1abs		( n -- +n )			(f6)(m,6)	'abs1'		forth
;-	absolute value wrt one-s complement - {invert} if -ve
abs1:GA M;:ok:
	test eax,eax
	jns .r
	not eax
.r:
M abs1
	ret
;-
;- not		( u1 -- u2 )					'invert'	fig
;-	return u2 1-s complement to u1
;-
;- invert  	( u1 -- u2 )			(a:c)(m,2)	'invert'	forth
;-	return u2 1-s complement to u1
invert:GA M;:ok:
	not eax
M invert
	ret
;-
;- d1abs  	( d1 -- +d1 )			(m,9)				forth
;-	ret one-s complement of double if -ve
d1abs:GA M;:ok:
	test eax,eax
	jns .r
	not dword TOS(1)
	not eax
.r:
M d1abs
	ret
;-
;- 2not		( ud1 -- ud2 )			(f6)(m,5)	'dnot'		fig
;-	return ud2 1-s complement of ud1
;-
;- 2invert	( ud1 -- ud2 )			(f6)(m,5)	'dnot'		forth
;-	return ud2 1-s complement of ud1
dnot:GA M;:ok:
	not dword TOS(1)
	not eax
.r:
M dnot
	ret
;-
;- and		( u1 u2 -- u3 )			(m,6)		'aand'		forth
;-	bitwise <and> u1 w/ u2
aand:GA M;:ok:
	and eax,TOS(1)
	TOSPULL(1)
M aand
	ret
;-
;- 2and		( d1 d2 -- d3 )			(f6)		'twoand'	forth
;-	bitwise <and> d1 w/ d2
twoand:GA;:ok:
	and eax,TOS(2); d1.h & d2.h
	mov ecx,TOS(1); d1.l
	TOSPULL(2)
	and TOS(1),ecx; d1l & d2.l
	ret
;-
;- or		( u1 u2 -- u3 )			(m,6)		'oor'		forth
;-	bitwise <or> u1 w/ u2
oor:GA M;:ok:
	or eax,TOS(1)
	TOSPULL(1)
M oor
	ret
;-
;- 2or		( d1 d2 -- d3 )			(f6)		'twoor'		forth
;-	bitwise <or> d1 w/ d2
twoor:GA;:ok:
	or eax,TOS(2)
	mov ecx,TOS(1)
	TOSPULL(2)
	or TOS(1),ecx
	ret
;-
;- xor		( u1 u2 -- u3 )			(m,6)		'xxor'		forth
;-	bitwise <xor> u1 w/ u2
xxor:GA M;:ok:
	xor eax,TOS(1)
	TOSPULL(1)
M xxor
	ret
;-
;- 2xor		( d1 d2 -- d3 )			(f6)		'twoxor'	forth
;-	bitwise <xor> d1 w/ d2
twoxor:GA;:ok:
	xor eax,TOS(2)
	mov ecx,TOS(1)
	TOSPULL(2)
	xor TOS(1),ecx
	ret
;-
;- andx	( n1 n2 n3 -- n4 )					'andx'		forth
;-	n4 := (( n1 AND n2 ) XOR n3 )
andx:GA M
	mov edx,eax
	mov ecx,TOS(1)
	spop 2
	and eax,ecx
	xor eax,edx
M andx
	ret
;-
;- toggle  	( p c -- )			(fg)				fig
;-	xor byte at ptr p with byte c
;-	(fig-4th opr w. inverse operands order: byte c @tos, ptr @nos)
toggle:GA;:ok:
; c: ecx
	mov ecx,TOS(1)
	xor [kref ecx],al
	spop 2
	ret
;-
;- and!		( n p -- )			(f6)		'ands'		forth
;-	<and> n to cell at p in memory
ands:GA;:ok:
	mov ecx,TOS(1)
	and [kref eax],ecx
	spop 2
	ret
;-
;- or!		( n p -- )			(f6)		'ors'		forth
;-	<or> n to cell at p in memory
ors:GA;:ok:
	mov ecx,TOS(1)
	or [kref eax],ecx
	spop 2
	ret
;-
;- xor!		( n p -- )			(f6)		'xors'		forth
;-	<xor> n to cell at p in memory
xors:GA;:ok:
	mov ecx,TOS(1)
	xor [kref eax],ecx
	spop 2
	ret
;-
;- on		( p -- ) 			(83)(m,)			forth
;-	store -1 to cell at p
on:GA M;:ok:
	mov dword[kref eax],-1
	spop
M on
	ret
;-
;- off		( p -- ) 			(83)(m,)			forth
;-	store 0 to cell at p
off:GA M;:ok:
	mov dword[kref eax],0
	spop
M off
	ret
;-
;- @0=		( p -- flg )			(m,)		'zeqf'		forth
;-	tf if cell @ptr p is zero
zeqf:GA M;:ok:
	cmp dword[kref eax],byte 1
	sbb eax,eax
M zeqf
	ret
%endif;__ALOGIC_I
;-
;-=====================================================================;
;- l4alogic.inc <eof>
