;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4bignum.inc
;   created	: 14-jul-01
;   modified	: -> re "version.inc"
;   version	: 0.01(+)
;   latest	: 1.33, 0.16.1+10
;   switches	: __BIGNUM_I
;   assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;   description	: "lib4th" shared library Forth kernel
;   author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;   comment	: doc re "glossary.txt" and "l4...html" in ./doc dir,
;		: "lib4words.inc" for short reference and which words
;		: belong to which vocabulary, or in 4th { help help }.
;		: generate docs w/ <make doc> and/or <make html>.
;   source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __BIGNUM_I
%define __BIGNUM_I
%define bignuminc

;-
;-----------------------------------------------------------------
;--- memory
;-

;-
;- n@+		( p1 -- p2 .xx. N )				'nfetchp'	bignum
;-	fetch counted int from memory w/ count at l.s.cell
;-	ret p2 as p1 advanced by size of N+1 cells.
nfetchp:GA;:ok:
	push PS
	push eax
	spush
	call nfetch
	pop ecx
	pop edx
	lea ecx,[ecx+eax*4+4]
	mov [edx],ecx
	ret
;-
;- n@		( p -- .xx. N )					'nfetch'	bignum
;-	fetch counted int from memory w/ count at l.s.cell
nfetch:GA;:ok:
	add eax,byte 4
	spush
	mov eax,[kref eax-4]
	push eax
	shl eax,2
	call mtos
	spush
	pop eax
	ret
;-
;- n!+		( .x. N p1 -- p2 )				'nstorep'	bignum
;-	store counted int w/ count at l.s.cell to memory
;-	ret p2 as p1 advanced by size of N+1 cells.
nstorep:GA;:ok:
	mov edx,TOS(1)
	lea edx,[eax+edx*4+4]
	push edx
	call nstore
	spush
	pop eax
	ret
;-
;- n!		( .x. N p -- )					'nstore'	bignum
;-	store counted int w/ count at l.s.cell to memory
nstore:GA;:ok:
	mov ecx,TOS(1)	; count
	mov [kref eax],ecx
	add eax,byte 4
	mov TOS(1),eax
	lea eax,[ecx*4]	;:continues:
;-
;- s>m		( n1..nn p n -- )				'stom'		forth
;-	store n bytes from tos to memory location p
;-	beginning w. cell aligned n1
stom:GA;:ok:
	push esi
	mov ecx,eax
	lea esi,TOS(1)
	lea eax,[eax+3]
	mov edi,[esi]	; PS := dst
	and al,-4
	add edi,KR
	sub esi,eax	; dst := tos - bytecount, -cell aligned
	lea edx,[esi-4]
	rep movsb
	mov PS,edx	; = edi
	pop esi
	mov eax,[PS]
	ret
;-
;- nn>r		( n1..nn N -- )(R: -- nn..n1 N M )		'ntor'		bignum
;-	transfer counted cells group from tos to tor, additioanlly store items count M
;-	does{ dup 1+ dup minus roll dup 2 + n>r }
nntor:GA;:ok:
	spush
	inc eax
	spush
	neg eax
	call roll
	spush
	lea eax,[eax+2]	;:continues: in <ntor>
;-
;- n>r		( n1..nn n -- )(R: -- nn..n1 )			'ntor'		forth
;-	transfer n cells from tos to tor, ready to fetching back w. {r>}, etc.
ntor:GA;:ok:
	mov ecx,eax
	pop edx
	jr .n
.l:
	add eax,KR
	push eax
.n:
	lea PS,[PS-4]
	dec ecx
	mov eax,[PS]
	jns .l
	jmp edx
;-
;- nnr>		( -- n1..nn N )(R: n1..nn N M -- )		'nnrfrom'	bignum
;-	transfer counted cells group from tor to tos, as stored w. {nn>r}.
;-	does{ r> nr> }
nnrfrom:GA;:ok:
	pop edx		; ret
	pop ecx
	sub ecx,KR	; count
	jr nrfrom.n
;-
;- nr>		( n -- n1..nn )(R: nn..n1 -- )			'nrfrom'	forth
;-	transfer n cells from tor to tos
nrfrom:GA;:ok:
	mov ecx,eax
	pop edx
	spop
	jr .n
.l:
	lea PS,[PS+4]
	pop eax
	sub eax,KR
.n:
	dec ecx
	mov [PS],eax
	jns .l
	jmp edx
;-
;- nr@		( n -- n1..nn )(R: nn..n1 -- )			'nrfetch'	forth
;-	copy n cells from tor to tos
nrfetch:GA
	mov ecx,eax
	lea edx,[esp+4]
	spop
	jr .n
.l:
	mov eax,[edx]
	lea PS,[PS+4]
	lea edx,[edx+4]
	sub eax,KR
.n:
	dec ecx
	mov [PS],eax
	jns .l
	ret
;-
;- n@sp		( .n.N ... .m.M n(n) -- .n.N ... .m.M n2 )	'nnposn'	bignum hidden
;-	find disp n2 to <pick> the multi-cell item no. n(n)
nnposn:A;-:ok:<nnroll, nnpick)
; i:	eax = dstn group no.
; o:	eax = cells disp to l.s.cell of dstn group
;	edx = dstn group's count
	mov ecx,eax	; count of groups
	mov eax,0
.l:
	mov edx,eax
	not edx
	mov edx,[PS+edx*4]; cells count of group
	dec ecx
	lea eax,[edx+eax+1]
	jns .l
	dec eax
	ret
;-
;- nnsdrop	( .n.N .m.M -- .m.M )						bignum
;-	drop next @tos counted int item
nnsdrop:GA
	call nnswap
	jmp ndrop
;-
;- n2dup  	( .nn.N .mm.M -- .nn.N .mm.M .nn.N .mm.M )	'nn2dup'	bignum
;-	duplicate top two multi-cell integers (4th, by { nnover nnover })
nn2dup:GA;:ok:
	call nnover	;:continued:
;-
;- nnover 	( .nn.N .mm.M -- .nn.N .mm.M .nn.N )				bignum
;-	copy next on tos counted int to tos
nnover:GA;:ok:
	spush
	mov eax,1	;:continued:
;-
;- nnpick 	( .xX .mM .. .nn.N m -- .xX .mM .. .nn.N .mm.M )		bignum
;-	m > 0: copy indexed counted int to tos, { 0 nnpick } == { nndup }.
;-	m < 0: replace indexed item by counted int @tos.
nnpick:GA;:ok:
	test eax,eax	;
	js .m
	call nnposn	;o: d=dstn group's count, a=cells disp to l.s. cell above src group
.n:
	push esi	; save lp
	not eax
	lea ecx,[edx+1]
	lea esi,[edi+eax*4]
	rep movsd
	pop esi
	spop
	ret
.m:A;:ok:
	push eax
	neg eax
	inc eax
	call nnroll
	call ndrop
	spush
	pop eax
	jr nnroll
;-
;- nn-rot 	( .n.N .m.M .o.O -- .o.O .n.N .m.M .o.O )	'nnmrot'	bignum
;-	roll counted int @tos to 3rd, by { -2 nnroll }.
nnmrot:GA
	spush
	mov eax,-2
	jr nnroll
;-
;- nnrot  	( .n.N .m.M .o.O -- .m.M .o.O .n.N )				bignum
;-	roll 3rd counted int to tos, by { 2 nnroll }.
nnrot:GA;:ok:
	spush
	mov eax,2
	jr nnroll
;-
;- nnswap 	 ( .nn.N .mm.M -- .mm.M .nn.N )					bignum
;-	exchange top two counted ints at tos, by { 1 nnroll }.
nnswap:GA
	spush
	mov eax,1	;:continued:
;-
;- nnroll 	( .nn.N .xx.N-1 ... .mm.M n -- .xx.N-1 ... .mm.M .nn.N )	bignum
;-	n > 0: roll indexed counted int to tos, { 1 nnroll } == { nnswap }.
;-	n < 0: roll counted int @tos into indexed stack posn
nnroll:GA;:ok:
	test eax,eax	;
	js .m
	call nnposn
	push esi	; save reg
	push edx	; last size
	push eax	; cells disp
	call nnpick.n
	pop ecx		; disp
	pop edx		; size
	neg ecx		; -(disp)-1
	not edx		; -(size)-1
	lea esi,[edi+ecx*4];src = tos - cells.disp'
	lea edi,[esi+edx*4];dst = tos - cells.disp' - cells.size'
	neg ecx
	rep movsd
	pop esi		;reg
	mov [PS],eax
	ret
.m:A;:ok:
	neg eax		;i: a=dstn group no.
	call nnposn	;o: d=dstn group's count, a=cells disp to l.s. cell above src group
	push esi	; save lp
	push edi	; save ps
	lea ecx,[eax+1]	; count of all cells + top count cell to move, bot up by size of top item
	not eax		; -ve cells disp to l.s.cell
	lea eax,[edi+eax*4]; bot of item to sto back
	push eax
	lea esi,TOS(1)	; src top dn, top item
	mov edx,TOS(1)	; size of top item
	mov eax,edi	; save bot of temp top
	lea edi,[edi+edx*4]; dst = temp top, top dn
	std		; copy top dn
	rep movsd
	cld		; copy bot up
	pop edi		; dst for top item
	mov esi,eax	; src to sto back, bot up
	lea ecx,[edx+1]	; count of top item
	rep movsd	; sto top to posn
	pop edi		; rst ps
	pop esi		; rst lp
	spop		; drop index cell
	ret
;-
;- nndup  	( nn..n0 +N -- nn..n0 +N nn..n0 +N )		(f6)		bignum
;-	duplicate +N cells sized multi-cell integer, 0 < N < 32768 cells.
nndup:GA;:ok:
; c: ecx
	spush
	inc eax		;:continued:
;-
;- ndup		( nn..n0 m -- nn..n0 nm..n0 )	(f6)				forth
;-	duplicate group of 0 < n < 32768 cells,
;-	else drop group of 0 > n > -32768 cells
; c: ecx
ndup:GA;:ok:
	test eax,eax
	js .d
	mov ecx,eax
	shl eax,2
	jz .r
	jo .r
	push esi,ecx
	lea ecx,[PS+ecx*4+4]
	call qstack.n
	pop ecx		; count
	mov eax,ecx
	neg eax
	lea esi,[PS+eax*4]
	rep movsd
	pop esi
.r:
	spop
	ret
.d:;(alignment)
	lea PS,[4*eax+PS-4]; drop -n cells
	mov eax,[PS]
	ret
;-
;-----------------------------------------------------------------
;--- (signed) numbers, counted 'multi-cell integers'
;-

;-
;- nndrop 	( .xx. N ... .xx. M n -- )					bignum
;-	drop n counted int-s
nndrop:GA;:ok:
	test eax,eax
	push eax
	spop
	jng .r
.l:
	call ndrop
	dec dword[esp]
	jg .l
.r:
	lea esp,[esp+4]
	ret
;-
;- n->q		( .n.N -- qn )					'n2q'		bignum
;-	signed extend/truncate bignum to un-counted quad
n2q:GA;:ok:
	spush
	mov eax,4
	call nton
	spop
	ret
;-
;- q->n		( qn -- .n.N )			(m,6)		'q2n'		bignum
;-	convert un-counted quad to bignum
q2n:GA M;:ok:
	spush
	mov eax,4
M q2n
	ret
;-
;- 2q->2n 	( q1 q2 -- .n1.N .n2.M )			'q22n'		bignum
;-	convert a pair of un-counted quads to bignum
q22n:GA;:ok:
	spush
	mov eax,4
	spush 2
	not eax
	jmp roll
;-
;- 2q->2un 	( q1 q2 -- .n1.N .n2.M )			'q22un'		bignum
;-	convert a pair of un-counted quads to unsigned bignum
q22un:GA;:ok:
	spush
	mov eax,0
	lea ecx,[eax+5]
	spush 3
	mov TOS(2),ecx
	mov eax,ecx
	spush
	lea eax,[ecx-8]
	jmp roll2
;-
;- 2n->2q 	( .n1.N .n2.M -- q1 q2 )			'n22q'		bignum
;-	signed extend/truncate two bignums to un-counted quad
n22q:GA;:ok:
	call q2n
	call nton
	call nnswap
	call q2n
	call nton
	call n22d
	jmp swap4
;-
;- 2n->2uq 	( .n1.N .n2.M -- q1 q2 )			'n22uq'		bignum
;-	unsigned extend/truncate two bignums to un-counted quad
n22uq:GA;:ok:
	call q2n
	call untoun
	call nnswap
	call q2n
	call untoun
	call n22d
	jmp swap4
;-
;- nn->dd 	( .nn.N .mm.M -- .nn .mm )			'n22d'		bignum
;-	convert two bignums to un-counted data
n22d:GA;:ok:
	call roll
	spop
	ret
;-
;- n->un  	( .n.N -- .un.M )				'n2un'		bignum
;-	convert signed to unsigned bignum: if .n.N < 0 push top cell := 0, M := N+1
n2un:GA
	cmp dword TOS(1),byte 0
	lea ecx,[eax+1]
	jns .r
	mov eax,0
	spush
	mov eax,ecx
.r:
	ret
;-
;- n->n		( .nn1. N M -- .nn2. M )			'nton'		bignum
;-	truncate/signed extend from +N cells to +M cells size
nton:GA;:ok:
; c: ecx,edx
	mov edx,eax
	sub eax,TOS(1)
	TOSPULL(1)
	jz .x
	jng untoun.d
	mov ecx,eax
	mov eax,TOS(1)
	sar eax,31
	rep stosd
.x:
	mov [PS],edx
	mov eax,edx
	ret
;-
;- un->un 	( .un1. N M -- .un2. M )			'untoun'	bignum
;-	unsigned extend/truncate from +N cells to +M cells size
untoun:GA;:ok:
; c: ecx,edx
	mov edx,eax
	sub eax,TOS(1)
	TOSPULL(1)
	jz .x
	jng .d
	mov ecx,eax
	mov eax,TOS(1)
	mov eax,0
	rep stosd
.d:
	lea PS,[4*eax+PS]; drop index, add/drop n cells
.x:
	mov [PS],edx
	mov eax,edx
	ret
;-
;- n->		( .nn1. N -- .nn2. N2 )				'ntoq'		bignum
;-	reduce multi-cell signed integer to smallest size w/o. loss of precision
ntoq:GA;:ok:
	test eax,eax
	jz nminus.r	;?; zero count
	mov ecx,TOS(1)
	test ecx,ecx
	push ecx
	jns .p		;?; +ve
	call nminus
.p:
	cmp eax,1
	jng .x		;?; ready, min size
	cmp dword TOS(1),0
	jnz .x		;?; ready
	cmp dword TOS(2),0
	js .x
	TOSPULL(1)
	dec eax
	jr .p
.x:
	pop ecx
	test ecx,ecx
	jns nminus.r	;:continued:
;-
;- nminus  	( n1..nn n -- m1..mn n )	(4)				bignum
;-	2-s cpl of group of n cells on tos
nminus:GA;:ok:
	push eax
	call ninvert
	call nonep
	pop eax
.r:
	ret
;-
;- ninvert	( n1..nn n -- m1..mn n )	(4)				bignum
;-	1-s cpl of group of n cells on tos
ninvert:GA;:ok:
	push eax	; count
	mov ecx,eax	; n cells
	not eax
	lea PS,[PS+eax*4]
.l:
	lea PS,[PS+4]
	dec ecx
	not dword[PS]
	jnz .l
	mov eax,[PS]
	spush
	pop eax
	ret
;-
;- nabs		( n1..nn n -- n1'..nn' n )	(4)				bignum
;-	negate counted integer if -ve
nabs:GA;:ok:
	cmp dword TOS(1),byte 0
	js nminus	;?; -ve
	ret
;-
;- n+-		( n1..nn n m -- n1'..nn' n )	(4)		'npm'		bignum
;-	apply sign of sing m to counted integer
npm:GA;:ok:
	test eax,eax
	spop
	js nminus
	ret
;-
;- n0=		( n1..nn N -- flg )				'nzeq'		bignum
;-	test group of n cells on tos for all zero
nzeq:GA;:ok:
	call qnzeq
	mov PS,edx
	ret
;-
;- ?n0=		( .n.N -- .n.N f )				'qnzeq'		bignum
;-	non-destructively test group of n cells on tos for all zero
qnzeq:GA;:ok:
	mov ecx,eax
	mov edx,PS
	spush
	xor eax,eax
.l:
	dec ecx
	lea edx,[edx-4]
	js .r
	or eax,[edx]
	jz .l
	mov eax,-1
.r:
	neg ecx
	not eax
	lea edx,[edx+ecx*4]; edx = tos w/o tested item
	ret
;-
;- n=		( n1..nn N m1..mn M -- flg )	(e)		'neq'		bignum
;-	compare two groups of n=m cells ea, on tos.
;-	normalizing for signed multi-cell integers if n=/=m.
;-	(e: Z if n.N = m.M
neq:GA;:ok:
	call nnorm
	mov ecx,eax	; n cells
	neg eax
	push esi
	lea esi,[PS+eax*4]
	lea edi,[esi+eax*4-4]
	mov edx,edi
	mov eax,0
	repz cmpsd
	mov PS,edx
	pop esi
	jnz .r
	lea eax,[eax-1]
.r:
	mov [PS],eax
	ret
;-
;- n0>		( n1..nn n -- flg )				'nzgt'		bignum
;-	test group of n cells on tos for +ve non-zero bignum
nzgt:GA;:ok:
	mov ecx,eax
	xor eax,eax
	cmp eax,TOS(1)
	jg .r
.l:
	dec ecx
	lea PS,[PS-4]
	js .r
	or eax,[PS]
	jz .l
	mov eax,-1
.r:
	neg ecx
	lea PS,[PS+ecx*4]
	ret
;-
;- n0<		( n1..nn n -- flg )				'nzlt'		bignum
;-	test group of n cells on tos for -ve bignum
nzlt:GA;:ok:
	mov ecx,eax
.n:
	neg ecx
	mov eax,TOS(1)
	lea PS,[PS+ecx*4]
.r:
	sar eax,31
	ret
;-
;- ?ncmp  	( .n.N .m.M -- .n.N .m.M flg )	(4)(e)		'qncmp'		bignum
;-	normalize and non-destructively signed compare two counted integers, on tos.
;-	ret flg according to sgn(subtraction)
;-	(e: eflags reflect sgn(flg)
qncmp:GA;:ok:
	call nn2dup	; :continued:
;-
;- ncmp		( n1..nn n m1..mn m -- flg )	(4)(e)				bignum
;-	normalize and signed compare two counted integers, on tos.
;-	ret flg according to sgn(subtraction)
;-	(e: eflags reflect sgn(flg)
ncmp:GA;:ok:
	call nsub	; :continued:
;-
;- nsgn		( n1..nn n -- f )		(e)				bignum
;-	ret signum of counted int, .n.N <0: f:=-1, =0:f:=0, >0: f:=1.
;-	(e: eflags reflect sgn(flg)
nsgn:GA;:ok:
	push dword TOS(1)
	call nzeq
	inc eax
	pop ecx
	jz .r
	mov eax,ecx
	sar eax,31
	or eax,byte 1	; flags
.r:
nnorm.x:
	ret
nnorm.g:A
	push eax
	call nnswap
	spush
	pop eax
	call nton
	jmp nnswap
;-
;- nnorm  	( n1..nn N m1..mn M -- n1..nn N m1..mn N )			bignum
;-	adjust two counted integers to minimal, equal size w/o loss of precision.
nnorm:GA;:ok:
	mov ecx,eax
	neg ecx
	mov edx,[PS+ecx*4-4]
	cmp eax,edx
	jz nnorm.x	;?; already equal sizes
	jg .g		;?; c(m) > c(n)
.n:
	spush
	mov eax,edx
	jmp nton
;-
;- n1+		( n1..nn N -- m1..mn N )			'nonep'		bignum
;-	increment multi precision integer of n cells on tos by one
nonep:GA;:ok:
	xor edx,edx	;(alignment)
	inc edx;(alignment);
.s:;	;<other sing. add's)
	mov ecx,eax	; n cells
	neg eax
	lea PS,[PS+eax*4]
	add dword[PS],edx
	mov edx,ecx
	lea eax,[eax+ecx]
	jr .n
.l:A
	lea PS,[PS+4]
	adc [PS],eax
.n:
	dec ecx
	jnz .l
	mov eax,[PS]
	spush
	mov eax,edx
	ret
;-
;- n1-		( n1..nn N -- m1..mn N )			'nonem'		bignum
;-	increment multi precision integer of n cells on tos by one
nonem:GA;:ok:
	xor edx,edx;(alignment)
	inc edx	;(alignment)
.s:
	mov ecx,eax	; n cells
	neg eax
	lea PS,[PS+eax*4]
	sub dword[PS],edx
	mov edx,ecx
	lea eax,[eax+ecx]
	jr .n
.l:A
	lea PS,[PS+4]
	sbb [PS],eax
.n:
	dec ecx
	jnz .l
	mov eax,[PS]
	spush
	mov eax,edx
	ret
;-
;- n-		( n1..nn N m1..mn M -- k1..kn N )		'nsub'		bignum
;-	subtract normalize multi precision integers of N cells ea, on tos.
nsub:GA;:ok:
	call nnorm
	mov ecx,eax	; n cells
	neg eax
	push ecx,esi
	lea esi,[PS+eax*4]
	lea edi,[esi+eax*4-4]
	push esi
	or eax,eax	; clr carry
.l:
	mov eax,[edi]
	sbb eax,[esi]
	mov [edi],eax
	lea edi,[edi+4]
	lea esi,[esi+4]
	dec ecx
	jg .l
	pop eax,esi,PS
	TOSPULL(1)
	ret
;-
;- n+		( n1..nn N m1..mn M -- k1..kn N )		'nadd'		bignum
;-	add normalized multi precision integers of N cells ea, on tos.
nadd:GA;:ok:
	call nnorm
	mov ecx,eax	; n cells
	neg eax
	push ecx,esi	; save count & reg
	lea esi,[PS+eax*4]
	lea edi,[esi+eax*4-4]
	push esi	; new tos
	or eax,eax	; clr carry
.l:A
	mov eax,[edi]
	adc eax,[esi]
	mov [edi],eax
	lea edi,[edi+4]
	lea esi,[esi+4]
	dec ecx
	jg .l
	pop eax,esi,PS
	TOSPULL(1)
.r:
	ret
;-
;- lg(nn) 	( n1..nn N nm -- nk )		(4)		'lgnn'		bignum
;-	ret nk := numerus lg(.n.N) to the base of nm,
;-	i.e nk := count of digits the rsp number would be converted to at radix of nm.
;-	only conventionally named "lg" after "logarithm" - which this opr yields not!
lgnn:GA;:ok:
	push eax	; cells count
	push_ibyte(-1)	; log count
	spop		; drop nm
	jr .entry
.begin:A
	mov eax,[esp+4]	; n2
	call unsbym	; uns/mod
	spop		; drop
	inc dword[esp]	; log 1+
.entry:
	call ntoq	; n->
	call qnzeq
	test eax,eax
	jz .begin
	mov eax,TOS(1)
	call ndrop
	pop eax
	add esp,byte 4
	ret
;-
;- s>nfactorial	( u -- .ux. n )					's2nfac'	bignum
;-	factorial of n,
;-	u*(u-1)*...(u-(u-1)) counted integer .ux. n from sing integer u.
s2nfac:GA;:ok:
	mov ecx,eax
	sub eax,byte 1
	mov eax,1	; 0!, 1! = 1
	spush
	jle .r		;?; 0, 1, -ve
	push ecx
.b:
	spush
	mov eax,[esp]
	call unsmul
	call ntoq
.e:
	dec dword[esp]
	jg .b
	add esp,byte 4
.r:
	ret
;-
;- uns/mod	( n0..nn N m -- q0..qn N rm )			'unsbym'	bignum
;-	unsigned divide multi-cell dividend by sing divisor, 
;-	leave sing rmd & N-sized quot.
;-	ret +ve or -ve max if m=0; rm:=0 if N=0; trivial case div-1
unsbym:GA;:ok:
	cmp dword TOS(1),byte 0
.f:; N=0, div 1
	jng near rfalse	;?; not a +ve cells count
	test eax,eax
	mov ecx,eax	; divisor
	spop
	jz .n		;?; div0
	cmp ecx,byte 1
	jz .f		;?; div by 1
	push eax,ebx,PS,eax; save count & reg & tos & counter
	mov edx,0
.l:A
	lea PS,[PS-4]
	mov eax,[PS]
	div ecx
	dec dword[esp]
	mov [PS],eax
	jg .l		;?; more
	lea esp,[esp+4]
	pop eax,ebx,PS
	spush
	mov eax,edx	; rmd
	ret
.n:A; div-0 hnd
	push dword TOS(1); sign
	call nndup
	call qnzeq
	push eax	; dvd Z-flag
	spop
	call ninvert
	call nadd	; => -1
	pop ecx
	inc ecx
	jz .m		;?; zero dvd
	spush
	mov eax,1
	call nrshift	; => 7ff...fff
	spush
	call roll
	spop
	dec eax
	cmp dword[esp],byte 0; sign
.m:
	lea esp,[esp+4]
	js near nminus
	ret
;-
;- uns*		( n0..nn N m -- p0..pn ov P )			'unsmul'	bignum
;-	unsigned multiply counted int by sing, 
;-	leave P:=(N+1) sized prod; P:=0 if N=0; trivial cases mul 1, mul 0
unsmul:GA;:ok:
	cmp dword TOS(1),byte 0
	jz near rfalse
	test eax,eax
	jz .n		;?; mul by zero
	cmp eax,byte 1
	jz .o		;?; mul by 1
	push esi,ebx,edi,eax; reg & kref & tos(prd) & mkr
	spop
	mov edx,eax	; cells count
	mov ecx,[esp]	; multiplicator
	neg eax		; index to l.s.cell
	mov [esp],edx	; save count
	push edx	; store counter
	lea PS,[PS+eax*4]; multiplicand l.s.cell => prod.
	xor edx,edx	; clr reg & flags
	mov ebx,edx	; clr carry reg
.l:A
	mov eax,[PS]
	neg ebx
	mov [PS],edx
	mul ecx
	add eax,ebx	; ripple carry
	adc edx,byte 0
	add [PS],eax
	adc edx,byte 0
	sbb ebx,ebx	; save carry
	dec dword[esp]
	lea PS,[PS+4]
	jnz .l		;?; more
	mov [PS],edx	; ovf cell
	pop esi,ebx,edi,eax,ebx; drop counter, fetch count, reg, kref, tos
	inc eax
	mov [PS],eax
	ret
.o:A; mul 1
	mov eax,TOS(1)
	sub TOS(1),eax
	inc eax
	ret
.n:A; mul zero
	mov eax,TOS(1)
	call ndrop
	xor eax,eax
	spush
	inc eax
	ret
;-
;- nshift 	( nm .. n0 M -s -- ov n0' .. nm' M+1 )				bignum
;-	 	( nm .. n0 M +s -- n0' .. nm' ov M+1 )
;-	n > 0: -re- {nlshift}
;-	n < 0: -re- {nrshift}
nshift:GA;:ok:
	test eax,eax
	jns nlshift
	neg eax
;-
;- nrshift 	( nm .. n0 M +n -- ov n0' .. nm' +(M+1) )			bignum
;-	right shift block of M cells by unlimited +ve no. n bits, leave ovf bits in ov.
;-	NOTE: this is a bit-wise op which might be used for but, is not "arithmetic".
nrshift:GA;:ok:
	test eax,eax
	mov ecx,eax	; shift count
	spop
	js .r		;?; invalid bitcount
.c:
	sub ecx,byte 32
	jc .s
	call nd32
	jr .c
.s:A
	add ecx,byte 32
.d:; <nn2div)
	mov edx,eax	; cells count
	neg eax
	lea PS,[PS+eax*4]; l.s. cell
	push ebx,edx
	mov ebx,0
.l:A
	mov eax,[PS]
	shrd ebx,eax,cl
	mov [PS],ebx
	mov ebx,eax
	dec edx
	lea PS,[PS+4]
	jnz .l		;?; more
	shr eax,cl
	mov [PS],eax
	lea PS,[PS+4]
	pop ebx,eax
	inc eax
.r:
	ret
;-
;- nlshift 	( nm .. n0 M +n -- n0' .. nm' ov +(M+1) )			bignum
;-	left shift block of m cells by unlimited +ve no. n bits, leave ovf bits in ov.
;-	NOTE: this is a bit-wise op which might be used for but, is not "arithmetic".
nlshift:GA;:ok:
	test eax,eax
	mov ecx,eax	; shift count
	spop
	js nn2mul.r	;?; invalid bitcount
.c:
	sub ecx,byte 32
	jc nn2mul.s
	call nm32
	jr .c
;-
;- n2*		( .nn. n -- .mm. m )				'nn2mul'	bignum
;-	multiply counted int by 2
nn2mul:GA;:ok:
	mov ecx,-31
.s:A
	spush
	add ecx,byte 32
	push PS
	spop
	mov edx,eax	; cells
	push ebx,eax	; reg & kref & count
	mov ebx,0	; begin
.l:A
	mov eax,[PS-4]	; from m.s. cell "down"
	shld ebx,eax,cl
	mov [PS],ebx
	lea PS,[PS-4]
	mov ebx,eax
	dec edx
	jns .l		;?; more
	shld edx,eax,cl
	pop PS,ebx,eax
	inc eax
.r:
	ret
;-
;- n2/		( .nn. n -- .mm. m )				'nn2div'	bignum
;-	unsigned divide counted int by 2
nn2div:GA;:ok:
	mov ecx,1
	call nrshift.d
	spush		; dup
	call roll	; get ovf
	spop		; drop
	dec eax		; adjust count
	ret
;-
;- n/2^32 	( .n.N -- .m.N-1 )				'nd32'		bignum
;-	n:=n/(2^32), divide counted int by cell-size (2^32)
;-	does{ dup roll drop 1- }, ret n={0,1} if N u< 2
nd32:GA;:ok:
	cmp eax,byte 1
	jna disp unsmul.n	; leave min. size of cell + count
	push ecx
	spush
	call roll
	spop
	pop ecx
	dec eax
	ret
;-
;- n*2^32  	( .n.N -- 0.n.N+1 )				'nm32'		bignum
;-	n:=n*(2^32), multiply counted int by cell-size (2^32)
;-	does{ 0 over not roll 1+ }
nm32:GA;:ok:
	mov edx,eax
	spush
	push ecx
	xor eax,eax
	not edx
	spush
	mov eax,edx
	call roll
	inc eax
	pop ecx
	ret
;-
;- n*		( .nn. N .mm. M -- .xx. X )	(4)		'nmul'		bignum
;-	signed multiply two counted integers
nmul:GA;:ok:
	push dword TOS(1)	; sg(m)
	call nabs
	call nnswap
	push dword TOS(1)	; sg(n)
	call nabs
	call nnswap
	spush 2
	push dword[byte uref uv.s0]
	inc eax			; w/ count cell
	call toloc
	mov eax,[esp]
	call nstore		; .mm.M to local: count @l.s.cell, little endian cells follow
	add [esp],KR		; true addr
	spush
	xor eax,eax
	spush
	inc eax			; 0 1 init accu
.l:A; ( .n.N .a.A -- .. ) loc{ .m.M }
	call nm32		; a:=a*(2^32)
	call nnover		; ( .n.N .a.A' .n.N -- .. )
	mov ecx,[esp]		; locmem base. l.s. cell
	spush
	mov edx,[ecx]		; cells count
	mov eax,[ecx+4*edx]	; next m.s. cell of mt
	call unsmul		; ( .. -- .n.N .a.A .p.P ) mul mpd by sing
	call nadd		; ( .. -- .n.N .a.A" )
	call ntoq		; reduce size
	mov ecx,[esp]
	dec dword[ecx]
	jg disp .l
	call nnswap
	call ndrop
	pop ecx
	pop edx
	xor edx,[esp]		; sg(m x n) = sgn(prod)
	mov [esp],ecx
	jns .o
	call nminus
.o:
	spush
	pop eax
	sub eax,KR
	sub eax,[byte uref uv.s0]
	sar eax,2
	jmp toloc
;-
;- n^		( n.N m -- x.X )		(4)		'npow'		bignum
;-		raise counted int n.N to the power of sing m.
npow:GA
	mov edx,TOS(1)
	test edx,edx
	mov ecx,eax
	spop
	jng .z		;?; empty or false data, ret 0
	dec eax
	mov ecx,eax
	push eax	;
	spop
	jng .e		;?; n.N^0, n.N^(-ve) := 1
	dec ecx
	jz .u		;?; n.N^1
	mov ecx,TOS(1)
	cmp ecx,byte 1
	jnz .p		;?; multi cell, pow > 1
	cmp ecx,TOS(2)
	jz .r		;?; 1^m
.p:
	call nndup
.l:
	call nnover
	call nmul
	call ntoq
	dec dword[esp]
	jg .l
	call nnswap
	jmp ndrop
.e:
	call ndrop
	lea esp,[esp+4]
.z:
	spush
	mov eax,0
	spush
	inc eax
.r:
	ret
.u:
	lea esp,[esp+4]
	inc eax		; rst count
	ret
;-
;- n/	  	( .d.D .s.S -- .q.Q )		(4)		'nby'		bignum
;-	signed division of dividend .d.D by divisor .s.S, leaves quot .q.Q.
nby:GA;:ok:
	call nbym
	call nnswap
	jmp ndrop
;-
;- n/mod  	( .dD .sS -- .rR .qQ )		(4)		'nbym'		bignum
;-	signed division of dividend .dD by divisor .sS, floored rounding mode
;-	leaves quot .qQ, remainder .rR
;-	fast trivial cases div +/- 1 and, if .sS>.dD ret sgn(.s)*.dD).
;-	div 0: 0/0 ret 0,1; non-zero/0 ret 0,sgn(.d)*.max(D+S).
nbym:GA;:ok:
; :sign: +/+ > +,+ ; +/- > -,+; -/- > +,- ; -/+ > -,- ::; (dvs/dvd > qot,rmd)
;:FEHLER:	   ;	   -,-;		  ;	  -,+
;  n-> ?n0=
	call ntoq
	mov ecx,eax	; sz(dvs)
	mov edx,TOS(1)	; sg(dvs)
	dec ecx
	jnz .g		;?; dvs w/ more than one cell
	test edx,edx
	jnz .z		;?; dvs =/= 0.1 = 0
;  if >rr ndrop 						\ div 0
	spop 2		; drop zero dvs
;   ?n0=
	call qnzeq
;    if rdrop ndrop 1 1 					\ 0/0
	test eax,eax
	spop
	jz .x		;?; non-zero dvd
.r:; dvs=dvd: ret 0.1 1.1
	call ndrop	; drop dvd
	spush
	xor eax,eax
	spush		; rmd:=0
	inc eax
	spush 2		; qot:=1
	ret
.q:; |dvs|=|dvd|, ecx=dvs x dvd, edx=
	call ndrop	; drop dvd
	jr .r		; drop dvs, ret rmd=0.1,qot=1.1
.x:; ret rmd=0.1, qot=sg(dvd) * maxnum, size = sz(dvd)+1
;    else r> nmx endif 						\ n/0
;	: nmx >r >rr ndrop 0 1 -1 1 r> r> + n->n sdrop -0 1- swap ;
	lea ecx,[eax+1]
	push dword TOS(1); sg(dvd)
	push ecx	; size+1
	call .r		; drop dvd, get 0.1, 1.1
	spush
	cmp dword[esp+4],byte 0; sg(dvd)
	pop eax		; size
	js .1
	neg dword TOS(2); 0.1 -1.1
.1:
	call nton	; 0 -1.D
	pop ecx		; sg(dvs)
	sar ecx,31
	not ecx
	lea ecx,[ecx*2+0x80000001]; 7f..ff or 80..01
	mov cl,ch	; 7f..ff or 80..00
	mov TOS(1),ecx	; +/- max.
	ret
;   exit
;  endif
.z:A; sz(dvs) = 1, edx = dvs =/= 0
;  nndup n1- n0= if n1- nnswap exit endif		\ div 1
	dec edx		; (single cell number)
	jnz .m		;?; dvs =/= 1
.p:
	mov TOS(1),edx
	jmp nnswap
;  nndup n1+ n0= if n1+ nnswap nminus exit endif	\ div -1
.m:A
	add edx,byte 2
	jnz .g		;?; dvs =/= -1
.n:;	<-ve dvs=dvd, ret -1,0
	call .p
	jmp nminus	; -ve qot
.g:A;<; sz(dvs) =/= 1, dvs =/= 0 and =/= +/- 1
;  nnswap n-> ?n0= if nd0 exit endif			\ 0 div n
	call nnswap
	call ntoq	; reduce dvd
	call qnzeq
	test eax,eax
	spop
	jz .o		;?; dvd =/= 0; operate...
; 2 nndrop 0 1 2dup ;					\ drop, ret 0
	call .q		; 2 nndrop 0 1
	dec dword TOS(1); ret 0, 0
	ret
.d:;<; dvs > dvd, ret dvd,0 w/ sg(rmd) = sg(dvd)
	push edx	; sg(dvd)
	call ndrop
	spush
	pop eax
	call npm	; sg(rmd) = sg(dvd)
	spush
	mov eax,0
	spush
	inc eax
	ret
.o:A;<
;  over >r nabs nnswap over >r nabs			\ R: s(s) s(d), sign-cells dvs, dvd
	push dword TOS(1); sg(dvd)
	call nabs	; |dvd|
	call nnswap
	push dword TOS(1); sg(dvs)
	call nabs	; |dvs|
;  ?ncmp 1+ case: nd0 nd1 false ; -dup if exit endif	\ z<n, z=n
	call qncmp	; normalize, compare
	spop
	pop edx		; dvS
	pop ecx		; dvD
	js .d		;?; dvd<dvs, ret 0,dvd
	jnz .c		;?; continue
	xor ecx,edx	; c: dvd x dvs, d: dvd
	jns disp .q	;?; +dvs=dvd	ret 1,0
	call .q
	neg dword TOS(1); ret -1,0
	ret
.c:A
	push ecx	; sg(dvs)
	push edx	; sg(dvd) => sg(rmd)
	push_ibyte(1)	; shift dvs up as long as not greater than dvd
;  0 >r							\ R: b, bitcount till remainder
	jr .u
;  enter begin r> 1+ >r n2* n->				\ shift dvs 'up' until m.s.bit
.b:A
	inc dword[esp]	; count bit(s)
	call nn2mul
	call ntoq
;  entry ?ncmp 0> 0= until				\ at same (or next l.s.) posn
.u:
	call qncmp
	spop
	jg .b		; more bits
;  ?ncmp 0< if n2/ endif  				\ c(n) := c(z)
	jz .t
	call nn2div
	dec dword[esp]
.t:
;  nnorm >rr						\ R: c:cells count, b:qot bitcount
	push eax	; dvd cells count
;  0 1 nn-rot						\ q
	spush
	xor eax,eax
	spush
	inc eax		; intial quot = 0.1 (single cell zero)
	call nnmrot
;  enter begin						\ q z n'
	jr .e
.v:
;   nnrot n2* r n->n nn-rot 				\ q:=2q, rst count
	call nnrot
	call nn2mul
	spush
	mov eax,[esp]	; rst dvd size
	call nton
	call nnmrot
;   ?ncmp 0< 0=
	call qncmp
	spop
	js .h
;    if
;     nndup nn>r >rr n- r> n->n nnr> 			\ z:=z-n'
	call nndup
	call nntor
	push eax
	call nsub
	spush
	pop eax
	call nton
	call nnrfrom
;     nnrot n1+ nn-rot
	call nnrot
	call nonep
	call nnmrot
;    endif
.h:
;    n2/						\ n:=n/2
	call nn2div
;  entry
.e:
;   r> 
;   r 0< 0= if -1 r+! endif				\ continue w/ remainder
	dec dword[esp+4]
.i:
;   r swap >r
;  0< until
	jns .v
;  2rdrop						\ drop c, b
	lea esp,[esp+8]
;  ndrop
	call ndrop
;  n-> r ( sg dvs ) n+-
	call ntoq
	spush
	mov eax,[esp+4]	; sg(dvd)
	call npm	; sg(rmd) = sg(dvd)
;  nnswap n-> r> r> xor n+- ;
	call nnswap
	call ntoq
.s:
	spush
	pop eax		; sg(dvs)
	pop ecx		; sg(dvd)
	xor eax,ecx
	jmp npm		; sg(qot) = sg(dvs x dvd)
%if 0
    mov ecx,AC_NL<<24|':I:'
    call .xx
    mov ecx,':PM:'
    call .xx
; --------------------- ;
.xx:
    spush
    mov eax,ecx
    call emit
    call nndup
    call ndot
    call dots
    call key
    cmp eax,27
    spop 
    jz near quit
    jmp cr
%endif
;-
;- ngcd		( n.N m.M -- o.O )			(4)		'ngcd'		bignum
;-	returns the greatest common denominator o.O of the absolute value of two signed 
;-	bignums. If either number is zero, returns the absolute value of the other one.
ngcd:GA;:ok:
; nabs nnswap nabs enter begin nnswap nnover n/mod ndrop entry ?n0= until ndrop ;
;{ 1 0 2004189184 4871 -401362253 206 501077504 76 2dup 3 2pick udm* 2rot 5 2roll md* 4rot md* q+ 4swap 2q->2n ngcd } = 10752
	call nabs	; abs
	call nnswap	; swap
	call nabs	; abs
	jr .e		; ENTER
.b:;			; BEGIN
	call nnswap
	call nnover	;  sover
	call nbym
	call ndrop	;  mod
.e:;			; ENTRY
	call qnzeq	;  dup n0=
	test eax,eax
	spop
	jz .b		; until
	jmp ndrop	; ndrop ;s
;-
;- nsqrt  	( .n.N -- .r.R j )		(4)				bignum
;-	+ve square root .r.R of signed counted integer .n.N and j=0,-1 if x<0.
;-	.m.M is nearest integer the sqare of which is lower or equal |.x.X|*(2j-1).
nsqrt:GA;:ok:
;input x, p := 0, a := 0
;loop 1 to #bits(x)/2
;	shift left x into a 2 places
;	shift left p one place
;	if a > 2*p then a:=a-(2*p+1); p:=p+1
;end
;return p
	call ntoq		;  n->
	lea ecx,[eax*8]
	shl ecx,1		;  bitcount/2
	push dword TOS(1)	;  over >r nabs		\ save signed, top cell for "j"
	push ecx		;  >rr 			\ bitcount X*16 for do..loop
	call nabs	;
	spush
	xor eax,eax		;  0 1 			\ a
	spush
	inc eax
	spush
	dec eax			;  0 1 			\ p
	spush
	inc eax
.l:		;  r> 16* 0 do	\ do bitcount/2 loops
	call nnrot		; nnrot
	spush		
	mov eax,2	
	call nlshift		; 2 nlshift		\ x:= x * 4
	dec eax			; 1-
	push dword TOS(1)
	TOSPULL(1)		; swap >r 
	call nnrot		; nnrot
	spush
	mov eax,2
	call nlshift		; 2 nlshift		\ a := a * 4
	spush
	pop eax			; r>
	spush
	mov eax,1		; 1
	call nadd		; n+			\ a:= a + ovf-bits
	call nnrot		; nnrot
	call nn2mul		; nn2*			\ p:= p * 2
	call ntoq		; n->
	call nn2dup		; n2dup
	call nn2mul		; nn2*
	call ncmp		; ncmp			\ ( x a p' a 2p -- f
	spop
	jng .e	;   0> if	\ a > 2p
	call nndup
	call nn2mul		; nndup nn2*
	call ntoq		; n->
	call ninvert		; ninvert		\ ( x a p -(2p+1)
	call nnrot
	call nadd		; nnrot n+		\ ( x p a'
	call nnswap
	call nonep		; nnswap n1+		\ ( x a' p+1
.e:			;   endif
	dec dword[esp]
	jg disp .l		;  loop
	call nnswap
	call nnrot
	spush
	lea esp,[esp+4]
	mov eax,2
	call nndrop
	spush
	pop eax			;  r> 0<		\ ( x -- p j )
	sar eax,31		; "j"
	ret
;-
;- --
;- counted integer stg/numeric conversion
;-	

;-
;- n#max  	( -- n )					'nhmax'		bignum
;-	max cells the size of which a counted integer can be converted
;-	to numeric string, at current {base} and memory setup.
;-	range can be extended after {allocate} and storeing new data-top
;-	to vari {n-top}, e.g.
;-		20000 allocate 
;-		0= if dup heaphead-size - @ dup d-top +! n-top +! endif drop
nhmax:GA;:ok:
	call pad
	call ntopf
	neg eax		; -pad
	add eax,edx	; d-top @ +
	sub eax,[byte uref uv.padsz]
	dec eax
	shr eax,2
	_mov ecx,0
	mov edx,ecx
	bsr ecx,[byte uref uv.base]
	mul ecx
.s:
	shrd eax,edx,6
	ret
ntopf:A
	mov ecx,[kref ntopv-Y]
	add ecx,databot
	mov edx,[kref ecx]
	test edx,edx
	jz .v
	ret
.v:A
	mov edx,[byte uref uv.dtop]
	mov [kref ecx],edx
	ret
;-
;- n#size 	( N -- N n )					'nhsize'	bignum
;-	no. of address units (bytes) required for stg conversion
;-	of N cells counted integer
nhsize:GA;:ok:
	spush
	_mov edx,0
	mov ecx,edx
	shld edx,eax,5
	shl eax,5
	bsr ecx,[byte uref uv.base]
	div ecx
	lea eax,[eax*4+4]
	ret
;-
;- n#?		( N -- N p flg )				'nhq'		bignum
;-	check whether counted int of size N can be converted to stg.
;-	flg=0 if unsufficient space for conversion at current radix
;-	else p = base (top down) address for multi-cell numeric stg.
;-	all of {pad}, {pad2} and {here}/{hld} storage protected.
nhq:GA;:ok:
	call nhsize
	call pad
	add eax,TOS(1)
	add eax,[byte uref uv.padsz]
	mov TOS(1),eax
	call ntopf
	cmp eax,edx
	sbb eax,eax
	ret
;-
;- n<#		( .nn. N -- .nn. N | abort-2 )			'nlth'		bignum
;-	check and set-up buf. on top of {pad} area, 
;-	use cells count adjusted to minimal size. 
;-	abort if unsufficient space for current conversion radix. 
nlth:GA;:ok:
	call ntoq
	call nhq
	not eax
	spush
	_mov eax,2
	call qerror
	mov [byte uref uv.hld],eax
	mov [byte uref uv.hld0],eax
	spop
	ret
;-
;- n#  		( .n1. N -- .n2. N )		(4)		'nh'		bignum
;-	append l.s. digit to number stg output storage
nh:GA;:ok:
	spush
	mov eax,[byte uref uv.base]
	call unsbym
	call h2a
	jmp hold
;-
;- n#s		( .n1. N -- .n2. N )		(4)		'nhs'		bignum
;-	convert all significant digits of a multi-cell integer 
;-	to stg, leave .n2. = 0
nhs:GA;:ok:
	call nh
	call ntoq
	call qnzeq
	test eax,eax
	spop
	jz nhs
	ret
;- (un.)  	( .n1. N -- p u )		(4)		'pundotp'	bignum
;-	convert unsigned multi-cell integer to numeric stg
pundotp:GA;:ok:
	push_ibyte(0)
	jr pndotp.u
;-
;- (n.)		( .n1. N -- p u )		(4)		'pndotp'	bignum
;-	convert signed multi-cell integer to numeric stg
pndotp:GA;:ok:
	push dword TOS(1)
	call nabs
.u:
	call nlth
	call nhs
	spush
	pop eax
	call asign	;:continues: in <nhgt>
;-
;- n#>		( .n1. N -- p u )		(4)		'nhgt'		bignum
;-	terminate multi-cell number conversion, leave resultant stg (p,u)
nhgt:GA;:ok:
	call ndrop
	spush
	mov eax,[byte uref uv.hld]
	spush
	mov eax,[byte uref uv.hld0]
	sub eax,TOS(1)
	ret
;-
;- n.		( .n1. N -- )			(4)		'ndot'		bignum
;-	display signed multi-cell integer as numeric stg
ndot:GA;:ok:
	call mone	;:continues: in <ndotr>
;-
;- n.r		( .n1. N n2 -- )		(4)		'ndotr'		bignum
;-	right aligned to field w. n2 chars of width 
;-	display signed multi-cell integer as numeric stg
ndotr:GA;:ok:
	call tor
	call pndotp
	call rfrom
.r:A; r 1abs bl.r type r> 0< 0= -exit space ;
	push eax
	call abs1
	call bldotr
	call type
	pop ecx
	test ecx,ecx
	js near space
	ret
;- un.		( .n1. N -- )			(4)		'undot'		bignum
;-	display unsigned multi-cell integer as numeric stg
undot:GA;:ok:
	call mone	;:continues: in <undotr>
;-
;- un.r		( .n1. N n2 -- )		(4)		'undotr'	bignum
;-	right aligned to field w. n2 chars of width 
;-	display unsigned multi-cell integer as numeric stg
undotr:GA;:ok:
	call tor
	call pundotp
	call rfrom
	jr ndotr.r
;-
;- n>number	( nx1 N1 p1 u1 -- nx2 N2 p2 u2 )		'ntonumber'	bignum
;-	basic multi-cell string to number conversion routine:
;-	accumulate to (nx1,N1) from numeric chars at stg(p1,u1)
;-	until 1st non-numeric char encountered or, at most u1 chars.
;-	u2 = 0 if successful, else p2 ptr to 1st invalid char code.
;-	cells count N2 adjusted to fit.
ntonumber:GA;:ok:
	test eax,eax
	jng .r		;?; no buf. space
	push eax
	spop		; ( nx N 	{ u
.l:
	spush
	movzx eax,byte[kref eax]
	call digit
	test eax,eax
	spop
	jz .x		;?; not a digit
	push eax	; save digit
	spop
	push eax	; save buf
	mov eax,[byte uref uv.base]
	call unsmul
	spush
	mov eax,[esp+4]	; digit
	spush
	_mov eax,1
	call nadd	; n+
	cmp dword TOS(1),byte 0
	jnz .d
	TOSPULL(1)
	dec eax
.d:
	spush
	pop eax		; p'
	add esp,4	; digit
	inc eax
	dec dword[esp]	; u'
	jg .l
	spush
.x:
	pop eax		; rmg count
.r:
	ret
;
; <[annumber]>	( -- .xx. X fl )		(k)		'abnnumberb'	
;	bignum from input stream; entry from ans-voc {[number]}, w/ @dpl = -1.
abnnumberb:A
	and dword[byte uref uv.bits],~(mbign)	;:continues:
;-
;- nnumber	( ptr +u -- .nn. N flg )					bignum
;-	convert string (ptr,+u) to multi-cell number w/ as much 
;-	cells as required to fully represent the string value.
;-	store last dp posn OR-ed w. 0xff800000 to dpl,
;-	leading number radix chars valid as w/ standard words.
;-	valid chars for fractional part marker are ',.'.
;-	leaves flg = 0 after success, else any number =/= 0.
;-	input cannot be a '-ve zero'.
nnumber:GA;:ok:
	mov dword[byte uref uv.dpl],-1
	push dword[byte uref uv.base]
	spush
	xor eax,eax
	spush
	inc eax
	call swap2	; 0x800000 dpl ! 0 1 2swap
	call abase	; base@ >r a-base dup w->s base! >r
	test ax,ax
	push eax	; sign
	movzx eax,ax
	jz .b
	mov [byte uref uv.base],eax; local radix
.b:
	spop
.l:
	call ntonumber	; n>number
	test eax,eax	; dup
	jz .x		;?;IF
	mov ecx,TOS(1)
	mov cl,[byte kref ecx]; over 1 s" ,.-/" 2swap search
    %if 1
	cmp cl,','
	jz .p
	cmp cl,'.'
	jnz .x
    %else; alternatively:
; -	valid chars for fractional part marker are ',.-/'.
; -	multiple sign chars permitted, each of which which negate the result.
	mov edx,[kref nfmarkt-Y]; (re at top of file)
	cmp cl,dl	; drop c@ [char] - = if r> negate >r endif
	jz .p
	cmp cl,dh
	jz .p
	bswap edx
	cmp cl,dl
	jz .p
	cmp cl,dh
	jnz .x
	neg dword[esp]	; negate sign flag
    %endif
.p:
	inc dword TOS(1)
	dec eax		; 1- dup dpl w! swap 1+ swap 0 
	mov [byte uref uv.dpl],ax
	jg .l		; else -rot 2drop over endif	;else over endif
.x:; ( ... --  .n. N ptr cnt; prevent false -ve 
	mov ecx,TOS(2)	; N
	mov edx,TOS(1)	; ptr
	spush
	inc ecx
	mov TOS(1),edx	; expand to 'unsigned'
	mov TOS(2),ecx	; N+1
	mov TOS(3),dword 0; top cell of integer
	cmp dword[esp],byte 0
	mov [esp],eax
	jns .r		;?; +ve
	push dword TOS(1)
	spop 2
	call nminus	; ret -ve bignum
	spush
	pop eax
	spush
	mov eax,[esp]
.r:
	lea esp,[esp+4]	; drop sign
	test eax,eax
	jnz .q		;?; failure
	and dword[byte uref uv.dpl],0xff80ffff
	xor ecx,ecx	; flag success
.q:
	pop ecx
	mov [byte uref uv.base],ecx; r> base!
	TOSPULL(1)	;  sdrop ;s
	ret
;-
;- ninteger	ccc(C: .nn. N -- )(X: -- .nn. N )				bignum
;-	counted integer alterable cons; {to}, {+to} NOT applicable.
;-	can be modified, e.g. with { ( .nn. N -- ) ' name >body n! }.
;-	data @body: [ N ][ .nn. ]
ninteger:GA;:ok:
	call create
	mov ecx,dataptr
	spush 2
	push ecx
	lea eax,[eax*4+4]
	call allot
	pop eax
	call nstore
	call pscodep
doncons:
	jmp nfetch
;-
;- nliteral	(C: .x. N -- )(X: -- .x. N )		(i)			bignum
;-	cpl: compile bignum (counted integer) as a literal, interp: noop
nliteral:GA;:ok:
	call ccplq
	jnz near c_nlit
	ret
%endif;__BIGNUM_I
;-
;-================================================================
;- l4bignum.inc <eof>
