;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4block.inc
;-  created	: 14-jul-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __BLOCK_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
%ifndef __BLOCK_I
%define __BLOCK_I
%define blockinc
;-	
;-	

;-
;-----------------------------------------------------------------
;- -- blockfile (screenfile) support
;-

;-
;- -- cons
;-

; {block-lk} @ => buffer-link @ bufhead-size + => datafield of block numbers & ptrs
;	[ owner fd ][ no. of blocks ][ blkptr1-* ]...[ blkptr4-* ]
; 	top bit of blk# is update-flag, block is no. of b/blk contiguos bytes from file

;-
;- b/b		( -- u )			(c)		'bbyb'		system
;-	bits per byte, "address unit" (8, max 32)
bbyb:GA;:ok:;cons
	call dobcons
	db b_b
;-
;- b/c		( -- u )			(c)		'bbyc'		editor
;-	bytes per character (1, max 4 bytes and max 32 bits)
bbyc:GA;:ok:;cons
	call dobcons
	db b_c
;-
;- c/l		( -- u )			(fg)		'cbyl'		fig
;-	characters per editor/load line (64)
cbyl:GA;:ok:;cons
	call dobcons
	db c_l
;-
;- b/scr  	( -- u )			(c)(fg)		'bbyscr'	fig
;-	blocks per editor/load screen (1)
bbyscr:GA;:ok:;cons
	call dobcons
	db b_scr
;-
;- b/buf  	( -- u )			(c)		'bbybuf'	forth
;-	bytes per block-buffer (1024 per ANS)
bbybuf:GA;:ok:;cons
	call doscons
	dd b_buf
;-
;- c/scr  	( -- u )			(kv)		'cbys'		editor
;-	chars per blockfile-editor screen (1024).
cbys:A
; o: ecx := c/scr
	mov ecx,[kref cbysv-Y]
	add ecx,databot
	mov ecx,[kref ecx]
	ret
;-
;- -- block access
;-

prev.n:; ;<nextbuf)
	call prev.z
	call use.z
	lea esp,[esp+4]
	spop
	call ibbuf	;:continued: <prev>
;-
;- prev		( -- p )			(fg)				fig
;-	most recently used block buf, i.e. block which contains latest data 
;-	of at time of last access valid owner, -re- {scr-chan}, {use}, {block}.
;-	if unset (zero) defaults to 1st unused, new block if none exists.
;-	can forcedly be re-fetched w/ { 0 to [prev] }.
prev:GA;:ok:
	mov ecx,prevv-Y
	call pdoval
	test eax,eax
	jnz .r
.b:
	mov eax,-1
	call blk2head	; 1st unused
	test eax,eax
	jnz .p		;?; found
	call ibbuf	; add new block
	jr .b		; ret 1st new buf
.p:
	call head2buf	
.z:
	mov ecx,[kref prevv-Y]
	add ecx,databot
	mov [kref ecx],eax; new [prev] default
.r:
use.r:
	ret
;-
;- use		( -- p )			(fg)				fig
;-	least recently used block buffer, i.e. block buf. to be used next.
;-	if {[use]} unset (zero), defaults to last buffer in chain of blocks.
;-	actual {use} represents a block of scr-chan valid immediatly before 
;-	{prev} assignment, used by {block} only if owner channel un-changed.
;-	can forcedly be re-fetched w/ { 0 to [use] }.
use:GA;:ok:
	mov ecx,usev-Y
	call pdoval	; [use]
	test eax,eax
	jnz .r
; initiate {[use]} to last block in chain:
	call prev
.l:; <block)
	mov TOS(1),eax
	call plusbuf
	test eax,eax
	spop
	jnz .l
	spop
.z:
	mov ecx,[kref usev-Y]
	add ecx,databot
	mov [kref ecx],eax
first.r:
	ret
;-
;- first  	( -- p )			(fg)				fig
;-	bottom address of 1st block buffer in buffer-block of -re- {prev}, 
;-	defaults to 1st buffer in chain of buffer-blocks if [prev] unset (zero).
;-	can forcedly be re-fetched w/ { 0 to [first] }.
;-	NOTE: different value if {prev} moved to a different block of buffers.
;-	provided for minimal compatibility wrt f.i.g.-4th blockfile access, 
;-	pointing to the 1st byte of the 1st, currently used block buffer.
;-	{first} is NOT necessarily THE BEGINNING of assigned chained buffer 
;-	blocks, the base address of which can be found with { block-lk @  }.
;-	-re- {ibbuf} in {blkfile} voc.
first:GA;:ok:
	mov ecx,firstv-Y
	call pdoval		; [first]
	test eax,eax
	jnz .r			;?; 2nd-ary access
; initiate {[first]}
	push ecx
	spop
	call prev
	call buf2lnk
	add eax,blkbuf.f	; 1st buffer in block
	call head2buf
	pop ecx
	mov [kref ecx],eax
limit.r:
	ret
;-
;- limit  	( -- p )			(fg)				fig
;-	next to top address of last block buffer in buffer-block of -re- {prev}.
;-	can forcedly be re-fetched w/ { 0 to [limit] }.
;-	NOTE: different value if {prev} moved to a different block of buffers.
;-	provided for minimal compatibility wrt f.i.g.-4th blockfile access, 
;-	pointing to the next after last byte of currently used block buffer.
;-	{limit} is most probably NOT THE LIMIT of assigned chained buffer blocks 
;-	which, at any time may be extended to any convenient amount and, cannot 
;-	be assumed to represent a contiguous data-space, wrt eachother.
;-	-re- {ibbuf} in {blkfile} voc.
limit:GA;:ok:
	mov ecx,limitv-Y
	call pdoval		; {[limit]}
	test eax,eax
	jnz .r			;?; already initiated
	push ecx
	call first
	call buf2lnk
	mov ecx,[kref eax+blkbuf.b]; count of buffers
	lea eax,[eax+ecx*blkhead_size+blkbuf.f-blkhead_size+blkhead.d]
	add eax,[kref eax]	; last buffer in block
	lea eax,[eax+b_buf]	; after end of buffer space
	pop ecx
	TOSPULL(1)
	mov [kref ecx],eax
	ret
;-
;- block-lk	( -- p )			(kv)				blkfile
;-	block buffer link base, ptr to header of 1st block of buffers
blocklk:KV; i: -/- ; o: edx ;<limit, first)
;-
;- blk>lnk	( n -- p )					'blk2lnk'	blkfile
;-	buffer-block link-address p of block-buffer of block no. n.
blk2lnk:GA;:ok:
	call blk2head
	jz .r
	mov eax,ecx
.r:
	ret
;-
;- buf>lnk	( p1 -- p2 )					'buf2lnk'	blkfile
;-	buffer-block link-address p2 of block-buffer at p1.
;-	can be used to checking whether a page-aligned ptr belongs to a 
;-	block of screenfile buffers, p2 := 0 if not.
buf2lnk:GA;:ok:
	call buf2head
	jz .r
	mov eax,ecx
.r:
	ret
;-
;- lnk>num	( p -- n )			(m,)		'lnk2num'	blkfile
;-	ret count n of buffers by buffer block link-ptr p -re- {blk>lnk>}, {buf>lnk}.
lnk2num:GA M;:ok:
	mov eax,[kref eax+blkbuf.b]
M lnk2num
	ret
;-
;- lnk>chn	( p -- n )			(m,)		'lnk2chn'	blkfile
;-	ret owner channel n of buffer block by link-ptr p -re- {blk>lnk>}, {buf>lnk}.
lnk2chn:GA M;:ok:
	mov eax,[kref eax+blkbuf.c]
M lnk2chn
	ret
;-
;- blk>buf	( n -- p )					'blk2buf'	blkfile
;-	buffer address p of block no. n.
;-	p := 0 if block n not buffered.
blk2buf:GA;:ok:;<bblockb)
	call blk2head	;:continued: head>buf
;-
;- head>buf	( p1 -- p2 )			(m,)		'head2buf'	blkfile
;-	convert ptr p1 to header to ptr to the rsp screenfile buffer. 
;-	if p1=0 ret p2 := 0, if word not compiled as a macro.
head2buf:GA M;:ok:
	test eax,eax
	jz .r
M,head2buf
	lea eax,[eax+blkhead.d]
	add eax,[kref eax]
M head2buf
.r:
	ret
;-
;- buf>blk	( p -- n )					'buf2blk'	blkfile
;-	block no. n. by ptr p to buffer. n := -1 if buffer not used.
buf2blk:GA;:ok:
	call buf2head	;:continued: head>blk
;-
;- head>blk	( p -- n )			(m,)		'head2blk'	blkfile
;-	convert ptr p to buffer header to block no. assigned to the rsp screenfile buffer. 
;-	n := -1 if buffer not used, if word not compiled as a macro.
head2blk:GA M;:ok:
	test eax,eax
	jz .r
M,head2blk
	mov eax,[kref eax+blkhead.u]
M head2blk
	inc eax
.r:
	dec eax
	ret
;-
;- blk>head	( n -- p )					'blk2head'	blkfile
;-	ptr p to header of screenfile buffer to which file-block no. n is assigned.
;-	ret 0 if p1 not a screenfile buffer.
blk2head:GA;:ok:;<blk2buf)
	lea edx,[kref b2hblk-Y]
	call scrcc
	jr scanblk
;-
;- buf>head	( p1 -- p2 )					'buf2head'	blkfile
;-	ret ptr p2 to header of screenfile buffer at p1.
;-	ret 0 if p1 not a screenfile buffer.
buf2head:GA;:ok:;<bbufb,buf2lnk,buf2blk,rws,update,updateoff,updateq)
	lea edx,[kref b2hsub-Y]
.k:
	xor ecx,ecx	; ret 0 if no success
; common proc scanning the linked list of screenfile buffer blocks
;	( n1 -- n2 n3 )
; i:	eax	working data, passed at [esp+4]
;	ecx	ditto, passed in reg <esi>, signal for failure flag if =/= -1
;	edx	true address of working routine
; o:	eax	reg <ebp> from working routine or, after failure last block's 1st buffer header
;	ecx	reg <edx> ditto. or, 
;	edx	flag, 0 after failure else edx := (ebp-eax) =/= 0 from working proc
%ifdef __MK_DBG
	global scanblk:function,b2hsub:function,b2hblk:function
%endif
scanblk:A;:ok:
; @tor:
%define .ar	;0	; eax = input @tos	\ 1st @tor!
%define .ip	+4	; edx = initial block	 ) ...
%define .xc	+8	; working routine	/ top item dropped @exit
%define .kr	+12	; saved KR
	push esi,edi,ebp,ebx,edx; save 34th-regs, subroutine @edx
	mov edx,databot
	add edx,[kref blocklkv-Y]; block-lk		;042024
	mov esi,[kref edx]; disp to 1st linkfield
	add edx,esi	; ptr to 1st linkfield	;0436A0
	test esi,esi	; Z if no buffer block exists
	mov esi,ecx	; posn in buf.ptr field
	push edx,eax	;			[ @tos lk
	lea eax,[edx+blkbuf.f]
	jr .n
.b:; edx: buffer-chain, [esp]: item to search for, [esp+4]: 1st block in chain for loop termination
	add edx,[kref edx]; next block of buffers
	cmp edx,[esp .ip]; initial ptr
.n:;
	lea ebp,[edx+blkbuf.f]; ptr to block headers
	jz .q		;?; chain ends or, no block linked
	mov ecx,[kref edx+blkbuf.b]; no. of blocks in this buffer
.i:
	call [esp .xc]	; -re- examples <b2h...>	[:ald: segfault or early ret while single stepping w/ "n"]
	jz .x		;?; found
	mov ebx,[esp .kr]; rst KR
	mov eax,ebp	; previous buf
	dec ecx
	lea ebp,[ebp+blkhead_size]
	jg .i		;?; next buf
	jr .b		; next block
.q:; end of chain, item not found
	inc esi
	jz .x		;?; ret w/o flag
	mov eax,ebp	; force d=0
;;    PRNUM ebp,l,':B:'
.x:
	mov ecx,edx	; buffer block base
	mov edx,ebp	; buf. header
	lea esp,[esp .xc+4]; drop base ptr, p1, subroutine
	pop esi,edi,ebp,ebx
; a' = previous.B if end of chain w/o success, else a' from subroutine
; ret after no success may be deduced from:
;	a' =/= 1st.blk+blkbuf.f AND Cy set
; eax := ebp = found buf-head,
; ecx := block-header
; edx := disp to previous buf-head. (prevhead=eax+edx)
; {use} : eax := ebp = 1st buf-head, ecx := block-header, edx := disp eax to last buf-head.
	sub edx,eax	; d := B-a', 0 at end of chain if ecx=0 on entry
	jnz .r		;?; found or end data request
	mov eax,edx	; ret 0
.r:
	add eax,edx	; a := a'+B-a' = B; a' = -(d-B) = B-d = eax-edx, Z after end of chain
	ret
; the working routine @[esp+8] receives:
; i:	eax	ptr to previous buf-head, i.e. previous <ebp>
;	ebx	KR
;	ecx	count(er) of buffers per block.
;		inner loop counter, decremented after subroutine, 
;		ecx < 2 early terminates the buffers/block loop.
;	edx	ptr to head of current block of buffers
;	edi	-unchanged-
;	esi	working data, passed from ecx
;	ebp	ptr to header of buffer in a block
; 	[esp+4]	working data, passed from eax
;	[esp+8]	ptr to link of 1st block of buffers
; c:	eax,ebx(=KR),esi(=LP),edi(=PS),
;		may be modified, regularly
;	ecx,ebp(=UP)
;		may be modified if terminating (Z flag set on exit!)
; o:	ebp	result1, ptr to header of tested block-buffer
;	ecx	result2, ptr to header of the rsp block of buffers
;	edx	flag (edx:=edx-eax)
;	eflags	Z terminates the loop, else 
;		ebp,edx forced to frame's defaults at buffers' chain end
;	esi(=LP),edi(=PS)
;		preserved across loops
b2hblk:A;<blk>head)
; o: eflags	eax=arg @tos and Z if blk# found
	mov eax,[kref ebp+blkhead.u]; ptr to blk#
	cmp dword[esp+4],byte -1; "wildcard" arg.
	jz .u		;?; any owner's buffer may be valid
	and eax,~(mupdt); clr "update" bit
	cmp dword[kref edx+blkbuf.c],-1
	jz .u		;?; unused buffer block
	cmp esi,[kref edx+blkbuf.c]; check owner
	jnz .r		;?; other channel
.u:
	cmp eax,[esp+4]	; by <eax> passed item for comparison
.r:
	ret
b2hsub:A;<buf>head)
; o: eflags	eax=arg @tos and Z if buf found
	lea eax,[ebp+blkhead.d]; disp to buffer
	add eax,[kref eax]; ptr = disp + address
	cmp eax,[esp+4]	; by <eax> passed item for comparison
	ret
b2hpbuf:A;<+buf)
	lea eax,[ebp+blkhead.d]
	add eax,[kref eax]; ptr to buffer
	cmp edi,[esp+4]	; by <eax> passed item for comparison
	mov edi,eax
	ret
b2hflsh:A;<flush)
	mov ecx,[kref edx+blkbuf.c]
	test ecx,ecx	; whether unused
	js .r		;?; next buffer-block
	mov eax,ecx	; channel
	xor ecx,ecx	; terminate
.r:
	ret
b2hmty1:A;<empty-buffers, clr blocks)
	mov eax,[kref edx+blkbuf.c]
	cmp eax,[esp+4]
	jnz .r		;?; other owner or unused
	mov dword[kref ebp+blkhead.u],-1; mark block unused
	test edx,edx	; NZ
.r:
	ret
b2hmty2:A;<empty-buffers, clr owner)
	mov eax,[kref edx+blkbuf.c]
	cmp eax,[esp+4]
	jnz .r		;?; other owner or unused
	mov dword[kref edx+blkbuf.c],-1; mark buffer unused
	test edx,edx	; NZ
.r:
	ret
;-
;- +buf		( p1 -- p2 tf|ff )		(fg)(e)		'plusbuf'	fig
;-	find next buffer in chain of buffer-blocks, regardless of the resp owner.
;-	ff := 0 if p2 = prev, else any no-zero value for tf (flg := prev - p2 ).
;-	(e: Z if p2=prev)
plusbuf:GA;:ok:
;fig: b/buf 4+ + dup limit = if drop first endif dup prev @ - ;
	mov ecx,-1	; <esi> no end of chain flag
	lea edx,[kref b2hpbuf-Y]
	call scanblk
	call head2buf
	call prev
	sub eax,TOS(1)	; flg
	ret
;-
;- [upd]  	( -- )						'bupdb'		hidden
;-	mark most recently changed block, -re- {prev}, to saveing before re-useing.
;-	default opr of deferred {update}.
bupdb:GA;:ok:
	call prev
	call buf2head
	jz .r
	or dword[kref eax+blkhead.u],mupdt; bit#31 is 'update'-bit
.r:
	spop
	ret
;-
;- update  	( -- )				(a:b)(d)	'bupdb'		forth
;-	mark most recently changed block, -re- {prev}, to saveing before re-useing.
;-	default opr -re- {[updt]}.
update:GA;:ok:
	mov ecx,updtv-Y
	jmp pdodefer
;-
;- update-off  	( -- )				(a:b)		'updateoff'	blkfile
;-	clear -re- {update} flag of most recently changed block.
updateoff:GA;:ok:
	call prev
	call buf2head
	jz .r
	and dword[kref eax+blkhead.u],~(mupdt)
.r:
	spop
	ret
;-
;- update?  	( p -- f )					'updateq'	blkfile
;-	ret tf if screenfile buffer at p is in use and {update}d.
updateq:GA;:ok:
	call buf2head
	jz .r		;?; p not a screenfile buffer
	mov eax,[kref eax+blkhead.u]
	inc eax
	jz .r		;?; not used
	sar eax,31
.r:
	ret
;-
;- ro?		( p -- f )			n.i.		'roq'		blkfile
;-	ret tf if at p buffered screenfile is write-protected.
roq:GA
	call buf2lnk
	jz .r		;?; buffer block not used
	mov eax,[kref eax+blkbuf.c]
	call chid
	inc eax
	jz .r		;?; no fd in channel
	sar eax,31
.r:
	ret
;-
;- nextbuf	( p1 -- p2 )							blkfile
;-	find next unused or by {scr-chan} owned buffer =/= p1 or, install new block.
;-	always succeeds, otherwise fatal error: system out of memory
nextbuf:GA;:ok:
	push eax	; passed buffer
	call buf2lnk
	jz near prev.n	;?; no such buffer exists, ret w/ new buffer bloc
	mov ecx,[kref scrcnv-Y]
	add ecx,databot
	mov ecx,[kref ecx]
.o:
	mov eax,[esp]	; buf p1
	push ecx	; scr-chan
	jr .n		;		( p1 		[ ch p1
.b:
	call ibbuf
.p:
	spop		; drop block-header
.n:
	call plusbuf	; +buf
	jz .p		;?; exclude {prev}
	mov eax,TOS(1)	; drop dup	( p' p'		[ ch p1
	cmp eax,[esp+4]
	jz .b		;?; all checked, install new buffer
	call buf2lnk	; block-header
.i:
	lea ecx,[kref eax+blkbuf.c]
	mov eax,[ecx]	; channel no.
	cmp eax,byte -1
	mov edx,[esp]	; owner flag
	jz .u		;?; buffer block not used
	cmp eax,edx
	jnz .p		;?; not {scr-chan}-s own block
.u:
	mov [ecx],edx	; sto channel & flag(s), mark owner
.r:
	spop		; drop block-header
	lea esp,[esp+8]	; drop p1 & flag
	ret
;
r.ow:; ;< error 'of working area'
	jmp r.oa
;-
;- rw@		( p -- )					'rwf'		blkfile
;-	fetch assigned block to buffer p from file by buffer owner or {scr-chan}.
rwf:GA;:ok:
	lea ecx,[kref rws.d-Y]
	jr rws.n
;-
;- rw!		( p -- )			(s)		'rws'		blkfile
;-	clr {update} flag, save block from ptr p to owner file or, {scr-chan}.
;-	defaults to {scr-chan}, stored into block header if no channel assigned.
;-	syscall write, llseek -re- man (2), restricted to file size less than 4G.
rws:GA;:ok:<bbufb)
	spush
	call buf2head
	jz r.ow		;?; no such buffer
	mov ecx,[kref eax+blkhead.u]; block no.			[    bb | b S x
	and ecx,~(mupdt); clr update flag
	mov [kref eax+blkhead.u],ecx; block no.			[    bb | b S x
	lea ecx,[kref .w-Y]; write
	spop
.n:
	push ecx	; xec addr
	call bbybuf	; b/buf
	push esi,ebx,eax; save LP,KR, b/buf			[ b/buf | b S x
	mov eax,TOS(1)	; >r dup		{ bf bf		[ ...
	call buf2head	; buffer header
	jz r.ow		 ;?; buffer doesn't exist
	mov eax,[kref eax+blkhead.u]; block no.			[    bb | b S x
	mul dword[esp]
	jc r.ow		;?; fptr > 2^31
	push eax	; fptr := blk# * b/buf			[ fp bb | b S x
	mov eax,TOS(1)	; >r dup		{ bf bf		[ ...
	call buf2lnk	; buffers' block header
	mov ecx,eax
	mov eax,[kref eax+blkbuf.c]; owner channel
	test eax,eax
	jns .c		;?; owner channel set
	lea ecx,[kref ecx+blkbuf.c]
	mov eax,[kref scrcnv-Y]; {scr-chan}
	add eax,databot
	mov eax,[kref eax]
	mov [ecx],eax	; store dft into buffer block header
.c:
	and eax,cdt_num-1; mask flags off channel-no.
	spush		; dup			{ ch ch bf	[ ...
	mov ecx,eax	; r> swap 0 swap -> unsigned d.fpos & chid
	mov eax,[esp]	;			{ fp ch bf	[ fp bb | b S x
	spush
	xor eax,eax	; d.fposn		{ fp.fp ch bf	[ fp bb | b S x
	spush
	mov eax,ecx	; chan			{ ch fp.fp ch bf [ fp bb | b S x
	call reposnf
	spop		; drop ernum		{ ch bf		[ fp bb | b S x
	call chid	; edx := cdt		{ fd bf		[ fp bb | b S x
	mov ecx,TOS(1)
	add ecx,KR	; buffer
	add edx,KR	; true addr of cdt
	mov ebx,eax	; fd
	mov eax,edx	; cdt
	mov esi,[esp]	; fptr					[ fp bb | b S x
	mov edx,[esp+4]	; b/buf
	mov [esp],eax	; save cdt				[ ct bb | b S x
	mov [eax+cdt.fpos],esi; fptr to cdt
	mov dword[eax+cdt.fpos+4],0; clear hi cell, i.e. file size less than 4G
	call [esp+16]	;					[ ct bb | b S x
	pop ecx		; cdt
	pop edx		; b/buf
	test eax,eax
	js .r		;?; access error
	add dword[ecx+cdt.fpos],eax
.r:
	mov [ecx+cdt.rnum],eax; ernum to cdt
	mov [byte uref uv.ierr],eax	;;;
	pop esi,ebx	; saved regs
	lea esp,[esp+4]	; drop xec-addr
	spop 2		; drop buffer & ernum
	ret
.w:
	sys_write
	ret
.d:
	sys_read
	ret
.o:; dummy xec, for write protected screenfile
	xor eax,eax
	ret
;-
;- scr-chan	( -- n )			(kv)		'scrn'		forth
;-	kernel value which holds the current screenfile channel no.
scrcn:GA;:ok:<eding,embufs,flush,relip,truncate,using)
	spush
	mov ecx,[kref scrcnv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
scrcc:A; ;<bblockb,buf2head)
	mov ecx,[kref scrcnv-Y]
	add ecx,databot
	mov ecx,[kref ecx]
	ret
; -----------------------------	;

;-
;- [buffer]	( +u -- p )			(k)(83)		'bbufb'		hidden
;-	assign buffer at ptr p to block no. u of owner {scr-chan}; dft for {buffer} 
;-	beginning w/ {use}, find next buffer =/= {prev} to useing next, set new {use}; 
;-	ret previous {use} as new {prev}, content stored back to file if {update}ed.
bbufb:GA;:ok:
;fig: use @ dup >r begin +buf until use ! r @ 0< if r 2+ r @ 7fff and 0 r/w endif r ! r prev ! r> 2+ ;
	call use
	spush
	call buf2lnk
	call scrcc
	mov edx,eax
	cmp ecx,[kref eax+blkbuf.c]
	spop
	jz .p		;?; currently owned block
	cmp [kref edx+blkbuf.c],byte -1
	jz .p		; buffer in unused block
	call nextbuf	; get unused or own buffer
.p:
	push eax	; save for ret
	spush
	call updateq	; whether {update}d
	spop
	jz .b		;?; may be used
	spush
	call rws	; {  0 r/w }, save {use} buffer to block of owner file
.b:
; test whether unused or current screenfile channel @scr-chan,
; else try next, else install new block of buffers
; <nextbuf> is +buf in chain of own or unused blocks
	call nextbuf	; find next currently owned buffer or, install new block
; sto buf to {prev}, old {prev} to {use}
	mov edx,databot
	mov ecx,[kref usev-Y]
	add ecx,edx
	add edx,[kref prevv-Y]
	mov [kref ecx],eax; buf is new {use}
	mov eax,[esp]	; previous {use}
	mov [kref edx],eax; new {prev}
	call buf2head
	mov ecx,eax	; header
	spop
	mov [kref ecx+blkhead.u],eax; block# to header
	pop eax		; ret {use} buffer as {prev}
	ret
;-
;- [block]	( n -- p )			(k)(83)		'bblockb'	hidden
;-	ret ptr p =: {prev} to buffer for block n, assigned to current {scr-chan}.
;-	no. of fetched bytes or file access error can be read by { scr-chan ch-rnum }.
;-	either, finds already present block u of {scr-chan} or, assigns and fetches 
;-	block no. u into p = {use} or, new buffer of rsp scr-chan, re-setting {use}. 
;-	dft for -re- {block}.
bblockb:GA;:ok:
; offset @ + >r prev @ dup @ r - dup + if 
;  begin +buf 0= if drop r buffer dup r 1 r/w 2- endif dup @ r - dup + 0= until dup prev ! 
; endif rdrop 2+ ;
	spush		; dup
	call blk2lnk
	jz .c		; block not present
	call scrcc	; ecx:=scr-chan
	lea edx,[kref eax+blkbuf.c]
	cmp ecx,[edx]
	jz .r		; buffer exists
	cmp dword[edx],-1
	jnz .c		;?; other owner or updateflag
	mov [edx],ecx	; assign block to owner-channel
.r:
	spop
	call blk2buf	; drop dup, get block#
	jnz .s
	spush
.c:A
	mov eax,TOS(1)	; block#
	call buffer	; get a useable buffer, assigned to block u
	mov TOS(1),eax	; save buffer ptr
	call rwf	; rw@, fetch data into buffer
.s:
	mov ecx,[kref prevv-Y]
	add ecx,databot
	mov [kref ecx],eax; buf.ptr to {prev}
	ret
; [source]
; dpans A.6.1.2216 SOURCE : { BLK @  IF BLK @ BLOCK 1024  ELSE TIB #TIB @ THEN }
bsc.blk:A;<bsc.fp)
	mov eax,ecx	; block no.
	call block
	call bbybuf
	cmp dword TOS(1),0
	jnz .b
	xor eax,eax
.b:
	mov ecx,[byte uref uv.span]
	mov [uref uv.uintib],ecx
	ret
bsc.file:A;:ok:	; file input, ecx +ve is blockfile
	test ecx,ecx	; -ve: @blk is 1-s complement of file pointer for stdin
	jns bsc.blk	;?; +ve block no.
bsc.fd:; <[source] stdin is file)
	mov edx,eax	; save channel no.
	not ecx
	mov eax,ecx	; fposn
bsc.bl:; <block: eax=block-fpos, edx=chan)
; eax=fptr, edx=source-id
	and edx,cdt_num-1; stdin 0 <- ch_num(1<<cdt_bnum);
	spush
	xor eax,eax
	spush		; ud fposn
	mov eax,edx
	call reposnf	; reposition-file
	spush		; dup
	call qerror	; ?error
	spush
	jr bsourceb.e
;-	
;-
;- [source]	( -- p +u )			(k)(83)		'bsourceb'	hidden
;-	dft for {source}, ret input buffer(p,u). does not modify {in}, {span} &c.
;-	dpans A.6.1.2216 : { BLK @  IF BLK @ BLOCK 1024  ELSE TIB #TIB @ THEN } modified 
;-	such that -ve {BLK}@ is 1-s cpl of file ptr and notifies input from 'stream'-file 
;-	and, initial state of commandline args at program invocation by -ve @{SOURCE-ID}.
bsourceb:GA;:ok::preliminary:
	spush			; save tos
	mov eax,[byte uref uv.sourceid]	; -ve: evaluate, 0: kbd, +ve: (input channel no.|max-channels)
	mov ecx,[byte uref uv.blk]	; +ve: blockfile
	test eax,eax
	jg bsc.file		;?; channel id
	js .a			;?; job's arguments or evaluate :ok:
	test ecx,ecx		; -ve @blk is 1-s complement of file pointer
	js bsc.fd		;?; stdin from file
.t:A	;   0: tib, file or keyboard :ok:
	mov dword[byte uref uv.blk],0
	mov dword[byte uref uv.sourceid],0
.r:
	lea eax,[uref uv.utibot]
	sub eax,KR
	mov [byte uref uv.tib],eax
.e:; <end of args, evaluate)
	mov eax,[byte uref uv.tib]
	spush
	mov eax,[byte uref uv.span]
	mov [uref uv.uintib],eax
	ret
.a:;:ok:
	inc eax
	jz .e			;?; {evaluate}
	mov ecx,[uref uv.uargn]
	test ecx,ecx
	jz .t
	js .e			;?; stdin if no more args
	dec dword[uref uv.uargn]
	sub ecx,dword[uref uv.uargc]
	neg ecx
	mov eax,[uref uv.uargp]
	add eax,KR
	mov eax,[eax+ecx*4+4]
	mov ecx,eax
	dec ecx
.l:
	inc ecx
	cmp byte[ecx],BL	; any ctrl char
	jnb .l			;?; not yet at <eol>
	sub ecx,eax
	sub eax,KR
	spush
	mov [uref uv.uintib],ecx
	mov eax,ecx
	ret
;-
;- --
;-	

;-
;- dr0		( -- )						'dr0c'		forth
;-	activate 1st screenfile channel, -re- {dr1}, {dr?}
dr0c:GA;:ok:
; o: ecx	channel no.
	mov ecx,scr1ch
	jr dr1c.n
;-
;- dr1		( -- )						'dr1c'		forth
;-	activate 2nd screenfile channel, -re- {dr0}, {dr?}
dr1c:GA;:ok:
; o: ecx	channel no.
	mov ecx,scr2ch
.n:
	mov edx,[kref scrcnv-Y]      
	add edx,databot
	mov [kref edx],ecx
	ret
;-
;- dr?		( -- n )					'schq'		blkfile
;-	ret +ve {dr0}-channel, -ve {dr1}-chanel 
;-	or, zero if neither one is {scr-chan}.
schq:GA;:ok:
	spush
	mov ecx,[kref scrcnv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	cmp eax,scr1ch
	jz .r
	neg eax
	cmp eax,-(scr2ch)
	jz .r
	xor eax,eax
.r:
	ret
;-
;- empty-blk	( n -- )			(fg)		'emblk'		blkfile
;-	in-validate buffer of blk# n, set {use} to it's buffer, don't save.
;-	forces re-fetching blk# n from {scr-chan} file.
emblk:GA;:ok:
	call blk2head
	jz .r
	mov dword[kref eax+blkhead.u],-1
	call head2buf
	mov ecx,databot
	mov eax,TOS(1)
	add ecx,[kref usev-Y]
	mov ecx,eax
.r:
	spop
	ret
;-
;- empty-buffers ( -- )				(fg)		'embufs'	forth
;-	don't save, make all buffers of {scr-chan} available, 
;-	force re-fetching of buffered blocks from file.
embufs:GA;:ok:
	call scrcn
	spush
	lea edx,[kref b2hmty1-Y]
	call scanblk	; invalidate buffers
	mov eax,TOS(1)
	lea edx,[kref b2hmty2-Y]
	call scanblk	; clear channel entries
	spop 2
	ret
;-
;- CLEAR	( n -- )							editor
;-	fill screen no. n (=block) with blanks.
;-	can extend beyond <eof>, itermediate scr-s filled w/ <nul>-bytes.
clear:GA;:ok:
	call block
	call update
	call bbybuf
	call blanks	;:continued: <flush>
    global sbufs:function
;-
;- flush	( -- )								forth
;-	write all updated blocks back to the rsp. screenfile, 
;-	un-assign. sync file w/ cached content.
flush:GA;:ok:
;- save-buffers	( -- )				(a:b)		'sbufs'		forth
;-	write all updated blocks back to the rsp. screenfile, fig-4th "flush".
;-	(useless, same as -re- {flush} but, required for ANS-Forth)
sbufs:;:ok:
;fig: b/scr 1+ 0 do 0 buffer drop loop ;
	call prev
	spush
.l:
	mov eax,TOS(1)
	call updateq
	spop
	jz .p
	spush
	call rws
.p:
	call plusbuf
	jnz .l
	call scrcn
	call flushf	; sync file content
	spop 3
	ret
;-
;- reading	( -- )				(4)		'rding'		blkfile
;-	re-open {scr-chan} for non-blocking, read-only access at un-changed file ptr.
;-	rsp {block} cleared to be refetched; -re- {editting} for r/w mode.
rding:GA;:ok:
	push_ibyte(O_RDONLY)
	jr eding.n
;-
;- editting	( -- )				(4)		'eding'		blkfile
;-	re-open {scr-chan} for non-blocking, read/write access at un-changed file ptr.
;-	rsp {block} cleared to be refetched; -re- {reading} for r/o mode.
eding:GA;:ok:
	push_ibyte(O_RDWR)
.n:
	call scrf
	call emblk
	call pathmax
	push eax
	shr eax,2
	push eax
	push dword[byte uref uv.s0]
	call toloc	; PATH_MAX bytes of local buffer
	spush
	mov eax,[esp]	; p
	spush
	mov eax,[esp+8]	; u-max
	call scrcn
	call chname
	test eax,eax
	jz .q		;?; scr-chan not accessible
	call scrcn
	call fposf
	call scrcn
	call closef
	spop 2
	call swap2
	spush
	mov eax,[esp+12]
	call using.o
	call scrcn
	call reposnf
	pop ecx		; drop p
.x:
	pop eax		; locals
	lea esp,[esp+8]	; drop path-max, open-flag
	neg eax
	jmp toloc
.q:
	pop eax		; drop p
	call m.nf
	jr .x
;-
;- using	( p u -- )							forth
;-	pass screenfile name to use w/ subsequent {load} and editting opr.
;-	open {scr-chan} in non-blocking, read-only mode. -re- {editting}.
using:GA;:ok:
	call flush	; flush modified buffers
	call embufs
	call scrcn
	call closef	; close previous screen channel
	mov eax,O_RDONLY
.o:
	or eax,O_NONBLOCK; provisional for file locking
	call openf
	lea ecx,[eax+1]
	mov eax,TOS(1)
	mov TOS(1),ecx
	call qerror
	call scrcn
	test eax,eax
	jns .s		;?; re-use channel no.
	call dr0c
	mov eax,ecx	; @{dr0} = scr1-chan
.s:
	jmp ischan
;-
;- load		( n -- )			(d)				forth
;-	make file by {scr-chan} current input, beginning at screen no. n.
load:GA
	mov ecx,loadv-Y
	jmp pdodefer
;-
;- [load]	( n -- )							hidden
;-	make file by {scr-chan} current input, beginning at screen no. n.
;-	dft for deferred -re- {load}.
bloadb:GA;:ok:
	spush
	call block
	spush
	test eax,eax
	jz .r		;?; no block present
	call scrcc
	mov eax,ecx
	spush
	call chrnum
	test eax,eax
	jg .l		;?; load
.r:
	spop 3		; error or eof
	ret
.l:
	spush		; received count
	call intf
	xor eax,eax	; min. local tib
	call pushio
	mov edx,TOS(3)	; block#
	mov ecx,TOS(2)	; buffer
	mov [byte uref uv.span],eax
	mov [byte uref uv.tib],ecx
	mov [byte uref uv.blk],edx
	mov ecx,TOS(1)
	or ecx,cdt_num
	mov dword[uref uv.utibmax],b_buf
	mov [byte uref uv.sourceid],ecx
	mov dword[byte uref uv.ierr],0	; for {interpret} to terminating after a single block
	spop 4
	jmp bquitb.c
;-
;- -->		( -- )				(d)		'nxld'		root
;-	-re- {load} next screen. deferred, dft -re- {[-->]}.
;-	e.g. within a word defn, use {[-->]} for immediate execution.
nxld:GA
	mov ecx,nxldv-Y
	jmp pdodefer
;-
;- [-->]	( -- )				(i)		'bnxldb'	forth
;-	cpl from next screen. dft for deferred -re- {-->}
;-	using initial, {load}ing block-buffer for tib.
bnxldb:GA;:ok:
	spush
	mov eax,[byte uref uv.blk]
	inc eax
	call block
	test eax,eax
	jz .r		; block not present
	cmp dword[byte uref uv.ierr],0
	jng .r		; eof or error
	call scrcc
	or ecx,cdt_num
	mov [byte uref uv.tib],eax
	mov [byte uref uv.sourceid],ecx
	inc dword[byte uref uv.blk]
	mov dword[byte uref uv.span],b_buf
	mov dword[byte uref uv.in],0
.r:
	mov dword[byte uref uv.ierr],0
	spop
	ret
esemis:KV
;-
;- thru		( n1 n2 -- ) 			(4)(a:b)(n.i.)			forth
;-	-re- {load} screens n1 till n2 (inclusive, dpans 7.6.2.2280).
;-	NOTE: n.i, pse, use nestable -re- {load} and {-->}, instead.
thru:GA
;	jmp nimp
	inc eax
	call swap
	call pdop
.do:
	call r
	call load
	call ploopp
	dd .do-$
	ret
;
; ---
;

;-
;- buffer  	( +u -- p )			(v)(fg)				forth
;-	leave ptr to useable screenfile buf.,
;-	flush oldest buf. and re-assign if none free,
;-	else leave ptr to next free in list of #buf-s.
;-	execute true address uot vector (buffer), dft is {[buffer]}.
buffer:GA;:ok:;
	spush
	mov eax,[uref uv.ubuffer]
	jmp execute.c
;-
;- block  	( +u -- p )			(v)(fg)				forth
;-	ptr to block for sreenfile i/o
;-	execute true address uot vector (block), dft is {[block]}.
block:GA;:ok:;
	spush
	mov eax,[uref uv.ublock]
	jmp execute.c
;-
;-----------------------------------------------------------------
;- -- the f.i.g. line editor, -re- {editor} voc.
;-	

;- 
;- 

;-
;- r#@		( -- n )			(m,7)		'rnf'		editor
;-	content of uvari {r#}, the editor screen disp from top to cursor.
rnf:GA M;:ok:
	spush
	mov eax,[uref uv.unr]
M rnf
	ret
;-
;- scr@		( -- n )			(m,7)		'scrf'		editor
;-	content of uvari {scr}, the current editor screen no.
scrf:GA M;:ok:
	spush
	mov eax,[uref uv.uscr]
M scrf
	ret
;-
;- scr!		( n -- )			(m,11)		'scrs'		editor
;-	sto n to uvari {scr}, the current editor screen no.
scrs:GA M;:ok:
	mov [uref uv.uscr],eax
	spop
M scrf
	ret
;-
;- (line)  	( n1 n2 -- p u )				'plinep'	editor
;-	ret ptr to and characters count of line no. n1 in screen n2.
plinep:GA;:ok:
;b/buf * swap c/l * + b/buf /mod block + c/l ;s
	call bbybuf
	push eax
	call smul
	call swap
	call cbyl
	mov ecx,[esp]
	mov [esp],eax
	push ecx
	call smul
	add TOS(1),eax
	pop eax
	call bymod
	call block
	add TOS(1),eax
	pop eax
	ret
;-
;- .LINE  	( n -- )					'dotline'	editor
dotline:GA;:ok:
; (line) -trailing type ;
	spush
	mov eax,[uref uv.uscr]
	call plinep
	call mtrailing
	jmp type
;-
;- LINE   	( n -- p | abort )		(4)				editor
;-	relative to scr. leave address of line
line:GA;:ok:
;dup -16 and $17 ?error scr @ (line) drop ;
	spush
	and eax,byte -16
	spush
	mov eax,23
	call qerror
	spush
	mov eax,[uref uv.uscr]
	call plinep
	spop
	ret
;-
;- where  	( cur blk -- )			n.i.				editor
;-	while {load}ing, print screen # and image of error, sto
;-	@blk to uvari {scr}, make {editor}-voc current and quit.
where:GA
;dup b/scr /mod dup scr ! dup list ." scr # " decimal .
;c/l * r# ! swap c/l /mod c/l * rot block + cr c/l type cr
;dup r# +! here c@ - spaces 94 emit [compile] editor quit ;
	jmp nimp

;-
;- #locate	( -- n1 n2 )			(4)		'hlocate'	editor
;-	leave cursor offset n2, line n1
hlocate:GA;:ok:
; r# @ c/l /mod ;
	and dword[uref uv.unr],1023
	spush
	mov eax,[uref uv.unr]
	call cbyl
	jmp bymod
;-
;- #lead  	( -- p n )			(4)		'hlead'		editor
;-	line address p, offset n from beginning to cursor.
hlead:GA;:ok:
;#locate line swap ;
	call hlocate
	call line
	jmp swap
;-
;- #lag		( -- p u )					'hlag'		editor
;-	cursor address p, count u after cursor till eol.
hlag:GA;:ok:
;#lead dup >r + c/l r> - ;
	call hlead
	mov ecx,eax
	add TOS(1),eax
	mov eax,c_l
	sub eax,ecx
	ret
;-
;- TRUNCATE	( n -- )							editor
;-	truncate current screenfile at or, extend & <nul>-fill till block # n.
truncate:GA;:ok:
	spush
	call block
	call update
	mov eax,TOS(1)
	inc eax		; for fpos at end of scr
	call cbys	; ecx := c/scr
	mul ecx  	; eax := end of scr # n
	spush
	mov eax,edx
	call scrcn	; scr-chan
	call resizef	; resize-file
	spop 2		; drop ernum & block-ptr
	ret
;-
;- L+		( -- )						'relip'		editor
;-	display next screen, from {scr}.
relip:GA;:ok:
	call scrcn
	call cheof
	add [uref uv.uscr],eax	; don't count beyond last screen
	spop
	add dword[uref uv.uscr],byte 2;:continued: <relim>
;-
;- L-		( -- )						'relim'		editor
;-	display previous screen, from {scr}.
relim:GA;:ok:
	dec dword[uref uv.uscr]	;:continued: <relip>
;-
;- L		( -- )						'relist'	editor
;-	re-display latest screen, from {scr}.
relist:GA;:ok:
	spush
	mov eax,[uref uv.uscr]	;:continued: <list>
;-
;- list		( n -- )			(a:b)		'elist'		forth
;-	display screen n, store max(0,n) to uvari -re- {scr}.
elist:GA;:ok:
; 16 0 do i 3 .r i scr @ .line cr loop ;
	test eax,eax
	jns .p		;?; {0max}
	xor eax,eax
.p:
	mov [uref uv.uscr],eax
%if (b_scr=1)&&(b_buf=1024)&&(c_l=64);only if c/scr=b/buf=16*c/l :ok:
	call block
	jz .r		;?; no screenfile
	push_ibyte(-(b_buf/c_l))
.l:
	spush 3		; dup <c/l> <r>
	mov eax,[esp]
	add eax,(b_buf/c_l); r 16 +
	spush
	mov eax,~4	; field of 3 chars & trailing blank
	call dotr
	mov eax,c_l
	call type
	call cr
	lea eax,[eax+c_l]
	inc dword[esp]
	js .l
	lea esp,[esp+4]
.r:
	spop
	ret
%else
	push eax
	push_ibyte(-16)
	jr .n
.l:
	spush
.n:
	mov eax,[esp]
	add eax,byte 16
	spush 3
	mov eax,~4	; field of 3 chars & trailing blank
	call dotr
	mov eax,[esp+4]
	call plinep
	call type
	call cr
	inc dword[esp]
	js .l
	lea esp,[esp+8]
	ret
%endif
;-
;- I		( n -- )			(4)		'eI'		editor
;-	insert text from pad onto line # n.
;-	spread (S) and replace empty line n from pad (R)
eI:GA;:ok:
; dup s r ;
	spush
	call eSe
	jr eR
;-
;- P		( n -- )ccc			(4)		'eP'		editor
;-	put following text to line n
eP:GA;:ok:
; 1 text r ;
	spush
	mov eax,1
	call text	;:continued: <eR>
;-
;- R		( n -- )			(4)		'eR'		editor
;-	replace (overwrite) line no. ( n ) with text from pad
eR:GA;:ok:
;pad 1+ swap -move ;
	call pad
	inc eax
	call swap	;:continued: <mmove>
;-
;- -move  	( p n -- )					'mmove'		editor
;-	move c/l chars of text from ptr p to line n of current screen.
mmove:GA;:ok:
;line c/l cmove update ;
	call line
	call cbyl
	call xcmove
	jmp update
;-
;- H		( n -- )			(4)		'eH'		editor
;-	hold numbered line at pad. copy to pad
eH:GA;:ok:
; line pad 1+ c/l dup pad c! cmove ;
	call line
	call cbyl
	call pad
	jmp place
;-
;- S		( n -- )			(4)		'eSe'		editor
;-	spread making line n blank
eSe:GA;:ok: w/ b/buf=c/scr
;fig: dup 1- 14 do i line i 1+ -move -1 +loop e ; <= :fehler: bei { 0 S }, { 15 S }
;=> dup line dup c/l + 0 line c/scr + over- move E ;
	spush 		; dup for {E}
	call line	; src
	spush
	add eax,byte c_l; dst
	spush
	xor eax,eax
	call line
	call cbys	; ecx:=bytes per screen
	sub eax,TOS(1)	; line(n1)
	add eax,ecx	; count := c/scr-(line(n1)-line(0)) = c/scr+line(0)-line(n1)
	call move	;:continued: <eE>
;-
;- E		( n -- )			(4)		'eE'		editor
;-	erase line n with blanks
eE:GA;:ok:
; line c/l blanks update ;
	call line
	call cbyl
	call blanks
	jmp update
;-
;- D		( n -- )			(4)		'eD'		editor
;-	delete line n, but hold in pad
eD:GA;:ok:
; dup h 15 dup rot do i 1+ line i -move loop e ; <= :fehler: bei { 15 D }
	cmp eax,byte 15
	jz eE		;?; erase bottom line
	spush 2		; dup (dup dup for space)
	call eH		; H
	inc eax		; n+1
	call line	; ptr to line n+1
	spush
	sub eax,byte c_l; ptr to line n
	call prev	; line 0
	call cbys	; ecx := bytes per screen
	add eax,ecx	; end of scr
	sub eax,TOS(2)	; count
	call move
	mov eax,15	; clr bottom line
	jr eE
;-
;- TOP		( -- )				(4)		'etop'		editor
;-	home cursor to top left of screen
etop:GA;:ok:
; 0 r# ! ;
	mov dword[uref uv.unr],0
	jr em0
;-
;- F		( -- )ccc			(4)		'eF'		editor
;-	find occurence of following text
eF:GA;:ok:
; 1 text n ;
	call t1		;:continued: <eN>
;-
;- N		( -- )				(4)		'eN'		editor
;-	find next occurence of previous text
eN:GA;:ok:
; find 0 m ;
	call efind
	jr em0
;-
;- T		( n -- )			(4)		'eT'		editor
;-	type line by # n, save also in pad
eT:GA;:ok:
;  dup c/l * r# ! dup h 0 m ;
	spush
	call cbyl
	call smul
	mov [uref uv.unr],eax
	mov eax,TOS(1)
	call eH
	jr em0p
;-
;- M		( n -- )			(4)		'eM'		editor
;-	move cursor by signed amount n, print its line
eM:GA;:ok:
;  r# +! cr space #lead type 95 emit #lag type #locate . drop ;
	add [uref uv.unr],eax
em0p:
	spop
em0:
	mov ecx,m0v-Y
	jmp pdodefer
;-
;- [0m]		( -- )				(d)		'bm0b'		hidden
;-	terminal cursor line re-listing. default for deferred {0m}.
bm0b:GA;:ok:
	spush 2
	mov ecx,[kref hlmgv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	call cr
	call spaces
	call hlead
	call type
	mov eax,0+'^'
	call emit
	call hlag
	call type
	call hlocate
	TOSPULL(1)
	jmp dot
;-
;- B		( -- )				(4)		'eB'		editor
;-	backup cursor by text in pad
eB:GA;:ok:
; pad c@ minus m ;
	call pad
	movzx eax,byte[kref eax]
	neg eax
	jr eM
;
t1:A;<eF,eX,eC)
	spush
	mov eax,1	;:continued: <text>
;-
;- text		( c -- )			(4)				editor
;-	accept following text to pad, w/ blanks filled up to c/l chars,
;-	until <eol> or char-code c received.
text:GA;:ok:
;here c/l 1+ blanks word here pad c/l 1+ cmove ;
;=>
;word count >rr pad >rr place r> r + 1+ c/l r> - blanks ;
	call XWORD
	call ccount
	push eax
	call pad
	push eax
	call place
	spush
	pop edx
	mov ecx,[esp]
	lea eax,[ecx+edx+1]
	call cbyl
	pop ecx
	sub eax,ecx
	jmp blanks
;-
;- C		( -- )ccc			(4)		'eC'		editor
;-	insert following text at cursor.
eC:GA A 1;:ok:
; pad count #lag 4dup 2swap sdrop -rot 2 pick - 0 2over + sdrop swap move 2drop swap cmove 0 m ;
	call t1
	call pad
	call ccount	;	( -- p1 u1 
;-
;- Y		( p u -- )			(4)		'eY'		editor
;-	insert stg(p,u) at cursor.
eY:GA;:ok:
	call hlag
	call dup4
	call swap2	;	( p u ps us  ps us p1 u1
	mov edx,TOS(3)	;   ps
	mov ecx,TOS(2)	;   us
	TOSPULL(1)	; sdrop p1
	add edx,eax	;d: ps+u1
	sub ecx,eax	;c: us-u1
	mov TOS(1),edx
	mov eax,ecx
	call move	;	( p u ps us  ps ps+u1 us-u1 -- p u ps us
	push eax	; us
	spop
	call swap	;	( p ps u
	pop ecx
	sub ecx,eax
	jg .m
	add eax,ecx
.m:
	call move
	call update
	jmp disp em0
;-
;- DELETE	( n -- )			(4)				editor
;-	backwards at cursor by count n, spanning full screen.
delete:GA;:ok:
;  >r #lag + forth r -              ( save blank fill location )
;  #lag r minus r# +!                          ( backup cursor )
;  #lead + swap cmove r> blanks update ; ( fill from end of txt)
	push eax
	spop		; >r
	call hlag
	add TOS(1),eax	; #lag +
	spop
	sub eax,[esp]	; r -
	call hlag
	mov ecx,[esp]
	sub [uref uv.unr],ecx; r minus r# +!
	call hlead
	add TOS(1),eax
	spop
	call swap
	call xcmove
	spush
	pop eax
	call blanks
	jmp update
;-
;- -text	( p1 u1 p2 -- f )		(4)		'mtext'		editor
;-	compare stg(p1,u1) whith data at p2, ret f =/= 0 if matching.
;-	letter case dependency as set w/ {lc-depend}/{lc-ignore}.
mtext:GA;:ok:
;  swap -dup if		( leave boolean matched=non-zero, nope=zero )
;   over + swap
;   do dup c@ forth i c@ - if 0= leave else 1+ endif loop
;  else drop 0= endif ;
	spush
	mov eax,TOS(2)	; over
	call compare
	jmp zeq
;-
;- match	( p1 u1 p2 u2 -- f u3 )		(4)				editor
;-	search memory(p1,u1) for stg(p2,u2),
;-	ret tf and disp u3 wrt p1 if stg found, else ff and u3 := u1
;-	letter case dependency as set w/ {lc-depend}/{lc-ignore}.
match:GA;:ok:(1line)
;  >r >r 2dup r> r> 2swap over + swap
;  do 2dup forth i -text if >r 2drop r> - i swap - 0 swap 0 0 leave endif loop
;  2drop swap 0= swap ;
	push dword TOS(2)	; save u1 for result disp.
	call search		; -- p3 u3 f ) tf:p3=stg u3 rmg chars, ff:p3=p1, u3=u1
	mov ecx,TOS(1)
	mov TOS(2),eax		; flg
	pop edx			; u1
	test eax,eax
	jz .r
	sub ecx,edx
	neg ecx			; u3 := u1-u3
.r:
	mov TOS(1),ecx		; u3
	spop
	ret
;-
;- 1line	( -- f )			(4)		'e1line'	editor
;-	scan line with cursor for match to pad text, update cursor, return boolean
e1line:GA;:ok:
;  #lag pad count match r# +! ;
	call hlag
	call pad
	call ccount
	call match
	add [uref uv.unr],eax
	spop
	ret
;-
;- scrfpos	( -- dn )							blkfile
;-	ret file-ptr at current screenfile cursor posn, from {scr} and {r#}.
scrfpos:GA;:ok:
; o: ecx = c/scr
	spush
	call cbys	; ecx := c/scr
	mov eax,dword[uref uv.uscr]
	push ecx	; save c/scr
	mul ecx
	add eax,[uref uv.unr]; curpos
	adc edx,byte 0
	spush
	mov eax,edx	; fposn of screen
	pop ecx		; pass c/scr to caller
	ret
;-
;- FIND		( -- )				(4)(e)		'efind'		editor
;-	string at pad over full screenfile range, else error message.
;-	(e: Z if found)
efind:GA;:ok:
; begin 1023 r# @ < if
;  top pad here c/l 1+ cmove 0 error endif ( display text & error message )
; 1line until ;
;1line: #lag pad count match r# +! ;
	call scrfpos	; get fptr at curs
	push ecx	; c/scr
	call scrcn	; scr-chan
	push eax	; save ch
	call reposnf	; fpos at cursor+1
	spop
	call pad
	call ccount	; stg
	spush
	pop eax		; ch
	call searchf	; ret disp to starting fpos, if found
	mov ecx,eax
	test eax,eax
	pop eax		; c/scr
	jnz d.nf	; text not found or file error
	mov ecx,[uref uv.unr]
	add dword TOS(2),ecx; cur disp+pos, to after stg
	adc dword TOS(1),byte 0
	call mby	; ( d.fpos c/scr -- curs d.scr# )
	mov ecx,TOS(1)
	mov eax,TOS(2)
	add [uref uv.uscr],ecx	; scr
	mov [uref uv.unr],eax	; r#
	cmp eax,eax	; ret Z
	spop 3
	ret
d.nf:
	spop 3		; drop ernum & false file-ptr
p.nf:; < text & 'not found'
	call pad
m.nf:
	call print
	spush 2
	mov eax,' : '
	call emit
	mov eax,19
	jmp message	; & ret NZ
;-
;- COPY		( n1 n2 -- )			(4)				editor
;-	duplicate screen n1, onto screen n2. 
;-	can extend beyond <eof>, itermediate scr-s filled w/ <nul>-bytes.
copy:GA;:ok:-?-
;  b/scr * offset @ + swap b/scr * b/scr over + swap
;  do dup forth i block 2 - ! 1+ update loop drop flush ;
	spush
	call clear		; append to file, if necessary
%if b_scr = 1;:ok:
	call swap
	call block
	call swap
	call block
	call bbybuf
	call xcmove
	jmp update
%else; -?-
	call bbyscr
	push eax		;		[ bs
	call smul
	mov ecx,databot
	add ecx,[kref offsetv-Y]
	add eax,[kref ecx]	; offset @ + ( n2' := n2*b/scr+@offset )
	call swap
	spush
	mov eax,[esp]		; b/scr
	call smul
	spush
	pop eax
	add eax,TOS(1)		; b/scr over +
	call swap
	call pdop
.do:
	spush
	call block
	call i
	call block
	call bbybuf
	call xcmove
	call update
	inc eax
	call ploopp
	dd .do-$
	spop
	jmp flush
%endif
;-
;- TILL		ccc( -- )			(4)		'till'		editor
;-	delete on cursor line, from cursor till text beginning.
till:GA;:ok:
; #lead + 1 text 1line 0= 0 ?error #lead + swap - delete 0 m ;
	call hlead	; p u
	add TOS(1),eax	; p'
	mov eax,1
	call text
	call e1line	; ( p' n
	test eax,eax
	spop
	jz near drop	;?; text not found
	call hlead
	add TOS(1),eax
	spop
	call swap
	sub TOS(1),eax
	mov eax,TOS(1)
	call delete
	jmp disp em0
;-
;- X		( -- )ccc			(4)		'eX'		editor
;-	delete following text, spanning full screenfile range.
;-	-re- {TILL} for del on cursor line, only
eX:GA;:ok:
; 1 text find pad c@ delete 0 m ;
	call t1
	call efind
	call pad
	movzx eax,byte[kref eax]
	call delete
	jmp disp em0
;-
;- [edit]	( -- )				(4)		'beditb'	editor
;-	initial runtime of deferred {edit}, loading the screenfile editor.
;-	message ENOENT if file <edit.f8> not found, else {edit} set to {noop}.
beditb:GA;:ok:
	call pad
	call pathmax
	call psqtep
	db 7,'edit.f8'
	call bfnameb
	test eax,eax
	jz .e
	spush
	mov eax,Ieditor
	call orderq
	test eax,eax
	jz .v
	call also
	call editor
.v:
	mov ecx,[kref editv-Y]
	add ecx,databot
	spop
	mov word[kref ecx],Inoop
	jmp included
.e:
	mov eax,ENOENT
	TOSPULL(1)
	jmp message
;-
;- edit		( -- )				(d)		'edit'		root
;-	loading/executing the screenfile editor. deferred to -re- {[edit]}.
;edit:KV;GA

;	mov ecx,editv-Y
;	jmp pdodefer
;-
;- [f]		(X: -- p f )			(i)		'bfb'		editor
;-		(C: -- )
;-	-re- {-find} and ret flg & lfa = p if in interpreting state, 
;-	else compile/execute next word w/ forth voc. preferrence.
bfb:GA;:ok:
	call also
	call forth
	call dfind
	call ccplq
	jz .j
	test eax,eax
	jz .j
	mov eax,TOS(1)
	TOSPULL(1)
	jg .x
	call lfa2idx
	call compilec
.j:
	jmp previous
.x:
	call execute
	jr .j
	
%endif; __BLOCK_I
;-
;-=====================================================================;
;- l4block.inc <eof>
