;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4consio.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __CONSIO_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	

%ifndef __CONSIO_I
%define __CONSIO_I
%define consioinc
;-
;-----------------------------------------------------------------
;- -- standard i/o (more or less...)
;-

;-	

;-
;- beep		( n1 n2 -- )			(s)				forth
;-	sound generation, pitch n1[84s], duration n2[ms]
;-	(standard console ^G sound is n2:=125, n1:=1591)
;-	syscall ioctl -re- man 2 ...
beep:GA;:ok:
;:HIER:	(m-tasking direct duration control not possible! :HERE: maybe "pcsp" program might help -?-
;	(emulated with endless timo & sigalarm? or nanosleeping sig.hnd?
;pc speaker with: msw(n2):t*ms, lsw(n1):t/1.193.180 => 838,1ns = 0,84s => 11,9kHz
; c: -/-
	pushad
	mov edx,eax
	shl edx,16
	spop
	mov dx,ax
	call getstdin
	mov ebx,[kref ecx+cdt.chid-cdt]
	sys_ioctl EMPTY,KDMKTONE
	popad
	spop 2
	ret
;-
;- space  	( -- )								forth
;-	send a blank space to @{stdout}
space:GA;:ok:
	call bbl
	jr emit
;-
;- cr		( -- )								forth
;-	send new-line command to @{stdout}
cr:GA;:ok:->{emit}
	spush
	mov eax,[uref uv.ueol]
;-
;- emit		( c -- )			(e)(v)		'vemit'		forth
;-	send single char to @{stdout}
;-	initially vectored to {[emit]}
emit:GA;:ok:
	jmp [uref uv.uemit]
;-
;- ?cr		( n -- )					'qcr'		forth
;-	new-line if (@out + n) not less than main window width
qcr:GA;:ok:
	push eax
	add [byte uref uv.nout],eax
	call qeol
	pop ecx
	TOSPULL(1)
	sub [byte uref uv.nout],ecx
;-
;- -cr		( flg -- )			(f6)		'dcr'		forth
;-	{cr} if flg =/= 0
dcr:GA;:ok:->{cr}
	test eax,eax
	spop
	jnz cr
.r:
	ret
;-
;- o-cr		( -- )						'odcr'		forth
;-	{cr} if output device is a console.
odcr:GA;:ok:
	call stdoutc
	call chconsq
	jr dcr
;-
;- ?eol		( -- flg )					'qeol'		forth
;-	true flag if max-x not larger than @out
qeol:GA;:ok:
	call maxxy
	spop
	movzx eax,byte[kref eax-whead.l+whead.n]
	add eax,[byte uref uv.nout]
	sub eax,TOS(1)
	sar eax,31
	not eax
	ret
;-
;- cbytes 	( c -- c u )					'cbytes4'	hidden
;-	ret count u of significant characters in cell-sized (32-bit) c.
cbytes4:GA
	spush
	call cbytes
	mov eax,edx
	ret
; edx := no of consecutive non-zero bytes in @tos = eax
cbytes:A;<bemitb, writechar)
; c: edx
	mov edx,eax	; byte-cnt or zero
	bsr edx,eax	; find top bit to send
	add edx,byte 7	; round up to 8bit-byte size
	shr edx,3	; bytewise
	jnz .e		;?; non-zero cnt
	inc edx		; at least one byte
.e:
	ret
;-
;- [emit]  	( c -- )			(e)		'bemitb'	hidden
;-	send single char to @{stdout}, update {out} accordingly.
;-	"single char" may be up to four non-zero bytes of cell c
;-	on tos, which will be sent lsb 1st, until rmng @cell = 0
;-	(this is in preparation for unicode capabilities)
bemitb:GA;:ok:
	call cbytes
	mov [PS],eax	; pass bytes to send
	call getstdout	; ecx:=chp
	push ebx,ecx	; save KR and used channel ptr
	mov [byte uref uv.outchp],ecx; chp in use
	mov ebx,[kref ecx+cdt.chid-cdt]
	mov ecx,PS	; buf
	sys_write
	pop ebx		; chp
	add ebx,[esp]	; kref'd
	mov [ebx+cdt.rnum-cdt],eax; store ernum
	pop ebx
	mov [uref uv.uoerr],eax; save ernum
	test eax,eax
	jng .r		;?; error
	add [byte uref uv.nout],eax; cnt sent bytes
; {out} & {rows} correction according to char code
; chars other than printing, <lf> and <tab> should be handled by the rsp. words
; in:	ecx	ptr to string
;	edx	cnt
; out:	ecx	0
.c:
	push ebx
.d:; ;<[type])
	push esi
	mov esi,ecx	; buf
	mov ecx,eax	; count
	jecxz .q
.o:
	lodsb
	cmp al,32
	jnc .l
	test al,al
	jz .q
	dec dword[byte uref uv.nout]
.f:
	cmp al,AC_LF	; 10
	jnz .t
	inc dword[byte uref uv.rows]
	dec ecx
	mov dword[byte uref uv.nout],ecx
	jg .o
.t:
	cmp al,TAB
	jnz .l
	mov eax,ecx
	mov edx,0
; <tab> spacing from console ioctl -?-
	mov ebx,[uref uv.untab]
	test ebx,ebx
	jg .v
; falling back to hard coded <tab>
	lea ebx,[edx+8]
.v:
	div ebx
	sub edx,ebx
	not edx
	add [byte uref uv.nout],edx
.n:
	add ecx,edx
.l:
	dec ecx
	jg .o
.q:
	pop esi
.x:; ;<[type])
	pop ebx
.r:; ;<[type])
	spop
	ret
;-
;- [ctype] 	( p +n -- )			(s)		'bctypeb'	hidden
;-	send string (p,u) to stdout, alternate vector for -re- {type}.
;-	silently return if -ve, zero or > 64K stg length.
;-	error code stored to cdt -re- {ch-rnum}, and {(oerr)}.
;-	syscall write -re- man 2 ...
bctypeb:GA;:ok:
	movzx edx,ax		; cnt
	cmp eax,edx
	spop
	jnz bemitb.r		;?; error -> noop
	mov ecx,eax
	test edx,edx
	jle bemitb.r		;?; nothing
	add ecx,KR
	push ebx,edx,ecx	; save kref
	call getstdout
	mov [byte uref uv.outchp],ecx; chp in use
	lea eax,[kref ecx]
	mov ebx,[eax+cdt.chid-cdt]
	mov ecx,[esp]
	push eax
	sys_write
	pop ecx
	mov [ecx+cdt.rnum-cdt],eax
	mov [uref uv.uoerr],eax	; save ernum
	pop edx,ecx
	test eax,eax
	js bemitb.x		;?; rst KR and ret
	add [byte uref uv.nout],edx; cnt sent bytes
	jmp disp bemitb.d
;-
;- [etype] 	( p +u -- )			(l)		'btypeb'	hidden
;-	{[type]} string(p,u) to stdout w/ by -re- {e\stg} converted escaped tokens.
;-	silently return if -ve, zero or > 64K stg length.
;-	error code readable immediately after, from -re- {ch-rnum}, {(oerr)}.
;-	numeric token:
;-		\nnn	character code, numeric sequence as valid for a forth single 
;-			integer at octal radix, any other w/ the appropriate praefix.
;-			for instance
;-				s" hick\46"	displays	hick!
;-				s" hick\#33"	does the same
;-	special symbols, letter case dependent:
;-		\a	<bel>
;-		\b	<bs>
;-		\c	w/o trailing <lf> - suppressed, not applicable
;-		\e	<esc>
;-		\f	<ff>
;-		\n	<new line>, Linux: <lf>
;-		\r	<cr>
;-		\t	<ht>
;-		\v	<vt>
;-		\\	double escape symbol, sending the escape symbol as a char.
;-		\other	character following the escape symbol will be sent.
betypeb:GA;:ok:
	spush 2
	mov ecx,[byte uref uv.s0]
	add eax,byte 3
	shr eax,2
	mov TOS(1),ecx
	push eax
	call toloc
	call estg
	call bctypeb
	spush
	pop eax
	neg eax
	jmp toloc
;
dotype:A
	mov ecx,btypebv-Y
	jmp pdodefer
;-
;- [ok]		( -- )				(79)		'bokb'		hidden
;-	kbd input prompt: ">" while interpreting, "=" compiling.
;-	true xec address by dft stored to (ok).
bokb:GA;:ok:
	spush
	mov eax,toki-Y	; interpreting prompt stg immediately before compiling prompt stg
	call ccplq
	jz print	;?; interpreting
	movzx ecx,byte[kref eax]
	lea eax,[eax+ecx+1];:continued: <print>
;-
;- print  	( p -- )			(f6)				forth
;-	display asciz or counted string
print:GA;:ok:
	call ccount	;:continued: <type>
;-
;- type		( p +n -- )			(kd)				forth
;-	send string (p,+n) to stdout. uvari vector {(type)}, which defaults to 
;-	deferred {[type]}, initially set to xec {[etype]} and is-dft {[ctype]}.
;-	{out} updated according to printing chars, <nl> and <tab>.
;-	NOTE wrt windowing: NO orderly means to finding out about 'pending newline'! 
;-	     neither by console routines not by vt-sequences, only feasible solution 
;-	     would be useing {emit} an doing the accounting, within an apllication.
type:GA;:ok:
	jmp [uref uv.utype]
;-
;- zprint  	( p -- )							forth
;-	display <nul> terminated string
zprint:GA;:ok:
	call zcount
	jr type
;-
;- bl.r		( p u n -- p u )		(f6)		'bldotr'	forth
;-	output as many blanks as required to right justified {type}ing 
;-	stg(p,u) in a field of n chars. no extra chars if n <= u.
;-	-re- {+blanks}, which builds a blank string at -re- {here}
bldotr:GA;:ok:
	sub eax,TOS(1)	;:continued: <spaces>
;-
;- spaces  	( +u -- )							forth
;-	-re- {type} +u blank spaces.
spaces:GA;:ok:
	mov ecx,eax
	test eax,eax
	jng .x		;?; -ve or zero, drop & noop
	shr ecx,8
	and eax,255
	push ecx
	mov edx,eax
.l:
	spush
	mov eax,blkst-Y
	spush
	mov eax,edx
	call type
	mov edx,256
	dec dword[esp]
	jns .l
.r:
	lea esp,[esp+4]
.x:
	spop
	ret
;-
;- emit-m  	( +u c -- )			(f6)		'emitm'		forth
;-	{emit} +u characters c, 0 < +u < 32768, no-op if > 32767
;-	char c may be up to 4 bytes, full size of a sing.
emitm:GA;:ok:
	mov ecx,eax
	spop
	movsx edx,ax	; limit to 64K
	cmp edx,eax
	jnz .r		;?; error
	test edx,edx	; count
	jng .r		;?; error
	push ecx	; >r	char
	push eax	; >r 	count
.l:
	mov eax,[esp+4]	; code
	call emit
	spush
	dec dword[esp]
	jg .l
	lea esp,[esp+8]
.r:
	spop
	ret
;-
;- edlin   	( p1 u1 -- p2 u2 )		(4)				forth
;-	accept to buf(p1,u1), edit expanded default <nul>-terminated text at p1. 
;-	cursor posn at dft text 2nd column, expand to eol w/ <end> or curs-down.
;-	passing buffer(p1,u1), ret stg(p2,u2) with received len u2 < u1.
edlin:GA;:ok:
	call xy
	spush
	mov eax,TOS(4)	; p
	call zprint
	call atxy
	call qkey 	; clr last key input
	mov ecx,TOS(2)	; u1
	mov edx,TOS(3)	; p1
	mov TOS(1),ecx
	mov TOS(2),edx	; 2drop over swap <tab> key! accept
	_mov eax,AC_TAB
	call keysto	;:continued: <osaccept>
;-
;- [accept]	( p +u1 -- +u2 )	(e)			'osaccept'	hidden
;-	default for accept - preliminary, until terminfo implemented.
;-	fetch at most u1 chars input from stdin, or until <eol>, to buf. at p.
;-	required space is 0 < u1 < u1+(len-of-<eol>-sequence).
;-	overwriting kbd input, termination after <eol>, <cr> or <nul> received,
;-	buf. protected to not writing beyond bounds 0 < #chars < +u1.
;-	trailing <eol> zeroed, +u2 not including length of <eol> sequence.
;-	input echo can be dis/enabled w/ {echo-off}/{echo-on}.
;-	ctrl actions for minimal editting capabilities w/ Linux console and "rxvt":
;-		<bs>	^H	back-space, not overwriting, moves cursor left.
;-		<e>[D	cur.lft	ditto (vt102 type terminal).
;-		<e>[C	cur.rgt	non-overwriting space, moves cursor right.
;-		<tab>	^I	non-overwriting space, moves cursor right, 
;-				enabling "last line recall" if 1st keystroke.
;-		<e>[1~	pos1	cursor back to line beginning.
;-		<e>[4~	end	expands till (previous) end of line.
;-		<del>	'<='	backup and delete char under cursor.
;-		<lf>	^J	line feed, exec <cr/lf>
;-		<cr>	^M	carriage return, exec <cr/lf>
;-		<ff>	^L	clear screen
;-		<bel>	^G	noise
;-	NOTE:	xterm-4.xx vt control broken! others in-consistent; though, almost 
;-		orderly w/ "rxvt" but, <tab> and <del> work for cursor positioning, 
;-		<enter> cuts off at cursor posn, <del-left> ok, and delete w/ <^H>.
osaccept:GA;:ok:
;c: ecx, edx
	movsx edx,ax		; cnt
	mov dword[byte uref uv.nout],0
	spop
	push esi		; save lp
	lea esi,[kref eax]	; buf.
	xor eax,eax		; +u2 on error
	mov ecx,EADDRNOTAVAIL
	test edx,edx
	jle near .r		;?; invalid chars cnt
	lea eax,[kref key-Y]	; vectored {key}
	call teststdin
	jz .i			;?; if input not console
	lea eax,[kref a1st-Y]	; initial "last line recall" routine
.i:
	push eax		; set to {key} after 1st keystroke
	push esi		;S0: buf.
	push edx		;S4: maxcount
	push edx		;S8: loop counter	@tor[ cnt max buf
	jr .h
.k:A
	mov eax,' '
.l:
	call ekey2char		; ascii code range check
	test eax,eax
	spop
	jnz .e			;?; not a control code
; prevent <tab> in files for cursor control
	call teststdin
	jz .c			;?; input not from console
	call actrl		; specific to line editting
	jc .h			;?; cc4th substituted
	call cc4th		; do ctrl action
	test eax,eax
	spop
	jz .h
	cmp al,' '
	jc .h
.e:
	call teststdin
	jz .c			;?; input not from console
	cmp byte[esi-1],0
	jnz .o
	and dword[uref uv.uiovtim],~(medif);reset eol-flag if 0-byte reached
.o:
	spush
	call emit		; echo visible chars
	or eax,eax		; NC - for {query}
.c:
	mov [esi-1],al		; store char
	jnc .h
	call tinqry
	jnz .x			;?; {query}
.h:; ;<initialentry)
	spop
	call keyq		; check pending input
	jnz .y			;?; char present
	mov ecx,[byte uref uv.ierr]
	cmp ecx,byte EAGAIN
	jnz .f			;?; <eof>
.y:
	spop
	call dword[esp+12]	; vectored {key} for 'last line recall'
;:: -?- tmp store key at {here} or above tos, handle ctrl-s, then copy to buf. -?- ::
	sub dword[esp],byte 1	; dec loop cnt & set C-flag!
	sbb ecx,ecx		; remain within buf. bounds, -1 if ovf
	jz .ti			;?; buffer ovf
	call del
	mov ecx,-1
	call tinqry
	jz .ti			;?; not {query}
	movzx eax,al
	lea esi,[esi+1]
	jr .xi
.ti:A
	sub [esp],ecx		; dont exceed..
	lea esi,[esi+ecx+1]	;	..upper bound
	test al,al
	jz .x			;?; <nul> terminates w.o. condition
	mov ecx,[uref uv.ueol]
	cmp al,cl		; <nl> code
	jnz near .l
.x:				; ready
	xor eax,eax
.xi:
	mov word[esi-1],ax
	pop eax			; count
	pop ecx			; maxcount
	sub eax,ecx
	not eax			; cnt w.o. trailing <nul>
	mov ecx,[byte uref uv.ierr]
	lea esp,[esp+8]		; drop buf & input vec.
.r:
	mov esi,[byte uref uv.inchp]
	and dword[kref esi+cdt.iofl-cdt],~(mqryf); clr query mode
	pop esi			; lp
	and dword[uref uv.uiovtim],~(medif|mqryf)
	mov [byte uref uv.ierr],ecx; store ernum
	ret
.f:A
	inc esi
	dec dword[esp]
	jr .x			; at eof
; -
actrl:A; ret C if action taken	; minimal editting, <bs> and <del>
	cmp eax,byte AC_ESC
	jz .v			;?; vt ctrl sequence
	cmp al,AC_CSI
	jz .q
.n:
	mov ecx,[esp+8]		; limit
	mov edx,[esp+4]		; rmd
	cmp al,AC_HT		; ^I non-overwriting space
	jz .t
	cmp al,byte AC_BS	; ^H backspace
	jz .s
	cmp al,AC_US		; ^_  non-overwriting space
	jz .t
	cmp al,byte AC_DEL	; <del> ascii 127
	jnz .d
.s:; ;<bs>
	call .b			; backup cursor
.d:				; discard ctrl char
	call .b
	mov [esp+4],edx
	ret
.b:A
	bts dword[uref uv.uiovtim],bedif
	jc .m
	mov word[esi-1],0
.m:
	cmp edx,ecx
	jnc .i			;?; no action, ret C
	inc edx
	dec esi
	cmc
	ret
.i:A;< signal at beginning of line
	pushad
	call bel
	popad
	stc
	ret
.t:A;< ^I => non overwriting space
	mov al,[esi-1]
	cmp al,1
	jc .w			;?; eol
	sub edx,byte 1		; rmd:=rmd-1
	jnc .r			;?; end of buf., no action, ret C
	dec edx
.w:
	inc edx
	dec esi
	mov [esp+4],edx
.r:
	ret
.a:A;<curs back
	mov al,AC_BS
	jr actrl.n
.p:A;<advance curs 
	mov al,AC_HT
	jr actrl.n
.q:A;<actrl w/ <csi>
	spush
	jr .z
.v:A;<actrl) ; evaluate curs, pos1, end; suppress other csi-s
; i: [esp+8] limit, [esp+4] rmd
	call nxk
	jz .u			;?; eof or error
	cmp al,'['
	jnz .j			;?; store this key back, use previous
.z:;after <csi>
	call keyq
	jz .g			;?; eof or error, as above
	spop 3			; drop <esc> and '['
	call key
	cmp al,'D'
	jz .a			;?; cursor left -> ^H
	cmp al,'C'
	jz .p			;?; cursor right -> ^I
	cmp al,'A'
	mov ecx,[esp+8]		; limit
	mov edx,[esp+4]		; rmd
	jz .cu			;?; cursor up
	cmp al,'1'
	jz .ku			;?; "pos1"
	cmp al,'B'
	jz .cd			;?; cursor down
	cmp al,'4'
	jz .kd			;?; "end"
	stc			; discard this char
.u:
	ret
.g:A
	spop
.j:
	jmp keysto
.kd:; <end>
	call dky
.cd:; <cursor down - expand previous text until <eol>
	mov al,[esi-1]
	cmp al,1;:HIER: cmp al,32 -?- fuer ctrls -> nee, nutzlos
	jc .de			;?; eol
	pushad
	spush
	call emit
	popad
	sub edx,byte 1		; rmd:=rmd-1
	inc esi
	jnc .cd			;?; more
.de:
	dec esi
	inc edx
.ce:
	or edx,edx		; NC
	mov [esp+4],edx
	stc
	ret
.ku:; <pos1>
	call dky
	jr .cu
.nu:A
	cmp edx,ecx
        jz .cu
	pushad
	call bs
	popad
.cu:; <cursor up - back to beginning of line
	call .b
	jnc .nu
	jr .ce
a1st:A;<last line recall, to expand w/ cur-right
	lea ecx,[kref key-Y]
	mov [esp+16],ecx
	call ecx
	cmp eax,byte AC_HT	; <tab>
	jz .r
	cmp eax,byte AC_ESC
	jz .e
.x:
	mov ecx,[esp+12]	; buf
; clear beginning of line, prevent ^H/^I exceeding beyond used area
	mov word[ecx],0		; proceed w/ regular input
.r:
	ret
.xd:
	spop
.xk:
	call nxk
	jz .xp
	spop
.xp:
	pop edx,esi
	jr .x
.e:; ;1st <esc> -> e[4~ = eol :: kommt mit <nl> am ende ::
	mov eax,AC_HT
	push edx,esi
	call nxk
	jz .xp
	cmp al,'['
	jnz .xd		;?; false key
	spop
	call nxk
	jz .xk
	call nxk
	jz .xd
	cmp byte TOS(1),'4'
	spop 2
	pop edx,esi
	jnz .x
	inc esi
	dec edx
	mov word[esi+edx],0
	mov eax,[esp+4]
	push eax	; pass rmd
	call actrl.cd
	pop eax
	mov [esp+4],eax
	mov eax,AC_HT	; advance to last char, prevent <nl>
	ret
dky:A;<discard next key
	push ecx,edx
	call key
	spop
	pop ecx,edx
	ret
nxk:A
	call keyq
	spop
	jz .r			;?; eof or error
	call key
	test eax,eax
.r:
	ret
tinqry:A
	push ecx
	mov ecx,[byte uref uv.inchp]
	test dword[kref ecx+cdt.iofl-cdt],mqryf; (#24)
	pop ecx
	ret
;-
;- accept 	( p +u1 -- +u2 )		(GA)		'vaccept'	forth
;-	fetch at most u1 chars input from stdin, or until <nl>
;-	execute true address uot vector (accept), dft -re- {[accept]}.
vaccept:GA;:ok:
	jmp [uref uv.uaccept]
;-
;- query  	( -- )								forth
;-	fetch at most {#tib}-1 chars, or until <nl>, from stdin to {tib}.
;-	returns if either of eol <nl> or {#tib 1-} chars received.
;-	store no. of received chars in user vari {span}. store 0 to #in and >in
query:GA;:ok:
	call source		; count buf.
	mov ecx,stdin
	shl ecx,cdt_bsize+2
	mov eax,[uref uv.utibmax]
	add ecx,[byte uref uv.chans]
	mov [byte uref uv.inchp],ecx
	or dword[kref ecx+cdt.iofl-cdt],mqryf; flag query mode
	call vaccept
	mov [byte uref uv.span],eax
	mov dword[byte uref uv.in],0
	spop	
	ret
;-
;- expect  	( p +u1 -- )							forth
;-	fetch at most u1 chars input from stdin, or until <nl>
;-	store no. of received chars in user vari {span}.
expect:GA;:ok:
	call [uref uv.uaccept]
	mov [byte uref uv.span],eax
	spop
.r:
	ret
;-
;- ---
;-
;- -- keyboard input
;-
;-	{uc-map}, {uc-unmap}
;-		emulated hardware mapping if input from stdin is console.
;-	{lc-ignore}, {lc-depend}
;-		sofware evaluation of any input source case ignoring or not.
;-

;-
;- uc-map 	( -- )				(e)(s)		'uclm'		linux
;-	map upper case to lower case chars of input to (work) channel.
;-	re {uc-unmap}
uclm:GA;:ok:
	pushad
	mov ecx,[uref uv.utattr+termios.c_iflag]
	or dword[ecx],IUCLC
	jr uclu.n
;-
;- uc-unmap	( -- )				(e)(s)		'uclu'		linux
;-	separate upper and lower case chars mapping of input to (work) channel
;-	store true error code to (oerr). -re- {uc-map}
;-	syscall ioctl -re- man 2 ...
uclu:GA;:ok:
	pushad
	mov ecx,[uref uv.utattr+termios.c_iflag]
	and dword[ecx],~(IUCLC)
.n:
	xor eax,eax
	dec eax
	mov ecx,kbdch
	call toutcons
	js .q
	mov ebx,[kref edx+cdt.chid-cdt]
	mov edx,[uref uv.utattr]
	sys_ioctl EMPTY,TCSETS	; set new tty mode
	mov [uref uv.uoerr],eax
.r:
	test eax,eax
.q:
	popad
	ret
;-
;- lcase@  	( -- n )					'lcasef'	forth
;- LCASE@  	( -- n )					'LCASEF'	forth
;-	fetch letter-case flag, zero if case-dependency set. -re- {caps?}
;-	e.g, use{ lcase@ lc-ignore ..words.. lcase@ 0= if lc-depend endif }
lcasef:GA;:ok:
	spush
	mov eax,mcaps
	and eax,[byte uref uv.bits]
	ret
;-
;- lc-ignore	( -- )				(m,7)		'lcignore'	root
;- LC-IGNORE	( -- )				(m,7)		'LCIGNORE'	root
;-	force case in-dependent search and string comparison. -re- {lc-depend}.
lcignore:GA M;:ok:
	or dword[byte uref uv.bits],mcaps
M lcignore
	ret
;-
;- lc-depend	( -- )				(m,7)		'lcdepend'	forth
;-	case dependent search and string comparison mode - default state.
;-	If case independent search mode enabled	1st match takes precedence,
;-	which for L4 kernel applies to lower case version. -re- {lc-ignore}.
lcdepend:GA M;:ok:
	and dword[byte uref uv.bits],~(mcaps)
M lcdepend
	ret
;-
;- caps?  	( -- flg )					'capsq'		forth
;- CAPS?  	( -- flg )					'capsq'		forth
;-	tf if text comparison is case dependently, for {find}, {compare}, etc
capsq:GA;:ok:
	spush
	mov eax,mcaps
	and eax,[byte uref uv.bits]
	mov eax,-1
	jz .r
	inc eax
.r:
	ret
;-	
;-
;-================================================================
;- -- i/o selectable timeout
;- -- raw/single characters intput
;-	

wait4t.c:A
	TICK wait4x
	jmp compilec
;-
;- wait4' 	ccc( n1 -- n2 n3 )		(i)		'wait4t'	linux
;-	xec any i/o related word ccc w/ the {(stimo)} uvari set to n1 no. of 
;-	seconds for specific time-out. restores previous setting, after return.
;-	effective for all words which call a {wait4..} word, e.g. {key}, etc.
;-	ret n2 := n3 := -1 if xt was not an executeable item.
wait4t:GA
	call tick
	call ccplq
	jnz .c
;-
;- wait4x 	( n1 xt -- n2 n3 )				'wait4x'	linux
;-	xec any i/o related word w/ the {(stimo)} uvari set to n1 no. of seconds
;-	for specific time-out. restores previous setting, after return.
;-	effective for all words which call a {wait4..} word, e.g. {key}, etc.
;-	ret n2 = n3 := -1 if xt was not an executeable item.
wait4x:GA
	call xecq
	mov ecx,TOS(1)	; xec
	mov edx,TOS(2)	; timo
	inc eax
	spop 3
	jnz near dmone
.n:
	push dword[uref uv.ustimo]
	mov [uref uv.ustimo],edx
	call ecx
	pop dword[uref uv.ustimo]
	ret
;-
;- wait4ti	( ch n1 -- n2 n3 )		((s))				linux
;-	wait for n1 seconds until input pending to channel ch, -re- {wait4ch} 
wait4ti:GA
	push dword[uref uv.ustimo]
	mov [uref uv.ustimo],eax
	spop
	call wait4ch
	pop dword[uref uv.ustimo]
	ret
;-
;- wait4ch	( ch -- n1 n2 )			((s))				linux
;-	wait for @(stimo) seconds until input pending to (any type) channel ch, 
;-	ret n1=n2=0 after timeout, else remaining seconds(n2) and nanosec(n1).
;-	limited to channel identifiers (fd-s) in range [0,31].
;-	syscall -re- man 2 select, remaining time is a feature specific to linux.
wait4ch:GA;:ok:
	call chid
	spush
	mov ecx,edx
	call wait4fd.s
	mov TOS(1),ecx
	ret
;-
;- wait4kbd	( -- )				(s)				linux
;-	wait until keyboard (console channel) input pending
wait4kbd:GA
	mov ecx,kbdch
	call getstdin.n
	jr wait4in.n
;-
;- wait4in	( -- )				(s)				linux
;-	wait until input pending if stdin is console
;-	limited to stdin channel identifiers (fd-s) in range [0,31].
;-	actual FILE_MAX can be read from /proc/sys/fs/file-max
wait4in:GA;:ok:
;man 2: int  select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout);
;4th: pad 100 erase 10 pad 20 + ! pad 20 + 4th>abs pad 8+ ! pad 12 + 4th>abs dup 4- dup 4- dup 4- 2 sp@ #142 syscall
; c: ecx,edx
	call getstdin	; ecx:=chp
.n:
	cmp dword[kref ecx+cdt.chid],byte -1
	jz wait4fd.x	;?; channel not valid
	test dword[kref ecx+cdt.iofl],mcons|micon
	jz wait4fd.x
	spush
	mov eax,[kref ecx+cdt.chid]
;-
;- wait4fd	( fd -- )			(s)				linux
;-	wait until input pending to channel's fd. limited to fd-s in range [0,31].
;-	time-out from uvari {(stimo)} [s] and {(stimo) 4-} [s]. max ea $7fffffff.
;-	executes the (pause) vector if idle and, if an(y) alternate uot exists.
;-	syscall select -re- man 2 ...
wait4fd:GA;:ok:
	cmp eax,byte 32
	push ecx
	jnc .r
	call .s		; ecx|eax = rmg time
%if 1;;;-.. ; ..not overly useful, yet
	jnz .r		;?; no time-out
	or eax,[uref uv.upause]
	jz .r		;?; vector disabled
	mov ecx,[ebp]
	add ecx,KR
	cmp ebp,ecx
	jz .r		;?; no other context
	push ebx
	push ebp
	mov [uref uv.uesi],esi; LP
	mov [uref uv.uedi],edi; PS
	mov ebp,ecx	; next user
	push esi
	push edi
	call eax
	pop edi
	pop esi
	xor ecx,ecx	; rst Z
	pop ebp
	pop ebx
	mov esi,[uref uv.uesi]; LP
	mov edi,[uref uv.uedi]; PS
%endif;;;-''
.r:
	pop ecx
	spop
.x:
	ret
.s:A;<wait4fd,wait4ch)
	push esi,edi,edx,ebx; save lp, PS, BP
	mov ecx,eax
	movzx eax,ah
	bts eax,ecx	; in
	mov ecx,PS	; ->readfds, also excep, write is NULL
	spush		; r/w/x flags
	mov eax,[uref uv.ustimo]
	mov ebx,[uref uv.ustimm]
	mov edx,PS	; ->time
	spush		;timeval.sec
	mov [PS],ebx;;eax	;	.usec
	sys_select 32,EMPTY,NULL,ecx,edx; ebx:=32, ecx=TOS(2), edx:=0, esi:=ecx=TOS(2), edi:=edx=TOS(1)
	test eax,eax
	mov ecx,eax
	js .p
	jnz .z
	test ecx,ecx	; ret Z if timer elapsed
.z:
	mov eax,[edi]
	mov ecx,[edi+4]
.p:
	pop esi,edi,edx,ebx	; lp, PS, BP
	ret
;-
;- ?input 	( ch -- flg )			(s)(fg)		'qinput'	forth
;-	input from supplied channel, otherwise identical -re- {?terminal}
qinput:GA;:ok:
	call chid	; edx:=cdt, ecx:=flags
	push edx
	jr qterminal.n
;-
;- ?terminal	( -- flg )			(s)(fg)		'qterminal'	forth
;-	tf if char available from stdin channel, else ff
;-	pre-fetches avail char to stdin <cdt.ibuf>, if not already in buf.
;-	syscall read -re- man 2 ...
;-	stores ernum to uv.ierr and cdt.rnum:
;-		 >0:(no of) char(s) received
;-		  0:no input (VMIN=0), <eof>
;-		 -9:invalid file descriptor (ebadf)
;-		-11:no input (VMIN>0) - (eagain)
;-	{?terminal} is default to vectored {ekey?}
qterminal:GA;:ok:;<stdinq,refill,uv.uekeyq)
; c: ecx/cl, C set if char received
	push edx
	call inid	; edx:=cdt, ecx:=iofl
.n:
	and ecx,~(mcrgt)
	mov [byte uref uv.inchp],edx; chp in use
	mov [kref edx+cdt.iofl-cdt],ecx
	test ecx,mcflg
	mov ecx,1
	jnz .r		;?; char present
	push edx,ebx	; save chp, kref
	lea ecx,[kref edx+cdt.ibuf-cdt]; buf
	mov edx,0
	mov [ecx],edx	; clear charbuf
	mov ebx,eax
	inc edx		; buflen 1
	sys_read
	test eax,eax
	pop edx,ebx
	mov [byte uref uv.ierr],eax; ernum
	mov [kref edx+cdt.rnum-cdt],eax
	setg cl		; 1:chars 0:err/no input
	jng .r
	or dword[kref edx+cdt.iofl-cdt],mcflg; flag new char present
	test dword[kref edx+cdt.iofl],mcons|micon
	jnz .r
	cmp dword[byte uref uv.blk],byte 0
	jns .r		;?; stdin not from file
	dec dword[byte uref uv.blk]; inc fileptr
.r:
	pop edx
	movzx eax,cl	; flag tf:ok, ff:no input
	neg eax		; tf => $ffffffff and C, 00ff and NC if no char
	ret
;
neky:
	call wait4in
	spop
;-
;- ekey		( -- c )							forth
;-	return any char, wait until received (from stdin)
;-	re DPANS 10.4.1.1: char encoding as set for the rsp terminal/linux console.
ekey:GA;:ok:
; c:	ecx
	call [uref uv.uekeyq]	; ?terminal
	test eax,eax
	jz neky
	call getstdin
	mov eax,[kref ecx+cdt.ibuf-cdt]; char
	and dword[kref ecx+cdt.iofl-cdt],~(mcflg); clr char received flag
	ret
;-
;- cc4th  	( c -- c flg )			(83)				forth
;-	vectored action on input ctrl chars in range of [0,31],127.
;-	checking whether char is ctrl code and corresponding vector set;
;-	ret char and tf if not ascii ctrl or if no vector defined, else
;-	ctrl-code and ff w. other stack results depending on ctrl action.
;-	leave <nul> w.o. action, transpose @tos <del> to vector for <nul>
;-	vector ( .yy. c c' -- .yy. .xx. c n )
;-		must preserve char code @nos, @tos may be overwritten.
;-		can be de-activated by kref(zero) ptrs at either posn.
cc4th:GA;:ok:
; c: ecx
	spush
	cmp eax,byte 32
	jnc .x		;?; > 31
	test eax,eax
	jz .r
.f:
	mov ecx,[byte uref uv.icc4th]
	add ecx,KR
	jz .r
	cmp ecx,KR
	jz .r
	mov ecx,[eax*4+ecx]
	add ecx,KR
	jz .r
	cmp ecx,KR
	jz .r
	call ecx
	xor eax,eax	; Z if active ctrl done
	ret
.x:A
	cmp eax,byte AC_DEL
	jnz .r		;?; 32..126,128..
	xor eax,eax
	jr .f		;?; ascii 127 = <del>, delete
.r:A
	mov eax,-1
	ret
;-
;- ekey?  	( -- flg )			(GA:fax)(v)	'ekeyq'		forth
;-	test pending input, don't wait. 
;-	xec uot vector (ekey?); by default vectored to {?terminal}:
;-	buf. 1st available char to stdin <cdt.ibuf> & return tf, else ff.
ekeyq:GA;:ok:(xec ?terminal)
; c: ecx
	jmp [uref uv.uekeyq]; vectored character input
;-
;- ?ekey  	( -- c flg )					'qekey'		forth
;-	return any char c1 and tf, else invalid c1 and ff, don't wait
;-	tests pending input & copies currently available char to tos.
;-	does{ ekey? IF ibuf c@ ELSE false ENDIF dup 0=/= }
qekey:GA;:ok:
; c: ecx, eflags Z if no char
	call [uref uv.uekeyq]; tos{ flg
	push eax
	call getstdin
	mov eax,[kref ecx+cdt.ibuf-cdt]; char
	pop ecx
	spush
	mov eax,ecx
	test ecx,ecx	; set NZ
	ret
;-
;- key?	  	( -- flg )			(GA:fa)		'keyq'		forth
;-	test pending input, don't wait
;-	tf if pending ascii char in terminal/standard input
keyq:GA;:ok:
; c: ecx
; o: eflags
	call qekey	; NZ if char received, char from ?terminal
	TOSPULL(1)	; swap drop
	jz .r		;?; no key
	call cc4th
	spop
	call ekey2asci	; C*Z+NC*NZ
	spop
	jnz .r		;?; valid char
	call getstdin
.r:
	test eax,eax	; flg (Z+NZ)*NC
	ret
;-
;- ?key   	( -- c1 flg )			(83)		'qkey'		forth
;-	ascii range char c1 and tf, else invalid c1 and ff
;-	remove char from input queue
qkey:GA;:ok:<key)
; c: ecx, edx
	call keyq	; tf = char received
	call getstdin	;ecx:=incons-cdt (eflags!)
	test eax,eax
	push eax
	mov eax,[kref ecx+cdt.ibuf-cdt]; char
	spush
	pop eax		; flag
	jns .r
	and dword[kref ecx+cdt.iofl-cdt],~(mcflg); valid char off buf.
.r:
	ret
;
nky:A;:ok:-in accept etc-:check: at :eof:!
	call teststdin	;Z if stdin is not from console
	jnz .p
	test dword[kref ecx+cdt.rnum-cdt],-1; error code
	jnz .p		;?; not <eof>
	mov eax,AC_CAN	;EOF, preload with exit code
	cmp byte[byte uref uv.aux+klok],1
	jnz key.r
	mov byte[byte uref uv.aux+skey],0; flg := 1 if stdin is r/o file, etc
;	PRSTG'eofK'
	call otty
;	nop		;(alignment) for key
;	nop		;(alignment) for key
;    call iocons
.p:
	call wait4in
	spop
;-
;- key		( -- c )							forth
;-	ascii range char, wait until received (from stdin)
;-	open console if <eof> -?-
key:GA;:ok:(accept, etc.)
; c: ecx
	call qkey	; ecx,edx from <getstdin>
	test eax,eax
	spop
	jz nky
	or dword[kref ecx+cdt.iofl-cdt],mcrgt; retry after invalid char
.r:
	ret
;-
;- kbd?key   	( -- c1 flg )					'kbdqkey'	forth
;-	forcedly query the keyboard channel, -re- {?key}
kbdqkey:GA
	call kbdkey.x
	call qkey
	jr kbdkey.x
;-
;- kbd-key	( -- c )					'kbdkey'	forth
;-	{key} forcedly from keyboard channel, -re- {key}
kbdkey:GA;:ok: but echoing
	call .x
	call wait4in
	call key
.x:A
	push edx	; some kernel words expect edx preserved -!-
	call stdinc
	call kbdc
	call xgchan
	pop edx
	ret
;-
;- ekey>asci	( u -- u ff | c1 tf )				'ekey2asci'	forth
;-	{ekey>char} with lower low bound := 0, bounds from uvari (pchr).
ekey2asci:GA;:ok:
; c: ecx
; o: eflags C*Z+NC*NZ, eax, @nos
	spush
	mov ecx,[uref uv.upchr]
	mov cl,0
	jr ekey2char.t
;-
;- ekey>char	( u -- u ff | c1 tf )		(GA:fcx)	'ekey2char'	forth
;-	return tf if u is in range of printing chars, bounds from uvari (pchr).
ekey2char:GA;:ok:
; o: eflags C*Z+NC*NZ, eax,  @nos
	spush
.e:
	mov ecx,[uref uv.upchr]
.t:
	test ecx,ecx
	jz .s		; always valid if no range given
.c:
	cmp al,cl
	mov cl,0
	jb .n		;?; invalid below lower bound
	cmp al,ch
	jbe .s		;?; valid at and below upper bound
.l:
	shr ecx,16
	jnz .c		;?; try 2nd range
	add ecx,byte 1;;inc ecx		; not a valid char, force NZ, C already cleared -?-
%if 1
.s:
	setbe cl	; u<
.n:
	movzx eax,cl	; generate {true} value
	neg eax		; C*Z or NC*NZ
	ret
%else; -?- 
	mov eax,ecx	; Z not a valid char, force NZ, C already cleared -?-
	ret
.s:
.n:
	mov eax,0	; generate {true} value
	sub eax,1	; C*Z or NC*NZ
	ret
%endif
;-
;- kbd!		( n -- )					'kbdsto'	forth
;-	non-overwriting store char n back to keyboard channel charbuf,
;-	if input buf. empty, for later retrieval w. {kbd-key} etc.
kbdsto:GA
	mov ecx,kbdch
	shl ecx,cdt_bsize+2
	add ecx,[byte uref uv.chans]
	jr keysto.n
;-
;- key!		( n -- )					'keysto'	forth
;-	non-overwriting store char n back to std input charbuf,
;-	if input buf. empty, for later retrieval w. {key} etc.
keysto:GA;:ok:
; c: 	ecx
	call getstdin
.n:
	test dword[kref ecx+cdt.iofl-cdt],mcflg; valid char received flag
	jnz .r		;?; can't back-store
	inc dword[byte uref uv.blk]
	js .b		;?; stdin from file
	dec dword[byte uref uv.blk]
.b:
	mov [kref ecx+cdt.ibuf-cdt],eax; char
	or dword[kref ecx+cdt.iofl-cdt],mcflg; valid char received flag
	and dword[kref ecx+cdt.iofl-cdt],~(mcrgt); valid char received flag
.r:
	spop
	ret
;-
;- can?		( -- flg )					'canq'		forth
;-	tf if consecutive w@(uv.ucan+2) times the char w@uv.ucan received
;-	defaults to four times <ctrl>Y = ascii 25.
;-	NOTE: false flag if stdin is not a console channel.
canq:GA;:ok:
; c:	ecx
	call teststdin
	jz near false
	call stdinq
	spop
	jnz .r
	push dword[uref uv.ucan]
.l:
	call getstdin
	and dword[kref ecx+cdt.iofl-cdt],~(mcflg); force state of valid char received flag
	call qkey
	mov ecx,eax
	spop
	cmp al,[esp]	; EM:=:ctr/y, FS:=:altr/ctrl"\":$1c
	jnz .m
	dec word[esp+2]
	jz .t
.m:
	spop
	test ecx,ecx
	js .l
	inc ecx		; flags to NS and NZ
.t:
	lea esp,[byte esp+4];(flags!)
.r:
	setne cl
	spush
	dec cl
	movsx eax,cl
.q:
	ret
;-
;- esc?		( -- flg )			(e)		'escq'		forth
;-	tf if <esc> char input, if <bl> wait until other char
;-	NOTE: false flag if {kbd} is not a readable console channel.
;-	(e:flags reflect result)
; c:	ecx
escq:GA;:ok:
	call ipushio	; save i/o state
	call kbdc
;	spush		; dup
	call stdinc
	call tochan	; stdin := kbd
	call qkey	; ?key
	test eax,eax
	jz .r		;?; no key is no <esc>
	mov eax,TOS(1)
	cmp eax,byte ' '
	jnz .t		;?; don't wait
.b:
	spop		; drop
	call key	; key
	cmp eax,byte BL	; bl - until
	jz .b		;?; pause continues
.t:
	sub eax,byte ESC
	jz .x		;?; <esc>
	mov eax,1
.x:
	dec eax
.r:
	TOSPULL(1)
.q:
	call popio	; rst i/o state from RS
	test eax,eax
	ret
;-
;- kbd'		ccc( -- )			(i)		'kbdt'		forth
;-	cpl/interpret next word to executing w. stdin from console (keyboard).
;-	immediate words <ccc> will be rejected.
kbdt:GA;:ok:
	call blwf	; find	{ ix flg | ix 0
	spop
	jns near nimp	;?; not found, or imed
	call ccplq	; cpl?
	jz .i		;?; interpreting
	TICK bkbdt
	call compilec
	call compilec	; word to execute, cpl by index
	TICK kbdtb
	jmp compilec	; cpl kbd input mod.
.i:
	call bkbdt
	call execute
	call kbdtb
	ret
;-
;- kb]		(R: ch -- )			(k)		'kbdtb'		novoc
;-	restore from console input
kbdtb:A;:ok:<kbdt)
	jmp popio
;-
;- [kb		(R: -- ch )			(k)		'bkbdt'		novoc
;-	enter console input, leave new-chan on t.o.r.
bkbdt:A;:ok:<kbdt)
	spush
	pop eax
	call ipushio
	push eax
	spop
	call kbdc
	call stdinc
	call tochan
	jmp siocons
;-
;-----------------------------------------------------------------
;- -- i/o channels tests
;-	

;-
;- out-pty?	( -- flg )			(4)		'outptyq'	linux
;-	tf if STDOUT is to a pty-device, 
outptyq:GA;:ok:
	call stdoutc
;-
;- pty?		( c -- flg )			(4)		'ptyq'		linux
;-	query channel c by file-name, ret tf if c is a pty device
;-	set cdt.iofl in chp accordingly.
ptyq:GA;:ok:
	push eax
	call chid
	pop eax
	js near false
	push edx	; chp
	mov ecx,eax
	mov eax,dataptr
	spush
	mov eax,d_npad	; u p
	spush
	mov eax,ecx
	call chname
	call ptystgq.n
	pop edx		; chp
	and dword[kref edx+cdt.iofl],~(mptyf)
	or dword[kref edx+cdt.iofl],ecx; pty-flg, bit #25
	test eax,eax	; rst result & eflags
.r:
	ret
	nop ;(4-aligned -1)
ptystgq:;ecx:=pty-flag <open-file)
	call dup2
.n:;	;<pty?)
	call dup2
	spush
	mov eax,ptst-Y
	spush
	mov eax,ptst.e-ptst	; xx/pts/N (unix98 pty)
	call .t
	jnz .r
	mov eax,ttypt-Y		; /dev/ttypN ("old style BSD...")
	spush
	mov eax,ttypt.e-ttypt
.t:A
	call search
	mov ecx,eax
.r:
	TOSPULL(2)
	and ecx,mptyf
	ret
;-
;- iomode  	( -- )				(s)				linux
;-	setting stdin to non-blocking r/o, stdout to sync. w/o
;-	syscall fcntl -re- man 2 ...
iomode:GA
	call inid
	mov [byte uref uv.inchp],edx
	push ebx
	push edx
	sys_fcntl eax,F_SETFL,O_RDONLY|O_NONBLOCK
	pop edx
	mov ebx,[esp]
	or dword[kref edx+cdt.iofl],(mchrf)
	call outid
	mov [byte uref uv.outchp],edx
	push edx
	sys_fcntl eax,F_SETFL,O_WRONLY|O_SYNC
	pop edx
	pop ebx
	test eax,eax
	jns .r
	and dword[kref edx+cdt.iofl],~(mcons|micon);;;
.r:
	and dword[kref edx+cdt.iofl],~(mchrf)
	spop 2
	ret
;-
;- emit?  	( -- flg )			(a:fx)((s))	'emitq'		forth
;-	return tf if non-blocking output available,
;-	by syscall <access>, ret true if stdout is writable.
emitq:GA;:ok:
	call wok
	call stdoutc
	jr stdinq.n
;-
;- stdin?  	( -- flg )			(e)((s))	'stdinq'	forth
;-	test whether stdin readable, @tos := error code.
;-	does not affect stdin channel's handling state.
;-	(e: Z if stdin readable)
stdinq:GA;:ok:
; c: ecx	;<canq,keycall)
	call rok
	call stdinc
.n:
	call faccessfd
	test eax,eax
	not eax
	jz .r
	mov eax,0
.r:
	ret
;-
;- io-cons	( -- )				((s))		'siocons'	forth
;-	determine stdin and stdout handling state flags.
;-	should be executed after any new i/o re-direction.
siocons:GA;:ok:
; c: ecx
	call outconsq
	jz .r
	call inconsq.i
	jz .r
	mov byte[byte uref uv.aux+skey],-1
.r:
	spop
.x:
	ret
;-
;- io-cons?	( -- flg )					'ioconsq'	forth
;-	ret previously determined i/o handling state,
;-	tf if both, stdin and stdout connected to a console
ioconsq:GA;:ok:
	spush
	movsx eax,byte[byte uref uv.aux+skey]
	ret
;-
;- cons?  	( fd -- flg )			(s)		'consq'		forth
;-	tf if fd is a console channel.
;-	does not affect fd channel's handling state.
;-	by test w/ syscall ioctl,TIOCGSID
consq:GA;:ok:
; o: edx	cdt
	mov ecx,eax
	xor eax,eax
	call toutcons
	js .r
	dec eax
.r:
	ret
;-
;- <toutcons>	( -- )				(s)(u)(k)
;-	eflags: NZ if fd = bx is not a console
;-	used to check whether stdin is console.
;-	syscall ioctl -re- man 2 ...
toutcons:A;:ok:;<consq, outconsq, inconsq, intro)
; o: edx	cdt
;    eflags	NS,Z if not a console channel
	pushad
	push ecx
	mov eax,ecx
	call ptyq
	pop ecx
	mov eax,ecx
	call chid
	js .r
	writead edx,edx
	mov ebx,eax
	sys_ioctl EMPTY,TIOCGSID,PS;valid for the console as well as for pty-s
	test eax,eax
	jz .r		;;;
	readad edx,edx	;;;
	readad ebx,ebx
	and dword[kref edx+cdt.iofl],~(mcons|micon);;;
	test eax,eax	;;;
.r:
	popad
	ret
;-
;- in-cons?	( -- flg )			((s))		'inconsq'	forth
;-	tf if stdin is console,
;-	set stdin-type flag in cdt, clear flag in uvari (aux) if not a console
;-	syscall ioctl -re- man 2 ...
inconsq:GA;:ok:
; c: edx
;    eflags	NS,Z if not a console channel
	spush
.i:
	xor eax,eax
	mov ecx,stdin
.n:
	call toutcons
	js .r
	or dword[kref edx+cdt.iofl],mcons
	dec eax
	ret
.r:
	and dword[kref edx+cdt.iofl-cdt],~(mcons|micon);
	and byte[byte uref uv.aux+skey],0
	ret
;-
;- out-cons?	( -- flg )			((s))		'outconsq'	forth
;-	tf if stdout is console,
;-	set stdout-type flag in cdt, clear flag in uvari (aux) if not a console
;-	syscall ioctl -re- man 2 ...
outconsq:GA;:ok:
; c: ecx
	spush
.i:
	mov ecx,stdout
.e:
	xor eax,eax
	jr inconsq.n
;-
;- echo-on	( c -- )			((s))		'echon'		forth
;-	test whether c is a console channel, modify cdt-flags to echoing, if appropriate.
;-	w/ syscall ioctl,TIOCGSID
echon:GA;:ok:
	call chid
	mov ecx,eax
	call outconsq.e
	spop
	ret
;-
;- echo-off	( c -- )					'echof'		forth
;-	if c is a console channel, modify cdt-flags to non-echoing
echof:GA;:ok:
	mov ecx,~(mcons)
	shl eax,cdt_bsize+2
	add eax,[byte uref uv.chans]
	lea eax,[kref eax+cdt.iofl-cdt]
	cmp [eax],byte 0; flags
	js .r		;?; invalid
	and [eax],ecx
.r:
	spop
	ret
;
; testing console & fount ioctls 
; sp> pad dup 1000 erase 4th>abs $4b6b 0 sp@ 256     pad ! pad 4th>abs pad 4+ ! sys ioctl pad 333 dump ndump
%endif;__CONSIO_I
;-
;-=====================================================================;
;- l4consio.inc <eof>
