;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4cpl.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __CPL_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: compiler support
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __CPL_I
%define __CPL_I
%define cplinc

;-
;- udisp?  	( n -- n flg )					'udispq'	forth
;-	check whether disp n to uref(uot) is in range and 4-aligned.
udispq:GA;:ok:
	spush
	xor ecx,ecx
	call tudisp
	mov eax,ecx
	jnz .r
	dec eax
.r:
	ret
;-
;- data?  	( p -- fl )					'dataq'		hidden
;-	true flag if kref'd p points to allocated r/x data-memory.
dataq:GA;:ok:
	call drange
	sbb eax,eax
	not eax
	ret
;-
;- ?prog  	( p -- p | abort-21 )				'qprog'		forth
;-	abort if p not in job's allocated compile-memory
qprog:GA;:ok:
	spush
	call progq
	not eax
rpd:	; 'protected dictionary'
	spush
	mov eax,21
	jmp qerror
;-
;- prog?  	( p -- fl )					'progq'		hidden
;-	true flag if kref'd p points to allocated compile-memory.
progq:GA;:ok:
	call prange
	sbb eax,eax
	not eax
	ret
;-
;- xec?		( n -- addr flg )				'xecq'		forth
;-	test whether n is -most probably- an {execute}able item,
;-	either one of an uot index, kref'd or, true address of code entry.
;-	n = 0 is not considered valid, thus ix of {root} not recognized.
;-	ret flg (any non-zero value!) and executable true address.
xecq:GA;:ok:
; c: ecx
	test eax,eax
	mov ecx,eax
	spush
	jz .z		;?; zero value
	add ecx,KR
	jz .z		;?; zero ptr
	cmp eax,[byte uref uv.lkindex]; abs. max index, locals exclusively
	jnc .p		;?; u< 0=
	call idx2xec;(noreg)
	test eax,eax
.z:
	stc
	jz kernelq.r	;?; not in u.o.t.
	mov TOS(1),eax; xec
.p:
	call kprange
	jnc .k		;?; kref'd in range
	sub eax,KR
	jz krange.r	;?; zero ptr
	mov TOS(1),eax
.k:
	add TOS(1),KR
;-
;- kprog? 	( p -- fl )					'kprogq'	hidden
;-	true flag if kref'd p in kernel or allocated compile-memory.
kprogq:GA;:ok:
	call prange
	jnc kernelq.r
;-
;- kernel?	( p -- fl )					'kernelq'	hidden
;-	true flag if kref'd p points to library kernel memory.
kernelq:GA;:ok:
	call krange
.r:
	sbb eax,eax	; -1 if C
	not eax		; ff and Z(!) if C
	ret
; -
kprange:A
	call prange
	jnc krange.r
; check memory bounds
; C if addr <eax> is in library memory
krange:A;:ok:
	cmp eax,[kref mincode.b-Y]; bottom of xec l4 code
	jc .r
	cmp eax,YE-Y
	cmc		; C if mincode > p > top
.r:
	ret
;-
kdrange:A;:ok:	; kernel header or dataspace
	cmp eax,vocbase-Y
	jc drange
	cmp eax,tophead-Y
	cmc
	jnc drange.r
; C if addr <eax> in r/w data memory
drange:A;:ok:
	cmp eax,databot
	jc .r
	cmp eax,dataptr
	cmc
.r:
	ret
; C if addr <eax> in r/w/x compile-memory
prange:A;:ok:
	cmp eax,codebot
	jc .r
	cmp eax,codeptr
	cmc
.r:
	ret
;-
;- static?	( p -- fl )					'staticq'	system
;-	true flag if program statically linked to L4.
staticq:GA
	spush 2
	mov ecx,[KR]
	lea eax,[kref ecx+PAGE_SIZE-1]
	and eax,PAGE_MASK
	mov ecx,codebot
	mov TOS(1),eax
	lea eax,[kref ecx]
	and eax,PAGE_MASK
	jmp equal
;-	

;-
;-================================================================
;--- compiler - preliminary impl.
;-
;-
;-	cpl stores new {cp}, cinterpret calls bcp, restores cp from bcp
;-
;-	compiling variables:
;-
;-	initiated	from
;-		c0	system	at initial setup
;-		cp	c0	running ptr to actual posn of code compiling
;-		head	cp	backup before 1st entering compiling state
;-		tail	cp	backup before last entering compiling state
;-		bcp	head	backup before entering compiling state
;-		btail	cp	backup at initial entry into compilestate
;-
;-	copied to	from
;-		cp	explicit value by
;-				codespace alignment words and
;-				codespace allocation words,
;-			bcp	after immediate execution of standard words,
;-			btail	error exit, discarding erroneous code,
;-			c0	emergency cold (re-)start.
;-		head	bcp	when entering compilestate with {[}, 'lbrac'.
;-		tail	cp	when entering 1st nesting level
;-		bcp	explicit value by
;-				codespace allocation words
;-		btail	
;-
;-	compiling values, {state} bits & bitmasks re "constants.inc"
;-
;-
;-================================================================
;--- runtime primitives
;-

;-	

;-
;- dobcons	( -- n )			(k)
;-	fetch unsigned single byte cons from codespace
dobcons:A
	spush
	pop eax
	movzx eax,byte[eax]
	ret
;-
;- doscons	( -- n )			(k)
;-	fetch single cell cons from codespace
doscons:A;:ok:
	spush
	pop eax
	mov eax,[eax]
	ret
;-
;- dosvari	( -- p )			(k)
;-	fetch ptr to single cell vari from codespace
dosvari:A
	spush
	pop eax
	mov eax,[eax]
	add eax,UR
	sub eax,KR
	ret
;-
;- douvari	( -- p )			(k)
;-	fetch ptr to user vari
douvari:A;:ok:
	spush
	pop eax
	mov eax,[eax]
	lea eax,[uref eax]
	sub eax,KR
	ret
;-
;- doucons	( -- n )			(k)
;-	fetch cell content of user vari
doucons:A;ok;<w. defer-lk)
	spush
	pop eax
	mov eax,[eax]
	mov eax,[uref eax]
	ret
;-
;- (vr	  	( -- p )			(k)		'pdovar'	novoc
;-	return ptr to item in dataspace, kernel def'd {variable}
pdovar:A;:ok:
	spush
.n:
	mov eax,[kref ecx]
	add eax,databot
	ret
;-
;- (vl		( -- n )			(k)		'pdoval'	novoc
;-	fetch cell content from dataspace, kernel def'd {value}
pdoval:A;:ok:
	spush
.n:
%if 1
	mov ecx,[kref ecx]
	add ecx,databot
	mov eax,[kref ecx]
%else
	mov eax,[kref ecx]
	add eax,databot
	mov eax,[kref eax]
%endif
	ret
;-
;- <pdo2val>	( -- n )			(k)
;-	fetch double cell content from dataspace, kernel def'd {value}
pdo2val:A
	spush
	mov ecx,[kref ecx]
	add ecx,databot
	mov eax,[kref ecx]
	spush
	mov eax,[kref ecx+4]
	ret
;-
;- ulitb  	( -- u )			(k)
;-	fetch unsigned(!) literal byte from compilespace
ulitb:GA;:ok:
; c: ecx
%ifdef __K6__	; keeping callers call/ret sequence in order - timing test w/ no noticeable difference
	spush
	mov ecx,[esp]
	inc dword[esp]
	movzx eax,byte[ecx]
	ret
%else
	spush
	pop ecx
	movzx eax,byte[ecx]
	inc ecx
	jmp ecx
%endif
; (ulitb)
pulitbp:M
	spush
	xor eax,eax
	mov al,0
M pulitbp
;-
;- litw		( -- n )			(k)
;-	fetch signed literal word from compilespace
litw:GA;:ok:
; c: ecx
%ifdef __K6__
	spush
	mov ecx,[esp]
	add dword[esp],byte 2
	movsx eax,word[ecx]
	ret
%else
	spush
	pop ecx
	movsx eax,word[ecx]
	lea ecx,[ecx+2]
	jmp ecx
%endif
;-
;- lit		( -- n )			(m,6)				hidden
;-	fetch cell from compilespace
lit:GA;:ok:
; c: ecx
%ifdef __K6__
	spush
	mov ecx,[esp]
	add dword[esp],byte 4
	mov eax,[ecx]
	ret
%else
	spush
	pop ecx
	mov eax,[ecx]
	lea ecx,[ecx+4]
	jmp ecx
%endif
; (lit)
plitp:M
	spush
	mov eax,dword 0
M plitp
;-
;- flit			( -- df )						hidden
;-	fetch double from compilespace
; c: ecx
flit:GA
;-
;- dlit			( -- dn )						hidden
;-	fetch double from compilespace
dlit:GA;:preliminary:
; c: ecx
	spush
	pop ecx
	mov eax,[ecx]
	lea ecx,[ecx+4]
	spush
	mov eax,[ecx]
	lea ecx,[ecx+4]
	jmp ecx
;-
;- qlit			( -- qn )						hidden
;-	fetch quad from compilespace
qlit:GA;:ok:
%if 0;def __K6__	; no noticeable difference...
	mov ecx,[esp]
	mov [esp],esi
	mov esi,ecx
	mov [PS],eax
	movsd
	movsd
	movsd
	movsd
	mov esi,[esp]
	mov [esp],ecx
	ret
%else
	mov [PS],eax
	pop ecx
	push esi
	mov esi,ecx
	movsd
	movsd
	movsd
	movsd
	mov ecx,esi
	pop esi
	jmp ecx
%endif
;-
;- nlit			( -- .xx. N )						hidden
;-	compiled by {nliteral}, fetch bignum from compilespace
nlit:GA;:ok:
	spush
	mov eax,[esp]
	mov ecx,[eax]
	lea edx,[eax+ecx*4+4]
	mov [esp],edx
	sub eax,KR
	jmp nfetch
;-
;- (.")		( -- )(X: -- )					'pdotqtep'	hidden
;- (.()		( -- )(X: -- )							hidden
;-	print compiled <nul> terminated string, for {."} and {.(}
pdotqtep:GA;:ok:
; c: ecx
	spush
	pop eax
	sub eax,KR
	call ccount
	jz .s
	dec ecx
.s:
	add ecx,eax
	lea ecx,[kref ecx+1]
	push ecx
	jmp type
;-
;- (")		(X: -- p )			(fg)		'pqtep'		hidden
;- (c")		(X: -- p )			(a:ce)		'pqtep'		hidden
;-	compiled by {"} and {c"}, fetch non-alterable string ptr to stack
pqtep:GA;:ok:
; c: ecx
	spush
	pop eax
	sub eax,KR
	call ccount
	jz .s
	dec ecx
.s:
	add ecx,eax
	lea ecx,[kref ecx+1]
	spop
	dec eax		; ptr to cnt-1
	jmp ecx
;-
;- (s")		(X: -- a u )			(a:ce)		'psqtep'	hidden
;-	fetch non-alterable string ptr & cnt to stack
;-	compiled by {s"}, {z"} 
psqtep:GA;:ok:
; c: ecx
	spush
	pop eax
	sub eax,KR
	call ccount	; Z if asciz string, zero length or zero address doesn't apply
	jz .s		;?; asciz string
	dec ecx	
.s:
	add ecx,eax
	lea ecx,[kref ecx+1]
	jmp ecx
;-	

;-
;-----------------------------------------------------------------
;--- branches
;-	compiled displacement is (destn - {c-here} at posn of disp)
;-

;-
;- bcc						(n.i.)
;-	compare current eflags with compiled bytes pair,
;-	emulates { 1st eflags and 2nd xor 0= if ... },
;-	1st unmasks the relevant flagbits, 2nd defines
;-	which flag should be set to taking the branch.
bcc:GA
	jmp nimp
;-
;- optimizing branches, not available to hi-level 4th
;-
;- 0>branch					(f6)(k)		'gtbra'		hidden
;-	: 0> 0= IF
gtbra:GA;:continues:
	dec eax
	jns g32 pnbra
	jmp j32 pbra
;-
;- +branch					(f6)(k)		'plbra'		hidden
;-	: 0< IF
plbra:GA;:continues:
	test eax,eax
	js pnbra
	jmp j32 pbra
;-
;- 0-branch					(f6)(k)		'lebra'		hidden
;-	: 0>    IF
lebra:GA;:continues:
	dec eax
	jns pnbra
	jr pbra
;-
;- 0<branch	( n -- )(C: -- ptr 'bran' )	(fg)(k)		'mibra'		hidden
;-	: 0< 0= IF
mibra:GA
	test eax,eax
	js pbra
	jr pnbra
;-
;- 0branch	( n -- )(C: -- ptr 'bran' )	(fg)		'zbra'		hidden
;-	: IF
;-	conditionally branch if flag = 0, for WHILE, UNTIL
zbra:GA;:ok:
	test eax,eax
	jz pbra
pnbra:;:ok:	; entry from conditional branches, w. drop
	spop
nbra:
	pop ecx
	lea ecx,[ecx+4]
	jmp ecx
;-
;- ?0branch					(f6)(k)		'zdupbra'	hidden
;-	: -dup IF
;-	branch & drop if zero else dup
zdupbra:GA;:ok:
	test eax,eax
	jnz nbra
	jr pbra
;-
;- -branch	( n -- )(C: -- ptr 'bran' )	(fg)		'nzbra'		hidden
;-	: 0= IF
;-	conditionally branch if flag =/= 0, for IF
nzbra:GA;:ok:
	test eax,eax
	jz pnbra
pbra:;:continues:; entry from conditional branches, w. drop
	spop
;-
;- branch 	( -- )(C: -- )			(fg)				hidden
;-	unconditionally branch, for AGAIN, ELSE, REPEAT
branch:GA;:ok:
	pop ecx
	add ecx,[ecx]
	jmp ecx
;-
;- ?1branch					(f6)(k)		'nzdb'		hidden
;-	: -dup 0= IF
;-	discard @tos & enter true part if @tos = 0, else leave @tos
nzdb:GA
	test eax,eax
	jz pnbra
	pop ecx
	add ecx,[ecx]
	jmp ecx
;-
;- ?=branch					(f6)(k)		'neqbra'	hidden
;-	: over  = IF drop
;-	enter true part and 2drop if @tos = @nos, for {?case}
neqbra:GA;:ok:
	cmp eax,TOS(1)
	spop
	jz g32 pnbra
	pop ecx
	add ecx,[ecx]
	jmp ecx
;-
;--- hi-level loop runtimes
;-

;-
;- (+loop)	( n -- ) (XR: n1 n2 -- n1 n3 | -- )		'pploopp'	hidden
;-	increment loop index by n from datastack and branch
;-	until next index(n2) increment overflows or equals limit(n1).
pploopp:GA;:ok:
	test eax,eax
	mov ecx,TOR(0)	; loop index
	lea ecx,[ecx+eax]; + increment
	spop
	jz .n		;?; use +1 at zero inc
.c:
	mov TOR(0),ecx	; index
	js .p		; -ve inc
	cmp ecx,TOR(1)
	jc .b		; count not > limit
.r:
	mov ecx,[esp]	; fetch ret
	lea ecx,[ecx+4]	; advance past disp
%ifdef __K6__
	mov [esp+8],ecx
	lea esp,[esp+8]	; drop ret,idx,limit
	ret
.p:A
	cmp ecx,TOR(1)
	jc .r
.b:
	mov ecx,[esp]
	add ecx,[ecx]
	mov [esp],ecx
	ret
%else
	lea esp,[esp+12]; drop ret,idx,limit
	jmp ecx
.p:A
	cmp ecx,TOR(1)
	jc .r
.b:
	pop ecx
	add ecx,[ecx]
	jmp ecx
%endif
.n:A
	lea ecx,[ecx+1]
	jr .c
;-
;- (loop)  	(XR: n1 n2 -- n1 n3 | -- )			'ploopp'	hidden
;-	increment loop index by one and branch
;-	until next index(n2) increment equals limit(n1).
ploopp:GA;:ok:
	mov ecx,TOR(0)
	inc ecx		; loop index
	mov TOR(0),ecx
	cmp ecx,TOR(1)
	mov ecx,[esp]	; fetch ret
	jz .r
	add ecx,[ecx]
	mov [esp],ecx
	ret
.r:A
	lea ecx,[ecx+4]	; advance past disp
	lea esp,[esp+12]; drop ret,idx,limit
	jmp ecx
;-
;- (?do)  	( n1 n2 -- n1 n2 | )				'pqdop'		hidden
;-	if n1 =/= n2 push loop parameter to returnstack,
;-	else jump to after end of loop structure
pqdop:GA;:ok:
	pop ecx
	cmp eax,TOS(1)
	jnz .l		;?; proceed into loop
	spop 2
	add ecx,[ecx]
	jmp ecx
.l:
	lea ecx,[ecx+4]
	jr pdop.n
;-
;- (do)								'pdop'		hidden
;-	push loop parameter to returnstack
;-	adjust limit to limit+1 if start = limit, to looping once only
;-	does{ swap 2dup = IF 1+ ENDIF >r >r }
pdop:GA;:ok:
	pop ecx
.n:
	mov edx,TOS(1)	; limit
	cmp eax,TOS(1)
	lea edx,[kref edx]
	lea eax,[kref eax]
	jnz .p
	inc edx		; loop once only if n1 = n2
.p:
	push edx
	push eax
	spop 2
	jmp ecx		; ret
;-
;- (leave)					(m,8)		'pleavep'	hidden
;-	prepare loop parametres such that next {loop} will terminate,
;-	stores limit := (1+index) to leaving structure at next {loop}
pleavep:GA M;:ok:
	mov ecx,[esp+4]
	inc ecx
	mov [esp+8],ecx
	ret
M,pleavep
	mov ecx,[esp]
	inc ecx
	mov [esp+4],ecx
M pleavep
;-
;- (-leave)	( flg -- )			(fg)(m,15)	'pmleavep'	hidden
;-	conditionally prepare loop parametres such that next {loop} will terminate,
;-	if flg =/= 0 store limit := (1+index) to leaving structure at next {loop}
pmleavep:GA M;:ok:
; c: -/-
	test eax,eax
	spop
	jz .r
	mov ecx,[esp+4]
	inc ecx
	mov [esp+8],ecx
.r:
	ret
M,pmleavep
	test eax,eax
	spop
	jz .x
	mov ecx,[esp]
	inc ecx
	mov [esp+4],ecx
.x:
M pmleavep

;-
;-----------------------------------------------------------------
;--- compiler control
;-

;-	
;-
;--- state flags
;-	-re- {compiler} words and, source-file "constants.inc"
;-

;-
;- <hpts>					(k)
;-	increment state level cnt, backup head and tail ptrs
;-      called by {]}, {]]}, structures and {create}.
hpts:A;:ok:
; c: ecx
	cmp word[byte uref uv.state],0
	jg .s				;?; compiling
	js .r				;?; not valid
	mov ecx,codeptr			;	running ptr into compile-space
	cmp dword[byte uref uv.state],0
	mov [byte uref uv.tail],ecx	; cp -> tail, unchanged until next cpl start
	mov [byte uref uv.head],ecx	; cp -> head, unchanged until next cpl start
	jnz .t
	mov [byte uref uv.btail],ecx	; cp -> btail, currently latest executeable entry
.t:
	cmp word[byte uref uv.state],0	; rst flags
.s:
	mov ecx,codeptr
	mov [byte uref uv.bcp],ecx	; cp -> bcp, entry into latest executeable code
.r:
	ret
;-
;- <hptr>					(k)
;-	reset state, restore cp from backup tail
;-	by {quit}, {error}
hptr:A;:ok:
; c: ecx
	mov ecx,[byte uref uv.btail]	; remove compiled code
	mov dword[byte uref uv.state],0
	mov dword[uref uv.ucmc],0
	mov [byte uref uv.cp],ecx
	ret
;-	

;-
;- ]]		( -- )						'rrbrac'	compiler
;-	for structures start, instead of {]}.
;-	should be used for any construct designed to before teminating the 
;-	currently defined word permitting an intermediately interpreting sequence.
;-	use{ ]] start .. [ .. (interpreting) .. ] .. end [[ }.
rrbrac:GA;:ok:
; c: ecx
	mov dword[uref uv.ucmc],0
	call rbrac
	jr rbrac.s
;-
;- ]		( -- )						'rbrac'		forth
;-	enter compiling state.
;-	increment compile-state level by one,
;-	clear intermediately interpreting state.
rbrac:GA;:ok:
; c: ecx
	and dword[byte uref uv.state],~(mbrac)
	call hpts
	js .z
.s:
	inc word[byte uref uv.state]
	ret
.z:A
	mov dword[byte uref uv.state],0
	ret
;-
;- [[		( -- )				(i)		'llbrac'	compiler
;-	for structures termination, instead of {[}, re {]]}.
llbrac:GA;:ok:
; c: ecx
	and dword[byte uref uv.state],~(mbrac)
	dec word[byte uref uv.state]
;-
;- c[		( -- )						'clbrac'	root
;-	decrement compile-state level by one,
;-	leave state flags unchanged.
;-	at zero level of state execute until then compiled code.
;-	use{ ]] start .. ] ..next level cpl.. c[ .. end [[ }.
clbrac:GA;:ok:
; c: ecx
	xor ecx,ecx
	jr lbrac.n
;-
;- [		( -- )				(i)		'lbrac'		root
;-	decrement compile-state level by one,
;-	if non-zero @state set interpreting state to intermediate,
;-	else execute until then compiled code.
;-	cannot be used as a structures terminator - re {[[} !
lbrac:GA;:ok:
; c: ecx
	mov ecx,mbrac
.n:
	dec word[byte uref uv.state]
	jz .i
	js rbrac.z
	or dword[byte uref uv.state],ecx
.c:
	mov ecx,codeptr
	mov [byte uref uv.bcp],ecx
	ret
.i:A
	mov ecx,[byte uref uv.btail]
	add ecx,KR
	jmp ecx
;-
;- cpl		( -- )				(i)				hidden
;-	advance {c-here} to current codeptr
cpl:GA;:ok:
	or dword[byte uref uv.state],mword; flag persistency
.n:
	mov ecx,codeptr			; compile ptr
	mov byte[kref ecx],x_ret	; temporary <ret>
	mov [byte uref uv.btail],ecx	; ret from compiler
	mov [byte uref uv.tail],ecx	; lock codespace
	ret
;-
;- cpl;		( -- )				(i)				hidden
;-	terminate permanent code compilation
csemis:GA;:ok:
	and dword[byte uref uv.state],~(mword|mbrac)
	jr cpl.n
;-
;- <xcomp>	( -- )				(k)
;-	subroutine call, early ret to caller if not in compiling state
xcomp:A
	cmp word[byte uref uv.state],0
	jz .q		;?; interpreting
	test dword[byte uref uv.state],mbrac
	jz .r		;?; not intermediately interpreting
.q:
	lea esp,[esp+4]	; return to previous caller
.r:
	ret
;-
;- <ccplq>	( -- )				(e)(k)		'ccplq'
;-	subroutine call for {cpl?} & internal use,
;-	eflags only, compiling state if NZ.
ccplq:A;:ok:
	push ecx
	mov ecx,[byte uref uv.state]
	cmp cx,0
	jz .r		;?; interpreting
	xor ecx,mbrac
	test ecx,mbrac	; whether not intermediately interpreting
.r:
	pop ecx
	ret    
;-
;- cpl?		( -- flg )			(e)		'cplq'		root
;-	returns true if caller is in permanently compiling state
;-	(e: NZ if cpl)
cplq:GA;:ok:
	spush
	xor eax,eax
	call ccplq
	jz .r
	dec eax
.r:
	ret
;-
;- call-by-disp?  ( -- flg )					'cbdispq'	compiler
;-	true flag if call by immediate displacement compiling enabled.
cbdispq:GA;:ok:
	spush
	mov eax,dword[byte uref uv.bits]
	and eax,mcall
	jnz near rtrue
	ret
;-
;- call-by-disp	( -- )				(i)		'cbdisp'	compiler
;-	enable call by displacement compiling mode.
;-	code executes 10%+ (up to about 50%) faster than by ref to u.o.t. 
;-	and compiles one byte shorter code for 4th words calls (5 bytes).
;-	call-by-disp mode compiles calls by PC-relative displacements which 
;-	thus depends on a particular lib4th version.
cbdisp:GA;:ok:
	or dword[byte uref uv.bits],mcall
	ret
;-
;- call-by-uot	( -- )				(i)		'cbuot'		compiler
;-	enable call to indirect uot reference compiling mode.
;-	system default, 4th word calls are 6 bytes, ea.
;-	call-by-uot mode compiles indirect calls to uot cells by index 
;-	which is independent of a particular lib4th version.
;-	(re "lib4th.mac", <callx> and <jmpx> macroes for assembly level 
;-	(application programming.
cbuot:GA;:ok:
	and dword[byte uref uv.bits],~(mcall)
	ret
;-
;- macro  	( -- )				(i)		'cmacro'	compiler
;-	enable expanded code compiling, flag in (bits) uvari.
cmacro:GA
	or dword[byte uref uv.bits],mcmac
	ret
;-
;- end-macro	( -- )				(i)		'cendmacro'	compiler
;-	disable expanded code compiling, flag in (bits) uvari.
cendmacro:GA
	and dword[byte uref uv.bits],~(mcmac)
	ret
;-
;- mac-disp	( ix -- n )			(i)		'macdisp'	compiler
;-	ret n = disp of macro code from xec entry,
;-	-ve n if word cannot be compiled as a macro.
macdisp:GA;:ok:
	call idx2lfa
	call lfa2mfa
	movsx eax,word[kref eax]
	ret
;-
;- mac-size	( ix -- n )			(i)		'macsize'	compiler
;-	ret n = no. of bytes to be cpl'd in macro compiling mode,
;-	n = 0 if word cannot be compiled as a macro.
macsize:GA;:ok:
	call idx2lfa
	call lfa2mfa
	movzx eax,word[kref eax+2]
	ret
;-
;- asciz  	( -- n )					'hasciz'	compiler
;-	state-flag value for the asciz string compiling state,
;-	for {z"} & {place}, flg valid until next change of state.
hasciz:GA
	spush
	mov eax,masciz	; 28
	ret
;-
;-

;-
;-----------------------------------------------------------------
;--- codespace management, code insertion
;-	align & allot in dataspace
;-	calign & c-allot in codespace
;-	l-align & l-allot in local mem
;-
;-	

;-
;-	
;--- local mem
;-

;-
;- l-align  	( -- )				(a)		'xalign'	forth
;-	adjust ptr to local memory to lowest multiple of cellsize 
;-	greater or equal { lp @ }
lalign:GA
	mov edx,3
	mov ecx,[byte uref uv.lp]
	spush
	lea eax,[kref ecx]
	and eax,edx
	not edx			; edx=4:  -4
	jz palignm.r
	lea eax,[eax+edx-1]	;	 r-5
	not eax			;	-r+4
;-
;- l-allot  	( n -- | abort )				'lallot'	forth
;-	allocate +/- b bytes in local memory, abort if out of memory
lallot:GA
	mov ecx,[byte uref uv.s0]; top
	lea edx,[byte uref uv.lp]; ptr
	jmp j16 allot.n
;-
;-	
;--- data-space, extendable range [@d0,@d-top)
;-
;-	

;-
;- paligned  	( x -- x )					'palignp'	linux
;-	align-up true address/value @tos to memory page bounds
palignp:GA;:ok:
; c: -/-
	add eax,PAGE_SIZE-1
;-
;- paligned-	( x -- x )			(m,)		'palignm'	linux
;-	truncate true address/value @tos to memory page bounds
palignm:GA M;:ok:
; c: -/-
	and eax,PAGE_MASK
M palignm
.r:
	ret
;-
;- aligned	( x1 -- x2 )			(a:c)(m,)			forth
;-	adjust x2 to lowest multiple of cellsize greater or equal x1
aligned:GA M;:ok:
; c: -/-
	lea eax,[eax+3]
;-
;- aligned-	( x1 -- x2 )			(a:c)(m,)	'alignedm'	forth
;-	adjust x2 to lowest multiple of cellsize greater or equal x1
alignedm:M
; c: -/-
	and eax,byte -4
M alignedm
M aligned
	ret
;-
;- FALIGNED	( x1 -- x2 )			(a:f)(m)			float
;-	adjust true address/value x2 to lowest multiple of 
;-	f.p.-item size (double cell) greater or equal x1.
faligned:GA M;:ok:
; c: -/-
	lea eax,[eax+7]
	and eax,byte -8
M faligned
	ret
sfaligned: equ faligned
;-
;- DFALIGNED	( x1 -- x2 )			(a:f)(m)			float
;-	adjust true address/value x2 to lowest multiple of 
;-	f.p.-item size (double cell) greater or equal x1.
dfaligned:GA M;:ok:
; c: -/-
	lea eax,[eax+15]
	and eax,byte -16
M dfaligned
	ret
;-
;- naligned	( a1 +u -- a2 )							forth
;-	adjust true address/value a2 to lowest multiple of +u greater or equal a1
;-	abort if +u is not a power of 2.
naligned:GA;:ok:
; c: edx,ecx
	call pow2q
	jns near ram	;?; not a power of 2
	mov ecx,edx	; +u
	spop	
	lea eax,[eax+ecx-1]
	neg ecx
	and eax,ecx
	ret
;-
;- FALIGN  	( -- )				(a:e)				float
;-	adjust ptr to dataspace to lowest multiple of f.p. size
;-	greater or equal {here}, and erase allocated memory.
falign:GA;:ok:
; c: edx,ecx
	mov edx,7
	jr xalign.n
;-
;- DFALIGN  	( -- )				(a:e)				float
;-	adjust ptr to dataspace to lowest multiple of double f.p. size
;-	greater or equal {here}, and erase allocated memory.
dfalign:GA;:ok:
; c: edx,ecx
	mov edx,15
	jr xalign.n
;-
;- align  	( -- )				(a)		'xalign'	forth
;-	adjust ptr to dataspace to lowest multiple of cellsize 
;-	greater or equal {here}.
xalign:GA;:ok:
; c: edx,ecx
	mov edx,3
.n:
	mov ecx,dataptr
	spush
	lea eax,[kref ecx]
	and eax,edx
	not edx		; edx=4:  -4
	jz allot.r
	lea eax,[eax+edx-1];	 r-5
	not eax		;	-r+4
;-
;- allot  	( n -- | abort )						forth
;-	allocate +/- b bytes in dataspace, abort-12 if out of memory
allot:GA;:ok:
; o:	eax	
;	ecx	updated running ptr
; c: ecx,edx
	call qresizem	; test and, expand data memory if required
	lea edx,dataptr
	mov ecx,datatop
.n:
	add eax,[edx]
	lea ecx,[kref ecx]
	mov [PS],eax
	lea eax,[kref eax]
	cmp ecx,eax	; c:upper bound, a:new ptr posn
	mov ecx,[PS]
	jnc .a		;?; ptr fits
	mov eax,ENOMEM
	jmp errorf
.a:
	mov [edx],ecx
.r:
	spop
	ret
;-
;- ,		( n -- )					'comma'		forth
;-	allocate next cell in dataspace and store n
;-	abort-12 if unsufficient memory
comma:GA;:ok:
; c: edx,ecx
	spush
	mov eax,4
	call allot
	mov [kref ecx-4],eax
	spop
	ret
;-
;- w,		( n -- )					'wcomma'	forth
;-	allocate next half-cell in dataspace and store l.s. 16 bits of n
;-	abort-12 if unsufficient memory
wcomma:GA;:ok:
; c: edx,ecx
	spush
	mov eax,2
	call allot
	mov [kref ecx-2],ax
	spop
	ret
;-
;- b,		( n -- )					'bcomma'	forth
;-	allocate and store l.s.b. of n to next byte in dataspace
;-	abort-12 if unsufficient memory
bcomma:GA;:ok:
; c: edx,ecx
	spush
	mov eax,1
	call allot
	mov [kref ecx-1],al
	spop
	ret
;-
;- c,		( c -- )					'ccomma'	forth
;-	allocate and store c to next char in dataspace
;-	abort-12 if unsufficient memory
ccomma:GA;:ok:
; c: edx,ecx
	spush
	mov eax,b_c
	call allot
	mov [kref ecx-b_c],eax
	spop
	ret
;-	

;-
;-	
;--- code-space, fixed range [@c0,@d0)
;-	
;-

;-
;- cfalign	( -- )								forth
;-	adjust ptr to codespace to lowest multiple 
;-	of f.p.-item size (double cell) greater or equal {c-here}.
cfalign:GA;:ok:
; c: edx,ecx
	spush
	mov edx,8
	jr cnalign.n
;-
;- calign  	( -- )								forth
;-	adjust ptr to codespace to lowest multiple 
;-	of cellsize greater or equal {c-here}.
calign:GA;:ok:
; c: edx,ecx
	call pfour
;-
;- cnalign 	( n -- | abort-3)						forth
;-	align codeptr to nearest greater or equal multiple of n,
;-	abort if n not a power of 2.
cnalign:GA;:ok:
; ( n ) c-here 2dup + 1- rot negate and - negate c-allot
; c: edx,ecx
	call pow2q
	jns near ram		;?; not a power of 2
.n:
	mov ecx,codeptr
	mov eax,edx		; n
	lea ecx,[kref ecx]	; c-here addr
	lea edx,[ecx+eax-1]	; a+n-1
	neg eax
	and eax,edx		; a:=aligned ptr
	sub eax,ecx
;-
;- c-allot	( n -- | abort )				'callot'	forth
;-	allocate +/- b bytes in compilespace, alignment remaining unchanged.
;-	abort-12 if out of memory - which might abort due to a segfault, anyway...
callot:GA;:ok:?
; c: ecx,edx	ret ecx := new top ptr
	lea edx,codeptr	;ptr to ptr
	mov ecx,codetop	; upper bound
	call allot.n
	test dword[byte uref uv.state],mbrac
	jnz .b
	cmp word[byte uref uv.state],0
	jnz .b
	mov [byte uref uv.tail],ecx
.b:
	mov [byte uref uv.bcp],ecx
	mov byte[kref ecx],x_ret
	ret
;-
;- 4,		( n -- ) 					'cpl4'		forth
;-	compile dword n, append <ret> code, advance cp by 4
;-	allocate next cell in code-space and store n
cpl4:GA;:ok:
; c: edx,ecx
	mov ecx,4
	jr cpl1.n
;-
;- 2,		( n -- ) 					'cpl2'		forth
;-	compile l.s. word of n, append <ret> code, advance cp 2
;-	allocate next 2 bytes in code-space and store l.s. 16 bits of n
cpl2:GA;:ok:
; c: edx,ecx
	mov ecx,2
	jr cpl1.n
;-
;- 1,		( n -- ) 					'cpl1'		forth
;-	compile l.s.byte of n, append <ret> code, advance cp by 1
;-	allocate next byte in code-space and store l.s.byte of n
cpl1:GA;:ok:
	mov ecx,1
.n:; c: ecx,edx
	spush
	mov eax,ecx
	spush
	call callot
	sub ecx,eax
	spop
	mov [kref ecx],eax
	spop
ret
;-
;-
;------------------------------------------------------------------
;--- compiler-primitives, accessible at source assembly level, only
;-	
;-	NOTE: do not attempt to 'know' about the actually compiled
;-	code, since %macro-defs involved might have been modified!
;-	

;-
;-	
;--- compiling (executeable) data to codespace
;-

;-
;- <cplp> 	( n -- ) or ( -- ), (R: addr -- )	(k)
;-	compile from pattern, n=/=0 supplies variable bytes if required
;-	compiles to {cp}+, adjusts {cp} to end of code.
;-	NOTE: do not attempt to 'know' about the actually compiled
;-	code, since %macro-defs involved might have been modified!
cplp:A;:ok:
	pop edx			; ptr to code pattern
	movzx ecx,byte[edx]	; bytecount
	lea edx,[edx+2]
	call cpll
	movzx ecx,byte[edx-1]	; variable towards code end
	test ecx,ecx
	jz cpll.r		;?; no variable data
	sub codeptr,ecx		; adjust compilespace
	mov [PS],eax
	mov edx,PS		; vari to tmp storage
	spop
;-
;- <cpll>	(ecx:ptr edx:cnt) (R/S: -- )	(k)
;-	compile supplied length @nor of code @tor, -re- kernel word <cplp>
cpll:A;:ok:
	push edi,esi
	mov edi,ecx		; cnt
	mov esi,edx		; ptr
	mov ecx,edi
	mov edi,codeptr
	add edi,KR
	rep movsb
	mov byte[edi],x_ret	; terminate w. un-counted ret code
	sub edi,KR
	mov codeptr,edi
	pop edi,esi
.r:
	ret
;-
;- ---
;-
;-	c_<name>:	;compiling subroutine
;-		call cplp	; the code compiler
;-		db b1		; length of code
;-		db b2		; variable bytes within length, at end of code.
;-	d_<name>:	;compile patterns, don't execute!
;-		codebytes	; code- & vari-block to compile.
;-	.e:
;-

;-	

;-
;-	conditional jumps, dft ref'd to next location, i.e. disp = 0
;-	(B:"below", A:"above", L="less", G:"greater")
;-  O           0	O		(overflow flag set)
;-  NO          1	 		(overflow flag not set)
;-  C  B   NAE  2 U<	C		(carry flag set)
;-  NC NB  AE   3 U>=			(carry flag not set)
;-  Z  E        4 =	Z		(zero flag set)
;-  NZ NE       5 0=/=			(zero flag not set)
;-     BE  NA   6 U<=	C+Z		(either of carry or zero flags set)
;-     NBE A    7 U>	~C*~Z		(neither of carry or zero flags set)
;-  S           8 0<	S		(sign flag set)
;-  NS          9 0>=			(sign flag not set)
;-  P  PE      10	P		(even parity, flag set)
;-  NP PO      11			(odd parity, flag not set)
;-     L   NGE 12 <	S*~O+~S*O	(either one of sign or overflow set)
;-     NL  GE  13 >=			(neither one of sign or overflow set)
;-     LE  NG  14 <=	Z+S*~O+~S*O	(zero or either one of sign or overflow set)
;-     NLE G   15 >	~Z*(S*O+~S*~O)	(not zero and both, sign and overflow either unset or set)
;-

;-	

;- d_jccs  					(k)
;-	jcc short disp	$70(db)
c_jccs:A
	call cplp
	db 2,1
d_jccs:
	jo short .j
.j:
;- d_jccn  					(k)
;-	jcc near disp	$0f80(dd)
c_jccn:A
	call cplp
	db 2,1
d_jccn:
	jo near .j
.j:
;- d_jmps  					(k)
;-	jmp short disp	$eb(db)
c_jmps:A
	call cplp
	db d_jmps.e-d_jmps,1
d_jmps:
	jmp short .e
.e:
;- d_jmpn  					(k)
;-	jmp near disp	$e9(dd)
c_jmpn:A
	call cplp
	db d_jmpn.e-d_jmpn,4
d_jmpn:
	jmp near .e
.e:
;- d_calln  					(k)
;-	call dword disp	$e8(dd)
c_call:A
	call cplp
	db d_call.e-d_call,4
d_call:
	call .e
.e:
;-	

;-
;--- stack, exec.
;-
;- tospush  					(k)
;-	TOSPUSH(1)	$8d7f04
c_tospush:A
	call cplp
	db d_tospush.e-d_tospush,0
d_tospush:
	TOSPUSH(1)
.e:
;- tospull  					(k)
;-	TOSPULL(1)	$8d7ffc
c_tospull:A
	call cplp
	db d_tospull.e-d_tospull,0
d_tospull:
	TOSPULL(1)
.e:
;- spush  					(k)
;-	free @tos for data item	$ab
;-	(may change, according to %macro spush, file <lib4th.mac>)
c_spush:A
	call cplp
	db d_spush.e-d_spush,0
d_spush:
	spush		; 1
.e:
;- spush2  					(k)
;-	free @tos for double data item, $abab (stosd, stosd)
;-	(may change, according to %macro spush, file <lib4th.mac>)
c_spush2:A;:ok:
	call cplp
	db d_spush2.e-d_spush2,0
d_spush2:
	spush 2
.e:
;- spop						(k)
;-	drop @nos into @tos, $8B47FC8D7FFC
;-	patch -ve bytecount at byte {c-here}-1 and {c-here}-8, after compilation
;-	NOTE: code may be different, according to %macro spop, file <lib4th.mac>.
c_spop:A
	call cplp
	db d_spop.e-d_spop,0
d_spop:
	spop		; 6
.e:
;-
;- c-dodoes	( -- )				(k)		'c_dds'
;-	cpl dodoes frame, pfa to be patched in at start+2
;-	cpl: lea edx,[pfa]; lea ecx,[edx+4]; add edx,[kref edx]; add edx,KR; jmp edx
c_dds:
	call cplp
	db d_dds.e-d_dds,0
d_dds:
	mov edx,0x08000000	; pfa	 BA00000008
	lea ecx,[edx+4]		; body	 8D4A04
	add edx,[kref edx]	; xec	 031413
	add edx,KR		;	 01DA
	jmp edx			;	 FFE2
.e:
;- kcall  					(k)
;-	call [dword kref (disp)], $ff93(dd)
c_kcall:A
	call cplp
	db d_kcall.e-d_kcall,4
d_kcall:
	call [dword kref 0]
.e:
;- xcalla  					(k)
;-	call [uref 4*reg+uv.ultab], $ff9485(ltab+dd)
c_xcalla:A
	call cplp
	db d_xcalla.e-d_xcalla,0
d_xcalla:GL
	call [dword uref 4*eax+uv.ultab]
.e:
;- xcall  					(k)
;-	call [uref uv.ultab+(disp)], $ff95(ltab+dd)
c_xcall:A;:ok:
	call cplp
	db d_xcall.e-d_xcall,4
d_xcall:
	call [dword uref uv.ultab+0]
.e:
;- xjmp						(k)
;-	jmp [uref uv.ultab+(disp)], $ffa5(ltab+dd)
c_xjmp:A;:ok:
	call cplp
	db d_xjmp.e-d_xjmp,4
d_xjmp:
	jmp [dword uref uv.ultab+0]
.e:
;- ret 						(k)
;-	ret-code, $c3
c_ret:A
	call cplp
	db 1,0
d_ret:
	ret
;- itld						(k)
;-	load T register: mov edx,imed, $ba(dd)
c_itld:A;:ok: {code}
	call cplp
	db d_itld.e-d_itld,4
d_itld:
	mov edx,dword 0
.e:
;- iwld						(k)
;-	load W register: mov ecx,imed, $b9(dd)
c_iwld:A;:ok: {code}
	call cplp
	db d_iwld.e-d_iwld,4
d_iwld:
	mov ecx,dword 0
.e:
;- xtos						(k)
;-	mov eax,[uref uv.ultab+(disp)], $8b85(ltab+dd)
c_xtos:A
	call cplp
	db d_xtos.e-d_xtos,4
d_xtos:
	mov eax,[dword uref uv.ultab+0]
.e:
;- utosa  					(k)
;-	add eax,[uref (disp)], add user vari, $0385(dd-link)
c_utosa:A
	call cplp
	db d_utosa.e-d_utosa,4
d_utosa:
	add eax,[dword uref 0]
.e:
;- utos 					(k)
;-	mov eax,[uref (disp)], $8b85(dd-link)
c_utod:A
	call cplp
	db d_utos.e-d_utos,4
d_utos:
	mov eax,[dword uref 0]
.e:
;- ktos1  					(k)
;-	mov ecx,[kref (disp)], $8b8b(dd)
c_ktos1:A;:ok:
	call cplp
	db d_ktos1.e-d_ktos1,4
d_ktos1:
	mov ecx,[dword kref 0]
.e:
;- ktosm  					(k)
;-	mov TOS(1),ecx, $894ffc
c_ktosm:A;:ok:
	call cplp
	db d_ktosm.e-d_ktosm,0
d_ktosm:
	mov TOS(1),ecx
.e:
;- ktos						(k)
;-	mov eax,[kref (disp)], $8b83(dd)
c_ktos:A;:ok:
	call cplp
	db d_ktos.e-d_ktos,4
d_ktos:
	mov eax,[dword kref 0]
.e:
;- itos1  					(k)
;-	mov TOS(1),imed, $c747fc(dd)
c_itos1:A;:ok:
	call cplp
	db d_itos1.e-d_itos1,4
d_itos1:
	mov dword TOS(1),dword 0
.e:
;- tos						(k)
;-	mov eax,imed, $b8(dd)
c_itos:A;:ok:
	call cplp
	db d_itos.e-d_itos,4
d_itos:
	mov eax,dword 0
.e:
;- c-litp   	( n -- ) 			(k)		'c_litp'
;-	cpl single literal, add eax,dword n, $abb8(dd)
c_litp:A;:ok:
	call cplp
	db d_litp.e-d_litp,4
d_litp:
	add eax,dword 0
d_litp.e:
;-
;- c-plsto   	( n -- n ) 			(k)		'c_plsto'
;-	prepare cpl add and drop @tos to "value", add [kref dword n],eax\ <spop>
;-	caller must save c-here and store n to this ptr+2, after compilation,
;-	re {to} implementation for an example.	$8983(dd)8d7ffc8b07
c_plsto:A;:ok:
	call cplp
	db d_sto.e-d_sto,0
d_plsto:
	add [dword kref 0],eax
	spop
d_plsto.e:
;-
;- c-sto   	( n -- n ) 			(k)		'c_sto'
;-	prepare cpl store and drop @tos to "value", mov [kref dword n],eax\ <spop>
;-	caller must save c-here and store n to this ptr+2, after compilation,
;-	re {to} implementation for an example.	$8983(dd)8d7ffc8b07
c_sto:A;:ok:
	call cplp
	db d_sto.e-d_sto,0
d_sto:
	mov [dword kref 0],eax
	spop
d_sto.e:
;-
;- c-valu 	( n -- ) 			(k)		'c_valu'
;-	cpl single "value": <spush> : mov eax,[kref dword n], $ab8b83(dd)
c_valu:A;:ok:
	call cplp
	db d_valu.e-d_valu,4
d_valu:
	spush
	mov eax,[dword kref 0]
d_valu.e:
;-
;- c-lit   	( n -- ) 			(k)		'c_lit'
;-	cpl single literal, <spush> mov eax,dword n, $abb8(dd)
c_lit:A;:ok:
	call cplp
	db d_lit.e-d_lit,4
d_lit:
	spush
	mov eax,dword 0
d_lit.e:
;-	
;-

;-
;- c-qvlit  	( qv -- )			(k)		'c_qvlit'
;-	cpl quad ranum (four cells' pair) literal
c_qvlit:A
	call swap4
	call c_qlit	;:continued: c_qlit
;-
;- c-qlit  	( q -- )			(k)		'c_qlit'
;-	cpl quad literal
c_qlit:A;:ok:
	call swap2
	call c_dlit	;:continued: c_dlit
;-
;- c-dlit 	( d -- ) 			(k)		'c_dlit'
;-	cpl double literal
c_dlit:A;:ok:
	mov ecx,eax		; swap
	mov eax,TOS(1)
	mov TOS(1),ecx
;-
;- c-flit 	( f -- ) 			(k)		'c_flit'
;-	cpl double precision f.p. literal (n.i.)
c_flit:A;(:ok:) by c_dlit
	call c_spush2
	call c_itos1
	jmp disp c_itos
;-
;- c-locv 	( n -- ) 			(k)(7)		'c_locv'
;-	cpl single local "value",
;-	<spush> : mov eax,[lref disp]
c_locv:A;:ok:<dolocc)
	call cplp
	db d_locv.e-d_locv,4
d_locv:
	spush				; 1
	mov eax,[LP+0x08000000]		; 6 => 7
d_locv.e:
;-
;- c-locp 	( n -- ) 			(k)(10)		'c_locp'
;-	cpl ptr to a local "value",
;-	<spush> : mov eax,LP: sub eax,KR: sub eax,disp
c_locp:A;:ok:<locptr)
	call cplp
	db d_locp.e-d_locp,4
d_locp:
	spush				; 1
	mov eax,LP			; 2
	sub eax,KR			; 2
	sub eax,0x08000000		; 5 => 10/9
d_locp.e:
;-
;- c-locto 	( -n -- ) 			(k)(13)		'c_locto'
;-	cpl local "to",
;-	mov ecx,@tos; spop;mov [lref disp],ecx
c_locto:A;:ok:
	call cplp
	db d_locto.e-d_locto,4
d_locto:
	mov ecx,eax			; 2
	spop				; 5
	mov [LP-0x08000000],ecx		; 6 => 13
d_locto.e:
;-
;- c-plocto 	( -n -- ) 			(k)(13)		'c_plocto'
;-	cpl local "+to",
;-	mov ecx,@tos; spop; add [lref disp],ecx
c_plocto:A;:ok:
	call cplp
	db d_plocto.e-d_plocto,4
d_plocto:
	mov ecx,eax			; 2
	spop				; 5
	add [LP+0x08000000],ecx		; 6 => 13
d_plocto.e:
;-
;- c-nlit	( .x. N -- )			(k)		'c_nlit'
;-	cpl counted integer
c_nlit:A
	TICK nlit
	call compilec
	lea ecx,[eax*4+4]
	spush
	mov eax,codeptr
	add ecx,eax
	mov codeptr,ecx
	mov byte[kref ecx],x_ret
	jmp nstore
;-
;- -compile,	( ix -- )					'mcompilec'	forth
;-	compile call to ix if in compiling state, else execute.
mcompilec:GA;:ok:
	call ccplq
	jz near execute
	jr bcb.c
;-
;- [compile]	ccc( -- )			(i)		'bcb'		forth
;- POSTPONE	ccc( -- )			(i)		'bcb'		ans
;- [_]		ccc( -- )			(i)(h4)		'bcb'		root
;-	suppress immediacy of next word <ccc> in input stream or,
;-	compile non-immediate word to being compiled at runtime.
;-	([compile], postpone and, [_] execute all the same code)
bcb:GA;:ok:
	call xcomp
	call dfind
	spop
	movzx eax,word[kref eax-whead.l+whead.i];lfa>idx
	jz compilec.r	;?; ret r.ni
	jns compilec	;?; compile imed word
	call literal
	TICK compilec	; cpl { literal compile, }
.c:
	mov dword[uref uv.ucmc],0;;;
;-
;- compile,	( ix -- )			(a:cx)		'compilec'	forth
;-	cpl subroutine call by execution token ix:
;-	call [uref disp], w. disp=ix*4+uv.ultab-(uv.link)
compilec:GA;:ok:
	cmp eax,[byte uref uv.lkindex]
.r:
	jnc near r.ni
.c:
	call ccplq
	jz .f		;?; not compiling
;:HIER:  whether macro, via find+{bits}-flg <mfmac> -?- :HERE: ;
;    test dword[byte uref uv.bits],mfmac
;    jnz cplxmac
;    btr dword[byte uref uv.bits],bfmac
;    jc cplxmac
;;.i:
	test dword[byte uref uv.bits],mcall
	jnz cplxcall
.f:
;;	cmp dword[eax*4+uv.ultab-(uv.link)],0;
;;	jz cplxpfa;
%if 0
	lea eax,[eax*4+uv.ultab-(uv.link)]
	jmp c_xcall
%else
	lea eax,[eax*4+uv.ultab-(uv.link)]
	cmp dword[uref eax+uv.link],0
	jz cudds		;?; xec by ref to pfa
	jmp c_xcall
%endif
;-
;- [,]		( ix | ix s1 -- | abort )	(i)		'bcompilecb'	root
;-	{[,]} takes uot index from tos or from below colon compiler structure
bcompilecb:GA;:ok:-?-
	cmp eax,[byte uref uv.lkindex]
	jc compilec.c	; {compile,}
	call ccplq
	jz compilec.r
	mov eax,TOS(2)	; try item below colon compiler structure
	jr compilec
; compile pc-relative immediate jump
; eax = k-rel. ptr to code entry
;;cplijmp:A	;
;;	mov ecx,codeptr	;
;;	lea ecx,[ecx+5]	; dst := pc after disp-dw of jmp op-code
;;	sub eax,ecx	;
;;	jmp c_jmpn	; cpl : jmp <dovoc>
;
; compile pc-relative immediate call
; eax = uot index
cplxcall:A
	call idx2xec	; Z if not found
	jz near r.am
	btr dword[byte uref uv.bits],bfmac
	jc cplxmac
; compile pc-relative immediate call
; eax = k-rel. ptr to code entry
cplicall:A;:ok:<vocabulary)
	mov ecx,codeptr
	lea ecx,[ecx+5]	; dst := pc after disp-dw of call op-code
	sub eax,ecx
	jmp c_call		; cpl : call <do-word>
cplxmac:A
	jr compilec.f;;i	;>;macro cpl yet not implemented
; <cudds>
;	entry to <cdodoes> from {compile,}
cudds:A
	sub eax,uv.ultab-(uv.link)
	sar eax,2		; uds>idx
; <cdodoes>
;	for <pfax> words: call from pfa, pass ecx := body to xec routine.
;	cpl: lea edx,[pfa]; lea ecx,[edx+4]; add edx,[kref edx]; add edx,KR; jmp edx
cdodoes:
	push dword[byte uref uv.cp]
	call c_dds
	call idx2pfa
	pop ecx
	inc ecx
	mov [kref ecx],eax
	mov eax,[byte uref uv.cp]
	mov word[kref eax],(x_ret<<8)|x_ret
	spop
statex.r:
	ret
;-
;- statex	( ..xx.. -- ..yy.. )		(k)
;-	if @state =/= 0 execute from cell where RS points to,
;-	else do nothing. does{ r> state@ IF execute ENDIF }
statex:A;:ok:
	call ccplq	; Z if interpreting
	pop ecx
	mov ecx,[ecx]
	jz .r
	lea ecx,[kref ecx]
	jmp ecx
;-
;-================================================================
;--- compiler
;-	process control, immediate & defining words
;-
; ---------------------------------------------------------------------------- ;
; NOTE:
;	words which compile more than 6 bytes must remove ret-address
;	to codespace from returnstack to returning immediately or,
;	adjust return address to kref<ecx> as returned from <csemis>.
;	this limit reduced to 5 bytes in call-by-disp mode!
; ---------------------------------------------------------------------------- ;
;

;-	

; ecx := latest words flag-bits address, NZ if headerless
latestf:A
	mov ecx,[byte uref uv.current]	; context @
	lea ecx,[ecx-vhead.m+vhead.p]	; voctop
	sub ecx,[kref ecx]		; disp to..
	lea ecx,[kref ecx+whead.f-whead.l]; ..top lfa
	test dword[byte uref uv.state],mnnam
	ret
;-
;--- wordheader flags modifiers
;-

;-
;- smudge  	( -- )				(e)				fig
;-	toggle smudge bit of -re- {latest} word-header.
;-	(e: Z flag reflects smudge state)
smudge:GA;:ok:
; c: ecx
	call latestf
	jnz .r
	xor word[ecx],smud
.r:
	ret
;-
;- unsmudge	( -- )								root
;-	mark -re- {latest} word-header ready for -re- {find}.
unsmudge:GA;:ok:
; c: ecx
	call latestf
	jnz .r
	and word[ecx],~(smud)
.r:
	ret
;-
;- immediate	( -- )								root
;-	modify -re- {latest} word to immediately executing. -re- {imed?}
immediate:GA;:ok:
; c: ecx
	call latestf
	jnz .r
	or word[ecx],imed
.r:
	ret
;-
;- casedep	( -- )								compiler
;-	modify -re- {latest} word to only case dependently being found.
casedep:GA
; c: ecx
	call latestf
	jnz .r
	or word[ecx],wcas
.r:
	ret
;-
;- no-opt	( -- )				(i)		'noopt'		root
;-	xec: modify -re- {latest} word to not optimizing jmp for ret.
;-	cpl: suppress latest word's jmp-modification by resetting the cpl'd 
;-	     words' counter to zero, uvari {(cmc)}.
noopt:GA;
; c: ecx
	call ccplq
	jnz .c
	call latestf
	jnz .r
	or word[ecx],nopt
.r:
	ret
.c:
	mov dword[uref uv.ucmc],0
	ret
;-
;- - header flags
;-	set   ( flag lfa -- ) { lfa>ffa or! }
;-	clear ( flag lfa -- ) { swap not swap lfa>ffa and! }
;-

;-	
;- cmac		( -- n )					'hcmac'		compiler
;-	header flag: compile-macro, bitmask to compiling inline code.
hcmac:GA;:ok:
	spush
	mov eax,cmac	; 15
	ret
;-
;- imed		( -- n )					'himed'		compiler
;-	header flag: immediately executing, set by {immediate}; -re- {imed?}
himed:GA;:ok:
	spush
	mov eax,imed	;  7
	ret
;-
;- ixec		( -- n )					'hixec'		compiler
;-	header flag: do not compiled execute when in interpreting state;
;-	provision for the emergency case - e.g. {bye}, {cold}.
hixec:GA;:ok:
	spush
	mov eax,ixec	; 10
	ret
;-
;- pfax		( -- n )					'hpfax'		compiler
;-	header flag: execute from pfa in header, pass ptr to body in ecx
hpfax:GA;:ok:
	spush
	mov eax,pfax	;  6
	ret
;-
;- nopt		( -- n )					'hnopt'		compiler
;-	header flag: suppress backward code optimizing
hnopt:GA;:ok:
	spush
	mov eax,nopt	; 14
	ret
;-
;- smud		( -- n )					'hsmud'		compiler
;-	header flag: word finding disabled, un-set by {unsmudge}, toggled by {smudge}
hsmud:GA;:ok:
	spush
	mov eax,smud	;  5
	ret
;-
;- inif		( -- n )					'hinif'		compiler
;-	header flag: word which requires initialization from header data
;-	@pfa is ptr to init. code wrt beginning of module
hinif:GA;:ok:
	spush
	mov eax,inif	; 11
	ret
;-
;- varf		( -- n )					'hvarf'		compiler
;-	header flag: 2nd-ary pfa holds disp to beginning of variables dataspace.
;-	used for default values and ptr's of kernel defined variables
hvarf:GA;:ok:
	spush
	mov eax,varf	; 12
	ret
;-
;- vocf		( -- n )					'hvocf'		compiler
;-	header flag: word is vocabulary, fieldcount is no. of wordheaders.
;-	set by -re- {vocabulary} or {wordlist} definition.
hvocf:GA;:ok:
	spush
	mov eax,vocf	;  4
	ret
;-
;- wcas		( -- n )					'hwcas'		compiler
;-	header flag: word may only be found case dependently
hwcas:GA;:ok:
	spush
	mov eax,wcas	;  8
	ret
;-
;- xwfl		( -- n )					'hxwfl'		compiler
;-	header flag: additional parameter-fields count mask (0..15),
;-	no. of dwords (cells) after concerning word header's pfa.
hxwfl:GA;:ok:
	spush
	mov eax,xwfl	;  0..3
	ret
;-
;- imed?  	( n -- f )					'imedq'		compiler
;-	ret f =/= 0 if lfa or ix n points to an -re- {immediate} word.
;-	other flags -re- 
;-		{cmac} {imed}, {ixec}, {pfax}, {nopt}, {smud}, 
;-		{inif}, {varf}, {vocf}, {wcas}, {xwfl}
;-	flags set by the rsp. defining words; 
;-	a program should safely modify no other than {imed}, {smud}, {wcas}:
;-		set   ( flag lfa -- ) { lfa>ffa or! }
;-		clear ( flag lfa -- ) { swap not swap lfa>ffa and! }
imedq:GA;:ok:
; c: ecx
	call qlfa	; lfa2xec
	jnz .r
	call lfa2ffa
	mov eax,[kref eax]
	and eax,imed
.r:
	ret
;-
;-----------------------------------------------------------------
;- -- compiler
;-	

;-
;- -compile	ccc( -- )					'mcompile'	forth
;-	compile call to ccc if in compiling state, else execute.
mcompile:GA;:ok:
	call ccplq
	jnz compile
	ret
;-
;- compile	( -- | abort-27 )						forth
;-	at runtime compiles next compiled word, abort-17 if not compiling.
;-	don't use in macro cpl mode: {compile} is not an immediate word and
;-	thus cannot catch any errors which might result from false opcodes.
;-	re {compile,} for compiling by execution token (ix, uot index).
compile:GA;:ok:<mcompile)
	call qcomp
	mov dword[uref uv.ucmc],0
	spush
	pop ecx			; ret & cplptr
	mov dx,[ecx]
	mov eax,[ecx+2]
	lea ecx,[ecx+6]		; advance over next xec uot-ref'd word
	cmp dx,x_ucall
	jz .u			;?; call by uot
	cmp dx,x_ujmp
	jz .u			;?; jmp by uot
	dec ecx
	mov dl,[ecx-5]
	cmp dl,x_dcall
	jz .d			;?; call by disp
	cmp dl,x_djmp
	jnz r.sa		;?; not  jmp by disp
.d:
	mov eax,[ecx-4]
	add eax,ecx		; xec address
	sub eax,KR		; ptr
	push ecx
	call xec2idx
	call idx2uds
	pop ecx
.u:
	push ecx		; adjusted ret
	push eax		; uds
	call uds2idx
	push eax		; ix
	call idx2uds
	mov ecx,eax
	pop eax			; ix
	pop edx			; uds
	cmp ecx,edx
	jz near compilec
r.sa:
	mov eax,27		; "sys addr error"
	jmp errorf
;-
;- [CHAR]  	ccc(C: -- )(X: -- c )		(i)		'char'		ans
;-	compile leading char of next word in input stream as a literal
;- ascii  	ccc(C: -- )(X: -- c )		(i)		'char'		forth
;-	dependent on @{state} compile or... -re- {CHAR} in ans-voc.
;- CHAR		ccc(C: -- )(X: -- c )						ans
;-	fetch to stack the 1st char of next word in input stream as a literal.
;-	control chars can be fetched if (upper case) char prefixed w/ "^".
char:GA;:ok:
	call BLWORD
	cmp byte[kref eax],1
	mov ecx,eax
	jc near rfalse
	movzx eax,byte[kref ecx+1]
	cmp al,'^'
	jnz literal
	cmp byte[kref ecx],1
	jna literal
	movzx eax,byte[kref ecx+2]
	sub al,64	;:continued:
;-
;- literal	(C: x1 -- )(X: -- x1 )		(i)				forth
;-	cpl: compile sing as a literal, interp: noop
literal:GA;:ok:
	call statex
	dd c_lit-Y
;-
;- 2literal	(C: d1 -- )(X: -- d1 )		(i)		'dliteral'	forth
;-	cpl: compile double as a literal, interp: noop
dliteral:GA;:ok:
	call statex
	dd c_dlit-Y
;-
;- 2LITERAL	(C: f1 -- )(X: -- f1 )		(i)		'adliteral'	ans
;-	cpl: compile float item (double cell) as a literal, interp: noop
;-	ANS mode, reversed (big endian) cells order.
adliteral:GA;:ok:(=2literal)
	call statex
	dd c_flit-Y
;-
;- 4literal	(C: q1 -- )(X: -- q1 )		(i)		'qliteral'	forth
;-	cpl: compile quad (four cells) as a literal, interp: noop
qliteral:GA;:ok:
	call statex
	dd c_qlit-Y
;-
;- qvliteral	(C: qv1 -- )(X: -- qv1 )	(i)		'qvliteral'	rational hidden
;-	cpl: compile four cells' pair (quad ranum) as a literal, interp: noop
qvliteral:GA;:ok:
	call statex
	dd c_qvlit-Y
;-
;-----------------------------------------------------------------
;--- imed, compiler
;-	immediate execution blocking until resolved
;-
;-	

;-
;- drop;s  	( n -- )			(m,)(k)		'drops'		hidden
;-	{drop} and {;s}
drops:GA M;:ok:
	lea esp,[esp+4]
M,drops
	spop
	ret
M drops
;-
;- -exit  	( flg -- )			(i)		'dexit'		forth
;-	cpl/execute conditional {exit}, to exitting a word if flg =/= 0.
dexit:GA
	call ccplq
	jz pdexitp
	call locrmv
	TICK pdexitp
	jmp compilec
;-
;- (-exit)  	( flg -- )			(m,8)		'pdexitp'	hidden
;-	exit if flg =/= 0
pdexitp:GA M;:ok:
	test eax,eax
	spop
	jz .r
	lea esp,[esp+4]
.r:
	ret
M,pdexitp
	test eax,eax
	spop
	jz .x
	ret
.x:
M pdexitp
;-
;- ;s		( -- )(R: a -- )		(m,1)		'semis'		forth
;-	immediately leave current & return to calling word at true address a.
;-	(e: no flags affected, pop RS & ret)
semis:GA M;:ok:
	lea esp,[esp+4]
M,semis
	ret
M semis
;-
;- exit		( -- )(R: a -- )		(i)		'exitf'		forth
;-	immediately leave current word & return to caller.
;-	compiles local values de-allocation and apropriate return opr.
;-	modifies last <call & ret> sequence to jmp into last word, 
;-	dependent on content of {(cmc)} vari and {nopt} header-flag.
;-	interpretive compiling locals de-allocation left to {;}.
;-	-re- conditional {-exit}, {else;}, {endif;}
exitf:GA;:ok:
	call ccplq
	jz semis		;?; interpreting
	call locrmv		; cpl locals de-alloc
cplxit:A;<semi)
; modify last call to jmp
	cmp dword[uref uv.ucmc],0
	jg .o
	mov edx,[byte uref uv.current]
	lea edx,[kref edx+vhead.p-vhead.m]
	sub edx,[edx]
	movzx edx,word[edx-vhead.l+vhead.i]
	cmp edx,[uref uv.uprevix]
	jnz .c			;?; not a recursion
.o:;	;<bfnameb)
	mov dword[uref uv.ucmc],0
	mov ecx,codeptr
	test dword[byte uref uv.bits],mcall
	lea ecx,[kref ecx-6]
	jnz .d
	cmp word[ecx],x_ucall
	jnz .c
	mov word[ecx],x_ujmp	; leave <ret> in place in case of final {endif}, etc.
	jr .c
.d:
	cmp byte[ecx],x_ucall&255
	jz .c
	inc ecx
	cmp byte[ecx],x_dcall
	jnz .c
	mov byte[ecx],x_djmp
.c:
	jmp c_ret		; include terminating <ret>
%endif;__CPL_I
;-
;-=====================================================================;
;- l4cpl.inc <eof>
