;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4cstruct.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __CSTRUCT_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: control structures compiler
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __CSTRUCT_I
%define __CSTRUCT_I
%define cstructinc

;-
;- mark		( -- ptr )			(79)				forth
;-	push current {c-here} for later resolution of back reference
mark:GA;:ok:
	spush
	mov eax,codeptr
	ret
;-
;- back		( ptr -- )					'back'		forth
;-	cpl back reference at {c-here}, append <ret>.
back:GA;:ok:
	sub eax,codeptr	; disp from after c-here to after ptr
.c:
	call cpl4
	mov byte[kref ecx+4],x_ret; terminate w. rts
	ret
;-
;- ahead  	( -- ptr )			(a:te)				forth
;-	push current {c-here},
;-	reserve cell in code-space for branch disp, append <ret>.
ahead:GA
	spush
	mov eax,codeptr
	spush
	jr back.c
;-
;- >resolve	( ptr -- )			(79)		'toresolve'	forth
;-	resolve forward branch from ptr to {c-here}
toresolve:GA;:ok:
	mov ecx,codeptr	; disp
	sub ecx,eax
	mov [kref eax],ecx
	spop
	ret
;-
;- recurse	( -- )				(i)(a:c)			forth
;-	recursion (self call) efficient wrt programming effort, but... (resources!)
;-	one exception, though, is "tail recursion", which lib4th compiles to a branch.
;-	cannot be used with headerless words; should not be used in a {does>} part.
;-	recursive execution of run-time part in a defining word should be set up 
;-	separately to circumventing re-execution of the _defining_ word - example:
;-		: (lit>!) 1- dup 1 < if drop 1 1 else dup recurse endif dup * ;
;-		: lit>! create , does> @ (lit>!) ;
;-		10 lit>! 10!	10!	.
;-	tail recursion:
;-		: tt dup 0> if dup . endif 1- recurse ;
;-		see tt \ ">;" notifies the JMP instr. to latest word, "tt"
recurse:GA;:ok:
;test: 1 depth? 0= abort" -stack- " dup 2 < if 1 else dup 1- recurse endif dup . * ;
	call qcomp
	mov dword[uref uv.ucmc],0
	call latest
	call lfa2idx
	mov ecx,codeptr
	mov [uref uv.uprevix],eax; for end-recursion resolution
	mov [uref uv.uprevcp],ecx
	jmp compilec
;-
;- begin  	( -- )(C: -- p f )		(i)				forth
;-	enter a loop structure, terminated by {again}, {repeat} or {until}.
begin:GA;:ok:
	call mark
	spush
	mov eax,'bgin'
	jmp rrbrac
;-
;- while  	- re {if} -			(i)				forth
;-	Forth (lib4th, etc) mode of operation:
;-	conditionally, if tf, continue execution until next {while},
;-	{repeat} or {until}, else terminate loop and continue after.
;-	multiple {while}s in a begin/until or begin/repeat structure
;-	resolved at end of loop, w/o additional words and, terminate 
;-	the loop execution as soon as one {while} part not entered.
;-	{while} cannot point to beyond the end of a {begin}...-loop!
;-	-re- {WHILE} in ans voc.
while:GA;:ok:
	call if
	mov eax,'whil'
	ret
;-
;- WHILE   	- re {if} -			(i)		'whilea'	ans
;-	ANS-4th mode of operation:
;-	conditionally, if tf, continue execution until next {while},
;-	{repeat} or {until}, else terminate loop and continue after.
;-	resolve excess {while}s in a begin/until or begin/repeat structure
;-	by an equal number of additional {else}/{then}, after loop end.
whilea:GA
	call if
	mov eax,'wila'
	ret
rept.i:;:ok:
	spop
	neg ecx
	push eax
	push ecx
	spop
;-
;- repeat  	(C: pn fn .. p f -- pm fm )( -- )(i)				forth
;-	unconditionally branch back to after {begin}.
;-	{repat} resolves {begin} and as many {while}s as used after the 
;-	corresponding {begin}. -re- ans {WHILE} for ANS-style resolution.
repeat:GA;:ok:
	cmp eax,'whil'
	mov ecx,eax
	jz rept.i	;?; multiple {while}
	cmp eax,'wila'
	jz rept.i	;?; dto, ans mode
	mov eax,'bgin'
	call again
	jr until.r
;-
;- again  	(C: p f -- )( -- )		(i)				forth
;-	unconditionally branch back to after {begin}.
again:GA;:ok:
	movzx edx,word[kref Lbranch-Y+whead.i-whead.l]
	jr until.n
until.i:;:ok:
	spop
	neg ecx
	push eax
	push ecx
	spop
;-
;- until  	(C: .. p1 f1 -- .. pm fm )(X: flg -- )(i)			forth
;-	{until} resolves {begin} and as many {while}s as used after the 
;-	corresponding {begin}. -re- ans {WHILE} for ANS-style resolution.
until:GA;:ok:
	cmp eax,'whil'
	mov ecx,eax
	jz until.i
	cmp eax,'wila'
	jz until.i
	movzx edx,word[kref Lzbra-Y+whead.i-whead.l]
.n:;		;<again)
	spush
	mov eax,'bgin'; put pairs value to tos
	call xloop.p; check & compile
.r:;		;<repeat)
	xor edx,edx
.w:
	pop ecx
	cmp ecx,-('whil')
	jz .f
	cmp ecx,-('wila')
	jz .u	;?; ans-mode multiple {while}
	jmp ecx
.f:A
	spush 2
	pop dword TOS(1)
	mov eax,'ifif'
	call endif
	jr .r
.u:
	spush
	pop eax
	spush
	mov eax,'ifif'
	cmp edx,ecx
	jz .w
	push ecx
	call endif
	pop edx
	jr .w
;-
;-
;- (unloop)					(ac)		'rdrop2'	hidden
;-	immediately discard loop parameters
;-	{unloop} runtime, executes {2rdrop}.

;-
;- UNLOOP  	( -- )				(i)(a:c)(wc)			ans
;-	discard loop parameters, i.e. limit and index on RS. requires 
;-	{exit}ting before next LOOP or explicit branch to after the LOOP disp.
unloop:GA;:ok:(?)
	spush
	mov dword[uref uv.ucmc],0
	movzx eax,word[kref Lpunloopp-Y+whead.i-whead.l]
	jmp compilec
;-
;- ?do		( n1 n2 -- )(R: -- n1' n2' )	(i)(83)		'qdo'		forth
;-	enter a LOOP structure if n1 =/= n2, else continue after it's end
;-	compiles{ (?do) ... (loop?) }.
qdo:GA;:ok:
	push dword'kudu'
	movzx ecx,word[kref Lpqdop-Y+whead.i-whead.l]
	jr do.q
;-
;- +loop  	( n -- )			(i)		'ploop'		forth
;-	increment loop counter by n, branch back to after {do}/{?do} if
;-		n > 0: counter unsigned less than limit,
;-		n < 0: counter unsigned greater than limit.
ploop:GA;:ok:
	cmp eax,'kudu'
	jnz .l
	call dup2
	mov eax,'dudu'
.l:
	movzx edx,word[kref Lpploopp-Y+whead.i-whead.l];
	jmp xloop.n
;-
;- do		(C: -- p f )(X: n1 n2 -- )(R: -- n1' n2' )	(i)		forth
;- 	C: f:"dudu" p:c-here
;-	X: push loop limit and start to tor and enter LOOP structure,
;-	   if n1 = n2 adjust n1 := n1+1 to loooping once, only
do:GA;:ok:
	movzx ecx,word[kref Lpdop-Y+whead.i-whead.l]
.d:
	push dword 'dudu'
.q:
	push ecx
.n:;		;<if) entry w. cpl routine & control word on tor
	call rrbrac	; {cmc} set 'off'
	spush
	pop eax		; cpl
	call compilec
	spush 2
	mov ecx,codeptr	; entrypoint
	mov TOS(1),ecx
	pop eax		; control word
	cmp eax,'kudu'
	jnz .r
	push eax
	spush
	xor eax,eax
	call compilec
	mov eax,'ifif'
	call dup2
	mov ecx,codeptr
	dec ecx
	mov codeptr,ecx
	mov TOS(1),ecx
	pop eax
.r:
	ret
;-
;- if		( n -- )			(i)				forth
;-	if n =/= 0 enter code immediately following {if}
;-	else branch forward to next {else} or {endif}, rsp.
if:GA;:ok:
	movzx ecx,word[kref Lzbra-Y+whead.i-whead.l]
.e:;		;<else>,-if,?case)
	push dword if.n-Y
	add [esp],KR
	push dword'ifif'
	push ecx
	jr do.n
.n:A;		;<if)
	spush
	jmp back.c
;-
;- 0if						(i)		'nif'		forth
;-	substitutes { 0= if }, used in {-enter}
nif:GA;:ok: in {-entry}
	movzx ecx,word[kref Lnzbra-Y+whead.i-whead.l]
	jr if.e
;-
;- ?case  	( n1 n2 -- n1 | )		(i)(k6)		'qcase'		forth
;-	does{ over = if drop }, resolved with {else}/{endif}.
;-	use{ nn ..words.. n1 ?case ..words.. else n2 ?case ... endif endif }
qcase:GA;:ok:
	movzx ecx,word[kref Lneqbra-Y+whead.i-whead.l]
	jr if.e
;
ebra:A
	movzx ecx,word[kref Lbranch-Y+whead.i-whead.l]
	jr if.e
;-
;- branch  	( n -- )			(i) n.i.	'ibranch'	forth
;-	branch by disp = n wrt PC after {branch} code.
ibranch:GA;:ok:
	movzx ecx,word[kref Lbranch-Y+whead.i-whead.l]
	jr if.e
;-
;- loop		(C: p f -- )(X: -- )		(i)		'xloop'		forth
;-	increment loop counter by one, terminate { do .. loop } if counter = limit.
xloop:GA;:ok:
	cmp eax,'kudu'
	jnz .l
	call dup2
	mov eax,'dudu'
.l:
	movzx edx,word[kref Lploopp-Y+whead.i-whead.l]
.n:
	spush
	mov eax,"dudu"	;	{ f1 f2 p
.p:;		;<again,repeat,until)
	call qpairs	; ?pairs
	spush		; 	{ p
	mov eax,edx	; (loop)
	mov dword[uref uv.ucmc],0
	call compilec
	call back	; cpl back-ref
	cmp eax,'kudu'
	jnz near llbrac	; if state -> 0 execute from @bcp
	spop 2
	jr endif	; resolve {?do}
;-
;- else		(C: p1 f1 -- p2 f2 )(X: -- )	(i)				forth
;-	begins alternate part of {if .. else .. endif } decision
else:GA;:ok:
	call ebra
	not eax
	call swap2
	test eax,eax
	jns endif
	shl eax,1	; force error + message
;-
;- endif  	(C: p f -- )(X: -- )		(i)				forth
;- THEN		synonym to {endif}				'endif'		ans
;-	terminates { if .. else .. endif } decision
endif:GA;:ok:
	cmp eax,'noop'
	jz .s
.e:
	spush
.i:
	mov eax,'ifif'
	spush
	mov dword[uref uv.ucmc],0
	not eax
	call qpairs2
.r:
	call toresolve
.l:
	jmp llbrac
.s:
	spop 2
	jr .l
;-
;- endif;  	(C: p f -- )(X: -- )		(i)		'endifs'	forth
;-	substitutes { ... exit endif }. saves the {exit} code.
endifs:GA;:ok:
	call cplxit
	jr endif
;-
;- else;  	(C: p f -- )(X: -- )		(i)		'elses'		forth
;-	substitutes { ... exit else ... },
;-	saves {exit} and eliminates the unconditional branch to after {endif}.
elses:GA;:ok:
	call cplxit
	call dup2
	mov dword TOS(2),'noop'
	call rrbrac
	jr endif.e
;-
;- enter  	(C: -- p1 f1 )(X: -- )		(i)		'nter'		forth
;-	unconditionally branch forward into a structure to
;-	resolving multiple inner unconditional branches of
;-		{ .. begin ... while ... repeat }
;-	to a single one on entry, e.g.
;-		{ .. enter begin ... entry ... until }
nter:GA;:ok:
	movzx ecx,word[kref Lbranch-Y+whead.i-whead.l]
	call if.e
	mov eax,'nter'
	ret
;-
;- -enter    	(C: -- p f )(X: flg -- )	(i)		'mnter'		forth
;-	conditionally, if true flg, branch forward to after {entry}
;-	else continue execution immediately after {-enter}. for instance:
;-	{ .. .nn. -dup -enter .. enter begin .. entry .. entry ... until }
;-	                  | false  `----->--------'        |
;-	                  `->--true------>-----------------'
;-	{enter} and {-entry} may multiply be nested around a single {begin}.
mnter:GA;:ok:
	call nif
	mov eax,'nter'
	ret
;-
;- entry  	(C: p1 f1 p2 f2 -- p2 f2 )(X: -- )(i)		'ntry'		forth
;-	entrypoint from branch into a BEGIN.. or IF.. construct via {enter}.
ntry:GA;:ok:
	call swap2
	spush
	mov eax,'nter'
	call qpairs
	call toresolve
	jmp llbrac
;-
;- -leave  	(X: flg -- )			(i)		'dleave'	fig
;-	leave if non-zero flag - fig-4th mode!
;-	use{ DO ..words1.. condition -LEAVE ..words2.. LOOP }
;-	wherein {words2} will be executed, whether leaving or not.
dleave:GA;:ok:
	spush
	movzx eax,word[kref Lpmleavep-Y+whead.i-whead.l];
	jmp compilec
;-
;- leave  					(i)		'xleave'	forth
;-	NOTE: fig-4th mode of operation, i.e. actually leaving a
;-	DO..LOOP structure is delayed till next {loop} execution.
;-	use{	DO ..words1.. condition 
;-		 IF ..words2.. LEAVE ..words3.. \ executes words2 and(!) words3
;-		 ELSE ..words4.. ENDIF
;-		..words5.. LOOP }
;-	wherein {words5} will be executed, whether leaving or not.
xleave:GA;:ok:
	spush
	movzx eax,word[kref Lpleavep-Y+whead.i-whead.l]
	jmp compilec
;-
;--- ANS-4th case construct

;-
;- CASE		( n -- n )(C: -- f f )		(i)				ans
;-	CASE-group is nothing really useful, just a simplification to writing
;-	a sequence of cascaded IF/ELSE decisions, required by DPANS 6.2.0873:
;-	  { n1 CASE n2 OF ..true.. ENDOF  n3 OF ..n3-true.. ENDOF .. ENDCASE }
;-	-re- {case:} in forth-voc. for a much more efficient 'substitute'...
case:GA;:ok:
	spush 2
	mov eax,1
	call toloc
	mov eax,'case'
	spush
	ret
;-
;- OF		( n1 n2 -- | n1 )(C: -- p f )	(i)				ans
;-	if n1 = n2 drop both and enter code after {of},
;-	continue until {endof} and branch to after {endcase}
;-	if n1 =/= n2 drop n2 and continue after next {endof}
;-	substitutes { over = if drop .... }
of:GA;:ok:
	mov ecx,[byte uref uv.s0]
	inc dword[kref ecx-4]
	jmp qcase
;-
;- ENDOF  	( -- )(C: p f -- p' f )		(i)				ans
;-	terminate code sequence after {of}, synonym to {else}.
	global default:function,endof:function
endof:GA {equ else}

;-
;- ENDCASE	( n1 -- )(C: p'1 f1 ... p'n fn -- )(i)				ans
;-	terminate and resolve a {case} structure.
;-	compiles a {drop} of the comarison item and, as many {endif}s as 
;-	{OF}s were encountered, after the latest initiating {CASE}.
endcase:GA;:ok:
	TICK drop
	call compilec
.l:
	mov ecx,[byte uref uv.s0]
	dec dword[kref ecx-4]
	js .r
	call endif
	jr .l
.r:
	spush
	mov eax,1
	call toloc.d
	jmp qpairs
;-
;- DEFAULT	( n1 -- n1 )			(i)				ans
;-	a {noop} word, may be used to point to last statement before {endcase}, 
;-	section after last {endof} receives the {case} comparison value which 
;-	will be dropped from tos by {endcase}.
default:equ noop

;-
;-	
;-
%endif;__CSTRUCT_I
;-
;-=====================================================================;
;- l4cstruct.inc <eof>
