;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4defining.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __DEFINING_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __DEFINING_I
%define __DEFINING_I
%define defininginc
;-
;--- word header ---
;-

;-
;- build  	( p -- )							hidden
;-	build a word header at {here}, linked to {current} wordlist.
;-	pass ptr p to subroutine for fetching a valid name which should move 
;-	the name stg to build-time {here} and leave that ptr and true flag 
;-	on tos - entry to {head} w/o error handling.
;-	{build} in contrary to {head} does neither, {align} nor {!csp}.
build:GA;:ok:
	lea ecx,[kref eax]	; pass name acquisition address
	spop
	jr head.b
;-
;- head		ccc( -- )							hidden
;-	{align} data-space to multiple of cellsize,
;-	store data-stack ptr to {csp} - re {ds-pick},
;-	build a word header at {here}, linked to {current} wordlist,
;-	leave header 'smudged', not ready to {find}.
;-	NOTE:
;-	{head} implements semantics like fig-forth "create".
;-	  [idx|flg][macflg][link][cnt+name][data...] - [items] 4-aligned.
;-	  ^ifa^ffa ^-mfa   ^-lfa ^-nfa     ^-pfa
;-	initial flg=smud, macflg=0, pfa={ here cell - }
;-	>body returns pfa+4, pfa is storage of executeable words' xec-ptr.
head:GA;:ok:
; c: ecx,edx	ret ecx:=lfa
	call xalign		; align dataspace
	call scsp		; !csp
	lea ecx,[kref dfind-Y]	; pass address of -find
.b:
	mov edx,[byte uref uv.dp]; reserve space for header flags
	lea edx,[kref edx+whead.n-whead.i+3]
	and edx,-4
	sub edx,KR
	mov [byte uref uv.dp],edx
	call ecx		; header-name acquisition vector
	test eax,eax
	spop
	jz .n			;		{ -- here | lfa }
	call space
	call iddot
	call space
	spush
	mov eax,4
	call message
	call here
.n:
	mov dword[uref uv.ucmc],0
	mov edx,eax		; here
	call latest		; fetch latest lfa		{ latest h
	lea ecx,[edx-whead.n+whead.l]; this linkfield
	push ecx		; save for ecx return value
	neg eax			; lfa latest -
	add eax,ecx		; disp to prevlink		{ prevlink h
	mov [kref ecx],eax	; link this header, previous is { lfa dup @ - }
	mov [byte uref uv.last],ecx; @{last}
	mov eax,[byte uref uv.current]; wid
	lea eax,[eax-vhead.m+vhead.p]; top lfa			{ toplfa h
	sub ecx,eax
	neg ecx			; new @{latest}
	mov [kref eax],ecx	; link this to wordlist
	add edx,KR		; 'here' is nfa
	movzx ecx,byte[edx]	; name cnt
	inc ecx
	add dataptr,ecx		; allot & advance over lfa
	lea eax,[edx+ecx]
	mov dword[eax],0
	lea ecx,[eax+3]
	and ecx,byte -4
	sub ecx,eax
	jz .a			;?; already aligned
	add dataptr,ecx		; align
	add eax,ecx
.a:
	add dword dataptr,byte 4; allot till after pfa
	mov ecx,codeptr		; kref'd c-here
	push ecx		; save xec address for uot
	mov [eax],ecx		; kref(xec+@c0)={c-here} to pfa	{ pfa h
	movzx ecx,word[byte uref uv.lkindex]
	inc dword[byte uref uv.lkindex]; prepare next index
	mov [edx-whead.n+whead.i],cx; store this words index
	mov word[edx-whead.n+whead.f],smud; mark 'not ready'
	pop edx			; xec ptr
	add edx,KR		; true address
	mov [ecx*4+uref uv.ultab],edx; to uot
	pop ecx			; pass lfa to lo-level caller
	spop 2
	ret
;-
;- alias  	ccc( ix -- )							tools
;-	defines a synonymous word, ccc, executing xec token ix.
;-	no code, header only: 28 + 4-aligned (namelength+1) bytes in data-space.
;-	NOTE: {alias}ed words conflict w/ {forget} and {:forget}.
alias:GA;:ok:
	call head		; ecx := lfa
	mov [kref ecx-whead.l+whead.i],eax; index field
	dec dword[byte uref uv.lkindex]	; back to previous index
	mov eax,[byte uref uv.lkindex]
	dec eax				; this word's ix
	TICK fga			; forget-handler
	call fgcomma
	jmp unsmudge
fga:A
	inc dword TOS(1)
	xor eax,eax
	ret
;-
;--- defining words ---
;-

;
; check cpl state and equal stack balance, remove {:} control structure
qcolon:A;<semi, doesgt)
	call qcomp		; abort if not compiling
	call smudge		; make word findable
	call locres		; resolve local mem & control structure
	TOSPULL(1)		; swap drop
	call spfetch		; sp@
	sub TOS(1),eax		; =/=
	jnz near rdn.n		;?; abort" definition not finished"
	spop 2
	ret
;-
;- end-code							'endcode'	assembler
;-	terminates assembly sequence, after ;code, in interpreting state
endcode:GA;:ok:(-?-)
	call c_ret
	call csemis
	jmp hpts
;-
;- ;		( -- )(S: s1 [ s2 ] -- |abort )	(i)		'semi'		forth
;-	terminates a colon definition.
;-	modifies last <call> to <jmp> to preventing call+ret sequence,
;-	which is disabled after a structure's end-ptr or, {cmc-off}.
;-	leaves final <ret> in place, for safe return from compiler.
semi:GA;:ok:
	call qcolon
	call cplxit		; cpl <ret> or <jmp last_word>
.r:;	;<;code)
	call csemis
	and dword[byte uref uv.state],~(mnnam)
	jmp llbrac
;-
;- :		ccc( -- s1 )					'colon'		forth
;-	begin a named hi-level definition, linked to {current} wordlist.
;-	ANS mode: leaves s1 = ptr-s to codespace and datastack on datastack!
;-	for fig-compatible handling and {;code}, stores initial SP to {csp}.
colon:GA;:ok:
	or dword[byte uref uv.state],mword; cpl persistent code
	mov dword[uref uv.ulocvc],0; reset local locals counter
	call rrbrac		; ]]
	call head
	spush
	mov eax,loclptr	; locals initial ref
	jmp spfetch
;-
;- code		ccc( -- )(C: -- )		(a:tx)				forth
;-	cpl:
;-	begin named lo-level definition which, initially is not <ret> terminated 
;-	and does not set-up nor leave any control data in the compile-stack. 
;-	word ccc can be found immediately after creation by {code} and, should 
;-	be orderly continued and terminated before return to the interpreter.
;-	xec:
;-	W-register <ecx>, [kref ecx] = pfa+cell = >body of word ccc
;-	T-register <edx>, [kref edx] = code entry
;-	(an entirely useless word, just to complying ANS-4th setup and, not tested!)
code:GA
	call qexec
	or dword[byte uref uv.state],mword
	call rrbrac
	call head		;(ecx:=lfa)
	spush
	lea esp,[esp+4]		; immediate return, <ret> in codespace overwritten :HIER: -?-
	mov eax,ecx		; latest
	call lfa2pfa
	lea eax,[eax+4]		; body
	call c_iwld		; cpl W = ecx := body
	spush
	mov eax,codeptr
	lea eax,[eax+5]
	call c_itld		; cpl T = edx := code entry
	jr ecode
;-
;- create  	ccc( -- p )(C: -- )		(w)				forth
;-	xec:
;-	word ccc leaves ptr to {here} in dataspace at definition time which is 
;-	the same ptr which {>body} would return, one cell after the word's pfa.
;-	cpl:
;-	build by name executeable word header, linked to @{current} wordlist.
;-	compiles literal ptr to after <pfa> & <ret> to {c-here}, 7 bytes in codespace.
;-	data arranged such that
;-		{ ' ccc idx>pfa @ } equals { ' ccc idx>xec }
create:GA;:ok:
	call qexec
	or dword[byte uref uv.state],mword
	call rrbrac
	call head		;(returns ecx:=lfa)
	spush
	mov eax,ecx		; latest
	call lfa2pfa
	lea eax,[eax+4]	; body
	call c_lit		; cpl: spush\ mov eax,imed
	inc dword codeptr	; include terminating <ret> code
ecode:A;<ibbuf,twocons/rcode,code)
	call csemis
	call hpts		; make the code permanent
	call llbrac
	jmp unsmudge
;-
;- 2constant	( -- dn )(C: dn -- )		(a:d)		'twocons'	forth
;-	non-alterable const double value, compiled to codespace.
twocons:GA;:ok:
;asm->:: does not return to caller! ::
	call swap
	call qconstant.n
	call rrbrac
.l:; <4constant)
	call c_lit		; cpl: spush\ mov eax,imed
rcode:; ;<value,vocabulary)
	mov ecx,codeptr
	inc ecx
	mov codeptr,ecx		; include terminating <ret> code
.c:
	lea ecx,[kref ecx-1]
	mov [esp],ecx		; adjust to safely return
	mov byte[ecx],x_ret	; terminate w. un-counted ret code
	jr ecode
;-
;- 4constant	( -- qn )(C: qn -- )		(a:d)		'qconstant'	forth
;-	non-alterable cons quad value, compiled to codespace.
qconstant:GA;:ok:
;asm->:: does not return to caller! :: re above, 2constant
	call swap2s
	call .n
	call rrbrac
	call c_lit
	call c_lit		; cpl: spush\ mov eax,imed
	jr twocons.l
;
.n:A;<2constant, 4constant)
	call create
	mov ecx,codeptr		; c-here
	mov [kref ecx-5],eax	; patch pfa-ptr to cons
	dec ecx
	mov codeptr,ecx
	spop 			; drop cons parm
	ret
;-
;- :noname	(C: -- ix )					'noname'	forth
;-	begin headerless hi-level definition, leave uot index after terminating {;}.
;-	{:noname} definitions occupy an uot item, ea., but no header-space.
;-	for fig-compatible handling and {;code}, stores initial SP to {csp}.
;-	use{ :noname ... ;    : word .. [ 0 cs-pick ] [,] ... ;    drop }
noname:GA;:ok:
	call qexec
	or dword[byte uref uv.state],mword|mnnam
	call rrbrac
	call scsp		; !csp
	call chere
	call fth2abs
	mov ecx,eax
	mov eax,[byte uref uv.lkindex]
	inc dword[byte uref uv.lkindex]
	mov [uref eax*4+(uv.ultab)],ecx
	spush
	jmp spfetch
;-
;- variable	ccc( -- p )(C: -- )		(a:c)				forth
;-	create named ptr to a cell of reserved dataspace (ANS: initiated to zero)
;-	execution of name ccc leaves same ptr as { ' ccc >body }.
variable:GA;:ok:
	call create
zcomma:A
	spush
	xor eax,eax
	jmp comma
;-
;- 2variable	( -- p )(C: -- )		(a:c)		'twovari'	forth
;-	create named ptr to double cell of reserved dataspace (ANS: initiated to zero)
twovari:GA;:ok:
	call variable
	jr zcomma
;-
;- 4variable	( -- p )(C: -- )		(a:c)		'qvari'		forth
;-	create named ptr to quad cell of reserved dataspace (ANS: initiated to zero)
qvari:GA;:ok:
	call twovari
	call zcomma
	jr zcomma
; -
;- <with  	ccc( -- )			(f6)(n.i.)	'ltwith'	forth
ltwith:GA
; -	extend currently defined words semantics to using defining part 
; -	of word <ccc> prior to those newly given. inheritance of defining method
	jmp nimp
;-
;- wordlist	ccc(C: -- )(X: -- wid )		(a:s)				forth
;-	create an empty, named wordlist, linked into {voc-link} chain
;-	re DPANS 16.6.1.1595: number of wordlists is not limited.
wordlist:GA;:ok:
	call qexec
	call xalign
	spush
	mov eax,dataptr
	spush
	mov eax,whead_size+vhead_size
	call erase
	spush				; dummy value
	call create
	mov ecx,codeptr			; c-here
	mov [kref ecx-5],eax		; patch pfa-ptr to cons
	push dword[byte uref uv.dp]	; vifa
	mov eax,vhead_size
	call allot
	call latest
	mov word[kref eax-whead.l+whead.f],vocf
	pop edx				; vifa
	mov ecx,edx
	sub ecx,eax			; disp to lfa
	mov dword[kref edx],ecx		; vifa : disp to word-lfa
	lea ecx,[edx-vhead.i+vhead.m]	; wid
	mov eax,codeptr
	mov [kref eax-5],ecx		; patch cons value to wid
	lea ecx,[byte uref uv.voclink]
	mov eax,ecx
	sub eax,KR
	sub eax,[ecx]			; previous voclink
	push eax			; save old
	lea eax,[kref edx-vhead.i+vhead.m]; this voclink posn
	sub eax,ecx
	neg eax
	mov [ecx],eax			; new {voc-link}
	pop ecx
	lea eax,[edx-vhead.i+vhead.m]; this voclink posn = wid
	sub eax,ecx
	mov [kref edx-vhead.i+vhead.m],eax	; vmfa : voclink <- wid
	mov dword[kref edx-vhead.i+vhead.l],0	; vlfa : wordlist base
	mov dword[kref edx-vhead.i+vhead.n],rgvnfa; vnfa : dummy name
	mov dword[kref edx-vhead.i+vhead.p],vhead.p-vhead.l; vpfa : top lfa
	spop
	ret
;-
;- vocabulary	ccc( -- )(X: -- )						forth
;-	re DPANS 16.6.1.1595: the number of vocabularies is not limited!
vocabulary:GA;:ok:
	push dword[byte uref uv.lkindex]
	push dword[byte uref uv.cp]
	call wordlist
	call rrbrac
	spush
	pop ecx			; initial cp
	mov [byte uref uv.cp],ecx
	mov eax,dovoc-Y
	call cplicall		; cpl : call <dovoc>
	mov ecx,codeptr
	pop dword[kref ecx]	; store uot index
	lea ecx,[ecx+4]
	mov [byte uref uv.cp],ecx
	call c_ret
	dec dword[byte uref uv.cp]
	jmp rcode
;-
;- constant	( -- n )(C: n -- )		(a:c)				forth
;-	defines a named, non-alterable sing numeric literal.
constant:GA;:ok:
	call create
	mov ecx,codeptr
	mov [kref ecx-5],eax	; patch data-ptr to literal
	dec ecx
	mov codeptr,ecx
	spop
	ret
;-
;- 4integer	ccc( -- dn )(C: dn -- )		(h4)((i))	'integer4'	forth
;-	immediate, non-alterable quad cons, which executes a 4literal.
;-	value can be altered e.g. w/ { ' name >body 4! }, before compiling.
integer4:GA A 3;:ok:
	call qexec
	call create
	call immediate
	spush 2
	push dword dataptr
	mov eax,16
	call allot
	pop eax
	call store4
	call pscodep
;-
;- (i4		( p -- qn )			(i)		'dointeger4'	novoc
;-	immediately fetch {4integer} as a {4literal} from data-space.
dointeger4:
	mov ecx,esi
	lea esi,[kref eax]
	movsd
	movsd
	movsd
	mov eax,[esi]
	mov esi,ecx
	jmp qliteral
;-
;- 2integer	ccc( -- dn )(C: dn -- )		(h4)((i))	'integer2'	forth
;-	immediate, non-alterable double cons, which executes a 2literal.
;-	value can be altered e.g. w/ { ' name >body 2! }, before compiling.
integer2:GA A 1;:ok:
	call qexec
	call create
	call immediate
	call swap
	call comma
	call comma
	call pscodep
;-
;- (i2		( p -- dn )			(i)		'dointeger2'	novoc
;-	run-time code: immediately fetch {2integer} as a {2literal} from data-space.
dointeger2:
	lea ecx,[kref eax]
	mov eax,[ecx]
	spush
	mov eax,[ecx+4]
	jmp dliteral
;-
;- integer	ccc( -- n )(C: n -- )		(h4)((i))			forth
;-	non-alterable sing cons, IMMEDIATEly compiling a literal.
;-	literal value can be altered with {to}, {+to}, before compiling, 
;-	e.g. to determining a constant at compiletime from program data. 
;-	modification does NOT affect already compiled {integer}s.
;-	can substitute { create , } but, would also declare ccc 'immediate'!
integer:GA A 3;:ok: (align <dointeger>)
	call qexec
	call create
	call immediate
	call comma
	call pscodep
;-
;- (i1		( -- n )			(i)		'dointeger'	novoc
dointeger:
	mov eax,[kref eax]	; fetch literal from data-space
	jmp literal
;-
;- val+  	ccc( n -- )(X: n1 -- n2 )			'valp'		forth
;-	increments @tos n1 by "value" n
valp:GA;:ok:
	call qexec
	call create
	call comma
	call pscodep
dovalp:; 4-aligned by <valp> alignment
	mov eax,[kref eax]
	add eax,TOS(1)
	TOSPULL(1)
	ret
;-
;- cons+  	ccc( n -- )(X: n1 -- n2 )	((i))		'consp'		forth
;-	increments @tos by non-alterable literal value n
;-	{cons+} compiles an IMMEDIATE word which compiles a literal increment.
consp:GA A 3;:ok: (align <doconsp>)
	call qexec
	call create
	call immediate
	call comma
	call pscodep
doconsp:; 4-aligned
	mov eax,[kref eax]
	call ccplq
	jnz near c_litp		;?; cpl add @tos,literal
	add eax,TOS(1)
	TOSPULL(1)
	ret
; <integer=value)
; ret NZ if {integer} already exists in {current} voc.
; i:	-/-
; o:	ecx	-?- | ix
;	eflags	Z   | NZ
tiname:A
	call ccplq
	jnz .q			;?; compiling
	call saveinput
	call blwf
	mov TOS(1),ecx		; rst lfa
	call tor2
	call restoreinput
	spop
	call rfrom2
	test eax,eax
	spop
	jz .r			; not found
	call idx2xec
	mov ecx,eax
.r:
	spop
.q:
	ret
tito:A
	call tiname
	jz .r
	mov ecx,[kref ecx]
	sub ecx,[kref d_valu-Y]
	shl ecx,8
	jz .x
	inc ecx			; NZ
.x:
	dec ecx
.r:
	ret
; -
intc:A
	call qexec
	or dword[byte uref uv.state],mword
	call rrbrac
	call head		;(returns ecx:=lfa)
	spush
	mov eax,ecx		; latest
	call lfa2body
	mov ecx,TOS(1)
	mov [kref eax],ecx	; store value
	ret
;-
;- =:		ccc( -- n )(C: n -- )		(h4)(p4)	'valdef'	forth
;-	store n to previously defined {value} or {integer} if found in 
;-	current search order. otherwise, in interpreting state, define 
;-	a new, alterable -re- {value}.
valdef:GA;:ok:
	call ccplq
	jnz g32 tov
	call tito
	jnz g32 tov		;?; name known or compiling state
;-
;- value  	ccc( -- n )(C: n -- )		(a:cx)				forth
;-	self fetching variable, w/ -re- {to}, {+to}, {=:} alterable cons.
;-	value of ccc stored to dataspace and alterable at { ' ccc >body }.
;-	{value}s can be altered with {to} and {+to}, which also affects 
;-	already compiled words. fetching and storeing a {value} is usually 
;-	faster than the rsp opr on a {variable}.
value:GA;:ok:
	call intc		; build header, sto value to body
	call c_valu		; cpl: spush\ mov eax,[kref disp]
	inc dword[byte uref uv.cp]; include terminating <ret> code
	mov eax,4
	call allot
	jmp rcode
;-
;- +>		ccc( n -- )			(i)(p4)		'pltov'		ans
;- +to		ccc( n -- )			(i)(a:cx)	'pltov'		forth
;-	add n to a {value}, equivalent to { ' name >body +! }.
;-	can substitute {+!} to a {variable}, by about 10% faster.
;-	defines new {value} if in interpreting state and ccc unknown.
pltov:GA;:ok:
	call ccplq
	jnz .c
	call tito
	jz value
.c:
	lea ecx,[kref c_plsto-Y]
	jr tov.n
;-
;- ->		ccc( n -- )			(i)(p4)		'tov'		ans
;- to		ccc( n -- )			(i)(a:cx)(a:l)	'tov'		forth
;-	store new number to a {value} - equivalent: { ' name >body ! }.
;-	also substitutes {!} to {variable}s or {defer}red words.
;-	values storeing with { to name } is a (tiny) bit faster than with 
;-	{ [ ... ] literal ! } and by about 10% faster than { vari-name ! }.
;-	(this would have been much simpler with just a compile-flag if ANSI didn't 
;-	(decide differently, due to whatever reason -if any- re DPANS, A.6.2.2295.
tov:GA;:ok:
;:HIER: auf cpl per index fuer uot/disp umbauen :HIER: -?-
	lea ecx,[kref c_sto-Y]
.n:
	push ecx		; save compiler, to or +to
	mov dword[uref uv.ucmc],0
	call blwf		; fetch name
	spop
	jz near tick.r		;?; not found
	call tobody		; ptr to data-field (c: ecx)
.c:
	call ccplq
	lea ecx,[kref c_plsto-Y]
	jz .i			;?; interpreting, eax = body
	cmp edx,kvalf
	jz .w
.t:
	cmp edx,locvf		; header-flags for local value
	mov edx,[esp]		; compiler
	jnz .v			;?; std. value
	mov eax,[kref eax]	; @body
	cmp edx,ecx
	lea edx,[kref c_locto-Y]; to
	jnz .l			;?; cpl {to}
	lea edx,[kref c_plocto-Y]; +to
.l:
	lea esp,[esp+4]		; drop initial compiler
	neg eax			; -ve disp
	jmp edx			; compile xec routine
.v:A
	mov ecx,codeptr
	mov [esp],ecx
	call edx		; cpl: to/+to
	pop ecx
	mov [kref ecx+2],eax	; patch disp into code
	spop
	ret
.w:A;<cpl to kernel value)
	push edx
	lea edx,[eax-4]		; pfa
	add edx,[kref edx]	; ca
	cmp edx,pdoval-Y
	pop edx
	jnz .t			;?; not a kernel value
	mov eax,[kref eax]
	add eax,databot		; data-body
	mov edx,[esp]
	jr .v			; store/add
.i:A;<interpreting), eax = data-body
	cmp edx,kvalf
	jz .k			;?; kernel {value}
	cmp edx,locvf
	jnz .g			;?; not a local value
	mov eax,[kref eax]
	neg eax			; -ve disp
	add eax,[byte uref uv.s0]; ptr to local data
.g:
	pop edx
	cmp edx,ecx
	mov ecx,TOS(1)
	jz .p			;?; +to
	mov [kref eax],ecx
.x:
	spop 2
	ret
.k:A;<kernel value)
	lea edx,[eax-4]
	add edx,[kref edx]
	cmp edx,pdoval-Y
	jnz .x			;?; not a value
	mov eax,[kref eax]
	add eax,databot
	jr .g			; store/add
.p:A
	add [kref eax],ecx
	spop 2
	ret
qloc:A
	cmp dword[kref eax-4],dolocv-Y; compiled
	jz .r
	cmp dword[kref eax-4],dolocc-Y; interpreted
.r:
	ret
;-
;- 2value  	ccc( -- n )(C: n -- )				'value2'	forth
;-	double, alterable cons.
;-	value of ccc stored to dataspace and alterable at { ' ccc >body 2! }.
;-	altered {2value}s also affect already compiled words.
value2:GA
	call swap
	call intc		; build header, sto value to body (=eax)
	push eax
	call c_valu		; cpl: spush\ mov eax,[kref disp]
	mov ecx,TOS(1)
	pop edx
	lea eax,[edx+4]
	mov [kref eax],ecx
	call c_valu
	inc dword[byte uref uv.cp]; include terminating <ret> code
	mov eax,8
	call allot
	jmp rcode
;-
;- 2to		ccc( dn -- )			(i)		'to2'		forth
;-	store double to a {2value}
;-	does{ ' name >body 2! }  w/o error checks!
to2:GA
	call blwf
	spop
	jz near tick.r		;?; not found
	call tobody
	call literal
	call ccplq
	jz near twosto
	TICK twosto
	jmp compilec
;-
;- does>  	( -- )(C: s1 -- s1 )(X: -- p )	(i)		'doesgt'	forth
;-	terminates defining part compilation of defining word, 
;-	compiles (;code) ;S, continues defining hi-level runtime procedure.
;-	defining word will not be found until after terminating {;}.
;-	the {does>} part of a word can be provided by a separate definition.
;-	NOTE: implemented such that it does not permit multiple {does>} in a
;-	single defn - which may be enabled w/ <__mdoes>, re lib4th Makefiles.
doesgt:GA;:ok:
;	...multiple does> not at all useful but, there were some 'experts'
;	who insisted on such a 'feature', in c.l.f. 'discussions'.
	mov dword[uref uv.ucmc],0
%ifdef __mdoes	; multiple DOES> permitted
	call qcolon		; terminate defining (resolve locals), begin runtime
	spush
	mov eax,'does'
	spush
	lea eax,[PS-4]
	sub eax,KR		; sp@
	call unsmudge		; combined w. {smudge} to hiding latest word
%else; abort w. multiple does>
	push dword TOS(1)
	call qcolon		; terminate defining (resolve locals), begin runtime
	spush
	pop eax			; control word
	mov ecx,'does'
	cmp ecx,eax
	mov eax,ecx
	spush
	jz .s			;?; force abort after repeated DOES>
	lea eax,[PS-4]
	sub eax,KR		; sp@
%endif
	call smudge		; hide this word
.s:
	TICK pscodep		; ['] (;code)
	jmp compilec
;-
;- (;code)							'pscodep'	assembler
;-	cpl'd by {does>}; appends runtime semantics to current defining word.
;-	NOTE: definitions using a such created word cannot compile access 
;-	      by uot and thus should reside in the same binary module.
pscodep:GA;:ok:
; :HIER: => uot-indizierung: zuweisung eines index durch {does>}, cpl dann mit {compile,}
;	 => mit vorsicht! war heikel genug, die sache ueberhaupt zum spielen zu bringen...
; cpl < jmp c-here >
	call rbrac
	call latest
	call lfa2idx		; ix
	push eax		; >rr
	call idx2xec
	spush			; save xec ptr for cmp at exit
	mov codeptr,eax		; restart after latest {create}
	pop eax			; ix
	call tobody		; ptr to after pfa
	call c_lit		; cpl literal
	call rfrom		; r>	kref'd ptr to after (;code)
; no jump by uot index:
	mov ecx,codeptr
	lea ecx,[ecx+5]
	sub eax,ecx
	call c_jmpn		; branch to after does>
	call c_ret
; -- ;
	call cpl		; prepare xec via <btail> from <codeptr>
	call llbrac		; resolve cpl level, jmp <btail> if @state=0
	mov ecx,codeptr
	mov [byte uref uv.bcp],ecx; backup codeptr
; check cpl state, ret appropriately
	add eax,KR		; latest xec
	add ecx,KR		; c-here
	cmp eax,[esp]		; test whether ret-adr within just compiled word
	spop
	jnc .r			;?; ret < latest-xec
	cmp ecx,[esp]		; test c-here > ret-adr
	jc .r			;?; ret > c-here (provision for local words)
	lea esp,[esp+4]		; imed. return
.r:
	ret
;-
;- ddo		( -- )				(r)		'dodoes'	novoc
;-	xec from pfa: [ dodoes ][ kref<pfa> ]
;-		 pfa: [ kref<cfa> ][ ..data.. ]
;-	compiled to code-space in RAM at startup time, re <iwlink>
;-	arg passing by W-reg, <ecx>
dodoes:A;:ok:
	pop ecx
	mov edx,[ecx]		; pfa
	lea ecx,[edx+4]		; W-reg, ptr to body
	sub ecx,KR
	add edx,[edx]		; @pfa pfa + -> xec
	jmp edx
;-
;- (code)  					(4)		'pcodep'	assembler
;-	terminates a colon definition, returns to interpreting state. -?-
pcodep:GA
	call rbrac
	call latest
	push eax
	call lfa2xec
	mov codeptr,eax
	pop eax
	call lfa2idx
	call tobody
	call c_lit
	call rfrom		; r> (kref'd ptr!)
;:HIER: auf cpl per index fuer uot/disp umbauen :HIER: -re- {compile} -?-
	mov ecx,[byte uref uv.cp]
	lea ecx,[ecx+5]
	sub eax,ecx
	call c_jmpn
	call csemis
	jmp clbrac
;-
;- ;CODE  					(a:tx)(ix)(4)	'scode'		ans
;-	terminates a colon definition, returns to interpreting state. -?-
scode:GA
	call semi
	dec dword[byte uref uv.cp]
	ret
;-
;- defer  	ccc(C: -- )(X: ..xx.. -- ..x.. )(ix)				forth
;-	execute word the uot index or true xec-address of which is stored in data body.
;-	at runtime reverts to default action if defined word not in valid memory range.
;-	null ptr de-activates the deferred word and does not revert to default action.
;-	compiles an initial {forget>} which removes the word from the {defer-lk} chain,
;-	additional forget-handlers may be assigned, which execute prior to the default.
;-	xec de-activated if ix = 0 or (fastest) { 0 4th>abs } at data body.
;-	{defer}red words linked to {defer-lk}:
;-		defer-lk dup @ +	-> ptr to last link
;-		<lastlink> dup @ +	-> ptr to previous link, <zero> at end of chain
defer:GA;:ok:
%assign dfer_lk_body 8; disp from body to link, for {deferred?}
%assign kdfer_lk_body 12; kernel-deferred
	call create		; word name, push pfa at runtime
	call rrbrac		; ]]
	TICK noop
	call comma		; ['] noop ,
	TICK ndferd
	call comma		; ['] not-deferred ,
	mov ecx,dferlkv-Y
	call pdovar		; defer-lk
	mov edx,[kref eax]	; @
	add edx,eax		; old-link
	mov ecx,dataptr		; here
	sub edx,ecx		;edx := disp old-link to here
	sub ecx,eax		;ecx := disp vari to new-link
	mov [kref eax],ecx	; store disp to new top-link
	mov eax,edx		; cpl disp here to old-link
	call comma
	spush
	xor eax,eax
	mov al,Idfdfg;(align <dodefer>, subst. "mov eax,Idfdfg")
	TICK fgl		; cpl { forget> dfg }
	call compilec
	call llbrac
	call pscodep		; (;code) & ret while defining
;-
;- ddf		(X: p -- xx )			(k)		'dodfer'	novoc
;-	runtime entry from {defer}.
;-	cplmem: ptr to pc(data) -> [ xec(code) | xec(default) | linkage ]
;-	reverts to default action if defined word not in valid memory range.
;-	{noop} if p is null-ptr (=kref+0) or zero.
dodfer:;:ok:;4-aligned, immediately after <pscodep>!
	spush			; dup @
	mov eax,[kref eax]
	mov ecx,eax
	add ecx,KR
	jz .z			;?; inactive, 2drop & ret
	sub ecx,KR
.z:
	jz near dfdfg.r		;?; inactive, 2drop & ret
	call xecq
	inc eax
	mov ecx,TOS(1)
	spop 2
	jz .x
	mov ecx,eax
	mov eax,[kref eax+4]
	mov [kref ecx],eax	; reset to default action
	call xecq
	inc eax
	mov ecx,TOS(1)
	jnz ndferd.r
	spop
.x:
	spop
	jmp ecx
;-
;- ddk		(X: -- p )			(k)		'pdodefer'	novoc
;-	kernel deferred -re- {ddf}
pdodefer:A;:ok:
	spush
	mov eax,[kref ecx]
	add eax,databot
	jmp disp dodfer
;-
;- ndf    	( -- error-26 )			(k)		'ndferd'	novoc
;-	#26 error, "xec vec undefined"
ndferd:A
	spush
.r:
	mov eax,26
	jmp errorf
;-
;- ?deferred	( ix -- | error-26 )				'qdferd'	forth
qdferd:GA;:ok:
	call dferdq
;-
;- df?		( f -- | error-26 )		(k)		'dfder'		novoc
;-	0= #26 ?error
dfder:A
	test eax,eax
	spop
	jz ndferd
	ret
;-
;- dfg		( ix -- ix )			(k)		'dfdfg'		novoc
;-	deferred default forget action
dfdfg:A
	spush
	call dfdbody		; dataptr
	mov ecx,dferlkv-Y
	call pdovar		; defer-lk
	mov ecx,TOS(1)
	add ecx,dfer_lk_body
	add ecx,[kref ecx]	; prevlink
	sub ecx,eax
	mov [kref eax],ecx
.r:
	spop 2
	ret
; defer-lk chain init (forget-handler not applicable)
dfdini:A
	mov ecx,dferlkv-Y
	call pdovar		; defer-lk
	test dword[kref eax],-1
	jnz .r
	sub [kref eax],eax
.r:
	spop
	ret
;-
;- deferred?	( ix -- f )			(4)		'dferdq'	forth
;-	tf if word the xec token of which is ix is linked to the deferred words chain.
dferdq:GA;:ok:
; c: ecx, edx
	call tobody		; >body -dup
	test eax,eax		; data field
	jz .q			;?;IF
	spush			; dup
	call .s			; search linkfield
	test eax,eax
	jnz .q			;?; found
	mov edx,TOS(1)		; body
	lea edx,[edx-4]		; pfa
	add edx,[kref edx]	; wa
	cmp edx,pdodefer-Y
	jnz .r			;?; not a kernel deferred word
	mov eax,TOS(1)
	mov eax,[kref eax]	; disp
	add eax,databot		; RAM data-field
	mov TOS(1),eax		; save df'd data ptr
	call .s
	test eax,eax
	jz .k			;?; init kernel-deferred
.q:
	TOSPULL(1)
	ret
.k:A; initiate a kernel deferrred data-field
	mov ecx,TOS(1)
	add ecx,dfer_lk_body	; [ defer-lk @ ] literal + ; disp to link
	mov eax,[kref dferlkv-Y]
	add eax,databot		;eax := defer-lk vari-ptr		; defer-lk
	mov edx,[kref eax]	; @	disp				; dup @
	add edx,eax		; ptr to old-link			; +
	sub edx,ecx		;edx := disp from old-link to here	; here -
	mov [kref ecx],edx	; disp here to previous lfa, in lfa	; here !
	sub ecx,eax		;					; here defer-lk -
	mov [kref eax],ecx	;mem := disp df-lk to here, in defer-lk	; defer-lk !
	jmp .r
.s:A; scan defer-lk chain
	add eax,dfer_lk_body	; [ defer-lk @ ] literal + ; disp to link
	mov ecx,dferlkv-Y
	call pdovar		; defer-lk
	mov edx,eax
.begin:;			; begin
	mov ecx,[kref eax]	;  dup @
	jz .until		;?; chain ends
;				; while
	add ecx,eax		;  +
	jz .until		;?; chain ends
	mov eax,ecx
	cmp eax,TOS(1)
	jnz .begin		;  2dup = until
.r:
	mov ecx,-1		;  drop true
.until:;			; ENDIF
	mov eax,ecx
	TOSPULL(1)		; sdrop ;s
isf.r:
	ret
;-
;- is?		ccc( -- ix )					'isq'		forth
;-	fetch xec token which the deferred word ccc will execute
isq:GA;:ok:
	call dfdtick
	jr isf.f
;-
;- is@		( ix -- ix1 ix2 | abort-26 )			'isf'		forth
;-	fetch default ix1 and xec token x2 of a deferred word.
isf:GA;:ok:
	call dfdtick.n
	spush
	add eax,byte 4
	call .f
	call swap
.f:
	mov eax,[kref eax]	; ix or addr
	call xecq
	sub eax,byte 1
	spop
	jc .r
	sub eax,KR
	jmp xec2idx
;-
;- is		ccc( -- | abort-26 )		(4)(i)				forth
;-	check and store an execution token to deferred word.
;-	use{ ' name is deferred-word }, for the default opr -re- {is-default}.
;-	use{ ['] name is deferred-word } while ANS-mode compiling.
;-	{>body} and {to} may be used, instead, which do no error checking.
is:GA A 1;:ok:
	call dfdc		; ( p
	jnz dois.c
;-
;- (is		( p -- )					'dois'		novoc
dois:A
	mov ecx,TOS(1)		; >body !
	mov [kref eax],ecx
	spop 2
	ret
.c:
	TICK dois
	jmp compilec
dfdtick:A
	call itick		; '
.n:
	spush			; dup
	call qdferd		; ?deferred
dfdbody:A
	call tobody
	lea ecx,[eax-4]
	add ecx,[kref ecx]
	cmp ecx,pdodefer-Y
	jnz .r
	mov eax,[kref eax]
	add eax,databot
.r:
	ret
;-
;- default-is	ccc( -- )			(4)(i)		'defaultis'	forth
;-	restore a deferrred words default action
;-	use{ default-is deferred-word }
defaultis:GA A 1;:ok:
	call dfdc
	jnz dfi.c
;-
;- (di		( pp -- )					'dfi'		novoc
dfi:
	mov ecx,[kref eax+4]
	mov [kref eax],ecx
	spop
	ret
.c:
	TICK dfi
	jmp compilec
;-
;- is-default	ccc( ix -- )			(4)(i)		'isdefault'	forth
;-	store default action to deferred word
;-	use{ ' action is-default deferred-word }
isdefault:GA A 1;:ok:
	call dfdc
	jnz ddi.c
;-
;- (dd		( pp -- )					'ddi'		novoc
ddi:
	mov ecx,TOS(1)
	mov [kref eax+4],ecx
	spop 2
	ret
.c:
	TICK ddi
	jmp compilec
dfdc:
	call dfdtick
	call literal
	jmp ccplq
;-
;-	
;---- multi-decision lists, case:, range:, associative:
;-	

%xdefine casec_num 0x0ff
;-
;- casehead-size	( -- n )					'caseheadsz'	hidden
;-	size in bytes of a {case:}-list header -re- <inc/l4struc.inc>
caseheadsz:GA;:ok:
	spush
	mov eax,casec_size
	ret
;-
;- is-case	ccc( n1 n2 -- )			(4)(i)			'iscase'	forth
;-	store list item n1 to list w/ name ccc at posn by index n2
;-	 
;-	data type flag at 1st cell of data body,
;-	over-all list bytes size stored at 2nd cell of data body, -re- {cases,}.
;-	
;-	forcing to numeric type:
;-		{ -1 ' listname >body c! }
;-	testing whether an index fits into list:
;-		{ ( caseindex ) 4* casehead-size + ' listname >body 4+ @ 4- > }
;-	  which returns tf if index was out of range.
;-	  list entries follow at {casehead-size} bytes disp to data body.
;-	modifying a list entry:
;-		{ ( new-entry caseindex ) 4* casehead-size + ' listname >body + ! }
iscase:GA;:ok:
	test eax,eax
	js .n			;?; false index
	call csetick
	spop			; ( n1 n2 ix
	jnz .n			;?; unsuitable name
	call ccplq
	jnz .c			;?; compiling
doisc:A; n1 n2 ix -- 
	call tobody		; ( n1 n2 body
	mov ecx,TOS(1)		; index to case cell
	cmp ecx,[kref eax+casec.size]
	lea eax,[eax+ecx*4+casec_size]; disp to data cell
	jg iscase.n
	mov edx,TOS(2)
	mov [kref eax],edx
	spop 3
	ret
iscase.c:
	mov dword[uref uv.ucmc],0; don't 'optimize'
	call literal
	TICK doisc
	jmp compilec
iscase.n:
	jmp disp ndferd
csetick:A;:ok:
	call itick		; '
	spush
; test whether @pfa = xec-entry, then 2+ 4+ 1+ @+ + => does-entry of the rsp list-type
	call idx2do		; get ptr to DOES>-entry
	cmp eax,docasec-Y
	jz .c			;?; "case:"
	cmp eax,dorangec-Y
	jz .c			;?; "range:"
	cmp eax,doassoc-Y	; error if not "associative:"
.c:
	ret
;-
;- #case	( -- n )					'ncase'		forth
;-	if 1st in a compiled word, case:-index by which the word was called.
;-	(can be used to reading the <edx> register content)
ncase:GA
	spush
	mov eax,edx
	ret
;
caseco.i:A
	call here		; cpl'd case:
	call literal
	TICK docasec
	call compilec
	jmp casesc
;-
;- case:  	xec: ( caseindex -- xx )	(i)(f6)(h4)	'caseco'	forth
;- 		cpl: cc1 cc2 ... ;	( -- )	while compiling
;- 		def: name cc1 cc2... ;	( -- )	while interpreting
;-	initiate a case decision by index list of executeables or numeric data:
;-	cpl:
;-		compile typed list of xec tokens or sing numeric values.
;-		list may span multiple lines w/ words separated by {bl}s, only, 
;-		and may be interleved w/ {\ ...} comment lines.
;-	def:
;-		compile header for later execution by name, procede as above.
;-	xec:
;-		check range and according to type, push indexed value or execute.
;-	
;-	Executeable type list is a 'jump table' of xt-s, by ix or true address, 
;-	numeric type returns the rsp. entry corresponding to the supplied index.
;-	state dependently built as a named word or, compiled into a forth word. 
;-	valid range: -1 < caseindex < no. of items,
;-	at list execution ovf indices will be fetching from the rsp bounds.
;-	(rather hypothetical) max. no. of list items is 2^31-3.
;-	A numeric mode list will be compiled if any one list item was not found 
;-	as a forth word in current search order, but could be converted to some
;-	numeric value, compiling it's least significant cell, any excess data 
;-	silently discarding. That list returns the rsp. xec token (uot index) in 
;-	any case an executeable word was specified.
;-	
;-	defining a list
;-		compiles each non-immediate item w/ -re- {cases,} until
;-		1st {;} which terminates the list, from input stream;
;-		executes other immediates - which should not leave more 
;-		than one cell on data-stack.
;-	  while compiling
;-	  compiles a {case:} list into the current word:
;-		..words.. case: case1name case2name .. lastcasename ; ..words..
;-	  while interpreting
;-	  builds the by <listname> executeable {case:} list:
;-		case: listname case1name case2name .. lastcasename ;
;-	execution:
;-		xec indexed compiled word or, push indexed value to data-stack.
;-		-ve index refers to 1st, +ve beyond last refers to last item.
;-	modifying a list
;-		-re- {is-case}
caseco:GA A 1;:ok:; align <docasec>
	call ccplq
	mov dword[uref uv.ucmc],0
	jnz .i			;?; create named list
	call ccasec		; build named header and data field
	call pscodep		;:continued: -> docasec
;-
;- (dc		( ix p -- xx )			(f6)		'docasec'	novoc
;-	according to list type fetch or execute by ix indexed cell
docasec:;4-aligned, immediately after <pscodep>
; c: eax, @nos	ptr =: @tos
;    ecx	disp to indexed cell in bytes
;    edx	casec.flag
	mov ecx,TOS(1)		; supplied index
	shl ecx,2		; cells
	mov edx,[kref eax+casec.size]; ptr to list size
	jns .e			;?; within lower bound
	xor ecx,ecx		; adjust to 1st item
.e:
	sub edx,byte casec_size	; size of list header
	cmp ecx,edx
	jc .u			;?; within upper bound
	lea ecx,[edx-4]		; last item
.u:
	cmp byte[kref eax+casec.flag],casec_num; type flag
	lea eax,[eax+ecx+casec_size]
	mov edx,TOS(1)		; for {#case}
	TOSPULL(1)
	jz near fetch		;?; numeric type
	mov ecx,[kref eax]
	spush			; dup @
	mov eax,ecx
	add ecx,KR
	jz .r			;?; null-ptr, inactive
	sub ecx,KR
	jz .r			;?; zero entry, inactive
	call xecq		; convert entry to an executeable address
	inc eax
	mov ecx,TOS(1)
	spop 2
	jnz .f			;?; false/invalid entry
	spop
	jmp ecx			; execute
.r:A
	spop 2
	ret
.f:
	mov ecx,Indferd		; error handler
	mov dword[kref eax],ecx	; store error handler
	jmp ndferd		; exit w/ error message
;-
;- (kc		(X: -- p )			(k)		'pdocasec'	novoc
;-	kernel deferred -re- {(dc}
pdocasec:A
	spush
	mov eax,[kref ecx]
	add eax,databot
	jmp disp docasec
;-
;- cases, 	ccc..( -- p )			(f6)		'casesc'	hidden
;-	initiate and build a 'cnum'-list, from uot indices ('xec tokens') or 
;-	numeric values as found in current search order, from input stream, 
;-	cpl list items into data-space until {;}, xec other immediates. 
;-	ret ptr to list header.
;-	preceded w/ a header field of top case-index and list type:
;-		[ list-type flags][ listsize bytes ][ 1st item ] .. [ last ]
;-		all entries of cell-size, <listsize> inclusive header cells.
;-		l.s.b. flags =/= 0 is numeric type, remaining space reseved.
;-	list structure subject to change, -re- {casehead-size}, {case:}.
casesc:GA;:ok:
; o: edx,@tos	ptr to list header
	call here	; data-ptr
	spush 2		; dup
	xor eax,eax	; 0
	call comma	; cpl flags field
	neg eax		; -ve initial code-ptr for list size
.l:
	call comma	; cpl code-ptr
.c:
	call cpasfq	; S: valid ix or numeric, Z:end of list
	js .l		;?; cpl
	jnz .c		;?; next after imed word
	mov edx,TOS(1)
	spop 2		; done - drop ix of {;}
	mov ecx,dataptr
	add [kref edx+4],ecx; cpl list size
	ret
%ifdef __DBG
	global cpasfq:function,cpasfqf:function
cpasfqf:A 1
cpasfq.f:
%else
cpasfq.f:A 1
%endif
	call refill	; next input line
	spop 3		; drop flag and ptr/count from find
	jz cpasfq.r	;?; eof
;-
;- (#,		( p -- p n | p )		(k)(f6)		'cpasfq'	novoc
;-	fetch a single list item, ret eax & S if ix or num fetched
;-	empty stack and Z after {;}, NS*NZ*NZ ok & no data after imed
;-	abort-31 if neither word nor number
cpasfq:;:ok:
	and dword[byte uref uv.bits],~(mbign|mranu|mquad); suppress bignum, float, quad
	call blwf	; fetch next name from input stream
	spop		; drop flag
	jnz .w		;?; word found
	call cfetchp	; c@+
	test eax,eax
	jz .f		;?; refill
	push PS		; save for truncation to sing
	call abnumberb	; fig-mode [number]
	mov ecx,eax	; Z if not a number
	pop PS		; rst PS, 2 cells
	spop		; drop flag, leave sing
	jnz .e		;?; not found, display name & abort
	mov edx,TOS(1)	; list ptr
	dec ecx		; number fetched, ret S
%if casec_num = 0xff
	mov byte[kref edx],cl; numeric type flg = 0x0ff
%else
	mov byte[kref edx],casec_num; numeric type flg
%endif
	ret
.w:A; ix found
	jng .r		;?; non-immediate word index, ret S*NZ
	cmp eax,Isemi
	jz .r
	push PS
	call execute	; immediate word
	pop PS
	lea PS,[PS-4]	; rst data-stack depth
	xor ecx,ecx
	inc ecx		; NZ, NC, NS <- +ve non-zero: no data, no fault
.r:
	ret
.e:
	call here
	call print
	call space
	jmp rno
;-
;- (ch		( -- p )			(f6)		'ccasec'	novoc
;-	build named list for {associative:}, {case:}, {range:}
ccasec:A
	call create		; named header
	call rrbrac
	call casesc		; data field
	call unsmudge
	jmp llbrac
;-
;- range: 	xec: ( n1 -- n2 )		(f6)		'rangec'	forth
;- 		def: cc1 cc2 ... ;	( -- )	while compiling
;- 		def: name cc1 cc2... ;	( -- )	while interpreting
;-	{ range: rg n1 n2 .. nn ; nx --rg-- nn : nn := index(nm), nm <= nx < nm+1
;-	i.e. ret. index is 0 for values below 1st, num(items) for values above top, 
;-	and, within list range the no. of greatest item which is equal or below n1.
;-	list structure and compiling rules otherwise as numeric -re- {case:}.
	global rangec:function; (aligned by <ccasec> posn)
rangec:;:ok:
	call ccplq
	mov dword[uref uv.ucmc],0
	jz .c			;?; create named list
	call here		; cpl'd case:
	call literal
	TICK dorangec
	call compilec
	jmp casesc
.c:;				; named {range:}
	call ccasec		; build named header and data field
	call pscodep		;:continued:
;-
;- (dr		( n1 p -- n2 )			(f6)		'dorangec'	novoc
dorangec:;(4-aligned, immediately after <pscodep>)
	mov ecx,TOS(1)		; parm
	push edi
	lea edi,[kref eax+casec_size]; ptr to data
	mov eax,ecx
	mov ecx,[edi-4]		; bytecount+header
	shr ecx,2
	lea edx,[ecx+1]
.l:
	dec ecx
	jz .r			;?; top ovf
	scasd	; cmp eax,[edi], edi:= edi+4
	jnl .l			;?; index not < val
	inc ecx			; top found
.r:
	dec ecx
	js .t			;?; top range
	inc ecx
.t:
	pop edi
	sub edx,ecx		; index
	TOSPULL(1)
	lea eax,[edx-1]
	ret
;-
;- (rc		(X: -- p )			(k)		'pdorangec'	novoc
;-	kernel deferred -re- {(dr}
pdorangec:A
	spush
	mov eax,[kref ecx]
	add eax,databot
	jmp disp dorangec
;-
;- associative:	xec: ( n1 -- n2 )		(f6)		'assocc'	forth
;- 		def: cc1 cc2 ... ;	( -- )	while compiling
;- 		def: name cc1 cc2... ;	( -- )	while interpreting
;-	return list index n2 of found value n1, else top index+1
;-	list structure and compiling rules otherwise as numeric -re- {case:}.
assocc:GA A 1;:ok:
	call ccplq
	mov dword[uref uv.ucmc],0
	jz .c			;?; create named list
	call here		; cpl'd case:
	call literal
	TICK doassoc
	call compilec
	jmp casesc
.c:				; named {associative:}
	call ccasec		; build named header and data field
	call pscodep		;:continued:
; (da		( n1 p -- n2 )			(f6)		'doassoc'	novoc
doassoc:;4-aligned, immediately after <pscodep>
	mov ecx,TOS(1)		; parm
	push edi
	lea edi,[kref eax+casec_size]; ptr to data
	mov eax,ecx
	mov ecx,[edi-4]		; bytecount+header
	shr ecx,2
	sub ecx,byte 2		; max-count
	mov edx,ecx		; save for result
	repnz scasd
	pop edi
	jz .r
	dec ecx
.r:
	sub edx,ecx		; index
	TOSPULL(1)
	lea eax,[edx-1];;mov eax,edx
	ret
;-
;- (ac		(X: -- p )			(k)		'pdoassoc'	novoc
;-	kernel deferred -re- {(dc}
pdoassoc:A
	spush
	mov eax,[kref ecx]
	add eax,databot
	jmp disp doassoc
;	
array:GA
	jmp nimp
;
dim:GA
	jmp nimp
;
%endif;__DEFINING_I
;-
;-=====================================================================;
;- l4defining.inc <eof>
