;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4dsarith.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __DSARITH_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __DSARITH_I
%define __DSARITH_I
%define dsarithinc
;-
;-================================================================
;--- arithmetic opr
;-

;-
;-
;-----------------------------------------------------------------
;--- single integers
;-

; min/max/abs w. neither division nor conditional code (tf=-1), but slow...
; : max 2dup < >rr and swap r> not and or ;
; : min 2dup > >rr and swap r> not and or ;
; : abs dup 0 min minus swap 0 max max ;

;-
;- abs		( n1 -- n2 )			(e)(m,10)	'cabs'		forth
;-	negate n1 if -ve; { dup 0 min minus swap 0 max max }
;-	"negative zero" case: ret 0 if n1 = -0
;-	(e: C if n1 was -0)
cabs:GA M;:ok:
	test eax,eax
	jns .r
	neg eax
	jns .r
	add eax,eax	;shl eax,1	; -0
.r:
M cabs
	ret
;-
;- ABS		( n1 -- n2 )			(e)(m,6)	'aabs'		ans
;-	ret n2 := | n1 | = { dup 0 min minus swap 0 max max }
;-	no "negative zero", ret n2 := n1 if n1 = -0
;-	(e: C if n1 was -0)
aabs:GA M;:ok:
	test eax,eax
	jns .r
	neg eax
.r:
M aabs
	ret
;-
;- 0max		( n1 -- n2 )			(k6)(m,6)	'zmax'		forth
;-	n2 = n1 larger than or adjusted equal to zero
zmax:GA M;:ok:
	test eax,eax
	jns .r
	mov eax,0
.r:
M zmax
	ret
;-
;- max		( n1 n2 -- n3 )							forth
;-	ret the larger of two signed single integers
max:GA;:ok:
; c: ecx
	cmp eax,TOS(1)
	TOSPULL(1)
	jge .r
	mov eax,[PS]
.r:
	ret
;-
;- umax		( u1 u2 -- u3 )							forth
;-	ret the larger of two unsigned single integers
umax:GA;:ok:
; c: ecx
	cmp eax,TOS(1)
	TOSPULL(1)
	jnc .r
	mov eax,[PS]
.r:
	ret
;-
;- min		( n1 n2 -- n3 )							forth
;-	ret the smaller of two signed single integers
min:GA;:ok:
; c: ecx
	cmp eax,TOS(1)
	TOSPULL(1)
	jle .r
	mov eax,[PS]
.r:
	ret
;-
;- umin		( n1 n2 -- n3 )							forth
;-	ret the smaller of two unsigned single integers
umin:GA;:ok:
	cmp eax,TOS(1)
	TOSPULL(1)
	jbe .r	;umax:ja .r
	mov eax,[PS]
.r:
	ret
;-
;- +-		( n1 n2 -- n3 )			(fg)		'spm'		forth
;-	apply sign =/= 0 of n2 to n1.
spm:GA;:ok:
	cmp eax,0;(alignment)
	spop
	jns negate.r
;-
;- negate  	( n1 -- n2 )							forth
;-	2-s complement sing @tos
;-
;- minus  	( n1 -- n2 )			(fg)(m,2)	'negate'	forth
;-	2-s complement sing @tos
negate:GA M;:ok:
; c: -/-
	neg eax
.r:
M negate
	ret
;-
;- +		( n1 n2 -- n3 )			(e)(m,8)	'plus'		forth
;-	2-s cpl add single cells
plus:GA M;:ok:
	add eax,TOS(1)
	TOSPULL(1)
M plus
	ret
;-
;- -		( n1 n2 -- n3 )			(e)(m,10)	'ssub'		forth
;-	2-s cpl subtract sing
ssub:GA M;:ok:
	sub TOS(1),eax
	spop
M ssub
	ret
;-
;- swap-  	( n1 n3 -- n3 )			(e)(m,8)	'swsub'		forth
;-	reverse subtract: leave result of { n3 n1 - }
swsub:GA M
	sub eax,TOS(1)
	TOSPULL(1)
M rsub
	ret
;-
;- mod		( n1 n2 -- r )			(e)				forth
;-	2-s cpl div sing remainder
mod:GA;:ok:
; c: ecx,edx
	call by
	mov eax,edx
	ret
;-
;- /mod		( n1 n2 -- r q )		(e)		'bymod'		forth
;-	2-s cpl div sing remainder & quotient, 'floored'
bymod:GA;:ok:
; c: ecx,edx
	call by
	spush
	mov TOS(1),edx
	ret
;-
;- /		( n1 n2 -- q )			(m,11)		'by'		forth
;-	2-s cpl div sing quotient, 'round near zero' mode
;-	if n2=0 ret q = sgn(n1) ovf value or, 1 if n1=n2
;-	check zero or ovf conditions before, if macro compiling.
by:GA;:ok:(w/ sighnd)
; c: ecx,edx, ovf by signal hnd
	mov ecx,eax
	spop
	mov edx,eax
	sar edx,31
	push ebx
	lea ebx,[kref .h-Y]
	idiv ecx
	pop ebx
	ret
.h:A;:ok: sigfpe handler entry
	pop ebx
	mov eax,[PS]
	inc ecx
	jz .n		;?; div -1
	dec ecx
	jz .z		;?; div 0
	dec ecx
	jns .r		;?; div >0
.n:
	neg eax
.r:
	mov edx,eax
	sar edx,31
	ret
.z:A
	cmp eax,ecx
	mov eax,1
	jz .e
	mov eax,0x7fffffff
	jg .q
	neg eax
.q:
	mov edx,eax
.e:
	ret
;-
;- gcd		( n1 n2 -- u3 )			(e)		'gcd'		forth
;-	greatest common denominator u3 of the absolute value of two signed sings. 
;-	If either number is zero, returns the absolute value of the other number.
;-	(e: Z if u3 = 1)
gcd:GA;:ok:
;alternative: 2dup = -exit 2dup > if 2dup - -rot sdrop else 2dup swap - sdrop endif recurse ;
;enter begin swap over mod entry -dup 0= until ;
; m,n >0, dann ggt:
;	1) m mod n => rest r
;	2) ist r=0 dann n=ggt, fertig
;	3) m:=n, n:=r und nach 1)
	test eax,eax	;m
	mov edx,eax	;r := nom
	mov ecx,TOS(1)	;n
	TOSPULL(1)
.n:
	jns .1		; abs
	neg edx
.1:			; swap
	test ecx,ecx	; abs
	jz .x		;?; ret den if nom = 0
	jns .e
	neg ecx
	jr .e		; enter
.b:A;			; begin
	mov ecx,edx	;d:=n:=r
	xor edx,edx
	div ecx		;d:=r'
.e:;
	test edx,edx	;r
	mov eax,ecx	;m:=n
	jnz .b		; -dup until
.x:
	dec ecx		; Z if unit
	ret
;-
;- *		( n1 n2 -- n3 )			(m,6)		'smul'		forth
;-	multiply sings
smul:GA M;:ok:
; c: ecx,edx
	imul dword TOS(1)
	TOSPULL(1)
M smul
	ret
;-
;- */mod  	( n1 n2 n3 -- n4 n5 )		(e)(a:c)	'mulbym'	forth
;-	signed divide by n2 double product of signed multiplication n1 by n2
;-	leave signed quotient n5 and remainder n4; rounding mode near zero.
;-	(e: overflow into double quot can be checked with NZ flag from {flg}.
mulbym:GA;:ok:
	push eax
	mov eax,TOS(1)
	imul dword TOS(2)
	mov TOS(1),edx
	mov TOS(2),eax
	pop eax
	call mby
	neg eax		; C: ovf, O: -0
	spop		; drop
	ret
;-
;- */		( n1 n2 n3 -- n4 )				'mulby'		forth
;-	signed divide by n2 double product of signed multiplication n1 by n2
;-	division result rounding mode as with m/.
mulby:GA;:ok:
	call mulbym
	TOSPULL(1)	; swap drop
	ret
;- 
;- lg(base)	( n1 -- n2 )							forth
;-	numerus of log n1 wrt actual {base}
;-	(which is one less than count of digits of the printed number n1)
;-	hi-level defn:
;-		: lg(base) 0 begin swap base @ / ?dup while swap 1+ repeat ;
lgbase:GA;:ok:
	mov ecx,[byte uref uv.base]
	jr log.n
;- 
;- log		( n1 n2 -- n3 )							forth
;-	numerus of log n1, to the base of n2; if n1 = 0 ret n3 := 0
log:GA;:ok:
	mov ecx,eax
	spop
.n:
	cmp ecx,byte 2
	jl .r		;?; false @base
	push_ibyte(-1)
.l:A
	xor edx,edx
	div ecx
	inc dword[esp]
	test eax,eax
	jnz .l
	pop eax
.r:
	ret
;-
;- lg2    	( d1 -- n2 )					'lg2'		forth
;-	ret n2 = numerus of log double, d1 to the base of 2
lg2:GA;:ok:
	mov edx,TOS(1)
.n:
	xor ecx,ecx
	bsr ecx,eax
	spop
	jnz .r
	bsr ecx,edx
	lea ecx,[ecx-32]
.r:
	lea ecx,[ecx+32]
	mov eax,ecx
	ret
;-
;- 2^	  	( n1 -- d2 )					'pow2'		forth
;-	ret d2 := n1-th power of two, -1 < n1 < 64
pow2:GA;:ok:
	mov ecx,eax
	mov eax,1
	spush
	dec eax
	spush
	mov eax,ecx
	jmp dlshift
;-
;- pow2?  	( n -- flg )			(e)		'pow2q'		forth
;-	tf if signed n is a power of 2 (including n=1, n=80000000h).
;-	(e: flg value)
pow2q:GA;:ok:
; o: edx := eax, n preserved for caller (align.. words)
	mov edx,eax
	sub eax,byte 1
	jc .z		;?; n = 0
	jns .p		;?; +ve or 80000000h
	not edx
	not eax
.p:
	and eax,edx
.z:
	mov eax,1
	jnz .r		;?; not a power of 2
	dec eax		; S flag!
.r:
	dec eax		;; ! eflags for <aligned> etc ! ;
	ret
;-
;- bsf		( n1 f -- n2 | -1 )				'bsff'		forth
;-	scan for bit-no. n2 of 1st set bit, beginning at least significant bit,
;-	use n1 as a cell value if f=0 -re- {bsfr}-
;-	else n1 = ptr to memory and cells range = f -re- {bsfm}:
;-		{ n 0 bsf } aequivalent to { n sp@ 1 bsf }.
bsff:GA;:ok:
	test eax,eax
	jz bsfr.n
;-
;- bsfm		( p n1 -- n2 | -1 )						forth
;-	bit scan forward in memory, range limited by count n1 > 0 cells.
;-	ret bitcount n2 from ptr p to the 1st non-zero bit in address range [p,p+n1*4].
bsfm:GA;:ok:
	lea edx,[eax*4+4]; maxcount bytes
	spop
	lea ecx,[kref eax]
	push ecx	; initial ptr
	lea ecx,[ecx+edx-4]; top cell+1
	mov eax,-1	; 'not found'
	neg edx		; -ve disp+4
.c:
	add edx,byte 4	; next cell, bot up
	jns .r
	bsf eax,[ecx+edx]
	jz .c
	add ecx,edx
	sub ecx,[esp]	; bytes
	lea eax,[eax+ecx*8];posn+bits
.r:
	lea esp,[esp+4]
	ret
bsfr.n:
	spop
;-
;- bsfr		( n1 -- n2 | -1 )		(m,10)				forth
;-	bit scan register
bsfr:GA M;:ok:
	mov ecx,eax
	mov eax,-1
	bsf eax,ecx
M bsfr
	ret
;-
;- bsr		( n1 f -- n2 | -1 )				'bsrf'		forth
;-	scan for bit-no. n2 of most significant set bit, beginning at m.s. bit,
;-	n1 = cell value if f = 0, else n1 = ptr to memory, scanning up to f cells.
bsrf:GA;:ok:
	test eax,eax
	jz bsrr.n
;-
;- bsrm		( p n1 -- n2 | -1 )						forth
;-	bit scan reverse in memory, range limited by count n1 > 0 cells.
;-	ret bitcount n2 from ptr p to the last non-zero bit in address range [p,p+n1*4].
bsrm:GA;:ok:
	lea edx,[eax*4]	; maxcount bytes
	spop
	lea ecx,[kref eax]; base address
	push ecx
	mov eax,-1	; 'not found'
.c:
	sub edx,byte 4	; next cell, top dn
	js .r
	bsr eax,[ecx+edx]
	jz .c
	add ecx,edx
	sub ecx,[esp]	; bytes
	lea eax,[eax+ecx*8];posn+bits
.r:
	lea esp,[esp+4]
	ret
bsrr.n:
	spop
;-
;- bsrr		( n1 -- n2 | -1 )		(m,10)				forth
;-	bit scan register
bsrr:GA M;:ok:
	mov ecx,eax
	mov eax,-1
	bsr eax,ecx
M bsrr
	ret
;-
;- tstb		( p1 n2 -- flg )		(m,13)				forth
;-	the cpu BT memory instruction, flg =/= 0 if specified bit is set.
;-	test bit no. n2 in range of [-2^31..2^31-1] wrt memory ptr p1.
tstb:GA;:ok:
	mov ecx,eax
	spop
	bt [kref eax],ecx
	sbb eax,eax
M btst
	ret
;-
;- setb		( p1 n2 -- flg )		(m,13)		'setb4'		forth
;-	the cpu BTS memory instruction, set bit n2 wrt n1 - re {tstb}
setb4:GA M;:ok:
	mov ecx,eax
	spop
	bts [kref eax],ecx
	sbb eax,eax
M bset
	ret
;-
;- resb		( p1 n2 -- flg )		(m,13)		'resb4'		forth
;-	the cpu BTR memory instruction, reset bit n2 wrt n1 - re {tstb}
resb4:GA M;:ok:
	mov ecx,eax
	spop
	btr [kref eax],ecx
	sbb eax,eax
M bres
	ret
;-
;- cplb		( p1 n2 -- flg )		(m,13)				forth
;-	the cpu BTC instruction, complement bit n2 wrt n1 - re {tstb}
cplb:GA M;:ok:
	mov ecx,eax
	spop
	btc [kref eax],ecx
	sbb eax,eax
M cplb
	ret
;-
;-
;-----------------------------------------------------------------
;--- double integers
;-

;-	

;-
;- d+-		( d1 n1 -- d2 )			(79)		'dpm'		forth
;-	apply sign of sing n1 to double d1.
dpm:GA;:ok:
	test eax,eax
	spop
	js dnegate
	ret
;-
;- dabs		( d1 -- +d1 )			(f6)				forth
;-	negate double d1 if -ve
dabs:GA;:ok:
	test eax,eax
	jns dnegate.r	;:continues:
;-
;- dminus  	( d1 -- -d1 )			(fg)(m,8)	'dnegate'	fig
;- dnegate	( d1 -- -d1 )			(a)(m,8)			forth
;-	2-s complement of double
;-	(eflags reflect result)
dnegate:GA M;:ok:
	neg dword TOS(1)
	adc eax,byte 0
	neg eax
.r:
M dnegate
	ret
;-
;- u+		( dn1 u2 -- dn3 )		(e)		'uplus'		forth
;-	unsigned add dn1 + u2 := dn3
;-	(eflags reflect double cell addition)
uplus:GA;:ok:
; c: ecx
	spush
	xor eax,eax
	jr dplus
;-
;- m+		( dn1 n2 -- dn3 )		(e)(79)		'mplus'		forth
;-	signed add n2 + dn1 := dn3
;-	(eflags reflect double cell addition)
mplus:GA;:ok:
; c: ecx
	spush
	sar eax,31
;-
;- d+		( dn1 dn2 -- dn3 )		(e)		'dplus'		forth
;-	add d2 to d1
;-	(eflags reflect double cell addition)
dplus:GA;:ok:
; c: ecx
	mov ecx,TOS(1)
	add TOS(3),ecx
	adc TOS(2),eax
	spop 2
	ret
;-
;- d-  		( dn1 dn2 -- dn3 )		(e)		'dsub'		forth
;-	subtract d2 from d1
;-	(eflags reflect double cell subtraction)
dsub:GA;:ok:
; c: ecx
	mov ecx,TOS(1)	; 2l
	sub TOS(3),ecx	; 1l-2l
	sbb TOS(2),eax	; 1h-2h-c
	spop 2
	ret
;-
;- ud*		( ud1 ud2 -- ud3 )		(4)		'udmul'		forth
;-	unsigned multiply doubles, ret double prd
udmul:GA;:ok:
	call umdmul
	jmp drop2
;-
;- d*		( d1 d2 -- d3 )			(4)		'dmul'		forth
;-	signed multiply doubles, ret double prd
dmul:GA;:ok:
	call mdmul
	jmp drop2
;-
;- dmod   	( d1 d2 -- d3 )			(4)				forth
;-	divide doubles d1 by d2, leave signed double rmd d3
dmod:GA;:ok:
	lea ecx,[kref drop2-Y]
	push ecx
	jr dbym
;-
;- d/		( d1 d2 -- d3 )			(4)		'dby'		forth
;-	divide doubles d1 by d2, leave signed double qot d3
dby:GA;:ok:
; { d/mod 2sdrop }
	lea ecx,[kref sdrop2-Y]
	push ecx	;:continues: in dbym
;-
;- d/mod  	( d1 d2 -- d3 d4 )		(4)		'dbym'		forth
;-	divide doubles d1 by d2, leave signed doubles rmd d3 and qot d4
dbym:GA;:ok:
;2 pick >r >rr 2>r dabs d->q 2r> dabs uqd/mod r> r xor d+- 2swap r> d+- 2swap ;
	mov ecx,TOS(2)	; 2 pick
	push ecx	; >r		sg(dvd)
	xor ecx,eax	; dup r xor
	push ecx	; >r		sg(dvd x dvs)
	call tor2	; 2>r
	call dabs	; dabs
	spush
	xor eax,eax	; d->q
	spush
	call rfrom2	; 2r>
	call dabs	; dabs
	call uqdbym	; uqd/mod
	spush
	pop eax		; r>
	call dpm	; d+-		sg(qot) = sg(dvd x dvs)
	call swap2	; 2swap
	spush
	pop eax		; r>
	call dpm	; d+-		sg(rmd) = sg(dvd)
	jmp swap2	; 2swap ;s
;-
;- dgcd  	( d1 d2 -- d3 )			(4)		'dgcd'		forth
;-	returns the greatest common denominator d3 of the absolute value of two signed 
;-	doubles. If either number is zero, returns the absolute value of the other one.
dgcd:GA;:ok:
;dabs 2swap dabs enter begin 2sover dmod entry 2dup d0= until 2drop ;
	call dabs	; abs
	call swap2	; swap
	call dabs	; abs
	jr .e		; ENTER
.b:;			; BEGIN
	call sover2	;  swap over
	call dmod	;  mod
.e:;			; ENTRY
	mov ecx,eax	;  -dup
	or ecx,TOS(1)	;  0=
	jnz .b		; until
	spop 2		; ;s
	ret
;-	
;-
;-----------------------------------------------------------------
;--- mixed double and sing
;-	

;-
;- m/mod  	( dn n1 -- n2r n3q )		(v)		'mbym'		forth
;-	double by sing to sing rmd and quot division,
;-	xec from uvari (m/mod) which initially is vectored to {sm/rem}.
mbym:GA
	jmp [uref uv.upmbymp]
;-
;- sm/rem  	( d1 n1 -- r2 q3 )		(e)(a:c)	'smbyrem'	forth
;-	signed divide d1 by n1,
;-	leave symmetric quotient q3 and remainder r2.
;-	Dividend Divisor | Quotient Remainder
;-	-------- ------- + -------- ---------
;-	   20       7    |     2        6    
;-	  -20       7    |    -2       -6    
;-	   20      -7    |    -2        6    
;-	  -20      -7    |     2       -6    
;	  dvd      dvs   |   qot      rmd    
;-	 dvs = 0 ovf:
;-		dvd > 0  $7fffffff  $7fffffff
;-		dvd < 0  $80000001  $80000001
;-		dvd = 0          1   0
;-	(e: overflow into double quot. if NZ flags)
smbyrem:GA;:ok:(w/ sighnd)
;sm/rem  ( d1 n1 -- n2 n3 )
; over >r dup >r abs -rot dabs rot um/mod
; r> r@ xor 0< IF      negate      THEN
; r>        0< IF swap negate swap THEN ;
; --
	mov ecx,eax
	lea eax,[kref .r-Y]
	push eax	; ret from exception hnd
	push ebx
	mov eax,TOS(2)
	mov edx,TOS(1)
	lea ebx,[kref mby.x-Y]
	mov [PS],edx
	idiv ecx
	pop ebx
	_add esp,4
.r:
	TOSPULL(1)
	mov TOS(1),edx
	ret
;-
;- m/		( d1 n2 -- n3 d4 )		(e)		'mby'		forth
;-	signed divide double d1 by single n2,
;-	leave symmetric double quotient d4 and remainder n3.
;-	Dividend Divisor | Quotient Remainder
;-	-------- ------- + -------- ---------
;-	   20       7    |     2        6    
;-	  -20       7    |    -2       -6    
;-	   20      -7    |    -2        6    
;-	  -20      -7    |     2       -6    
;	  dvd      dvs   |   qot       rmd   
;-	 dvs = 0 ovf:
;-		dvd > 0  $7fffffff.ffffffff  $7fffffff
;-		dvd < 0  $80000000.00000001  $80000001
;-		dvd = 0          0.00000001   0
;-	[hopefully, all that sick "exceptions" stuff will safely be intercepted...
mby:GA;:ok:(no sighnd)
	mov edx,TOS(1)	; dvd.hi
	mov ecx,eax	; dvs
	xor eax,edx
	shr eax,16
	mov dx,ax	; sgn(r) = sgn dvd|sgn(q) = sgn(dvd xOR dvs)
	push edx
	test ecx,ecx
	mov eax,TOS(2)
	mov edx,TOS(1)
	jz .n
	jns .s		;?; +ve dvs
	neg ecx
.s:
	test edx,edx
	jns .d		;?; +ve dvd
	neg eax
	adc edx,byte 0
	neg edx
	mov TOS(2),eax	; save d.lo
.d:
	mov eax,edx
	sar edx,31
	div ecx
	mov TOS(1),eax	; q.hi
	mov eax,TOS(2)	; d.lo
	div ecx
	mov TOS(2),edx	; r
	mov ecx,eax
	mov eax,TOS(1)
	mov TOS(1),ecx
	pop ecx		; sgn flags
	test ecx,ecx
	jns .q		;?; +ve rmd
	neg dword TOS(2)
.q:
	shl ecx,16
	jns .r		;?; +ve qot
	neg dword TOS(1)
	adc eax,byte 0
	neg eax
.r:
	ret
.n:A; div 0
	call rzreg
	mov [esp],ebx	; drop sgn flags
.b:; <umby)
	pop ebx
	mov TOS(2),edx	; rmd
	movsx eax,al
	mov TOS(1),eax	; q.lo
	mov dl,dh	; +ve: 7fffffff -ve: 80000000
	mov eax,edx	; q.hi
	ret	
; hnd passed by <ebx> =/= kref, common hnd-s re <fpecx>, <fpecxi>
;in: stored dvd.hi in [PS], dvd.l in TOS(2), kref reg @tor
.x:A; idiv 0, idiv ovf	; by fm/mod, sm/rem
	pop ebx		; rst saved kref
	mov edx,[PS]	; d.hi or q.hi (after 1st fpe)
	mov eax,TOS(2)	; d.lo
	test ecx,ecx
	jz .z		;?; div0
	inc ecx
	jz .m
	sub ecx,byte 2
	jz .p
	xor ecx,edx
	mov edx,1
	jns .z		;?; +ve qot
	neg edx
.z:; < div0
	or eax,edx
	jz .e		;?; 0/0
	sar edx,31
	not edx		; +ve:-1 -ve:0
	lea edx,[edx*2+0x80000001];r: +/- max
	movsx eax,dl	; q.lo
	mov ecx,edx	; q.hi
	mov cl,dh	; ext'd to 64bit value
	ret
.e:A;	;< 0/0
	mov ecx,eax	; q.hi
	inc eax		; q.lo
	ret
.m:A;	dn -1 /
	neg eax
	adc edx,0
	neg edx
.p:;	dn 1 /
	ret
;-
;- um/mod  	( ud u1 -- u2 u3 )		(e)(GA)		'umbymod'	forth
;-	unsigned divide ud by u1
;-	leave quotient u3 and remainder u2.
;-	 dvs = 0 ovf:
;-		dvd > 0  $ffffffff $ffffffff
;-		dvd = 0          1 0
umbymod:GA;:ok:(umby)
; 	 edx returns overfow flag
	lea ecx,[kref drop-Y]
	push ecx	;:continues: in um/
;-
;- um/		( ud1 u2 -- u3 ud4 )		(e)		'umby'		forth
;-	unsigned divide double ud1 by single u2,
;-	leave unsigned double quotient ud4 and remainder u3.
;-	[this is "lib4th" basic unsigned division operation.]
;-	 dvs = 0 ovf:
;-		dvd > 0  $ffffffff.ffffffff $ffffffff
;-		dvd = 0          0.00000001 0
umby:GA;:ok:(w/ sighnd)
; edx returns overfow flag,
; div0 and ovf 'exceptions' handled by signal handler
	mov ecx,eax	; ecx:=dvd
	spop
	lea edx,[kref .f-Y]
	push ebx
	push edx
	lea ebx,[kref fpecx-Y]
	_mov edx,0	; unsigned dvd.hi
	div ecx		; a:=qot.h d:=r.h=q.lh
.h:
	mov [PS],eax	; qh
	mov eax,TOS(1)	; tos0:=q.h a:dvd.lo
	div ecx		; a:=qot.l d:=rmd
	_add esp,4
	pop ebx
.r:
	mov ecx,[PS]	; qh
	spush		; push qot.lo, tos:=qot.h
	mov TOS(2),edx	; rmd
	mov edx,eax	; 
	mov eax,ecx	; d:=qot.h tos1:=rmd
	ret
.f:A;<umby) 2ndary sighnd
	test ecx,ecx
	jz .z		;?; div0
	push dword .y-Y
	add [esp],KR
	jr .h
.y:A
	test ecx,ecx
.z:
	pop ebx
	jnz .r		;?; regular ovf
	mov eax,[PS]	; ret 1.,0 or -1. -1
	or eax,TOS(1)
	spush		; q.lo := 1
	jz .x		;?; 0/0
	_mov eax,-2
.x:
	inc eax
	mov TOS(1),eax	; q.lo f..f or 1
	mov al,ah	; $ffffffff or 0
	mov TOS(2),eax	; rmd
	ret
;-
;- fm/mod  	( d1 n1 -- n2 n3 )		(e,?)(GA)	'fmbymod'	forth
;-	signed divide dvd=d1 by dvs=n1, leave floored quotient n3 and remainder n2.
;-	fm/mod from sm/rem and, if quot 0<, rmd := rmd + divisor, quot := quot - 1.
;-	 Dividend Divisor   Quotient Remainder
;-	 -------- ------- | -------- ---------
;-	     10      7    |    1        3
;-	    -10      7    |   -2        4
;-	     10     -7    |   -2       -4
;-	    -10     -7    |    1       -3
;-	 dvs = 0 overflow values:
;-		dvd > 0    $7fffffff  $7fffffff
;-		dvd < 0    $80000001  $80000001
;-		dvd = 0     1          0
fmbymod:GA;:ok:(w/ sighnd)
;fm/mod  ( d1 n1 -- n2 n3 )
; dup >r dup 0< IF negate >r dnegate r> THEN
; over       0< IF tuck + swap THEN
; um/mod  r> 0< IF swap negate swap THEN ;
; --
; sgn(r) = sgn dvs|sgn(q) = sgn(dvd xOR dvs)
; 	 edx returns overfow flag
	mov ecx,eax
	lea eax,[kref .q-Y]
	push eax	; ret from xep-hnd
	push ebx	; kref
	mov eax,TOS(2)	; dvd.lo
	mov edx,TOS(1)	; dvd.hi
	lea ebx,[kref mby.x-Y]
	mov [PS],edx	; pass dvd.hi to fault hnd
	idiv ecx
	pop ebx
	_add esp,4	; drop ret from xep-hnd
	test edx,edx
	jz .r		;?; zero rmd
	test eax,eax
	jns .r		;?; +ve quot 
	dec eax		; qot := qot-1
	add edx,ecx	; rmd := rmd+dvs
.r:
	mov TOS(2),edx	; rmd
	TOSPULL(1)
	ret
.q:;	;<ret from xep-hnd)
	test edx,edx
	jz .r
	mov eax,edx
	jr .r
;-
;- m*		( n1 n2 -- d3 )			(f6)(a)(m,8)	'mmul'		forth
;-	signed multiply sings n1 by n2 to double d3
mmul:GA M;:ok:
; c: edx
;;	mov edx,TOS(1);
	imul dword TOS(1);;edx;;same timing-?-)
	mov TOS(1),eax
	mov eax,edx
M mmul
	ret
;-
;- u*		( n1 n2 -- d3 )			(f6)(m,8)	'umul'		forth
;- um*		( n1 n2 -- d3 )			(a)(m,8)	'umul'		forth
;-	unsigned multiply sings, leave double
umul:GA M;:ok:(=ummul)
; c: edx
	mul dword TOS(1)
	mov TOS(1),eax
	mov eax,edx
M umul
	ret
;-
;- dm*		( d1 n2 -- d3 ovf )		(e)		'dmmul'		forth
;-	signed multiply double by sing to double.
;-	ovf is m.s. cell of treble product.
;-	(e: NZ*S if ovf =/= 0)
dmmul:GA;:ok:
; c: ecx,edx
	spush
	sar eax,31
	call mdmul	; -> quad
	mov edx,eax	; save qhh
	spop
	ret
;-
;- udm*		( ud1 u2 -- ud3 ovf )		(e)		'udmmul'	forth
;-	unsigned multiply double by sing to double,
;-	ovf is m.s. cell of treble product.
;-	(e: NZ*S if ovf =/= 0)
udmmul:GA;:ok:<btonumb)
; c: ecx,edx, ecx:=ovf, edx:=prd.hi
	mov ecx,eax	; u2
	mul dword TOS(2); a:=mll d:=mlh
	mov TOS(2),eax	; pd.lo.lo
	mov eax,TOS(1)
	mov TOS(1),edx	; pd.lo.hi intermediate @nos
	mul ecx		; a:=mhl d:=mhh
	add TOS(1),eax
	sbb ecx,ecx	; flag ovf
	mov eax,edx
	ret
;-
;- udm*/mod	( ud1 u2 u3 -- u4 ud5 )		(4)		'udmmulbym'	forth
;-	unsingned double ud5 = ud1*u2/u3, u4=rmd; w/ quad intermediate opr.
udmmulbym:GA
	call tor
	call udmmul
	spush
	xor eax,eax
	call rfrom
	call uqsmby
	spop 2
	ret
;-
;- m*/		( d1 n2 n3 -- d4 )				'mmulby'	forth
;-	singned double d4 = d1 * n2 / n3; w/ quad intermediate opr.
mmulby:GA;:ok:
; >r s->d md* >rr 0< if qnegate endif r> r> swap over xor >r
; 0< if negate endif uqsm/ r> if qnegate endif 2drop rot drop ;
	push eax
	mov eax,TOS(1)
	sar eax,31	; ( d2 )
	call mdmul	; md* -> quad
	test eax,eax
	push eax	; sgn(d1*n2)
	jns .s
	call qnegate
.s:
	spush
	mov eax,[esp+4]	; n3
	test eax,eax
	jns .m
	neg eax
.m:
	call uqsmby	; qn4
	pop ecx
	pop edx
	xor ecx,edx
	jns .r
	call qnegate
.r:
	spop 2
	jmp rotd
;-
;- um*/		( u1 u2 u3 -- ud4 )				'ummulby'	forth
;-	unsingned double ud4 = u1*u2/u3
ummulby:GA;:ok:
	mov ecx,eax	; save u3
	mov edx,TOS(1)	; u2
	mov eax,TOS(2)	; u1
	mul edx
	mov TOS(2),eax
	mov TOS(1),edx
	mov eax,ecx
	call umby
	jmp rotd
;-
;-	
;-

;-
;- u^		( u1 +u2 -- +d3 )				'upow'		forth
;-	return double +d3 as u1 to the pover of u2,
;-	ud3 := 0 if u2 < 0
upow:GA;:ok:
	mov ecx,eax
	dec ecx
	js .o		; u1^0, u1^(-ve)
	mov edx,ecx	;d:=u2-1
	jz .s		; u1 ^ 1
	mov eax,TOS(1)
	mov edx,0
	cmp eax,byte 1
	jna .s		;?; 1^u2=1 or 0^u2=0; u2>0
	push ebx
	mov ebx,eax	; exp
	mov eax,0
.l:A
	mul ebx		; a:=hl d:= hh
	mov edx,eax	; p.hl
	mov eax,TOS(1)	;   ll
	mov TOS(1),edx	; p.hl
	mul ebx
	add edx,TOS(1)	; p.h
	mov TOS(1),eax	; p.l
	mov eax,edx
	dec ecx
	jg .l
	pop ebx
.s:
	mov eax,edx
	ret
.o:A
	xor eax,eax
	mov TOS(1),eax
	inc ecx
	jnz .r		; n1^(-n2) < 1 := 0
	inc dword TOS(1)
.r:
	ret
;-
;- ^		( n1 +n2 -- d3 )				'pow'		forth
;-	return signed double d3 as n1 to the pover of n2,
;-	d3 := 0 if n2 < 0, d3 := 1 if n2 = 0
pow:GA;:ok:
	mov ecx,TOS(1)
	test ecx,ecx
	mov edx,ecx
	sar ecx,31
	and ecx,eax
	push ecx
	test eax,eax
	jns .a
	neg eax
.a:
	test edx,edx
	jns .d
	neg edx
	mov TOS(1),edx
.d:
	call upow
	pop ecx
	and cl,1
	jnz near dnegate
	ret
;-
;- udsqrt  	( ud1 -- ud2 j )		(x)				forth
;-	ud2 := sqare root of unsigned double |ud1|; j := -1 if top bit of ud1 set (-ve).
;-	note: ud2 m.s. sing always zero, left in place for -re- {vsqrt}, {dvsqrt}.
udsqrt:GA;:ok:
%ifdef __MMX
;
input x
;
p := 0; 
a := 0
;
loop 1 to #bits(x)/2
;
        shift left x into a 2 places
;
        shift left p one place
;	 t := 2*p
;
        if a > t then a:=a-(t+1); p:=p+1
;
end
;
return p
	mov edx,TOS(1)		; x.lo
	test eax,eax
	push ebx		; save KR
	push eax		; save "j"
	mov ecx,31
	jnz .c			;?; source more than 32 bit
;::-:: could further be optimized by count/2 of leading zeroes
	sub ecx,byte 16
	mov TOS(1),eax
	mov eax,edx		; swap drop 0 
.c:
	spush
	xor eax,eax		; x=TOS(1)
	movq mm0,TOS(2)		;m0: x
	spush 3			; a=TOS(3), p=TOS(5)
	push esi		; save lp
	push ecx		; bit-count
	mov edx,eax		; a := edx.hi:eax.lo
	mov TOS(5),eax
	mov TOS(4),eax
	movd mm1,eax		; p
.l:A
	movq TOS(1),mm0		;@tos: x
	movq mm2,mm1		; t:=p
	psllq mm0,2		; x:=x*4
	psllq mm1,1		; p:=2p
	psllq mm2,2		; t=4p
	mov esi,TOS0		; x.hi
	movq TOS(3),mm2		; t
	shld edx,eax,2		; a.hi<-a.lo
	shld eax,esi,2		; a.lo<-x.hi
	mov ecx,TOS(3)		;c: t.lo
	mov ebx,TOS(2)		;b: t.hi
	sub ecx,eax
	sbb ebx,edx		; t-a
	movq TOS(5),mm1		; p':= p
	jnc .n			; a < t
	not ecx
	not ebx			; t := -(t-a)-1 = a-(2p+1)
	add dword TOS(5),byte 1
	adc dword TOS(4),byte 0	; p':= p+1
	mov eax,ecx		; a := t
	mov edx,ebx
	movq mm1,TOS(5)		; p := p'
.n:
	dec dword[esp]
	jns disp .l
	lea esp,[esp+4]		; drop counter
	pop esi
	pop eax
	TOSPULL(3)		; drop temps
	pop ebx
	sar eax,31		; "j"
	ret
%endif;
;-
;- udsqrt  	( ud1 -- ud2 j )				'fudsqrt'	non-mmx
;-	ud2 := +ve sqare root of double |ud1|; j := -1 if ud1 signed < 0.
;-	note: ud2 m.s.cell always zero, left in place for -re- {vsqrt}, {dvsqrt}.
;-	ud2 is nearest integer the sqare of which is lower or equal |ud1|*(2j-1).
fudsqrt:GA
; non-mmx substitute, do not take apart from above code
; non-mmx variant by about 15% slower than <dsqrt>
%define p_hi TOS(2)
%define p_lo TOS(3)		; result accu
%define t_hi ebx
%define t_lo ecx
%define a_hi edx
%define a_lo eax
%define x_hi esi
%define x_lo ebp
	push ebx		; save KR
	push eax		; save "j"
	spush
	xor eax,eax
	spush
	push esi		; save lp
	push ebp		; save UP
	push_ibyte(31)		; bit-count
; - might be optimized by count/2 of leading zeroes
	mov esi,TOS(2)		; x.hi
	mov ebp,TOS(3)		; x.lo
	mov p_lo,eax
	mov p_hi,eax		; p := TOS(3)
	mov edx,eax		; a := edx.hi:eax.lo
.l:A
	mov t_hi,p_hi		; p.temp=ecx|ebx
	mov t_lo,p_lo
%ifdef __K6__; ...is a tiny bit faster (ca.0,5%)
	add t_lo,t_lo
	adc t_hi,t_hi		; p := 2p
%else
	shl t_lo,1
	rcl t_hi,1		; p := 2p
%endif
	mov p_lo,t_lo		; save p, intermediate result
	mov p_hi,t_hi
%ifdef __K6__
	add t_lo,t_lo
	adc t_hi,t_hi		; p := 2p
%else
	shl t_lo,1
	rcl t_hi,1		; t := 4p
%endif
	shld a_hi,a_lo,2	; a.hi<-a.lo
	shld a_lo,x_hi,2	; a.lo<-x.hi	a=4*a+[2 m.s.b (x)]
	shld x_hi,x_lo,2	; x.hi<-x.lo
	shl x_lo,2		; <- x := 4x
	sub t_lo,a_lo
	sbb t_hi,a_hi		; t-a
	jnc .n			; a <= t
	not t_lo
	not t_hi		; t := -(t-a)-1 = a-(2p+1)
	mov a_lo,t_lo		; a := t
	mov a_hi,t_hi
	add dword p_lo,byte 1
	adc dword p_hi,byte 0	; p':= p+1
.n:
	dec dword[esp]
	jns disp .l
	lea esp,[esp+4]		; drop counter
	pop ebp			; rst UP
	pop esi			; rst lp
	TOSPULL(1)		; drop temp
	pop eax			; sign x
	pop ebx			; rst KR
	sar eax,31		; "j"
	ret
;-	
;-
;-----------------------------------------------------------------
;--- fast & special arithmetics
;-
;-	

;-
;- flg		( -- u )			(m,3)				tools
;-	fetch eflags register from latest opr
;-	can efficiently be used immediately after those words
;-	which are marked "(e)".
flg:GA M;:ok:
	pushfd
	spush
	pop eax
M flg
	ret
;-
;- flg>smb	( n -- p u )					'flg2smb'	tools
;-	convert eflags value to symbolic string at {hld}+,
;-	string corrupted by consecutive numeric conversions.
flg2smb:GA;:ok:
	call ltsharp	; <#
	spush		;
	mov ecx,32	; 31
	mov TOS(1),ecx	; swap
.l:
	add eax,eax	; dup 0< swap 2*
	jc .h		; swap dup 0< 0= while ...
.c:
	dec ecx		;            ...repeat
	jnz .l
	jmp sharpgt	; #>
.h:A
	movzx edx,byte[kref ecx+efsymt-Y-1]
	test edx,edx	; over efsymt c@ -dup
	jz .c		; if
	spush
	mov TOS(2),ecx
	mov eax,edx
	call hold	;  hold
	mov ecx,TOS(1)	; endif
	jr .c		;       ->  ...repeat
;-
;- chars  	( n1 -- n2 )			(a:c)(m,0)			forth
;-	multiply n1 by charsize (1, i.e. a {noop} word)
;-
;- 2*		( n1 -- n2 )			(e)(m,2)	'mul2'		forth
;-	multiply n1 by 2
mul2:GA M;:ok:
	add eax,eax
M mul2
	ret
;-
;- noop		( -- )				(fg)(m,1)			forth
;-	most efficiently do nothing,
;-	hi-level call immediately returns,
;-	macro inserts a <nop> operation.
noop:GA M;:ok:
	ret
M,noop
nopcode:
	nop
M noop
;-
;- 6*		( n1 -- n2 )			(e)(m,4)	'mul6'		forth
;-	multiply n1 by 6
mul6:GA M;:ok:
	add eax,eax
;-
;- 3*		( n1 -- n2 )			(e)(m,3) 	'mul3'		forth
;-	multiply n1 by 3
mul3:GA M;:ok:
	lea eax,[eax+eax*2]
M mul3
M mul6
	ret
;-
;- 10*		( n1 -- n2 )			(e)(m,3)	'mul10'		forth
;-	multiply n1 by 10
mul10:GA M;:ok:
	add eax,eax	;;shl eax,1
;-
;- 5*		( n1 -- n2 )			(e)(m,3)	'mul5'		forth
;-	multiply n1 by 5
mul5:GA M;:ok:
	lea eax,[eax+eax*4]
M mul5
M mul10
	ret
;-
;- 7*		( n1 -- n2 )			(e)(m,5)	'mul7'		forth
;-	multiply n1 by 7
mul7:GA M;:ok:
	mov ecx,eax
	shl eax,3
	sub eax,ecx
M mul7
	ret
;-
;- 9*		( n1 -- n2 )			(e)(m,3)	'mul9'		forth
;-	multiply n1 by 9
mul9:GA;:ok:
	lea eax,[eax+eax*8]
M mul9
	ret
;-
;- cells  	( n1 -- n2 )			(a:c)(m,3)	'mul4'		ans
;-	multiply n1 by cellsize(4)
;- 4*		( n1 -- n2 )			(e)(m,3)	'mul4'		forth
;-	multiply n1 by 4
mul4:GA M;:ok:
	shl eax,2
M mul4
	ret
;-
;- FLOATS 	( n1 -- n2 )			(e)(m,3)	'mul8'		ans
;- SFLOATS	( n1 -- n2 )			(e)(m,3)	'mul8'		ans
;- 8*		( n1 -- n2 )			(e)(m,3)	'mul8'		forth
;-	multiply n1 by 8
;- abits  	( n1 -- n2 )			(m,3)		'mul8'		ans
;-	n2 := no. of n1 address units bits
mul8:GA M;:ok:
	shl eax,3
M mul8
	ret
    global floats:function,dfloats:function,sfloats:function
floats:  equ mul8
sfloats: equ mul8
;-
;- DFLOATS	( n1 -- n2 )			(e)(m,3)	'mul16'		ans
;- 16*		( n1 -- n2 )			(e)(m,3)	'mul16'		forth
;-	multiply n1 by 16
mul16:GA M;:ok:
	shl eax,4
M mul16
	ret
dfloats: equ mul16
;-
;- @lshift	( p +u -- )			(e)(m,9)	'mlshift'	forth
;-	shift cell in memory at ptr p by +u < 32 bits
mlshift:GA M;:ok:
	mov ecx,eax
	spop
	shl dword[kref eax],cl
	spop
M mlshift
	ret
;-
;- @rshift	( p +u -- )			(e)(m,9)	'mrshift'	forth
;-	shift cell in memory at ptr p by +u < 32 bits
mrshift:GA M;:ok:
	mov ecx,eax
	spop
	shr dword[kref eax],cl
	spop
M mrshift
	ret
;-
;- @rshifta	( p +u -- )			(e)(m,9)	'mrshifta'	forth
;-	shift cell in memory at ptr p by +u < 32 bits, propagate sign bit
mrshifta:GA M;:ok:
	mov ecx,eax
	spop
	sar dword[kref eax],cl
	spop
M mrshifta
	ret
;-
;- rot-l  	( x1 u -- x2 )			(e)(m,9)	'rotl'		forth
;-	left rotate x1 by u bits
rotl:GA M;:ok:
	mov ecx,eax
	spop
	rol eax,cl
M rotr
	ret
;-
;- shift  	( x1 n -- x2 )							forth
;-	32 > n > 0 : {lshift}, -32 < n < 0 {rshift}; x2 := 0 if |n| > 31.
shift:GA;:ok:
	neg eax
	jns rshift.s
	neg eax
	cmp eax,byte 32
	jnc near dzneq	;?; ret x2:=0 if |n| > 32
;-
;- lshift  	( x1 u -- x2 )			(e)(a:c)(m,9)			forth
;-	unmodified cpu opr, i.e. x1 unchanged for u = any multiple of 32
lshift:GA M;:ok:
	mov ecx,eax
	spop
	shl eax,cl
M lshift
	ret
;-
;- 2/		( n1 -- n2 )			(e)(m,12)	'div2'		forth
;-	signed divide n1 by 2. 'round towards zero' mode.
;-	(NOTE: correct result incl. odd -ve, not ANS-4th compliant)
div2:GA M;:ok:
	cmp eax,0x80000000; add 1 if n1<0 { .. dup 0< - .. }, for odd -ve n1
	sbb eax,-1	; C := ^sg(n1)
;-
;- 2/		( n1 -- n2 )			(e)(m,2)	'shr2'		ans
;-	signed divide n1 by 2, 'floored rounding' mode -re- DPANS 6.1.0330.
shr2:M;:ok:
	sar eax,1
M shr2
M div2
	ret
;-
;- 4/		( n1 -- n2 )			(e)(m,13)	'div4'		forth
;-	signed divide n1 by 4 (cellsize). 'round towards zero' mode.
div4:GA M;:ok:
	mov edx,eax
	sar edx,31
	and edx,byte 3
	add eax,edx
;-
;- CELL/  	( n1 -- n2 )			(e)(m,3)	'cellby'	ans
;-	signed divide n1 by 4 (cellsize). w/ floored rounding.
cellby:M;:ok:
;by macro 'M':    global cellby:function
	sar eax,2
M cellby
M div4
	ret
;-
;- 8/		( n1 -- n2 )			(e)(m,) 	'div8'		forth
;-	signed divide n1 by 8. 'round towards zero' mode.
div8:GA M;:ok:
	mov edx,eax
	sar edx,31
	and edx,byte 7
	add eax,edx
	sar eax,3
M div8
	ret
;-
;- 16/mod 	( n1 -- n2 n3 )			(e)(m,) 	'div16m'	forth
;-	signed divide n1 by 16, ret quot n2, rem n3. 'round towards zero' mode.
div16m:GA M;:ok:; AMD
	mov edx,eax	;sign extend into EDX
	spush
	sar edx,31	;EDX = 0xFFFFFFFF if dividend < 0
	and edx,byte 15	;mask correction abs(div)-1
	add eax,edx	;pre-correction
	mov ecx,eax
	sar eax,4	; quot
	and ecx,byte 15	;remainder
	sub ecx,edx	;pre-correction
	mov TOS(1),ecx
M div16m
	ret
rshift.s:
	cmp eax,byte 32
	jnc near dzneq	;?; ret x2:=0 if |n| > 32
;-
;- rshift  	( x1 u -- x2 )			(e)(a)(m,9)			forth
;-	unmodified cpu opr, i.e. x1 unchanged for u = any multiple of 32
rshift:GA M;:ok:
	mov ecx,eax
	spop
	shr eax,cl
M rshift
	ret
;-
;- rshifta	( n1 u -- n2 )			(e)(a)(m,9)			forth
;-	unmodified cpu opr, i.e. x1 unchanged for u = any multiple of 32
rshifta:GA M;:ok:
	mov ecx,eax
	spop
	sar eax,cl
M rshift
	ret
;-
;- rot-r  	( x1 u -- x2 )			(e)(m,9)	'rotr'		forth
;-	right rotate x1 by u bits
rotr:GA M;:ok:
	mov ecx,eax
	spop
	ror eax,cl
M rotr
	ret
%ifdef __MMX
dmul8:GA;:ok:
	mov cl,3
	jr dmul2.n
dmul4:GA;:ok:
	mov cl,2
 %if 1
dmul2:GA
	shl dword TOS(1),1
	rcl eax,1
	ret
 %else
	jr dmul2.n
dmul2:GA;:ok:
	mov cl,1
 %endif
dmul2.n:
	spush
	movzx eax,cl	;:continues:
;-
;- 2lshift	( dx1 u -- dx2 )		(e)(x)		'dlshift'	forth
;-	logic shift left double
dlshift:GA;:ok:
	movq mm0,TOS(2)
	movd mm1,eax	; zero xtd to 64bit
	psllq mm0,mm1
	movq TOS(2),mm0
	spop
	ret
%else
;-
;- d8*		( d1 -- d2 )			(x)		'dmul8'		forth
;-	multiply double by 8
dmul8:GA;:ok:
	mov cl,3
	jr dmul2.n
;-
;- d4*		( d1 -- d2 )			(x)		'dmul4'		forth
;-	multiply double by 4
dmul4:GA;:ok:
	mov cl,2
	jr dmul2.n
;-
;- d2*		( d1 -- d2 )			(e)(x)(a:d)	'dmul2'		forth
;-	multiply double by 4
dmul2:GA;:ok:
	mov cl,1
.n:
	spush
	movzx eax,cl	;:continues:
dlshift:GA;:ok:
%endif
;-
;- 2lshift	( dx1 u -- dx2 )		(e)		'dlshiftf'	non-mmx
;-	logic shift left double
dlshiftf:GA;:ok:
	mov cl,al
	mov ch,al
	spop
	sub ch,32
	js .s
	mov eax,TOS(1)
	mov dword TOS(1),0
	jz .r
	mov cl,ch
.s:
	mov edx,TOS(1)
	shld eax,edx,cl
	shl edx,cl
	mov TOS(1),edx
.r:
	ret
;-
;- dslbits	( d1 -- d1 n )							forth
;-	ret no. n of bits by which d1 may be shifted left w/o change of sign.
dslbits:GA
; for scaling +ve double:
; find top set bit and cell, complement less 2, +ve or zero, is max scaling power of 2
	mov edx,PS
.e:; <dvsqrt, 2nd)
	spush
.n:; <dvsqrt, 1st)
	mov ecx,0
	mov eax,29	; 63 bits max, less sign bit, less start bsr:0 vs. shift:1
	cmp ecx,[edx]	; @@tos
	jnz .b
	lea edx,[edx-4]
	lea eax,[eax+32]
.b:
	bsr ecx,[edx]
	sub eax,ecx
	jns .r
	mov eax,0
.r:
	ret
;-
;- 2rshifta	( d1 u -- d2 )			(e)(x)		'drshifta'	forth
;-	double shift right, arithmetic
;-	floored, signed division by 2^(u), u < 64.
drshifta:GA;:ok:
%ifdef __MMX;:ok: gain depends on shift count, though always better than w/o mmx
	cmp al,32
	jc .s
	mov ecx,TOS(1)
	sar ecx,31
	TOSPUSH(2)	;	{ u dx dn
	mov TOS(2),ecx
	mov TOS(1),ecx	;	{ u dsgn dn
	movq mm0,TOS(4)	; dn
	psrlq mm1,32	; clr hi.d
	movd mm1,eax	; shift count
	movq mm2,TOS(2)	; -1
	movq mm3,TOS(2)	; -1
	psrlq mm2,mm1	; 0bits (2)
	psrlq mm0,mm1	; dn'	(0)
	pxor mm2,mm3	; 1bits (2)
	por mm0,mm2	; signed dn'
	movq TOS(4),mm0
	spop 2
	spop
	ret
.s:A;:ok:
	mov ecx,eax
	spop
.n:
	shrd dword TOS(1),eax,cl
	sar eax,cl
	ret
%endif
;-
;- 2rshifta	( dx1 u -- dx2 )		(e)		'drshiftfa'	non-mmx
;-	shift right a double, sign preserving
drshiftaf:GA;:ok:
	mov cl,al
	mov ch,al
	spop
	sub ch,32
	js .n
	mov TOS(1),eax
	sar eax,31
	sar eax,1
	mov cl,ch
.n:
	shrd dword TOS(1),eax,cl
	sar eax,cl
.r:
	ret
;-
;- d8/		( d1 -- d2 )			(e)		'deightby'	forth
;-	divide double by 8
deightby:GA;:ok:
	mov cl,3
	jr drshiftaf.n
;-
;- d4/		( d1 -- d2 )			(e)		'dfourby'	forth
;-	divide double by 4
dfourby:GA;:ok:
	mov cl,2
	jr drshiftaf.n
;-
;- d2/		( d1 -- d2 )			(e)(a:d)	'dtwoby'	forth
;-	divide double by 2
dtwoby:GA;:ok:
	mov cl,1
	jr drshiftaf.n
;-
;- 2rshift	( dx1 u -- dx2 )		(e)(x)		'drshift'	forth
;-	logic shift right double, u mod 64
drshift:GA;:ok:
%ifdef __MMX
	movq mm0,TOS(2)
	movd mm1,eax
	psrlq mm0,mm1
	movq TOS(2),mm0
	spop
	ret
%endif
;-
;- 2rshift	( dx1 u -- dx2 )		(e)		'drshiftf'	non-mmx
;-	logic shift right double, u mod 64. non-mmx subst.
drshiftf:GA;:ok:
	mov cl,al
	mov ch,al
	spop
	sub ch,32
	js .s
	mov TOS(1),eax
	mov eax,0
	jz .r
	mov cl,ch
.s:
	shrd dword TOS(1),eax,cl
	shr eax,cl
.r:
	ret
;-
;- 1+		( n1 -- n2 )			(e,fg)(m,1)	'add1'		forth
;- char+  	( n1 -- n2 )			(e,C)(m,1)	'add1'		forth
;-	increment sing by 1
add1:GA M;:ok:
	inc eax
M add1
	ret
;-
;- 1-		( n1 -- n2 )			(e,fg)(m,1)	'sub1'		forth
;- char-  	( n1 -- n2 )			(e,C)(m,1)	'sub1'		forth
;-	decrement sing by 1
sub1:GA M;:ok:
	dec eax
M sub1
	ret
;-
;- 2+		( n1 -- n2 )			(e)(fg)(m,3)	'add2'		forth
;-	increment sing by 2
add2:GA M;:ok:
	add eax,byte 2
M add2
	ret
;-
;- 2-		( n1 -- n2 )			(e)(fg)(m,3)	'sub2'		forth
;-	decrement sing by 2
sub2:GA M;:ok:
	sub eax,byte 2
M sub2
	ret
;-
;- 4+		( n1 -- n2 )			(e)(m,3)	'add4'		forth
;- cell+  	( n1 -- n2 )			(GA)(m,3)	'add4'		forth
;-	n2 := n1 + cellsize(4), increment sing by 4
add4:GA M;:ok:
	add eax,byte 4
M add4
	ret
;-
;- 4-		( n1 -- n2 )			(e)(f6)(m,3)	'sub4'		forth
;- cell-  	( n1 -- n2 )			(a)(m,3)	'sub4'		forth
;-	n2 := n1 - cellsize(4), decrement sing by 4
sub4:GA M;:ok:
	sub eax,byte 4
M sub4
	ret
;-
;- 8+		( n1 -- n2 )			(e)(f6)(m,3)	'add8'		forth
;-	increment sing by 8
add8:GA M;:ok:
	add eax,byte 8
M add8
	ret
;-
;- 8-		( n1 -- n2 )			(e)(f6)(m,3)	'sub8'		forth
;-	decrement sing by 8
sub8:GA M;:ok:
	sub eax,byte 8
M sub8
	ret
;-
;- d1+		( d1 -- d2 )			(e)(f6)(m,7)	'donep'		forth
;-	increment double by 1
donep:GA M;:ok:
	add dword TOS(1),1
	adc eax,byte 0
M donep
	ret
;-
;- d2+		( d1 -- d2 )			(e)(f6)(m,7)	'dtwop'		forth
;-	increment double by 2
dtwop:GA M;:ok:
	add dword TOS(1),2
	adc eax,byte 0
M dtwop
	ret
;-
;- d4+		( d1 -- d2 )			(e)(f6)(m,7)	'dfourp'	forth
;-	increment double by 4
dfourp:GA M;:ok:
	add dword TOS(1),4
	adc eax,byte 0
M dfourp
	ret
;-
;- d8+		( d1 -- d2 )			(e)(f6,m,7)	'deightp'	forth
;-	increment double by 8
deightp:GA M;:ok:
	add dword TOS(1),byte 8
	adc eax,byte 0
M deightp
	ret
;-
;- d1-		( d1 -- d2 )			(e)(f6)(m,7)	'donem'		forth
;-	decrement double by 1
donem:GA M;:ok:
	sub dword TOS(1),byte 1
	sbb eax,byte 0
M donem
	ret
;-
;- d2-		( d1 -- d2 )			(e)(f6)(m,7)	'dtwom'		forth
;-	decrement double by 2
dtwom:GA M;:ok:
	sub dword TOS(1),byte 2
	sbb eax,byte 0
M dtwom
	ret
;-
;- d4-		( d1 -- d2 )			(e)(f6)(m,7)	'dfourm'	forth
;-	decrement double by 4
dfourm:GA M;:ok:
	sub dword TOS(1),byte 4
	sbb eax,0
M dfourm
	ret
;-
;- d8-		( d1 -- d2 )			(e)(f6)(m,7)	'deightm'	forth
;-	decrement double by 8
deightm:GA M;:ok:
	sub dword TOS(1),byte 8
	sbb eax,byte 0
M deightm
	ret
%endif;__DSARITH_I
;-
;-=====================================================================;
;- l4dsarith.inc <eof>
