;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4dstack.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __DSTACK_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __DSTACK_I
%define __DSTACK_I
%define dstackinc
;-
;-================================================================
;--- basic data-stack opr
;-
;-	dpans94: stacks should not be accessed directly!
;-	returnstack grows towards lower addresses, datastack grows
;-	towards higher addresses, thus they differ in cell order.
;-	@tos stored in eax register, @(sp@) may temporarily be valid
;-	further, data will be stored to returnstack offset by <kref>
;-	and adjusted back when fetched to datastack.
;-
;-----------------------------------------------------------------
;- datastack
;-

;-
;- depth  	( -- n )							forth
depth:GA;:ok:
	spush
	lea eax,[PS-4]
	sub eax,KR
	sub eax,[byte uref uv.s0]
	sar eax,2
	ret
;-
;- -depth  	( -- n )					'mdepth'	forth
;-	free space on datastack
mdepth:GA;:ok:
	call depth
	neg eax
	add eax,65536/4
	ret
;-
;- cdepth  	( -- n )							compiler
;-	no. of -double- items in compilestack, valid only after {!csp} or {:}.
cdepth:GA;:ok:
	spush
	mov eax,[uref uv.ucsp]
	test eax,eax
	jz .r
	lea eax,[kref eax+4]
	neg eax
	add eax,PS
	sar eax,3
.r:
	ret
;-
;- sp!		( | p -- )			(fg|a:c)	'spstore'	forth
;-	vectored, either to forth or ans version of {sp!}.
spstore:GA;:ok:
	spush
	mov eax,[uref uv.ustosp]
	jmp execute.c
;-
;- sp>		( | p -- )			(m,5)		'fspstore'	forth
;-	re-initiate datastack from @{s0}
;-
;- sp!		( -- )				(fg)		'fspstore'	fig
;-	f.i.g.-mode: re-initiate datastack from @s0 - F83 mode by ans-word {SP!}.
;-	being the safer one this is in {forth} and {fig} vocs for precedence over
;-	-re- ANS-mode {SP!}, which is in {ans} vocabulary.
fspstore:GA M;:ok:
; c: -/-
	mov PS,KR
	add PS,[byte uref uv.s0]
M spstore
	ret
;-
;- SP!		( p -- )			(a:c)		'aspstore'	ans
;-	reset datastack to kref(p). ANS-std word, in {ans} wordlist.
;-	(e: eflags remain unchanged)
aspstore:GA M;:ok:
	lea PS,[kref eax]
M spstore
	ret
;-
;- sp@		( -- a )			(fg)		'spfetch'	forth
;-	get datastack ptr
;-	ANS: programs shall NOT access datastack by address!
;-	@tos stored to corresponding address but, might get lost with operations to 
;-	follow. references to above @tos will be safe. "above" is to descending addr!
spfetch:GA M;:ok:
	spush
	lea eax,[PS-4]
	sub eax,KR
M spfetch
	ret
;-
;- @csp		( -- )				(f6)		'fcsp'		hidden
;-      restore datastack ptr from saved in {csp} compilestack ptr
;-	(e: eflags remain unchanged)
fcsp:GA;:ok:
; c: ecx
	mov ecx,[uref uv.ucsp]
	lea PS,[kref ecx]
	ret
;-
;- !csp		( -- )				(m,12)		'scsp'		forth
;-	store datastack ptr to uvari csp.
;-	executed by {head} and {:noname}, thus by all named defining words.
scsp:GA M;:ok:
	mov [uref uv.ucsp],PS
	sub [uref uv.ucsp],KR
M scsp
	ret
;-
;- csp@		( -- p )					'cspf'		hidden
;-      fetch kref'd compilestack ptr to tos
;-	(e: eflags remain unchanged)
cspf:GA;:ok:
	spush
	mov eax,[uref uv.ucsp]
	ret
;-
;- -drop  	( n | 0 -- 0 )					'mdrop'		forth
;-	drop n if non-zero
mdrop:GA;:ok:
	test eax,eax
	jnz drop
	ret
;-
;- 0drop  	( n | 0 -- n )			(m,9)		'zdrop'		forth
;-	drop if zero
zdrop:GA M;:ok:
	test eax,eax
	jnz drop.r
;-
;- drop		( n1 -- )			(e)(m,5)			forth
;-	remove item from tos
;-	(e: eflags remain unchanged)
drop:GA M;:ok:
	spop
.r:
M drop
M zdrop
	ret
;- ?DUP		( n1|0 -- n1 n1 | 0 )		(a:c)(83)	'qdup'		ans
;-	duplicate n1 if non-zero
qdup:GA M;:ok:
;-
;- -dup		( n1|0 -- n1 n1 | 0 )		(fg)(m,4)	'mdup'		forth
;-	duplicate n1 if non-zero
mdup:GA M;:ok:
	test eax,eax
	jz cdup.r
;-
;- dup		( n1 -- n1 n1 )			(m,1)		'cdup'		forth
;-	duplicate item on tos
;-	(e: eflags remain unchanged)
cdup:GA M;:ok:
	spush
M cdup
.r:
M mdup
M qdup
	ret
;-
;- ddup		( n1 -- n1 n1 n1 )		(m,2)		'ddup'		forth
;-	twice duplicate sing on tos
;-	(e: eflags remain unchanged)
ddup:GA M;:ok:
	spush 2
M ddup
	ret
;-
;- d2dup  	( d1 -- d1 d1 d1 )		(m,x)		'd2dup'		forth
;-	twice duplicate double on tos
;-	(e: eflags remain unchanged)
d2dup:GA M;:ok:
	mov ecx,TOS(1)
	spush 4
	mov TOS(1),ecx
	mov TOS(3),ecx
M d2dup
	ret
;-
;- 2sdup  	( d1 d2 -- d2 d1 d1 )		(x)		'sdup2'		forth
;-	does{ 2swap 2dup } or, { 2over 2swap }
;-	(e: eflags remain unchanged)
sdup2:GA;:ok:
%ifdef __MMX;:ok: 2*
	spush 2
	movq mm1,TOS(5)
	movq mm0,TOS(3)
	movq TOS(1),mm1
	movq TOS(3),mm1
	movq TOS(5),mm0
	mov eax,[PS]
	ret
%endif
;- 2sdup  	( d1 d2 -- d2 d1 d1 )				'twosdupf'	non-mmx
;-	does{ 2swap 2dup } or, { 2over 2swap }; non-mmx subst.
twosdupf:GA;:ok:
	call swap2
;-
;- 2dup		( d1 -- d1 d1 )			(a:d)(m,8)	'dup2'		forth
;-	duplicate double @tos
dup2:GA M;:ok:
; c: ecx
	mov ecx,TOS(1)
	spush 2
	mov TOS(1),ecx
M dup2
	ret
;-
;- 3dup		( n1 d2 -- n1 d2 n1 d2 )	(f6)(x)		'dup3'		forth
;-	duplicate top three cells @tos
;-	(e: eflags remain unchanged; mmx mode w/o timing gain)
dup3:GA;:ok:
%ifdef __MMX; :ok: 0*
	movq mm0,TOS(2)
	spush 3
	movq TOS(2),mm0
	ret
%endif
;-
;- 3dup		( n1 n2 n3 -- n1 n2 n3 n1 n2 n3 )		'dup3f'		non-mmx
;-	duplicate top three cells @tos; non-mmx subst.
;-	(e: eflags remain unchanged)
dup3f:GA;:ok:
	spush 3
	mov ecx,TOS(4)
	mov edx,TOS(5)
	mov TOS(1),ecx
	mov TOS(2),edx
	ret
;-
;- swap		( n1 n2 -- n2 n1 )		(m,8)				forth
;-	exchange contents of tos and nos cells
;-	(e: eflags remain unchanged)
swap:GA M;:ok:
; c: ecx
; simple and short "xchg eax,TOS(1)" is 4 times slower
; => but may be required when multi-tasking!
	mov ecx,TOS(1)
	mov TOS(1),eax
	mov eax,ecx
M swap
	ret
;-
;- 2swap   	( d1 d2 -- d2 d1 )				'swap2'		forth
;-	exchange top two doubles at tos
;-	(e: eflags remain unchanged)
swap2:GA;:ok:
; c: ecx	  32 10    10 32
	mov ecx,TOS(3)
	mov edx,TOS(1)
	mov TOS(1),ecx	; 3->1
	mov TOS(3),edx	; 1->3
	mov ecx,TOS(2)	; 2->0
	mov TOS(2),eax	; t->2
	mov eax,ecx
	ret
;-
;- over		( n1 n2 -- n1 n2 n1 )		(m,4)				forth
;-	copy sing @nos to tos
;-	(e: eflags remain unchanged)
over:GA M;:ok:
	spush
	mov eax,TOS(2)
M over
	ret
;-
;- 2over   	( d1 d2 -- d1 d2 d1 )				'over2'		forth
;-	copy double @nos to tos
;-	(e: eflags remain unchanged)
over2:GA;:ok:
	spush
	mov eax,TOS(4)
	spush
	mov eax,TOS(4)
	ret
;-
;- 2rot		( d1 d2 d3 -- d2 d3 d1 )	(x)(a:dx)	'rot2'		forth
;-	roll 3rd double to tos
;-	(e: eflags & non-mmx regs remain unchanged)
rot2:GA;:ok:
%ifdef __MMX; :ok: 5*
	mov [PS],eax
	movq mm0,TOS(1)
	movq mm1,TOS(3)
	movq mm2,TOS(5)
	movq TOS(3),mm0
	movq TOS(1),mm2
	movq TOS(5),mm1
	mov eax,[PS]
	ret
%endif
;-
;- 2rot		( d1 d2 d3 -- d2 d3 d1 )	(a:dx)		'rot2f'		non-mmx
;-	roll 3rd double to tos; non-mmx subst.
rot2f:GA;:ok:
	spush
	mov eax,2
	jr roll2
;-
;- 2-rot  	( d1 d2 d3 -- d3 d1 d2 )			'mrot2'		forth
;-	reverse roll double @tos to 3rd
mrot2:GA;:ok:
; edc preserved! (path+file)
	spush
	mov eax,-2
;- 2roll   	( nx...nm +N -- nx...nm nN )			'roll2'		forth
;-	 	( nx...nm -N -- nN nx...nm )
;-	roll double, re {roll}, index range -16383 < N < 16382
roll2:GA;:ok:
; c: ecx
	shl eax,1
	jo roll.d	;?; bounds ovf
	inc eax
	jns .s
	sub eax,byte 2
	jns roll.d	;?; bounds ovf
.s:
	push eax
	call roll
	spush
	pop eax
;-
;- roll		( n1 nN n3 ... nm +N -- n1 n2 n3 ... nm nN )			forth
;-	move +N-th to tos, move N+1 items one cell up.
;- roll		( n1 nN n3 ... nm -N -- n1 nN n2 n3 ... nm )
;-	move -N items one cell up, move @tos to -N-th cell above tos.
;-	N including 0, in either case.
roll:GA;:ok:
; c: ecx (edx preserved!)
	test eax,eax
	mov ecx,eax
	js .s		;?; roll back
	jz .d		;?; nothing, drop index
	not ecx		; pick
	mov eax,[PS+4*ecx]
	mov [PS],eax	;PS = edi
	push edi,esi
	lea edi,[PS+ecx*4]
	lea esi,[edi+4]
	neg ecx
	rep movsd
	pop regl
.d:
	spop
	ret
.s:A
	spop
	push esi
	push edi
	lea esi,[PS-4]
	lea edi,[esi+4]
	neg ecx
	std
	rep movsd
	cld
	mov [edi],eax
	pop edi
	mov eax,[PS]
	pop esi
	ret
;-
;- 2rotd  	( d1 d2 d3 -- d2 d3 )		(x)		'rotd2'		forth
;-	{ 2rot 2drop }
;-	(e: eflags remain unchanged; mmx mode w/o timing gain)
rotd2:GA;:ok:
%ifdef __MMX;:ok: 0*
	mov [PS],eax
	movq mm1,TOS(3)
	movq mm0,TOS(1)
	TOSPULL(2)
	movq TOS(3),mm1
	movq TOS(1),mm0
	ret
%endif
;- 2rotd  	( d1 d2 d3 -- d2 d3 )				'rotd2f'	non-mmx
;-	(e: eflags remain unchanged)
rotd2f:GA;:ok:
	push dword TOS(1)
	mov ecx,TOS(3)
	mov edx,TOS(2)
	TOSPULL(2)
	pop dword TOS(1)
	mov TOS(3),ecx
	mov TOS(2),edx
	mov [PS],eax
	ret
;-
;- rotd		( n1 n2 n3 -- n2 n3 )		(m,9)				forth
;-	{ rot drop }
;-	(e: eflags remain unchanged)
rotd:GA M;:ok:
; c: ecx
	mov ecx,TOS(1)
	TOSPULL(1)
	mov TOS(1),ecx
M rotd
	ret
;-
;- -rot		( n1 n2 n3 -- n3 n1 n2 )	(79)		'mrot'		forth
;-	(e: eflags remain unchanged)
mrot:GA;:ok:
; c: ecx
	mov ecx,TOS(2)	;n1
	mov TOS(2),eax	;n3
	mov eax,TOS(1)	;n2
	mov TOS(1),ecx	;n1
	ret
;-
;- rot		( n1 n2 n3 -- n2 n3 n1 )					forth
;-	(e: eflags remain unchanged)
rot:GA;:ok:
; c: -/-
	mov ecx,TOS(1)
	mov TOS(1),eax
	mov eax,TOS(2)
	mov TOS(2),ecx
	ret
;-
;- pick		( nm..n1 n -- nm..n1 n{n} )	(m,5)				forth
;-	fetch n-th item to tos, { 0 pick } is {dup}.
;-	NOTE:	meaningless data if -ve index passed to cpl'd macro variant
;-		use -re- {-pick}, instead
;-	(e: eflags remain unchanged)
pick:GA M;:ok:
; c: -/-
	test eax,eax
	js mpick.n
M,pick
	not eax
	mov eax,[PS+4*eax]
M pick
	 ret
;-
;- 2pick   	( nm..n1 n -- nm..n1 n{n} )	(m,12)		'pick2'		forth
;-	(e: eflags remain unchanged)
pick2:GA M;:ok:
; c: ecx
	not eax
	mov ecx,[PS+eax*8+4]
	mov eax,[PS+eax*8]
	spush
	mov eax,ecx
M pick2
	ret
;-
;- -pick  	( nm..n1 n -- nm..n1..n2 )	(m,23)		'mpick'		forth
;-	store n1 as n-th item @tos, position of n1 (item to store) counts for zero.
;-	drop count and n1 if n < 0
mpick:GA M;:ok:
; c: -/-
	neg eax
.n:
	mov ecx,TOS(1)
	jz .p
	jns .r
	mov [PS+4*eax-4],ecx
.r:
	spop
.p:
	spop
M mpick
	ret
;-
;- 2-pick   	( dm..d1 n -- dm..d1..d2 )			'mpick2'	forth
;-	store d1 as n-th item @tos, position of d1 (item to store) counts for zero.
;-	drop count and d1 if n < 0
mpick2:GA;:ok:
; c: ecx,edx
	neg eax
	mov ecx,TOS(1)
	mov edx,TOS(2)
	jz .p
	jns .r
	mov [PS+eax*8-8],edx
	mov [PS+eax*8-4],ecx
.r:
	spop 2
.p:
	spop
	ret
;-
;- third  	( n1 n2 n3 -- n1 n2 n3 n1 )	(m,4)				forth
;-	fetch 3rd item from stack. substitutes { 2 pick }
;-	(e: eflags remain unchanged)
third:GA M;:ok:
; c: -/-
	spush
	mov eax,TOS(3)
M pthird
	 ret
;-
;- 2third  	( d1 d2 d3 -- d1 d2 d3 d1 )	(m,8)		'third2'	forth
;-	fetch 3rd item from stack. substitutes { 2 2pick }
;-	(e: eflags remain unchanged)
third2:GA M;:ok:
; c: -/-
	spush
	mov eax,TOS(6)
	spush
	mov eax,TOS(6)
M third2
	 ret
;-
;--- compile-time stack ref.
;-

	global cspick:function,csroll:function,csdrop:function

;-
;- ds-pick	( n1 -- n2 )(C: ..x.. -- ..x.. )		'dspick'	forth
;-		( -n -- )(C: ..x.. n1 -- n1 ..x.. )		'dspick'	forth
;-	pick from/to saved top of datastack (as stored by {head} or user/appl)
;-	for use while {:}-compiling, in intermediately interpreting state.
;-	storeing an item back to data-stack by -n with -1 referring to @tos.
;-	(e: eflags remain unchanged)
dspick:GA;:ok: (+ve)
	neg eax
	mov ecx,[uref uv.ucsp]
	jg .s
	lea eax,[kref eax*4]
	mov eax,[ecx+eax]
	ret
.s:; store back
	shl eax,2
	sub ecx,eax
	mov eax,TOS(1)
	mov [kref ecx+4],eax;;+4
;-
;- cs-drop	( n -- )(C: s(n)..s1 -- s(n).. )(a:tx)				forth
;-	drop control item (double cell) on datastack
csdrop:
	spop 2
	ret
;-
;- cs-pick	( n -- )(C: s(n)..s1 -- s(n)..s1 s(n) )(a:tx)			forth
;-	pick control item (double cell) from datastack
cspick:equ pick2;(unused, mark for the doc's generation)
;-
;- cs-roll	( n -- )(C: s(n)..s1 -- ..s1 s(n) )(a:tx)			forth
;-	roll control item (double cell) on datastack
csroll:equ roll2;(unused, mark for the doc's generation)
;-
;- cs-depth  	( -- n )					'csdepth'	forth
;-	no. of compile-stack items (doubles) after colon defn. initialization.
csdepth:GA;:ok:
	spush
	lea eax,[PS-4]
	sub eax,KR
	sub eax,[uref uv.ucsp]
	sar eax,3
	dec eax
	ret
;-
;--- 'sillywords' and misc. combined stack opr.
;-

;-
;- 3drop   	( n1 n2 n3 -- )			(e)(m,6)	'drop3'		forth
;-	(e: eflags remain unchanged)
drop3:GA M;:ok:
	spop 3
M drop3
	ret
;-
;- dmin		( d1 d2 -- dn )			(a:d)(15)			forth
;-	drop larger one of two doubles.
dmin:GA;:ok:
; dmin{ 4dup d+ 2rot 2rot d- dabs d- d2/ } 
; dmax{ 4dup d+ 2rot 2rot d+ dabs d- d2/ }
	cmp eax,TOS(2)
	jg drop2
	jl sdrop2
	mov ecx,TOS(1)
	cmp ecx,TOS(3)
	jl sdrop2	;:continued:
;-
;- 2drop  	( dn -- )			(a:c)(m,5)	'drop2'		forth
;-	drop double @tos
;-	(e: eflags unchanged)
drop2:GA M;:ok:
	spop 2
M drop2
	ret
;-
;- dmax		( d1 d2 -- dn )			(a:d)				forth
;-	drop smaller one of two doubles.
dmax:GA;:ok:
	cmp eax,TOS(2)
	mov ecx,TOS(1)
	jl .d
	jg .n
	cmp ecx,TOS(3)
	jg .n
.d:
	mov ecx,TOS(3)
	mov eax,TOS(2)
.n:
	TOSPULL(2)
	mov TOS(1),ecx
	ret
;-
;- 2sdrop  	( d1 d2 -- d2 )			(f6)(m,9)	'sdrop2'	forth
;-	drop double @nos; does{ 2swap 2drop }
;-	(e: eflags remain unchanged)
sdrop2:GA M;:ok:
	mov ecx,TOS(1)
	TOSPULL(2)
	mov TOS(1),ecx
M sdrop2
	ret
;-
;- 2dropd  	( d1 d2 -- d1 d1 )				'dropd2'	forth
;-	copy double @nos to @tos, does{ 2drop 2dup }
;-	(e: eflags unchanged)
dropd2:GA M;:ok:
	mov ecx,TOS(3)
	mov eax,TOS(2)
	mov TOS(1),ecx
M dropd2
	ret
%ifdef __MMX;:ok: 3*
sover2:GA M;:ok:
; c: ecx
	spush 2
	movq mm1,TOS(3)
	movq mm0,TOS(5)
	movq TOS(1),mm1
	movq TOS(5),mm1
	movq TOS(3),mm0
M sover2
	ret
%else
;-
;- 2sover  	( d1 d2 -- d2 d1 d2 )		(f6)(x)(m,10)	'sover2'	forth
;-	{ 2swap 2over } - macro compiling only if mmx enabled.
;-	(e: eflags remain unchanged)
sover2:GA
%endif
;-
;- 2sover  	( d1 d2 -- d2 d1 d2 )		(4)		'twosoverf'	non-mmx
;-	{ 2swap 2over }
;-	(e: eflags remain unchanged)
twosoverf:GA;:ok:
	TOSPUSH(2)	; 5)d1 3)d2 1)xx -- 5)d2 3)d1 1)d2
	mov edx,TOS(4)	; d1h
	mov ecx,TOS(5)	; d1l
	mov TOS(2),edx	; d1h<-
	mov TOS(4),eax	; d2h<-
	mov edx,TOS(3)	; d2l
	mov TOS(3),ecx	; d1l<-
	mov TOS(1),edx	; d2l<-
	mov TOS(5),edx	; d2l<-
	ret
;-
;- 2oswap  	( n1 n2 -- n1 n1 n2 )		(m,14)(x)	'oswap2'	forth
;-	does{ 2over 2swap }
;-	(e: eflags remain unchanged; mmx mode w/ minor timing gain )
oswap2:GA;:ok:
	spush 2
	mov edx,TOS(4)	;n1h	<= { 0:n2h 1:n2h 2:n2h 3:n2l 4:n1h 5:n1l
	mov ecx,TOS(3)	;n2l
	mov TOS(2),edx	;n1h	=> { 0:n2h 1:n2l 2:n1h 3:n1l 4:n1h 5:n1l
	mov TOS(1),ecx	;n2l
	mov edx,TOS(5)	;n1l
	mov TOS(3),edx	;n2l
	ret
;
%ifdef __MMX;:ok: 2*
srot2:GA M;:ok:
; c: ecx
	movq mm0,TOS(5)
	mov ecx,TOS(1)
	mov TOS(4),eax
	movq TOS(1),mm0
	mov TOS(5),ecx
	mov eax,[PS]
M srot2
	ret
%else
;-
;- 2srot  	( d1 d2 d3 -- d3 d2 d1 )	(m,8)(x)	'srot2'		forth
;-	{ 2swap 2rot }, i.e. exchanges @tos and 2nd
;-	(e: eflags remain unchanged; mmx mode w/o timing gain)
srot2:GA
%endif
;-
;- 2srot  	( d1 d2 d3 -- d3 d2 d1 )	()		'srot2f'	non-mmx
;-	{ 2swap 2rot }, i.e. exchanges @tos and 2nd
;-	(e: eflags remain unchanged)
srot2f:GA
	mov edx,TOS(4)
	mov ecx,TOS(5)
	mov TOS(4),eax
	mov eax,TOS(1)
	mov TOS(5),eax
	mov eax,edx
	ret
%ifdef __MMX;:ok: 0*
rots2:GA M
; c: ecx,edx
	movq mm0,TOS(5)
	movq mm1,TOS(3)
	movq TOS(3),mm0
	movq TOS(5),mm1
M rots2
	ret
%else
;-
;- 2rots  	( d1 d2 d3 -- d2 d1 d3 )	(m, )(x)	'rots2'		forth
;-	{ 2rot 2swap }, i.e. swap next on tos doubles
;-	(e: eflags remain unchanged)
rots2:GA
%endif
;-
;- 2rots  	( d1 d2 d3 -- d2 d1 d3 )	()		'rots2f'	non-mmx
;-	{ 2rot 2swap }, i.e. swap next on tos doubles
;-	(e: eflags remain unchanged)
rots2f:GA
	mov ecx,TOS(2)
	mov edx,TOS(4)
	mov TOS(2),edx
	mov TOS(4),ecx
	mov ecx,TOS(3)
	mov edx,TOS(5)
	mov TOS(2),edx
	mov TOS(3),ecx
	ret
;-
;- rotd2s  	( n1 n2 n3 -- n1 n1 n2 n3 )	(m)(x)				forth
;- quck-tuck					(a?)(n.i.)
;-	duplicate 3rd on tos, supporting {skip}, {scan}, etc: { rot dup 2swap }
;-	(e: eflags remain unchanged; mmx mode w/ minor timing gain)
rotd2s:GA M;:ok:
;	( bonni's ranch? - nee, ANS... )
%ifdef __MMX
	movq mm0,TOS(2)
	spush
	movq TOS(2),mm0
M rotd2s
	ret
%endif
;-
;- rotd2s  	( n1 n2 n3 -- n1 n1 n2 n3 )	(m)		'rotd2sf'	non-mmx
;-	substitutes { rot dup 2swap }, non-mmx
;-	(e: eflags remain unchanged)
rotd2sf:GA M;:ok:
	spush
	mov ecx,TOS(2)
	mov edx,TOS(3)
	mov TOS(1),ecx
	mov TOS(2),edx
M rotd2sf
%ifndef __MMX
M rotd2s
%endif
	ret
;-
;- swap2s 	( n1 n2 n3 n4 -- n4 n3 n2 n1 )	(m,20)				forth
;-	revert cells order of top four cells
;-	does{ swap 2swap swap }
;-	-re- {bswap} for conversion of bytes order in a cell
;-	(e: eflags remain unchanged)
swap2s:GA M
	mov ecx,eax	; a->T3, T1->T2, T2->T1,T3->a
	mov eax,TOS(3)	;                       T3->a
	mov edx,TOS(2)	;
	mov TOS(3),ecx	; a->T3
	mov ecx,TOS(1)	;
	mov TOS(1),edx	;                T2->T1
	mov TOS(2),ecx	;        T1->T2
M swap2s
	ret
;-
;- >|<		( n1 -- n2 n3 )			(m,9)		'w2b'		forth
;-	split n1 lo.word into n3:=hi and n2:=lo bytes
;-	(e: eflags remain unchanged)
w2b:GA M;:ok:
; c: ecx
	movzx ecx,ah
	movzx eax,al
	spush
	mov eax,ecx
M w2b
	ret
;-
;- >||<		( n1 -- n2 n3 )			(m,11)		'd2w'		forth
;-	split n1 lo into n3:=hi and n2:=lo words
d2w:GA M;:ok:
; c: ecx
	mov ecx,eax
	movzx eax,ax
	spush
	mov eax,ecx
	shr eax,16
M d2w
	ret
;-
;- bswap  	( n1 -- n2 )			(m,2)		'bbswap'	forth
;-	swap bytes, convert big <-> little endian byte order.
;-	-re- {swap2s} for conversion by cells
;-	(e: eflags remain unchanged)
bbswap:GA M;:ok:
	bswap eax
M bbswap
	ret
;-
;- ><		( n1 -- n2 )			(lm)(m,2)	'xgb'		forth
;-	swap l.s. bytes (8bit portions) of cell n1.
;-	(e: eflags remain unchanged)
xgb:GA M;:ok:
	xchg al,ah
M xgb
	ret
;-
;- wswap  	( n1 -- n2 )			(c1)(m,3)			forth
;-	swap lo and hi words (16bit portions) of n1.
wswap:GA M;:ok:
	ror eax,16
M wswap
	ret
;-
;- unpack  	( n1 -- n2 )							forth
;-	expand l.s. 4 4bit-nibbles to cell of 4 8-bit bytes
unpack:GA;:ok:
	shl eax,4
	shr ax,4
	shl eax,8
	shr ax,4
	shr al,4
	ret
;-
;- pack		( n1 -- n2 )							forth
;-	pack a cells 4 l.s. 4bit-nibbles per 8bit-byte to 16bit word
pack:GA;:ok:
	shl al,4	; x1x2x34-
	shl ax,4	; x1x234--
	shl eax,4	; 1x234---
	shr eax,12	; ---1x234
	shl ax,4	; ---1234-
	shr eax,4	; ----1234
	ret
;-
;- srot		( n1 n2 n3 -- n3 n2 n1 )	(m,8)				forth
;-	{ swap rot }, i.e. exchanges @tos and 2nd
;-	(e: eflags remain unchanged)
srot:GA M;:ok:
; c: ecx
	mov ecx,TOS(2)
	mov TOS(2),eax
	mov eax,ecx
M srot
	ret
;-
;- rots		( n1 n2 n3 -- n2 n1 n3 )	(m,12)				forth
;-	{ rot swap }, i.e. swap next on tos cells
;-	(e: eflags remain unchanged)
rots:GA M;:ok:
; c: ecx,edx
	mov ecx,TOS(2)
	mov edx,TOS(1)
	mov TOS(2),edx
	mov TOS(1),ecx
M rots
	ret
;-
;- NIP						(a:ce)		'sdrop'		ans
;-	drop @nos; does{ swap drop }
;-
;- sdrop  	( n1 n2 -- n2 )			(m,3)				forth
;-	drop @nos; does{ swap drop }
;-	(e: eflags remain unchanged)
sdrop:GA M;:ok:
	TOSPULL(1)
M sdrop
	ret
;-
;- TUCK   					(a:ce)		'sover'		ans
;-	does{ swap over }
;-
;- sover  	( n1 n2 -- n2 n1 n2 )		(m,12)				forth
;-	{ swap over }
;-	(e: eflags remain unchanged)
sover:GA M;:ok:
; c: ecx
	mov ecx,TOS(1)	; n1
	TOSPUSH(1)
	mov TOS(1),ecx
	mov TOS(2),eax	; n2
M sover
	ret
;-
;- over-  	( n1 n2 -- n1 n3 )		(f6)(m,3)	'overm'		forth
;-	{over -}, leave n3 := n2-n1
;-	(e: eflags reflect result)
overm:GA M;:ok:
	sub eax,TOS(1)
M overm
	ret
;-
;- over+  	( n1 n2 -- n1 n3 )		(m,3)		'overp'		forth
;-	{over +}, leave n3 := n2+n1
;-	(e: eflags reflect result)
overp:GA M;:ok:
	add eax,TOS(1)
M overp
	ret
;-
;- oswap  	( n1 n2 -- n2 n2 n1 )		(m,8)				forth
;-	does{over swap}
;-	(e: eflags remain unchanged)
oswap:GA M;:ok:
	mov ecx,eax
	mov eax,TOS(1)
	spush
	mov eax,ecx
M oswap
	ret
;-
;- sdup		( n1 n2 -- n2 n1 n1 )		(m,9)				forth
;-	does{swap dup} or, {over swap}
;-	(e: eflags remain unchanged)
sdup:GA M;:ok:
; c: ecx
	mov ecx,TOS(1)
	mov TOS(1),eax
	mov eax,ecx
	spush
M sdup
	ret
;-
;- dropd  	( n1 n2 -- n1 n1 )		(m,3)				forth
;-	copy @nos to @tos, does{ drop dup }
;-	(e: eflags unchanged)
dropd:GA M;:ok:
	mov eax,TOS(1)
M dropd
	ret
;-
;- m>s		( p n -- n1..nn )				'mtos'		forth
;-	fetch n bytes from ptr p to tos, beginning w. cell aligned n1.
mtos:GA;:ok:
	push esi
	mov ecx,eax
	lea PS,TOS(1)
	mov esi,[PS]
	add esi,KR
	mov dword[PS],0	; for count < 4 and 255 < address
	rep movsb
	pop esi
	dec PS
	and PS,-4
	mov eax,[PS]
	ret
;-
;- ndrop  	( x y(n)...y(1) n -- x | abort )				forth
;-	reserve -n cells, stack content indeterminate or,
;-	drop +n cells from datastack. stack bounds check.
ndrop:GA;:ok:
	not eax
	lea PS,[4*eax+PS]; drop index, add/drop n cells
	mov eax,[PS]	; fetch @tos
	jmp qstack
;-
;- nswap  	( nx0 nx1..nxm ny0 ny1..nym m -- ny0..nym nx0..nxm )		forth
;-	multiple cells swap, swap blocks of m cells
nswap:GA;:ok:
; c: ecx
	lea ecx,[eax*4]	; bytecount
	mov eax,PS
	sub eax,KR
	sub eax,ecx
	spush
	sub eax,ecx
	spush
	mov eax,ecx	;:continues:
;-
;- c-xchg  	( p1 p2 +u -- )					'cxchg'		forth
;-	exchange +u bytes in ranges p1,p1+u2 and p2,p2+u2.
cxchg:GA;:ok:
; c: ecx,edx
	mov ecx,eax	; cnt
	test ecx,ecx
	spop
	mov edx,eax	; destn
	spop		; eax:=src
	jle .r
	push esi,edi
	lea esi,[kref eax]; src
	lea edi,[kref edx]; dst
.l:
	mov al,[edi]
	movsb
	mov [esi-1],al
	dec ecx
	jnz .l
	pop esi,edi
.r:
	spop
	ret
;-
;- swapn  	( nx nN nm .. n0 M -- nx n0 nm .. nN  )				forth
;-	swap N-th cell with cell on tos
swapn:GA;:ok:
	mov ecx,eax
	neg ecx
	spop
	jnc .r
	mov edx,[PS+ecx*4]
	mov [PS+ecx*4],eax
	mov eax,edx
.r:
	ret
;-
;- bounds  	( x1 n -- x2 x1 )		(fg)(m,4)			forth
;-	leave x1 and x2 := x1+n, ready for do..loop etc.
;-	{ over + swap }
;-	(e: eflags reflect result of addition)
bounds:GA M;:ok:
; c: -/-
	xadd TOS(1),eax
M bounds
	ret
%endif;__DSTACK_I
;-
;-=====================================================================;
;-  l4dstack.inc <eof>
