;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-
;- file  	: l4-envq.inc
;- requires	: lib4th.asm
;-=====================================================================;
;-

;- -----------------------------------------------------------------

	global aub:function,ecore:function,floored:function,icore:function,mxu:function,mxud:function,mxc:function
;-
;-----------------------------------------------------------------
;-  ANS-Forth environmental queries
;-	
;- ANS defined environment strings are word names in the {environment} vocabulary
;- which is a {hidden} sub-vocabulary.
;- user definitions which extend lib4th to further standard compliance should
;- thus simply supply the respective query strings by new words defined in the 
;- {environment} vocabulary which return the appropriate values which with no
;- further precautions will then be handled by {environment?}.
;- non-present features can simply be left undefined because the {environment?} 
;- query returns zero for any unknown such word-names' strings.
;-	

;-
;- ENVIRONMENT?	( p u -- ff | .xx. flg )	(a:c)(v)	'envirq'	root
;-	letter case independently find and, execute ANS-4th environment string.
;-	returns zero for any non-present capability.
envirq:GA;:ok:
	call here
	push eax
	call place
	spush
	pop eax
	spush
	mov eax,Ienvironment
	call idx2wid
	call wid2top
	call pfindp
	mov ecx,TOS(1)
	TOSPULL(1)
	jz l4e.r
	mov eax,ecx
	call lfa2idx
	call execute	;:continues:
;-
;- LIB4TH 	( -- 1 -1 )							environment
;-	whether a program is using/based on "lib4th"
l4e:GA
	spush
	mov eax,-1
.r:
	ret
;-
;- CORE 	( -- 0 )							environment
acore:	equ false;:ok:
;- CORE-EXT	( -- 0 )							environment
acorex:	equ false;:ok:
;- FLOORED  	( -- ff )							environment
floored:equ false;:ok:
;-
;- queries which return non-zero values

;-
;- ADDRESS-UNIT-BITS ( -- n )							environment
aub:	equ bbyb;:ok:

;-
;- /COUNTED-STRING ( -- n )							environment
scstg:GA;:ok:
	spush
	mov eax,255
	ret
;-
;- STACK-CELLS	( -- n )							environment
;- DATASTACK-SIZE ( -- n )							environment
;-	(NOTE: data- and return-stack share a common memory area.)
dss:GA
sss:GA
	call depth
	call unused
	jmp plus
;-
;- BLOCK	( -- fl )							environment
bcore:	equ true
;- BLOCK-EXT	( -- fl )							environment
bcorex:	equ true
;-
;- DOUBLE 	( -- fl ) 						 	environment
dcore:	equ true;:ok:
;- DOUBLE-EXT	( -- fl ) 						 	environment
dcorex:	equ true;:ok:
;-
;- EXCEPTION 	( -- fl ) 						 	environment
;-	untested, hi-level substitutes available
xcore:	equ false
;- EXCEPTION-EXT	( -- fl ) 						 	environment
xcorex:	equ true
;-
;- FACILITY	( -- fl )							environment
ccore:	equ true;:ok:
;- FACIILITY-EXT ( -- fl )							environment
ccorex:	equ true;:ok:
;-
;- FILE 	( -- fl )       						environment
fcore:	equ true;:ok:
;- FILE-EXT	( -- fl )       						environment
fcorex:	equ true;:ok:
;-
;- FLOATING	( -- fl )							environment
flting:	equ true
;- FLOATING-EXT	( -- fl )							environment
fltingx:equ true
;-
;- LOCALS 	( -- fl )							environment
lcore:	equ true;:ok:
;- LOCALS-EXT	( -- fl )							environment
lcorex:	equ true;:ok:
;-
;- MEMORY_ALLOC	( -- fl )							environment
mcore:	equ true;:ok:
;- MEMORY_ALLOC-EXT ( -- fl )							environment
mcorex:	equ false;:ok:
;-
;- SEARCH-ORDER ( -- fl )							environment
score:	equ true;:ok:
;- SEARCH-ORDER-EXT ( -- fl )							environment
scorex:	equ true;:ok:
;-
;- STRING 	( -- fl )							environment
gcore:	equ true;:ok:
;-
;- TOOLS	( -- fl )							environment
tcore:	equ true
;- TOOLS-EXT	( -- fl )							environment
tcorex:	equ true

;-
;- MAX-N  	( -- n )							environment
;-	max. sing integer, MIN-N is -ve MAX-N (i.e. hex 80000000 not a valid integer!).
mxn:GA;:ok:
	spush
	mov eax,0x7fffffff
	ret
;-
;- MAX-D  	( -- d )							environment
;-	max. double integer, MIN-D is -ve MAX-D (hex 80...00 is not a valid double!).
mxd:GA;:ok:
	call mxud
	shr eax,1
	ret
;-
;- MAX-FLOAT  	( -- df )							environment
;-	a "float" is a pair of double cells, representing a rational (real) number
;-	by fraction of signed numerator (@nos) and unsigned denumerator (@tos).
;-	smallest representable number is 1/max-float, displayed w/ up to {#fplaces} 
;-	fractional digits. beyond |2^62-1| only integers representable.
mxflt:
	call mxd
	jmp d2dv
;-
;- MAX-CHAR  	( -- n )							environment
mxc:	equ scstg;:ok:
;-
;- MAX-U  	( -- n )							environment
mxu:	equ mone;:ok:
;-
;- MAX-UD  	( -- d )							environment
mxud:	equ dmone;:ok:
;-
;- /PAD    	( -- n )							environment
spad:GA;:ok:
	spush
	mov eax,[byte uref uv.padsz]
	ret
;-
;- /HOLD  	( -- n )							environment
;-	space for numeric to stg conversion, between {pad2} and {here}.
shold:GA;:ok:
	call pad2
	call here
	jmp ssub
;-
;- RETURNSTACK-CELLS	( -- n )						environment
;-	 query returns half size of data- and return-stack shared memory area.
rsc:GA;:ok:
	call rpfetch
	call spfetch
	call ssub
	shr eax,3
	ret
;-
;- #LOCALS  	( -- n )							environment
;-	cells count remaining available for allocation in local dataspace.
;-	n is taken as half of unused memory between data- and return-stack, where 
;-	dependent on both stacks actual useage more or less locals may be allocated.
;-	This limit is about 1/8 that value while compiling, due to additional space 
;-	being used for the temporary named word-headers. A Local requires 8 cells 
;-	plus the name's length, while at runtime occupying only a single cell.
nlocals:GA;:ok:
	spush
	mov eax,esp
	sub eax,PS
	shr eax,4
	ret
;-
;- environment 	( -- )				(x)(i)				hidden
;-	vocabulary for the ANS-4th environment query.
environment:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ienvironment
;
;-	
; =====================================================================;
; l4-envq.inc <eof>
