;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4interpret.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __INTERPRET_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __INTERPRET_I
%define __INTERPRET_I
%define interpretinc
;-
;-------------------------------------------------------------------------
;--- interpreter, forget, find
;-	

;-
;- marker 	ccc( -- )			(a:cx)(4)			forth
;-	defining: save stacked search-order, current, fence values, latest.
;-	execution: forget till marker, discard new, rst saved system state.
;-	NOTE:	{fence} values stored after the rsp marker was defined won't 
;-		protect against implied forget by the marker word.
;-		use -re- {empty} to {forget}ting everything after {fence}s.
marker:GA
;in@ blk@ bl word find -execute blk ! in !
;(lkindex) @ create here >r GET-ORDER ocurrent 4- 4@ 4 roll 4+ dup 1+ cells allot r> n!
;fence@ here >r 4 cells allot r> 4!		( index-, data-, code-fence, own ix )
;does> also bignum n@+ previous 4- -4 roll ocurrent 4- 4! SET-ORDER 4@ fence! iforget drop
	call infetch
	call blkfetch
	call blwf
	call dexecute
	mov ecx,TOS(1)
	mov [byte uref uv.blk],eax
	mov [byte uref uv.in],ecx
	spop
	mov eax,[byte uref uv.lkindex]
	call create
	call rrbrac
	push dword dataptr; here
	call gorder
	push eax	; count
	lea eax,[byte uref uv.ocurrent-4]
	sub eax,KR
	call fetch4
	spush
	pop eax		; count
	add eax,byte 4	; 4+
	spush
	lea eax,[eax*4+4]; 1+ cells
	call allot
	spush
	pop eax		; here
	call nstore
	call fencef
	spush
	mov eax,dataptr
	spush
	mov eax,16
	call allot
	call store4
	call llbrac
	call pscodep
domark:
	call nfetchp
	sub eax,4
	spush
	mov eax,-4
	call roll
	spush
	lea eax,[byte uref uv.ocurrent-4]
	sub eax,KR
	call store4
	call sorder
	call fetch4
	call fences
	call iforget
	spop
	ret
;-
;- forget  	ccc( -- )			(a:t)				forth
;-	remove words defined after and including word w. name {ccc} and 
;-	{allot}ted data from dictionary, i.e. code- and data-space, including 
;-	and running through all vocabularies (orderly defined, i.e. not just 
;-	"wordlists"). noop if word not found, fenced or in non-alterable memory.
;-	special cases error handling can be achieved w/ -re- {sforget}.
;-	see also -re- {empty}, {:forget}, {forget>}, {marker}.
forget:GA;:ok:
	call BLWORD
	call ccount
	spush
	mov eax,[byte uref uv.lkindex]
	call sforget
	cmp eax,-1
	jl near errorf	;?; error
	spop
	ret
;-
;- empty  	( -- )								forth
;-	forget all above {fence}d ptrs.
;-	NOTE: current state of program code & data can be saved w/ -re- {fence!}
empty:GA;:ok:
	lea ecx,[kref drop-Y]
	spush
	push ecx
	mov eax,[byte uref uv.ifence]; :continues: to iforget
;-
;- iforget	( ix -- flg )							hidden
;-	forget words defined after and including word by execution token ix
;-	in range (last in kernel) < @ifence <= ix < @(lkindex).
;-	if ix is a voc replace occurences in voc-stack and current with {local}-wid.
;-	header-less words can only be removed in code-space, no data-ref exists.
;-	flg: -1 success,    1 search-order changed, 0 not found
;-	     21 {fence}d,  23 kernel/non-writable memory.
iforget:GA;:ok:
	mov ecx,[byte uref uv.lkindex]
	spush
	push ecx,ecx,ecx
	jr sforget.i
;-
;- sforget	( p u xt -- flg )						hidden
;-	forget words defined after and including word w. name of string (p,u)
;-	in range (last in kernel) < @ifence < ix(p,u) < min(xt,@(lkindex)).
;-	if (p,u) is a voc replace occurrences in voc-stack with {local}-wid - which is
;-	short (i.e. fastly searched) and, neutral because forcedly searched 1st, anyway.
;-	{context} will be restored from {ocontext}, if possible, else both set to {forth}.
;-	flg: -1 success,    1 search-order changed, 0 not found
;-	     21 {fence}d,  23 kernel/non-writable memory.
;-	NOTE:	unsafe if local values exist in the chain of words to forget.
;-		which applies to locals assigned while interpreting or, those 
;-		which may be left at error abortion while defining a word.
sforget:GA;:ok:
;scan and invalidate uot entries until word to forget reached:
; 1) find defining voc, modify top lfa
; 1.1) execute concerning words from {fence-lk} linkage chain. (n.i.)
; 2) if word is voc check whether in search order and if so,
;	'invalidate' by replacing w/ 'local' or, 'forth' if {current}, {ocurrent}
;::->	invalidate by overwriting w/ remaining v-stack or, terminate if at it's end.
; 3) save xec entry ptr
; 4) store last xec entry to @cp, now beginning of free code-space
; 5) store last >body ptr to @dp, now beginning of free data-space
; 6) store words index to (lkindex), now being 1st un-used
	call qexec	; not while compiling
	cmp eax,[byte uref uv.lkindex]
	jbe .e
	TOSPULL(2)	; 3drop 0
	xor eax,eax
	ret
.e:A
	push eax	;	[ ix
	mov eax,TOS(1)
	add eax,byte 4
	shr eax,2
	push eax	; 	[ #locals ix
	push dword[byte uref uv.s0];[ p2 #l ix 
	call toloc	; p1 u1
	spush
	mov eax,[esp]	; p1 u1 p2
	call place	; local, counted copy of name to forget
	spush
	mov eax,[esp]	; p
	call find	; i +/- 1
	spush
	mov eax,[esp+4]
	call toloc.d	; rmv local mem
	test eax,eax
	jz .n		;?; not found
	mov eax,TOS(1)	;	{ i2 i2	[ f #l i1
.i:;	;<iforget)
	cmp eax,[esp+8]
	mov dword[esp],0; flag
	ja .x		;?; not in given range
	cmp eax,[byte uref uv.ifence]
	dec dword[esp]	; flag unchanged search order
	jc .f		;?; fenced
	cmp eax,[kref tophead.i-Y]
	jc .k		;?; kernel word
	call fgchain	; execute fall-back definitions
	test eax,eax
	jz .x		; 'alias'ed word
	call localwl	; neutral substitution..
	mov [esp+4],eax	;  ..for removed vocabularies
	spop
	mov [esp+8],eax
.l:
	mov eax,[byte uref uv.lkindex]; i3 i2
	dec eax
	cmp eax,[esp+8]
	jnc .c
	call idx2lfa
	mov ecx,codeptr
	mov [byte uref uv.last],eax
	mov [byte uref uv.bcp],ecx
	mov [byte uref uv.btail],ecx
	mov [byte uref uv.tail],ecx
.x:
	mov eax,[esp]	; flg
.n:
	TOSPULL(1)	; drop ptr/i2, leave flg
	lea esp,[esp+12]; 3r> 3drop
	ret
.f:;	;<fenced)
	add dword[esp],byte 22
	jr .x
.k:;	;<in kernel)
	add dword[esp],byte 24
	jr .x
.c:A
	TOSPULL(1)	;	{ i3
	mov ecx,eax	; save ix for headerless
	call idx2lfa
	spush
	jz .o		;?; headerless word
	call wvocr	; find voc
	call lfa2wid
	add eax,vhead.p-vhead.m; voc-top: ptr to disp to top lfa in voc
	mov edx,TOS(1)	; this lfa
	mov ecx,eax	; save ptr to ptr to voc-top
	sub edx,[kref edx]; prev. lfa
	sub eax,edx	; disp to ptr to voc-top
	mov [kref ecx],eax; @voc-top
	mov eax,TOS(1)
	call lfa2ffa	; flags
	test word[kref eax],vocf
	jnz .v		;?; vocabulary
.w:;	;<after voc. ref)
	mov eax,TOS(1)
	call lfa2ifa	; header
	mov dataptr,eax	; free data-space
	mov eax,TOS(1)
	call lfa2xec
.p:;	;<after headerless)
	mov codeptr,eax	; free code-space
	mov dword[kref eax],x_ret|x_ret<<8|x_ret<<16|x_ret<<24
	dec dword[byte uref uv.lkindex]
	jmp .l		; try next word
.o:A;	;headerless: no data-ref, remove code
	mov eax,ecx
	call idx2xec
	jr .p		; free codespace
.v:A; forget vocabulary/wordlist
; unlink from vocabulary linkage:
; { voc-link dup @ - dup @ - voc-link - negate voc-link ! }
	lea ecx,[byte uref uv.voclink]
	mov edx,ecx
	sub ecx,[ecx]
	sub ecx,[ecx]
	sub ecx,edx
	neg ecx
	mov [edx],ecx
; replace voc-stack occurrences w. local-wordlist
	mov eax,TOS(1)
	call lfa2wid
	mov ecx,[kref cnvocs-Y]	; vocstack size
	lea edx,[uref uv.context];vocstack top
	push ebx
	mov ebx,[esp+8]	; provide default wid - {local}
.m:
	cmp [edx],eax
	jnz .d		;?; other voc
	mov [edx],ebx	; replace w/ passed default voc
.d:
	dec ecx
	lea edx,[edx-8]
	jnz .m		;?; more stacked vocs
; set defining vocs to {forth}, if required or, to {prevdef} if {ocurrent} still useable
	pushad
	call forthwl
	writead ecx,eax
	popad
	lea edx,[uref uv.ocurrent]
	cmp [edx],eax	; test ocurrent
	jnz .u		;?; other wid
	mov [edx],ecx	; replace ocurrent w/ {forth}
.u:
	cmp [edx+8],eax	; current
	jnz .r		;?; other wid
	mov ecx,[edx]	; copy ocurrent to current
	mov [edx+8],ecx
.r:
	pop ebx
	jmp .w		; continue unlinking sequence
; check forget-handlers
fgchain:A; ( ix -- ix ) :ok:
	mov ecx,forgetlkv-Y
	call pdovar
	push eax	; save link base
	push eax
.l:
	mov eax,[esp]
	mov ecx,[kref eax]
	mov dword[kref eax],0; new end
	mov eax,[esp+4]	; {forget-lk}
	mov [kref eax],ecx ; unlink this item
	test ecx,ecx
	jz .r		;?; list ends
	mov [esp],ecx	; ptr to next..
	mov edx,[kref ecx+4]; fetch ix
	cmp edx,TOS(1)	;target ix
	jc .r		;?; not the requested handler
	push PS
	mov eax,edx
	spush
	mov eax,[kref ecx+8]
	call execute	; ( ix1 ix2 -- ix1 xx ) top item may be modified, ix1 must remain
	pop PS		; just in case..; and for the final <spop>
	jr .l
.r:A
	lea esp,[esp+8]	; drop linkptr & link base
	spop		; and workspace, leave ix!
	ret
;-
;- :forget	ccc( -- )			(ix)		'colforget'	forth
;-	initiates a header-less word, linked to the {forget-lk} chain, 
;-	which will be executed before {forget} removes the word ccc.
;-	the handling word receives ix of the word to {forget} @tos and @nos
;-	and, should not modify the @nos item nor drop any item from stack; 
;-	except, a word can be protected by returning @nos cell set to zero.
;-	additional data left on stack will be discarded, internally.
colforget:GA;:ok:
	call tick
	push eax		; target ix
	spop			; clr stack for :noname
	call noname		; enter compiling state	( -- ix2 ix2 sp@ )
;.n:; ;<ix by ix, for headerless word, e.g. <ibbuf>
	spush
	mov ecx,dataptr
	_mov eax,fgtlink_size	; allocate linkptr & ix storage
	push ecx		; save {here}
	call allot
	mov ecx,forgetlkv-Y	; fetch link
	call pdovar
	pop edx			; new ptr to linkage body
	mov ecx,[kref eax]	; old ptr
	mov [kref eax],edx	; store new
	pop eax			; word to forget
	mov [kref edx+fgtlink.lfa],ecx	; link to previous ptr
	mov [kref edx+fgtlink.ixw],eax	; sto ix
	mov eax,TOS(2)		; hnd
	mov [kref edx+fgtlink.ixh],eax	; sto hnd
; -
	mov ecx,TOS(1)
	mov TOS(2),ecx
	mov TOS(1),eax
	sub dword[uref uv.ucsp],byte 4
	spop 2			; drop 2nd ix
	sub eax,byte 4		; adjust ctrl ptr
	ret
;-
;- forget>	ccc( -- )			(i)		'forgetgt'	forth
;-	to be placed before {does>} and after {create} in a definig word, to 
;-	setting up a link to the chained word-specific forget handlers. 
;-	At execution the word being defined will be assigned the word ccc to, 
;-	which,in turn, will be executed immediatly before the newly compiled 
;-	word is to be removed by {forget}. -re- {:forget}, {sforget}.
;-	compiles a preliminary structure, 3 cells, the last of which contains 
;-	the execution token ix of the handler word. example:
;-		: defn ..words1.. create ..words2.. forget> noop does> ..words3.. ;
forgetgt:GA
	call qcomp
	call tick		; fetch hnd, ix2
	TICK fgl
	jmp compilec
;-
;- fgl		( ix -- )					'fgl'		novoc
;-	by {forget>} cpl'd runtime code, ix is handler word's xec-token.
;-	(asm level useage example -re- {ibbuf}, buffer-blocks allocation)
fgl:A
	call latest
	call lfa2idx		; current word, ix1
	call swap		;:continued:
;-
;- forget,	( ix1 ix2 -- )					'fgcomma'	forth
;-	allocate linkage field and link hnd ix2 for word ix1 into forget-lk chain.
;-	-re- {:forget}, {forget}, {sforget}.
fgcomma:GA;:ok:
	spush
	mov ecx,dataptr
	_mov eax,fgtlink_size	; allocate linkptr & ix storage
	push ecx		; save {here}
	call allot
	mov ecx,forgetlkv-Y	; fetch link
	call pdovar
	pop edx			; data-ref, new link
	mov ecx,[kref eax]	; old ptr
	mov [kref eax],edx	; store new
	mov [kref edx+fgtlink.lfa],ecx	; link to previous ptr
	mov eax,TOS(2)			; word to forget
	mov ecx,TOS(1)			; hnd
	mov [kref edx+fgtlink.ixw],eax	; sto ix
	mov [kref edx+fgtlink.ixh],ecx	; and hnd
	spop 3
	ret
;-	

;-
;- SEARCH-WORDLIST ( p1 u1 p2 -- p3 flg ) 	(a:s)		'swlist'	ans
;-	find name(p1,u1) in headers chain beginning at lfa p2
;-	this is ANS-4th version of (find) - re below
swlist:GA;:ok:
; o:	ecx	lfa
;	edx	@ffa
	mov ecx,eax	; save top-lfa
	mov edx,TOS(1)	; cnt
	spop		; drop
	push PS		; save PS w. 2 cells
	call placeps	; convert p,u to asciz
	dec eax		; eax: old PS
	mov [eax],dl	; prepend countbyte
	sub eax,KR	; abs>4th
	spush
	mov eax,ecx	; voc-lfa
	call pfindp	; (find)
	mov ecx,eax	; fetch result
	mov edx,TOS(1)
	pop PS		; drop local stg
	mov eax,ecx	; ret result
	test ecx,ecx
	mov ecx,edx	; save lfa for <value>, etc
	jz .n		;?; not found
	movzx ecx,word[kref edx+whead.i-whead.l]
	movzx edx,word[kref edx+whead.f-whead.l]; header-flags
.n:
	mov TOS(1),ecx
	ret
;-
;- (find)  	( p1 p2 -- p3 flg )		(e)		'pfindp'	hidden
;-	find name at p1 in headers chain beginning at lfa p2
;-	return lfa p3 and flg =/= 0, else strg p1 and 0;
;-	set macro-flag in {state} according to words capabilities.
;-	flg := 1 if word is immediately executing in either -re- {state}.
;-	default search mode letter case in-dependent, which can be forced to 
;-	case dependent w/ a header flag in the words' defn or, by {lc-depend}.
;-	upper case synonyma should conventionnaly be defined earlier than their 
;-	lower case counterparts, as in kernel, such that the lower case word 
;-	would take precedence before upper case defn, in case-independent mode.
;-	char and case conversion re-mappable, using tables (lo>hi) and (hi>lo).
;-	(e: eflags reflect flag value)
pfindp:GA;:ok:
; c: ecx,edx
;%define __exit_find;:preliminary: to terminating if (find) runs into loop
%undef  __exit_find	; emergency exit disabled (re below)
	mov ecx,mcaps
	and ecx,dword[byte uref uv.bits]
	mov edx,[uref uv.ulo2hi]	; lo>hi char conversion
	add edx,KR
	push dword[uref uv.uhi2lo]	; hi>lo
	add [esp],KR
	pushad
;; ;:EBX::preserved:;;for KRef, ALL OTHER REGS USED!;
	lea ebp,[kref eax]		; ebp := lfa link ptr
	spop
	lea edx,[kref eax]		; edx := string
.l:;outer loop: start comparing stg with nfa
%ifdef __exit_find; enabling emergency exit
	pushad
	mov edi,[esp s_edi s_tor]
	mov ebp,[esp s_ebp s_tor]
	call escq
	popad
	jnz .x
%endif
	xor eax,eax
	cmp byte[edx],al
	jz .x		;?; exit not-found if zero length string
	mov eax,[ebp-whead.l+whead.f]
	test eax,smud
	jnz .n		;?; word not ready
	lea edi,[ebp+whead.n-whead.l]; advance to nfa
	mov esi,edx	; cnt-b of string
	cmpsb
	jnz .n		;?; next, cnt not equal
	movzx ecx,byte[esi-1]
	jecxz .n	;?; =/=
.d:;inner loop: compare entire string
	dec ecx
	cmpsb
	jnz .s		;?; try case independently
	jecxz .f	;?; found
	jr .d
.s:A;try case independently, single chars
	cmp dword[byte esp s_ecx],mcaps
	jnz .n		;?; case dependency forced by user setup
;for case dependent words lower case takes precedence over upper case by sequence of defn's
	test word[ebp-whead.l+whead.f],wcas
	jnz .n		;?; by word defn forced case dependency
.i:;:ok:- w. re-definable character code tables
	movzx eax,byte[edi-1]
	mov ah,al
	xor al,[esi-1]
	jz .q		;?; found
	shr eax,8
	add eax,[byte esp s_edx]; lo>hi
	movzx eax,byte[eax]
	cmp al,[esi-1]
	jz .q		;?; ok
	add eax,[byte esp s_tor]; hi>lo
	mov al,[eax]
	cmp al,[esi-1]
	jnz .n		;?; not found
.q:
	jecxz .f	;?; found
	jr .d		;?; next char
.n:A; fetch next lfa
	mov eax,ebp
	sub ebp,[ebp]
	cmp eax,ebp
	jnz .l		;?; next word
.x:; word not found
	mov ebp,edx
	xor ecx,ecx
	dec ecx
.f:; word found, ecx = 0
	readad edi,edi
	sub ebp,KR	; 4th ptr
	mov TOS(1),ebp	; lfa , else stg
	mov [byte esp s_ecx],ebp;pass lfa
	not ecx		; -1:found, else 0
	mov eax,[kref ebp+whead.f-whead.l]
	xor edx,edx
	test eax,cmac
	jz .c
	or edx,mfmac
.c:
	test eax,ixec
	jz .e
	readad UR,ebp	; for uref in <ccpl?>
	call ccplq
	jz .g
.e:
	test eax,imed
	jz .m
.g:
	neg ecx		;+1 if immediate(ly executing) word
.m:; not immediate
	mov [byte esp s_edx],edx
	mov [byte esp s_eax],ecx
	popad
	and dword[byte uref uv.bits],~(mfmac)
	test dword[byte uref uv.bits],mcmac
	jz .o
	or [byte uref uv.bits],edx
.o:
	lea esp,[esp+4]
	test eax,eax
	ret
;-
;- blwf		( -- n1 f )			(k)				novoc
;-	does{ bl word find }, eflags set according to {find} flag result
blwf:A
	call BLWORD
;-
;- find		( p -- p 0 | ix 1 | ix -1 )	(e)(4)(a:c)			forth
;-	find uot index from string at p searching the stacked search order. 
;-	values returned by FIND while compiling or interpreting may differ.
;-	{local} searched 1st, unconditionally, {root} last if not in search 
;-	order. for wordlist specific search -re- {(find)}, {search-wordlist}.
;-	letter case dependency:
;-	dft {lc-ignore} and {uc-unmap} mode, true chars case independently,
;-	which can be modified by
;-	 {lc-depend} case dependent
;-	 {lc-ignore} case independent
;-	 {uc-map}    input mapping upper to lower chars
;-	 {uc-unmap}  input upper and lower chars
find:GA;:ok:
; o:	ecx	lfa
;	edx	@ffa
;	eflags	ret.-flag status
	call dfind.f	; (-find) dup
	mov ecx,TOS(1)	;  >r lfa>idx
	jz .r		; IF
	movzx edx,word[kref ecx+whead.f-whead.l]	
	movzx ecx,word[kref ecx+whead.i-whead.l]
	mov TOS(1),ecx	;  r>
	test edx,nopt
	jz .n
	mov dword[uref uv.ucmc],-1
.n:
	test eax,eax
.r:			; ENDIF
	ret		; ;s
;-
;- -find  	ccc( -- p flg)			(e)(4)(fg)	'dfind'		forth
;-	find word w. name string <ccc> in the stacked search order,
;-	if found ret lfa and flg := 1 if word is immedate, else -1.
;-	ret ptr p to counted string <ccc> and flag 0, if not found.
;-	(e: result flag status)
dfind:GA;:ok:
	call BLWORD		; bl word here  { s
.f:
	spush
	mov eax,pfindp-Y	; execution vector: (find)
;-
;- scanwlist	( n1 p1 -- n2 n3 )						hidden
;-	common routine to scanning vocabulary stack
;-	expects xec-ptr p1 of working routine w. stack flow
;-		( n1 toplfa -- n1 0 | n2 n3 )
;-	after {locals} wordlist, searching in stacked lists order,
;-	toplfa from vocstack and, if no success, lastly from {root},
;-	ret n2 as determined by routine p1, n3 flag according to result.
;-	scanning terminates early if supplied routine returns non-zero.
;-	last opr if no success is {perform}ing user-vari (scanwl).
scanwlist:GA;:ok:<by -find)
; c: ecx,edx
	call rootwl
	push eax		; dummy, zeroed if root in search order
	spop
	lea eax,[kref eax]
	push eax		; >r			{ x stg	[ xec
	lea eax,[uref uv.context]; vocstack top
	mov ecx,[kref cnvocs-Y]	; vocstack size
	lea ecx,[ecx+2]		; context/ocontext
	push ecx
	push ecx
	push eax		; vocstack top			[ vt vb xec
; search locals first, unconditionally
	mov eax,Ilocal		; local-ix
	call idx2wid		; local-wordlist
	call wid2top		; wid>top
	call [esp+12]		; (supplied routine)
	test eax,eax
	jnz .m			;?; name found locally
.l:;				; BEGIN			{ x s	[
	mov ecx,[esp]		;  r> r> r
	mov eax,[esp+8]		;  swap >r over >r
	test eax,eax		; cnt
	jz .r			;?; end of voc-stack
	mov edx,eax		;  rmng cnt
	sub eax,[esp+4]		;
	lea ecx,[ecx+eax*8]	;  ptr to wid in vocstack
	mov eax,[ecx]		;  @
	test eax,eax		;  dup			{ v v s	[ t
	jz .r			; WHILE			{ v s	[ t
	dec dword[esp+8]	;  2r> r> 1- >r	2>r	{ f s 	[ t b x
	bts dword[ecx-4],bsvoc
	jc .l			;?; already searched
	call markwlist
	cmp eax,[esp+16]
	jnz .w
	sub [esp+16],eax
.w:
	call wid2top		;  voc>top		{ l v s	[ t
; xec vector w. tos{ toplfa n1 }
	call [esp+12] 		;  (supplied routine)	{ f p 	[ t
	mov edx,ecx		;
	test eax,eax		; -dup UNTIL
	jz .l			;			{ p s	[ t b x
	inc dword[esp+4]	;			{ f p	[
	test eax,eax
	jnz .m			;?; success
.r:
; search root last, if name not found and not already done
	mov eax,[esp+16]	; root-wl or 0
	test eax,eax
	jz .s			;?; root already searched
	call wid2top		; top lfa
	call [esp+12]		; (supplied routine)
	test eax,eax		; result
	jnz .m			;?; ok
.s:
	spush			; final hnd
	mov eax,[uref uv.uscanwl]; (scanwl) @ 0> if ..
	call execute.c		; execute from +ve non-zero address ..endif
.m:
	call clrwlist
	lea esp,[esp+20]
	test eax,eax
	ret
;
; <scanwlist> subroutines
;
; ( wid -- wid ) mark multiply found wid-s
; i:	eax=voc, ecx=vtos, edx=rmng cnt
; c:	all regs preserved
markwlist:A
	push ecx
.l:
	cmp [ecx],eax
	jnz .n
	or dword[ecx-4],msvoc; mark voc item 'scanned'
.n:
	dec edx
	lea ecx,[ecx-8]
	jg .l
	pop ecx
	ret
; clear 'scanned' bits
clrwlist:A
	lea edx,[uref uv.context]	; vocstack top
	mov ecx,[kref cnvocs-Y]		; vocstack size
	lea ecx,[ecx+2]			; context/ocontext
.c:
	and dword[edx-4],~(msvoc)	;clear searched-flag
	dec ecx
	lea edx,[edx-8]
	jnz .c
	ret
;-
;- scanvocs	( n1 xt -- n2 )			(e)				hidden
;-	common routine to scanning vocabulary linkage, latest to 1st.
;-	expects xec token (uot index) xt of a working routine w. stack flow
;-		( n1 wid lfa lfa --      exit: nn xxx xxx  0	or
;-				 -- next word: nn wid lfa 0=/=	or
;-				 -- next voc.: nn xxx  0  0=/=	)
;-	scanning terminates if supplied routine returns zero or all vocs done.
;-	(eflags reflect state of n2)
scanvocs:GA;:ok:
	call idx2xec
	test eax,eax
	jz bscanvb.x
;-
;- [scanvocs]	( n1 p2 -- n2 )					'bscanvb'	hidden
;-	as -re- {scanvocs} but, w/ p2 execution ptr instead of ix.
bscanvb:GA;:ok:
	lea ecx,[byte uref uv.voclink]
	sub ecx,KR
	lea eax,[kref eax]
	push ecx
	push eax		; >r			{ x stg	[ xec vocl
.l:;outer loop, wordlists
	mov eax,[esp+4]
	mov ecx,[kref eax]
	test ecx,ecx
	jz .r
	sub eax,ecx
	jz .r			;?; end of linkage, else wid
	mov [esp+4],eax
	spush			; pass wid
	call wid2top		; and top nfa
.i:;inner loop, words
	spush
	call [esp]		; ( nn wid lfa lfa -- nx wid lfa flg )
	test eax,eax
	spop
	jz .q			;?; ready
	test eax,eax
	jz .w			;?; force next wl
	sub eax,[kref eax]	; next lfa
	cmp dword [kref eax],byte 0
	jnz .i			; next word
.w:
	spop
	jr .l			;?; next wordlist
.q:
	spop
.r:
	add esp,byte 8
.x:
	spop
	test eax,eax
	ret
;-
;- [debug]	( flg ix -- flg ix )				'bdebugb'	compiler
;-	executed before compiling any non-immediate words. initially a noop.
;-	if non-zero execute true address or xec token from uot vector ([debug]).
;-	then store current cp and ix @nos to uvari (prevcp) and (previx).
;-	debugging routine receives previous ix in ECX reg, current ix is @nos; 
;-	items on stack may be modified to any appropriate (valid!) value.
bdebugb:GA;:ok:
; i:	edx	headerflags, preserve! - re <pfax>
	test dword[uref uv.updebugp],-1
	jz .c
	push edx			; save header-flags
	spush
	mov ecx,[uref uv.uprevix]	; fetch previously cpl'd uot index
	mov eax,[uref uv.updebugp]	; previous c-here in uvari uv.uprevcp
	call perform.n
	pop edx
.c:
	mov ecx,codeptr
	mov [uref uv.uprevcp],ecx	; sto current cp
	mov ecx,eax
	mov [uref uv.uprevix],ecx	; sto this words index
	ret
;-
;- ]interpret	( p u -- p u )					'binter'	compiler
;-	execute if non-zero, true address or xec token in uot vector (]interpret).
;-	last action, called from <[interpret]> if neither, a word nor a number found.
binter:GA;:ok:
	mov ecx,[uref uv.ubinterpret]
	test ecx,ecx
	jnz .g
	ret
.g:A
	spush
	mov eax,ecx
	jmp perform.n
;-
;- interpret[	( ix f -- ix f )				'interb'	compiler
;-	execute if non-zero, true address or xec token in uot vector (interpret[).
;-	1st action, called from <[interpret]> immediately after {find}.
interb:GA;:ok:
	mov ecx,[uref uv.uinterpretb]
	test ecx,ecx
	jnz .g
	ret
.g:A
	spush
	mov eax,ecx
	jmp perform.n
;-
;- interpret					(v)				forth
;-	execute true address uot vector (interpret), default is <[interpret]>.
interpret:GA;:ok:
	jmp [uref uv.uinterpret]
;-
;- [']		ccc(C: -- )( -- ix )		(i)		'tick'		ans
;-	compile execution index of word ccc as a literal
;- '		ccc(C: -- )(X: -- ix )		(i)		'tick'		forth
;-	find execution token of word ccc and return a literal, else abort
tick:GA;:ok:
	call itick
	call ccplq
	jnz near literal
.r:
	ret
    global itick:function
;-
;- '		ccc( -- ix )			(a:c)		'atick'		ans
;-	find execution token of word ccc, else abort
;-	ANS-4th version, non-immediate.
atick:GA
itick:; tick for immediate words, w/o literal compiling
	call blwf	; +1:imed : -1:std
	spop
	jnz tick.r
p.ni:
	inc byte[kref eax]
	call print
r.ni:
	spush
	jr osi.j
; binterpretb, errors
osi:A
	mov ecx,codeptr
	mov eax,dataptr
	mov dword[kref ecx],x_ret|x_ret<<8|x_ret<<16|x_ret<<24; prevent repeated execution
	call binter		; ]interpret
	inc byte[kref eax]
	call space
	call print
	spop 2			; drop dnum
.j:
	mov eax,ENOSYS		; '...not implemented'
	jmp errorf
;
.q:A; rst interpret vector from recovered segfault
	call getrefb
	lea ecx,[kref bquitb-Y]
	mov [uref uv.uquit],ecx
	ret
;
.r:A;<binterpretb,binterpretb.iw)
	spop
;-
;- [interpret]	( -- )				(k)		'binterpretb'	hidden
;-	default 'outer interpreter' loop ('inner interpreter' is the cpu).
;-	reads word at ptr a1, pre-compiles exec token (uot index) or literal.
;-	executes -re- {interpret[} immediately after initial {find},
;-	and -re- {]interpret} if the token was neither a word nor a number.
binterpretb:GA;:ok:<.p)
	mov ecx,[byte uref uv.in]
	mov [byte uref uv.in0],ecx
	cmp ecx,[byte uref uv.span]
	jnc osi.q		;?; ready
	call BLWORD		; bl word here
%ifdef __DBG
%define __DBG__
%endif
%undef __DBG
;;-.;; obsolete after {word} bugfix v0.0.58) -?-
	cmp byte[kref eax],0
	jz osi.r		;?; don't find zero length string, retry
;;-';;
	call find		; find		{ flg ix } header-flags in edx
	call interb		; interpret[
	mov ecx,eax		; swap over
	mov eax,TOS(1)
;;	test ecx,ecx;
	mov TOS(1),ecx
	test ecx,ecx;;;
	jg .m	; imed		; 0> IF...
	js near .w; 4th		; 0< 0= IF...	{ ix flg
	inc eax			; 1+
	call bnumberb		; [number]	{ flg dn|qn|qv flg
	jnz disp osi		;?; neither word nor number
;    jnz near osi			;?; neither word nor number
	mov ecx,[byte uref uv.dpl]
	inc dword[uref uv.ucmc]	;:: literals don't add to (cmc) ::
	cmp ecx,byte -1
	spop			; drop flag
	jl .f			;?;counted int
	jnz .d			;?; not a sing
;:SINGLE: integer
	spop			; drop m.s. cell
	call c_lit		; cpl the literal, noop if interpreting
	jmp .c			; leave in code-space, or execute
.f:;:BIGNUM:; counted int
	call c_nlit	; variable no. of cells: counted int literal
	jr .c
.m:;:IMMEDIATE: 4th word
	mov ecx,[byte uref uv.state]
	test ecx,mimed
	jnz .iw			;?; delayed by "[_]"/"[compile]"/"postpone"
	TOSPULL(1)		; drop index & flag
	call execute.d		; execute imed word
	cmp word[byte uref uv.state],0
	mov ecx,[byte uref uv.tail]
	jz .p			;?; ready
	mov ecx,codeptr
.p:
	mov [byte uref uv.cp],ecx
	jmp disp binterpretb	; re-enter the interpreter
.q:;:QUAD: integer or double ranum
	call c_qlit	; 4 cells: quad int, double ranum
	jr .c
.v:;:QUAD: ranum
	call c_qvlit	; 8 cells, quad ranum
	jr .c
.d:;:DOUBLE:+:QUAD: -> quad: @dpl=0x0080xxxx, int $00808xx, ranum $008088xx
	cmp ch,0x80
	jz .q			;?; quad integer
	cmp ch,0x08
	jz .q			;?; double ranum
	cmp ch,0x88
	jz .v			;?; quad ranum
	call c_dlit		; double int or sing ranum
	jr .c
.w:;:FORTH:			;<;		{ ix flg
	call bdebugb		; [debug]
.iw:;<:immediate:
	call compilec.c		; cpl 4th word
	inc dword[uref uv.ucmc]	; words counter
	mov eax,[byte uref uv.state]
	btr eax,bimed		; delayed execution
	mov [byte uref uv.state],eax
	jc .c			;?;postponed
	test eax,mbrac
	jnz .c			;?; tmp interpreting state
	test eax,mword		; cpl
	jnz disp osi.r		;?; persistent code
.c:;<;common entry after compiling calls)
; decide whether to just compiled code immediately executing or 
; leaving in place: leave if state > 1, else execute if not postponed
	mov edx,codeptr
	mov ecx,[byte uref uv.tail]
	mov eax,[byte uref uv.state]
	test ax,ax
	js quit.h		;?; error
	jz .i			;?; interpreting
; {state} is "on", somehow:
	mov ecx,[byte uref uv.bcp]
	test eax,mbrac		; tmp interpreting state
	jnz .i
	mov ecx,edx
.i:;:INTERPRET: i.e. execute & remove the compiled code
	spop
	mov byte[kref edx],x_ret
	mov [byte uref uv.cp],ecx
	mov [byte uref uv.bcp],ecx
	add ecx,KR
	dec dword[uref uv.ucmc]	; words counter
	call ecx
.k:
	mov ecx,[byte uref uv.bcp]; remove latest code
	jmp disp .p
; -- ;
quit.r:A
	mov dword[byte uref uv.iolink],0
	call rpstore
	call fspstore
	jr quit.q
quit.t:A; resolve nested input levels
	movzx eax,byte[byte uref uv.sourceid]
	and eax,cdt_num-1
	jz quit.q
	call closef		; source-id -(maxchan) and close-file
	test eax,eax
	js quit.r		;?; 'emergency case'
	call popio		; pop-io drop
	spop
quit.s:A;<[error],?stack,abort),pushed to rp
	call getrefb
	cmp dword[byte uref uv.iolink],byte 0
	spush
	jnz quit.t		;?; resolve input levels
quit.q:
	mov ecx,[byte uref uv.span]
	mov dword[byte uref uv.sourceid],0
	mov [byte uref uv.in],ecx
quit.h:;	; <osi)
	call hptr		; interpreting state
	mov eax,[byte uref uv.base]
	lea eax,[eax-2]		; 1 -> ff..ff
	cmp eax,0xff
	jc .o
	call decimal
.o:
	spop
;-
;- \\		ccc( -- )					'quit'		forth
;-	comment till eof, exits loadfile (to next lower nesting level)
;-
;- quit		( .xx. -- .xx. ) (R: .xx. -- ) (C: .xx. -- ) (v)		root
;-	the outer interpreter, dft xec {[quit]}.
;-	
quit:GA
	cmp dword[uref uv.uquit],0
	jz bquitb
	jmp [uref uv.uquit]
;-
;- [squit]	( .xx. -- )(R: .xx. -- )			'bsquitb'	hidden
;-	entry from abortion by SEGV signal handler
bsquitb:GA; rst quit vector for terminating after (fatal) signals
	call getrefb		;>:KREF:
	lea ecx,[kref bquitb-Y]
	mov [uref uv.uquit],ecx
;-
;- [quit] 	( .xx. -- .xx. ) (R: .xx. -- ) (C: .xx. -- )	'bquitb'	hidden
;-	the outer interpreter (dft to {quit}):
;-	setup kernel ref, enable interrupts, 
;-	store zero to {blk}, {id} and {state}, 
;-	store {btail} to {cp}, clear return-stack, 
;-	enter interpreting loop, w. data-stack bounds check.
bquitb:GA;:ok:
	call getrefb		;>:KREF:
	and dword[byte uref uv.state],~(mword)
%ifdef __lockthread; if defined :threading:::locked: via iopl 0, cli/sti
	call mtunlock		; get safe, in case of locked interrupts always try unlocking 
%endif
.e:
	mov ecx,[byte uref uv.sourceid]
	mov dword[byte uref uv.blk],0
	mov dword[byte uref uv.sourceid],0
	test ecx,ecx
	jng near .z		;?; stdin or eval
	and ecx,byte cdt_num-1	; < max-chan
	spush
	movzx eax,cl		; source-chan, w/o flags
	push eax
	call chrnum
	test eax,eax
	pop eax
	jg .f			;?; not at eof
.o:
	call closef
.f:
	spop
	call popio
	jnz .c			;?; NZ: by popio all levels resolved
;;	cmp [byte uref uv.blk],byte 0;;;
;;	jg .l			;?; blockfile
.z:
	call hptr
	call rpstore
	lea ecx,[kref bquitb.l-Y]
	push ecx
.c:;	;<included)
	call siocons
.l:; ;----------<---------------.
	call qstack
	mov ecx,[uref uv.span]
	cmp ecx,[byte uref uv.in]
	jg .i			;?; more to interpret
	call scrcc
	or ecx,byte cdt_num
	cmp ecx,[byte uref uv.sourceid]
	jz .e			;?; input from screenfile
	call teststdin
	jz .q			;?; input not from console
	call bokb
	spush
	mov eax,[uref uv.upokp]	; echo prompt
	call execute.c
.q:; ;-----<---'
	call query
	cmp dword[byte uref uv.ierr],0
	jng near .e;;jz near .e		;?; eof: close file, pop io state
	call teststdin
	jz .i			;?; input not from console
	call cr
.i:; ;----<------'
	call source
	spop			; drop tib-max
	mov [byte uref uv.tib],eax
	spop
	call interpret
	jmp .l	;--->-----------'
%ifdef __DBG__
%define __DBG
%endif
%undef __DBG__
;-
;- [evaluate]	( p +u -- ..xx.. )				'bevalb'	hidden
;-	execute command or fetch number or send error.
;-	may be nested to any no. of levels, limited by available memory.
;-	saves/restores {blk},{in},{span},{stdin},{tib} on return-stack
;-	using p and +u, temporarily stored to locally assigned {tib} and {span}.
;-	initial vector for -re- {evaluate}.
bevalb:GA;:ok:
	call ipushio
	mov [byte uref uv.span],eax
	mov [uref uv.uintib],eax
	spop
	mov [byte uref uv.tib],eax
	spop
	mov dword[byte uref uv.in],0
	mov dword[byte uref uv.sourceid],-1
	call interpret
	call popio	; no <jmp>, data on RS!
	nop
	ret
;-
;- evaluate	( p +u -- ..xx.. )		(a:c)(v)			forth
;-	interpret string at p of length +u; dft -re- {[eval]}
;-	save/restore {source-id}, execute true address uot vector (eval).
evaluate:GA
	jmp [uref uv.ueval]
;-
;- -execute	( p flg -- )					'dexecute'	forth
;-	-re- {execute} p if flg =/= 0, else drop p and flg
dexecute:GA
	test eax,eax
	spop
	jnz execute
	spop
	ret
;-
;- perform	( p -- ..xx.. )							forth
;-	execute by ptr to uot index, drop & do nothing if p = 0 or @p = 0.
;-	ptr p is the ptr to index field (ifa) of a word header or any other 
;-	variable the l.s.w of which contains an execution token (uot index).
;-	example: { ' name sp@ perform }.
;-	NOTE:	because lib4th vectors may be de-activated by ptr to 0 
;-		{perform} cannot execute a call to {root} by uot index.
perform:GA;:ok:
	test eax,eax
	jz execute.r
	movzx eax,word[kref eax]
.n:
	test eax,eax
	jz execute.r
;-
;- execute	( ix | a -- ..xx.. )						forth
;-	execute 4th word by +ve uot index or, by true address a.
;-	noop if zero or kref'd zero or, if address not in regular program memory.
;-	{execute} can be used by index or by true execution address:
;-		{ -find name IF lfa>idx execute ENDIF }
;-	or,	{ c" name" find IF execute ENDIF }
;-	or,	{ (ok) @ execute }.
;-	or,	by ptr from <pfa> in word header if uot entry = 0.
;-	asm: <eax> if = ix converted to address, no (other) regs changed.
execute:GA;:ok:
	cmp eax,%$origin;0x08048000
	ja .c		;?; executeable addr
	cmp eax,32767
	ja .r		;?; noop if neither index nor address
	cmp dword[uref uv.ultab+4*eax],0;:ald: exits w/ code 139 -> -117 'struct needs cleaning' - silly?
	jz .p		;?; xec from pfa if zero entry @uot
	mov eax,[uref uv.ultab+4*eax]
.c:;	;<xec true addr or, entry from pfa)
	test eax,eax
	jz .r		;?; noop if <zero>
	cmp eax,KR
	jz .r		;?; noop if kref<zero>
	cmp eax,[byte uref uv.sstop]; no silly 'tricks' in stacks, tib, uot...
	jnc near rabort
	push eax
.r:
	spop
	ret
.d:A 1
	test dx,pfax
	jz execute	;?; standard xec
.p:; xec from pfa if <pfax> header-flag set
	call idx2pfa
	lea ecx,[eax+4]
	mov eax,[kref eax]
	add eax,KR
	jr .c
;-
; ---
;-	

;-
;- <tudisp>	( n -- n )			(e)(k)
;-	eflags: disp into uot out of bounds if NZ
;-	for {see}, {dcm}, etc.
tudisp:A;:ok:
	test eax,eax
	jg .r
	cmp eax,((((x_nop<<8|x_nop)<<8)|x_nop)<<8)|x_nop
	jz .v
; 0 <= eax-(uv.ultab-(uv.link)) < idx*4
	push ecx
	mov ecx,eax
	sub ecx,(uv.ultab-(uv.link))
	sar ecx,2
	cmp ecx,[byte uref uv.lkindex]
	pop ecx
	jnc .v
	test al,3
	jz .r
.v:
	test eax,eax
.r:
	ret
;-
;-================================================================
;--- conditional interpretation, comment, shell execution
;-

;-	

;-
;- defined	ccc( -- f )			(w3)				forth
;- [DEFINED]	ccc( -- f )			(i)				ans
;-	tf if wordd ccc found in current search order.
defined:GA;:ok:
	call dfind
	TOSPULL(1)
	ret
;-
;- o(		ccc( -- )			(i)		'ocommentp'	root
;-	dependent on type of output device print or silently discard a comment.
;if-nfound o( : o( stdout cons? if [_] .( else [_] ( endif ;     immediate
ocommentp:GA
	call stdoutc
	call chconsq
	test eax,eax
	jz paren.n
	jr commentp.n
;-
;- (		ccc( -- )			(i)		'paren'		root
;-	discard multiple lines spanning comment till matching ")"
paren:GA;:ok:
	spush
	jr .n
.f:
	call refill
.d:
	jz g16 iftrue.n	;?; drop ;
.n:
	mov eax,0+')'
	call parse
	add eax,TOS(1)
	cmp byte[kref eax],')'
	spop 2
	jnz .f
	ret
;-
;- .(		ccc( -- )			(i)		'commentp'	root
;-	display multiple lines spanning comment till matching ")"
commentp:GA;:ok:
; enter begin 1+ type refill 0= -exit entry parse 2dup + c@ [char] ) = until type
	spush
	jr .n
.f:
	inc eax
	call type	; display
	call refill
	jz iftrue.n	;?; drop ;
.n:
	_mov eax,0+')'
	call parse
	mov ecx,TOS(1)
	add ecx,eax
	cmp byte[kref ecx],')'
	jnz .f
	jmp type	; last text
;-
;- .\		ccc( -- )			(i)		'dotbslash'	root
;-	display comment till e.o.l.
dotbslash:GA;:ok:
	spush
	_mov eax,0
	call parse
	jmp type
;-
;- if-nfound	ccc( -- )			(i)(f6)		'ifnfound'	root
;-	conditionally include text until <eol> to interpretation
ifnfound:GA;:ok:
	call defined
	jz iftrue.n
	jr bslash.n
;-
;- if-found	ccc( -- )			(i)(f6)		'iffound'	root
;-	conditionally include text until <eol> to interpretation
iffound:GA;:ok:
	call defined
;-
;- if-true	( flg -- )			(i)(f6)		'iftrue'	root
;-	conditionally include text until <eol> to interpretation
iftrue:GA;:ok:
	test eax,eax
	jz bslash.n
.n:
	spop
	ret
;-
;- \\		ccc( -- )					'quit'		root
;-	comment till eof, exits loadfile (to next lower nesting level)
.x:

;-
;- #!		( -- )						'hb'		root
;-	enable shell script execution. requires {ss} at end of top line:
;-		#! /forth-executeable ..optional args.. ss
;-	note that {#!} requires at least one trailing space (blank, tab).
;-	{#!} introduces a comment, as -re- {\}, such that an executeable 
;-	script alternatively can be {included} as a hi-level source file. 
hb:GA

;-
;- \		ccc( -- )			(i)		'bslash'	root
;-	comment till e.o.l.
bslash:GA;:ok:
	spush
.n:;	;<if{{n,}found,true})
	_mov eax,0
	call parse
	jmp drop2
;-
;- [IF]		( flg -- )			(i)(4)		'bifb'		root
;-	immediate, non-compiling decision construct, for conditional cpl, etc.
;-	([if]/[else]... construct transposed from -re- dpans94 document)
bifb:GA;:ok:
	test eax,eax	;   0= IF POSTPONE [ELSE] THEN ;  IMMEDIATE
	jnz iftrue.n
	jr belseb.n
;-
;- [ELSE] 	( -- )				(i)(4)		'belseb'	root
;-	enter execution alternative after false flag to [IF].
belseb:GA;:ok:
	spush		; : [ELSE]  ( -- )
.n:
	_mov eax,1	;1;    1 BEGIN				\ level
.b1:
.b2:			;2;      BEGIN
	call BLWORD	;          BL WORD COUNT DUP		\ level adr len
	inc eax
	spush
	movzx eax,byte[kref eax-1]
	test eax,eax
	jz near .r2	;2;      WHILE
	call dup2	;         2DUP S" [IF]" COMPARE 0=
	call psqtep
	db 4,'[IF]'
	call compare
	test eax,eax
	spop
	jnz .e1		;1;       IF				\ level adr len
	spop 2		;          2DROP 1+			\ level'
	inc eax
	jr .n1
.e1:;			;1;       ELSE				\ level adr len
	call dup2	;          2DUP  S" [ELSE]"
	call psqtep
	db 6,'[ELSE]'
	call compare	;          COMPARE 0=			\ level adr len
	test eax,eax
	spop
	jnz .e2		;2;        IF
	spop 2		;           2DROP 1- DUP IF 1+ THEN	\ level'
	dec eax
	jz .r;;jz .n2;
	inc eax
	jnz .b2;;jr .n2;
	jr .r;;;
.e2:;			;2;        ELSE				\ level adr len
;	push dword TOS(1);;;
;	push eax	;;;
	call psqtep	;           S" [THEN]" COMPARE 0=	\ level
	db 6,'[THEN]'
	call compare
	test eax,eax
;	pop ecx		;;;
;	pop edx		;;;
;	jz .n31		;;;
;	mov eax,edx	;;;
;	spush		;;;
;	mov eax,ecx	;;;
;	call psqtep	;;;           S" [ENDIF]" COMPARE 0=	\ level
;	db 7,'[ENDIF]'	;;;
;	call compare	;;;
;	test eax,eax	;;;
;.n31:
	spop		;;;
	jnz .n3
	dec eax		;3;         IF 1- THEN			\ level'
.n3:
.n2:;			;2;        THEN
.n1:;			;1;       THEN ?DUP 0= IF EXIT THEN	\ level'
;;	test eax,eax;
	jnz near .b2;;jz .r
	jr .r;;jmp .b2		;2;      REPEAT
.r2:
	spop 2		;        2DROP				\ level
	call refill	;        REFILL 0=			\ level
	spop
	jnz near .b1	;     UNTIL DROP ;  IMMEDIATE
.r:
	spop
	ret
;-
;- [THEN] 	( -- )				(i)		'noop'		root
;-	terminate [IF] construct.
bthenb:GA {equ noop}; : [THEN]  ( -- )  ;  IMMEDIATE

;-
%endif;__INTERPRET_I
;-
;-=====================================================================;
;- l4interpret.inc <eof>
