;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4memfsto.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __MEMFSTO_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __MEMFSTO_I
%define __MEMFSTO_I
%define memfstoinc

;-
;-================================================================
;---  memory access
;-	false data may result if F8 stacks accessed directly!
;-	explicit addresses shall be passed as ptr's wrt kref.
;-	fetched data items not sign extended, unless mentioned.
;-

;-
;- c@+		( a1 -- a2 u )			(m,7)		'cfetchp'	forth
;-	fetch byte pointed to by @tos and increment a1 by byte size
;-	(this is traditional {count} implementation)
cfetchp:GA M;:ok:
	inc eax
	spush
	movzx eax,byte[kref eax-1]
M cfetchp
	ret
;-
;- c@		( ptr -- u )			(m,4)		'cfetch'	forth
;-	fetch byte pointed to by @tos
cfetch:GA M;:ok:
	movzx eax,byte[kref eax]
M cfetch
	ret
;-
;- w@+		( a1 -- a2 u )			(m,10)		'wfetchp'	forth
;-	fetch half cell pointed to by @tos and increment a1 by its size
wfetchp:GA M;:ok:
	movzx ecx,word[kref eax]
	lea eax,[eax+2]
	spush
	mov eax,ecx
M wfetchp
	ret
;-
;- w@		( ptr -- u )			(m,4)		'wfetch'	forth
;-	fetch half cell item pointed to by @tos
wfetch:GA M;:ok:
	movzx eax,word[kref eax]
M wfetchp
	ret
;-
;- @+		( a1 -- a2 u )			(m,8)		'fetchp'	forth
;-	fetch data item pointed to by @tos and increment a1 by cell size
fetchp:GA M;:ok:
	lea eax,[eax+4]
	spush
	mov eax,[kref eax-4]
M fetchp
	ret
;-
;- @		( ptr -- u )			(a:c)(m,3)	'fetch'		forth
;-	fetch data item pointed to by @tos
fetch:GA M;:ok:
	mov eax,[kref eax]
M fetch
	ret
;-
;- 2@		( ptr -- ud )			(m,10)		'twofetch'	fig
;-	fetch double pointed to by ptr.
;- 2@		( ptr -- ud )			(a:d)(m,10)	'twofetch'	forth
;-	fetch double pointed to by ptr.
;-	-re- 2@ in {ans} voc, for the ANS-4th big-endian cells order words.
twofetch:GA M;:ok:
; c: ecx
	mov ecx,[kref eax+4]
	mov eax,[kref eax]
	spush
	mov eax,ecx
M twofetch
	ret
;-
;- 2@		( ptr -- ud )			(a:d)(m,10)	'twofetcha'	ans
;-	fetch double pointed to by ptr.
;-	NOTE: re DPANS94 .6.1.0350, inconsistent, BIG ENDIAN cells order!
twofetcha:GA M;:ok:
; c: ecx
	mov ecx,[kref eax]
	mov eax,[kref eax+4]
	spush
	mov eax,ecx
M twofetcha
	ret
;-
;- a@		( a -- n )			(m,2)		'afetch'	forth
;-	fetch cell from true address
afetch:GA M;:ok:
	mov eax,[eax]
M afetch
	ret
;-
;- a!		( n a -- )			(m,10)		'astore'	forth
;-	store cell to true address
astore:GA M;:ok:
	mov ecx,TOS(1)
	mov [eax],ecx
	spop 2
M astore
	ret
;-
;- c!		( uc ptr -- )			(a:c)(m,11)	'cstore'	forth
;-	store byte @nos to address pointed to by @tos
cstore:GA M;:ok:
; c: ecx
	mov cl,TOS(1)
	mov [kref eax],cl
	spop 2
M cstore
	ret
;-
;- w!		( uc ptr -- )			(m,13)		'wstore'	forth
;-	store l.s.half cell @nos to address pointed to by @tos
wstore:GA M
; c: ecx;:ok:
	mov cx,TOS(1)
	mov [kref eax],cx
	spop 2
M wstore
	ret
;-
;- !		( u ptr -- )			(a:c)(m,11)	'store'		forth
;-	store item @nos to address pointed to by @tos
store:GA M;:ok:
; c: ecx
	mov ecx,TOS(1)
	mov [kref eax],ecx
	spop 2
M store
	ret
;-
;- 2!		( d ptr -- )			(a:d)(m,18)	'twosto'	fig
;-	store double @nos to address pointed to by @tos.
;- 2!		( d ptr -- )			(a:d)(m,18)	'twosto'	forth
;-	store double @nos to address pointed to by @tos.
;-	-re- 2! in {ans} voc, for the ANS-4th big-endian cells order words.
twosto:GA M;:ok:
; c: ecx
	mov ecx,TOS(2)
	mov edx,TOS(1)
	mov [kref eax],ecx	; l.s. cell to l.s. address
	mov [kref eax+4],edx
	spop 3
M twostore
	ret
;-
;- 2!		( d ptr -- )			(a:d)(m,18)	'twostoa'	ans
;-	store double @nos to address pointed to by @tos, 
;-	NOTE: re DPANS94 .6.1.0310, inconsistent, BIG ENDIAN cells order!
twostoa:GA M;:ok:
; c: ecx
	mov ecx,TOS(2)
	mov edx,TOS(1)
	mov [kref eax],edx	; m.s. cell to l.s. address
	mov [kref eax+4],ecx
	spop 3
M twostoa
	ret
;-
;- c!-		( x p1 -- p2 )			(m,10)		'cstorem'	forth
;-	pre-decrement p1 by one and store byte x, leave new ptr p2.
cstorem:GA M;:ok:
; c: -/-
	dec eax
	mov ecx,TOS(1)
	mov [kref eax],cl
	TOSPULL(1)
M cstorem
	ret
;-
;- w!-		( x p1 -- p2 )			(m,13)		'wstorem'	forth
;-	pre-decrement p1 by half size of a cell and store l.s. half cell x, leave new ptr p2.
wstorem:GA M;:ok:
; c: -/-
	lea eax,[eax-2]
	mov ecx,TOS(1)
	mov [kref eax],cx
	TOSPULL(1)
M wstorem
	ret
;-
;- !-		( x p1 -- p2 )			(m,12)		'storem'	forth
;-	pre-decrement p1 by cell size and store x, leave new ptr p2.
storem:GA M;:ok:
	lea eax,[eax-4]
	mov ecx,TOS(1)
	mov [kref eax],ecx
	TOSPULL(1)
M storem
	ret
;-
;- c!+		( x p1 -- p2 )			(m,10)		'cstorep'	forth
;-	store byte x, post-increment p1 by one, leave new ptr p2.
cstorep:GA M;:ok:
	mov ecx,TOS(1)
	mov [kref eax],cl
	TOSPULL(1)
	inc eax
M cstorep
	ret
;-
;- w!+		( x p1 -- p2 )			(m,13)		'wstorep'	forth
;-	store l.s. half cell x, post-increment p1 by half size of a cell, leave new ptr p2.
wstorep:GA M;:ok:
	mov ecx,TOS(1)
	mov [kref eax],cx
	TOSPULL(1)
	lea eax,[eax+2]
M wstorep
	ret
;-
;- !+		( x p1 -- p2 )			(m,12)		'storep'	forth
;-	 store x, post-increment p1 by cell size, leave new ptr p2.
storep:GA M;:ok:
	mov ecx,TOS(1)
	mov [kref eax],ecx
	TOSPULL(1)
	lea eax,[eax+4]
M storep
	ret
;-
;- c@!		( p1 p2 -- )					'cfsto'		forth
;-	copy byte (char) at ptr p1 to ptr p2
cfsto:GA;:ok:
	mov ecx,TOS(1)
	mov cl,[kref ecx]
	mov [kref eax],cl
	spop 2
	ret
;-
;- w@!		( p1 p2 -- )					'wfsto'		forth
;-	copy 16bit word (l.s.half cell) at ptr p1 to ptr p2
wfsto:GA;:ok:
	mov ecx,TOS(1)
	mov cx,[kref ecx]
	mov [kref eax],cx
	spop 2
	ret
;-
;- @!		( p1 p2 -- )					'fsto'		forth
;-	copy cell at ptr p1 to ptr p2
fsto:GA;:ok:
	mov ecx,TOS(1)
	mov ecx,[kref ecx]
	mov [kref eax],ecx
	spop 2
	ret
;-
;- 2@!		( p1 p2 -- )			(x)		'twofsto'	forth
;-	copy double cell from ptr p1 to ptr p2
;-	( mmx mode w/o timing gain )
twofsto:GA;:ok:
%ifdef __MMX;:ok: 0*
	mov ecx,TOS(1)
	movq mm0,[kref ecx]
	movq [kref eax],mm0
	spop 2
	ret
%endif
;-
;- 2@!		( p1 p2 -- )					'twofstof'	float
;-	copy double cell from ptr p1 to ptr p2
twofstof:GA;:ok:
	mov ecx,TOS(1)
	mov edx,[kref ecx]
	mov ecx,[kref ecx+4]
	mov [kref eax],edx
	mov [kref eax+4],ecx
	spop 2
	ret
;-
;- c!!		( p1 p2 -- )					'cxsto'		forth
;-	exchange contents of byte at ptr p1 with byte at ptr p2
cxsto:GA;:ok:
	lea ecx,[kref eax]
	spop
	mov dl,byte[ecx]
	mov dh,[kref eax]
	mov [ecx],dh
	mov [kref eax],dl
	spop
	ret
;-
;- w!!		( p1 p2 -- )					'wxsto'		forth
;-	exchange contents of word at ptr p1 with word at ptr p2
wxsto:GA;:ok:
	lea ecx,[kref eax]
	spop
	mov dx,[ecx]
	xchg dx,[kref eax]
	mov [ecx],dx
	spop
	ret
;-
;- !!		( p1 p2 -- )					'xsto'		forth
;-	exchange contents of cell at p1 with cell at p2
xsto:GA;:ok:
	lea ecx,[kref eax]
	spop
	mov edx,[ecx]
	xchg [kref eax],edx
	mov [ecx],edx
	spop
	ret
;-
;- 4!!		( p1 p2 -- )			(x)		'qxsto'		forth
;-	exchange contents of quad at p1 with quad at p2
qxsto:GA;:ok:
%ifdef __MMX; :ok: 5*
	mov ecx,TOS(1)
	movq mm0,[kref eax]
	movq mm1,[kref eax+8]
	movq mm2,[kref ecx]
	movq mm3,[kref ecx+8]
	movq [kref ecx],mm0
	movq [kref ecx+8],mm1
	movq [kref eax],mm2
	movq [kref eax+8],mm3
	spop 2
	ret
%endif
;-
;- 4!!		( p1 p2 -- )					'qxstof'	float
;-	exchange contents of quad at p1 with quad at p2, non-mmx
qxstof:GA;:ok:
%ifdef __MMX
	mov ecx,TOS(1)
	add eax,KR
	add ecx,KR
	movq mm0,[eax]
	movq mm1,[eax+8]
	movq mm2,[ecx]
	movq mm3,[ecx+8]
	spop 2
	movq [ecx],mm0
	movq [ecx+8],mm1
	movq [eax],mm2
	movq [eax+8],mm3
	ret
%else
	mov ecx,TOS(1)
	spush 2
	add ecx,byte 8
	add eax,byte 8
	mov TOS(1),ecx
	call twoxstof	;:continued:
%endif
;-
;- 2!!		( p1 p2 -- )			(x) 		'twoxsto'	forth
;-	exchange contents of double at p1 with double at p2
twoxsto:GA;:ok:
%ifdef __MMX;:ok: 3*
	lea ecx,[kref eax]
	spop
	movq mm0,[ecx]
	movq mm1,[kref eax]
	movq [ecx],mm1
	movq [kref eax],mm0
	spop
	ret
%endif
;-
;- 2!!		( p1 p2 -- )					'twoxstof'	float
;-	exchange contents of double at p1 with double at p2, non-mmx
twoxstof:GA;
	push dword[kref eax]
	push dword[kref eax+4]
	push dword TOS(1)
	call twofstof
	pop ecx
	pop dword[kref ecx+4]
	pop dword[kref ecx]
	ret
;-
;- c+!		( u p -- )			(m,15)		'cplsto'	forth
;-	add byte u to byte pointed to by p
cplsto:GA M;:ok:
; c: ecx
	lea ecx,[kref eax]
	spop
	add [ecx],al
	spop
M cplsto
	ret
;-
;- +!		( u p -- )			(m,15)		'plsto'		forth
;-	add u to cell pointed to by p
plsto:GA M;:ok:
; c: ecx
	lea ecx,[kref eax]
	spop
	add [ecx],eax
	spop
M plsto
	ret
;-
;- m+!		( u ptr -- )			(f6)		'mplsto'	forth
;-	add sign extended sing @nos to double pointed to by @tos
mplsto:GA;:ok:
; c: ecx,edx
	mov ecx,eax
	mov eax,TOS(1)
	sar eax,31
	spush
	mov eax,ecx	;:continues:
;-
;- 2+!		( ud ptr -- )			(f6)		'twoplsto'	forth
;-	add double @nos to address pointed to by @tos
twoplsto:GA;:ok:
; c: ecx,edx
	lea ecx,[kref eax]
	spop
	mov edx,eax
	spop
	add [ecx],eax
	adc edx,0
	add [ecx+4],edx
	spop
	ret
;-
;- (1+)		( p -- )			(e,fg)(M,8)	'b1pb'		forth
;-	increment cell at p by one
b1pb:GA M;:ok:
	inc dword[kref eax]
	spop
M b1pb
	ret
;-
;- (1-)		( n1 -- n2 )			(e,fg)(m,8)	'b1mb'		forth
;-	decrement cell at p by one
b1mb:GA M;:ok:
	dec dword[kref eax]
	spop
M b1mb
	ret
;-
;- (d1+)  	( p -- )			(e,fg)(M,8)	'bd1pb'		forth
;-	increment double cell at p by one
bd1pb:GA M;:ok:
	add dword[kref eax],byte 1
	adc dword[kref eax+4],byte 0
	spop
M bd1pb
	ret
;-
;- (d1-)  	( n1 -- n2 )			(e,fg)(m,8)	'bd1mb'		forth
;-	decrement double cell at p by one
bd1mb:GA M;:ok:
	sub dword[kref eax],byte 1
	sbb dword[kref eax+4],byte 0
	spop
M bd1mb
	ret
;-	
%endif;__MEMFSTO_I
;-
;-=====================================================================;
;- l4memfsto.inc <eof>
