;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4numconv.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __NUMCONV_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" numeric i/o and conversion words
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __NUMCONV_I
%define __NUMCONV_I
%define numconvinc
;-
;-	
;-----------------------------------------------------------------
;--- numeric input modifiers
;-

;-
;- big		( -- )				(i)(l4)				root
;-	next numeric input will be taken as a counted integer.
;-	set bignum-flag in uvari (bits) for numeric stg input,
;-	flag reset immediately after next literal number or, at eol.
big:GA;:ok:
	or dword[byte uref uv.bits],mbign
	ret
;-
;- quad		( -- )				(i)(f6)				root
;-	next numeric input will be taken as a quad integer or a double 'ranum'.
;-	set quad-flag in uvari (bits) for numeric stg input,
;-	the flag can accumulate w/ -re- {ranum} or "\" number input praefix.
;-	reset immediately after next literal number or, at eol.
quad:GA;:ok:
	or dword[byte uref uv.bits],mquad
	ret
;-
;- sf		( -- )				(i)(fs)		'sing'		root
;-	next numeric input will be taken as a SINGLE RATIONAL NUMBER.
;-	set 'ranum'-flag in uvari (bits) for numeric stg input -re- {rational}.
;-	flag can accumulate for quad ranum input w/ -re- {quad}.
;-	reset immediately after next literal number or, at eol.
sing:GA;:ok:
	or dword[byte uref uv.bits],mranu
	ret
;-
;- df		( -- )				(i)(fs)		'ranum'		root
;-	next numeric input will be taken as a DOUBLE RATIONAL NUMBER.
;-	set 'ranum'-flag in uvari (bits) for numeric stg input -re- {rational}.
;-	flag can accumulate for quad ranum input w/ -re- {quad}.
;-	reset immediately after next literal number or, at eol.
ranum:GA;:ok:
; - ranum	( -- )				(i)(fs)				root
	or dword[byte uref uv.bits],mranu|mquad
	ret
;-
;- qf		( -- )				(i)(fs)		'double'	root
;-	next numeric input will be taken as a QUAD RATIONAL NUMBER.
double:GA;:ok:
	or dword[byte uref uv.bits],mqranu|mranu|mquad;=mdfloat
	ret
;-
;- flt    	( -- )				(i)(4)		'fltf'		root
;-	all numeric input will be taken as a double 'ranum', until xec. of {int}.
;-	pushes the -re- {FLOAT} vocabulary 1st to voc-stack if not already there.
;-	NOTE that in {flt} state any numeric input will return a rational number 
;-	(double) cells pair, which can be forded to quad integer w/ '\' praefix.
;-	Other numeric types can only be fetched in standard state -re- {int}.
fltf:GA;:ok:
	TICK float
	call orderq
	test eax,eax
	spop
	jz .f	;?; {float} 1st on voc-stack
	call also
	call float
.f:
	or dword[byte uref uv.bits],(mfloatm)
	ret
;-
;- int    	( -- )				(i)(4)		'intf'		root
;-	clear {flt} state to default, integers interpreting mode.
;-	takes {FLOAT} from top of voc-stack, if applicable.
;-	-re- {sf}, {df}, {qf} for single or, {flt} permanent fp input mode.
intf:GA;:ok:
	TICK float
	call orderq
	test eax,eax
	spop
	jnz .f	;?; {float} not 1st on voc-stack
	call previous
.f:
	and dword[byte uref uv.bits],~(mfloatm)
	ret
;-
;- -flt    	( -- f )					'fltq'		root
;-	ret tf if numeric input mode not persistent floats -re- {flt}.
fltq:GA
	spush
	mov eax,[byte uref uv.bits]
	and eax,mfloatm
	cmp eax,mfloatm
	jmp zeq
;-
;-
;-----------------------------------------------------------------
;--- number conversion
;-

;-
;- base!  	( n -- )			(f6)(m,8)	'bases'		forth
;-	store @tos to uvari {base}.
bases:GA M;:ok:
	mov [byte uref uv.base],eax	
	spop
M bases
basef.r:
	ret
;-
;- base@  	( -- n )			(f6)(m,4)	'basef'		forth
;-	fetch current number conversion radix.
;-	with range check, 1 < @base < 257 if not compiled as a macro.
;-	expect abiguous numeric results if out of range 1 < @base < 197.
;-	base = 256 used for string data to character code conversion.
basef:GA M;:ok:
; c: -/-
	spush
.n:; ;<abnumberb)
	mov eax,[byte uref uv.base]
M basef
	cmp eax,256
	jnc .d	;?; fall back to dft @base := 10
	cmp eax,byte 2
	jnc .r	;(above, branch 'backward' for the cpu-s 'branch prediction' -?-)
.d:
	mov eax,10
	mov [byte uref uv.base],eax
	ret
;-
;- binary 	( -- )				(f6)				forth
;-	set number conversion radix := 2
binary:GA;:ok:
; c: -/-
	push_ibyte(2)
sbas:
	pop dword[byte uref uv.base]
	ret
;-
;- octal  	( -- )				(f6)				forth
;-	set number conversion radix := 8
octal:GA;:ok:
	push_ibyte(8)
	jr sbas
;-
;- decimal	( -- )				(f6)				forth
;-	set number conversion radix := 10
decimal:GA;:ok:
	push_ibyte(10)
	jr sbas
;-
;- hex		( -- )				(f6)				forth
;-	set number conversion radix := 16
hex:GA;:ok:
	push_ibyte(16)
	jr sbas
;-
;- u>d		( n1 -- ud1 )			(m,3)		'u2dn'		ans
;-	does{ s>d 2abs }
;-
;- u->d		( n1 -- ud1 )			(m,3)		'u2dn'		forth
;-	unsigned extend sing to double, does{ s>d 2abs }.
u2dn:GA M;:ok:
	spush
	xor eax,eax
M u2dn
	ret
;-
;- d->q		( dn -- qn )			(m,5)		'd2qn'		forth
;-	sign extend double to quad
d2qn:GA M
	spush
	sar eax,31
	spush
M d2qn
	ret
;-
;- s->d		( n1 -- d1 )			(fg)(m,4)	's2dn'		forth
;-	sign extend sing to double
;-
;- s>d		( n1 -- d1 )			(m,4)		's2dn'		forth
;-	sign extend sing to double
s2dn:GA M;:ok:
; c: -/-
	spush
	sar eax,31
M s2dn
	ret
;-
;- s->c		( n1 -- +n1 )			(m,3)		's2cn'		forth
;-	truncate cell to unsigned byte
;-	effect is extension of l.s.byte to unsigned single
s2cn:GA M;:ok:
	movzx eax,al
M s2cn
	ret
;-
;- s->w		( n1 -- +n1 )			(m,3)		's2wn'		forth
;-	truncate cell to unsigned word
;-	effect is extension of l.s. half cell to unsigned single
s2wn:GA M;:ok:
	movzx eax,ax
M s2wn
	ret
;-
;- c->s		( n1 -- s )			(e)(m,3)	'c2sn'		forth
;-	signed extend l.s.byte to single
c2sn:GA M;:ok:
	movsx eax,al
M c2sn
	ret
;-
;- w->s		( n1 -- s )			(e)(m,3)	'w2sn'		forth
;-	signed extend l.s.half cell to single
w2sn:GA M;:ok:
	movsx eax,ax
M w2sn
	ret
;-
;- d->s		( d -- s )			(e)		'd2sn'		forth
;-	truncate double cell to single
;-	(e: NZ if double doesn't fit into single)
;-
;- d>s						(a)		'd2sn'		forth
;-	truncate double cell to single.
;-	re DPANS 8.4.1.2: m.s. cell is lost, w/o indication of ovf.
d2sn:GA;:ok:
	test eax,eax
	jz .p
	inc eax
	jnz .p
	inc eax		; NZ if ovf
.p:
	spop
	ret
;-	
;- d->b		( d -- n1 .. n8 )				'd2bn'		forth
;-	unsigned extract 8 bytes of a double, m.s.b @tos
d2bn:GA
	push eax
	spop
	call s2bn
	spush
	pop eax
;-
;- s->b		( n -- n1 n2 n3 n4 )				's2bn'		forth
;-	unsigned extract 4 bytes of a sing, m.s.b @tod
s2bn:GA
	mov ecx,eax
	movzx eax,al
	spush
	movzx eax,ch
	shr ecx,16
	spush
	movzx eax,cl
	spush
	movzx eax,ch
	ret
;-
;- b->s		( n1 n2 n3 n4 -- n )				'b2sn'		forth
;-	combine 4 sings l.s.bytes to sing n, ms.b from tos
b2sn:GA
	shl eax,8
	mov al,byte TOS(1)
	shl eax,16
	mov ah,byte TOS(2)
	mov al,byte TOS(3)
	TOSPULL(3)
	ret
;-
;- b->d		( n1 .. n8 -- d )				'b2sn'		forth
;-	combine 8 sings l.s.bytes to double d, ms.b from tos
b2dn:GA
	call b2sn
	push eax
	spop
	call b2sn
	spush
	pop eax
	ret
;-
;- sign		( n dn -- dn )			(fg)				fig
;-	FIG mode - ANS-4th is different: ( n -- ) !
;-	when in case-independent search mode and {forth} on top of voc-stack 
;-	this variant takes precedence over ANS-style {sign}!
sign:GA;:ok:
	call rot	;:continues:
;-
;- SIGN		( n -- )			(a:c)		'asign'		forth
;-	ANS word, linked to {forth} vocabulary, fig-4th is different: ( n dn -- dn )!
;-	fig-4th {sign} takes precedence when in case-independent search mode and fig 
;-	on top of voc-stack.
asign:GA;:ok:
	test eax,eax
	jns g32 hold.p
	mov al,'-'
	jr hold
;-
;- q#s		( q1 -- q0 )					'qsharps'	forth
;-	convert q1 to numeric characters at @hld - re {#s}
qsharps:GA;:ok:
	call qsharp
	mov ecx,eax
	or ecx,TOS(1)
	or ecx,TOS(2)
	or ecx,TOS(3)
	jnz qsharps
	ret
;-
;- q#		( q1 -- q2 )					'qsharp'	forth
;-	{#} wrt quad no.
qsharp:GA;:ok:
	spush
	mov eax,[byte uref uv.base]
	call uqsmby
	spush
	mov eax,4
	call roll
	jr sharp.q
;-
;- #		( d1 -- d2 )					'sharp'		forth
;-	divide d1 by {base}, convert remainder to character,
;-	append char to beginning of numeric output string.
sharp:GA;:ok:
	spush
	mov eax,[byte uref uv.base]
	call umby
	call rot
.q:
	call h2a	;:continues:
;-
;- hold		( c -- )							forth
;-	prepend char c to numeric stg at @-{hld}
hold:GA;:ok:
; c: ecx
	mov ecx,[byte uref uv.hld]
	dec ecx
	mov [byte uref uv.hld],ecx
	mov [kref ecx],al
.p:
	spop
	ret
;-
;- <#		( -- )						'ltsharp'	forth
;-	init pictured number conversion, quad or sing (any size)
ltsharp:GA;:ok:
; c: -/-
	call pad2
	mov [byte uref uv.hld],eax
	mov [byte uref uv.hld0],eax
	spop
	ret
;-
;- #s		( d1 -- 0. )					'sharps'	forth
;-	convert d1 to numeric characters at @-{hld}
sharps:GA;:ok:
	call sharp
	test eax,eax
	jnz sharps
	cmp dword TOS(1),0
	jnz sharps
	ret
;-
;- h>a		( n -- c )			(h4)		'h2a'		forth
;-	character representation c of byte n
;-	for any n within range 1 < @{base} < 198
h2a:GA;:ok:
	mov cl,[byte uref uv.base]
	test cl,cl
	jz .r		;?; base is ascii/utf
	cmp cl,198
	jnc .r		;?; cannot convert
	movzx eax,al
	cmp al,10
	jc .n
	add al,7
.n:
	add al,'0'
.r:
	ret
;those are all much slower:
; --------------------- ;
;    add al,0x90	;t46;
;    daa
;    adc al,0x40
;    daa
; --------------------- ;
;    add al,40h		;t33;
;    movsx ax,al	;t40;for eax
;    and ah,09h
;    add al,ah
;    and al,0Fh
; --------------------- ;

;-
;- .r		( n u -- )					'dotr'		forth
;-	right aligned display n in u chars field, filled up w/ blanks.
;-	pass 1-s complement of fieldwidth n to appending a trailing blank.
dotr:GA;:ok:
	mov ecx,eax
	mov eax,TOS(1)
	sar eax,31
	spush
	mov eax,ecx
	jr ddotr
;-
;- ?		( a -- )					'ques'		forth
;-	signed display content of cell at ptr a
ques:GA;:ok:
	mov eax,[kref eax];:continues:
;-
;- .		( n -- )					'dot'		forth
;-	signed sing numeric output
dot:GA;:ok:
	spush
	sar eax,31	;:continues:
;-
;- d.		( d -- )					'ddot'		forth
;-	signed double numeric output
ddot:GA;:ok:
; { >r r 2abs <# #s sign #> type space }
	call mone	; zero fieldwidth, trailing blank ;:continues:
;-
;- d.r		( d n -- )			(f6)		'ddotr'		forth
;-	right aligned display signed double numeric output string
;-	pass 1-s complement of fieldwidth n to appending a trailing blank.
ddotr:GA;:ok:
	lea ecx,[kref type-Y]
	push ecx	;:continues:
;-
;- (d.r)  	( d n -- p u )			(f6)		'pddotrp'	hidden
;-	ret double d as signed double numeric stg(p,u)
;-	right aligned to n bytes field, ready to {type}
;-	pass 1-s complement of fieldwidth n to appending a trailing blank.
pddotrp:GA;:ok:
	push dword TOS(1); sign
	call tor	; fieldwidth
	call dabs
	call rfrom
	jr puddotrp.n
;-
;- (ud.r)  	( ud n -- p u )			(f6)		'puddotrp'	hidden
;-	ret double d as unsigned double numeric stg(p,u)
;-	right aligned to n bytes field, ready to {type}
;-	pass 1-s complement of fieldwidth n to appending a trailing blank.
puddotrp:GA;:ok:
	push_ibyte(0)
; puddotrp.n
;	entry w. @tor := sign-flag, for (d.r)
.n:
	push eax	; fieldwidth & trailing space flag in sign bit
	spop
	call ltsharp
	mov ecx,[esp]
	test ecx,ecx
	jns .l
	not ecx
	mov [esp],ecx
	call bbl
	call hold
.l:
	call sharps
	spush
	mov eax,[esp+4]	; number sign flag
	call asign
	mov ecx,[esp]
	sub ecx,[byte uref uv.hld0]
	add ecx,[byte uref uv.hld]
	jle .s
	mov [esp],ecx
.h:
	call prpf	; (r)@
	call hold
	dec dword[esp]
	jg .h
.s:
	lea esp,[esp+8]	;:continues:
;-
;- #>		( xd -- caddr u )			(m)	'sharpgt'	forth
;-	finish numeric output string, ready to {type}
sharpgt:GA M;:ok:
	mov eax,[byte uref uv.hld]
	mov TOS(1),eax
	neg eax
	add eax,[byte uref uv.hld0]
M sharpgt
	ret
;-
;- q#>		( xq -- caddr u )			(m)	'qsharpgt'	forth
;-	finish numeric output string, ready to {type}
qsharpgt:GA;:ok:
	TOSPULL(2)
	jr sharpgt
;-
;- (q.)		( q n1 n2 -- p u )			(k)	'pqdotp'	hidden
;-	convert quad to (n1:+ve)un- (n1:-ve)signed numeric string
;-	1-s complement fieldwidth n2 to appending a trailing blank.
; c: ecx,edx
pqdotp:GA;:ok:
	mov ecx,eax
	spop
	test eax,eax
	push eax	; swap over 2>r	{ n1
	push ecx	;		{ n2 n1 
	spop
	jns .s		;?;unsigned
	call qabs	; 0< IF qabs ENDIF
.s:
	call ltsharp	; <#
	cmp dword[esp],0; r		{ n1
	jns .b		; 0< IF
	not dword[esp]	; r> minus >r
	call bbl	;  bl hold
	call hold	; ENDIF
.b:
	call qsharps	; q#s
	spush
	pop ecx		; r>
	mov eax,[esp]	; r>
	mov [esp],ecx	; swap >r
	call asign	; sign
	pop edx		; r>
	sub edx,[byte uref uv.hld0]; hld0 @ -
	add edx,[byte uref uv.hld];  hld @ +
.h:;			; begin
	dec edx		;  1-
	js qsharpgt	;?;0< if q#> exit endif
	call prpf	;  (r)@
	call hold	;  hold
	jr .h		;>;again
;-
;- q.		( q -- )					'qdot'		forth
;-	display signed quad numeric output string.
qdot:GA;:ok:
	spush		; save sign
.n:
	call mone	; no field, -ve: append blank
	call pqdotp
	jmp type
;-
;- uq.		( q -- )					'uqdot'		forth
;-	display unsigned quad numeric output string.
uqdot:GA;:ok:
	spush
	xor eax,eax	; unsigned
	jr qdot.n
;-
;- q.r		( q u -- )					'qdotr'		forth
;-	right aligned display signed quad numeric output string,
;-	pass 1-s complement fieldwidth u to appending a trailing blank
qdotr:GA;:ok:
	call over
.n:
	call swap
	call pqdotp
	jmp type
;-
;- uq.r		( uq u -- )					'uqdotr'	forth
;-	right aligned display unsigned double numeric output string
;-	pass 1-s complement fieldwidth u to appending a trailing blank
uqdotr:GA;:ok:
	spush
	xor eax,eax	; unsigned
	jr qdotr.n
;
sgndot:A
	test eax,eax
	js .c
	ret
.c:
	spush
	mov eax,0+'-'
	call emit
	jmp cabs
;-
;- s#.		( n -- )					'sdecdot'	forth
;-	pictured signed numeric decimal output
sdecdot:GA;:ok:
; c: ecx
	call sgndot	;:continues:
;-
;- #.		( n -- )					'decdot'	forth
;-	pictured unsigned numeric decimal output
decdot:GA;:ok:
; c: ecx
	call basedot
	db 10
;-
;- &.		( n -- )					'octdot'	forth
;-	pictured unsigned numeric octal output
octdot:GA;:ok:
; c: ecx
	call basedot
	db 8
;-
;- %.		( n -- )					'bindot'	forth
;-	pictured unsigned numeric binary output
bindot:GA;:ok:
; c: ecx
	call basedot
	db 2
;-
;- $.		( n -- )					'hexdot'	forth
;-	pictured unsigned numeric sedecimal output
hexdot:GA;:ok:
; c: ecx
	call basedot
	db 16
basedot:A
	pop ecx
	push dword[byte uref uv.base]
	movzx ecx,byte[ecx]
	mov [byte uref uv.base],ecx
	lea ecx,[kref sbas-Y]
	push ecx
;-
;- u.		( u -- )					'udot'		forth
;-	pictured unsigned sing numeric output
udot:GA;:ok:
	call zero
;-
;- ud.		( ud -- )					'uddot'		forth
;-	unsigned double numeric output
uddot:GA;:ok:
	call mone
;-
;- ud.r		( ud n -- )					'uddotr'	forth
;-	unsigned double numeric output right aligned to n bytes field
uddotr:GA;:ok:
	call puddotrp
	jmp type
;-
;- u.r		( u1 +u -- )					'udotr'		forth
;-	cnt and ptr to unsigned sing numeric output string
;-	right aligned to n bytes field, ready to {type}
udotr:GA;:ok:
	mov ecx,eax
	xor eax,eax
	spush
	mov eax,ecx
	jr uddotr
;-
;- u?		( a -- )					'uques'		forth
;-	unsigned display content of cell at ptr a
uques:GA;:ok:
	mov eax,[kref eax]
	jmp j32 udot
;-
;- -- numeric stg input
;-	

;-
;- digit  	( c -- n tf | c ff )						forth
;-	tf: c is a digit, n := numeric value corresponding to c
;-	ff: c not a digit, value remains un-changed
;-	base = 256 assumed and no conversion if { @base 256 mod } = 0
;-	abiguous result if 256 > @base > 96
digit:GA;:ok:
; c: ecx,edx
	movzx ecx,al
	cmp byte[byte uref uv.base],0
	jz .a		;?; no conversion if base@ = 256 (or zero)
	mov edx,[uref uv.ulo2hi]
	add edx,KR
	mov cl,[edx+ecx]
	sub ecx,byte'0'	; #48 $30
	jc .f		;?; not a valid char
	cmp ecx,byte 10
	jc .b		;?; decimal digit
	lea ecx,[ecx-7]	; #55 $35
	cmp ecx,byte 10
	jc .f		;?; ascii gap
.b:
	cmp ecx,[byte uref uv.base]
	jnc .f		;?; not in range
.r:
	mov eax,ecx	; char value
	jmp true	; true ;s
.a:
	cmp al,' '	; any ctrl char and {bl} terminates ascii mode
	ja .r
.f:
	jmp false	; false ;s
;-
;- number  	( p | p u -- dn | abort )	(v)		'number'	forth
;-	string to number conversion. - vectored by (number), re fig/ans words,
;-	actual xec mode dependent on ans/fig preference as set in "Makefile".
;-	default -re- {[number]} in {forth} voc.
number:GA;:ok:
	spush
	mov eax,[uref uv.unumber]
	jmp execute.c
;-
;- f-number	( p -- dn | abort )		(fg)		'fnumber'	fig
;- number  	( p -- dn | abort )				'fnumber'	fig
;-	counted string to number conversion, f.i.g.-Forth mode.
;-	abort if leading char, after opt. prefices, not a currently valid digit.
;-	detailed spec. -re- {[number]} in {forth} voc.
fnumber:GA;:ok:
	inc eax
	call bnumberb
	jr rnno
;-
;- a-number	( p u -- dn | abort )		(a)		'anumber'	ans
;- NUMBER  	( p u -- dn | abort )				'anumber'	ans
;-	counted string to number conversion.
;-	detailed spec. -re- {[number]} in {forth} voc.
anumber:GA;:ok:
	call abnumberb
rnno:
	spush
rno:
	mov eax,31
	jmp qerror
;
abnumberb.c:A; ctrl char
	mov ecx,TOS(2)
	movsx eax,ax
	movzx ecx,byte[kref ecx]
	add ecx,eax
	xor eax,eax
	mov TOS(2),ecx
	mov TOS(1),eax
	ret
;
abnumberb.f:A; rational number (float)
	push dword[byte uref uv.base]
	movzx ecx,ax	; working base value
	mov [byte uref uv.base],ecx
	call abranumb
	pop dword[byte uref uv.base]
	ret
;-
;- [number]	( p -- [dn|qn|cn] ff | 0. 1 )	(fg)		'bnumberb'	fig
;-	number conversion of chars at ptr p, until <bl> or low code ctrl char.
;-	f.i.g.-4th mode, otherwise identical -re- {[number]} in forth voc.
bnumberb:GA;:ok:
	call zblcount
;-
;- [number]	( p +u -- dn|qn|cn| f )		(a)(v)		'abnumberb'	forth
;-	integer number conversion of up to +u chars at ptr p. ANS-style word.
;-	returns flag f = no. of remaining, unconverted chars in stg(p,u).
;-	ascii-numeric prefixed string terminated by ctrl chars or blank.
;-	
;-	stg may contain a base exponent mark, either the char corresponding to 
;-	the rsp. @base figure as converted w/ {h>a} or, up to decimal base the
;-	char "e" or "E". the exponent |exp|<128 can be fetched w/ -re- {edpl@} 
;-	and will automatically be taken into accont in 'ranum' input mode, only.
;-	stg can be prefixed by char(s) for a specific base value, re {a-base}.
;-	valid quad, double and sing numbers include '-ve zero'.
;-	
;-	modifiers for numeric stgs w/ optional sign & base praefix:
;-	{ big ..nn.. }
;-		for unlimited size counted integers -re- {nnumber}.
;-	{ quad ..nn.. }
;-		quad cells integer.
;-	{ sf ..nn.. } or, { \nn.. }
;-		praefix for rational number ordered pair,
;-		{ \nn.. } for a quad int while in persistent f.p. mode -re- {flt}.
;-	{ df ..nn..} or, { quad \..nn.. }
;-		next input is double rational number,
;-	{ qf ..nn..}, { sf ..nn.. }
;-		for quad and single rational number, rsp.
;-	
;-	dpl settings, 'nn' no. of digits after frac. marker:
;-		0xff80nnnn	counted, umlimited size integer
;-		0x000080nn	quad integer
;-		0x008000nn	rational number ('ranum'), cells pair
;-		0x008080nn	float, double ranum by pair of doubles
;-		0x008088nn	eight cells, quad ranum by pair of quads
;-		  -1		sing
;-		  000000nn	any other +ve is a double
;-	edpl@
;-		returns the optional, trailing exponent.
abnumberb:GA;:ok:
; c: ecx,edx
	mov dword[byte uref uv.dpl],-1	;flag sing
	mov byte[uref uv.bits+edpl],0	; clr -ve exp count
; counted int
	test dword[byte uref uv.bits],mbign
	jnz near abnnumberb ;?; counted integer
	call abase	;		{ ptr' u' base
	test ax,ax	; temp. base
	jg .z		;?; +ve non-zero, i.e. base set
	js .c		;?; ctrl-char
; prevent fatal conversion errors and, get valid base in case of recursed 'float' type input
	mov ecx,eax	; save sign
	call basef.n	; get safe base value
	mov ax,cx	; sign|base
.z:
%ifdef __RATIONAL
	test dword[byte uref uv.bits],mranu
;=> float, to different radices, implemented as 'ranum' (reelle zahl) by praefix 'ranum' or 'vulgar'
	jnz .f		;?; rational number (float)
%endif
	push eax	; save sign flg
	push dword[byte uref uv.base];save current base		[ b
	movzx eax,ax
	mov [byte uref uv.base],eax;store tmp. radix
	xor eax,eax
	mov ecx,TOS(1)	; save cnt				[ u b
	mov edx,TOS(2)	; and ptr
	mov TOS(1),eax	; init..			{ 0 00
	mov TOS(2),eax
	test dword[byte uref uv.bits],mquad|mqranu
	jz .q
	or dword[byte uref uv.bits],mquad
	spush
	spush
.q:
	mov eax,edx	; ptr
	spush
	mov eax,ecx	; max-len		{u p 00(00q) [ b
	dec dword[byte uref uv.dpl];compensate..
.l:
	inc eax		;  ..dpl and cnt of '.' chars
.b:
	inc dword[byte uref uv.dpl]
	js .n
	mov dword[byte uref uv.dpl],0; (re)start dpl cnt
.a:
	mov ecx,[byte uref uv.bits]
	and ecx,mquad
	jz .n		;?; sing
	mov dword[byte uref uv.dpl],0x00008000; force quad int
.n:
	dec eax		; discard fraction mark		{u- p accu
	jz .r		;?; ready after dpl-ed
	call tonumber	; do the conversion ( dn|qn p u -- dn'|qn' p' u' )
	test eax,eax
	jz .r		;?; zero stg count
	mov edx,TOS(1)	; ptr to last char
	inc dword TOS(1)
	mov ecx,[uref uv.udplm]; fraction marker char(s)
.d:
	cmp byte[kref edx],cl
	jz .b		;?; more digits after @dpl
	shr ecx,8
	jnz .d		;?; another marker char
	push eax	; stg count
	movzx eax,byte[kref edx]
	cmp al,d_bl	; <bl>
	jbe .e		;?; terminator
	call digit
	test eax,eax
	spop
	jz .x		;?; invalid
.e:
	dec dword[esp]
.x:
	pop eax		; stg count
	call .t		; mul by base exponent, NZ if not found
.r:
	TOSPULL(1)	; swap drop, drop ptr & leave count = flag
	pop dword[byte uref uv.base]; rst gloval value	{u accu	[
	lea edx,[byte uref uv.dpl]
	shl dword[esp],1; sign flag
	mov ecx,[byte uref uv.bits]
	lea esp,[esp+4]	; rdrop
	jnc .p		;?; +ve number
	and ecx,mquad
	mov ecx,TOS(1)
	jz .m		;?; -ve double
	or ecx,TOS(4)
	or ecx,TOS(3)
	neg dword TOS(4)
	adc dword TOS(3),0
	neg dword TOS(3)
	adc dword TOS(2),0
.m:
	neg dword TOS(2)
	adc dword TOS(1),0
	neg dword TOS(1)
	or ecx,TOS(2)
	jnz .p
	mov dword TOS(1),0x80000000; '-ve zero'
	cmp dword[edx],-1
	jnz .p		;?; double or dpl'd quad
	test dword[byte uref uv.bits],mquad
	jnz .p		;?; not a sing
	mov dword TOS(2),0x80000000; sing '-ve zero'
.p:A
	btr dword[byte uref uv.bits],bquad
	jnc .o
	cmp dword[edx],0; dpl
	jns .o
	mov dword[edx],0x00008000; flag 'quad'
.o:
	test dword[byte uref uv.bits],mfloat
	jz .g
	or dword[byte uref uv.bits],mfloatm;=mfloat|mquad|mranu	; rst persistent ranum flags
.g:
	test eax,eax	; flags for caller
	ret
; find base exponent
.t:A; ( dn|qn p u' -- dn|qn" p" u" )
	spush		; base@
	mov eax,[byte uref uv.base]
	call h2a	; h>a
	movzx eax,al	; base-char
	mov ecx,TOS(2)
	mov cl,[kref ecx-1]; marker char
	mov edx,[uref uv.ulo2hi]
	add edx,KR
	cmp cl,byte[edx+eax]
	jz .tx		;?; base-char exponent present
	mov edx,[uref uv.uhi2lo]
	add edx,KR
	cmp cl,byte[edx+eax]
	jz .tx
	cmp al,'E'+1;;'A':
	jnc .tr		;?; not decimal (or below)
	cmp cl,'e'
	jz .tx		;?; decimal exponent present
	cmp cl,'E'
	jz .tx		;?; decimal exponent present
.tr:
	spop
	ret
.tx:A; calculate & store base exp. ( q|d p u char -- 
	push dword[byte uref uv.dpl]
	push dword[byte uref uv.bits]
	mov dword[byte uref uv.dpl],-1
	and dword[byte uref uv.bits],~(mquad); force double
	mov ecx,TOS(1)	; u
	mov edx,TOS(2)	; p
; 2>r base@ 0 0 2r >number drop c@ bl = if r1+ endif d->s u^ 
	mov eax,[byte uref uv.base];	( qd p u b
	mov TOS(2),eax	;		( qd b u b
	mov eax,0
	mov TOS(1),eax	;		( qd b 0 0
	cmp byte[kref edx],'-'
	spush		;		( qd b 0 0 x
	push_ibyte(0)	; sign flag
	jnz .tp		;?; +ve exponent
	dec dword[esp]
	inc edx		; 1/string
	dec ecx
.tp:
	mov eax,edx	; ptr
	spush		;		( qd b 0 0 p u
	mov eax,ecx	; max-len
	call tonumber	; get exponent	( qd b e.e p u
	mov ecx,TOS(1)	; p"
	push ecx	; p"  save (p,u)"
	push eax	; u"			| u p sgn bits dpl
	spop 3		; 2>r d->s	( qd b e
	movzx ecx,byte[kref ecx]
	cmp cl,d_bl
	ja .tv		;?; no terminator
	dec dword[esp]	; discard trailing <bl>
.tv:
	movzx ecx,al
	cmp dword[esp+8],byte 0;sgn
	jns .tt		;?; -ve exp	
	neg ecx
.tt:; 2drop 2r>
	pop eax		; u"
	pop dword TOS(1); p"		( q|d p u
	lea esp,[esp+4]	; drop sgn
	pop dword[byte uref uv.bits]
	pop dword[byte uref uv.dpl]
	mov [byte uref uv.bits+edpl],cl; exp
	ret
;-
;- convert	( dn1 ptr1 -- dn2 ptr2 flg )			'converta'	forth
;-	ANS-Forth mode, in {ans} and {forth} vocabularies.
;- CONVERT	( dn1 ptr1 -- dn2 ptr2 )	(a:cx)		'converta'	ans
;-	interpret string at @ptr as number to radix @base. -re- {[number]}.
converta:GA
	lea ecx,[kref drop-Y]
	push ecx
;-
;- convert	( dn1 ptr1 -- dn2 ptr2 flg )	(fg)				fig
;-	fig-Forth mode, dft in {fig} vocabulary, only. -re- {>number}.
;-	interpret string at @ptr as number to radix @base until 1st non-valid 
;-	char, leave accumulated double, ptr to after last valid char and flag.
;-	1-s complement of flag is length of digits string found valid.
;-	flag result such that {convert} could be continued after having the 
;-	terminating char analyzed and taken into account (i.e. flg:=flg+1).
convert:GA;:ok:
	call mone	; cnt := ffffffff
;-
;- >number	( dn1|qn1 ptr1 u1 -- dn2|qn2 ptr2 u2 )	(v)(m,6) 'tonumber'	forth
;-	interpret string at @ptr as number to radix @base
;-	until 1st non-valid and no more than u1 char(s).
;-	leave remaining cnt and ptr. flag u2 := 0 if all chars converted.
;-	execute true address uot vector (>number), dft {[>number]}
tonumber:GA M;:ok:
	jmp [uref uv.utonum]
M,tonumber
	call [uref uv.utonum]
M tonumber
;-
;- [>number]	( dn1|qn1 ptr1 u1 -- dn2|qn2 ptr2 u2 )		'btonumb'	hidden
;-	dft for {>number}, fetches quad integer immediately after {quad},
;-	quad flag in dpl: 80xx, w. (xx) for digits after fraction sign.
;-	-re- {sf}, {df}, {qf} for single or, {flt} permanent fp input mode.
btonumb:GA;:ok:
; c: ecx,edx
%if 1; 40% xec time wrt version w/ do..loop; :ok:
	push eax	; >rr		{ u p dd	[ u
	mov ecx,[byte uref uv.bits]
	and ecx,mquad
	spop		; drop		{ p dd
	jnz .qd		;?; quad
;	call pdop	; (do)		{ u p dd	[ i u
.dd:; double		;		{ n p dd
	inc eax
	spush		;  c@+
	movzx eax,byte[byte kref eax-1];{ c p+ dd
	call digit	; digit		{ f c p dd
	test eax,eax
	spop		;IF
	jz .xd		;?; not a digit	{ n p dd
	cmp dword[byte uref uv.dpl],0
	js .d		;   dpl @ 0< 0=
	inc dword[byte uref uv.dpl];IF 1 dpl +! ENDIF
.d:; double		;		{ n p dd
	call swap2	; 2swap		{ dd n p
	spush		; base@
	mov eax,[byte uref uv.base];	{ b dd n p
	call udmmul	; udm*		{ o db n p
	spop		; drop ovf
	call rot	; rot		{ d db p
	add TOS(2),eax
	spop
	adc eax,0	; m+		{ dd+ p
	call rot	; rot		{ p dd
.ed:
	dec dword[esp]
	jnz .dd;;jg .dd;
.xq:
	inc eax
	spush
.xd:
	dec dword TOS(1)
	pop eax
	ret
.qd:; quad
	inc eax
	spush		;  c@+
	movzx eax,byte[byte kref eax-1];{ c p+ qqqq
	call digit	; digit		{ f c p qqqq
	test eax,eax
	spop		;IF
	jz .xd		;?; not a digit	{ n p qqqq
	cmp dword[byte uref uv.dpl],0
	js .q		;   dpl @ 0< 0=
	inc dword[byte uref uv.dpl];IF 1 dpl +! ENDIF
.q:;			;		{ d p qqqq
	push dword TOS(1);		{ qqqq
	push eax	; swap 2>r			[ d p
	spop
	mov eax,[byte uref uv.base];	{ b qqqq 
	call uqsmmul	; uqsm*		{ o qqqq'
	pop eax		; drop r>	{ d qqqq'	[ p
	call qsmplus	; qsm+		{ qqqq'
	spush
	pop eax		; r>		{ p qqqq'	[-/-
.eq:
	dec dword[esp]	; i
	jnz .qd;;jg .qd;
	jr .xq
%else
	spush		; dup
	push eax	; >rr		{ u p dd	[ u
	spush		; 0		{ 0 u p dd
	dec dword[esp]	; compensate final c@+
	xor eax,eax
	call pdop	; (do)		{ p dd		[ i u u
.do:
	spop		; drop		{ p dd
	inc eax
	spush		;  c@+
	movzx eax,byte[byte kref eax-1];{ c p+ dd
	call digit	; digit		{ f c p dd
	test eax,eax
	spop		;IF
	jz .else	;?; not a digit	{ n p dd
	cmp dword[byte uref uv.dpl],0
	js .d		;   dpl @ 0< 0=
	inc dword[byte uref uv.dpl];IF 1 dpl +! ENDIF
.d:
	mov ecx,[byte uref uv.bits]
	and ecx,mquad
	jnz .q		;?; quad
; double		;		{ n p dd
	call swap2	; 2swap		{ dd n p
	spush		; base@
	mov eax,[byte uref uv.base];{ b dd n p
	call udmmul	; udm*		{ db o n p
	spop		; drop ovf
	call rot	; rot		{ d db p
	add TOS(2),eax
	spop
	adc eax,0	; m+		{ dd+ p
	call rot	; rot		{ p dd
	jr .endif	;
; quad
.q:		;		{ d p qqqq
	push dword TOS(1);		{ qqqq
	push eax	; swap 2>r			[ d p
	spop
	mov eax,[byte uref uv.base];{ b qqqq 
	call uqsmmul	; uqsm*		{ o qqqq'
	pop eax		; drop r>	{ d qqqq'	[ p
	call qsmplus	; qsm+		{ qqqq'
	spush
	pop eax		; r>		{ p qqqq'	[-/-
	jr .endif
.else:; not a digit ;ELSE	{ c p dd
	mov eax,[esp]
	mov [esp+4],eax	; (leave)
	dec eax		; and compensate count
	mov [esp],eax
	spop		; drop		{ p dd
	dec eax		; ptr back from c@+
.endif:
	spush
	mov eax,[esp]	; i
	sub eax,KR	;
	call ploopp	; (loop)	{ i p dd	[ u
	dd .do-$
	sub [esp],eax	;r> swap -	{ u2 p2 dn
	pop eax
	ret
%endif
;-
;- a-base  	( ptr1 u1 -- ptr2 u2 n )		(f6)	'abase'		forth
;-	ret base according to praefix to number, else ret current base.
;-	(almost) any sequence of praefices will do:
;-		-ve:	"{praefix}-nnn"	and, w/ the same effect
;-			"-{praefix}nnn"	(OTA-std permitting the 1st, only).
;-		e.g:	"-\&2.3"	double ranum -2.3 octal = -2,375 dec.
;-	n := lsw. (16-bit) corresponding radix value, sign flag in msw.
;-	valid prefices & corresponding radices:
;-		'%'	2		''	12
;-		'!'	4		'$'	16
;-		'@'	8 (OTA)		'0x'	16, w/ leading "-", only
;-		'&'	8		'"'	256  ascii, "- returns zero
;-		'#'	10		'^'	-64, convert to single ctrl char code
;-		'\'	-2,  2nd-ary praefix, fetch double 'ranum' (cells pairs fraction)
;-		'\'	quad integers modifier while in persistent ranum mode -re- {flt}.
;-	max-len +u adjusted according to praefix or, -ve to limit by 1st non-valid char
;-	NOTE: '^' praefix for a single ctrl-code, by next upper-case char, only.
abase:GA;:ok:	; "\"=ascii 92; f6:"ascibase"
	mov ecx,TOS(1)	; ptr
	mov edx,eax	; len, max
	spush		; save stg(p,u) figures
	add ecx,KR	; ptr->addr
	mov eax,[byte uref uv.base]; preset n to base@
	dec edx
	jz .x		;?;ret if single char
	cmp byte[ecx],'-'
	jnz .c		;?; +ve
	inc ecx		; cnt
	or eax,0xffff0000; sign flag
.c:;		;<+ve
	cmp byte[ecx],'0'
	jnz .o		;?; try single char praefix
	inc ecx
	cmp byte[ecx],'x'
	jz .s		;?; sedec
	cmp byte[ecx],'X'
	jnz .p		;?; not sedec
.s:
	mov ax,16	; sedecimal
.i:
	inc ecx
.r:
	sub ecx,TOS(2)
	sub ecx,KR	; diff.
	add TOS(2),ecx	; advance to digit chars
	sub TOS(1),ecx	; adjust cnt
.x:;		;<single char
	ret
.l:
	mov ax,-64
	jr .i
; base praefix may follow after '\' praefix
.f:A; 16.b bis <.p>
	xor dword[byte uref uv.bits],mranu
	jc .u
	or dword[byte uref uv.bits],mquad
.u:
	mov eax,edx
	call .t
	jr .n
.p:;		;<not '0x..'
	dec ecx
.o:;		;<not sedec.
	cmp byte[ecx],'^'
	jz .l
	call .t
	js .f
.n:; check sign after praefix
	test edx,edx
	js .m
	shr eax,8
	cmp ah,'-'	; 2nd char
	jnz .m
	inc ecx
	not edx
	not dx
.m:
	mov eax,edx	; radix
	jmp disp .r
; find f.p. praefix and/or base modifier n
.t:A; ( p u xx -- p u n )
	push ecx	; count
	push edi	;reg
	lea edi,[kref npref-Y]; number base praefices
	mov edx,eax	; save base@ & sign in edx
	mov eax,[ecx]	; leading chars of stg
	shl eax,8	;
	mov al,ah
	mov ecx,npref.e-npref
	repnz scasb	; scan for praefix char
	movsx ax,byte[edi+nbase-npref-1];radix value
	pop edi		;reg
	pop ecx
	jnz .g		;?; none
	test al,al	; radix
	jz .q		;?; praefix not valid
	lea ecx,[ecx+1]	;advance count over praefix
	js .q		;?; float
	mov dx,ax	; pass radix
	dec al
	jnz .g
	mov dx,256	; -> ascii
.g:
	xor al,al
.q:
	ret
%endif;__NUMCONV_I
;-
;-=====================================================================;
;- l4numconv.inc <eof>
