;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4qdarith.inc
;-  created	: 04-aug-01
;-  modified	: 0.16.1+13 -re- "version.inc"
;-  version	: 0.01(+)
;-  switches	: __QDARITH_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __QDARITH_I
%define __QDARITH_I
%define qdarithinc
;-	
;-
;-----------------------------------------------------------------
;--- mixed quad and double
;-	

;-
;- umd*		( ud1 ud2 -- uq3 )				'umdmul'	forth
;-	unsigned multiply doubles to quad
umdmul:GA;:ok:
	push ebp,esi,ebx
	push eax	;[s] := h2
	mov ecx,TOS(2)	;ecx := h1
	mov esi,TOS(3)	;esi := l1
	mov ebx,TOS(1)	;ebx := l2
	mov eax,ebx	; a=l2
	mul esi		; l1*l2
	mov TOS(3),eax	;p3 result-ll
	mov eax,esi	; a=l1
	mov esi,edx	;esi := l1l2o
	mov edx,eax	; d=l1
	mov eax,[esp]	; a=h2
	mul edx		; l1*h2
	add esi,eax	; l1h2+l1l2o
	adc edx,0	;edx := l1h2o+cy
	mov eax,ebx	;eax := l2
	mov ebx,edx	;ebx := l1l2o+cy
	mul ecx		; h1*l2
	add esi,eax	;esi := l1h2+l1l2o+h1l2
	adc ebx,edx	;ebx := l1l2o+c+h1l2o
	sbb ebp,ebp	;-1 = Cy
	mov eax,ecx	;eax := h1
	neg ebp		;ebp := Cy
	pop edx		; h2
	mul edx		; h1h2
	add eax,ebx	;p1
	adc edx,ebp	;p0
	mov TOS(2),esi	;p2
	mov TOS(1),eax
	mov eax,edx
	pop ebp,esi,ebx
	ret
;-
;- md*		( d1 d2 -- q3 )			(4)		'mdmul'		forth
;-	signed multiply doubles to quad
mdmul:GA;:ok:
	mov ecx,eax
	xor ecx,TOS(2)
	push ecx	; final negation flag
	js .a		;?; d1 or d2 -ve
	test eax,eax
	jns .m		;?; d1 and d2 +ve
.a:
	call dabs
	call swap2
	call dabs
.m:
	call umdmul	; ( d1 d2 -- q3
	pop ecx
	test ecx,ecx
	js near qnegate
	ret
;-
;- uqd*		( uq1 ud1 -- uq2 ud2 )		(4)		'uqdmul'	forth
;-	unsigned multiply quad by double, ret six cells prod.
;	does{ 6 nndup 2drop uqsm* 5 nnswap drop sdrop uqsm* 0 -5 roll 6 n+ drop }
uqdmul:GA;:ok:;0,65s
	push eax
	push dword TOS(1); >rr
	call umdmul	;  ( q1l phl.phh
	call rot2	;  ( phl.phh q1l
	spush
	pop eax
	spush
	pop eax		;  ( phl.phh q1l d1
	call umdmul	;  ( phl.phh pll.plh
	call swap4	;  ( pll.plh phl.phh
	spush 
	mov eax,0	;  ( pll.plh phl.phh 00
	spush
	call mrot2
	jmp qplus
p1:GA;:TEST: in 'forth' voc
	jmp nimp
vpl:GA;:TEST: in 'rational' voc
	jmp nimp
;-
;- uq*		( q1 q2 -- q2 qovf )		(X)		'uqmul'		forth
;-	unsigned multiply quad by quad, ret quad prod & quad ovf = 8 cells product.
;-	does{ 8 nndup 3drop uqd* 6 nnswap drop 2sdrop uqd* 0. -3 2roll 8 n+ drop }
uqmul:GA;:ok:(mmx:0,7s, else 2,3s)
%ifdef __MMX
	mov [PS],eax
	movq mm3,TOS(7)	; -> q1l
	movq mm2,TOS(5)	; -> q1h
	movq mm1,TOS(3)	; -> q2l
	movq mm0,TOS(1)	; -> q2h
	movq TOS(5),mm1	; <- q2l
	spop 4
	call umdmul	; q.pll
	spush
	mov eax,0
	spush 5		;	( q.pll 00 xx xx
	movq TOS(1),mm0	; q2h
	movq TOS(3),mm3	; q1l	( q.pll 00 q1l q2h
	mov eax,[PS]
	call umdmul	; q.plh
	call qplus	; 6.plldplh
	sbb ecx,ecx
	spush
	neg ecx		; Cy = 1
	mov eax,0
	push ecx	; save C.lh
	spush 3		; 6.pll.plh xx xx
	movq TOS(1),mm1	; q2l	( 6.plldplh q1l q2l
	movq TOS(3),mm2	; q1h
	mov eax,[PS]
	call umdmul	; phl
	call qplus	; 6.qpLdphl
	adc dword[esp],0; accumulate C
	spush
	mov eax,0
	spush 5		; 6.qpLdphl 00 xx xx
	movq TOS(1),mm1	; q1h
	movq TOS(3),mm0	; 6.qpLdphl 00 q1h q2h
	mov eax,[PS]
	call umdmul
	pop ecx
	test ecx,ecx
	jz near qplus
	call qonep	; add half-carry
	jmp qplus	; 8.prod
%else
	call dup8	; q1l q1h q2l q2h q1l q1h q2l q2h
	call sdrop2	; q1l q1h q2l q2h q1l q1h q2h
	call uqdmul	; q1l q1h q2l q2h p1h.p1l.o1h		; quad by double.hi
	spush
	mov eax,0
	spush 2		; q1l q1h q2l q2h p1l.p1h.o1h.000
	mov eax,-3
	call roll2	; q1l q1h q2l q2h 000 p1l p1h o1h	; mul hi.prd by 2^64
	call swap8	; 000 p1l p1h o1h q1l q1h q2l q2h
	spop 2		; 000 p1l p1h o1h q1l q1h q2l
	call uqdmul	; 000 p1l p1h o1l p0l.p0h.o1l		; lo.prod
	spush
	mov eax,0
	spush		; 000 p1l p1h o1l p0l p0h ovf 000
	jmp oplus
%endif
;-
;- q* 	 	( q1 q2 -- q2 qovf )				'qmul'		forth
;-	signed multiply quad by quad, ret quad prod & quad ovf = 8 cells product.
qmul:GA;:ok:
	test eax,eax
	push eax
	jns .t
	call qnegate
.t:
	mov ecx,TOS(4)
	test ecx,ecx
	jns .d
	xor [esp],ecx
	call qnegaten
.d:
	call uqmul
	cmp dword[esp],byte 0
	lea esp,[esp+4]
	js near onegate
	ret
;
umqpow.2:A
	mov eax,0
	spush 3
	mov eax,ecx
	jmp qlshift
;-
;- umq^  	( u1 u2 -- uq3 )		((x))(4)	'umqpow'	forth
;-	unsigned raise sing u1 to the power of sing u2, leave quad result.
;-	max ca. 1,7*10^39, ret quad $ff..ff as overflow indicator.
umqpow:GA;:ok:
	mov ecx,eax
	dec ecx
	js .o		;?; d1^0, d1^(-ve)
	jz .e		;?; d1^1
	mov edx,TOS(1)
	cmp edx,byte 1
	jna .v		;?; 1^n2=1 or 0^n2=0; n2>0
	cmp edx,byte 2
	jz .2		;?; 2^n2 -> shift left n2 places
.p:
	push ecx	; exponent
	mov eax,1
	spush 2
	mov TOS(1),edx	; bignums
.l:A
	call nnover
	call nmul
	call ntoq
	mov ecx,eax
	cmp eax,byte 5
	jnc .x		;?; overflow, ret max
	mov eax,0
	spush
	lea eax,[ecx+1]	; assure unsigned
	dec dword[esp]
	jg .l
	lea esp,[esp+4]
	cmp eax,4
	jz .s
	spush
	mov eax,4
	call untoun
.s:
	spop
	jmp rotd2
.v:; 0^n2, 1^n2
	lea ecx,[eax-2]
	jc .o		;?; 0^n2 := 0
	xor ecx,ecx	;   1^n2 := 1
.o:; n1^(-ve), n1^0
	xor eax,eax
	mov TOS(1),eax
	inc ecx
	js .r		; d1^(-ve) := 0
	inc dword TOS(1); d1^0     := 1
.r:
	lea esp,[esp+4]
	spush 2
	ret
.e:; n1^1
	mov eax,ecx
	jr .r
.x:; ovf ret $ff..ff
	lea esp,[esp+4]
	add eax,byte 2
	call ndrop	; drop result and n1
	spush
	mov eax,-1
	spush 3
	ret
;-
;- qsqrt  	( q1 -- q2 n3 )			((x))(4)			forth
;-	ret +ve square root of |q1|, n3 := 1|-1 = sgn(q1), 1 if q1=0
;-	q1 = q2*q2*n3+c | -1 < c < q2, j = [-1,1]
;-	q2 is nearest integer the sqare of which is lower or equal |q1|*n3.
;-	NOTE: {qsqrt} is different vs {nsqrt}, {dsqrt} wrt flag n3 =/= 0.
qsqrt:GA;:ok:
;tests, v1-0-0+1 -> v1-0-1+0, free ram ca. 500M
;	qf   919191,9191 qvsqrt qf. \ 46,2s -> 27,5s
;	quad 919191,9191  qsqrt qf. \ 15,7s -> 11,4s
;input x, p := 0, a := 0
;loop 1 to #bits(x)/2
;	shift left x into a 2 places
;	shift left p one place
;	if a > 2*p then a:=a-(2*p+1); p:=p+1
;end
;return p
	mov ecx,eax
	sar ecx,31
	lea ecx,[ecx*2+1]
	push ebx
	push ebp
	push ecx	; save sign for "j"
	jns .p
	call qnegate
.p:
	spush
	mov eax,0
	spush 12	; p:=0, a:=0, t:=0, workspace cell @tos
	push_ibyte(63)	; bitcount/2
;tos+cells: 16)qx 12)qa 8)qp 4)qt 0)xx | j
; --
.l:A; shift left x into a 2 places ( 16)qx 12)qa 8)qp 4)qt 0)xx -- qx/4 qa' qp xx
	mov eax,TOS(9)	; a.hh
	mov ecx,TOS(10)	; a.hl
	shld eax,ecx,2
	mov TOS(9),eax
	mov eax,ecx
	mov ecx,TOS(11)
	shld eax,ecx,2
	mov TOS(10),eax
	mov eax,ecx
	mov ecx,TOS(12)
	shld eax,ecx,2
	mov TOS(11),eax
	mov eax,ecx
	mov ecx,TOS(13)
	shld eax,ecx,2
	mov TOS(12),eax
	mov eax,ecx
	mov ecx,TOS(14)
	shld eax,ecx,2
	mov TOS(13),eax
	mov eax,ecx
	mov ecx,TOS(15)
	shld eax,ecx,2
	mov TOS(14),eax
	mov eax,ecx
	mov ecx,TOS(16)
	shld eax,ecx,2
	mov TOS(15),eax
	shl ecx,2
	mov TOS(16),ecx
; shift left p one place ;	( 16)qx 12)qa 8)qp 4)qt 0)xx -- qx/4 qa' qp*2 qp*4 xx
	mov eax,TOS(5)	; p.hh
	mov ecx,TOS(6)	; p.hl
	mov edx,eax	; copy for t
	shld eax,ecx,1	; p:=2p
	mov TOS(5),eax
	shld edx,ecx,2	; t:=4p
	mov TOS(1),edx
	mov ebx,edx	; t.hh
	mov eax,ecx
	mov ecx,TOS(7)
	mov edx,eax
	shld eax,ecx,1
	mov TOS(6),eax
	shld edx,ecx,2
	mov TOS(2),edx
	mov eax,ecx
	mov ebp,edx	; t.hl
	mov ecx,TOS(8)
	mov edx,eax
	shld eax,ecx,1
	mov TOS(7),eax
	shld edx,ecx,2
	mov TOS(3),edx	; t.lh
	shl ecx,1
	mov TOS(8),ecx
	shl ecx,1
	mov TOS(4),ecx	; t.ll
;tos+cells: 16)qx 12)qa 8)qp 4)qt 0)xx | j
; if a > t then a:=a-(t+1); p:=p+1
; u> if ...
;tos+cells: 16)qx 12)qa 8)qp 4)qt 0)xx | j
	cmp ebx,TOS(9)
	jnz .x
	cmp ebp,TOS(10)
	jnz .x
	cmp edx,TOS(11)
	jnz .x
	cmp ecx,TOS(12)
.x:
	jnc .c		;?;
;tos+cells: 15)qx 11)qa 7)qp 3)qt | j
	not ecx		; t:=-(t+1)
	not edx
	not ebp
	not ebx
; 15)qx 11)qp 7)-(t+1) 3)qa
	add TOS(12),ecx	; a:=a+t=a-(4*p+1)
	adc TOS(11),edx
	adc TOS(10),ebp
	adc TOS(9),ebx
	add dword TOS(8),byte 1; p:=p+1
	adc dword TOS(7),byte 0
	adc dword TOS(6),byte 0
	adc dword TOS(5),byte 0
.c:
	dec dword[esp]
	jns disp .l
;tos+cells: 16)qx 12)qa 8)qp 4)qt 0)xx | j
	mov eax,TOS(5)	; copy p
	mov ebp,TOS(6)
	mov ecx,TOS(7)
	mov edx,TOS(8)
	TOSPULL(12)
	lea esp,[esp+4]	; drop counter
	mov TOS(1),eax	; result to @tos
	mov TOS(2),ebp
	pop eax		; q1 sign cell
	pop ebp
	pop ebx
	mov TOS(3),ecx
	mov TOS(4),edx
	ret
;-
;- uqm/mod	( +q1 ud2 -- ud3 uq4 )		(X)		'uqmbym'	forth
;-	unsigned divide quad +q1 by ud2, ret quad quotient uq4, double remainder ud3.
;-	div 0: ret ud3:=0, uq4:=-1; fast exit if dvd = 1
;-	(w/ shortcut to {uqd/mod} if qot would fit into 2 cells)
uqmbym:GA;:ok:;<qfmbym,uqbym)
; c: ecx,edx,mm0..3
; filter double qot result:
;  1) significant dvd = 2 cells, 2) dvd = 3 and dvs = 2 significant cells
	mov ecx,TOS(1)
	mov edx,eax
	or edx,ecx
	jz .z		;?; div 0
	dec edx
	jnz .c		;?; not div 1
	dec ecx
	jz .e		;?; div 1
.c:
	mov edx,TOS(2)	; dvd.hh
	mov ecx,TOS(3)	; dvd.hl
	or ecx,edx	; m.s. dvd double
	jz .d		;?; qot not above double size
	test edx,edx	; dvd.hh
	jnz .q		;?; uncertain
	test eax,eax	; dvs.h
	jnz .d		;?; qot not above double size
.q:
%ifdef __MMX;:ok:
	mov [PS],eax
	pxor mm3,mm3	;3: #0
	movq mm1,TOS(3)	;1: qh
	movq mm0,TOS(5)	;0: ql
	movq mm2,TOS(1)	;2: d
	movq TOS(3),mm3	; #0->qh
	movq TOS(5),mm1	; qh->ql
	call uqdbym	; -> drh, dqh
	TOSPUSH(2)
	movq mm3,TOS(5)	;rm->
	movq mm1,TOS(3)	; save qot-h
	movq TOS(1),mm2	;d
	movq TOS(5),mm0	;ql
	movq TOS(3),mm3	;rm<-
	mov eax,[PS]	;dh
	call uqdbym	; ql rm d -- drl dql
	TOSPUSH(2)
	movq TOS(1),mm1	;qot-h
	mov eax,[PS]
	ret
%else;:ok:
	call dup4	; qd1l qd1h ud2 ud1h ud2
	spush 2		; qd1l qd1h ud2 ud1h ud2 dxx
	mov edx,0
	mov eax,TOS(2)	; d2h<-
	mov ecx,TOS(3)	; d2l
	mov TOS(2),edx
	mov TOS(3),edx	; q1h' := qd1h|00
	mov TOS(1),ecx	; d2l<-
	call uqdbym	; qd1l qd1h ud2 drh dqh
	mov ecx,TOS(1)	; dqhl
	mov edx,TOS(8)	; qdlh
	mov TOS(6),eax	;			; qdhh<-	; save dqh
	mov TOS(7),ecx	; qd1l dqh: ud2 drh dqh	; dqhl<-
	mov eax,TOS(4)	; d2h<-
	mov ecx,TOS(5)	; d2l
	mov TOS(4),edx	; qdlh<-
	mov edx,TOS(9)	; qdll
	mov TOS(1),ecx	; d2l<-
	mov TOS(5),edx	; qd1l dqh: qdl: drh ud2:
	call uqdbym	; qdll dqh  drl  dql
	call swap4	; drl  dql  qdll dqh 
	jmp sdrop2	; drl  dql  dqh 
%endif
.e:A	; dvs = 1 -> rmd:=0, qot:=q1
	mov TOS(1),ecx
	jmp mrot2
.z:A	; dvs = 0, ret ff.. ff..
	mov TOS(5),eax
	mov TOS(4),eax
	dec eax
; -1 = eax to TOS(1,2,3), ret
	mov TOS(3),eax
	mov TOS(2),eax
	mov TOS(1),eax
	ret
.d:A; quad in range of a double
	call uqdbym
	spush
	mov eax,0
	spush
	ret
;-
;- ud/		( ud1 ud2 -- ud3 )				'udby'		forth
;-	unsigned divide doubles, leave qot
udby:GA;:ok:
	lea edx,[kref sdrop2-Y]
	push edx
;-
;- ud/mod 	( ud1 ud2 -- ud3 ud4 )				'udbym'		forth
;-	unsigned divide doubles, leave qot ud4 & rmd ud3
udbym:GA;:ok:
	mov ecx,eax	; h2
	mov eax,0
	mov edx,TOS(1)	; l2
	mov TOS(1),eax	; h1l := 0, h1h := 0
	spush
	mov eax,edx	; l2
	spush
	mov eax,ecx	; h2
;-
;- uqd/mod	( uq1 ud2 -- ud3 ud4 )				'uqdbym'	forth
;-	unsigned divide quad by double, ret rmd ud3, quot ud4
;-	div 0: ret ud4=qot := -1., ud3=rmd := uq1.lo
;-	(bit-wise div. from F-PC, R.L.Smith)
uqdbym:GA;:ok:
	push ebp,ebx,esi,edi; @tor := PS
	mov ebp,eax	;B: dvs	h
	mov edx,TOS(1)	;d:	l
	mov eax,TOS(2)	;a: dvd	hh
	mov ebx,TOS(3)	;b:	hl
	mov esi,TOS(5)	;S:	ll
	mov edi,TOS(4)	;D:	lh
	cmp ebp,eax	;sh-dhh
	jc .s		;?; hi dvs > dvd
	jz .m		;?; lo dvs > dvd ?
.g:
	mov ecx,64
	or eax,eax
.o:A; -----------------.
	rcl esi,1
	rcl edi,1
	rcl ebx,1
	rcl eax,1
	jnc .d		;?; -.
.i:A
	sub ebx,edx	;hl-l
	sbb eax,ebp	;hh-h
	stc
	dec ecx
	jnz .o		;?; -'
	jr .n		;?; q:=2q+1, done -.
.m:A; ---< sh=dhh ---'
	cmp edx,ebx	;sl-dhl
	ja .g		;?; -'
.s:; ----< sh>dhh ---' ret q:=ffffffff, r=dlh|dll -?-
	mov ebx,esi	;ll
	mov esi,-1
	mov eax,edi	; rh = rmd = dvd.lo
	mov edi,esi	; qh = qot = -1
	jr .r		; - ret ovf
.d:A; -----------------'
	cmp eax,ebp	; hh'-h
	ja .i		;?; -'
	jc .v		;?; -.
	cmp ebx,edx
	jnc .i		;?; -'
.v:; ------------------'
	or eax,eax
	dec ecx
	jnz .o		;?; -'
.n:
	rcl esi,1	; ql
	rcl edi,1	; qh
.r:
	mov ecx,edi	; qh
	pop edi
	mov TOS(5),ebx	; rl
	mov TOS(4),eax	; rh
	mov TOS(3),esi	; ql
	mov TOS(2),ecx
	TOSPULL(2)
	pop ebp,ebx,esi
	mov [PS],ecx
	mov eax,ecx
	ret
;-
;- df/mod  	( d1 d2 -- d3r d2q )				'dfbym'		forth
;-	floored division of doubles, double remainder and double quotient
dfbym:GA;:ok:
	mov ecx,TOS(2)
	mov edx,TOS(1)
	sar ecx,31
	mov TOS(1),ecx
	xchg eax,ecx
	spush
	mov eax,edx
	spush
	mov eax,ecx	;:continued:
;-
;- qdfm/mod	( q1 d2 -- d3r d2q )				'qdfmbym'	forth
;-	floored division of quad by double, double remainder and double quotient
;-	 Dividend Divisor   Quotient Remainder
;-	 -------- ------- | -------- ---------
;-	     10      7    |    1        3
;-	    -10      7    |   -2        4
;-	     10     -7    |   -2       -4
;-	    -10     -7    |    1       -3
;-	 dvs = 0 overflow values:
;-		dvd > 0    $7fff.ffff  $7fffffff
;-		dvd < 0    $8000.0001  $80000001
;-		dvd = 0        0.0001          0
qdfmbym:GA;:ok:
	mov ecx,eax
	mov edx,eax
	xor ecx,TOS(2)
	or edx,TOS(1)
	jz .z		;?; div-0
	push eax	; sign of rmd
	push ecx	; sign of qot
	call mrot2	; save dvd for -ve qot floored adjustmt
	call qabs
	call third2	; dvs dvd dvs
	call dabs
	call uqdbym	; uqd/mod -> dr dq
	pop ecx		; sgn(q)
	test ecx,ecx
	jns .p		;?; +ve quot
	not dword TOS(1); 2not
	not eax		; qot := -qot-1
	call mrot2
	pop edx		; sgn(r)
	test edx,edx
	js .r		;?; +ve rmd
	call dnegate	; r := -r
.r:
	call dplus	; r := -(r-dvs)
	jmp swap2
.p:A
	pop edx		; sgn(r)
	test edx,edx
	jns .d		;?; +ve rmd
	neg dword TOS(3)
	adc dword TOS(2),byte 0
	neg dword TOS(2)
.d:
	jmp rotd2
.z:A; div-0
	sar ecx,31
	jnz .o		;?; -ve ovf
	or eax,TOS(2)
	or eax,TOS(3)
	or eax,TOS(4)
	or eax,TOS(5)
	jnz .o		;?; +ve ovf
	inc eax		; 0/0
	mov TOS(5),ecx
	mov TOS(4),ecx
	mov TOS(3),eax
	xor eax,eax	; ret 1,0
	TOSPULL(2)
	ret
.o:A;+ve: q;=r:=7fff.ffff -ve: q:=r:=8000.0001
	not ecx		; 0> 2* 1+ $80000000 + 
	lea ecx,[ecx*2+0x80000001]
	TOSPULL(2)
	mov eax,ecx
	mov al,ah	; 7fffffff or 80000000
	mov TOS(2),eax
	shl ecx,1
	sar ecx,1	; ffffffff or 00000001
	mov TOS(1),ecx
	mov TOS(3),ecx
	ret
;-
;- qfm/mod	( q1 d2 -- d3r q2q )		((x))		'qfmbym'	forth
;-	floored division of quad by double, leave double remainder and quad quotient
qfmbym:GA;:ok:
	mov ecx,eax
	mov edx,eax
	xor ecx,TOS(2)
	or edx,TOS(1)
	jz .z
	push eax	; sign of rmd
	push ecx	; sign of qot
	call mrot2	; save dvd for -ve qot floored adjustmt
	call qabs
	call third2	; dvs dvd dvs
	call dabs
	call uqmbym	; uqm/mod -> dr dq
	pop ecx		; sgn(q)
	test ecx,ecx
	jns .p		;?; +ve quot
	not dword TOS(3); 2not
	not dword TOS(2)
	not dword TOS(1)
	not eax		; qot := -qot-1
	call swap4
	pop edx		; sgn(r)
	test edx,edx
	js .r		;?; +ve rmd
	call dnegate	; r := -r
.r:
	call dplus	; r := -(r-dvs)
	jmp mrot2
.p:A
	pop edx		; sgn(r)
	test edx,edx
	jns .d		;?; +ve rmd
	neg dword TOS(5)
	adc dword TOS(4),byte 0
	neg dword TOS(4)
.d:
	spush
	mov eax,3
	call roll2
	spop 2
	ret
.z:A
	call qdfmbym.z
	spush 2		; drm dql dqh
	jz .o		;?; 0/0
	mov ecx,TOS(3)
	mov cl,ch
	mov TOS(1),ecx
	mov TOS(2),ecx
.o:
	ret
;-	
;-
;-----------------------------------------------------------------
;--- mixed quad and sing
;-	

;-
;- uqsm*  	( q1 n1 -- q2 ovf )		(f6)		'uqsmmul'	forth
;-	unsigned multiply quad by sing, leave quintuple prod, i.e. q2 & overflow
uqsmmul:GA;:ok:
	mov ecx,eax
	mov eax,TOS(4)
	mul ecx			; ll
	mov TOS(4),eax		;   p.lll
	mov eax,TOS(3)
	mov TOS(3),edx		; 3:p.ll.h
	mul ecx
	add TOS(3),eax		; 3:p.llh
	adc edx,byte 0		;cy
	sbb eax,eax
	neg eax
	push eax		;cy 3->2
	mov eax,TOS(2)
	mov TOS(2),edx
	mul ecx
	add TOS(2),eax
	pop eax
	adc edx,eax
	sbb eax,eax
	neg eax
	push eax
	mov eax,TOS(1)
	mov TOS(1),edx
	mul ecx
	add TOS(1),eax
	pop eax
	adc eax,edx		; ovf
	ret
;-
;-
;- uqsm/  	( q1 n2 -- n3 q4 )		(f6)		'uqsmby'	forth
;-	unsigned divide quad by sing, leave sing rmd n3 & quad quot.
;-	div by zero ret quot and rmd w. all bits set, i.e. max value.
uqsmby:GA;:ok:
	test eax,eax
	mov ecx,eax
	jz .o
	mov edx,0
	mov eax,TOS(1)	;dhh
	div ecx
	push eax	; qhh
	mov eax,TOS(2)	;dhl
	div ecx
	mov TOS(1),eax	; qhl
	mov eax,TOS(3)	;dlh
	div ecx
	mov TOS(2),eax	; qlh
	mov eax,TOS(4)	;dll
	div ecx
	mov TOS(3),eax	; qll
	mov TOS(4),edx	;rmd
	pop eax
	ret
.o:A
	dec eax
	mov TOS(4),eax
	mov TOS(3),eax
	mov TOS(2),eax
	mov TOS(1),eax
	ret
;-
;-
;-----------------------------------------------------------------
;--- quad
;-

;-	

;-
;- qgcd		( q1 q2 -- q3 )			(4)		'qgcd'		forth
;-	returns the greatest common denominator q3 of the absolute value of two signed 
;-	quads. If either number is zero, returns the absolute value of the other one.
qgcd:GA;:ok:
	call qabs	; abs
	call swap4	; swap
	call qabs	; abs
	jr .e		; ENTER
.b:;			; BEGIN
	call sover4	;  swap over
	call uqbym
	spop 4		;  umod
.e:;			; ENTRY
	mov ecx,eax	;  -dup
	or ecx,TOS(1)	;  0=
	or ecx,TOS(2)
	or ecx,TOS(3)
	jnz .b		; until
	spop 4		; ;s
	ret
; tuq/mod, special cases
uqbym.z:A; test dvd =/= 0, ret rmd=0, qot=max; 0/0 -> ret q=1,r=0
	spop 4
	call qsgn
	spush
	sar eax,31
	spush 3
	jr uqbym.pn
uqbym.p:A; q div 1, ret qot=dvd, rmd=0
	TOSPULL(3)
uqbym.pn:
	call uqbym.zn
	jmp swap4
uqbym.n:A; 0 div q, ret qot=0, rmd=0
	TOSPULL(7)
	call uqbym.zn
	spush
uqbym.zn:
	mov eax,0
	spush 3
	ret
uqbym.g:A; dvd < dvs, ret qot=0, rmd=dvd; dvd @tos -!-
	call swap4
	TOSPULL(3)
	jr uqbym.zn
uqbym.e:A; dvd=dvs, ret qot=1, rmd=0
	TOSPULL(7)
	call uqbym.zn
uqbym.u:
	spush
uqbym.un:
	mov eax,1
	spush
	dec eax
	spush 2
	ret
;-
;- uq/mod 	( uq1 uq2 -- uqr uqq )	((x))			'uqbym'		forth
;-	unsigned div quad uq1 by quad uq2, leave quad remainder (@nos) and quotient.
;-	div 0: ret uqr := -1, uqq = -1; fast exit if uq2=dvd = 1
uqbym:GA;:ok: 0,17s - error w/ -ve ovf
	call qqzeq
	spop
	jz disp .z	;?; dvd=0, dvs/0
	call dup4
	call qonem
	call qzeq	; 4dup q1- q0=
	spop
	jz disp .p	;?; dvd=1, dvs/1
	call swap4
	call qqzeq	; 4swap ?q0=
	spop
	jz disp .n	;?; dvs=0, 0/dvd
	call quqcmp	; dvd dvs f
	test eax,eax
	spop
	js disp .g	;?; dvs>dvd
	jz disp .e	;?; dvs=dvd
	call qslbits	; dvd free bits
	push eax
	call qlshift	;
	call swap4	; ( dvd dvs --
	call qslbits	; dvs free bits
	mov ecx,eax
	sub ecx,[esp]	; scaling bitcount
	inc ecx
	add [esp],ecx	; bitcount to restoring remainder
	push ecx	; qot bitcount
	call qlshift	; shift 'up' by free bits of dvs
.o:
	spush
	mov eax,0
	spush 3
	call mrot4	; ( q z n'
	jr .t
%define _uqbymc_
;%undef   _uqbymc_ ; no advantage!
.l:A; ( qot dvd dvs -- qot rmd xxx )
%ifdef _uqbymc_
	call qmul2t	; 3rd @tos		qot := qot*2
	call quqcmp	; 1 if q1 u< q2, dvd < dvs
	test eax,eax	;C:-1,Z:0;A:1
	spop
	jg .b		;?; dvs > dvd
	mov ecx,TOS(3)	; q 7)z 3)n
	mov edx,TOS(2)
	sub TOS(7),ecx
	mov ecx,TOS(1)
	sbb TOS(6),edx
	sbb TOS(5),ecx
	sbb TOS(4),eax	; q z' n, z' = z-n	dvd := dvd-dvs
	call q1pt	; q' z' n, inc 3rd @tos	qot := qot+1
.b:
	call qmul2n	; q' z' n'		dvd' := dvd*2
%else
	shl dword TOS(11),1
	rcl dword TOS(10),1
	rcl dword TOS(9),1
	rcl dword TOS(8),1
	cmp eax,TOS(4)
	mov edx,TOS(1)
	ja .b		;?; dvs > dvd
	jc .c
	cmp edx,TOS(5)
	mov ecx,TOS(2)
	ja .b
	jc .c
	cmp ecx,TOS(6)
	mov edx,TOS(3)
	ja .b
	jc .c
	cmp edx,TOS(7)
	ja .b
.c:
	mov ecx,TOS(3)	; q 7)z 3)n
	mov edx,TOS(2)
	sub TOS(7),ecx
	mov ecx,TOS(1)
	sbb TOS(6),edx
	sbb TOS(5),ecx
	sbb TOS(4),eax	; q z' n, z' = z-n	dvd := dvd-dvs
	add dword TOS(11),byte 1
	adc dword TOS(10),byte 0
	adc dword TOS(9),byte 0
	adc dword TOS(8),byte 0
.b:
	shl dword TOS(7),1
	rcl dword TOS(6),1
	rcl dword TOS(5),1
	rcl dword TOS(4),1
%endif
.t:;  entry
	dec dword[esp]
	jns disp .l
	lea esp,[esp+4]
	spop 3
	pop eax
	call qrshift	; rmd
	jmp swap4	; ( ... -- r q )
;-
;- qmod   	( q1 q2 -- qr )		((x))			'qmod'		forth
;-	signed div quad by quad, leave quad remainder. -re- {q/mod}
qmod:GA;:ok:
	call qbym	; q/mod 4sdrop
	spop 4
	ret
;-
;- q/	 	( q1 q2 -- qq )		((x))			'qby'		forth
;-	signed div quad by quad, leave quad quotient. -re- {q/mod}
qby:GA;:ok:
	call qbym	; q/mod 4drop
	jmp sdrop4
;-
;- q/mod  	( q1 q2 -- qr qq )	((x))			'qbym'		forth
;-	signed div quad by quad, leave quad remainder (@nos) and quotient.
;-	'round near zero' mode, sgn(qot) = sgn(dvd xOR dvs), sgn(rmd) = sgn(dvd).
qbym:GA;:ok:
; +/+ q+,r+ ; +/- q-,r+ ; -/- q+,r- ; -/+ q-,r-
	test eax,eax
	mov edx,eax
	jns .s		;?; +ve dvd
	push eax	; sg(rmd)
	call qnegate	; abs(dvd)
	pop edx
.s:
	mov ecx,TOS(4)
	xor edx,ecx
	test ecx,ecx
	push ecx	; sg(rmd)
	push edx	; sg(qot)
	jns .d		;?; +ve dvd
	call qnegaten	; abs(dvs)
.d:
	call uqbym	; div
	pop ecx
	test ecx,ecx
	jns .q		;?; +ve qot
	call qnegate
.q:
	pop ecx
	test ecx,ecx
	jns .r		;?; +ve rmd
	call qnegaten
.r:
	ret
;-
;- qf/mod  	( q1 q2 -- qr qq )				'qfbym'		forth
;-	floored, signed div quad by quad, leave quad remainder (@nos) and quotient.
;-	signed 'round near zero' division but, if qq < 0 ret qq:=qq-1, qr:=qr+q2.
qfbym:GA;:ok:
	call sover4	; 4swap 4over
	call qbym	; q/mod
	test eax,eax	; ( 11:dvs 7:rmd 3:qot
	jns .r		;?; +ve qot no adjustmt
	call qonem	; qot := qot-1
	mov ecx,TOS(11)
	mov edx,TOS(10)
	add TOS(7),ecx
	adc TOS(6),edx	; rmd := rmd + dvs
	mov ecx,TOS(9)
	mov edx,TOS(8)
	adc TOS(5),ecx
	adc TOS(4),edx
.r:
	call rot4	; drop dvs
	spop 4
	ret
;-
;- 8not		( o1 -- o2 )					'not8'		hidden
;-	q2 := one-s complement of eight-cells item q1
not8:GA;:ok:
	not dword TOS(3)
	not dword TOS(2)
	not dword TOS(1)
	not eax		;:C:
;-
;- nos-4not	( q1 qx -- q2 qx )				'not4n'		hidden
;-	one's cpl of quad q1 @nos. (double ranum support)
not4n:GA;:ok:
	not dword TOS(7)
	not dword TOS(6)
	not dword TOS(5)
	not dword TOS(4)
	ret
;-
;- nos-qnegate	( q1 qx -- q2 qx )				'qnegaten'	hidden
;-	negate quad q1 @nos. (double ranum support)
qnegaten:GA;:ok:<qv2uqvq, ranum)
	not dword TOS(7)
	not dword TOS(6)
	not dword TOS(5)
	not dword TOS(4);:C:
;-
;- nos-q1+	( q1 qx -- q2 qx )				'q1pn'		hidden
;-	increment quad @nos by one. (double ranum support)
q1pn:GA;:ok:
	add dword TOS(7),byte 1
	adc dword TOS(6),byte 0
	adc dword TOS(5),byte 0
	adc dword TOS(4),byte 0
	ret
;-
;- 3rd-q1+	( q1 qx qy -- q2 qx qy )			'q1pt'		hidden
;-	increment quad 3rd @tos by one. (uq/mod support)
q1pt:GA
	add dword TOS(11),byte 1
	adc dword TOS(10),byte 0
	adc dword TOS(9),byte 0
	adc dword TOS(8),byte 0
	ret
;-
;- nos-q2/	( uq1 qx -- uq2 qx )				'qdiv2n'	hidden
;-	unsigned divide q1 @nos by 2.
qdiv2n:GA;:ok:
	sar dword TOS(4),1
	rcr dword TOS(5),1
	rcr dword TOS(6),1
	rcr dword TOS(7),1
	jnc .r
	test eax,eax
	js q1pn
.r:
	ret
;-
;- nos-q1-	( q1 qx -- q2 qx )				'q1mn'		hidden
;-	dec @nos quad q1 by one. (double ranum support)
q1mn:GA;:ok:
	sub dword TOS(7),byte 1
	sbb dword TOS(6),byte 0
	sbb dword TOS(5),byte 0
	sbb dword TOS(4),byte 0
	ret
;-
;- qsm+		( q1 n1 -- q2 )			(e)		'qsmplus'	forth
;-	add sign extended sing n1 to quad q1
qsmplus:GA;:ok:
	spush
	sar eax,31
	spush 2		;:c: in qplus
;-
;- q+		( q1 q2 -- q3 )			(e)		'qplus'		forth
;-	add two quad numbers
qplus:GA;:ok:
	mov ecx,TOS(3)
	mov edx,TOS(2)
	add TOS(7),ecx
	adc TOS(6),edx
	mov ecx,TOS(1)
	adc TOS(5),ecx
	adc TOS(4),eax
	spop 4
	ret
;-
;- o+		( o1 o2 -- o3 )			(e)		'oplus'		hidden
;-	add two 8-cells numbers. 'ranum' support.
;-	(e: eflags from m.s. cells addition
oplus:GA;:ok:
	mov ecx,TOS(7)
	mov edx,TOS(6)
	add TOS(15),ecx
	adc TOS(14),edx
	mov ecx,TOS(5)
	mov edx,TOS(4)
	adc TOS(13),ecx
	adc TOS(12),edx
	mov ecx,TOS(3)
	mov edx,TOS(2)
	adc TOS(11),ecx
	adc TOS(10),edx
	mov ecx,TOS(1)
	adc TOS(9),ecx
	adc TOS(8),eax
	spop 8
	ret
;-
;- q-		( q1 q2 -- q3 )			(e)		'qsub'		forth
;-	subtract quad q2 from quad q1
qsub:GA;:ok:
	mov ecx,TOS(3)
	mov edx,TOS(2)
	sub TOS(7),ecx
	sbb TOS(6),edx
	mov ecx,TOS(1)
	sbb TOS(5),ecx
	sbb TOS(4),eax
	spop 4
	ret
;-
;- o-		( o1 o2 -- o3 )			(e)		'osub'		hidden
;-	subtract 8-cells o2 number from o1. 'ranum' support
;-	(e: eflags from m.s. cells subtraction
osub:GA
	mov ecx,TOS(7)
	mov edx,TOS(6)
	sub TOS(15),ecx
	sbb TOS(14),edx
	mov ecx,TOS(5)
	mov edx,TOS(4)
	sbb TOS(13),ecx
	sbb TOS(12),edx
	mov ecx,TOS(3)
	mov edx,TOS(2)
	sbb TOS(11),ecx
	sbb TOS(10),edx
	mov ecx,TOS(1)
	sbb TOS(9),ecx
	sbb TOS(8),eax
	spop 8
qdiv2.r:
	ret
;-
;- onegate	( o1 -- o2 )					'onegate'	hidden
;-	negate eight-cells item
onegate:GA
	call not8	;:continued: {o1+}
;-
;- o1+		( o1 -- o2 )					'o1p'		hidden
;-	increment eight-cells item by one
o1p:GA;:ok:
	call q1pn
	adc dword TOS(3),byte 0
	jr qonep.n
;-
;- q2/		( q1 -- q2 )					'qdiv2'		forth
;-	signed divide q1 by 2. 'round towards zero' mode.
;-	NOTE: correct result, including odd -ve figures (thus NOT ANS compliant).
qdiv2:GA;:ok:
	sar eax,1
	rcr dword TOS(1),1
	rcr dword TOS(2),1
	rcr dword TOS(3),1
	jnc .r
	test eax,eax
	jns .r		;:continued: {q1+}
;-
;- q1+		( q1 -- q2 )					'qonep'		forth
;-	increment quad by one
qonep:GA;:ok:
	add dword TOS(3),byte 1
.n:
	adc dword TOS(2),byte 0
	adc dword TOS(1),byte 0
	adc eax,byte 0
	ret
;-
;- q1-		( q1 -- q2 )					'qonem'		forth
;-	decrement quad by one
qonem:GA;:ok:
	sub dword TOS(3),byte 1
	sbb dword TOS(2),byte 0
	sbb dword TOS(1),byte 0
	sbb eax,byte 0
	ret
;-
;- uq2/		( uq1 -- uq2 )					'uqdiv2'	forth
;-	unsigned divide q1 by 2.
uqdiv2:GA;:ok:
	shr eax,1
	rcr dword TOS(1),1
	rcr dword TOS(2),1
	rcr dword TOS(3),1
	ret
;-
;- nos-uq2/	( uq1 qx -- uq2 qx )				'uqdiv2n'	hidden
;-	unsigned divide q1 @nos by 2.
uqdiv2n:GA;:ok:
	shr dword TOS(4),1
	rcr dword TOS(5),1
	rcr dword TOS(6),1
	rcr dword TOS(7),1
	ret
;-
;- 3rd-uq2/	( uq1 qx qy -- uq2 qx qy )			'uqdiv2t'	hidden
;-	unsigned divide q1 3rd @tos by 2.
uqdiv2t:GA
	shr dword TOS(8),1
	rcr dword TOS(9),1
	rcr dword TOS(10),1
	rcr dword TOS(11),1
	ret
;-
;- qrshift	( q1 n -- q2 )							forth
;-	shift quad q1 right by +ve n bits, i.e. unsigned floored div by 2^n.
;-	no overflow result, q2 := 0 if unsigned n > 127
qrshift:GA;:ok:
; o:edx := ovf
	mov edx,eax
	mov dword [PS],0
	shr edx,5	; bits/cell
	mov ecx,eax
	jz .c		;?; no cells
	push esi
.l:
	lea esi,[PS-12]
	lea edi,[PS-16]
	dec edx
	movsd
	movsd
	movsd
	movsd
	jg .l
	pop esi
.c:
	lea PS,[PS-4]
	and ecx,byte 31
	mov edx,TOS(3)	; (3)
	jng .r
	mov eax,TOS(2)
	shrd edx,eax,cl	; [d][a][.][.]
	mov TOS(3),edx
	mov edx,eax	; (2)
	mov eax,TOS(1)
	shrd edx,eax,cl	; [3][d][a][.]
	mov TOS(2),edx
	mov edx,eax	; (1)
	mov eax,TOS0
	shrd edx,eax,cl	; [3][2][d][a]
	mov TOS(1),edx
	mov edx,0
	shrd edx,eax,cl	; save ovf
	shr eax,cl
	mov TOS0,eax
.r:
	mov eax,TOS0
	ret
;-
;- qlshift	( q1 n -- q2 )							forth
;-	shift quad q1 left by +ve n bits, i.e. mul by 2^n.
;-	no overflow result, q2 := 0 if unsigned n > 127
qlshift:GA;:ok:; 55s
; o:edx := ovf
	mov ecx,eax
	lea eax,TOS(1)
	sub ecx,byte 32
	js .c		;?; none, check bits
	std
	push esi
.l:
	mov edi,eax	; m.s.cell
	lea esi,[eax-4]
	sub ecx,byte 32
	movsd
	movsd
	movsd
	mov dword[edi],0
	jns .l
	pop esi
	cld
.c:
	add ecx,byte 32
	mov PS,eax	; ptr to m.s. cell
	jng .r
	mov eax,TOS0	; hh
	mov edx,TOS(1)	; hl
	shld eax,edx,cl	; hh<<hl
	mov TOS0,eax	;hh
	mov eax,TOS(2)	; lh
	shld edx,eax,cl	; hl<<lh
	mov TOS(1),edx	;hl
	mov edx,TOS(3)
	shld eax,edx,cl	; lh<<ll
	mov TOS(2),eax	;lh
	shl edx,cl
	mov TOS(3),edx	;ll
.r:
	mov eax,TOS0
	ret
;-
;- qslbits	( q1 -- q1 n )							forth
;-	ret no. n of bits by which +/- q1 can be shifted left w/o change of sign.
qslbits:GA;:ok:+/-
	mov edx,PS
.e:; <qvsqrt, 2nd)
	spush
.n:; <qvsqrt, 1st)
	mov ecx,0
	mov eax,30	; 63 bits max, less sign bit, less start bsr:0 vs. shift:1
	cmp ecx,[edx]	; @@tos
	jg .s		;?; -ve
	jnz .b		;?; +ve, non-zero
	lea edx,[edx-4]
	lea eax,[eax+32]
	cmp ecx,[edx]
	jnz .b
	lea edx,[edx-4]
	lea eax,[eax+32]
	cmp ecx,[edx]
	jnz .b
	lea edx,[edx-4]
	lea eax,[eax+32]
	cmp ecx,[edx]
	lea ecx,[ecx+1]	; prevent -1 if no bits set
	jz .r		;?; all zero
.b:
	bsr ecx,[edx]
	sub eax,ecx
.r:
	ret
.s:; -ve, count m.s. set bits
	lea eax,[eax+128-30]
	lea edx,[edx-12]
.l:
	dec eax
	js .x		;?; all bits set
	bt [edx],eax
	jc .l		;?; next bit
.x:
	neg eax
	lea eax,[eax+126]
	ret
;-
;- nos-q2*	( q1 qx -- q2 qx )		(e)(m,0)	'qmul2n'	hidden
;-	multiply q1 @nos by 2.
;-	(e: C if ovf
qmul2n:GA;:ok:
	shl dword TOS(7),1
	rcl dword TOS(6),1
	rcl dword TOS(5),1
	rcl dword TOS(4),1
	ret
;-
;- 3rd-q2*	( q1 qx qy -- q2 qx qy )	(e)(m,0)	'qmul2t'	hidden
;-	multiply q1 3rd @tos by 2.
;-	(e: C if ovf
qmul2t:GA
	shl dword TOS(11),1
	rcl dword TOS(10),1
	rcl dword TOS(9),1
	rcl dword TOS(8),1
	ret
;-
;- q2*		( q1 -- q2 )			(e)(m,0)	'qmul2'		forth
;-	multiply q1 by 2.
;-	(e: C if ovf
qmul2:GA M;:ok:
	shl dword TOS(3),1
	rcl dword TOS(2),1
	rcl dword TOS(1),1
	rcl eax,1
M qmul2
q1abs.r:
	ret
;-
;- q1abs  	( q1 -- ~q1|q1 )						forth
;-	ret 1s-complement if q1 -ve
q1abs:GA;:ok:
	test eax,eax
	jns .r		;:continues:
;-
;- 4not		( q1 -- q2 )					'not4'		forth
;-	convert quad to one-s complement
not4:GA;:ok:
	not dword TOS(3)
	not dword TOS(2)
	not dword TOS(1)
	not eax
qabs.r:
	ret
;-
;- qabs		( q1 -- -q1|q1 )				'qabs'		forth
;-	negate if q1 -ve
qabs:GA;:ok:
	test eax,eax
	jns qabs.r	;:continues:
;-
;- qnegate  	( q1 -- q2 )			(4)				forth
;-	negate quad (2-s complement)
;-	q2 := q1 if only signbit of q1 set.
qnegate:GA;:ok:
	not dword TOS(3)
	not dword TOS(2)
	not dword TOS(1)
	not eax
	add dword TOS(3),byte 1
	adc dword TOS(2),byte 0
	adc dword TOS(1),byte 0
	adc eax,byte 0
	ret
;-
;- q+-		( q1 n2 -- q3 )					'qpm'		forth
;-	apply sign of n2 to q1, i.e. negate q if n < 0
qpm:GA;:ok:
	test eax,eax
	spop
	js qnegate
	ret
;-
;- ?uqcmp	( uq1 uq2 - uq1 uq2 f )		(e)		'quqcmp'	forth
;-	unsigned compare quads, ret f: 0 if q1=q2, -1 if q2 u< q1, 1 if q2 u> q1.
;-	(e: Z if q2=q1, C if @tos u< @nos; ~C*NS*NZ if @tos u> @nos, ecx=sg(q1-q2)
quqcmp:GA;:ok:
	mov ecx,eax
	sub ecx,TOS(4)	; sign of uq1 uq2 -
	mov edx,TOS(1)
	spush
	mov eax,-1
	jnz .c
	cmp edx,TOS(6)
	mov edx,TOS(3)
	jnz .c
	cmp edx,TOS(7)
	mov edx,TOS(4)
	jnz .c
	cmp edx,TOS(8)
	jz .p		; ret 0
.c:
	jc .r		; ret -1
	inc eax		; ret 1
.p:
	inc eax
.r:
	ret
;-
;- q=		( q1 q2 -- flg )		(e)(4)		'qeq'		forth
;-	tf if equal quads
;-	(e: Z if q1 = q2, ecx=sign cell of { q1 q2 - }
qeq:GA;:ok:
	call qsub	;:continues:
;-
;- q0=		( q -- flg )			(e)(4)		'qzeq'		forth
;-	tf if quad q = 0
;-	(e: Z if q = 0, ecx=sign cell
qzeq:GA;:ok:
	call qqzeq
	TOSPULL(4)
	ret
;-
;- o0=		( o -- flg )			(e)(4)		'ozeq'		hidden
;-	tf if eigth-cells int o = 0
;-	(e: Z if o = 0, ecx=sign cell
ozeq:GA
	call qqzeq
	TOSPULL(4)
	jnz .r
	spop
	call qqzeq
.r:
	TOSPULL(4)
	ret
;-
;- ?q0=		( q -- q flg )			(e)		'qqzeq'		forth
;-	non-destructively test quad for zero, tf if quad q = 0
;-	(e: Z if q = 0, ecx = sign cell | NZ*(ecx<0): q<0, NZ*((ecx>0)+(ecx=0)): q>0
qqzeq:GA;:ok:
; c: edx
	mov ecx,TOS(3)
	or ecx,TOS(2)
	or ecx,TOS(1)
	or ecx,eax
	mov ecx,eax	; sign
	spush
	mov eax,-1
	jz .x
	lea eax,[eax+1]	; preserve flags
.x:
	ret
;-
;- qsgn		( q -- 1|0|-1 )			(e)(4)				forth
;-	ret signed unit of |q|
;-	(e: Z if q=0, ecx=sign cell
qsgn:GA;:ok:
; c: ecx,edx; eflags: NZ
	call qzeq
	sar ecx,31
	inc eax
	jz .r		;?; 0
	lea eax,[ecx*2+1]
.r:
	ret
;-
;- q0>		( q -- flg )			(e)(4)		'qzgreat'	forth
;-	ret true flag if quad q > 0
;-	(e: NZ if q>0
qzgreat:GA;:ok:
; c: ecx,edx
	call qsgn
	jz .r		;?; 0 -> ff
	neg eax
	sar eax,1	; +1 -> 0
.r:
	ret
;-
;- qu<		( uq1 uq2 -- flg )		(e)		'quless'	forth
;-	ret true flag if quad uq1 < uq2
;-	(e: ret ^C*^Z if q1 u< q2, Z if q1=q2
quless:GA;:ok:
	cmp TOS(4),eax
	mov ecx,TOS(5)
	jnz .x
	cmp ecx,TOS(1)
	mov ecx,TOS(6)
	jnz .x
	cmp ecx,TOS(2)
	mov ecx,TOS(7)
	jnz .x
	cmp ecx,TOS(3)
.x:
	mov eax,0
	TOSPULL(7)
	jnc .r
	not eax
.r:
	ret
;-
;- qu>		( uq1 uq2 -- flg )		(e)		'qugreat'	forth
;-	ret true flag if quad uq1 < uq2
;-	(e: ret ^C*^Z if q1 u> q2 (tf), Z if q1=q2
qugreat:GA;:ok:
	cmp TOS(4),eax
	mov ecx,TOS(5)
	jnz .x
	cmp ecx,TOS(1)
	mov ecx,TOS(6)
	jnz .x
	cmp ecx,TOS(2)
	mov ecx,TOS(7)
	jnz .x
	cmp ecx,TOS(3)
.x:
	mov eax,0
	TOSPULL(7)
	jna .r
	not eax
.r:
	ret
;-
;- q<		( q1 q2 -- flg )				'qless'		forth
;-	ret true flag if quad q1 signed < q2
qless:GA;:ok:
	call qsub	;:continues:
;-
;- q0<		( q -- flg )					'qzless'	forth
;-	ret true flag if quad q1 is -ve
qzless:GA;:ok:
	TOSPULL(3)
	sar eax,31
	ret
;-
;- -4dup  	( q1 | 0  -- q1 q1 | 0 )			'qdup4'		forth
;-	 duplicate quad @tos if =/= zero.
qdup4:GA;:ok:
	mov ecx,eax
	or ecx,TOS(3)
	jnz dup4
	or ecx,TOS(2)
	or ecx,TOS(1)
	jz .r
;-
;- 4dup		( q1 -- q1 q1 )			(x) 		'dup4'		forth
;-	duplicate quad
dup4:GA;:ok:
%ifdef __MMX;:ok: 1,5*
	mov [PS],eax
	movq mm0,TOS(3)
	movq mm1,TOS(1)
	TOSPUSH(4)
	movq TOS(3),mm0
	movq TOS(1),mm1
	mov eax,[PS]
	ret
%endif
;- 4dup		( q1 -- q1 q1 )			 		'dup4f'		non-mmx
dup4f:GA;:ok: non-mmx float-voc alternative
	mov ecx,TOS(1)
	mov edx,TOS(2)
	push dword TOS(3)
	spush 4
	pop dword TOS(3)
	mov TOS(2),edx
	mov TOS(1),ecx
qdup4.r:
	ret
;-
;- 4drop  	( q1 q2 -- q1 )			(m,5)		'drop4'		forth
;-	drop quad cells
drop4:GA M;:ok:
	spop 4
M drop4
	ret
;-
;- 8drop  	( o1 o2 -- o1 )			(m,5)		'drop8'		forth
;-	drop top eight cells
drop8:GA M
	spop 8
M drop8
	ret
;-
;- 8sdrop 	( o1 o2 -- o2 )					'sdrop8'	forth
sdrop8:GA
%ifdef __MMX;:ok: 2*
	movq mm3,TOS(7)
	movq mm2,TOS(5)
	movq mm0,TOS(3)
	movq mm1,TOS(1)
	movq TOS(15),mm3
	movq TOS(13),mm2
	movq TOS(11),mm0
	movq TOS(9),mm1
	TOSPULL(8)
	ret
%endif
;-
;- 8sdrop 	( q1 q2 -- q2 )					'sdrop8f'	non-mmx
sdrop8f:GA;:ok:
	call swap8
	spop 8
	ret
;-
;- 8swap  	( qv1 qv2 -- qv1 qv2 )				'swap8'		forth
;-	{ 8 nswap }, 8 cells per item swap
swap8:GA
	spush
	mov eax,8
	jmp nswap
;-
;- 4swap  	( q1 q2 -- q2 q1 )		(x) 		'swap4'		forth
;-	swap quads
swap4:GA;:ok:
%ifdef __MMX;:ok: 2*
	movq mm3,TOS(7)
	mov [PS],eax
	movq mm2,TOS(5)
	movq mm1,TOS(3)
	movq mm0,TOS(1)
	movq TOS(7),mm1
	movq TOS(1),mm2
	movq TOS(5),mm0
	mov eax,[PS]
	movq TOS(3),mm3
	ret
%endif
;- 4swap  	( q1 q2 -- q2 q1 )		 		'swap4f'	non-mmx
swap4f:GA;:ok:
; c: ecx,edx
	push eax
	push dword TOS(1)
	push dword TOS(2)
	push dword TOS(3)
	mov eax,TOS(5)
	mov ecx,TOS(6)
	mov edx,TOS(7)
	mov TOS(1),eax
	mov TOS(2),ecx
	mov TOS(3),edx
	mov eax,TOS(4)
	pop dword TOS(7)
	pop dword TOS(6)
	pop dword TOS(5)
	pop dword TOS(4)
	ret
;-
;- 4sdrop 	( q1 q2 -- q2 )					'sdrop4'	forth
sdrop4:GA
%ifdef __MMX;:ok: 2*
	movq mm0,TOS(3)
	movq mm1,TOS(1)
	movq TOS(7),mm0
	movq TOS(5),mm1
	TOSPULL(4)
	ret
%endif
;-
;- 4sdrop 	( q1 q2 -- q2 )					'sdrop4f'	non-mmx
sdrop4f:GA;:ok:
	call swap4f
	spop 4
	ret
;-
;- uqmax   	( uq1 uq2 -- uq3 )		((X))				forth
;-	ret uq3 := (uq1+uq2+|uq1-uq2|)/2
uqmax:GA;:ok:
	call dup8
	call qugreat	; (the faster one)
	spop
	jc sdrop4
	spop 4
	ret
;-
;- uqmin   	( uq1 uq2 -- uq3 )		((X))				forth
;-	ret uq3 := (uq1+uq2-|uq1-uq2|)/2
uqmin:GA;:ok:
	call dup8
	call qugreat	; (the faster one)
	spop
	ja sdrop4
	spop 4
	ret
;-
;- 8dup		( n1..n8 -- n1..n8 n1..n8 )	(x)		'dup8'		forth
;-	duplicate top eight cells
dup8:GA;(for ranum)
%ifdef __MMX;:ok: 2*
	mov [PS],eax
	movq mm3,TOS(7)
	movq mm2,TOS(5)
	movq mm1,TOS(1)
	movq mm0,TOS(3)
	TOSPUSH(8)
	movq TOS(7),mm3
	movq TOS(5),mm2
	movq TOS(1),mm1
	movq TOS(3),mm0
	ret
%endif
;-
;- 8dup		( ra1 ra2 -- ra1 r2 r1 r2 )			'dup8f'		non-mmx
dup8f:GA;(for ranum)
	call over4f
%ifdef __MMX
	jr over4f
%endif			;else :continued:
;-
;- 4over  	( q1 q2 -- q1 q2 q1 )		(x) 		'over4'		forth
;-	duplicate 2nd on stack quad to @tos
;-	( mmx mode w/ minor timing gain )
over4:GA;:ok:
%ifdef __MMX
	mov [PS],eax
	movq mm0,TOS(7)
	movq mm1,TOS(5)
	TOSPUSH(4)
	movq TOS(3),mm0
	movq TOS(1),mm1
	mov eax,[PS]
	ret
%endif
;- 4over  	( q1 q2 -- q1 q2 q1 )		 		'over4f'	non-mmx
over4f:GA;:ok:
	spush
	mov eax,TOS(8)
	spush
	mov eax,TOS(8)
	spush
	mov eax,TOS(8)
	spush
	mov eax,TOS(8)
	ret
;-
;- 4sover  	( q1 q2 -- q2 q1 q2 )		(x)		'sover4'	forth
;-	does{ 4swap 4over }
sover4:GA;:ok: q1 q2 -- q2 q1 q2 ) = d1l d1h d2l d2h -- d2l d2h d1l d1h d2l d2h
;	;  	                     7   5   3   1      11  9   7   5   3   1 
%ifdef __MMX
	TOSPUSH(4)
	mov TOS(4),eax	; q2hh
	movq mm0,TOS(11); d1l
	movq mm1,TOS(9)	; d1h
	movq mm2,TOS(7)	; d2l
	movq mm3,TOS(5)	; d2h
	movq TOS(7),mm0	; d1l<-
	movq TOS(5),mm1	; d1h<-
	movq TOS(9),mm3
	movq TOS(1),mm3	; d2h<-
	movq TOS(11),mm2; d2l<-
	movq TOS(3),mm2
	ret
%endif
;-
;- 4sover  	( q1 q2 -- q2 q1 q2 )				'sover4f'	non-mmx
;-	does{ 4swap 4over }
sover4f:GA
	call swap4f
	jmp over4f
;-
;- 4rot   	( q1 q2 q3 -- q2 q3 q1 )	(x)		'rot4'		forth
;-	roll 3rd quad to @tos
rot4:GA;:ok:; q1 q2 q3 = d1l d1h d2l d2h d3l d3h -- d2l d2h d3l d3h d1l d1h
;	;	         11  9   7   5   3   1      11  9   7   5   3   1  
%ifdef __MMX
	movq mm0,TOS(11); d1l
	movq mm1,TOS(9)	; d1h
	movq mm2,TOS(7)	; d2l
	movq mm3,TOS(5)	; d2h
	mov edx,TOS(2)	; d3lh
	mov ecx,TOS(3)	; d3ll
	mov TOS(4),eax	; d3hh->
	mov TOS(6),edx	; d3lh->
	mov eax,TOS(1)	; d3hl
	mov TOS(7),ecx	; d3ll->
	mov TOS(5),eax	; d3hl->
	movq TOS(1),mm1	; d1h->
	movq TOS(3),mm0	; d1l->
	movq TOS(11),mm2; d2l->
	movq TOS(9),mm3	; d2h->
	mov eax,[PS]	; d1hh<-
	ret
%endif
;-
;- 4rot   	( q1 q2 q3 -- q2 q3 q1 )			'rot4f'		non-mmx
;-	roll 3rd quad to @tos
rot4f:GA;:ok:
	spush
	mov eax,5
	call rot2f
	spush
	mov eax,5
	jmp rot2f
;-
;- 4rotd   	( q1 q2 q3 -- q2 q3 )		(x)		'rot4d'		forth
;-	fast aequivalent to { 4rot 4drop }.
rot4d:GA
%ifdef __MMX
	mov [PS],eax
	movq mm2,TOS(7)	; q2l
	movq mm3,TOS(5)	; q2h
	movq mm0,TOS(3)	; q3l
	movq mm1,TOS(1)	; q3h
	movq TOS(11),mm2; q2l
	movq TOS(9),mm3	; q2h
	movq TOS(7),mm0	; q3l
	movq TOS(5),mm1	; q3h
	TOSPULL(4)
	ret
%endif
;-
;- 4rotd   	( q1 q2 q3 -- q2 q3 )				'rot4df'	non-mmx
;-	non-mmx aequivalent to { 4rot 4drop }.
rot4df:GA
	call rot4f
	spop 4
	ret
;-
;- 8sover 	( qv1 qv2 -- qv2 qv1 qv2 )	(x+x)		'sover8'	forth
;-	{ qvswap qvover }, 8 cells per item swap and over
sover8:GA;		; ( 15:qv1 7:qv2 -- 23:qv2 15:qv1 7:qv2
%ifdef __MMX
	mov TOS(0),eax
	movq mm0,TOS(15); qv1
	movq mm1,TOS(13)
	movq mm2,TOS(11)
	movq mm3,TOS(9)
	movq mm4,TOS(7)	; qv2
	movq mm5,TOS(5)
	movq mm6,TOS(3)
	movq mm7,TOS(1)
	TOSPUSH(8)	; -- 23:qv2 15:qv1 7:qv2
	movq TOS(23),mm4; 
	movq TOS(21),mm5
	movq TOS(19),mm6
	movq TOS(17),mm7
	movq TOS(15),mm0; qv1
	movq TOS(13),mm1
	movq TOS(11),mm2
	movq TOS(9),mm3
	movq TOS(7),mm4	; 
	movq TOS(5),mm5
	movq TOS(3),mm6
	movq TOS(1),mm7
	ret
%endif
;-
;- 8sover 	( qv1 qv2 -- qv2 qv1 qv2 )			'sover8f'	non-mmx
;-	{ qvswap qvover }, 8 cells per item swap and over
sover8f:GA
	call dup8	; qv1 qv2 qv2
	spush
	mov eax,-5
	call roll4
	spush
	mov eax,-5
	jr roll4
;-
;- 4-rot  	( q1 q2 q3 -- q3 q1 q2 )	(x)		'mrot4'		forth
;-	reverse roll quad @tos to 3rd
mrot4:GA;:ok:; q1 q2 q3 = d1l d1h d2l d2h d3l d3h -- d3l d3h d1l d1h d2l d2h
;	;	           11  9   7   5   3   1      11  9   7   5   3   1  
%ifdef __MMX
	movq mm0,TOS(11); d1l
	movq mm1,TOS(9)	; d1h
	movq mm2,TOS(7)	; d2l
	movq mm3,TOS(5)	; d2h
	mov edx,TOS(2)	; d3lh
	mov ecx,TOS(3)	; d3ll
	mov TOS(8),eax	; d3hh->
	mov TOS(10),edx	; d3lh->
	mov eax,TOS(1)	; d3hl
	mov TOS(11),ecx	; d3ll->
	mov TOS(9),eax	; d3hl->
	movq TOS(5),mm1	; d1h->
	movq TOS(7),mm0	; d1l->
	movq TOS(3),mm2	; d2l->
	movq TOS(1),mm3	; d2h->
	mov eax,[PS]	; d1hh<-
	ret
%endif
;-
;- 4-rot  	( q1 q2 q3 -- q2 q3 q1 )	(4)		'mrot4f'	non-mmx
;-	reverse roll quad @tos to 3rd
mrot4f:GA;:ok:
	spush
	mov eax,-2	;:continued:
;-
;- 4roll  	( qn .. q1 q0 +n -- .. qn-1 .. q1 q0 qn ) (4)	'roll4'		forth
;-		( qn .. q1 q0 -n -- .. q0 qn .. q1 )
roll4:GA;:ok:
	mov ecx,eax
	sar eax,31
	shl eax,1
	lea eax,[eax+ecx*2+1]
	push eax
	call roll2
	spush
	pop eax
	jmp roll2
;-
;- 8roll  	( on .. o1 o0 +n -- .. on-1 .. o1 o0 on ) (4)	'roll8'		forth
;-		( on .. o1 o0 -n -- .. o0 on .. o1 )
roll8:GA
	mov ecx,eax
	sar eax,30
	or eax,byte 1
	lea eax,[eax+ecx*2]
	push eax
	call roll4
	spush
	pop eax
	jr roll4
;-
;- 8over  	( n1..n8 n9..n16 -- n1..n8 n9..n16 n1..n8 )(x)	'over8'		forth
;-	copy 2nd @tos eight cells
over8:GA;:ok:
%ifdef __MMX
	mov [PS],eax
	movq mm3,TOS(15)
	movq mm2,TOS(13)
	movq mm1,TOS(9)
	movq mm0,TOS(11)
	TOSPUSH(8)
	movq TOS(7),mm3
	movq TOS(5),mm2
	movq TOS(1),mm1
	movq TOS(3),mm0
	mov eax,[PS]
	ret
%endif
;-
;- 8over  	( n1..n8 n9..n16 -- n1..n8 n9..n16 n1..n8 )	'over8f'	non-mmx
;-	copy 2nd @tos eight cells
over8f:GA;(for ranum)
	spush
	mov eax,2	;:continued:
;-
;- 8pick  	( ..on..o1 o0 +n -- ..on..o1 o0 on )	(4)	'pick8'		forth
;-		( ..on..o1 o0 oo -n -- ..oo..o1 o0 )
;-	n>0: fetch (+n)-th item to tos. index n = 0 @tos.
;-	n<0: store @tos item to (-n)-th position, index 0 refers to item to store.
pick8:GA;:ok: +/-
	test eax,eax
	jns .p
; 'pick back'
	not eax
	lea eax,[eax*2+1]
	push eax
	call pick4.n
	spush
	pop eax
	jr pick4.n
.p:
	lea eax,[eax*2+1]
	push eax
	call pick4.p
	spush
	pop eax
	jr pick4.p
;-
;- 4pick  	( ..qn..q1 q0 +n -- ..qn..q1 q0 qn )	(4)	'pick4'		forth
;-		( ..qn..q1 q0 qq -n -- ..qq..q1 q0 )
;-	n>0: fetch (+n)-th item to tos. index n = 0 @tos.
;-	n<0: store @tos item to (-n)-th position, index 0 refers to item to store.
pick4:GA;:ok:
	test eax,eax
	js .m
.p:; <pick8)
	lea eax,[eax*2+1]
	push eax
	call pick2
	spush
	pop eax
	jmp pick2
.m:; 'pick back'
	not eax
.n:; <pick8)
	lea eax,[eax*2+2]
	push eax
	call mpick2
	spush
	pop eax
	jmp mpick2
;-
;- 4@		( p -- uq )					'fetch4'	forth
;-	fetch quad at ptr p
fetch4:GA;:ok:
; c: ecx,edx - no mmx because of alignment requirements!
	push eax
	mov edx,[kref eax+12]
	mov ecx,[kref eax+8]
	mov eax,[kref eax]
	spush
	pop eax
	mov eax,[kref eax+4]
	spush
	mov eax,ecx
	spush
	mov eax,edx
	ret
;-
;- 4!		( q ptr -- )					'store4'	forth
;-	store quad @nos to address pointed to by @tos
store4:GA;:ok:
; c: ecx,edx - no mmx because of alignment requirements!
	lea ecx,[kref eax]
	mov eax,TOS(1)
	mov edx,TOS(2)
	mov [ecx+12],eax
	mov [ecx+8],edx
	mov eax,TOS(3)
	mov edx,TOS(4)
	mov [ecx+4],eax
	mov [ecx],edx
	spop 5
	ret
%endif;__QDARITH_I
;-
;-=====================================================================;
;- l4qdarith.inc <eof>
