;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4rational.inc
;-  created	: 27-jun-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __RANUM_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th", double & quad rational numbers ("floats")
;-		: words, re file l4sranum.inc for single ranum words.
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: rational numbers by ordered cells' pairs
;-  source	: Bronstein Semendjajev, Taschenbuch der Mathematik
;-		: Forth Scientific Library Algorithm #46
;-		: Htte Bd.1, Theoretische Grundlagen
;-		: bj, and others (re asm source).
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;
; -- quellen u.a.
;	Barkhausen, Elektronen-Rhren, Hirzel Leipzig 1937.3-1954.6 (5 Bd.)
;	Bronstein Semendjajev, Taschenbuch der Mathematik, 1980.19 (Nauka Moskau 1979)
;	Graetz, Die Elektrizitt, Union Dt. Verlagsg. Stuttgart 1928.23
;	Htte - Des Ingenieurs Taschenbuch, Bd I Theoretische Grundlagen, Bln 1955.28
;	Mach, Die Mechanik, Leipzig 1933.9 (Repro Wiss. Buchges. Darmstadt 1982)
;	Schmidt, Die Mathematik des Funktechnikers, Frankh Stuttg 1940 (Rechenfehler!)
;	Wallot, Schwachstromtechnik, Springer Bln 1948.5
;	Zurmhl, Praktische Mathematik, Springer Bln 1965.5
;    nebenbei,
;	eine recht hbsche Einordnung von "Experten" (welcher ich mich ohne Vorbehalt
;	anschliee) gibt Ernst Mach, im Schluwort zur "Mechanik"...
;-	
;-	

%ifndef __RANUM_I
%define __RANUM_I
%define ranuminc

;:NOTE: - to do
;	ovf flags -1,0 -ve & 1,0 +ve, leave 0,0 as 1,0

;-
;- rational	( -- )				(X)(i)				root
;-	vocabulary for Rational Numbers opr ('real', 'ranum', 'vulgar numbers').
;-	
;-	The "rational" word-set provides basic arithmetic, logical, type conversion 
;-	and numerical in-/output routines for rational numbers. Basic routines from
;-		Vulgar Maths Words Version FSL1.1", Forth Scientific Library 
;-		Algorithm #46, by Gordon Charlton - gordon@charlton.demon.co.uk.
;-	adapted to 'lib4th' and (much) extended.
;-	(Find the original FSL text in lib4th source, "./doc/src/vulgar.fth").
;-	
;-	'sing' "v", 'double' "dv", 'quad' "qv" refer to a 'ranum' of the rsp size 
;-	of ea, the numerator and denominator; prefix 'v...' from 'vulgar'.
;-	
;-	A 'ranum' in this context is a single or multiple cells pair in the order
;-		Numerator Denominator -- , denominator @tos (nenner, divisor), 
;-	which is (almost) always a reduced fraction. The numerator is a signed integer, 
;-	and the denominator a non-negative integer. Zero is represented by (0,1).
;-	Numbers too large to be represented are indicated by the special value (0,0),
;-	which will propagate through a program. Most words in this suite are tolerant
;-	of the overflow indicator, with the exception of those that convert rational 
;-	numbers to other numerical data types. These are flagged in the listing.
;-	
;-	The data type "rational number" is also referred to by "fractional number", 
;-	if displayed as an integer and fraction or, by "floating point" number. 
;-	The term "ranum" will be used for either display type of a rational number.
;-	
;-	In the range of improper fractions (1,..) the difference between successive 
;-	representable 'ranums' increases as the absolute value of the fraction 
;-	increases; i.e. above |2^(bitsize-2)-1| only integers are representable.
;-	Range of single cell pairs  'ranums' is thus quite limited, double and quad 
;-	pairs sizes provided for more accurate ranges and intermediate calculations.
;-	
;-	Further, -re- {vsplit} is provided, which separates the fractional and the 
;-	integral part of a rational fraction. If the loss of precision associated 
;-	with large numbers is not acceptable vsplit can be used to deal with the 
;-	fractional component separately, thus maintaining absolute precision.
rational:GA;:ok:
	call dovoc
	dd Irational
;-
;- hidden 	( -- )				(X)(i)		'rahidden'	rational
;-	{rational} sub-vocabulary for Rational Numbers, quad format and support.
rahidden:GA;:ok:
	call dovoc
	dd Irahidden
;-
;- cons   	( -- )				(X)(i)		'racons'	rational
;-	{rational} sub-vocabulary, Rational Numbers' constants.
racons:GA;:ok:
	call dovoc
	dd Iracons
;-
;- float  	( -- )				(X)(i)				ans
;-	ans FLOATING and FLOATING EXT words, aliased/copied from -re- {rational} voc.
;-	'float' representation is 'double ranum', common 4-cells operations apply, 
;-	'double float' by quad cells pairs 'quad ranum' w/ 8-cells basic operations.
;-	-re- {flt}/{int} for persistent floats/integers numeric input modes.
float:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ifloat
;-
;-	
;--- constants, by cells pairs e.g. for rational numbers and/or {*/} etc.
;-

;-
;- 0dv		( -- dv )					'v0dv'		rational cons
;-	cons double ranum 0
v0dv:GA;:ok:
	spush
	mov eax,0
	spush 2
	inc eax
	spush
	dec eax
	ret
;-
;- 0qv		( -- dv )					'v0dv'		rational cons
;-	cons quad ranum (double float) 0
v0qv:GA;:ok:
	spush
	mov eax,0
	spush 4
	inc eax
	spush
	dec eax
	spush 2
	ret
;-
;- 1dv		( -- dv )					'v1dv'		rational cons
;-	cons double ranum 1
v1dv:GA;:ok:
	call v0dv
	inc dword TOS(3)
	ret
;-
;- 1qv		( -- qv )					'v1qv'		rational cons
;-	cons quad ranum 1
v1qv:GA;:ok:
	call v0qv
	inc dword TOS(7)
	ret
;-
;- -1dv		( -- dv )					'vm1dv'		rational cons
;-	cons double ranum -1
vm1dv:GA;:ok:
	spush
	mov eax,-1
	spush 2
	add eax,byte 2
	spush
	dec eax
	ret
;-
;- -1qv		( -- qv )					'vm1dv'		rational cons
;-	cons quad ranum -1
vm1qv:GA;:ok:
	call v0qv
	jmp q1mn
;-
;- 2^(1/2)	( -- num den )					'casqr2n2'	rational cons
;-	1.41421356(2..)	+/- 1,5e-9
casqr2n2:GA;:ok:
	spush
	mov eax,707106781	;19601
	spush
	mov eax,500000000	;13860
	ret
;- 2^(1/12)	( -- num den )					'casqr12n2'	rational cons
;-	1,059...	+/- 1,0e-9
casqr12n2:GA;:ok:
	spush
	mov eax,26797
	spush
	mov eax,25293
	ret
;- 3^(1/2)	( -- num den )					'casqr2n3'	rational cons
;-	1,732050808...	+/- 0,5e-9
casqr2n3:GA;:ok:
	spush
	mov eax,216506351	;18817
	spush
	mov eax,125000000	;10864
	ret
;- 10^(1/2)	( -- num den )					'casqr2n10'	rational cons
;-	3,162...	+/- 5,7e-9
casqr2n10:GA;:ok:
	spush
	mov eax,22936
	spush
	mov eax,7253
	ret
;- bernulli(v)	( n -- v )							rational cons
;-	1st eleven Bernulli numbers by index 0<n<12 (?, "Bernulli'sche Zahlen")
bernulliv:GA;:ok:
	mov edx,[kref bernullit-Y+4+eax*8-8]
	mov eax,[kref bernullit-Y+eax*8-8]
	spush
	mov eax,edx
	ret
;- c		( -- num den )					'cac'		rational cons
;-	2,997...	+/- 1,6e-9
cac:GA;:ok:
	spush
	mov eax,24559
	spush
	mov eax,8192
	ret
;-
;- dv.e		( -- dv )					'caed'		rational cons
;-	2,7182818284(59) +/- 10^-12 (ref. Bronstein)
caed:GA;:ok:
	spush
	mov eax,3862497387
	spush
	mov eax,632
	spush
	mov eax,3567587328
	spush
	mov eax,232
	ret
;- dv.e^6	( -- dv )					'cae6d'		rational cons
;-	e to the power of six, ret dv := 403,42879349
cae6d:GA;:ok:
	spush
	mov eax,1688173685
	spush
	mov eax,9
	spush
	mov eax,100000000
	spush
	mov eax,0
	ret
;- dv.e^10	( -- dv )					'cae10d'	rational cons
;-	e to the power of ten, ret dv := 22026,4657948
cae10d:GA;:ok:
	spush
	mov eax,0xa133b253
	spush
	mov eax,0x162de54e
	spush
	mov eax,0x9cc31fa4
	spush
	mov eax,0x41fd
	ret
;- euler-c	( -- num den )					'caec'		rational cons
;-	0,57721566(4901532)...
caec:GA;:ok:
	spush
	mov eax,34491032
	spush
	mov eax,59754151
	ret
;- euler(v)	( n -- v )							rational cons
;-	1st seven Euler numbers by index 0<n<8 (?: "Euler'sche Zahlen")
eulerv:GA;:ok:
	lea eax,[eax-1]
	and eax,byte 7	; get safe
	mov eax,[kref eulert-Y+eax*4]
	spush
	mov eax,1
	ret
;- k(t)		( -- num den )					'cakptp'	rational cons
;-	1,64676  CORDIC trigonometric correction factor (by 15 iterations)
cakptp:GA;:ok:
	spush
	mov eax,41169
	spush
	mov eax,25000
	ret
;- k(h)		( -- num den )					'cakphp'	rational cons
;-	0,8281593  CORDIC hyperbolic correction factor (by 15 iterations)
cakphp:GA;:ok:
	spush
	mov eax,8281593
	spush
	mov eax,10000000
	ret
;- lg(2)  	( -- num den )					'calg2'		rational cons
;-	0,301029995(...)
calg2:GA;:ok:
	spush
	mov eax,188143747
	spush
	mov eax,625000000
	ret
;- lg(e)  	( -- num den )					'calge'		rational cons
;-	0,434294481(...)
calge:GA;:ok:
	spush
	mov eax,428781055
	spush
	mov eax,987304867
	ret
;- ln(2)  	( -- num den )					'caln2'		rational cons
;-	0,69314718(...)
caln2:GA;:ok:
	spush
	mov eax,34657359
	spush
	mov eax,50000000
	ret
;- dv.ln(2)  	( -- dv )					'cdvln2'	rational cons
;-	0,69314718 +/- 10^11
cdvln2:GA;0,69314718055
	spush
	mov eax,2492567714
	spush
	mov eax,826086822 
	spush
	mov eax,1204250187
	spush
	mov eax,1191791362
	ret
;- ln(10) 	( -- num den )					'caln10'	rational cons
;-	2,3025850930..
caln10:GA;:ok:
	spush
	mov eax,1006604908
	spush
	mov eax,437162957
	ret
;- dv.ln(10) 	( -- dv )					'cdvln10'	rational cons
;-	2,3025850930.. +/- 10^11
cdvln10:GA;2,30258509299..
	spush
	mov eax,-1593222911
	spush
	mov eax,889193778
	spush
	mov eax,1723559992
	spush
	mov eax,386171951
	ret

;- pi		( -- num den )					'cacpi'		rational cons
;-	3,1415926536...	+/- 10^-10
capi:GA;:ok:;			;  10^-10						;10^-8	;10^-2
	spush
	mov eax,1723588251	;461704645	;1043835	;62831853	;103993	;355	;22
	spush
	mov eax,548635180	;146965153	;332263		;20000000	;33102	;113	;7
	ret
;- qv.2pi	( -- q2 )					'cqv2pi'	rational cons
;-	quad ranum cons 2*pi 6,2831853072...	+/- 10^-10 (in range of a sing)
cqv2pi:GA;:ok:;
	spush
	mov ecx,0
	mov eax,1723588251
	spush
	mov eax,ecx
	spush 3
	mov eax,274317590
	spush
	mov eax,ecx
	spush 2
	ret
;- qv.2/3pi	( -- q2 )					'cqv23pi'	rational cons
;-	quad ranum cons 2*pi/3 2,09439510240...	+/- 10^-10 (in range of a sing)
cqv23pi:GA;:ok:;
	spush
	mov ecx,0
	mov eax,1723588251
	spush
	mov eax,ecx
	spush 3
	mov eax,822952770
	spush
	mov eax,ecx
	spush 2
	ret
;- qv.pi/2	( -- q2 )					'cqvpi2'	rational cons
;-	quad ranum cons pi/2 1,5707963268...	+/- 10^-10 (in range of a sing)
cqvpi2:GA;:ok:;
	spush
	mov ecx,0
	mov eax,1723588251
	spush
	mov eax,ecx
	spush 3
	mov eax,1097270360
	spush
	mov eax,ecx
	spush 2
	ret
;- qv.pi/4	( -- q2 )					'cqvpi4'	rational cons
;-	quad ranum cons pi/4 0,7853981634...	+/- 10^-10 (in range of a double)
;-	pi/4 = Sa(n->oo) (-1)^(n+1)/(2n-1); n=1...k|k>>0, after 2 days: 0,785398164328770
cqvpi4:GA;:ok:;
	spush
.n:; <atg(qv) if qv=1)
	mov ecx,0
	mov eax,1723588251
	spush
	mov eax,ecx
	spush 3
	mov eax,2194540720
	spush
	mov eax,ecx
	spush 2
	ret
;- qv.pi/180	( -- q2 )					'cqvpi180'	rational cons
;-	quad ranum cons pi/180 0,0174532925...	+/- 10^-10 (in range of a double)
cqvpi180:GA;:ok:;
	spush
	mov ecx,0
	mov eax,574529417
	spush
	mov eax,ecx
	spush 3
	mov eax,2853339728
	spush
	lea eax,[ecx+7]
	spush
	mov eax,ecx
	spush
	ret
;-
;- qvrdv  	( -- qv )					'qvmdv'		rational cons
;-	max quad to {qvround}ding a quad to size of a double ranum.
qvmdv:GA;:ok:
	spush	; quad $7fffffffffffffff
	mov eax,-1
	spush
	shr eax,1
;-
;- vovf		( -- v )							rational cons
;-	cons (+ve) overflow indicator
    global vofv:function
vovf:M;:ok:
	spush
	mov eax,0
	spush
M vovf
	ret
d.ovf:; <lndv)
	spop 4
;-
;- dvovf  	( -- dv )							rational cons
;-	cons (+ve) overflow indicator.
dvovf:GA;:ok:
	call vovf
	spush 2
	ret
;-
;- qvovf  	( -- qv )							rational cons
;-	cons (+ve) overflow indicator.
qvovf:GA;:ok:
	call dvovf
	spush 4
	ret
;-
;-	
;--- Literals
;-

;-	
;-
;--- text input
;-

;-
;- >double	( p u -- d n tf | 0. 0 ff )	(e)				rational
;-	translate string at p of length u to double number d in the current base
;-	and return n = @dpl and tf if conversion successful, else all zero and ff.
;-	fractional marker and, number prefices for sign and radix as w/ -re- {a-base}.
;-	stg may contain a base exponent mark, either the char corresponding to 
;-	the rsp. @base figure, e.g. as converted w/ {h>a}, or, at decimal base
;-	the char "e" or "E". the fetched rational number will then be multiplied 
;-	by @base raised to the power of |exp|<127, valid exponent -9 < exp < 10.
;-	tf if conversion successful or, stg(p,u) of zero length or fraction mark, only.
;-	ff indicates that an unvalid character was present in the string, in which case 
;-	d and n are the result of leading valid characters, if existent, else zero.
;-	NOTE: returns "TRUE" flag after SUCCESS, in contrary to any other conventional 
;-	      numeric conversion forth words - which may be a subject to change!
;-	(e: eflags reflect flag value, ecx:=p, edx:=u
toquad.n:
	or dword[byte uref uv.bits],mquad
todouble:GA;:ok:
; dup if forth quad [number] previous dpl@ swap 0= else 2drop 0. 0 true endif ;
	test eax,eax
	jz .n		;?; empty stg
	call abnumberb	; [number] ( p u -- dn|qn f )
	test eax,eax	; ff for successful conversion
	movsx eax,byte[byte uref uv.dpl]
	spush
	mov eax,0
	jnz .r		;?; ret un-tr
	cmp eax,TOS(1)
	jle .t
	mov TOS(1),eax	; [hp] no fraction = integer number
.t:
	dec eax
.r:
	ret
.n:
	mov TOS(1),eax	; 2drop 0.
	spush 2		; 0
	dec eax		; tr
	ret
;-
;- >quad  	( p u -- q n ff | q0 0 tf )					rational hidden
;-	fetch quad int or double ranum from stg(p,u); input modifiers apply as w/ {[number]}
;-	ret n := radix of current base and false if conversion successful, else true flag.
;-	max convertible figure e.g. 19 decimal, fractional+integral digits. -re- {>double}.
;-	valid decimal exponent in range -14 < exp < 19.
toquad:GA;:ok:
; dup if forth quad [number] previous dpl@ swap 0= else 2drop 0. 0 true endif ;
	test eax,eax
	jnz .n		;?; fetch quad ranum
	mov TOS(1),eax	; 2drop
	spush 3		; quad 0 0
	dec eax		; tr
	ret
; from abnumberb, after call to <abase>, w/ @dpl = -1
;( ptr u n -- rn ff | 0. tf ) n = sign|base
abranumb:GA;:ok:<abnumberb)
	push eax	; save sign from <abnumberb>
	spop		; drop (false) base figure
;:HIER: stg in zaehler/nenner aufteilen, wenn "/" enthalten ist, oder {#fd} fuer ganzzahligen teil, dann getrennt abhandeln
	and dword[byte uref uv.bits],~(mranu)	; clr ranum-bit for <todouble>
	call abase	;		{ ptr' u' base
	test ax,ax
	mov ecx,[byte uref uv.base]
	jg .b
	mov ax,cx
.b:
	or [esp],eax	; save sign flag and base
	movzx eax,ax	; temp base
	push ecx	; save current base		[ b s
	mov [byte uref uv.base],eax;store tmp. radix
	spop
	test dword[byte uref uv.bits],mqranu
	jz .u
	or dword[byte uref uv.bits],mquad
.u:
	call todouble	; ( p u -- dn dpl f | qn dpl f )
	test dword[byte uref uv.bits],mfloat
	jz .a
	or dword[byte uref uv.bits],mfloatm	; rst persistent {floats} state
.a:
	test eax,eax	; tf from successful conversion
	pop eax		; base
	mov ecx,[byte uref uv.base]; working base
	mov [byte uref uv.base],eax; rst global base value
	jz .n		;?; not a number
	mov edx,[byte uref uv.dpl]
	and edx,0x80008000
	mov eax,TOS(1)	; dpl
	js .d		;?; input by sing or bignum
	or edx,eax	; dpl count
.d:
	mov [byte uref uv.dpl],edx
	movsx edx,byte[byte uref uv.bits+edpl]
	sub eax,edx	; @dpl=divisor, @edpl=multiplicator
	push eax
	jns .p
	neg eax
.p:
	mov TOS(1),ecx	; base
	call upow	; { dpl base ^ }	( 0 q|dnum dexp
	pop edx		; sg(exp)
	mov ecx,[byte uref uv.dpl]
	shr ecx,12;;8
	cmp ecx,00000008h
	jz .q		;?; double float
	test edx,edx
	jns .s		; +ve exp
	call mdmul	; d.nom*d.den -> quad
	spush
	mov eax,1
	spush
	dec eax
	spush 2
	call qv2dv
.s:
	spush
	pop eax
	call dvpm	; apply sign
.r:
	spush
	or dword[byte uref uv.dpl],0x00800000
.x:
	and dword[byte uref uv.bits],~(mqranu)
	xor eax,eax	; flag success
	ret
.n:A;<.a) not a number
	lea esp,[esp+4]	; drop sign
	spop
	call .x
	mov TOS(2),eax
	mov TOS(1),eax	; 2drop 0.
	dec eax		; -1
	ret
.q:A;<quad, i.e. quad ranum	( q|dnum dexp
	test edx,edx
	jns .e		;?; +ve exp
	push dword TOS(2); sign
	call uqdmul	; q.nom*q.den -> 6.prd
	pop eax
	TOSPULL(1)	; quad
	call qpm	; rst sg(nom)
	spush
	mov eax,1	; d.denom = 1
	spush
	dec eax
.e:
	spush
	pop ecx
	mov eax,0	; d->q
	test ecx,ecx
	spush
	jns .v
	call qnegaten	; apply sign
.v:
	or dword[byte uref uv.dpl],0x0800	; -> xxxx88xx
	test dword[byte uref uv.bits],mqranu
	jnz disp .r	;?; ret quad ranum
	and dword[byte uref uv.dpl],~(0xfffff7ff); clr quad ranum modifier
	call qv2dv
	jmp disp .r
;-
;- >FLOAT 	( p u -- dv f )					'tofloat'	float
;-	convert stg(p,u) to a single 'float', i.e. double ranum.
;-	an empty stg, or all blanks, resulting to ovf indicator.
;-	valid, optional exponent in range -14 < exp < 19. -re- {>double}
tofloat:GA;:ok:
	call double
	call toquad
	test eax,eax
	jz .n		;?; not a number
	TOSPULL(1)	; discard dpl figure
.n:
	ret
;-
;- >DFLOAT 	( p u -- qv f )					'todfloat'	float
;-	convert stg(p,u) to a double 'float', i.e. quad ranum.
;-	an empty stg, or all blanks, resulting to ovf indicator.
;-	valid, optional exponent in range -21 < exp < 22. -re- {>double}
todfloat:GA;:ok:
	or dword[byte uref uv.bits],mdfloat
	call toquad
	test eax,eax
	jz .n		;?; not a number
	TOSPULL(1)	; discard dpl figure
.n:
	ret
;-
;- dv>udv?	( dv -- +dv f)					'dv2udvq'	rational hidden
;-	return absolute value of signed double rational no. tf if dv was negative. 
;-	A double ranum is represented on the stack as sign*d.numerator +d.denominator.
dv2udvq:GA;:ok:(vnormal)
; 2SWAP DUP 0< >R DABS 2SWAP R> ;
	spush		; nml nmh dnl dnh flg
	mov eax,TOS(3)
	mov ecx,eax
	sar eax,31
	jz .r
	neg dword TOS(4)
	adc ecx,byte 0
	neg ecx
	mov TOS(3),ecx
.r:
	ret
;-
;- v>dv		( v1 -- v2 )					'v2dv'		rational
;-	expand v1 to double cells size.
v2dv:GA;:ok:
	mov ecx,TOS(1)
	spush 2
	sar ecx,31
	mov eax,0
	mov TOS(2),ecx
	ret
;-
;--- Output
;-

;-
;- #digits 	( -- n )			(kv)				forth
;-	n is the maximal number of digits to be displayed in the numerator or
;-	denominator of a rational number, when displayed in fractional format.
hvdigits:KV

;-
;- PRECISION	( f1 -- f2 )			((kv))		'fprec'		ans float
;-	no. of displayable fractional digits, dpans EXT 12.6.2.2035
;-	by kernel value {#fplaces}
;-
;- #fplaces 	( -- n )			(kv)		'fprec'		rational
;-	returns the maximal number of fractional digits to be included in the 
;-	f.p. numeric output string by #vf. reads the kernel-value {#prec}.
fprec:GA;:ok:
	spush
	mov eax,[kref hprecv-Y]
	add eax,databot
	mov eax,[kref eax]
	ret
;-
;- SET-PRECISION  ( n -- )							ans float
;-	set no. of displayable f.p. mode fractional digits, dpans EXT 12.6.2.2035
sprec:GA;:ok:
	mov ecx,[kref hprecv-Y]
	add ecx,databot
	mov [kref ecx],eax
	spop
	ret
;-
;- #fround	( -- n )			(kv)		'hfround'	rational
;-	the number of f.p. fractional digits, rounded towards zero by -re- {dvfround}.
hfround:GA;:ok:;<sharpdvf)
	spush
	mov ecx,[kref vfroundv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
;-
;- #n!v  	( -- n )			(kv)		'snfv'		rational hidden
;-	max integer + 1 for which a vsing factorial can be calculated w/o loss praecision.
snfv:GA;:ok:;<epowv,lntv2)
	spush
	mov ecx,[kref snfvv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
;-
;- #dp		( -- c )			(kv)		'vpoint'	forth
;-	char c to insert as the fraction marker in floating point format, dft ",".
vpoint:GA;:ok:;<mzeroes,sharpvf,vdotfj,sharpdvf)
	spush
	mov ecx,[kref hdpv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
;-
;- truncation	( -- n )			(kv)		'vtrunc'	rational
;-	returns TRUE if trailing zeroes will be truncated when displaying 
;-	a rational number in floating point format.
vtrunc:GA;:ok:;<vdotfj)
	spush
	mov ecx,[kref vtruncv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
;-
;- fdelim  	( -- c )					'fdelim'	rational hidden
;-	ret the char to insert as the delimiter between integral and fractional 
;-	part in fractional format. reads kernel-value {#fd}.
fdelim:GA;:ok:;<sharpv,vdotj,sharpdv)
	spush
	mov ecx,[kref fdlv-Y]
	add ecx,databot
	mov eax,[kref ecx]
	ret
;-
;- >ch		( p u c -- n )					'gtch'		rational hidden
;-	search stg(p,u) for character c. If it is present then n is the position of the
;-	first occurrence of ch in the string. If it is not present then n is equal to u.
gtch:GA;:ok:;<vdotfj,vdotj)
; >ch rotd2s scan drop swap - ;
	mov edx,TOS(2)	; p
	mov ecx,TOS(1)	; u
	spush
	mov TOS(2),edx	; rot dup 2swap (=rotd2s)
	mov TOS(1),ecx
	call scan
	mov ecx,TOS(1)
	spop 2
	sub ecx,eax
	mov eax,ecx
	ret
;-
;- -zeroes	( p u1 -- p u2 )				'mzeroes'	rational
;-	the string at addr of length n1 is stripped of trailing "0"s, returning a
;-	string of length n2. If there is no point in the string no stripping takes
;-	place. At least one character will remain after the point.
mzeroes:GA;:ok:
; 2dup point >rr scan -dup 0= if rdrop drop exit endif
; "0 rskip drop dup c@ r> = if 1+ endif 1+ sdrop over- ;
	call dup2
	call vpoint
	push eax
	call scan
	pop ecx
	jz near drop2
	spush
	push ecx
	mov eax,'0'
	call rskip
	spop
	pop ecx
	cmp byte[kref eax],cl
	jnz .i
	inc eax
.i:
	inc eax
	TOSPULL(1)
	sub eax,TOS(1)
.r:
	ret	
;-
;- digits>	( -- u )					'vdigitsgt'	rational hidden
;-	u is the largest number that can be represented in the current base by 
;-	not more than {digits} number of digits.
;-	to preventing very odd values {digits} will be truncated to less than 256
;-	which, if {pad-size} not set to a value below default will always be safe.
vdigitsgt:GA;:ok:;<vdotj,dvdotj)
;digits 255 and dup to digits BASE @ SWAP um**  -1 M+ DROP ;
	spush
	call .v
	jg .r
	xor edx,edx
	mov eax,[byte kref uv.base]
	spush
	mov ecx,eax
	mov eax,32
	bsr ecx,ecx
	div ecx
	jr .p
.v:A
	mov eax,[byte uref uv.base]
	spush
	mov ecx,[kref vdigitsv-Y]
	add ecx,databot
	movzx eax,byte[kref ecx]
	mov [kref ecx],eax	; protect...
.p:
	call upow
	spop
	dec eax
.r:
	ret
;-	
;-
;--- double size rational numbers
;-

;-
;-	
;--- defining 
;-

	global dvliteral:function,dvconstant:function,dvvariable:function,dvinteger:function
;-
;- dvliteral	(C: dv -- ) (X: -- dv )						rational
;-	-re- 4literal
dvliteral:equ qliteral

;-
;- dvconstant	ccc(C: dv -- ) (X: -- dv )					rational
;-	-re- {4constant}
dvconstant:equ qconstant

;-
;- dvinteger	ccc(C: dv -- ) (X: -- dv )					rational
;-	double ranum modifiable constant (two cells L4 variant of a {value}).
;-	synonym to -re- {4integer}.
dvinteger:equ integer4

;-
;- dvvariable	ccc(C: -- ) (X: -- p )						rational
;-	-re- {4variable}
dvvariable:equ qvari

;-
;-	
;--- Memory
;-

    global dvstore:function,dvfetch:function
;-
;- dv@		( p -- dv ) 					'dvfetch'	rational
;-	fetch dv from memory at p. -re- {4@}.
dvfetch:equ fetch4
;-
;- dv!		( dv p -- ) 					'dvstore'	rational
;-	store dv to memory at p. -re- {4!}.
dvstore:equ store4

;-
;-	
;--- Stacks
;-

    global dvdrop:function,dvdup:function,dv2dup:function,dvmrot:function,dvover:function
    global dvpick:function,dvrot:function,dvroll:function,dvswap:function
;-
;- dv-rot  	( dv1 dv2 dv3 -- dv3 dv1 dv2 )			'dvmrot'	rational
;-	-re- {4-rot}
dvmrot:equ	mrot4
;- dvdrop  	( dv1 -- dv1 dv1 )						rational
;-	-re- {4drop}
dvdrop:equ	drop4
;- dvdup  	( dv1 -- dv1 dv1 )						rational
;-	-re- {4dup}
dvdup:equ	dup4
;- dv2dup  	( dv1 dv2 -- dv1 dv2 dv1 dv2 )					rational
;-	-re- {8dup}
dv2dup:equ	dup8
;- dvover  	( dv1 dv2 -- dv1 dv2 dv1 )					rational
;-	-re- {4over}
dvover:equ	over4
;- dvpick  	( ... n -- ... dvn )						rational
;-	-re- {4pick}
dvpick:equ	pick4
;- dvrot  	( dv1 dv2 dv3 -- dv2 dv3 dv1 )					rational
;-	-re- {4rot}
dvrot:equ	rot4
;- dvroll  	( ... n -- ... dvn )						rational
;-	-re- {4roll}
dvroll:equ	roll4
;- dvswap  	( dv1 dv2 -- dv2 dv1 )						rational
;-	-re- {4swap}
dvswap:equ	swap4
;-	quad cells opr. apply to double rational numbers.
;-

;-
;- dvdepth	( -- n )							rational
;-	aequivalent to the number of double ranums in data-stack.
dvdepth:GA
	call depth
	sar eax,2
	ret
;-
;-	
;--- arithmetics
;-

    global dvoverflow:function
;-
;- dvoverflow	( dv -- f )							rational
;-	tf if dv was the overflow indicator
dvoverflow:equ qzeq

;-
;- dvround 	( dv1 +dn -- dv2 )						rational
;-	returns an approximation to the rational number dv1 such that neither  
;-	the numerator nor the denominator of its fractional part exceeds the 
;-	+ve integer dn.
dvround:GA;:ok:
;  >R 2dup voverflow if rdrop exit endif
	mov ecx,TOS(3)
	or ecx,TOS(2)
	jz near drop2
;  setvars v>uv? ( v -- uv f ) R> also ans 2>R previous
	call mrot2	; 2-rot		save limit
	call dv2udvq	; |dv| & flg
	spush
	mov eax,1	; init p.num & den
	spush
;  { sg.dv prev.numerator denominator | numerator prev.denominator } [ = setvars ]
	kloci 11
.pnuh:	klocz		; prev.numerator.hi
.pnul:	klocv		;
.dnh:	klocz		; denominator.hi
.dnl:	klocv		;
.sg:	klocv		; sign flag from v>uv?
.nuh:	klocz		; numerator.hi
.nul:	klocz		;
.pdnh:	klocz		; prev.denominator.hi
.pdnl:	klocz		; 
	call rot2
.dvah:	klocv		; dn.hi
.dval:	klocv		;
.enter:;    enter
	jmp .entry
.begin:A;    begin	; ( d u			[ limit
;	OVER DUP IF /mod THEN		(1)
	call over2	; over
	mov ecx,eax	; dup
	or ecx,TOS(1)	; 0=/=
	jz .endif1	; if
	call dbym	; d/mod (_signed_ div, mul with den as well as num!)
.endif1:
;        DUP r@ snum>?
	call dup2	; dup (r@ subst. by local value)
;snum>? ( u1 u2--u3 f)
;  >R numerator * prev.numerator + DUP R> U> ;
	spush		; numerator
	mov eax,[lref .nul]
	spush
	mov eax,[lref .nuh]
	call dmul	; *
	mov ecx,TOS(1)
	add ecx,[lref .pnul]
	adc eax,[lref .pnuh]
	mov TOS(1),ecx	; prev.numerator +
	call dup2	; dup
	spush
	mov eax,[lref .dval]
	spush
	mov eax,[lref .dvah]
	call dugreat	; r@ ... >r ... r> u>
	test eax,eax
	spop
;	IF next!
	jnz .if2
;	else [ 2 cs-roll 2 cs-roll ]
;	    swap r@ sden>? 
	call swap2	; swap
;sden>? ( u1 u2 -- u3 f)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  >R denominator * prev.denominator + DUP R> U> ;
	spush
	mov eax,[lref .dnl]
	spush
	mov eax,[lref .dnh]
	call dmul	; denominator *
	mov ecx,TOS(1)
	add ecx,[lref .pdnl]
	adc eax,[lref .pdnh]
	mov TOS(1),ecx	; prev.denumninator +
	call dup2	; dup
	spush
	mov eax,[lref .dval]
	spush
	mov eax,[lref .dvah]
	call dugreat	; r@ ... >r ... r> u>
	test eax,eax
	spop
	jnz .if3
;	    IF next! else [ 2 cs-roll 2 cs-roll ] 
;		next! SWAP
;    entry dup 0= until
	call .nexts
	call swap2
.entry:;    entry
;	dup 0=
	mov ecx,eax
	or ecx,TOS(1)
.until:;    until
	jnz disp .begin
.endif23:;    endif endif
;    endif endif
;    2drop rdrop
;   prev.numerator R> +- prev.denominator ;
	cmp dword[lref .sg],byte 0
	mov ecx,[lref .pnul]
	mov eax,[lref .pnuh]
	jns .p
	neg ecx
	adc eax,byte 0
	neg eax
.p:
	mov edx,[lref .pdnl]
	mov TOS(2),eax
	mov TOS(3),ecx
	mov TOS(1),edx
	mov eax,[lref .pdnh]
	ret
.if2:A
.if3:
	call .nexts
	jr .endif23
.nexts:A
;make the new approximation the current approximation, and the current
;approximation the previous approximation.
	mov ecx,[lref .dnl]	;c: denominator..
	mov edx,[lref .dnh]
	mov [lref .pdnl],ecx	;d: ..to prev.denominator
	mov [lref .pdnh],edx	;
	mov ecx,TOS(1)
	mov [lref .dnh],eax	;c: ..to denominator..
	mov [lref .dnl],ecx
	mov edx,[lref .nul]	;d: numerator..
	mov ecx,[lref .nuh]	;
	mov [lref .pnul],edx	; to prev.numerator
	mov [lref .pnuh],ecx	;
	mov eax,TOS(2)
	mov ecx,TOS(3)
	mov [lref .nuh],eax	; to numerator
	mov [lref .nul],ecx
	spop 4
dvfround.r:
	ret
;-
;- dvfround 	( dv1 -- dv2 )			(4)				rational
;-	returns an approximation to the rational number dv1 such that the fractional 
;-	part rounded to {#fround} digits. simple rounding algorithm which increments
;-	last digit by one if next would be greater or equal to 0,5 * rounding factor:
;-	{ dvsplit base@ #fround u^ d>dv 4sover dv* df 0,5 dv+ dv>floor dvswap dv/ dv+ }
;-	noop if (kernel-)value {#fround} < 0.
dvfround:GA;:ok:
	call hfround	; in id fn fd sr
	test eax,eax
	spop
	js .r
	call dvsplit	; in id fn fd
	spush
	mov eax,[byte uref uv.base]
	call hfround	; in id fn fd sr
	call upow	; in id fn fd pr
	call d2dv	; in id fn fd pn pd
	call sover4	; in id pn pd fn fd pn pd
	call dvmul
	spush
	mov eax,1
	spush
	dec eax
	spush
	lea eax,[eax+2]
	spush
	lea eax,[eax-2]
	call dvplus
	call dv2floor
	call dvswap
	call dvby
	jmp dvplus
;-
;- dvcmp  	( dv1 dv2 -- dv1 dv2 f )	(e)				rational
;-	leave signum of ranum doubles subtraction.
;-	(e: eflags reflect signum
dvcmp:GA;:ok:
	call dup8
	call dvsub
;-
;- dvsgn  	( dv -- n )			(e)				rational
;-	ret sing signum of dv.
;-	(e: eflags reflect signum
dvsgn:GA;:ok:
	call dvzeq
	jz .i	;0= -> 0
	test ecx,ecx
	lea eax,[eax-1]
	js .r	; -> -1
	inc eax	; ->  1
.i:
	inc eax
.r:
	ret
;-
;- dvmin  	( dv1 dv2 -- dv3 )						rational
;-	ret dv3 the smaller of dv1,dv2.
dvmin:GA;:ok:
	call dvcmp
	spop
	jg near sdrop4
	spop 4
	ret
;-
;- dvmax  	( dv1 dv2 -- dv3 )						rational
;-	ret dv3 the greater of dv1,dv2.
dvmax:GA;:ok:
	call dvcmp
	spop
	js near sdrop4
	spop 4
	ret
;-
;- #dv		( dv -- 0.0 )					'sharpdv'	rational
;-	append signed double ranum dv to the numeric output string:
;-	  If dv is zero append "0".
;-	  If dv has an integral component but no fractional component, append that
;-	        number.
;-	  If dv has a fractional component but no integral component append "a/b"
;-	        where a is the numerator, and b the denominator.
;-	  If dv has both a an integral and a fractional component append "a b/c"
;-	        where a is the integral component, b the numerator and c the
;-	        denominator of the fractional component. There is a delimiting char
;-	        between the integral and fractional components,-re- {#fd}, {fdelim}.
;-	Sign applies to the entire figure; conversion according to current @BASE.
;-	Leave double zero, after conversion.
sharpdv:GA;:ok:
; 4dup dv0= if(1) 2drop "0 hold
	push dword TOS(2)	; sign
	call dvabs
	call dup4
	call dvzeq		; dv2dup dv0=
	spop
	jz near .s1		; if...
; else(1) 2dup d1- or
	test eax,eax
	jnz .e2
	cmp dword TOS(1),byte 1
	jz .e2			;?; no fraction
;	; if(2) 2dup #s 2drop "/ hold 4dup dmod #s 2drop
	call dup2
	call sharps
	TOSPULL(1)
	mov eax,"/"
	call hold
	call dup4
	call dmod
	call sharps
	spop 2
;	; 4dup d< 0if(3) bl hold endif(3)
	call .h
	jnz .e4
	call fdelim
	call hold
	jr .s4
.e2:;	;.endif(2)
;	; 4dup d< 0if(4) dv>d #s endif(4)
	call .h
	jnz .e4
.s4:
	call dv2d
	call sharps
	spush 2
.e4:
	TOSPULL(1)
	mov eax,[esp]
	call asign
	xor eax,eax
	lea esp,[esp+4]	; drop sign cell
	mov TOS(1),eax
	ret
.s1:;	; 2drop "0 hold
	TOSPULL(2)
	mov eax,"0"
	call hold
	xor eax,eax
	lea esp,[esp+4]	; drop sign cell
	spush
	ret
.h:A
	call dup4
	call dless
	test eax,eax
	spop
	ret
;
sharpdvf.v:A;<append overflow stg to num. output buffer)
	TOSPULL(1)
	xor eax,eax	; dvnl dvnl dvdl flg
	mov TOS(2),eax
	mov TOS(1),eax	; 0 0 dvdl flg
	mov eax,vovt-Y
	call ccount	; 0 0 dvdl pv' uv'
	mov edx,TOS(1)	; pv'
	mov ecx,[byte uref uv.hld]
	mov TOS(2),edx	; sto src
	sub ecx,eax	; dst
	dec ecx		; +countbyte
	mov TOS(1),ecx
	mov [byte uref uv.hld],ecx
	jmp xcmove	; ovf-stg to buf, leave 0,0
;-
;- #dvf		( +dv -- 0 0 )					'sharpdvf'	rational
;-	append the positive ranum +dv to the numeric output string as a	floating 
;-	point number w/ the current value of {base}. That is to say, the integral 
;-	component, followed by a marker char -re- {#dp}, the "decimal point", 
;-	followed by as many digits as specified by -re- {fplaces} w/ no rounding. 
;-	Returns 0 0 for consistency with other numeric strings conversion words. 
sharpdvf:GA;:ok:
; 2>rr ud/mod 2swap base@ udm* 0 2r> fplaces 0> 
; if fplaces >r 2>r 
;  begin 2r 0. uq/mod 3drop h>a hold base@ uqsm 2r> r> 1- dup >rr -rot >r 0> 0= until 
;  drop hld @ dup dup dup dup hld0 @ - hld0 @ >move< #dp hold rdrop 2rdrop
; endif 6 ndrop #s ;s
; -- ;
; 2>rr ud/mod 2swap base@ udm* 0
	call qdzeq
	jz .v			;?; store ovf string
	spop
	push eax		; save denum
	push dword TOS(1)
	call udbym		; ud/mod	( dr di
	call swap2		; 2swap		( di dr
	call basef		; base @	( di dr s.b
	call udmmul		; udm*		( di dr's.o	| d
	spush			; 		( di dr'd.o	| d
	mov eax,0
; 2r> fplaces
	spush			; double integer part, quad fraction*@base
	pop eax
	call fprec		; 2r> #fplaces	( di dr' d.o dd	| -/-
	mov ecx,eax
	pop eax
	test ecx,ecx
; 0> if fplaces >r 2>r 
	jng .e
	push ecx		; #fplaces >r
	push eax
	spop
	push eax		; 2>r				| d s.loop-count
;  begin 2r 0. uq/mod 3drop h>a hold base@ uqsm 2r> r> 1- dup >rr -rot >r 0> 0= until 
.do:A				;		( di dr' d.o	| dd count
	mov eax,[esp]		; drop ovf, copy denum.lo
	spush
	mov eax,[esp+4]		; 2r		( qr' d
	spush
	mov eax,0
	spush
	call uqbym		; uqm/mod	( qr" dn
	mov eax,TOS(3)		; drop
	TOSPULL(2)
	call h2a		; h>a hold
	call hold
	mov eax,[byte uref uv.base];base @
	call uqsmmul		; 5-cells prod
.loop:
	dec dword[esp+8]	; (places)-1
	jg .do
;  drop hld @ dup dup dup dup hld0 @ - hld0 @ >move< #dp hold rdrop 2rdrop	\ sto fraction
	mov ecx,[byte uref uv.hld0]
	mov eax,[byte uref uv.hld];		( ov h
	spush 4
	sub ecx,eax		; u
	mov eax,ecx
	call movex		;		( ov h h h h h u -- ov h h
	call vpoint
	call hold		; prepend fraction char
	lea esp,[esp+12]
.e:;( di dr'do d	| -/-
; endif 3drop 3drop #S ;	\ sto integer
	spop 6			; drop r' d, leave integer
	jmp sharps
;-
;- <#dv#> 	( dv -- p u )					'ltdvgt'	rational
;-	Signed convert double ranum to fractional format stg.
;-	Returns the string(p,u) representing the rational number dv. The string 
;-	is as described in -re- {#dv}, but will have a preceding minus sign if 
;-	dv was negative. If dv is the overflow indicator, returns stg "Overflow".
ltdvgt:GA;:ok:
	lea edx,[kref sharpdv-Y]
.n:
	push dword TOS(2); sign
	push edx
	call dvabs
	call dup2
	call dzeq
	spop
	jz sharpdovf
.c:
	call ltsharp
	pop edx
	call edx	; <sharpvf> or <sharpv>
	spush
	pop eax
	call asign	; (ans-version)
	jmp sharpgt
;
sharpdovf:A
	lea esp,[esp+8]	; drop xec-vec and sign
	mov eax,vovt-Y
	TOSPULL(3)
	jmp ccount
;-
;- <#dvf#>	( dv -- p u )					'ltdvfgt'	rational
;-	-re- {#dvf}, signed convert -re- {dvfround}ed double ranum to floating pt stg.
ltdvfgt:GA;:ok:
	lea edx,[kref sharpdvf-Y]
	jr ltdvgt.n
;-
;- df.j		( dv n1 n2 -- )					'dvdotfj'	rational
;-	display double ranum dv according to the rules described in -re- {#dvf},
;-	padded w/ up to n1 leading and up to n2 trailing blanks, none if -ve.
dvdotfj:GA;:ok:
	inc eax		; n2 1+
	mov ecx,TOS(1)
	push eax	; (2)
	push ecx	; pass tor{ n1 n2 }
	spop 2
	call dvfround
	call ltdvfgt	; raw stg
%ifdef __SRANUM
	jmp disp vdotfj.n
%else
.n:; <dvdotfj)
	call vtrunc
	test eax,eax
	spop
	jz .endif1
	call mzeroes
.endif1:
	call dup2
	call vpoint
	call gtch	; >ch
	mov ecx,eax
	spush
	pop eax
	sub eax,ecx	; over-
	call spaces
; over 2swap ; ( 1 2 3 -- 3 2 1 2
	mov ecx,TOS(2)	; (1)
	mov edx,TOS(1)	; (2)
	mov TOS(2),eax	; (3)
	mov eax,ecx	;(1)->
	spush
	mov eax,edx	;(2)->
	call type
	mov ecx,TOS(1)
	sub ecx,eax	; - r> +
	spop
	pop eax
	add eax,ecx
	jmp spaces
%endif
;-
;- df.		( dv -- )					'dvdotf'	rational
;-	display a rational number in floating point format as w/ -re- {df.j}
;-	but without left or right justification, and with one trailing space.
dvdotf:GA;:ok:
	spush
	mov eax,-1	; no justification
	spush
	call dvdotfj
	jmp space	; one trailing blank
;-
;- dv.j		( dv n1 n2)					'dvdotj'	rational
;-	display rational number dv in fractional format, as described in <#dv#>,
;-	padded w/ up to n1 leading and, up to n2 trailing spaces, none if -ve.
dvdotj:GA;:ok:
	push eax
	push dword TOS(1); 2>r
	spop 2
	call vdigitsgt	; digits>
	spush
	xor eax,eax	; u->d
	call dvsimplify
	call ltdvgt	; <#dv#>
%ifdef __SRANUM
	jmp vdotj.n
%else
.n:; <dvdotj)	; 2DUP [CHAR] / has-ch
	call dup2	; -- p u p u 
	spush
	mov eax,'/'
	call scan	; "/ hasch
	test eax,eax
	spop 2		; -- p u 
; IF 2DUP BL >ch 2DUP >R =
	jz .else1	;?; no fraction
	call dup2	; -- p u p u 
	call fdelim	; #fd
	call gtch	; -- p u u'
	push eax	; 2dup >r ; -- p u u' u	[ u' R2 R1
	cmp eax,TOS(1)	; = 	  ; -- p u f
;    IF 2R> >R 1+ SPACES TYPE 2R> - SPACES	; [2]
	jnz .else2	;?; bl not at end of stg, i.e. integer and fraction
;	; <no integral part>
	pop ecx		; Ru
	pop eax		; R2
	push ecx	;		[ Ru R1
	inc eax
	call spaces
	call type
	spush
	pop ecx
	pop eax
	sub eax,ecx
	jmp spaces
;    ELSE[2]  2R> DUP >R - SPACES TUCK TYPE 2R> + 1+ SWAP - SPACES THEN[2]
.else2:A
	pop edx		; 2R> dup >r (ans) 	:=>  2r> swap >rr	:=> r
	pop eax
	push edx
	sub eax,edx	;	{ R2 u		[ u R1
	call spaces	; spaces
	call sover	; tuck
	call type	; type
	pop edx
	pop ecx		; 2r>
	inc ecx
	add ecx,edx	; + 1+
	sub ecx,eax	; swap -
	mov eax,ecx
	jmp spaces
; ELSE[1] R> OVER - SPACES TYPE R> 1+ SPACES THEN[1] ;
.else1:A;<no fraction)	; ( p u 
	mov ecx,eax
	spush
	pop eax		; r>
	sub eax,ecx	; over-
	call spaces	; spaces
	call type	; type
	spush
	pop eax		; r> 1+
	inc eax
	jmp spaces	; spaces endif
%endif
;-
;- dv.		( dv -- )					'dvdot'		rational
;-	display a ranum in fractional format as described in dv.j but, 
;-	w/o left or right justification and, with one trailing space.
dvdot:GA;:ok:
	spush
	mov eax,-1	; no justification
	spush
	call dvdotj
	jmp space	; one trailing blank
;-
;- dvsimplify	( dv1 +dn -- dv2 )		(4)				rational
;-	returns approximation to the rational number dv1 such that neither the
;-	numerator nor the denominator of the fractional component exceeds dn.
dvsimplify:GA;:ok:
	push eax
	push dword TOS(1)
	spop 2
	call dvsplit
	spush 2
	pop dword TOS(1)
	pop eax
	call dvround
	jmp dvplus
;-
;- dvreduce	( dv -- dv)			(4)				rational
;-	normalises double ranum by greatest common divisor of numerator and denominator.
dvreduce:GA;:ok:; <dvplus)
; 2DUP gcd TUCK / >R / R> ;
	call dup4
	call dgcd
	test eax,eax
	mov edx,TOS(1)
	jnz .d		;?;
	cmp edx,byte 2	; 2 u< if... (ovf or unity)
	jc near drop2	;?; no gcd
.d:
	call sover2
	call dby
	call mrot2
	call dby
	jmp swap2
;-
;- s>dv		( n -- dv )					's2dv'		rational
;-	signed promote sing integer to double ranum equivalent. n --> n s->d 1. =: dv.
s2dv:GA;:ok:
	spush
	sar eax,31
;-
;- d>dv		( dn -- dv )					'd2dv'		rational
;-	promote double integer to double ranum equivalent. dn --> dn 1. =: dv.
d2dv:GA;:ok:
	spush
	mov eax,1
	spush
	dec eax
	ret
;-
;- dv>d		( dv -- d )					'dv2d'		rational
;-	return the floored, integral component of ranum dv as signed integer d.
dv2d:GA;:ok:
	mov edx,eax	; dh
	mov ecx,TOS(1)	; dl den
	mov eax,TOS(2)	; nh
	sar eax,31	; nhh d>q
	mov TOS(1),eax	; nhl
	spush
	mov eax,ecx	; dl
	spush
	mov eax,edx	; dh
	call qdfmbym
	jmp sdrop2
;-
;- dv>s  	( dv -- n )					'dv2n'		rational
;-	return the m.s. signed integral component of ranum dv as signed single integer n.
dv2n:GA;:ok:
	call dv2dn
	mov edx,1
	jr dv2dn.p
;-
;- dv>dn  	( dv -- d )					'dv2dn'		rational
;-	return the m.s. signed integral component of ranum dv as signed integer d.
dv2dn:GA;:ok:
	call dv2qv
	call qbym
	call sdrop4
	mov edx,3
	call .p
	mov edx,2
.p:
	test eax,eax
	mov ecx,eax
	jz .m
	inc ecx
	jz .m
	spush
	mov eax,edx
	call roll
.m:
	spop
dvfrac.r:
	ret
;-
;- dvsplit 	( dv1 -- dv2 dv3 )						rational
;-	floored integral dv2 and (non-negative) fractional component dv3 of ranum dv1. 
;-	w/ Floored rounding. word is tolerant of the overflow condition.
dvsplit:GA;:ok:
	call dup4
	mov ecx,eax
	or ecx,TOS(1)
	jz dvfrac	;?; ovf
	call dv2floor
	call swap4	;:continued:
;-
;- dvfrac  	( dv1 -- dv2 )							rational
;-	return the fractional component of dv1 as non-negative dv2.
;-	w/ Floored rounding. word is tolerant of the overflow condition.
dvfrac:GA;:ok:
	mov ecx,TOS(1)
	or ecx,eax
	jz .r		; 2dup d0= -exit
	call sover2
	mov edx,eax
	mov ecx,TOS(1)	; 2>r
	mov eax,TOS(2)
	sar eax,31
	mov TOS(1),eax	; d->q
	spush
	mov eax,ecx
	spush
	mov eax,edx	; 2r>
	call qdfmbym	; qdfm/mod
	spop 2		; 2drop
	jmp swap2	; 2swap

;- dv>floor	( dv1 -- dv2 )					'dv2floor'	rational
;-	double ranum to floored integral value
dv2floor:GA
	call dv2d
	jmp d2dv
;-
;- dv>int	( dv1 -- dv2 )					'dv2int'	rational
;-	round float to nearest integral value
dv2int:GA
	call dv2dn
	jmp d2dv
;-
;- FROUND	( dv1 -- dv2 )							float
;-	round float to nearest integral value
fround:equ dv2floor
	global dv2floor:function
;-
;- 1/dv		( dv1 -- dv2 )					'dvrecip'	rational
;-	ret dv2 the reciprocal of dv1. 
dvrecip:GA;:ok:
	mov ecx,TOS(2)
	test ecx,ecx
	jns .s
	call dnegate
.s:
	call swap2
	test eax,eax
	js near dnegate
	ret
;-
;- dv+-		( dv1 n -- dv2 )				'dvpm'		rational
;-	negate dv1 if n < 0
dvpm:GA;:ok:;(m,19)
	test eax,eax
	spop 
	jns dvnegate.r	;:continued:
;-
;- dvnegate	( dv -- dv )			(m,10)(e)			rational
;-	negate dv1
;-	(e: eflags reflect result of numerator negation
dvnegate:GA M;:ok:
	neg dword TOS(3)
	adc dword TOS(2),byte 0
	neg dword TOS(2)
M dvnegate
.r:
	ret
;-
;- dvabs  	( dv -- |dv| )							rational
;-	negate dv if -ve
dvabs:GA;:ok:
	cmp dword TOS(2),byte 0
	js dvnegate
	ret
;-
;- dv0=		( dv -- f )			(e)		'dvzeq'		rational
;-	tf if dv is zero
;-	(e: Z if dv=0, ecx=sign cell
dvzeq:GA;:ok:
; c: ecx,edx	;<eflags: dvsgn, sharpdv; ecx: dvsgn)
	test eax,eax
	mov ecx,TOS(1)	; sg=0 if ovf
	jnz .v
	cmp eax,ecx
	jz .r
.v:
	mov ecx,TOS(2)	; nh
	mov edx,TOS(3)	; nl
	or edx,ecx
.r:
	TOSPULL(3)
	mov eax,0
	jnz .x		; Z if zero dv
	lea eax,[eax-1]
.x:
	ret
;-
;- dv~		( dv1 dv2 -- f )				'dvaeq'		rational
;-	tf if dv1, dv2 approximately (within dv accuracy) equal or, if |dv1-dv2| = ovf.
dvaeq:GA;:ok:
	call dvsub
	call dvabs
	cmp eax,TOS(2)	; dh, nh
	mov ecx,TOS(1)
	jnz .n
	cmp ecx,TOS(3)	; dl, nl
	jz .z		;?; exactly equal
.n:
	call udbym	; -- udr udq
	or eax,TOS(1)	; qot
	or eax,TOS(2)	; rmd m.s.cell
.z:
	mov eax,0	; ff = non-zero dv
.o:
	jnz .g		; num =/= 0
	not eax
.g:
	TOSPULL(3)
	cmp eax,byte -1	; Z if zero dv
	ret
;
faeq.z:A;<f~) ret exact eqality
	spop 4		; drop relation
;-
;- dv=		( dv1 dv2 -- f )				'dveq'		rational
;-	tf if d1, dv2 exactly equal.
dveq:GA;:ok:
	call mrot2
	call mdmul
	call swap4
	call mdmul
	jmp qeq
;
; die dpans-beschreibung fuer F~ ist aehnlich bescheuert wie bei withIN.
; drum g4th dreist copiert - wo die sache offenbar ebensowenig verstanden 
; wurde, d.h. die "mathematische" dartellung aus dpans94 geradewegs nur 
; in forth umgesetzt (und falsch abgeschrieben) ist:
;: f~ ( r1 r2 r3 -- flag ) \ float-ext f-proximate
;    \G ANS Forth medley for comparing r1 and r2 for equality: r3>0:
;    \G @code{f~abs}; r3=0: bitwise comparison; r3<0: @code{fnegate f~rel}.
;    fdup f0=
;    IF \ bitwise comparison
;	fp@ float+ 1 floats over float+ -text 0=
;	fdrop fdrop fdrop
;    else fdup f0> IF f~abs ELSE fnegate f~rel THEN endif ;
; und mehr bloedsinn aus dem wissensborn der "gforth"-gesellschaft:
;   "Whilst every sane Forth has a separate floating-point stack, (...&c)
; ein grund hierfuer wird nicht genannt... (wie auch)
   
;-
;- F~		( f1 f2 f3 -- flg )				'faeq'		float
;-	should check for approximate equality according to dpans, 12.6.2.1640.
;-	apparently, w/ f1=test and f2=expected exact result:
;-		+ve: f3=abs. tolerance, f2-f3 < f1 < f2+f3.
;-		-ve: f3=factor of how much f1 may deviate from f2, 
;-		     (f1-f2)*|f3| < f2 < (f1+f2)*|f3|	-?-
;-	-re- {dv~} for an alternative approach.
;- --
;-				       |r1-r2|
;-	r3>0: |r1-r2|<r3;	r3<0: --------- < |r3| = |r1-r2|<|r3*(|r1|+|r2|)|
;-				      |r1|+|r2|
;- --
;-	(I do not understand a word of that silly dpans94 description,
;-	(section 12.6.2.1640; thus just providing F~, left un-tested, 
faeq:GA
	call dvdup
	call dvzeq
	spop
	jz .z		;?; ret exact eqality
	test ecx,ecx
	js .m		;?; relative error
;: f~abs ( r1 r2 r3 -- flag ) \ gforth
;    \G Approximate equality with absolute error: |r1-r2|<r3.
;    frot frot f- fabs fswap f< ;
	call mrot4	; rot rot
	call dvsub	; f-
	call dvabs	; fabs
	jr dvgt		; fswap f<
.m:A
;: f~rel ( r1 r2 r3 -- flag ) \ gforth		<= ...jedenfalls :muell:: falsch von dpans94 abgeschrieben
;    \G Approximate equality with relative error: |r1-r2|<r3*|r1+r2|
;	frot frot fover fabs fover fabs f+ frot frot
;	f- fabs frot frot f* f< ;
	call mrot4	; r3 r1 r2
	call over4
	call dvabs
	call over4
	call dvabs
	call dvplus	; |r1|+|r2|
	call dv2div	; arithmetisches mittel(?)
	call mrot4	;		( r3 |1|+|2| r1 r2
	call dvsub
	call dvabs	; |r1-r2|	( r3 |1|+|2| |1-2|
	call mrot4
	call dvmul
	call dvabs	; 		( |1-2| |3*(|1|+|2|)|
;;	call dv2div	; arithmetisches mittel(?)	;:continued: dv<
;-
;- dv<		( dv1 dv2 -- f )		(4)		'dvlt'		rational
;-	tf if dv1 signed smaller than dv2.
dvlt:GA;:ok:
	call swap4	;:continued:
;-
;- dv>		( dv1 dv2 -- f )		(4)		'dvgt'		rational
;-	tf if dv1 signed greater than dv2.
dvgt:GA;:ok:; 2-rot md* 4swap md* q<
	call mrot2
	call mdmul
	call swap4
	call mdmul
	jmp qless
;-
;- dv0<		( dv -- f )			(m,15)(e)	'dvzlt'		rational
;-	tf if dv is -ve.
;-	(e: S reflects result
dvzlt:GA M;:ok:
	cmp dword TOS(2),byte 0
	mov eax,0
	TOSPULL(3)
	jns .r
	dec eax
.r:
M dvzlt
	ret
;-
;- dv2*		( dv1 -- dv2 )					'dv2mul'	rational
;-	multiply dv1 by 2.
dv2mul:GA;:ok:
	mov ecx,TOS(1)
	shrd ecx,eax,1	; l<-h, den 2/
	jc .c		;?; praecision loss
	shr eax,1
	mov TOS(1),ecx
	ret
.c:A;<dv2*, preserve num l.s.b)
	mov ecx,TOS(2)	; hi
	mov edx,TOS(3)
	shld ecx,edx,1	; h<-l, num 2*
	lea edx,[edx*2]	; lo
	jo .o		;?; sign ovf
	mov TOS(2),ecx
	mov TOS(3),edx
	ret
.o:A;<dv2*, sign ovf)
	call dup4
	jmp dvplus
;-
;- dv2/		( dv1 -- dv2 )					'dv2div'	rational
;-	divide dv1 by 2.
dv2div:GA;:ok:
	mov ecx,TOS(2)	; hi
	mov edx,TOS(3)
	shrd edx,ecx,1	; l<-h, num 2/
	jc .c		;?; praecision loss
	sar ecx,1
	mov TOS(3),edx
	mov TOS(2),ecx
	ret
.c:A
	mov ecx,TOS(1)
	shld eax,ecx,1	; h<-l, den 2*
	lea ecx,[ecx*2]
	jo .o		;?; prevent praecision loss, by { 2 dvm/ }
	mov TOS(1),ecx
	ret
;-
;- dvm*		( dv dv -- qv)					'dvmmul'	rational
;-	multiply two double ranums, generating quad result.
dvmmul:GA;:ok:;<dv*,fasinh)
	call rot2
	call umdmul	; q.d1*d2
	call swap4
	call mdmul	; q.n1*n2
	call swap4
	jmp qvreduce	; qn qd
;
dvmul.m:A; ret -ve double
	mov ecx,TOS(3)
	mov eax,TOS(2)
	mov TOS(5),ecx
	mov TOS(4),eax
	TOSPULL(4)
	ret
dv2div.o:A; div by v=2
	spush
	mov eax,2
	spush
	dec eax		;:continued: dvm/
;-
;- dvm/		( dv1 v2 -- dv3 )				'dvmby'		rational
;-	divide double dv2 by sing v2 giving double quotient dv3.
dvmby:GA;:ok:;<dv2/)
	mov ecx,TOS(1)
	spush 2
	sar ecx,31
	xor eax,eax
	mov TOS(2),ecx	;:continued:
;-
;- dv/		( dv1 dv2 -- dv3 )				'dvby'		rational
;-	divide dv1 by dv2 giving quotient dv3.
;-	div-0: ret +/- max or, ret 1 if dv1=0.
dvby:GA;:ok:;<dvm/)
	mov ecx,TOS(1)
	or ecx,eax
	jz .c		;?; ret ovf
	mov ecx,TOS(4)
	or ecx,TOS(5)
	jz .c		;?; ret ovf
	mov ecx,TOS(2)
	or ecx,TOS(3)
	jz .z		;?; div 0
.c:
	call dvrecip	;:continued:
;-
;- dv*		( dv dv -- dv)					'dvmul'		rational
;-	multiply two doubles.
dvmul:GA;:ok:;<dv/)
; n1l.n1h d1l.d1h  n2l.n2h d2l.d2h -- npl.nph dpl.dph
	call dvmmul	; 7)num.ll.lh.hl.hh 3)den.ll.lh.hl.hh
	mov ecx,eax	; n.hh
	mov edx,TOS(4)	; d.hh
	or ecx,TOS(1)	; n.hl
	jnz .n		;?; quad den
	or ecx,edx	; d.hh
	or ecx,TOS(5)	; d.hl
	jnz .n		;?; num not in range of a double
	cmp ecx,TOS(6)	; d.lh
	jng .m		;?; +ve double
	cmp edx,byte -1
	jnz .n		;?; quad num
	xor edx,TOS(5)	; check whether dh = -1.
	jnz .n		;?; quad -ve num
	cmp edx,TOS(6)	; d.lh
	jng .m		;?; -ve double
.n:
	call qvsplit
	spush
	mov eax,0x7fffffff
	spush
	mov eax,0
	spush 2
	call qvround
	call qv2dv
	call mrot4
	call qv2dv
	jmp dvplus
dvby.z:A; dv1/0 -> +/- max or 1 for 0/0
	mov edx,TOS(6)	; sg(dv1)
	mov ecx,edx
	or edx,TOS(7)
	TOSPULL(4)	; drop dv2, leave eax=0=den.hi
	jz dvby.e	;?; 0/0
dvby.m:; eax=0, ecx = sign cell
	not ecx
	sar ecx,30
	or ecx,byte 1	; +ve:-1 -ve:1
	add ecx,0x80000000; -> 7fffffff or 80000001=-7fffffff
	movsx edx,cl
	mov dword TOS(3),edx
	mov cl,ch
	lea edx,[eax+1]	; 1
	mov TOS(2),ecx
	mov TOS(1),edx
	ret
dvby.e:; <dv/, 0/0 -> ret 1 - for tg(dv)
	inc ecx
	mov TOS(2),eax
	mov TOS(3),ecx
	mov TOS(1),ecx
	ret
;-
;- dv1+		( dv1 -- dv2 )					'dvonep'	rational
;-	add 1 to dv1
dvonep:GA;:ok:
	call dup4
	call sover2
	call dplus
	jo .o
	call swap2
	jmp sdrop4
.o:A
	spop 4
	call v1dv
	jr dvplus
;-
;- dv1-		( dv1 -- dv2 )					'dvonem'	rational
;-	subtract 1 from dv1
dvonem:GA;:ok:
	call dup4
	call sover2
	call dsub
	jo .o
	call swap2
	jmp sdrop4
.o:A
	spop 4
	call vm1dv
	jr dvplus
;-
;- dv-		( dv1 dv2 -- dv3 )				'dvsub'		rational
;-	subtract dv2 from dv1.
dvsub:GA;:ok:
	call dvnegate	;:continued:
;-
;- dv+		( dv1 dv2 -- dv3 )				'dvplus'	rational
;-	add two double ranums; fast (i.e. less slow), trivial case if den1 = den2.
dvplus:GA;:ok:;<dv1-,dv1+,dfvround,dvsimplify,dv*,...)
;	v1n*v2d + v2n*v1d
; v3 := ------------------
;	    v1d*v2d
	cmp eax,TOS(4)	; 7)n1 5)d1 3)n2 1)d2 
	mov ecx,TOS(5)	; den.lo
	jnz .m		;?; den.hi =/=
	cmp ecx,TOS(1)
	jnz .m		;?; den.lo =/=
; no mul if common denominator:
; 2drop d->q 4swap d->uq 2rot d->q 4rot q+ 4swap qv>dv
	spop
	mov eax,TOS(1)	; 2drop
	sar eax,31
	spush		; d->q
	call swap4	; 4swap
	spush
	mov eax,0
	spush		; d->uq
	call rot2	; 2rot
	spush
	sar eax,31
	spush		; d->q
	call rot4	; 4rot
	jr .q		; q+ 4swap qv>dv
.m:A; 2dup 3 2pick umd* 2rot 5 2roll md* 4rot md* q+ 4swap
	call dup2	; ( 1n 1d 2n 2d 2d
	spush
	mov eax,3
	call pick2	; ( 1n 1d 2n 2d   2d   1d
	call umdmul	; ( 1n 1d 2n 2d   21dl 21dh		; common denominator
	call rot2	; ( 1n 1d 2n 21dl 21dh 2d
	spush
	mov eax,5
	call roll2	; ( 1d   2n   21dl 21dh 2d   1n
	call mdmul	; ( 1d   2n   21dl 21dh 21nl 21nh	; signed den2*num1
	call rot4	; ( 21dl 21dh 21nl 21nh 1d   2n
	call mdmul	; ( 21dl 21dh 21nl 21nh 12nl 12nh	; signed den1*num2
.q:
	call qplus	; ( dl dh nl nh
	call swap4	; ( qnl.nh qdl.dh  ;:continued: qv>dv;	; d1*n2+d2*n1/(d1*d2)
;-
;-	
;-	
;--- quad (and intermediate) size rational numbers for doubles support
;-	
;-

;-
;- qv>dv  	( qv -- dv )					'qv2dv'		rational
;-	returns double which most closely approximates quad ranum.
;-	ret ovf indicator (0,0) if qv doesn't fit into dv.
qv2dv:GA;:ok:;<abranumb, dvmul, continued from <dvplus>
;  qv>uqv? >r				\ |qv| & sgn
	call qv2uqvq
	push eax
	spop
	call qvdround
	mov ecx,eax
	or ecx,TOS(1)
	jnz .x
	or ecx,TOS(4)
	jg .x		;?; > 0 -> quad
	mov edx,ecx
	jnz .g		;?; -ve
	mov ecx,TOS(5)
	dec ecx
.g:
	cmp ecx,byte -1
	jnz .x
	cmp TOS(5),edx
	jnz .x
.r:
;  endif 2rotd 2drop r> dv+-
	spop 2
	call sdrop2
	spush
	pop eax		; sgn
	jmp dvpm
.o:
	lea PS,[PS+ecx*4]
	xor eax,eax
	lea esp,[esp+4]	; drop sgn
	spush 3
	ret
.x:; :kommt::hier::nie: (oder fast nie?) an
	call qvsplit
	mov ecx,TOS(11)
	or ecx,TOS(10)
	mov ecx,-15
;    0if rdrop #16 qov exit endif	\ xit w/ ovf of integer part
	jnz .o		; integer part ovf, conversion not possible
;    31 >r 7 2pick bsrr dup 1+ 
	mov edx,-1
	bsr edx,dword TOS(12)
	mov ecx,32
;    0if rdrop 63 >r swap bsrr endif 1abs
	jnz .b		;?; bits set
	bsr edx,dword TOS(13)
	lea ecx,[ecx+32]
	jnz .b
	neg edx
;    sdrop r> swap - dup 0> + 2^ 0. qvround
.b:
	spush
	sub ecx,edx
	mov eax,ecx
	call pow2	; 2^
	spush
	xor eax,eax
	spush
	call qvround
;    8 nswap 4drop 4over q* q0=
	call swap8
	spop 4
	call over4
	call qmul
	call qzeq
;    0if rdrop #12 qov exit endif	\ xit w/ ovf
	spop
	mov ecx,-11
	jnz .o
;    2over d0< 
	cmp dword TOS(2),byte 0
	jns .a
;    if q2/ 4rot q2/ 4rot q2/ 4rot endif \ last resort...
	call uqdiv2
	call rot4
	call uqdiv2
	call rot4
	call uqdiv2
	call rot4
.a:
;    4rot q+ 4swap
	call rot4
	call qplus
	call swap4
;   endif
	jmp .r
;-
;- qvreduce	( qv -- qv)			(4)				rational hidden
;-	reduce quad ranum by greatest common divisor of numerator and denominator.
qvreduce:GA;:ok:; <dvplus, dvmul)
	call dup8	; 8dup
	call qgcd	; qgcd
	mov ecx,eax
	or ecx,TOS(1)
	or ecx,TOS(2)
	mov edx,TOS(3)
	jnz .d		; quad 1
	cmp edx,byte 2	; 2 u< if... (ovf or unity)
	jc near drop4	;?; no gcd
.d:;	; 11)qvn 7)qvd 3)gcd 
; (mmx alternative was slow)
	call rot4	; qvd gcd qvn
	call over4	; qvd gcd qvn gcd
	call qby	; qvd gcd qvn'
	call mrot4	; qvn' qvd gcd
	call uqbym
	jmp sdrop4	; qvn' qvd'
;-
;- dv>uqv?	( dv -- +qv f)					'dv2uqvq'	rational hidden
;-	return absolute quad value of signed double rational no. tf if dv was negative. 
;-	A ranum is represented on data-stack as ( sign*|numerator| |denominator| ).
dv2uqvq:GA		; 3)nl 2)nh 1)dl 0)dh
	mov ecx,TOS(1)	; dl
	spush 5		; 8)NL:NLL 7)NH:NLH 6)dl:NHL 5)dh:NHH  4)DL:DLL 3)DH:DLH 2)dh:DHL 1)dh:DHH 0)dh:FLG
	mov TOS(4),ecx	; dl,  NLL      NLH      dlh      dlh    DL:DLL   DH:DLH      dlh      dlh      dlh
	mov ecx,eax
	mov edx,0
	mov eax,TOS(7)	; nh, for the flag
	mov TOS(1),edx	; unsigned extd denum
	mov ecx,eax	; save nh for negation
	sar eax,31	; sign flag
	mov TOS(2),edx	;      NLL      NLH      dlh      dlh       DLL      DLH      DLH      DHH      FLG
	mov TOS(6),edx
	mov TOS(5),edx	;      NLL      NLH      DLH      DHL       DLL      DLH      DLH      DHH      FLG
	jns .r		;?; +ve dv
	neg dword TOS(8); nl
	adc ecx,byte 0
	neg ecx
	mov TOS(7),ecx	; nh
.r:
	ret
;-
;-	
;--- Memory, quad ranum
;-

;-
;- qvinteger	ccc( qv -- )(X: -- qv )						rational hidden
;-	quad ranum integer; modifiable constant (four cells L4 variant of a {value}).
qvinteger:GA
	spush
	mov eax,8
	call ninteger
	call pscodep
	call nfetch
	spop
	ret
;-
;- qv@		( p -- dv ) 					'qvfetch'	rational hidden
;-	fetch qv from memory at p.
qvfetch:GA;:ok:
	spush
	mov eax,8*4
	jmp stom
;-
;- qv!		( dv p -- ) 					'qvstore'	rational hidden
;-	store qv to memory at p.
qvstore:GA;:ok:
	spush
	mov ecx,8*4
	jmp mtos
;-
;- FLOAT+	( p1 -- p2 )			(m,3)		'floatp'	ans float
;-	add size of a single float (16) to ptr p1, dpans 12.6.1.1555
floatp:GA M;:ok:
	lea eax,[eax+16]
M floatp
	ret
;-
;- DFLOAT+	( p1 -- p2 )			(m,3)		'dfloatp'	ans float
;-	add size of a double float (32) to ptr p1
dfloatp:GA M;:ok:
	lea eax,[eax+32]
M dfloatp
	ret
;-
;-	
;--- Stacks, quad ranum
;-

;-
;- qvrot	( qv1 qv2 qv3 -- qv2 qv3 qv1 )	(x)(4)				rational hidden
;-	roll 3rd @tos to tos.
qvrot:GA;:ok:
	spush
	mov eax,2
	jmp roll8
;-
;- qv-rot 	( qv1 qv2 qv3 -- qv3 qv1 qv2 )	(x)(4)		'qvmrot'	rational hidden
;-	roll @tos to 3rd @tos.
qvmrot:GA;:ok:
	spush
	mov eax,-2
	jmp roll8
;-
;- qv2dup  	( qv1 qv2 -- qv1 qv2 qv1 qv2 )					rational hidden
;-	duplicate top two quad ranums
qv2dup:GA
	spush
	mov eax,16
	jmp ndup
; quad ranum stack opr synonymous words
    global qvdrop:function,qvdup:function,qvnegate:function,qvover:function
    global qvroll:function,qvsover:function,qvswap:function,qvpick:function
;-
;- qvpick 	( qvn .. n -- .. qvn )		8pick				rational hidden
;-	pick a quad ranum, i.e. 8 cells per item +/- pick
qvpick:equ pick8
;- qvroll 	( qvn .. qv1 n -- .. qv1 qvn )	8roll				rational hidden
;-	roll a quad ranum, i.e. 8 cells per item +/- roll
qvroll:equ roll8
;- qvswap 	( qv1 qv2 -- qv1 qv2 )		8swap				rational hidden
;-	{ 8 nswap }, swap 8 cells items
qvswap:equ swap8
;- qvover 	( qv1 qv2 -- qv2 qv1 qv2 )	8over				rational hidden
;-	pick @nos to tos
qvover:equ over8
;- qvdup  	( qv -- qv qv )			8dup				rational hidden
;-	duplicate quad ranum @tos
qvdup:equ dup8
;- qvdrop  	( qv -- )			8drop				rational hidden
;-	discard top quad ranum
qvdrop:equ drop8
;- qvnegate	( qv1 -- qv2 )			nos-qnegate			rational hidden
;-	negate quad ranum qv1
qvnegate:equ qnegaten
;- qvsover	( qv1 qv2 -- qv2 qv1 qv2 )	8sover				rational hidden
;-	{ qvswap qvover ), 8 cells per item swap and over
qvsover:equ sover8
;-	NOTE: qvnegate, qvsover, qvdup, etc. are synonymous to the rsp. operations 
;-	for eight-cells items. a quad ranum is a pair of quad integers, thus any 
;-	quad or eight-cells integer stack-opr. apply.
;-	re the rsp. aliased word for some further explanation of how to operate.
;-	(all of which involve the mmx registers if L4 was compiled w/ "__MMX" def'd)

;-
;- v>qv		( v1 -- qv2 )					'v2qv'		rational
;-	signed expand v1 to double cells size.
v2qv:GA;:ok:
	mov ecx,TOS(1)	; num
	mov edx,eax
	mov eax,ecx
	sar eax,31
	spush 3
	mov eax,edx
	spush
	mov eax,0
	spush 2
	ret
;-
;- dv>qv  	( dv1 -- qv2 )					'dv2qv'		rational
;-	signed expand dv1 to quad qv2.
dv2qv:GA;:ok:
; (mmx alternative was slow)
	mov ecx,TOS(2)	; num.hi
	mov edx,TOS(1)	; den.lo
	spush 4
	sar ecx,31
	mov TOS(3),edx
	xor eax,eax
	mov TOS(5),ecx
	mov TOS(4),ecx
	mov TOS(1),eax
	ret
;-
;- dv>uqv  	( dv1 -- qv2 )					'dv2uqv'	rational hidden
;-	unsigned expand dv1 to quad qv2.
dv2uqv:GA;:ok:
	spush
	mov eax,0
	spush 3
	jmp mrot2
;-
;- s>qv		( n -- qv )					's2qv'		rational
;-	signed convert single integer to quad ranum
s2qv:GA;:ok:
	spush
	sar eax,31
;-
;- d>qv		( dn -- qv )					'd2qv'		rational hidden
;-	signed convert double integer to quad ranum
d2qv:GA;:ok:
	spush
	sar eax,31
	spush
;-
;- q>qv		( qn -- qv )					'q2qv'		rational hidden
;-	signed convert quad integer to quad ranum
q2qv:GA;:ok:
	spush
	mov eax,1
	spush
	dec eax
	spush 2
	ret
;-
;- n>qv		( n.N -- qv )					'n2qv'		rational hidden
;-	truncate/modify/signed extend counted integer to quad ranum.
n2qv:GA;:ok:
	spush
	mov eax,4
	call nton
	mov eax,1
	spush
	dec eax
	spush 2
	ret
;-
;- qv>fq  	( qv -- q )			(4)		'qv2q'		rational hidden
;-	ret floored signed quad integral part of quad ranum.
qv2q:GA;:ok:
	call qfbym
	jmp sdrop4
;-
;- qv>s  	( qv -- n )					'dv2n'		rational hidden
;-	return the m.s. signed integral component of ranum qv as single integer n.
qv2n:GA;:ok:
	call qv2dn
	mov edx,1
	jr qv2dn.p
;-
;- qv>d  	( qv -- d )					'dv2dn'		rational hidden
;-	return the m.s. signed integral component of ranum qv as double integer d.
qv2dn:GA;:ok:
	call qv2qn
	mov edx,3
	call .p
	mov edx,2
.p:
	test eax,eax
	mov ecx,eax
	jz .m
	inc ecx
	jz .m
	spush
	mov eax,edx
	call roll
.m:
	spop
qvfrac.r:
	ret
;-
;- qv>q  	( qv -- q )					'dv2dn'		rational hidden
;-	return the signed integral component of ranum qv as quad integer.
qv2qn:GA;:ok:
	call qbym
	jmp sdrop4
;-
;- qvsplit 	( qv1 -- qv2 qv3 )						rational hidden
;-	floored integral qv2 and (non-negative) fractional qv3 component of qv1. 
;-	w/ Floored rounding, i.e. v1 = v2 + v3 and assuming +ve v3 is always valid. 
qvsplit:GA;:ok:
	call dup8
	call qqzeq
	spop
	jz qvfrac.r	;?; ovf
	call qv2q
	spush		; q>qv:
	mov eax,1	; quad 1
	spush
	dec eax
	spush 3
	lea eax,[eax+8]	; swap quads
	call nswap	;:continued: qvfrac
;-
;- qvfrac 	( qv1 -- qv2 )			(4)		'qvfrac'	rational hidden
;-	ret +ve fractional part of qv1, remainder of floored div'n.
qvfrac:GA;:ok:
	call qqzeq
	spop
	jz .r	;?; ovf
	call sover4
	call qfbym
	spop 4
	jmp swap4
;-
;- qvabs  	( qv1 -- qv2 )							rational hidden
;-	negate qv1 if -ve
qvabs:GA;:ok:
	cmp dword TOS(4),byte 0
	js near qnegaten
	ret
;-
;- qv>uqv?	( qv1 -- qv2 f)					'qv2uqvq'	rational hidden
;-	return absolute value of signed quad ranum. tf if qv1 was negative. 
qv2uqvq:GA;:ok:
	spush
	mov eax,TOS(5)
	sar eax,31
	jz .p
	not dword TOS(8)
	add dword TOS(8),byte 1
	not dword TOS(7)
	adc dword TOS(7),byte 0
	not dword TOS(6)
	adc dword TOS(6),byte 0
	not dword TOS(5)
	adc dword TOS(5),byte 0
.p:
	ret
;-
;- qvdround	( qv1 -- qv2 )			(X)				rational hidden
;-	returns approximation to qv1 such that qv2 fits into double ranum range.
qvdround:GA;:ok:
	call qvmdv	;:continued: {qvround}
;-
;- qvround 	( qv1 +qn -- qv2 )		(X)				rational hidden
;-	returns an approximation to the rational number qv1 such that neither  
;-	the numerator nor the denominator of its fractional part exceeds the 
;-	positive integer +qn. abort w/ addressing mode error if -ve den of qv1.
qvround:GA;:ok:
; macroes <qlpop>, <qlpush>, <qlget>, <qlput> def'd in "lib4th.mac"
; ------------------------------------------- ;
%undef x
%ifndef pnu
%ifndef dn
%ifndef nu
%ifndef pdn
%ifndef dva
%assign pnu	0	; prev.numerator
%assign dn	pnu+16	; denominator
%assign nu	dn+16	; numerator
%assign pdn	nu+16	; prev.denominator
%assign dva	pdn+16	; qv1.den.hi
.llpnu:	equ	pnu	; 1, prev.numerator
.lhpnu:	equ	.llpnu+4
.hlpnu:	equ	.lhpnu+4
.hhpnu:	equ	.hlpnu+4
.lldn:	equ	dn	; 1, denominator
.lhdn:	equ	.lldn+4	
.hldn:	equ	.lhdn+4	
.hhdn:	equ	.hldn+4	
.llnu:	equ	nu	; 0, numerator
.lhnu:	equ	.llnu+4	
.hlnu:	equ	.lhnu+4	
.hhnu:	equ	.hlnu+4	
.llpdn:	equ	pdn	; 0, prev.denominator
.lhpdn:	equ	.llpdn+4
.hlpdn:	equ	.lhpdn+4
.hhpdn:	equ	.hlpdn+4
.lldva:	equ	dva	; @tos, quad limit
.lhdva:	equ	.lldva+4
.hldva:	equ	.lhdva+4
.hhdva:	equ	.hldva+4
.sg:	equ	dva+16	;sgn from v>uv?	klocv
l_qvround: equ	.sg+4
%define x
%endif
%endif
%endif
%endif
%endif
%ifndef x
%error "qvround:" " pnu" pnu " qdn" qdn " nu" nu " pdn" pdn " dva" dva
%endif
; ------------------------------------- ;		; --------------------------------- ;
;: vround ( v1 +n -- v2 )				; local{ +qn | pn pd nu dn sg }
;  >R  2DUP voverflow					; 4to +qn ?q0= -exit		\ ovf
;  IF  R> DROP						;
;  ELSE  setvars  v>uv? R> 2>R				; setvars qv>uqv? to sg
;   BEGIN DUP						; begin 4dup 
;   WHILE OVER DUP IF /MOD  THEN			; while 4over ?q0= 0= if q/mod endif
;    DUP R@ snum>?  IF next! DONE(1)			;  4dup  +qn snum>? if next! else(1)
;    SWAP R@ sden>? IF next! DONE(2)			;   4swap +qn sden>? if next! else(2)
;    next! SWAP						;    next!
;   REPEAT THEN(2) THEN(1)				; repeat endif(2) endif(1)
;   2DROP R> DROP					; 8 ndrop
;   prev.numerator @ R> ?negate prev.denominator @	; pn sg q+- pd ;
;  THEN ;
; ------------------------------------- ;
;  >R 2dup voverflow if rdrop exit endif
%xdefine _divmod_ uqbym	; qfbym? qbym?
	mov ecx,3*4*4	; min. depth, 3 quad cells
	call qdepthb	; abort if less than 12 bytes on stack
	call over4	; qv den
.s:
	call qzeq	; q0=
	spop
	jz near drop4	;?; ovf, drop limit qn, leave qv
	spush
	mov eax,(l_qvround/4)
	push dword[byte uref uv.s0]
	call toloc	; allocate & clr local mem
	mov ecx,LP
	mov LP,[esp]	; previous @s0, for [lref+disp] w/ +ve disp by <ql..> macroes
	mov [esp],ecx	; save LP from <toloc> for de-allocation
	add LP,KR
	qlpush dva	; sto limit to locmem
; setvars  { sg.dv:=flg prev.numerator:=1 denominator:=1 | numerator:=0 prev.denominator:=0 }
	mov ecx,1	; init p.num & den
	mov [lref .llpnu],ecx
	mov [lref .lldn],ecx
; v>uv? ( v -- uv f ) R> also ans 2>R previous
	call qv2uqvq	; |qv| & flg
	mov [lref .sg],eax; save qv sign
	spop
.enter:;    enter
	jmp .entry
.begin:;    begin	; ( d u			[ limit
;	OVER DUP IF /mod THEN		(1)
	call over4	; over
	call qqzeq	; dup 0=/=
	spop
	jz .endif1	; if
	call _divmod_	; /mod - ( num rem.den/num qot.den/num
.endif1:
;        DUP r@ snum>?
	call dup4	; dup (r@ subst. by local value 'dva')
;snum>? ( u1 u2--u3 f)
;  >R numerator(i:0) * prev.numerator(i:1) + DUP R> U> ;
;  tf if qnum*lnum+lpnum > limit
	qlpop nu	; numerator, from loc to @tos
	call qmul	; q*
	qlget pnu	; overwrite @tos w/ prev.num
	call qplus	; prev.numerator +
	call dup4	; dup
	qlpop dva	; limit
	call qugreat	; r@ ... >r ... r> u>
	test eax,eax
	spop
;	IF next! done -> cont after until
	jnz near .if2
;	else [ 2 cs-roll 2 cs-roll ]
;	    swap r@ sden>? 
	call swap4	; swap
;sden>? ( u1 u2 -- u3 f)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  >R denominator * prev.denominator + DUP R> U> ;
	qlpop dn
	call qmul	; denominator *
	qlget pdn
	call qplus	; prev.denumninator +
	call dup4	; dup
	qlpop dva
	call qugreat	; r@ ... >r ... r> u>
	test eax,eax
	spop
;	    IF next!
	jnz .if3	;?; u>
;	else [ 2 cs-roll 2 cs-roll ] 
;		next! SWAP
	call .nexts
	call swap4
;    entry dup 0= until
.entry:;    entry
;	4dup 0=
	call qqzeq
	spop
.until:;    until
	jnz disp .begin
.endif23:;    endif endif
;    endif endif
;    2drop rdrop
	spop 4		; drop
;   prev.numerator R> +- prev.denominator ;
	qlget pnu	; drop prev.num
	cmp dword[lref .sg],byte 0
	jns .p
	call qnegate
.p:
	qlpop pdn
	spush
	pop LP
	mov eax,-(l_qvround/4)
	jmp toloc
.if2:A
.if3:
	call .nexts
	jr .endif23
.nexts:A
;( +qv)
;  denominator to prev.denominator to denominator
;  numerator   to prev.numerator   to numerator ;
;make the new approximation the current approximation, 
;and the current approximation the previous approximation.
%ifdef __MMX
	movq mm0,[lref .lldn]	; old den
	movq mm1,[lref .hldn]
	movq mm2,[lref .llnu]	; old num
	movq mm3,[lref .hlnu]
	mov [PS],eax
	movq [lref .llpdn],mm0	; prev den
	movq [lref .hlpdn],mm1
	movq [lref .llpnu],mm2	; prev num
	movq [lref .hlpnu],mm3
	movq mm0,TOS(3); new den
	movq mm1,TOS(1)
	movq mm2,TOS(7); new num
	movq mm3,TOS(5)
    	spop 8
	movq [lref .lldn],mm0
	movq [lref .hldn],mm1
	movq [lref .llnu],mm2
	movq [lref .hlnu],mm3
%else
	mov ecx,[lref .lldn]	; lo denominator..
	mov edx,[lref .lhdn]
	mov [lref .llpdn],ecx	;	to prev.denominator
	mov [lref .lhpdn],edx	;
	mov ecx,[lref .hldn]	; hi denominator..
	mov edx,[lref .hhdn]
	mov [lref .hlpdn],ecx	;	to prev.denominator
	mov [lref .hhpdn],edx	;
;
	mov edx,[lref .llnu]	; lo numerator..
	mov ecx,[lref .lhnu]	;
	mov [lref .llpnu],edx	;	to prev.numerator
	mov [lref .lhpnu],ecx	;
	mov edx,[lref .hlnu]	; hi numerator..
	mov ecx,[lref .hhnu]	;
	mov [lref .hlpnu],edx	;	to prev.numerator
	mov [lref .hhpnu],ecx	;
;
	mov ecx,TOS(1)
	mov [lref .hhdn],eax	; hi to denominator
	mov [lref .hldn],ecx
	mov eax,TOS(2)
	mov ecx,TOS(3)
	mov [lref .lhdn],eax
	mov [lref .lldn],ecx
;
	mov eax,TOS(4)
	mov ecx,TOS(5)
	mov [lref .hhnu],eax	; lo to denominator..
	mov [lref .hlnu],ecx
	mov eax,TOS(6)
	mov ecx,TOS(7)
	mov [lref .lhnu],eax
	mov [lref .llnu],ecx
	spop 8
%endif;__MMX
	ret
%undef pnu
%undef dn
%undef nu
%undef pdn
%undef dva
;-
;- qvsimplify	( qv1 +qn -- qv2 )		(4)				rational hidden
;-	returns approximation to the rational number qv1 such that neither the
;-	numerator nor the denominator of the fractional component exceeds qn.
qvsimplify:GA
;  >R vsplit R> vround v+ ;
	call mrot4
	call qvsplit	; qn qin qid qfn qfd
	spush
	mov eax,4
	call roll4
	call qvround
	jmp qvplus
;-
;- qvsqrt 	( qv1 -- qv2 )							rational hidden
;-	qv2 := sgn(qv1) * square root of quad ranum |qv1|.
;-	for instance, accurate within 19 integral plus fractional decimal digits.
;-	apply 'bignum' words to num/den separately for unlimited, higher range and 
;-	praecision:
;-		scale by common factor as great as possible w/o overflowing into 
;-		sign bit of denumerator, get square roots of num and den, reduce.
qvsqrt:GA;:ok:
; => grtmglichen faktor ermitteln, bruch erweitern, wurzeln, krzen.
	call qv2uqvq	; dn qd
	push eax
; max common scaling factor by power of 2
	lea edx,TOS(5)	; addr of m.s.cell of 2nd quad below cell @tos
	call qslbits.n	; max no. of bits to shift left (re-use @tos cell)
	lea edx,TOS(1)
	call qslbits.e	; denom, count m.s. un-set bits
	cmp eax,TOS(1)	; use the smaller scaling 2-s exponent
	mov ecx,eax
	jna .g		;?;
	mov ecx,TOS(1)
.g:
	spop
	push ecx
	mov eax,ecx
	call qlshift	; scale
	call qsqrt	; sqrt(den)
	spop
	call swap4
	spush
	pop eax
	call qlshift	; scale
	call qsqrt	; sqrt(num)
	pop eax
	call qpm	; apply sign
	jmp swap4
;-
;- qv2*		( qv1 -- qv2 )					'qv2mul'	rational hidden
;-	fast multiplication by two.
qv2mul:GA
	test dword TOS(3),1; denum l.s.bit
	mov ecx,TOS(4)
	jz .d		;?; div denum if no loss of praecision
	and ecx,0xc0000000; signed num
	jnz .s		;?;
	test eax,0xc0000000
	jnz .d		;?;
.m:
	call qmul2n	; mul num
	jmp mqvround
.s:A
	jns .d		;?; prevent ovf into -ve
	shl ecx,1
	js .m		;?; mul -ve num
.d:
	call uqdiv2
	jmp mqvround
;-
;- qv2/  	( uqv1 -- qv2 )					'qv2div'	rational hidden
;-	unsigned divide quad ranum by two.
qv2div:GA
	test dword TOS(7),1; num l.s. bit
	jz .d		;?; div num if no loss of praecision
	test eax,0xc0000000
	jnz .d		;?; div num, prevent ovf into -ve denum
	call qmul2	; mul denum
	jmp mqvround
.d:A
	call qdiv2n	; div num
	jmp mqvround
;-
;- 1/qv  	( qv1 -- qv2 )					'qvrecip'	rational hidden
;-	reciprocal
qvrecip:GA;:ok:
	cmp dword TOS(4),byte 0
	jns near swap4
	call qnegate
	call swap4
	jmp qnegate
;-
;- qv/		( qv1 qv2 -- qv3 )				'qvdiv'		rational hidden
;-	divide quad ranums
qvdiv:GA;:ok:
	call swap4
	test eax,eax
	jns qvmul
	call qnegate
	call qnegaten	;:continued: <qvmul>
;-
;- qv*		( qv1 qv2 -- qv3 )				'qvmul'		rational hidden
;-	multiply quad ranums
qvmul:GA;:ok:		; q1n q1d q2n q2d
	call rot4	; q1n q2n q2d q1d
	call uqmul
	spop 4		; q1n q2n qdd
	call mrot4
	call qmul
	spop 4		; qdd qnn
	call swap4
	jmp mqvround
;-
;- qv-		( qv1 qv2 -- qv3 )				'qvsub'		rational hidden
;-	subtract quad ranum qv2 from qv1.
qvsub:GA
	call qvnegate
;-
;- qv+		( qv1 qv2 -- qv3 )				'qvplus'	rational hidden
;-	add quad ranums
qvplus:GA;:ok:
	call .v
	call swap8
	call .v
	call dup4	; q1n q1d q2n q2d q2d	|tos
	spush
	mov eax,3	; 1st den
	call pick4	; q1n q1d q2n q2d q2d q1d
	call qeq
	spop
	jz near .c	;?; common den
	call dup4	; q1n q1d q2n q2d q2d	|tos
	spush
	mov eax,4
	call pick4	; q1n q1d q2n q2d q2d q1n
	call qmul	; q1n q1d q2n q2d q21lq21h
	call q1abs	; q1n q1d q2n q2d q21lq21h'
	call qzeq	; q1n q1d q2n q2d q21l f
	jnz .o		;?; ovf
	mov eax,3
	call pick4	; q1n q1d q2n q2d q21 q1d
	spush
	mov eax,3
	call roll4	; q1n q1d q2d q21 q1d q2n
	call qmul	; q1n q1d q2d q21 q12
	call q1abs
	call qzeq
	jnz .o		;?; ovf
	spop
	call qplus	; q1n q1d q2n q2d qnn
	spush
	mov eax,-3
	call pick4	; qnn q1d q2d
	call uqmul
	call qzeq
	spop
	jnz .d		;?; ovf
.s:; ; recover from single bit ovf
	test eax,eax	; sg(den)
	jns .p		;?; no ovf
	call uqdiv2	; revert from denum ovf into -ve:
	call uqdiv2n	; reduce by 2, i.e. unsigned shift right 1 bit num and den
.p:
	test eax,eax
	jnz .v		;?; round to range of double ranum
	cmp dword TOS(4),eax
	jz near qvreduce
.v:
	jmp qvdround
.d:A; denom prod ovf
	TOSPUSH(9)
.o:; intermediate nom prod ovf
	spop 21		; drop 5 quads
	jmp qvovf	; ret ovf indicator
.c:A;<common denominator
	spop 4		; q1n q1d q2n q2d -- q1n q1d q2n
	call rot4	; q1d q2n q1n
	call qplus
	call swap4
	jr .s		; test/recover from ovf, ret
;-
;- qv1+		( qv1 -- qv2 )					'qvonep'	rational hidden
;-	increment qv1 by one
qvonep:GA;:ok:
	call sover4
	call qplus
	call swap4
	jmp qvreduce
;-
;- qv1-		( qv1 -- qv2 )					'qvonem'	rational hidden
;-	decrement qv1 by one
qvonem:GA;:ok:
	call sover4
	call qsub
	call swap4	; qv1-
	jmp qvreduce
;-
;- qvcmp  	( qv1 qv2 -- qv1 qv2 f )	(e)				rational hidden
;-	leave signum of ranum doubles subtraction.
;-	(e: eflags reflect signum
qvcmp:GA;:ok:
	call over8
	call over8
	call qvsub	;:continued: qvsgn
;-
;- qvsgn  	( qv -- n )			(e)				rational hidden
;-	ret sing signum of qv.
;-	(e: eflags reflect signum
qvsgn:GA;:ok:;<qvcmp)
	call qqvzeq
	TOSPULL(8)
	jz .i	;0= -> 0
	test edx,edx
	lea eax,[eax-1]
	js .r	; -> -1
	inc eax	; ->  1
.i:
	inc eax
.r:
	ret
;-
;- ?qv0=  	( qv -- qv f )			(e)		'qqvzeq'	rational hidden
;-	tf if qv is zero or overflow indicator.
;-	use additional test w/ {?q0=} to determine whether ovf(tf) or valid zero(ff) value.
;-	(e: Z if qv=0 and tf, else NZ; edx:=sign cell
qqvzeq:GA;:ok:
	mov edx,TOS(4)	; n.hh
	mov ecx,TOS(5)	; n.hl
	or ecx,TOS(6)	; n.lh
	or ecx,TOS(7)	; n.ll
	or ecx,edx
	spush
	mov eax,0
	jnz .x
	lea eax,[eax-1]
.x:
	ret
;-
;- qv=		( qv1 qv2 -- f )		(e)		'qveq'		rational hidden
;-	tf if dv1, qv2 (within qv accuracy) equal or overflow.
;-	(e: Z if qv=0 and tf
qveq:GA;:ok:
; 15)n1 11)d1 7)n2 3)d2
	call mrot4	; 15)n1 11)d2 7)n2 3)d1
	call qmul	; n2*d1 octal prod
	call qvswap
	call qmul	; n1*d2
	call osub
	jmp ozeq
;-
;- qv~		( qv1 qv2 -- f )		(e)		'qvaeq'		rational hidden
;-	tf if dv1, qv2 approximately equal or, overflow condition met.
;-	(e: Z if qv=0 or ovf, and tf
qvaeq:GA;:ok:
	call qvsub	; qv diff
	call qvabs
	call qqvzeq	; qv f
	spop
	jz .z
	call qvsplit	; qvi qvf
	call qvswap	; qvf qvi 
	call qqvzeq
	spop 9		; qvf
	jnz .z		;?; non-zero integer
	call qvreduce
	call swap4	; qden qnom
	call uqbym	; 4r 4q
	or eax,TOS(1)
	or eax,TOS(2)
	mov ecx,eax
	jnz .n		; invert Z flag
	not ecx
.n:
	sub eax,ecx
.z:
	mov eax,0	; ff = non-zero dv
	jnz .g		; num =/= 0
	lea eax,[eax-1]
.g:
	TOSPULL(7)
	ret
;-
;- -qvround	( qv1 -- qv2 )					'mqvround'	rational hidden
;-	round quad ranum if numerator or denumerator out of range for double arithmetics
;-	normalize, i.e. reduce the fraction, otherwise.
mqvround:GA;:ok:
	mov ecx,eax
	or ecx,TOS(4)
	jnz .r
	or ecx,TOS(5)
	or ecx,TOS(1)
	jz near qvreduce
.r:
	jmp qvdround
;-
;- dv^s   	( dv1 +n -- dv2 )				'dvpovs'	rational
;-	raise dv1 to the power of non-negative single integer n. 
;-	dv2 := ovf if n < 0 or result out of range.
dvpows:GA;:ok:
	push eax
	spop
	call dv2qv
	spush
	pop eax
	call qvpows
	jmp qv2dv
;-
;- qv^s   	( qv1 +n -- qv2 )				'qvpovs'	rational hidden
;-	raise qv1 to the power of non-negative single integer n. 
;-	qv2 := ovf if n < 0 or result out of range.
qvpows:GA;:ok:
	push eax
	spop
	call v1qv
%if 1
;: qv^n >r 1qv qvswap enter begin r 1 and if qvsover qv* qvswap endif qvdup qv* r> 2/ >r entry r 0= until rdrop 8drop ;
	call qvswap
	jr .n		;?; dv^0
.l:
	call qvsover
	call qvmul
	call qqzeq
	spop
	jz .r
	call qvswap
.m:
	call qvdup
	call qvmul
.n:
	shr dword[esp],1
	jc .l
	jnz .m
.r:
	lea esp,[esp+4]
	spop 8
	ret
%else
	cmp dword[esp],byte 1
	jc .r		;?; dv^0
.l:
	call qvover
	call qvmul
	call qqzeq
	spop 
	jz .r		;?; ovf
	dec dword[esp]
	jg .l
.r:
	lea esp,[esp+4]
	jmp sdrop8
%endif

;-
;- s>dvfactorial	( n -- dv )					's2dvfac'	rational
;-	dv := n!, -1<n<21; dv := 0 if not in range of a double ranum.
;-	(quad: n -- { s>nfactorial dup 5 < if n->q else ndrop 0qv endif } -- qv )
s2dvfac:GA;:ok:
	call s2nfac	; ret bignum type
	dec eax
	jz .v
	dec eax
	jz .r
	call ndrop
	xor eax,eax
	mov TOS(1),eax
.v:
	spush
.r:
	inc eax
	spush
	dec eax
	ret

;-
;-	
;- -- irrationale zahlen, transzendente fu
;-	(ref to source texts & formulae in "l4rational.inc")
;-	
;-
;
; -- ein paar rechenregeln (dumm, was einer so alles vergit...):
;	log(x^y) = y log x;	log(xy)= log x + log y;	log(x/y)= log x - log y;
;	log 1 = 0;		log 0 = -oo;		exp 0 = 1
;	(a^x)^y = a^(xy);	a^(x+y) = a^x*a^y	(1+1/n)^x = ((n+1)/n)^x
;	x^n = (x^2)^(n div 2)*x^(n mod 2)
; -- ln(x)
;		    x^(1/n)-1
;	 ln x = lim ---------, n:=2^m => ln x = lim 2^m (x^(1/(2^m))-1)
;		n->oo (1/n)
; -- exponentialfunktionen
;		-- v^n
;	e^v   = >  ---, x >= 0, e = lim (1+1/n)^n
;		--  n!		    n->oo
;	mit:
;	 v^n        v^(n-1)   v           
;	---- = vn = ------ * ---, n=0 -> vn = 1; n=1: -> vn = v
;	 n!         (n-1)!    n
;
; -
;		-- (x ln a)^n
;	a^x   = >  ---------, x > 0
;		--    n!
; - Htte I pg 61:
;	             2k pi         2k pi
;	1^(-n) =cos ------ + j sin ----- = e^(2k pi j/n); k=0,...,n-1
;	               n             n    
; -
;	             pi(2k+1)         pi(2k+1)
;	-1^(-n)=cos -------- + j sin --------; k=0,...,n-1
;	                n                n    
; -- winkelfunktionen
;	               x^(2n+1)                               x^2n
;	sin x = (-1)^n -------- = cos(pi/2-x); cos x = (-1)^n -----
;	               (2n+1)!                                (2n)!
; - zyklometrische
;	         pi   --            m^n       x-1
;	atg x = --- + >  (-1)^(n+1) ----, m = ---; x=0, x>0 (Htte I pg 80)
;	         4    --            2n-1      x+1
;
;	          x    pi   --   (-1)^(n+1)
;	atan x = --- * -- + >  --------------; |x| > 1
;	         |x|   2    -- (2n+1)x^(2n+1)
;
;	         --        x^(2n+1)
;	atan x = >  (-1)^n --------; |x| < 1
;	         --          2n-1
; - hyperbolische
;	         e^x - e^(-x)   e^x    2            e^x + e^(-x)           e^x - e^(-x)
;	sinh x = ------------ = --- - ---; cosh x = ------------; tanh x = ------------
;	              2          2    e^x                2                 e^x + e^(-x)
; -- gamma
;	           1  _oo_ (1+1/n)^x
;	gamma(x) = -  |  | ---------	\/ x =/= -k, k=0,1,2,...
;	           x   n=1   1+x/n
; --

;-
;- dvsqrt  	( dv1 -- dv2 )							rational
;-	dv2 := square root of double ranum |dv1| * sgn(dv1).
;-	accuracy ok w/ less than 10 decimal digits, -re- {qvsqrt} for higher praecision.
dvsqrt:GA;:ok:
; => groesztmoeglichen faktor ermitteln, bruch erweitern, wurzeln, kuerzen.
	call dv2udvq	; dn dd
	push eax
; max common scaling factor by power of 2
	lea edx,TOS(3)	; m.s.cell of 2nd double below cell @tos
	call dslbits.n	; get no. of m.s. unset bits
	lea edx,TOS(1)
	call dslbits.e
	cmp eax,TOS(1)	; use the smaller scaling exponent
	mov ecx,eax
	jna .g		;?;
	mov ecx,TOS(1)
.g:
	spop
	push ecx
	mov eax,ecx
	call dlshift	; scaling figure
	call udsqrt
	spop
	call swap2
	spush
	pop eax
	call dlshift
	call udsqrt
	pop eax
	call dpm	; apply sign
	call swap2
	jmp dvreduce
;-
;- lg(dv) 	( dv1 -- dv2 )			(4)		'flog'		rational
;-	decimal logarithm of dv1, by dv2 := ln(dv1)/ln(10)
lgdv:GA;:ok:
	call lndv
	call caln10
	call v2dv
	jmp dvby
;-
;- FLNP1	( f1 -- f2 )			(4)				float
;-	f2 := ln (f1+1) dpans EXT 12.6.2.1554 -	not overly useful:
;-	ln(dv) is very accurate, thus FLNP1 exactly equal to the rsp result from FLN.
flnp1:GA;:ok:
	call dvonep	;:continues:
;-
;- ln(dv)  	( dv1 -- dv2 )					'lndv'		rational
;-	dv2 := ln dv1; dv1 > 0, otherwise ret overflow indicator, -re- {ln(qv)}.
lndv:GA;:ok:
;4; dv>qv 8dup q1- enter
	cmp TOS(2),byte 0
	js disp d.ovf	;?; ret ovf if -ve den.
	mov ecx,eax
	or ecx,TOS(1)
	jz disp d.ovf	;?; invalid, ovf
	call dv2uqv	; dv>uqv
.m:; lnqdv ( quad in, double out )
	call lnqv
	jmp qv2dv
lnqv.o:
	spop 8
	jmp qvovf
;-
;- ln(qv)  	( qv1 -- qv2 )					'lnqv'		rational hidden
;-	qv2 := ln qv1; qv1 > 0, otherwise ret overflow indicator.
;-	accurate to 12(+) fractional+integral decimal digits in range 0<qv<10^11.
;-	(bj) calculated w/ infinite product to highest possible accuracy within 
;-	range of quad ranum; terminated at ovf or, x^(1/2^n) == x^(1/2^(n-1)).
;-	-:-
;-		            _oo_       2
;-		ln x = (x-1)|  | -------------, 0<x<oo
;-			     n=1 x^(1/(2^n))+1
lnqv:GA;:ok:
	call qqvzeq	; (edx:=sign cell)
	spop
	jz .o		;?; ln(0) abfangen, ret ovf
	test edx,edx
	js .o		;?; -ve
	call dup8	; ( dvr dvp	; root, prod
	call qvonem	; qv1-
	jr .e
;4; begin 8sover 4sover q+ s->d if nos-qnegate qnegate endif qv2* qv* quad $7fff.ffffffff qvround 
.l:A		;	; ( p r
	call sover8	; ( r p rn.rd -- r p rd.rn rd
	call sover4	; qv1+	; ( r p r+
	call qplus
	test eax,eax	; 1/qv	; ( r p 1/r+
	jns .c
	call qnegaten
	call qnegate
.c:; qv2* qv*
	call qv2mul	; ( r p 2/r
	call qvmul	; ( r p*2/r
.e:;4; entry qvswap 8dup qvsqrt 8sover qv= until 8 ndrop qv>dv ;
	call qvswap	; ( p r
	call dup8	; ( p r r 
	call qvsqrt	; ( p r r'
	call sover8	; ( p r' r r'
	call qveq
	spop
	jnz .l		;?; continue until r = r^-2
	spop 8		; drop root item
	ret
;-
;- FEXPM1	( f1 -- f2 )							float
;-	f2 := e^(f1)-1, dpans EXT 12.6.2.1516 - {FEXP} always more accurate (and faster)
fexpm1:GA
	call epowdv
	jmp dvonem
;-
;- e^dv		( dv1 -- dv2 )					'epowdv'	rational
;-	returns approximation to e^dv1, -re- {e^qv}.
;-	tested in range -43,6 < dv1 < 43,6, accurate within min. +/- 10^-10.
epowdv:GA;:ok:
	call dv2qv
	call epowqv
	jmp qv2dv
;-
;- e^qv		( qv1 -- qv2 )					'epowqv'	rational hidden
;-	returns approximation to e^qv1, tested in range  -43,6 < qv1 < 43,6.
;-	max. deviation +/- 10^-9, best results with |qv|<6, deviation < 10^-10.
;-	-:-
;-			-- x^n
;-		e^x   = >  ---, x >= 0
;-			--  n!
;-	numeric accuracy limited by the last term x/n in
;-		x^n   x^(n-1)    x
;-		--- = ------- * ---
;-		 n!    (n-1)!    n
;-	which terminates the series if x/n cannot be resolved to =/= 0 within 
;-	range of +ve quad ranums and, by max. four cells range factorial, n=33.
epowqv:GA;:ok:
%assign _er_ 6
	call qv2uqvq
	push eax
	push_ibyte(0)	; prepare e^xx disp
; test whether denum dv1 = 1, then return simplified, integral power of e
	spop
	mov ecx,eax	;   den.hh
	or ecx,TOS(1)	;	hl
	or ecx,TOS(2)	;	lh
	or ecx,TOS(4)	;   num.hh
	or ecx,TOS(5)	;	hl
	or ecx,TOS(6)	;	lh
	jnz .v		;?; qv =/= 1
	cmp dword TOS(3),byte 1
	jnz .v
	cmp dword TOS(7),byte 1
; ret cons {qv.e} if dv1 = 1
	jz near .u	; dv=1, ret dv.e
.v:
	spush
	mov eax,_er_	; for reduction by ln(e^6)
	call s2qv
	call swap8
	call dup8	; qv
	call qv2q
	or eax,TOS(1)
	or eax,TOS(2)
	mov ecx,TOS(3)
	spop 4		; drop integral part
	jnz .c		;?; not in range to reduce
.e:
	sub ecx,byte _er_
	js .c		;?; noting to reduce
	inc dword[esp]
	push ecx
	call over8
	call qvsub	; reduce qv by ln(e^6)
	pop ecx
	jr .e
.c:A
	call sdrop8
	call v1qv	; v(n=0)
	call v1qv
; 33 is maximum for factorial in range of quads.
	push_ibyte(33)
	push_ibyte(32)	; 1st iteration := 1
.l:A;	;				( qv1 sum(n-1) v(n-1)
	call qqvzeq
	jz .z		;?; ovf or zero
	mov eax,2
	call pick8	; qv1		( qv1 sum(n-1) v(n-1) qv1
	spush
	mov eax,[esp+4]
	sub eax,[esp]	; n
; -. call s2qv:
	spush
	mov eax,0	; qh ql n 0 0.	( qv1 sum(n-1) v(n-1) qv1 (n)!
	spush 3
	inc eax
	spush
	dec eax
	spush 2
; -'
	call qvdiv	; x/n		( qv1 sum(n-1) v(n-1) x/n
	call qvmul	; x(n)		( qv1 sum(n-1) v(n)
	call qv2dup	;		( qv1 sum(n-1) v(n) sum(n-1) v(n)
	call qvplus	;		( qv1 sum(n-1) v(n) sum(n)
	call qqzeq	;		( qv1 sum(n-1) v(n) sum(n) f
	mov eax,-2	; arg to pick, ovw flag
	mov ecx,TOS(5)
	jz .x		;?; early terminate at ovf condition
	test ecx,ecx
	js .x		;?; terminate at ovf into -ve
	call pick8	;		( .. -- qv1 sum(n) v(n) )
	dec dword[esp]
	jns disp .l
	TOSPUSH(1)
.z:
	TOSPUSH(8)	;
.x:
	spop 17		;		( qv1 sum(n-1) v(n) sum(n) f -- qv1 sum(n-1)
	lea esp,[esp+8]
	call sdrop8
.r:
%xdefine _cv_ cae %+ _er_ %+ d; mul by e^6 or e^10...
	dec dword[esp]	; n*e^6 reduction
	js .s		;?; none
	call _cv_
	call dv2qv	; quad e^6
	call swap8	; ( .. -- qv.e^6 qv.e^dv
.a:
	call qvover
	call qvmul
	dec dword[esp]
	jns .a
	call sdrop8
.s:
	mov ecx,[esp+4]	; sg(dv)
	lea esp,[esp+8]
	test ecx,ecx
	js disp qvrecip
	ret
.u:A;dv1=1 ret e
	spop 8		; drop qv1
	call caed
	call dv2qv	; qv.e
	jr .r
;-
;- 10^dv  	( dv1 -- dv2 )			(4)		'powdv10'	rational
;-	raise ten to the power of dv1.
powdv10:GA;:ok:
;: falog ( f -- 10^f )  [ 10.0e0 fln ] FLiteral f* fexp ;
	spush
	mov eax,10
	call s2dv	;:continued: dv^
;-
;- dv^    	( +dv1 dv2 -- dv3 )		(4)		'dvpowdv'	rational
;-	dv3 := +ve float or -ve integral dv1 to the power of dv2; dv3 := 1 if dv1 = 0.
;-	ambiguous result if -ve dv1 is not an integer, -re- {re,im(v)}.
;-	-re- {qv^}, a^x = e^(x ln a); a=0 | a^x := 1
dvpowdv:GA;:ok:
	call dv2qv
	call rot4
	call dv2qv
	call qvpowqv.n
	jmp disp qv2dv
;-
;- qv^    	( +qv1 qv2 -- qv3 )		(4)		'qvpowdv'	rational hidden
;-	qv3 := +ve float or -ve integer qv1 to the power of qv2; qv3 := 1 if qv1 = 0.
;-	exact result within up to 12 decimal digits, integer + fractional part.
;-	for -ve dv1 get sign of real part and imaginary unit w/ -re- {re,im(v)}, then
;-	procede w/ abs(qv1), apply 're' value and use sign of 'im' as appropriate.
;-	currently, {re,im(v)} n.i, ambiguous result if -ve dv1 is not an integer.
;-	-:-
;-		a^x = e^(x ln a); a=0 | a^x := 1
qvpowqv:GA;:ok:
	call qvswap
.n:
	call qv2uqvq
	push eax
	spop
	call lnqv
	call qvmul
	call epowqv
	pop ecx
	test ecx,ecx
	js disp qvnegate
	ret	
;-
;- re,im(^)	( dv1 dv2 -- |dv1| dv2 n1 n2 )			'reimp'		reational
;-	n.i.
;-	ret |dv1|, n1 sign of dv1^dv2 and n2 = j, for imaginary result of -re- {^dv}.
;-	-:-
;-	re Htte I pg 61:
;-		             2k pi         2k pi
;-		1^(-n) =cos ------ + j sin ----- = e^(2k pi j/n); k=0,...,n-1
;-		               n             n    
;-		             pi(2k+1)         pi(2k+1)
;-		-1^(-n)=cos -------- + j sin --------; k=0,...,n-1
;-		                n                n    
reimp:GA
	mov ecx,TOS(6)
	mov edx,0
	test ecx,ecx
	spush
	lea eax,[edx+3]
	mov edx,TOS(6)	;
	jns .z
	and ecx,eax	; denom mod 4
	lea eax,[eax-4]
	jz .4		;?; re':= -1, im := 0
	dec ecx
	jz .1		;?; re':= 1, im := 1
	dec ecx
	jz .2		;?; re':= -1, im := 0
.3:;			; re':= -1, im := -1
	
.2:A;			; re':= -1, im :=  0
.1:A;			; re':=  1, im :=  1
	
.4:A; a=-1		; re':= -1, im :=  0


.z:A;	+ve exponent	; re :=  1, im :=  0
	lea eax,[edx+1]
	spush
	dec eax

	ret

;
; -- winkel-/hyperbelfunktionen
;

;-
;- deg>rad	( dv1|f1 -- dv2|f2 )				'deg2rad'	rational ans float
;-	convert to radians.
deg2rad:GA;:ok:
	call dv2qv
	call cqvpi180
	call qvmul
	jmp qv2dv
;-
;- rad>deg	( dv1|f1 -- dv2|f2 )				'rad2deg'	rational ans float
;-	convert to sexagesimal degrees.
rad2deg:GA;:ok:
	call dv2qv
	call cqvpi180
	call qvdiv
	jmp qv2dv
;-
;- tg(dv)	( dv1 -- dv2 )			(4)		'tgdv'		rational
;-	tangens, by tg(x) = sin(x)/cos(x)
tgdv:GA;:ok:
; nun ja, das neumodische "tan" sparen Wir Uns hier... s. 'float'
	call dv2qv
	call tgqv
	jmp qv2dv
;-
;- tg(qv)	( qv1 -- qv2 )			(4)		'tgqv'		rational hidden
;-	tangens, by tg(x) = sin(x)/cos(x)
tgqv:GA;:ok:
	call dup8
	call sinqv
	call swap8
	call cosqv
	jmp qvdiv
;-
;- sincos(dv)	( dv1 -- dv2 dv3 )		(4)		'sincosdv'	rational
;-	dv2 := sin(dv1), dv3 := cos(dv1)
sincosdv:GA;:ok:
	call dv2qv
	call dup8
	call sinqv
	call qv2dv
	call mrot4
	call cosqv
	jmp qv2dv
;-
;- sincos(qv)	( qv1 -- qv2 qv3 )		(4)		'sincosqv'	rational hidden
;-	qv2 := sin(qv1), qv3 := cos(qv1)
sincosqv:GA;:ok:
	call dup8
	call sinqv
	call swap8
	jmp cosqv
;-
;- sin(dv)	( dv1 -- dv2 )			(4)		'sindv'		rational
;-	sinus, -re- {sin(qv)}.
sindv:GA;:ok:
	call dv2qv
	call sinqv
	jmp qv2dv
;-
;- cos(dv)	( dv1 -- dv2 )			(4)		'cosdv'		rational
;-	cosinus, -re- {cos(qv)}.
cosdv:GA;:ok:
	call dv2qv
	call cosqv
	jmp qv2dv
;-
;- sin(qv)	( qv1 -- qv2 )			(4)		'sinqv'		rational hidden
;-	ret qv2 sinus of qv1 in rad, by sin(x)=cos(x-pi/2).
sinqv:GA;:ok:
	call cqvpi2	; pi/2
	call qvsub	;:continues: cos(qv)
;-
;- cos(qv)	( qv1 -- qv2 )					'cosqv'		rational hidden
;-	ret qv2 cosinus of qv1 in rad.
;-	        --       x^2n          x^2n     x^(2n-1)     x^2
;-	cos x = > (-1)^n -----; x(n) = ----- = --------- * ---------; n=1..33(or ovf)
;-	        --       (2n)!         (2n!)   (2(n-1))!   2n*(2n-1)
;-	numeric accuracy limited by the last term, x(n), which terminates the series 
;-	if x^2n/(2n)! cannot be resolved to =/= 0 within the range of +ve quad ranum;
;-	and by the max. factorial representable w/ four cells, n! = 33!.
cosqv:GA;:ok:
	call qv2uqvq
	spop
; reduce by k*2pi
	call cqv2pi	; qv 2pi
	call qv2dup	; qv 2pi qv 2pi
	call qvdiv
	call qv2qn
	call q2qv	; qv 2pi int(qv/2pi)=k
	call qvmul	; qv k*2pi
	call qvsub	; qv1
	call dup8
	call qvmul	; qv^2
	call v1qv	; sum(0)
	call v1qv	; v(0), 1st term
; 33 is maximum for factorial in range of quads.
	push_ibyte(33)
	push_ibyte(32)	; 1st iteration n=1
.l:A
	call qqvzeq	;			( qv^2 sum(n-1) v(n-1) f
	jz .z		;?; ovf or zero
	mov eax,2
	call pick8	; qv^2			( qv^2 sum(n-1) v(n-1) qv1
	spush
	mov eax,[esp+4]
	sub eax,[esp]	; n
	lea ecx,[eax*2]	; 2n
	lea eax,[ecx-1]	; 2n-1
	mul ecx		; 2n*(2n-1)
	spush
	mov eax,0	; qh ql nl nh 0.	( qv^2 sum(n-1) v(n-1) qv1 (n)!
	spush 3
	inc eax
	spush
	dec eax
	spush 2
	call qvdiv	; x^2/(2n*(2n-1))	( qv^2 sum(n-1) v(n-1) x^2/(2n*(2n-1))
	call qvmul	; x(n)			( qv^2 sum(n-1) v(n)
	call qv2dup	;			( qv^2 sum(n-1) v(n) sum(n-1) v(n)
	test dword[esp],1
	jnz .p		;?; +ve term
	call qnegaten	; negate summand
.p:
	call qvplus	;			( qv^2 sum(n-1) v(n) sum(n)
	call qqvzeq	;			( qv^2 sum(n-1) v(n) sum(n) f
	mov eax,-2	; pick index, ovw flag
	jz .x		;?; early terminate at ovf condition
	call pick8	; sum(n) to 2nd @tos	( qv^2 sum(n) v(n)
	dec dword[esp]
	jns disp .l
	TOSPUSH(1)	; compensate flg
.z:
	TOSPUSH(8)	; compensate sum()
.x:
	lea esp,[esp+8]
	spop 17		; drop v(v),sum(n+1),flg
	jmp sdrop8	; drop qv^2, leave sum
;-
;- FATAN2 	( f1 f2 -- f3 )			(4)				ans float
;-	f1/f2 = tg f3, i.e. f3 := arc tg (f1/f2), dpans EXT 12.6.2.1489 
;-	ret pi/4 if f1=f2=0
fatan2:GA
	call dvby	;:continued: atg(dv)
;-
;- atg(dv)	( dv1 -- dv2 )					'atgdv'		rational
;-	ret dv2 arcus tangens of dv1 in rad. -re- {atg(qv)}.
atgdv:GA;:ok:
	call dv2qv
	call atgqv
	jmp qv2dv
; <arc tg(1), ret pi/4
atgqv.p:A
	TOSPULL(15)
	jmp cqvpi4.n
;-
;- atg(qv)	( qv1 -- qv2 )					'atgqv'		rational hidden
;-	ret qv2 arcus tangens of qv1 in rad.
;-	         pi   --            m^n       x-1
;-	atg x = --- + >  (-1)^(n+1) ----, m = ---; x=0,x>0; (Htte I)
;-	         4    --            2n-1      x+1
;-	numeric accuracy limited by the last term, which terminates the series
;-	if either the n-th summand is zero within accuracy of a quad ranum or, 
;-	the running index excedes the (arbitrarily chosen) maximum of n = 255.
atgqv:GA;:ok:
; qv>uqv? >r  ?qv0= 0if
	push dword TOS(4); sign
	lea ecx,[kref .s-Y]
	push ecx
	call qvabs
	call qqvzeq	; atg(0) = 0
	spop
	jnz .c		;?; continue, try atg(1)
	ret
.c:; 8dup qv1- ?qv0=
	call dup8
	call qvonem
	call qqvzeq
	spop
; if 8drop qv.pi/4 else
	jz .p		;?; ret atg(1) = pi/4
; 8swap qv1+ qv/ 8dup 8dup qv* 8swap 8dup qv.pi/4 qv+ 8swap
	call swap8
	call qvonep
	call qvdiv	; m = (x-1)/(x+1)	( m
	call dup8
	call dup8
	call qvmul	; m^2
	call qvswap	;			( m^2 m
	call qvdup	;			( m^2 m m
	call cqvpi4
	call qvplus	; pi/4+m, 1st sum	( m^2 m sum(0)
	call qvswap	; v(0)=m		( m^2 sum(0) m(0)
	push_ibyte(2)	; 1st iteration n=2
.l:A;=m^(n-1)/(2(n1-)-1)=m
; 2 8pick qv* qv2dup   2 dup 1 and 0if negate endif 2* 1- s>qv qv/ qv+
	spush
	mov eax,2
	call pick8	; m			( m^2 sum(n-1) m(n-1) m^2
	call qvmul	;			( m^2 sum(n-1) m(n)
	call qv2dup	;			( m^2 sum(n-1) m(n) sum(n-1) m(n)
	spush
	mov ecx,[esp]	; n
	lea eax,[ecx+ecx-1]; 2n-1
	shr ecx,1
	jc .i		;?; (-1)^(n+1) = 1
	neg eax
.i:
	call s2qv
	call qvdiv	; m^(2n-1)/(2n-1)	( m^2 sum(n-1) m(n) sum(n-1) m(n)/(2n-1)
	call qvplus	;			( m^2 sum(n-1) m(n) sum(n)
	call qvdup
	spush
	mov eax,3
	call pick8	;			( m^2 sum(n-1) m(n) sum(n) sum(n) sum(n-1)
	call qveq	;			( m^2 sum(n-1) m(n) sum(n) f
	test eax,eax
	mov eax,-2	;(item to store is at zero index)
	jnz .x		;?; early terminate if no significant change of sum
	call pick8	; replace 2nd w/ sum(n)
	inc byte[esp]	; max 253 iterations
	jnz disp .l	;			( m^2 sum(n) m(n)
.z:
	TOSPUSH(9)	; compensate sum() and flg
.x:
	lea esp,[esp+4]
	spop 17		; drop v(v),sum(n+1),flg
	jmp sdrop8	; drop m^2, leave sum
.s:A
	pop ecx
	test ecx,ecx
	js near qnegaten
	ret
;-
;- asin(dv)	( dv1 -- dv2 )			(4)		'asindv'	rational
;- fasin  	( dv1 -- dv2 )							ans float
;-                               x
;-	arc sin x = arc tg -------------
;-	                   (1-x^2)^(1/2)
asindv:GA;:ok:
	call .s
	jmp atgdv
.s:A
	call dup4
	call dup4
	call dvmul
	call dvnegate
	call dvonep
	call dvsqrt
	jmp dvby
;-
;- acos(dv)	( dv1 -- dv2 )			(4)		'acosdv'	rational
;- facos  	( f1 -- f2 )							ans float
;-                         sqrt(1-x^2)
;-	arc cos x = arc tg -----------
;-	                       x
acosdv:GA;:ok:
	call asindv.s
	call dvrecip
	jmp atgdv
;-	
;-
;- -- hyperbelfunktionen
;-	
;-

;-
;- FSINH	( f1 -- f2 )							ans float
;-	hyperbelsinus, dpans EXT 12.6.2.1617
;-	         e^x - e^(-x)   1          1 
;-	sinh x = ------------ = - ( e^x - --- )
;-	              2         2         e^x  
fsinh:GA;:ok:
	lea ecx,[kref qvsub-Y]
.n:
	push ecx
	call dv2qv
	call epowqv	; e^qv
	call dup8
	call qvrecip	; e^(-qv)
	pop ecx
	call ecx	; qv+ or qv-
	call qv2div
	jmp qv2dv
;-
;- FCOSH	( f1 -- f2 )			(4)				ans float
;-	hyperbelcosinus dpans94 EXT 12.6.2.1494
;-	                 1            1 
;-	f2 := cosh f1 =  - ( e^f1 + ---- )
;-	                 2          e^f1  
fcosh:GA;:ok:
	lea ecx,[kref qvplus-Y]
	jr fsinh.n
;-
;- FACOSH	( f1 -- f2 )			(4)				ans float
;-	area cosh, f2:=ln(f1-sqrt(f1^2-1)), dpans EXT 12.6.2.1477
;-	                sqrt(x^2-1)
;-	arcosh = artanh ----------- = ln(x-sqrt(x^2-1)); |x| > 1
;-	                     x
facosh:GA;:ok:
	call dup4
	call dup4
	call dvmul
	call dvonem
	call dvsqrt
	call dvplus
	call dv2qv
	call lnqv
	jmp qv2dv
;-
;- FASINH	( f1 -- f2 )			(4)				ans float
;-	area sinh, dpans EXT 12.6.2.1487 
;-	                      x 
;-	 arsinh = artanh -----------
;-	                 sqrt(x^2+1)
fasinh:GA;:ok:
	call dup4
	call dv2qv
	call rot4
	call dup4
	call dvmmul	; f1^2
	call qvonep
	call qvsqrt	; sqrt(f1^2+1)
	call qvdiv
	call dfatanh
	jmp qv2dv
;-
;- FATANH	( f1 -- f2 )			(4)				ans float
;-	area tanh, dpans EXT 12.6.2.1491 
;-	           1      1+x
;-	artanh x = - ln ( --- ), |x| < 1
;-	           2      1-x
fatanh:GA;:ok:
	call dv2qv
	call dfatanh
	jmp qv2dv
;-
;- DFATANH	( df1 -- df2 )			(4)				ans float
;-	area tanh, dpans EXT 12.6.2.1491 
;-	           1      1+x
;-	artanh x = - ln ( --- ), |x| < 1
;-	           2      1-x
dfatanh:GA;:ok:
	call dup8
	call qvonep
	call v1qv
	call qvrot
	call qvsub
	call qvdiv
	call lnqv
	jmp qv2div
;-
;- FTANH	( f1 -- f2 )			(4)				ans float
;-	hyperbeltangens, dpans EXT 12.6.2.1626 
;-	         e^x - e^(-x)   e^2x-1
;-	tanh x = ------------ = ------
;-	         e^x + e^(-x)   e^2x+1
ftanh:GA;:ok:
	call dv2mul
	call fexpm1
	call dvdup
	call dvonep
	call dvonep
	jmp dvby
;-
;- dv>ieee 	( f1 -- dn )					'f2ieee'	rational
;- F>IEEE 	( f1 -- dn )					'f2ieee'	ans float
;-	convert floating pt. (double ranum) to IEEE 754 normalized 'double real' format.
;-	simple rounding method, increment significand by one if bit next to l.s.b set.
;-	f1 = +/- infinity represented by num=0 and den=0 or $80..00 (sign bit set), rsp.
f2ieee:GA;:ok:
	mov edx,TOS(1)
	mov ecx,eax
	or ecx,edx
	jz .v		;?; +ve infinity
	shl ecx,1
	jnz .d
	test edx,edx
	jz .v		;?; -ve inf
.d:
	call dv2udvq	; get |dv|
	sar eax,31
	shl eax,11
	push eax	; save sign bit
	mov ecx,TOS(3)	; sg(num)
	spop
	jns .p		;?; +ve
	test ecx,ecx
	js .n		;?; "negative zero"
.p:
	call dvdup
	call dugreat
	spop
	jns near .e	;?; dv \< 1
	push_ibyte(-1)	; init exp
	jr .b
.v:; ; +oo
	or eax,0x7ff00000
	jr .i
.n:; ;<-ve zero)
	mov eax,ecx
	lea esp,[esp+4]
.i:
	TOSPULL(2)
	shl ecx,1
	mov TOS(1),ecx
	ret
.l:A
	inc dword[esp]	; exp := exp+1
	call dv2div
.b:
	call dvdup
	call dugreat	; tf if den>num, i.e. dv <1
	spop
	js .l		;?; procede until dv < 1
	call dv2mul	; dv in range [1,2}
.x:
	add dword[esp],0x3ff; 'bias'
.u:; ;<zero)
	call dvfrac	; +ve, thus 'floored' mode doesn't matter
	mov ecx,[esp]	; exp
	or [esp+4],ecx	; merge exp w/ sign bit
	mov dword[esp],52; mnt bits
	call sover2
	call udbym
	xor eax,eax
	mov TOS(1),eax
	call mrot2	; ( ie3-num den rmd
.s:
	shl dword TOS(1),1
	rcl eax,1	; d2*
	call over2
	call udbym
	shr dword TOS(1),1; get digit
	rcl dword TOS(7),1;
	rcl dword TOS(6),1; into ie3-num
	spop 2
	dec dword[esp]
	jnz .s
; whether rounding last digit...
	shl dword TOS(1),1
	rcl eax,1	; d2*
	call over2
	call udbym
	cmp dword TOS(1),byte 0
	spop 6
	jz .r
	add dword TOS(1),byte 1
	adc eax,0
	test eax,0xfff00000
	jz .r		;?; no ovf
	shr eax,1
	rcr dword TOS(1),1; adjust for propagated Cy
	dec dword[esp+4]
.r:;
	mov ecx,[esp+4]
	lea esp,[esp+8]
	shl ecx,20	; merge sg & exp into bits 52..63
	or eax,ecx
	ret
.e:A;<less than or equal to 1.0)
	push_ibyte(0)	; init exp
	mov ecx,TOS(2)
	mov edx,TOS(3)
	cmp ecx,eax
	jnz .z		;?; dv =/= 1
	cmp edx,TOS(1)
	jz disp .x	;?; dv = 1 or ovf
.z:
	or ecx,edx
	mov ecx,eax
	jz disp .u	;?; zero, exp=0 & mnt = 0, no 'bias'
	or ecx,TOS(1)
	jz disp .x	;?; ovf
	jr .a
.m:A
	dec dword[esp]
	call dv2mul
.a:
	call dvdup
	call dugreat	; tf if den>num, i.e. dv <1
	spop
	jns .m		;?; procede until dv >/ 1
	jmp disp .x
;:tests::ok: w/ dv>ieee and ieee>dv
;	dq -0.0	nasm: == 0.0 !
;	dq 0.0		;0000000000000000:
;	dq 1.0		;3FF0000000000000:
;	dq 0.12345	;3FBF9A6B50B0F27C:
;	dq 1.2345e0	;3FF3C083126E978D:
;	dq 123.45e0	;405EDCCCCCCCCCCD:
;	dq 123.45e1	;40934A0000000000:
;	dq 123.45e2	;40C81C8000000000:
;	dq -1.0		;BFF0000000000000:
;	dq -0.12345 	;BFBF9A6B50B0F27C:
;	dq -1.2345e0	;BFF3C083126E978D:
;	dq -123.45e0	;C05EDCCCCCCCCCCD:
;	dq -123.45e1	;C0934A0000000000:
;	dq -123.45e2	;C0C81C8000000000:

;-
;- ieee>dv 	( dn -- f1 )					'ieee2f'	rational
;- IEEE>F 	( dn -- f1 )					'ieee2f'	ans float
;-	convert IEEE 754 normalized 'double real' to floating pt. (double ranum) format.
;-	valid f1 within range of a double ranum, overflow indicator otherwise.
;-	"negative zero" as (00..00/$80...00) which propagates zero to any other opr.
;-	any "nan" patterns return the overflow indicator (00/00).
;-	"signed infinity" returns +/- maximum.
ieee2f:GA;:ok:
	lea ecx,[eax+eax-0x7fe00000]; clr sgn, sub 'bias'
	mov edx,eax	; sgn
	cmp ecx,0x80000000
	jz .i		;?;  +/- infinity
	sar ecx,21	; signed exponent
	and eax,0x000fffff; hi bits
	jnz .e		;?; mnt =/= 0
	cmp eax,TOS(1)
	jnz .e		;?; mnt =/= 0
	test edx,0x7fffffff
	jz .z		;?; exp = 0
.e:; ;<non-zero
	or eax,0x00100000; add 1
	test edx,edx
	jns .p		;?; +ve mnt
	neg dword TOS(1)
	adc eax,byte 0
	neg eax
.p:
	spush		; dn = num
	mov eax,0
	spush
	mov eax,0x00100000; den
	spush		; expand to num/den to 1 + fraction
	mov eax,ecx	; exp
	neg ecx
	jns .s		;?; -ve
	call pow2	; 2^exp
	spush
	mov eax,1	; d>dv
	spush
	mov eax,0
	jmp dvmul	; mnt*exp
.s:A;<-ve exp)
	mov eax,1	; 1/(2^exp)
	spush
	dec eax		; num = 1.
	spush ;2
	mov eax,ecx	; den := 2^exp
	call pow2
	jmp dvmul	; mnt/exp
.z:A; zero: eax=0, edx=sgn
	spush		; expand dn to dv
	inc eax
	spush
	mov eax,edx	; represent -ve zero by denom w/ only sign bit set
	ret
.i:A; ret +/- max for infinity
	TOSPUSH(2)	; provide for denum
	mov ecx,edx
	xor eax,eax	; pass eax=0, ecx = sign cell
	jmp disp dvby.m
;-
;-	
;- -- quad numeric output
;-	
;-

;-
;- qv.		( qv -- )					'qvdot'		rational hidden
;-	display quad ranum in fractional format (auxilary word, not error free)
qvdot:GA;:ok:
	call sover4
	call qbym
	call qdot
	call qabs
	call swap4
	call qvdround
	call qv2dv
	test eax,eax
	jnz .d
	cmp TOS(1),byte 1
	jz near drop4
.d:
	jmp dvdot
;-
;- qf.		( qv -- )					'qfdot'		rational hidden
;-	display quad ranum in floating pt format (auxilary word, not error free).
qfdot:GA;:ok:
	spush
	mov eax,-1
	spush
	dec eax		;:continued: -> <qfdotj>
;-
;- qf.j		( qv n1 n2 -- )					'qfdotj'	rational hidden
;-	l/r justified display quad ranum in floating pt format (auxilary word).
qfdotj:GA;:ok:
	push eax	; trailing blanks
	push dword TOS(1); leading
	spop 2		; ( qv
	call qv2uqvq
	push eax
	spop
	call sover4
	call uqbym	; uq/mod	( den rmd int
	spush
	pop eax		; sg
	spush
	pop eax
	test eax,eax
	jns .o		;?; +ve, else prevent implied trailing blank
	xor eax,eax
.o:
	call pqdotp	; q.r
	call type
	call vpoint
	call emit	; dp
	push_ibyte(-1)	; digits count
.l:
	inc dword[esp]	;		( den rmd
	call basef
	call uqsmmul
	spop
	call over4
	call uqbym
	spop 3
	call h2a
	call emit
	call fprec
	cmp eax,[esp]
	spop
	jng .e
	call qqzeq	; ?q0=
	spop
	jnz .l		;?; more digits
.e:
	spush
	pop ecx		; sent digits
	pop eax		; max blanks
	mov edx,eax
	sub eax,ecx
	jg .p
	js .t
	test edx,edx
	jz .p
.t:
	mov eax,1
.p:
	call spaces
	spop 8
	ret
;-
;- df#.		( dv -- )					'dfdecdot'	rational
;-	display dv in f.p. format at decimal base
dfdecdot:GA;:ok:
	call dv2qv	;:continued: qf#.
;-
;- qf#.		( qv -- )					'dfdecdot'	rational hidden
;-	display qv in f.p. format at decimal base
qfdecdot:GA;:ok:
	push dword[byte uref uv.base]
	mov dword[byte uref uv.base],10
	call qfdot
	pop dword[byte uref uv.base]
	ret
;-
;- REPRESENT	( f1 p u -- n flg1 flg2 | xx xx 0 )		'frepresent'	rational ans float
;-	decode rational number, dpans 12.6.1.2143:
;-	sto characters representation of f1, integral and #fplaces length fractional 
;-	parts w/o trailing zeroes, w/ implied fractions' marker after 1st character 
;-	plus n places, to buffer(p,u), leave base-10 exponent n, 
;-	flg1 := tf if f1 <0, else 0 and, flg2 := tf for a valid result, 0 otherwise.
;-	NOTE: n is aequivalent to @dpl by numeric input and, in contrary to the 
;-	dpans description of REPRESENT, conversion follows current @base setting.
;-	buffer(p,u) size should be at least 4 chars, ret false flg2, otherwise.
;-	ambiguos condition if f1 is overflow indicator and p = u = 0: flg2 := 1.
frepresent:GA;:ok:
	cmp eax,byte 4
	mov ecx,TOS(1)
	jl near qzeq	;?; too small a buffer, (hopefully) ret xx xx 0
	mov dword[kref ecx],0
	call mrot2	; f1 p u
	call dv2udvq	; dv := |dv| and sign := @tos
	push eax
	spop		;		( |f1| p u	[ sg
	call dvsplit	; |f1| +ve, floored method doesn't matter
	call swap4	;		( p u f1f f1i
	TOSPULL(1)	; leave double int nominator, drop den=1
	xor eax,eax	; no trailing blanks
	call pddotrp	; int to stg	( p u f1f pi ui
	push eax	; integer digits count		[ sg ci
	spush
	mov eax,3	; pick buf(p,u)
	call pick2	;		( p u f1f pi ui p u
	sub eax,TOS(2)	; dec u by no. of digits
	js near .o	;?; buffer ovf			[ sg ci
	push eax	; save rmg count		[ sg ci cb
	spop		;		( p u f1f pi ui p
	call place	; sto leading part to buffer
	mov ecx,TOS(1)	;		( p u f1f
	or ecx,eax	; test for ovf indicator
	push ecx	; save flg			[ sg ci cb ov
	jnz .n		;?; valid number
	mov ecx,TOS(5)	; p
	mov dword[kref ecx],0; discard "0" char from ovf indicator
.n:; round to #fplaces (aka 'PRECISION')
	mov edx,databot
	mov ecx,[kref hprecv-Y]	; ptr to #fplaces
	add ecx,edx
	add edx,[kref vfroundv-Y]; ptr to #fround
	mov ecx,[kref ecx]	; get #fplaces
	push dword[kref edx]	; save #fround		
	push edx	;	and ptr			[ sg ci cb ov n pv
	mov [kref edx],ecx; rounding to #fplaces
	call dvfround	; round to max digits
	pop ecx
	pop dword[kref ecx]; rst #fround		[ sg ci cb ov
	call ltsharp	; <# ( p u f1f
	call sharpdvf	; #dvf
	call sharpgt	; #>
	call mzeroes	; discard trailing zeroes ( p u pf uf
	pop ecx		; ov				[ sg ci cb
	test ecx,ecx	; flg
	jz .f		;?; leave ovf-stg
	spush
	mov eax,2
	call sstring	; discard leading "0,"	( p u pf' uf'
.f:
	pop ecx		; cb				[ sg ci
	sub ecx,eax	; uf
	js .v		;?; buffer ovf			[ sg ci
	spush
	mov eax,TOS(4)	; ( p u pf uf p
	call pplace	; ( p u
; discard leading zeroes
	mov eax,TOS(1)
	call ccount	; ( p ps us
	push eax	; save stg len
	spush
	inc edx		; countbyte
	mov eax,0+'0'
	call skip	; ( p ps us -- p ps' us'
	pop ecx
	sub ecx,eax
	jz .l		;?; no "0"-s
	dec dword TOS(1); leave single leading "0"
	inc eax	
	sub [esp],ecx	; adjust exponent		[ sg ci'
	call rot	; ( ps' us' p
	call place
	spush 3
.l:
	xor eax,eax
	pop dword TOS(2); count of int digits, drop buf-size [ sg
	pop dword TOS(1); sg				[ -/-
	dec eax		; S
	ret	
; (10: p u f1f pi ui p u -- 4:p u f1f.num		[ sg ci
.o:A;<integral digits buffer ovf
	TOSPULL(6)
; (4: p u pf uf -- 3:xx xx 0				[ sg ci
.v:; <fractional digits buffer ovf
	lea esp,[esp+8]; drop count & sg
	TOSPULL(1)
	xor eax,eax	; Z
	ret
;-
;-	
;- -- program flow by double ranum, FOR/NEXT loop
;-	
;-

; items posn's on returnstack w/ ret-address @tor
%xdefine _dh	+16
%xdefine _dl	+12	; mmx denominator double
%xdefine _nh	+8
%xdefine _nl	+4 	; mmx numerator double
; limit:(R: lim:d1h.d1l.n1h.n1l .stp. .idx. (tor)
%xdefine lim	esp+32
; step:	(R: .lim. stp:d2h.d2l.n2h.n2l .idx. (tor)
%xdefine stp	esp+16
; index:(R: .lim. .stp. idx:d3h.d3l.n3h.n3l (tor)
%xdefine idx	esp

;-
;- (for)  	( f1 -- )(R: -- f1 f2 f3 )			'pforp'		novoc
;-	runtime introducing a FOR [ FROM ] [ STEP ] ... NEXT loop
;-	set initial looping range = limit = f1 (item type of 'double ranum').
;-	pushes loop limit dv1 to the return-stack, plus defaults of
;-		start = initial loop index = f3 := 0, step = f2 := 1.
;-	data on returnstack in little endian order, suitable to F@, F! wrt RP@.
;-	-re- {@NX} for orderly access to any levels loop ctrl. data structures.
pforp:GA;:ok:
	mov ecx,[esp]
	mov [esp],eax	; dh1, limit := @tos
	push dword TOS(1);dl1
	push dword TOS(2);nh1
	push dword TOS(3);nl1	(R: lim:dv1:dh.dl:nh.nl	| tor
	spop 4
	push_ibyte(0)	; dh2, step: = 1
	push_ibyte(1)	; dl2
	push_ibyte(0)	; nh2
	push_ibyte(1)	; dh2	(R: lim:dv1:dh.dl:nh.nl stp:dv2:dh.dl.nh.nl
	push_ibyte(0)	; d3h, index: = 0
	push_ibyte(1)	; d3l
	push_ibyte(0)	; n3h
	push_ibyte(0)	; n3l	(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
	push ecx	; ret
	ret
;-
;- (from) 	( dv -- )(R: dv1 dv2 dv3 -- dv1 d2 dv3' )	'pfromp'	novoc
;-	modifies loop start dv3 (initial lop index) in returnstack to dv from data-stack.
;-	loop limit remains unchanged, -re- {(+from)}.
pfromp:GA
; (R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
	mov ecx,TOS(1)	; dl
	mov edx,TOS(2)	; nh
	mov [idx _dh],eax
	mov [idx _dl],ecx
	mov eax,TOS(3)	; nl
	mov [idx _nh],edx
	mov [idx _nl],eax
	spop 4
	ret
;-
;- (+from) 	( dv -- )(R: dv1 dv2 dv3 -- dv1' d2 dv3' )(X)	'ppfromp'	novoc
;-	modifies loop start dv3 := dv and limit dv1 := dv3+dv from data-stack; i.e.
;-	looping range incremented by dv, from pre-set start to by dv displaced limit.
ppfromp:GA;	{ dvdup @lim dv+ 0 @nf dv! }
; (R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
%ifdef __MMX
	mov [PS],eax
	movq mm0,TOS(3)	; nl
	movq mm1,TOS(1)	; dl
	spush 4
	movq mm2,[lim _nl]
	movq mm3,[lim _dl]
	movq [idx _nl],mm0
	movq [idx _dl],mm1
	movq TOS(1),mm3
	movq TOS(3),mm2
	mov eax,[PS]
	call dvplus
	mov [PS],eax
	movq mm0,TOS(3)	; nl
	movq mm1,TOS(1)
	movq [lim _nl],mm0
	movq [lim _dl],mm1
	spop 4
	ret
%else;:ok:
	mov ecx,TOS(1)	; dl
	mov [idx _dl],ecx
	mov [idx _dh],eax
	mov edx,TOS(2)	; nh
	mov ecx,TOS(3)	; nl
	mov [idx _nh],edx
	mov [idx _nl],ecx
	spush
	mov eax,[lim _nl]
	spush
	mov eax,[lim _nh]
	spush
	mov eax,[lim _dl]
	spush
	mov eax,[lim _dh]
	call dvplus
	mov ecx,TOS(1)	; dl
	mov [lim _dl],ecx
	mov [lim _dh],eax
	mov edx,TOS(2)	; nh
	mov ecx,TOS(3)	; nl
	mov [lim _nh],edx
	mov [lim _nl],ecx
	spop 4
	ret
%endif
;-
;- (step) 	( dv -- )(R: dv1 dv2 dv3 -- dv1 d2' dv3 )(X)	'pstepp'	novoc
;-	modifies loop stepping dv2 in returnstack to dv from data-stack.
;-	exchanges limit and initial index if dv -ve.
pstepp:GA;:ok:
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
%ifdef __MMX;:ok:
	mov [PS],eax
	mov edx,TOS(2)
	movq mm0,TOS(3)	; n
	movq mm1,TOS(1)	; d
	test edx,edx	; get sign
	movq [stp _dl],mm1
	movq [stp _nl],mm0
	jns .rr		;?; ret if +ve inc, else xchg limit & initial index
	movq mm0,[idx _dl]
	movq mm1,[idx _nl]
	movq mm2,[lim _dl]
	movq mm3,[lim _nl]
	movq [lim _dl],mm0
	movq [lim _nl],mm1
	movq [idx _dl],mm2
	movq [idx _nl],mm3
.rr:
	spop 4
	ret
%else
	mov ecx,TOS(1)	; dl
	mov edx,TOS(2)	; nh
	mov [stp _dh],eax
	test edx,edx	; get sign
	mov eax,TOS(3)	; nl
	mov [stp _dl],ecx
	mov [stp _nh],edx
	mov [stp _nl],eax
	jns .r		;?; ret if +ve inc, else xchg limit & initial index
	mov ecx,[idx _dl]
	mov edx,[lim _dl]
	mov eax,[idx _dh]
	mov [idx _dl],edx
	mov [lim _dl],ecx
	mov edx,[lim _dh]
	mov ecx,[idx _nl]
	mov [idx _dh],edx
	mov [lim _dh],eax
	mov edx,[lim _nl]
	mov eax,[idx _nh]
	mov [lim _nl],ecx
	mov [idx _nl],edx
	mov ecx,[lim _nh]
	mov [lim _nh],eax
	mov [idx _nh],ecx
.r:
	spop 4
	ret
%endif
;-
;- (next) 	( -- )(R: dv1 dv2 dv3 -- ) 	(X)		'pnextp'	novoc
;-	terminates a FOR [ FROM ] [ STEP ] ... NEXT loop
pnextp:GA;:ok:
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
%ifdef __MMX;:ok:166.b; 38,8s; => no advantage w/ mmx
	mov [PS],eax
	TOSPUSH(8)
	movq mm3,[idx _dl]
	movq mm2,[idx _nl]
	movq mm1,[stp _dl]
	movq mm0,[stp _nl]
	movq TOS(1),mm3
	movq TOS(3),mm2
	movq TOS(5),mm1
	movq TOS(7),mm0
	mov eax,[PS]
	call dvplus
	mov ecx,TOS(1)	; dl
	mov [PS],eax
	or ecx,eax
	mov ecx,[esp]
	jz .oo		;?; exit on overflow
	movq mm0,TOS(3)	; sum.n = new idx
	movq mm1,TOS(1)	; sum.d
	TOSPUSH(4)
	movq mm2,[lim _nl]
	movq mm3,[lim _dl]
	movq [idx _nl],mm0
	movq [idx _dl],mm1
	cmp dword[stp _nh],byte 0
	movq TOS(1),mm3	; lim.d
	movq TOS(3),mm2	; lim.n
	jns .cc		;?; +ve limit
	movq TOS(7),mm2	; swap idx/limit if stepping 'top down'
	movq TOS(1),mm1	; sum.d
	movq TOS(5),mm3	; lim.d
	movq TOS(3),mm0	; sum.n
.cc:
	mov [PS],eax
	call dvlt
	test eax,eax
	spop
	mov ecx,[esp]
	jz .xx		;?; loop end
	add ecx,[ecx]
	mov [esp],ecx
	ret
.oo:A; ovf exit
	spop 4
.xx:A
	lea ecx,[ecx+4]
	lea esp,[esp+3*4*4]
	mov [esp],ecx
	ret
%else;:ok:;151.b; 38,3s
	spush
	mov eax,[stp _nl]
	spush
	mov eax,[stp _nh]
	spush
	mov eax,[stp _dl]
	spush
	mov eax,[stp _dh]
	spush
	mov eax,[idx _nl]
	spush
	mov eax,[idx _nh]
	spush
	mov eax,[idx _dl]
	spush
	mov eax,[idx _dh]
	call dvplus
	mov ecx,TOS(1)	; dl
	mov edx,eax
	or edx,ecx
	jz .o		;?; exit on overflow
	mov edx,TOS(2)	; nh
	mov [idx _dh],eax
	mov [idx _dl],ecx
	mov ecx,TOS(3)	; nl
	mov [idx _nh],edx
	mov [idx _nl],ecx
	spush
	mov eax,[lim _nl]
	spush
	mov eax,[lim _nh]
	spush
	mov eax,[lim _dl]
	spush
	mov eax,[lim _dh]
	cmp dword[stp _nh],byte 0
	jns .c		;?; +ve limit
	call dvswap	; swap idx/limit if stepping 'top down'
.c:
	call dvlt
	test eax,eax
	spop
	mov ecx,[esp]
	jz .x			;?; loop end
	add ecx,[ecx]
	mov [esp],ecx
	ret
.o:A; ovf exit
	mov ecx,[esp]
	spop 4
.x:A
	lea ecx,[ecx+4]
	lea esp,[esp+3*4*4]
	mov [esp],ecx
	ret
%endif
;-
;- LIM		( -- f1 )			(X)		'flim'		rational ans float
;-	get innermost FOR-loop limit.
flim:GA;:ok:
%ifdef __MMX;:ok: but no advantage
	spush 4
	movq mm1,[lim _dl]
	movq mm0,[lim _nl]
	movq TOS(1),mm1
	movq TOS(3),mm0
	mov eax,[PS]
	ret
%else
	spush 4
	mov eax,[lim _nl]	; n3l
	mov ecx,[lim _nh]	; n3h
	mov edx,[lim _dl]	; d3l
	mov TOS(3),eax
	mov TOS(2),ecx
	mov TOS(1),edx
	mov eax,[lim _dh]	; d3h
	ret
%endif
;-
;- END		( -- )				(m,)		'fend'		rational ans float
;-	leave innermost FOR-loop at next xec of NEXT,
;-	{END} executes like f.i.g-style {leave}, leaving the loop at next NEXT.
;-	(forced termination by storeing the overflow indicator to STP value)
fend:GA M;:ok:
	mov dword[stp _dl],0
	mov dword[stp _dh],0
	ret
M,fend
	mov dword[stp _dl-4],0
	mov dword[stp _dh-4],0
M fend
;-
;- !INC		( f1 -- )(R: dv1 dv2 dv3 -- dv1 d2 f1 )		'sinc'		rational ans float
;-	store f1 from data-stack to innermost FOR-loop increment
sinc:GA;:ok:
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
	mov [stp _dh],eax
	mov edx,TOS(1)
	mov eax,TOS(3)
	mov ecx,TOS(2)
	mov [stp _dl],edx
	mov [stp _nl],eax
	mov [stp _nh],ecx
	spop 4
	ret
;-
;- INC  	( -- f1 )			(i)(X)		'fstep'		rational ans float
;-	get innermost FOR-loop increment
fstep:GA;:ok:
%ifdef __MMX;:ok: but no advantage
	spush 4
	movq mm1,[stp _dl]
	movq mm0,[stp _nl]
	movq TOS(1),mm1
	movq TOS(3),mm0
	mov eax,[PS]
	ret
%else
	spush 4
	mov eax,[stp _nl]	; n3l
	mov ecx,[stp _nh]	; n3h
	mov edx,[stp _dl]	; d3l
	mov TOS(3),eax
	mov TOS(2),ecx
	mov TOS(1),edx
	mov eax,[stp _dh]	; d3h
	ret
%endif
;-
;- IDX		( -- f1 )(R: dv1 dv2 dv3 -- dv1 d2 f1 )	(X)	'fidx'		rational ans float
;-	push innermost FOR-loop index dv1 to data-stack, 
;-	loop index in a { FOR .. NEXT } loop.
fidx:GA;:ok:
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
%ifdef __MMX;:ok: but no advantage
	spush 4
	movq mm1,[idx _dl]
	movq mm0,[idx _nl]
	movq TOS(1),mm1
	movq TOS(3),mm0
	mov eax,[PS]
	ret
%else
	spush 4
	mov eax,[idx _nl]	; n3l
	mov ecx,[idx _nh]	; n3h
	mov edx,[idx _dl]	; d3l
	mov TOS(3),eax
	mov TOS(2),ecx
	mov TOS(1),edx
	mov eax,[idx _dh]	; d3h
	ret
%endif
;-
;- !IDX		( f1 -- )(R: dv1 dv2 dv3 -- dv1 d2 f1 )		'sidx'		rational ans float
;-	store f1 from data-stack to innermost FOR-loop index
;-	loop index in a { FOR .. IDX .. NEXT } loop.
sidx:GA;:ok:
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
	mov [idx _dh],eax
	mov edx,TOS(1)
	mov eax,TOS(3)
	mov ecx,TOS(2)
	mov [idx _dl],edx
	mov [idx _nl],eax
	mov [idx _nh],ecx
	spop 4
	ret
;-
;- @NX		( n1 -- p )					'fnx'		rational ans float
;-	get ptr p to indexed, n1-th level nested FOR-loop index. n1 = 0 for innermost loop.
;-	running index: n1 @NX F@, increment: n1 @NX FLOAT+ F@, limit: n1 @NX DFLOAT+ F@.
;-	can be used at runtime to modifying e.g, the loop increment:
;-		{ inc fdup f* inc f- 0 dv>n @nx float+ f! }
fnx:GA
	shl eax,4
	lea eax,[eax*2+eax+4]	; disp to base of indexed loop-data structure
	add eax,esp
	sub eax,KR
	ret
;-
;- NX		( n1 -- f1 )							rational ans float
;-	n1-th level nested FOR-loop index.
;-	n1 = 0 refers to innermost level, as { IDX }.
nx:GA
;(R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret [ .. ] [ (R: lim:d1h.d1l.n1h.n1l stp:d2h.d2l.n2h.n2l idx:d3h.d3l.n3h.n3l ] ret | tor
	spush 3
	shl eax,4
	lea eax,[eax*2+eax+4]	; disp to base of indexed loop-data structure
	mov edx,[eax+lim _nl]	; n3l
	mov ecx,[eax+lim _nh]	; n3h
	mov TOS(3),edx
	mov TOS(2),ecx
	mov edx,[eax+lim _dl]	; d3l
	mov eax,[eax+lim _dh]	; d3h
	mov TOS(1),edx
	ret
;-
;- FOR		(C: -- p f )(X: f1 -- )		(i)				rational ans float
;-	introducing a FOR [ FROM ] [ STEP ] .. [ IDX ][ nn NX ] .. NEXT loop.
;-	cpl: leaves back-ptr p for resolution by NEXT and control flag f.
;-	xec: -re- {(for)} runtime code
;-	further, related words -re- {INC}, {LIM}, {END} and the rsp. "!" and "@" variants.
for:GA;:ok:
	call rrbrac
	TICK pforp
.c:
	call compilec
	call mark
	spush
	mov eax,'forf'
	ret
;-
;- FROM		(C: p1 f -- p2 f )(X: f1 -- )	(i)				rational ans float
;-	immediately after {for} or {step}, modifies loop start to f1 from data-stack.
;-	cpl: checks flg f, modifies back-ptr p to pointing to immediately after {FROM}.
;-	xec: runtime code, -re- {(from)}
from:GA
	spush
	mov eax,'forf'
	call qpairs
	TICK pfromp
.c:
	mov TOS(1),eax
	TOSPULL(1)
	jr for.c
;-
;- +FROM  	(C: p1 f -- p2 f )(X: f1 -- )	(i)		'pfrom		rational ans float
;-	immediately after {for} or {step},
;-	set FOR-loop start index := dv and limit := limit + dv, limit := FOR-arg + dv.
;-	cpl: checks flg f, modifies back-ptr p to pointing to immediately after {+FROM}.
;-	xec: -re- runtime code, "(+from)"
pfrom:GA;:ok:
	spush
	mov eax,'forf'
	call qpairs
	TICK ppfromp
	jr from.c
;-
;- STEP		(C: p1 f -- p2 f )(X: f1 -- )	(i)				rational ans float
;-	immediately after {for} or {from}, {+from}, modifies loop increment to dv.
;-	cpl: checks flg f, modifies back-ptr p to pointing to immediately after {STEP}.
;-	xec: -re- runtime code, "(step)"
step:GA;:ok:
	spush
	mov eax,'forf'
	call qpairs
	TICK pstepp
	jr from.c
;-
;- NEXT		(C: p f -- )			(i)				rational ans float
;-	terminates a FOR [ FROM ] [ STEP ] ... NEXT loop
;-	cpl: checks flg f, resolve backward ptr.
;-	xec: -re- runtime code, "(next)"
next:GA;:ok:
	spush
	mov eax,'forf'
	call qpairs
	TICK pnextp
	call compilec
	call back
	mov dword[uref uv.ucmc],0
	jmp llbrac
;
%ifdef __ANSALIAS
; float voc synonymous words:
	global dtof:function,dftof:function,fstore:function,fmul:function,fplus:function,fsub:function
	global fby:function,fzlt:function,fzeq:function,flt:function,ftod:function,ffetch:function,fcons:function
	global fdepth:function,fdrop:function,fdup:function,fliteral:function,floor:function,fmax:function
	global fmin:function,fnegate:function,fover:function,frot:function,fswap:function,fvari:function
	global dfstore:function,dffetch:function,fpow:function,fdot:function,fabs:function,facos:function
	global falog:function,fasin:function,fatan:function,fcos:function,fonep:function,fonem:function
	global fedot:function,fexp:function,fln:function,flog:function,fsdot:function,fsin:function
	global fsc:function,fsqrt:function,ftan:function,fprec:function,sfstore:function
	global sffetch:function,sfalign:function,sfloatp:function,f2mul:function,f2by:function,ftodf:function
	global dfover:function,dfswap:function,dfrot:function,fpick:function,dfpick:function
;-
;- D>F		( dn -- f1 )			d>dv		'dtof'		ans float
dtof:	equ	d2dv	; dpans 12.6.1.1130  
;-
;- DF>F		( df1 -- f1 )			qv>dv		'dftof'		ans float
dftof:	equ	qv2dv
;-
;- F!		( p -- )			dv!		'fstore'	ans float
fstore:	equ	dvstore	; dpans 12.6.1.1400 
;-
;- F*		( f1 f2 -- f3 )			dv*		'fmulf'		ans float
fmulf:	equ	dvmul	; dpans 12.6.1.1410 
;-
;- F2*		( f1 f2 -- f3 )			dv2*		'f2mul'		ans float
f2mul:	equ	dv2mul
;-
;- F+		( f1 f2 -- f3 )			dv+		'fplus'		ans float
fplus:	equ	dvplus	; dpans 12.6.1.1420 
;-
;- F-		( f1 f2 -- f3 )			dv-		'fsubf'		ans float
fsubf:	equ	dvsub	; dpans 12.6.1.1425 
;-
;- F/		( f1 f2 -- f3 )			dv/		'fby'		ans float
fby:	equ	dvby
;-
;- F2/		( f1 f2 -- f3 )			dv2/		'f2by'		ans float
f2by:	equ	dv2div	; dpans 12.6.1.1410 
;-
;- F1+		( f1 -- f2 )			dv1+		'fonep'		ans float
fonep:	equ	dvonep 
;-
;- F1-		( f1 -- f2 )			dv1-		'fonem'		ans float
fonem:	equ	dvonem
;-
;- F0<		( f1 -- f2 )			dv0<		'fzlt'		ans float
fzlt:	equ	dvzlt	; dpans 12.6.1.1440 
;-
;- F0=		( f1 -- f2 )			dv0=		'fzeq'		ans float
fzeq:	equ	dvzeq	; dpans 12.6.1.1450 
;-
;- F<		( f1 -- f2 )			dv<		'flt'		ans float
flt:	equ	dvlt	; dpans 12.6.1.1460 
;-
;- F>D		( f1 -- dn )			d/		'ftod'		ans float
ftod:	equ	dby	; dpans 12.6.1.1470 
;-
;- F>DF		( f1 -- df1 )			dv>qv		'ftodf'		ans float
ftodf:	equ	dv2qv 
;-
;- F@		( p -- f1 )			dv@		'ffetch'	ans float
ffetch:	equ	dvfetch	; dpans 12.6.1.1472 
;-
;- FCONSTANT	ccc( f1 -- )(X: -- f1 )	(i)	dvconstant	'fcons'		ans float
fcons:	equ	dvconstant; dpans 12.6.1.1492 
;-
;- FDEPTH	(  -- n )			dvdepth				ans float
fdepth:	equ	dvdepth	; dpans 12.6.1.1497 
;-
;- FDROP	( f1 -- )			4drop				ans float
fdrop:	equ	drop4	; dpans 12.6.1.1500 
;-
;- FDUP		( f1 -- f1 f1 )			4dup				ans float
fdup:	equ	dup4	; dpans 12.6.1.1510 
;-
;- F2DUP	( f1 f2 -- f1 f2 f1 f2 )	8dup				ans float
f2dup:	equ	dup8
;-
;- DF2DUP	( df1 df2 -- df1 df2 df1 df2 )	qv2dup				ans float
df2dup:	equ	qv2dup
;-
;- FLITERAL	( f1 -- )		(i)	dvliteral			ans float
fliteral:equ	dvliteral; dpans 12.6.1.1552 
;-
;- FLOOR	( f1 -- f2 )			dv>d				ans float
floor:	equ	dv2d	; dpans 12.6.1.1558 
;-
;- FMAX		( f1 -- f2 )			dvmax				ans float
fmax:	equ	dvmax	; dpans 12.6.1.1562 
;-
;- FMIN		( f1 -- f2 )			dvmin				ans float
fmin:	equ	dvmin	; dpans 12.6.1.1565 
;-
;- FNEGATE	( f1 -- f2 )			dvnegate			ans float
fnegate:equ	dvnegate; dpans 12.6.1.1567 
;-
;- FOVER	( f1 -- f2 )			dvover				ans float
fover:	equ	dvover	; dpans 12.6.1.1600 
;-
;- DFOVER	( df1 -- df2 )			qvover				ans float
dfover:	equ	qvover
;-
;- FROT		( f1 f2 f3 -- f2 f3 f1 )	dvrot				ans float
frot:	equ	dvrot	; dpans 12.6.1.1610 
;-
;- DFROT  	( df1 df2 df3 -- df2 df3 df1 )	qvrot				ans float
dfrot:	equ	qvrot
;-
;- FPICK  	( f1 ... n -- ... f1 )		dvpick				ans float
fpick:	equ	pick4
;-
;- DFPICK  	( df1 ... n -- ... df1 )	qvpick				ans float
dfpick:	equ	pick8
;-
;- FSWAP	( f1 f2 -- f2 f1 )		dvswap				ans float
fswap:	equ	dvswap	; dpans 12.6.1.1620 
;-
;- DFSWAP	( df1 df2 -- df2 df1 )		qvswap				ans float
dfswap:	equ	qvswap
;-
;- FVARIABLE	ccc( -- )(X: -- p )	(i)	dvvariable	'fvari'		ans float
fvari:	equ	dvvariable; dpans 12.6.1.1630 
;-
;- DF!		( qv p -- )			qv!		'dfstore'	ans float
dfstore:equ	qvstore	; EXT 12.6.2.1203 
;-
;- DF@		( p -- f1 )			qv@		'dffetch'	ans float
dffetch:equ	qvfetch	; EXT 12.6.2.1204 
;-
;- F^		( f1 -- f2 )			dv^		'fpow'		ans float
;- F**		( f1 -- f2 )			dv^		'fpow'		ans float
fpow:	equ	dvpowdv	; EXT 12.6.2.1415 
;-
;- F.		( f1 -- )			df.		'fdot'		ans float
fdot:	equ	dvdotf	; EXT 12.6.2.1427 
;-
;- DF.		( df1 -- )			qf.		'dfdotf'	ans float
dfdotf:	equ	qfdot
;-
;- fabs		( f1 -- f2 )			dfabs				ans float
fabsf:	equ	dvabs	; EXT 12.6.2.1474 
;-
;- FACOS	( f1 -- f2 )			acos(dv)			ans float
facos:	equ	acosdv	; EXT 12.6.2.1476 
;-
;- FALOG	( f1 -- f2 )			10^dv		'falogf'	ans float
falogf:	equ	powdv10	; EXT 12.6.2.1484 
;-
;- FASIN	( f1 -- f2 )			asin(dv)			ans float
fasin:	equ	asindv	; EXT 12.6.2.1486 
;-
;- FATAN	( f1 -- f2 )			atg(dv)				ans float
fatan:	equ	atgdv	; EXT 12.6.2.1488 
;-
;- FCOS		( f1 -- f2 )			cos(dv)		'fcosf'		ans float
fcosf:	equ	cosdv	; EXT 12.6.2.1493 
;-
;- FE.	( f1 -- f2 )				f.				ans float
;-	preliminary, non-standard substitute
fedot:	equ	dvdotf	; EXT 12.6.2.1513 
;-
;- FEXP		( f1 -- f2 )			e^dv				ans float
fexp:	equ	epowdv	; EXT 12.6.2.1515 
;-
;- FLN		( f1 -- f2 )			ln(dv)				ans float
fln:	equ	lndv	; EXT 12.6.2.1553 
;-
;- FLOG		( f1 -- f2 )			lg(dv)				ans float
flog:	equ	lgdv	; dpans EXT 115 |12.6.2.1557 
;-
;- FS.		( f1 -- f2 )			f.				ans float
;-	preliminary, non-standard substitute
fsdot:	equ	dvdotf	; EXT 12.6.2.1613 
;-
;- FSIN		( f1 -- f2 )			sin(dv)		'fsinf'		ans float
fsinf:	equ	sindv	; EXT 12.6.2.1614 
;-
;- FSINCOS	( f1 -- f2 )			sincos(dv)	'fsc'		ans float
fsc:	equ	sincosdv; EXT 12.6.2.1616 
;-
;- FSQRT	( f1 -- f2 )			dvsqrt		'fsqrtf'	ans float
fsqrtf:	equ	dvsqrt	; EXT 12.6.2.1618 
;-
;- FTAN		( f1 -- f2 )			tg(dv)				ans float
ftan:	equ	tgdv	; EXT 12.6.2.1625 
;-
;- SF!		( f1 p -- )			dv!		'sfstore'	ans float
sfstore:equ	dvstore	; EXT 12.6.2.2202 
;-
;- SF@		( p -- f1 )			dv@		'sffetch'	ans float
sffetch:equ	dvfetch	; EXT 12.6.2.2203 
;-
;- SFALIGN	( p1 -- p2 )			falign				ans
sfalign:equ	falign	; EXT 12.6.2.2204 
;-
;- SFLOAT+	( p1 -- p2 )			float+		'sfloatp'	ans
sfloatp:equ	floatp	; EXT 12.6.2.2207 
%endif;__ANSALIAS
%endif;__RANUM_I
;-
;-================================================================
;- l4rational.inc <eof>
