;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4rstack.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  latest	: 1.9, 0.16.1+8
;-  switches	: __RSTACK_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __RSTACK_I
%define __RSTACK_I
%define rstackinc
;-
;---------------------------------------------------------------------
;---  returnstack
;-	push/pop to returnstack adds/subtracts <kref> value
;-	2>r, r> r> cannot be used to swap cells! re {rswap}.
;-

;-
;- rp!		( -- )				(fg)		'rpstore'	forth
;-	reset returnstack to @r0 from @r0. - f.i.g.-mode
;-	return to caller, other returnpoints cleared.
;-	pushes return to {quit} below tor, for safety.
rpstore:GA;:ok:
; c: edx
	mov edx,[byte uref uv.r0]
	mov ecx,[esp]
	lea esp,[kref edx]
	lea edx,[kref quit.s-Y]
	push edx
	push ecx
	ret
;-
;- RP!		( p -- )			(a:c)		'arpstore'	ans
;-	reset returnstack to kref'd p. - ANS mode
;-	return to caller, other returnpoints cleared.
;-	(e: flags unchanged)
arpstore:GA
	mov ecx,[RS]
	lea RS,[kref eax]
	spop
	push ecx
	ret
;-
;- >rr		( nn -- nn )( rp: -- nn )	(f6)(m,4)	'torr'		forth
;-	push duplicate tos item to returnstack
torr:GA M;:ok:
; c: ecx
%ifdef __K6__
; tested sequence { quad 111222333444 >r >rr r> rdrop } by about 20% faster!
	lea esp,[esp-4]
	lea edx,[kref eax]
	mov ecx,[esp+4]
	mov [esp+4],edx
	mov [esp],ecx
	ret
M,torr
	lea ecx,[kref eax]
	push ecx
M torr
%else
	pop ecx
M,torr
	push eax
	add [esp],KR
M torr
	jmp ecx
%endif
;-
;- >r		( nn -- )( rp: -- nn )		(M,9)		'tor'		forth
;-	push cell from tos to tor
;-	(e: flags unchanged)
tor:GA M;:ok:
; c: ecx
%ifdef __K6__
	lea eax,[kref eax]	; 8d0403
	mov ecx,[esp]		; 8b0c24
	mov [esp],eax		; 890424
	push ecx		; 51
	spop			; 8d7ffc 8b07
	ret			; c3
M,tor; <tor>+16
	lea eax,[kref eax]	; 8d0403
	push eax		; 50
	spop			; 8d7ffc 8b07
M tor
%else
	pop ecx
M,tor
	lea eax,[kref eax]
	push eax
	spop
M tor
	jmp ecx
%endif
;-
;- r>		( -- n )(R: n -- )		(m,4)		'rfrom'		forth
;-	pop cell ftom tor to tos
rfrom:GA M;:ok:
; c: -/-
%ifdef __K6__
	spush
	mov ecx,[esp]
	mov eax,[esp+4]	
	mov [esp+4],ecx
	add esp,byte 4
	sub eax,KR
	ret
; compilemacro r>
M,rfrom
	spush
	pop eax
	sub eax,KR
M rfrom
%else
	spush
	pop ecx
	pop eax
	sub eax,KR
	jmp ecx
; compilemacro r>
M,rfrom
	spush
	pop eax
	sub eax,KR
M rfrom
%endif
;-
;- r+!		( n -- )(R: n1 -- n2 )		(m,8)		'rps'		forth
;-	add n to @tor
;-	(e: flags after addition)
rps:GA M
	add [RS+4],eax
	spop
	ret
M,rps
	add [RS],eax
	spop
M rps
;-
;- r>>r		( n1 -- n2 )(R: n2 -- n1 )	(m,14)		'rxchg'		forth
;-	exchange cells on tos and on tor
rxchg:GA M;:ok:
	add eax,KR
	mov ecx,[RS+4]
	mov [RS+4],eax
	sub ecx,KR
	mov eax,ecx
M rxchg
	ret
;-
;- rp@		( -- a )			(fg)(m,5)	'rpfetch'	forth
;-	fetch returnstack ptr
;-	programs should not access returnstack by address!
rpfetch:GA M;:ok:
; c: -/-
	spush
	lea eax,[RS+4]
	sub eax,KR
	ret
M,rpfetch
	spush
	mov eax,RS
	sub eax,KR
M rpfetch
;-
;- i		( -- nn )(R: nn -- nn )  	(m,6)				forth
;-	copy loop index to tos
i:GA M
;- r		( -- nn ) (R: nn -- nn )	(m,6)				forth
;- r@		( -- nn ) (R: nn -- nn )			'r'		forth
;-	duplicate & push top item from returnstack to datastack
r:GA M;:ok:
; c: -/-
	spush
	mov eax,[RS+4]
	sub eax,KR
	ret
M,r
M,i
	spush
	mov eax,[RS]
	sub eax,KR
M r
M i
;-
;- j		( -- n2 )(R: n1 n2 -- n1 n2 )	(m,7)				forth
;-	copy next outer loop index to tos
j:GA M;:ok:
; c: --
	spush
	mov eax,[RS+12]
	sub eax,KR
M j
	ret
;-
;- ni		( ni -- nx )(R: nx .. nn -- nx .. nn )	(h4)(m,5)		forth
;-	copy ni-th outer loop index to tos, cnt begins w. zero for {i}
ni:GA M;:ok: w. 0..2
; c: --
	mov eax,[RS+8*eax+4]
	sub eax,KR
	ret
M,ni
	mov eax,[RS+8*eax]
	sub eax,KR
M ni
;-
;- li		( ni -- nx )(R: nx .. nn -- nx .. nn )	(h4)(m,5)		forth
;-	copy li-th outer loop limit to tos, cnt begins w. zero for innermost loop.
li:GA M;:ok: w. 0..2
; c: --
	mov eax,[RS+8*eax+8]
	sub eax,KR
	ret
M,li
	mov eax,[RS+8*eax+4]
	sub eax,KR
M li
;-
;- rdrop  	(R: n1 -- )			(f6)(m,3)			forth
;-	drop cell from return-stack
rdrop:GA M;:ok:
; c: ecx
%ifdef __K6__
	mov ecx,[esp]
	mov [esp+4],ecx
	lea esp,[esp+4]
	ret
%else
	pop ecx
	lea esp,[esp+4]
	jmp ecx
%endif
M,rdrop
	lea esp,[esp+4]
M rdrop
;-
;- rswap  	(R: n1 n2 -- n2 n1 )						forth
;-	swap cells on return-stack
rswap:GA;:ok:
; c: ecx,edx
	mov ecx,[esp+4]
	mov edx,[esp+8]
	mov [esp+8],ecx
	mov [esp+4],edx 
	ret
;-
;- !r		( p -- )(R: n -- )		(m,8)		'storer'	hidden
;-	store kref(n) - n+KR - to ptr p, remove both from rsp. stacks.
storer:GA M;:ok:
	pop ecx
M,storer
	pop dword[kref eax]
	spop
M storer
	jmp ecx
;-
;- r+		( -- n )			(f6)(m,)	'rplus'		hidden
;-	! while compiling and, for 2nd level compiling words, only !
;-	fetch n from code-space and advance beyond cell where next level return-ptr 
;-	points to; i.e. pass an execution token - or other data - to next level word.
;-	does{ r> r> dup cell+ >r swap >r @ }
;-	e.g, to passing different execution tokens to a generalized frame-word:
;-		: frame ..words..  r+ execute  ..words.. ;
;-		: caller ... frame [ ' .s 4, ] ... ; 
;-	then, {caller} will execute {frame} with the tick-ed {.s} in between.
rplus:GA M; macro -?-
; c: -/-
	spush
	mov eax,[esp+4]
	add dword[esp+4],byte 4
	mov eax,[eax]
	ret
M,rplus
	spush
	mov eax,[esp]
	add dword[esp],byte 4
	mov eax,[eax]
M rplus
;-
;-
;------------------------------------------------------------------------
;---  doubles
;-	cell order always(!) little endian, l.s.cell @tor,
;-	maintained by store+increment PS and decrement+store RS.
;-
;-	for ANS-Forth compliance, the only exceptions being the ordered 
;-	double cells storeing/fetching words 2@, 2!, 2>R, 2R@, 2R>, which by
;-	that standard were justified with some silly tricks programmery, 
;-	combining push/pop with an implied swap was 'natural programming..' 
;-	to those highly admired, deeply respected, "experts"...
;-	L4 mode:
;-		2>r and >r >r	PS:[ n2 n1 -- ]top - RS:[ -- n1 n2 ]top
;-		2r@		PS:[ -- n2 n1 ]
;-	ANS mode:
;-		2>r		PS:[ n2 n1 -- ]top - RS:[ -- n2 n1 ]top
;-		>r >r		PS:[ n2 n1 -- ]top - RS:[ -- n1 n2 ]top
;-	
;-	
;- 2R>		(R: n1 n2 -- )( -- n2 n1 )	(m,11)		'arfrom2'	ans
;-	in reversed cells order pop double from tos to returnstack,
;-	use w/ {ans} words 2>R and 2R@.
;-	DPANS94 17.6.2.0340: @tos corresponds to @tor, cells order on RS reversed.
;-	DPANS94  A.6.2.0340: justification of cells order is plain nonsense...
;-	ANS-mode push/pop vs l4/4th is { swap 2>r .. 2r@ swap ..  2r> swap }
arfrom2:GA
; c: ecx,edx
	pop edx		;ret
M,arfrom2
	spush 2
	pop eax		;hi (!falschrum!)
	pop ecx		;lo
	sub ecx,KR
	sub eax,KR
	mov TOS(1),ecx
M arfrom2
	push edx
	ret
;-
;- 2R@		(R: n1 n2 -- )( -- n2 n1 )			'ar2f'		ans
;-	copy double from tos to returnstack - in reversed cells order! 
;-	use w/ {ans} words 2>R and 2R>. ANS-mode push/pop vs l4/4th -re- {2R>}
ar2f:GA
; c: ecx
	spush 2
	mov eax,[RS+4]	;lo
	mov ecx,[RS+8]	;hi
	sub eax,KR
	sub ecx,KR
	mov TOS(1),ecx
	ret
;-
;- 2>R		( n1 n2 -- )(R: -- n2 n1 )			'ator2'		ans
;-	in reversed cells order push double from tos to returnstack,
;-	use w/ {ans} words 2R@ and 2R>. ANS-mode push/pop vs l4/4th -re- {2R>}
ator2:GA
; c: ecx,edx
	mov edx,TOS(1)
	mov ecx,[esp]	; ret
	add eax,KR
	add edx,KR
	mov [esp],edx	; >r (hi)
	push eax	; >r (lo)
	push ecx
	spop 2
	ret
;-
;- 2>r		( d -- )(R: -- d )				'tor2'		forth
;-	push double from tos to returnstack, PS: n1 n2 -- RS: -- n1 n2
;-	F8: runtime of {do}, does{ >r >r }
tor2:GA;:ok:
; c: ecx,edx
	mov edx,TOS(1)
	mov ecx,[esp]	; ret
	add eax,KR
	add edx,KR
	mov [esp],eax	; >r (hi)
	push edx	; >r (lo)
	push ecx
	spop 2
	ret
;-
;- 2>rr		( d -- d )(R: -- d )		(m,10)		'torr2'		forth
;-	copy double from data-stack to to return-stack,	{ 2dup 2>r }.
torr2:GA;:ok:
; c: ecx
	pop ecx
M,torr2
	push eax
	add [esp],KR
	push dword TOS(1)
	add [esp],KR
M torr2
	push ecx
	ret
;-
;- 2r>		( -- d )(R: d -- )		(m,11)		'rfrom2'	forth
;-	does{ r> r> }
rfrom2:GA M;:ok:
; c: ecx,edx
	pop edx		; ret
M,rfrom2
	spush 2
	pop ecx
	pop eax
	sub ecx,KR	;lo->nos
	sub eax,KR
	mov TOS(1),ecx	;hi->tos
M rfrom2
	push edx
	ret
;-
;- 2r>>2r  	( d1 -- d2 )(R: d2 -- d1 )			'r2xchg'	forth
;-	exchange doubles on tos and on tor
r2xchg:GA;:ok:
; c: ecx,edx
	add eax,KR
	mov ecx,[esp+8]
	mov edx,[esp+4]
	mov [esp+8],eax
	sub ecx,KR
	sub edx,KR
	mov eax,TOS(1)
	add eax,KR
	mov TOS(1),edx
	mov [esp+4],eax
	mov eax,ecx
	ret
;-
;- 2r		( -- d ) (R: d -- d )		(m,9)		'r2'		forth
;-	copy double from returnstack to datastack, does{ r> r over >r }.
;-	NOTE: consistent, little endian cells & bytes order; -re- {2R@} ANS-voc.
r2:GA M;:ok:
; c: -/-
	spush 2
	mov eax,[RS+4]	;lo
	sub eax,KR
	mov TOS(1),eax
	mov eax,[RS+8]	;hi
	sub eax,KR
	ret
M,r2
	spush
	mov eax,[esp]
	sub eax,KR
	spush
	mov eax,[esp+4]
	sub eax,KR
M r2
;-
;- 2rdrop  	(R: n1 n2 -- )			(f6)(m,3)	'rdrop2'	forth
;-	drop a double from return-stack
rdrop2:GA M;:ok:
; c: ecx
	pop ecx
	lea esp,[esp+8]
	jmp ecx
M,rdrop2
	lea esp,[esp+8]
M rdrop2
;-
;- 2rswap  	(R: d1 d2 -- d2 d1 )		(m,20)(X)	'rswap2'	forth
;-	swap doubles on return-stack (macro w/ __MMX mode, only)
rswap2:GA M;:ok:(mmx-mode)
%ifdef __MMX
; c: m0,mm1
	movq mm0,[esp+4]
	movq mm1,[esp+12]
	movq [esp+12],mm0
	movq [esp+4],mm1
M rswap2
	ret
%else
%undef MD
	mov ecx,[esp+4]
	mov edx,[esp+12]
	mov [esp+12],ecx
	mov [esp+4],edx
	mov ecx,[esp+8]
	mov edx,[esp+16]
	mov [esp+16],ecx
	mov [esp+8],edx
	ret
%endif
;-
;- 4>r		( q -- )(R: -- q )				'tor4'		forth
;-	push quad from tos to returnstack, PS: n1 n2 n3 n4 -- RS: -- n1 n2 n3 n4
tor4:GA;:ok:
; c: ecx,edx
	mov edx,TOS(1)	; hl
	mov ecx,[esp]	; ret
	add eax,KR
	add edx,KR
	mov [esp],eax	; hh >r
	push edx	; hl >r
	mov eax,TOS(2)	; lh
	mov edx,TOS(3)	; ll
	add eax,KR
	add edx,KR
	push eax	; lh >r
	push edx	; ll >r
	push ecx	; ret
	spop 4
	ret
;-
;- 4r>		( -- q )(R: q -- )				'rfrom4'	forth
rfrom4:GA;:ok:
; c: ecx,edx
	spush 4
	pop ecx		; ret
	pop edx		; ll
	pop eax		; lh
	sub edx,KR
	sub eax,KR
	mov TOS(3),edx
	pop edx		; hl
	mov TOS(2),eax
	mov eax,[esp]	; hh
	sub edx,KR
	sub eax,KR
	mov [esp],ecx
	mov TOS(1),edx
	ret
;-
;- 4r		( -- q ) (R: q -- q )				'r4'		forth
;-	copy double from returnstack to datastack, does{ r> r over >r }.
;-	NOTE: consistent, little endian cells & bytes order; -re- {2R@} ANS-voc.
r4:GA;:ok:
; c: ecx,edx		; R: hh hl lh ll
	spush 4
	mov eax,[esp+4]	; ll
	mov ecx,[esp+8]	; lh
	mov edx,[esp+12]; hl
	sub eax,KR
	sub ecx,KR
	sub edx,KR
	mov TOS(3),eax	; ll
	mov TOS(2),ecx	; lh
	mov eax,[esp+16]; hh
	mov TOS(1),edx	; hl
	sub eax,KR
	ret
;-
;- 4rdrop  	(R: q -- )			(f6)(m,3)	'rdrop4'	forth
;-	drop four cells from return-stack
rdrop4:GA M
; c: ecx
	mov ecx,[esp]
	mov [esp+16],ecx
M,rdrop4
	lea esp,[esp+16]
M rdrop4
	ret
;-
;- m>r		( p n -- )(R: -- n1..nn )			'mtor'		forth
;-	little-endian fetch n bytes(!) to tor, leaving data @RS un-changed,
;-	restore true figures w. {4th>abs} when fetched back with {r>}, etc.
mtor:GA;:ok:
; c: ecx,edx
	mov ecx,eax	; count
	lea eax,TOS(1)	; save tos+cell
; xchg [eax],esi,edx; kref(src), save count
	mov edx,[eax]
	mov [eax],esi
	mov esi,edx
	lea edi,[esp+3]
	mov edx,[esp]	; ret
	mov dword[esp],0
	sub edi,ecx
	and edi,-4
	add esi,KR	; src
	mov esp,edi	; storage space
	rep movsb
	mov esi,[eax]	; reg back
	lea PS,[eax-4]
	push edx
	mov eax,[PS]
	ret
;-
;- r>m		( p n -- )(R: n1..nn -- )			'rtom'		forth
;-	little-endian store n bytes(!) from tor to memory location p,
;-	beginning w. cell aligned n1; sto by true values.
rtom:GA;:ok:
; c: ecx,edx
	mov ecx,eax	; count
	pop edx		; ret
	lea eax,TOS(1)
	mov edi,[eax]	; dst
	add edi,KR
	mov [eax],esi	; save reg
	mov esi,esp
	rep movsb
	lea esi,[esi+4]
	and esi,-4	; -align
	mov esp,esi	; drop data from rs
	mov esi,[eax]
	lea PS,[eax-4]
	push edx
	mov eax,[PS]
	ret
%endif;__RSTACK_I
;-
;-=====================================================================;
;- l4rstack.inc <eof>
