;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4smem.inc
;-  created	: 14-jul-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __SMEM_I
;-  assembler	: nasm 0.98-3e
;-  description	: memory & blockfile buffers management
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: doc re "glossary.txt" and "l4...html" in ./doc dir
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-

%ifndef __SMEM_I
%define __SMEM_I
%define smeminc
	
;-	
;-
;-================================================================
;- -- memory
;-

;-
;-----------------------------------------------------------------
;- -- access mode & flags
;-

;-
;- prot-r  	( -- n )			(c)(m,6)	'protr'		linux
;-	enable reading memory access, OR-able flag to {mprotect}
protr:GA M;:ok:
	spush
	mov eax,PROT_READ
M protr
	ret
;-
;- prot-w  	( -- n )			(c)(m,6)	'protw'		linux
;-	enable writing memory access, OR-able flag to {mprotect}
protw:GA M;:ok:
	spush
	mov eax,PROT_WRITE
M protw
	ret
;-
;- prot-x  	( -- n )			(c)(m,6)	'protx'		linux
;-	enable executing memory access, OR-able flag to {mprotect}
protx:GA M;:ok:
	spush
	mov eax,PROT_EXEC
M protx
	ret
;-
;- page-size	( -- n )			(c)(m,6)	'pagesize'	linux
;-	Linux PAGE_SIZE system cons (4K w/ kernel 2.2.xx)
pagesize:GA;:ok:
	spush
	mov eax,PAGE_SIZE
	ret
;-
;- mprotect	( a n1 n2 -- er )		(s)		'mprot'		linux
;-	set memory access mode, range n1 at true address a, access mode n2.
;-	range and address expected page aligned. disabled for linux user "root".
;-	syscall -re- man (2) mprotect
mprot:GA;:ok:
	call uidf	; user id
	dec eax
	mov ecx,TOS(2)	; range
	mov edx,TOS(1)	; flags
	spop 3
	js near rtrue	;?; xit if 'root'
	push ebx
	mov ebx,eax	; base address
	sys_mprotect	; b=address, c:=range, d:=flags
	pop ebx
	ret
;-	

;-
;-----------------------------------------------------------------
;- -- common resources access routines -:preliminary:-
;-	caller should not store KR (ebx reg.) to RS immediately prior
;-	to call to one of below basic resources related calls.
;-

;other memory management:
;	mremap, (gut am offenen ende, mittendrin gibts probleme)
;	--
;			MREMAP_MAYMOVE	1	(linux/mman.h)
;	ml{,un}ock{,all}
;	-- (nur) sys_mlockall:
;			MCL_CURRENT     1  lock all current mappings
;			MCL_FUTURE      2  lock all future mappings
    a
;	; handler block, by init moved to ram w. true addr
hphnd:	equ $+handler.lxint-handler.lio
hphnd.lio:	dd 0	; ptr to next hnd, zero at end of chain
hphnd.aio:	dd mmresiz-Y
hphnd.aopen:	dd mmalloc-Y
hphnd.aclos:	dd mmrechp-Y
	times handler.dflen-handler.aclos db -1
hphnd.name:
	db 4
	db 'heap',0	; device name dummy
	dd hphnd-Y
;
hpfree0: dd 0	; base of free blocks list
hpused0: dd 0	; base of used blocks list
;- --
;- <rel4hnd>	( -- )				(k)
;-	push flag to notify that ptr rel kref passed into hnd routine to RS,
;-	should be called immediately before call to handler opr as of open/close/io.
;-	data-stack, regs & eflags preserved
rel4hnd:A
	_xchg KR,dword[esp]
	push KR
	mov KR,[esp+4]
	ret
; ret after hnd opr and convert @nos to kref relative
; stacks, other regs & eflags preserved
hndrel0:A
	jnz .r		;?; eax is error code
	not eax
	lea eax,[kref eax]
	not eax
.r:
	ret
; ret after hnd opr and convert @tos to kref relative
; other stacked data, regs & eflags preserved
hndrel1:A
	push eax
	not ebx
	mov eax,TOS(1)
	lea eax,[kref eax+1]
	mov TOS(1),eax
	not ebx
	pop eax
	ret
;-	
;- <utlink>
;-	link rel. ptr to linked list
utlink:A
; A1 @ A0 ! A0 A1 !
; i:	ecx	A0
;	edx	A1
	push dword[kref edx]
	mov [kref edx],ecx
	pop dword[kref ecx]
	ret
;- <utunlink>
;-	unlink rel. ptr from linked list
; i:	ecx	A0 base-addr
;	edx	A1 link-addr
; enter begin A1 @ to A1 entry A1 @ A0 = until A0 @ A1 ! 
	lea eax,[eax+2*eax];(alignment)
utunlnk.l:
	mov edx,[kref edx]
utunlnk:A
	cmp dword[kref edx],byte 0
	jz .r		;?; base ptr
	cmp ecx,[kref edx]
	jnz .l
	mov ecx,[kref ecx]
	mov [kref edx],ecx
.r:
    ret
; Z: main job or root or owner is owner, NZ: EACCESS, other owner
; i:	eax	ptr to heap header
; o:	eax	eax | EACCESS
;	edx	handler base-address | edx
tsthnd:A
	push ecx	; reg
	push eax	; heap-block address
	mov ecx,[eax+hphead.owner]
	cmp ecx,[uref uv.upid]
	jz .r		;?; main job
	sys_getpid	; current job
	test eax,eax
	jz .r		;?; root
	cmp eax,[ecx+hphead.owner]
	jz .r
	mov dword[esp],EACCES;?; not owner of resource
.r:
	pop eax
	pop ecx
	ret
;-
;-	
;- ----------------------------------------------------------------
;- -- basic handler access
;-	receiving args and returning result in data-stack,
;-	args passed to/received from handler routines by regs eax,ecx
;-	
;- -- common to all routines:
;-	in:
;-		eax	data input
;-		@tos	2ndary data input from data-stack
;-	out:
;-		eflags	Z or, error condition: NC & NZ, eax "above" -4096 (i.e. unsigned greater)
;-		eax	multiple args: 0 or error code, single arg: error code or return value
;-		@tos	optional, other return values
;-	

;-

;-
;- hndopen	(u1 a1 -- a2 er )		(k)
;-	allocate system resource by handler base address a1 and specific data u1
;-	ret a2 adress of item definition block and er error code.
hndopen:A;:ok:;<allocatem)
;:preliminary:!
	cmp [esp+4],KR
	jnz .a		;?; abs address passed
	add eax,KR
	add dword[esp+4],hndrel1-Y
.a:
	push ecx
	mov ecx,[eax+handler.aopen]
	test ecx,ecx
	jz .r		;?; ok if no 'open' entry in handler defn
	cmp ecx,YE-Y
	ja .c
	add ecx,KR
.c:
	mov eax,TOS(1)	; pass u1
	_xchg dword[esp],ecx
	call [esp]
	cmp eax,-4096
	ja .r		;?; error
	mov TOS(1),eax
	xor eax,eax	;(eflags!)
.r:
	pop ecx
	ret
;-
;- hndclose	( a -- 0 | er )			(k)
;-	call de-location routine of a device handler
;-	by a address of beginning of heaphead struc.
hndclose:A;:ok:<freem)
	cmp [esp+4],KR
	push edx
	jnz .a		;?; abs address passed
	add dword[esp+8],.r-Y
	add eax,KR
.a:
	call tsthnd
	jnz .r
	mov ecx,[eax+hphead.drivr]
	test ecx,ecx	; handler
	jz .q		;?; ok if no ptr to handler (non-removable system memory, etc)
	mov ecx,[ecx+handler.aclos]
	test ecx,ecx
	mov dword[eax+hphead.drivr],0; clr hnd
	mov dword[eax+hphead.owner],0; clr owner
	jz .q		;?; ok if no 'close' entry in handler defn
	cmp ecx,YE-Y
	ja .c
	add ecx,KR
.c:
	_xchg dword[esp],ecx
	call [esp]	; routine receives -all- registers, unchanged
	mov ecx,eax
.q:
	mov [esp],ecx
	pop eax
.r:
	cmp eax,-4096
	ja .x
	xor eax,eax	;eflags!
.x:
	ret
;-
;- hndio  	( a1 n1 -- a2 er )		(k)
;-	perform an operation coded by n1 on data item pointed to by p1
;-	eax=n1, TOS(1)=a1, edx=op-code
hndio:A;:ok:;<resizem)
	cmp [esp+4],KR
	jnz .a		;?; abs address passed
	add TOS(1),KR	; convert @no to true address
	add dword[esp+4],hndrel1-Y; provide back-conversion
.a:
	push eax	; save @tos = n1
	mov eax,TOS(1)	; @tos := address
	call tsthnd	; whether hnd present and eligible resource owner
	jnz .r		;?; other owner
	mov eax,[esp]	; rst @tos
	mov [esp],ecx	; save reg
	mov ecx,TOS(1)
	mov ecx,[ecx+hphead.drivr]
	test ecx,ecx
	jz .q		;?; ok if no handler address
	mov ecx,[ecx+handler.aio]
	test ecx,ecx
	jz .q		;?; ok if no 'io' entry in handler defn
	cmp ecx,YE-Y
	ja .c
	add ecx,KR
.c:
	mov [esp],ecx	; sto hnd entry
	mov ecx,TOS(1)	; pass a1
	call [esp]
	cmp eax,-4096
	mov ecx,eax
	ja .r		;?; error
	xor ecx,ecx	; ok & eflags
.q:
	mov eax,ecx
.r:
	lea esp,[esp+4]	; drop spare data
	ret
;-
;-----------------------------------------------------------------
;- -- resource specifc routines - access by reg's, only, no data-stack opr
;-	
;-	common to all routines:
;-	in:
;-		eax	data input
;-		ecx,edx	2ndary data input
;-	out:
;-		eflags	error condition: NC & NZ, eax "above" -4096 (i.e. unsigned greater)
;-		eax	multiple args: error code, single arg: errorc code or return value
;-		ecx,edx	optional, other return values
;-	

;-
;- system memory
;-	

;-
;- mmrechp	<eax:a -- eax:er>
;-	pass memory back to system heap, store new memtop to (top)
;- i:	eax	heaphead
;- o:	eax	er
mmrechp:A;:ok:
; mark and link to free user heap if not top block, else set memtop and pass memory back to system heap
	push ebx
	sys_brk eax
	pop ebx
	cmp eax,-4096
	ja .r		;?; ernum
	mov [byte uref uv.mtop],eax
	xor eax,eax
.r:
	ret
;-
;- mmalloc	<eax:n -- eax:er ecx:hphead>	(k)
;-	allocate +n bytes of r/w/x enabled memory in system heap (.bss),
;-	ret address of bottom of allocated area, error if 0 > er > -4096
;-	new memtop stored to (top).
;- i:	eax	n
;- o:	eax	heaphead | ernum
mmalloc:A;:ok:
	mov ecx,[byte uref uv.mtop]
	call allocbss; eax=heaphead=old top, ecx=page-aligned size
	cmp eax,-4096
	ja .r		;?; error
	add ecx,eax	;c:=head+size
	mov [uref uv.mtop],ecx
.r:
	ret
;-
;- mmresiz	<ecx:a1 eax:u -- ecx:a2 eax:er>	(k)
;- i:	eax	new size
;-	ecx	address of header of block to resize
;- o:	eax	new top | er
;-	ecx	(new) hphead
; --
; check:
;	1) heap header
;	2) heap-mem handler
;	3) location
;		3.1)intermediate block
;		3.2)top block
;	4)
;		4.1) size smaller
;		4.2)	equal
;		4.3)	greater
; action:
;	3.x) &  4.2)	ready
;	3.1) &  4.1)	reduce size entry,
;			link remainder into 'free' list
;		4.3)	allocate full size on top, copy from old,
;			link old into 'free' list
;	3.2) &  4.x)	sys-brk to reduced/extended size, ready
; --
; simplified:
;	smaller:	resize if top block else alloc new, mark free, change pointers
;	larger:	allocate new on top, link old into free list
mmresiz:A;:ok:
	call palignp	; page align requested size
	pushad
	mov dword[esp s_eax],EFAULT	; prepare error return
	mov edx,[ecx+hphead.len]	; current block-size
	add eax,ecx	; top requested addr
	test edx,edx
	jle .q		;?; false data, not a heap header
	mov ebx,eax		; requested top address
	cmp eax,[byte uref uv.mtop]
	mov eax,0
	jz .x		;?; ready if equal size
	sys_brk
	cmp eax,-4096
	ja .x		;?; error
	mov [byte uref uv.mtop],eax; new
	sub eax,ecx
	mov [ecx+hphead.len],eax
	add eax,ecx
.x:
	writead eax,eax
.q:
	popad
	ret
;-
;---------------------------------------------------------------------------------
;- -- 4th host system memory words
;-

;-	

; -----------------------------------------------------------------------------------
;	u < blocklen		u > blocklen
;if p is top block:
;	free remining len	alloc additional len
;if p is intermediate block:
;	do nothing		alloc new on top and move
;				mark and link old block to free memory chain
;ANS:
; Change the allocation of the contiguous data space starting at the address a-addr1,
; previously allocated by ALLOCATE or RESIZE, to u address units. u may be either larger
; or smaller than the current size of the region. The data-space pointer is unaffected
; by this operation.
; If the operation succeeds, a-addr2 is the aligned starting address of u address units
; of allocated memory and ior is zero. a-addr2 may be, but need not be, the same as a-addr1.
; If they are not the same, the values contained in the region at a-addr1 are copied to
; a-addr2, up to the minimum size of either of the two regions. If they are the same, the
; values contained in the region are preserved to the minimum of u or the original size.
; If a-addr2 is not the same as a-addr1, the region of memory at a-addr1 is returned to
; the system according to the operation of FREE.
; If the operation fails, a-addr2 equals a-addr1, the region of memory at a-addr1 is
; unaffected, and ior is the implementation-defined I/O result code.

;-
;- ?resize	( n -- )					'?resizem'	forth
;-	-re- {resize} if {allot}ting n bytes would leave less than PAGE_SIZE 
;-	of free data-space; - does not {allot} but, is called by {allot}.
qresizem:GA;:ok:
	mov edx,dataptr
	mov ecx,datatop
	lea edx,[edx+PAGE_SIZE*2]
	sub ecx,eax
	sub ecx,edx
	jns .r
	spush 2
	mov ecx,[byte uref uv.mbot]
	add eax,[ecx]
	add ecx,byte hphead_size
	sub ecx,KR
	mov TOS(1),ecx
	add eax,4*PAGE_SIZE
	call resizem
	test eax,eax
	jnz near errorf
	mov eax,[byte uref uv.mbot]
;;sub eax,byte hphead_size
	add eax,[eax]
	sub eax,KR
	mov [byte uref uv.dtop],eax
.x:
	spop 2
.r:
	ret
;-
;- resize  	( p1 u -- p2 er )		(a:m)		'resizem'	forth
;-	provided for top block, only:
;-	new memtop stored to (top), error if 0 > er > -4096
;-	p1 ptr to after heap-header of top {allocate}d block.
;-	example: { (bot) @ abs>4th dup headhead-size + swap @ 4000 + resize 0if }.
;-	No garbage collection - which lib4th, if efficiently used, does not require, 
;-	e.g, due to it's extended LOCALS and BLOCK words -re- {local:}, {ibbuf} &c.
resizem:GA;:ok:
	call rel4hnd
	sub dword TOS(1),byte hphead_size
;edx: no op-code required, single one heap-hnd io-opr is 'resize'
	call hndio
	add dword TOS(1),byte hphead_size
	ret
;-
;- allocate	( u -- p 0 | p er )		(a:m)		'allocatem'	forth
;-	allocate u bytes of memory in system heap (.bss section), aligned to multiples 
;-	of Linux PAGE_SIZE, minimum is PAGE_SIZE = 4Kb (for Linux-2.2.+ and 2.4.+).
;-	ret ptr p to beginning of free space, after heap header.
;-	new memtop stored to (top), error if 0 > er > -4096
;-	{d-top} remains unchanged! adjust if extension of program dataspace intended,
;-	new top of data-memory is { p heaphead-size - dup @ + }, which then can be
;-	stored to {d-top} to make it available to the rsp. words.
;-	-re- LOCALS wordsets, {locals|} &c, for temporary buffers allocation, 
;-	-re- {ibbuf} and BLOCK words, for an alternate/supplemental allocation scheme.
allocatem:GA;:ok:
;o: eax 0 | ernum
;   ecx size
;   edx valid @(bot)
	add eax,byte hphead_size
	spush
	lea eax,[kref hphnd-Y]
	call hndopen
	jnz .r
	pushad
	mov ebx,TOS(1)	; address
	sys_mprotect EMPTY,dword[ebx],PROT_WRITE|PROT_READ
	popad
	sub TOS(1),KR
	add dword TOS(1),byte hphead_size
	xor eax,eax
.r:
	ret
;-
;- free		( p -- er )			(a:m)(e)	'freem'		forth
;-	assign {allocate}d top memory block p back to system heap (.bss section)
;-	ok if p was top block new memtop stored to (top), error if 0 > er > -4096
;-	{d-top} adjusted to { p heaphead-size - } if new top below old.
;-	p is ptr to after heap structure data, as returned by {allocate}.
;-	(e: Z after success, S on error, ENOBUFS if ptrs not suitable
freem:GA;:ok:
; o: eax ernum
;    ecx p-hphead
;    edx kref(p)-phead - no-error, only
	sub eax,byte hphead_size
	push eax
	lea eax,[kref eax]
	cmp eax,[byte uref uv.mtop]
	jnc .b		;?;
	mov ecx,[eax]
	lea ecx,[eax+ecx]
	cmp ecx,[byte uref uv.mtop]
	jnz .b		;?; not the ptr to top block
	call hndclose
	pop ecx
	ja .r		;?; error
	lea edx,[kref ecx]
	mov eax,[byte uref uv.mtop]
	cmp edx,eax
	jc .x		;?; d-top accessible
	sub eax,KR
	mov datatop,eax
.x:
	xor eax,eax
.r:
	ret
.b:A
	pop ecx
	xor eax,eax
	or eax,byte ENOBUFS; set S
	ret
;-
;- heaphead-size ( -- n ) 			(c)		'hpheads'	linux
;-	size of heap header (16/8 bytes = 4/2 cells)
;-	allocated: [ len ][ handler ][ pid ][ flagptr ][ ..memory.. ]
;-	free:	   [ len ][ next block ][ ..memory.. ]
hpheads:GA;:ok:
	spush
	mov eax,hphead_size
	ret
;-
;- top-mblock	( -- p )					'topmblock'	forth
;-	find ptr to beginning of top {allocate}d memory block,
;-	p := "free space" posn, after heap-header, initially = @c0.
topmblock:GA;:ok:
	call mtop
	mov eax,[kref eax]
	call mbot
	mov ecx,[kref eax]
.l:
	mov eax,ecx
	add ecx,[ecx]
	cmp ecx,TOS(1)
	jb .l
	sub eax,KR
	add eax,byte hphead_size
	TOSPULL(1)
	ret
;-
;- bufhead-size	( -- )				(kv)		'bufheadsz'	blkfile
;-	default size of a block buffer header (by kernel).
bufheadsz:KV
;-
;- blkhead-size	( -- )				(kv)		'blkheadsz'	blkfile
;-	default size of a per screenfile-block header in block of screenfile buffers.
blkheadsz:KV
;-
;- fgb		( ix -- ix )			(k)				novoc
;-	buffer blocks forget handler
fgb:GA;:ok:
	spush
	xor eax,eax
	mov ecx,[kref blocklkv-Y]
	add ecx,databot		; link base in RAM
	mov edx,[kref ecx]	; disp lk to 1st block
	cmp eax,edx
	lea edx,[edx+ecx]	;d: 1st blockhead
	jz .r			;?; initial link lost
; 1st in chain is 1st, 2nd is latest buffers' block
	mov ecx,[kref edx]	; disp 1st to latest
	cmp eax,ecx
	lea ecx,[ecx+edx]	;c: ptr of block to remove
	jz .z			;?; clear primary link
	add ecx,[kref ecx]	; next header
	sub ecx,edx
	mov [kref edx],ecx	; link 1st to next, now latest block
.r:
	ret
.z:
	mov ecx,[kref blocklkv-Y]
	add ecx,databot		; link base in RAM
	mov [kref ecx],eax
	ret
;-
;- ibbuf  	( -- )								blkfile
;-	set-up a block of screenfile block buffers.
;-	a buffers' block is of {page-size} bytes, beginning at page-aligned true 
;-	address which thus may be used to store a minimal, {mmap}ed item.
;-	constitutes a quasi-nameless word (name is a <nul> byte) w/ forget-handler 
;-	such, that buffer blocks removeable by {iforget} per uot-index of the rsp 
;-	buffer or with {forget} of any older regular word.
;-	
;-	block of block-/screenfile buffers, header at 8-aligned address:
;-		struc blkbuf  [ link ][ reserve ][ channel ][ #blocks ]
;-	00: .l	circular link, pc-rel disp to next blockptr, zero if 
;-		no other block exists; initial ptr in -re- {block-lk}, 
;-		linkage such that {block-lk} disp pointing to 1st, further 
;-		linkfields to the rsp latest and, last to 1st buffer block:
;-		block-lk->[ oldest->most recent->previous->..->next to oldest ]
;-	04: .u	- till unused -
;-	08: .c	file channel the rsp buffer is assigned to
;-	0c: .b	no. of blocks in this buffer
;-	10: .f	#blocks(4) buffer-headers follow:
;-		struc blkhead [ updt|blk# ][ disp ]
;-		00: .u: update-flag(bit#31) ORed w/ block-number
;-		04: .d: pc-rel disp-to-blk
;-	next buffer in chain ( -- p f ), f=/=0 if next buffer found:
;-		{ block-lk @ begin dup @ over swap + dup rot - ... }
;-	-re- {buffer}, {block}, {blk>lnk}, {buf>lnk}, {blk>head}, {buf>head} &c.
;-	data/code structures re file <inc/l4struc.inc>.
ibbuf:GA;:ok:
	call qexec			; whether in executing state, else abort
	call rrbrac
	spush
	mov eax,.h-Y			; dummy name
	call build			; header, for forget-handler base (ecx:=lfa)
	call csemis			; cpl <ret>
	call hpts			; make code permanent
	call llbrac			; leave the word un-findable
	spush
	mov eax,[byte uref uv.lkindex]
	dec eax				; this word's ix
	TICK fgb			; forget-handler
	call fgcomma
	spush
	mov edx,databot
	mov eax,dataptr			; {here}
	lea eax,[kref eax+7]
	and eax,byte -8			; +8-aligned
	sub eax,KR			; -> blkbuf.l, link-ptr
	mov dataptr,eax
	mov ecx,edx			; c:=d0
	add edx,[kref blocklkv-Y]	; {block-lk}
	add ecx,[kref limitv-Y]		;c: {[limit]}
	mov dword[kref ecx],0		; force re-caculation of [limit]
	mov ecx,[kref edx]		;b: ptr to 1st link
	test ecx,ecx
	lea ecx,[ecx+edx]		; ptr<-disp
	jnz .b				;?; already initialized
	mov [kref edx],eax		; here to {block-lk}
	sub [kref edx],edx		; link relative
	mov ecx,eax			; chain begins
	mov [kref eax],eax
.b:
	mov edx,[kref ecx]		; prev. disp
	add edx,ecx			; -> ptr
	sub edx,eax			; pc-rel new buf
	mov [kref eax],edx		; back-ptr into new buf
	mov edx,eax			; this buf-blk, link ptr
	sub eax,ecx			; initially 0, ref to self
	mov [kref ecx],eax		; new disp into prev buf
	mov dword[kref edx+blkbuf.c],-1	; no owner channel
	mov ecx,(bbufmem/b_buf)		; count (4)
	mov dword[kref edx+blkbuf.b],ecx
	lea edx,[kref edx+blkbuf.f]	; true addr after last buf-head
	lea eax,[edx+PAGE_SIZE-1+(bbufmem/b_buf)*blkhead_size]
	and eax,PAGE_MASK		; +page-aligned
	sub eax,edx			; start buffers page aligned
	lea eax,[eax+ecx*blkhead_size-4-(bbufmem/b_buf)*blkhead_size]
.l:; [ l | u | c | f | p1 | p2 | p3 | p4 ]...[ b1 ][ b2 ][ b3 ][ b4 ]
	mov dword[edx+blkhead.u],-1
	mov [edx+blkhead.d],eax
	lea eax,[eax+b_buf+(bbufmem/b_buf)*blkhead_size-blkbuf_size+blkhead_size]
	dec ecx
	lea edx,[edx+blkhead_size]
	jnz .l
; ende auf <edx+blkhead.d+eax-KR>
	lea eax,[edx+eax+blkhead.d]
	sub eax,KR
	mov dataptr,eax			; allot buffer space
	spop
	ret
.h:
	call here
	mov dword[kref eax],0		; would break "words" if {find}able
	jmp zero
%endif; __SMEM_I
;-
;-=====================================================================;
;- l4smem.inc <eof>
