;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4sorder.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  latest	: 1.6, 0.16.1+8
;-  switches	: __SORDER_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __SORDER_I
%define __SORDER_I
%define sorderinc

;-
;-================================================================
;--- search order
;-	unless mentioned all vocabularies belong to {root},
;-	{root} vocabulary is sub-vocabulary of itself.
;-	lib4th supplied vocabularies are set "immediate".
;-
;-	

localwl:A; for {forget} words
	spush
	_mov eax,Ilocal
	jr idx2wid
;-
;- forth-wordlist ( -- wid )			(a:v)		'forthwl'	root
forthwl:GA;:ok:
	spush
	mov eax,Iforth
	jr idx2wid
;-
;- root-wordlist ( -- wid )			(a:-)		'rootwl'	hidden
rootwl:GA;:ok:
	spush
.i:
	mov eax,Iroot
; find xhead of voc in ram (glossary below), Z and ret zero if not found
;-
;- idx>wid	( ix -- wid | 0 )				'idx2wid'	hidden
;-	convert vocabulary name's index to wordlist identifier
idx2wid:GA;:ok:
	mov ecx,eax
	call rvocfind
	jnz .r
	xor ecx,ecx
.r:
	mov eax,ecx
	ret
; find wordlist linkage by voc-index
rvocfind:A
	push eax;
	push edx
	lea edx,[byte uref uv.voclink]
.l:
	mov eax,[edx]
	test eax,eax
	jz .r
	sub edx,eax		; voc-lfa = wid
	lea eax,[edx-vhead.l+vhead.m]
	sub eax,[eax]		; lfa
	cmp cx,[eax-whead.l+whead.i]; index
	jnz .l
	sub edx,KR
	mov ecx,edx
.r:
	pop edx
	pop eax
	ret
;-
;- dvo						(X)(k)		'dovoc'		novoc
;-	vocabulary runtime routine, executed by a vocabulary defn:
;-	save previous current in {ocurrent}, drop previous {ocurrent},
;-	store wid = voclink ptr of called voc to {current}.
dovoc:A;:ok:
; c: ecx, mm0
%ifdef __MMX
	pop ecx
	movq mm0,[byte uref uv.context-4]
	movq [byte uref uv.ocontext-4],mm0
%else
	mov ecx,[byte uref uv.context]
	mov edx,[byte uref uv.context-4]
	mov [byte uref uv.ocontext],ecx
	mov [byte uref uv.ocontext-4],edx
	pop ecx
%endif
	mov ecx,[ecx]	; uot index
	call rvocfind
	jz .r
	mov [byte uref uv.context],ecx
	mov dword[byte uref uv.context-4],0
.r:
;-
;- novoc  	( -- )				(i)				system
;-	inactive, dummy vocabulary for debugging purposes, sub-vocabulary of {system},
;-	list of words for kernel access, only, named to supporting {see} output and 
;-	w/ header data for internal, redirectably compiling (i.e. by ix) purposes.
;-	{novoc} is a noop. words linked to a pseudo-'wordlist' which should not be 
;-	accessed; the rsp words glossary entries will be displayed w/ { help name }.
;-	if activated - w/ some applied magic - would most certainly crash the system!
novoc:
	ret
;-
;- root		( -- )				(X)(i)				root
;-	the basic vocabulary which can't be excluded from search order, where it 
;-	will be searched last if not explicitely included. contains the words 
;-	essential to enabling orderly access to the forth system. 
;-	-re- {only} resets search order to no other vocabulary but {local} and {root}.
root:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Iroot
;-
;- linux  	( -- )				(X)(i)				root
;-	vocabulary providing words for host system access which are not defined 
;-	in a(ny) Forth standard. linux system calls and system specific constants.
linux:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ilinux
;-
;- forth  	( -- )				(X)(i)				root
;-	most commonly used and, all ANS-Forth words as described in "dpans94" doc.
forth:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Iforth
;-
;- bignum  	( -- )				(X)(i)				forth
;-	vocabulary for opr on counted signed integers.
;-	a 'counted integer' is a contiguous group of little endian ordered bits in 
;-	multiples of cell-size w/ sign bit at m.s. position (top address), in 2-s 
;-	complement notation, and a count cell, bearing the no. of the rsp. cells.
;-	located in data-stack w/ count cell @tos or, in memory w/ count cell at l.s. 
;-	address, for fast data access. -re- {n@} &c.
;-		Notation:	( .n.N )	count N of cells n(0..N)
;-	(genuine, new implementation, no "Art"work, nothing copied from nowhere...)
bignum:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ibignum
;-
;- hidden  	( -- )				(X)(i)				root
;-	vocabulary, words which support but aren't standard words.
hidden:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ihidden
;-
;- testvoc	( -- )				(X)(i)				hidden
;-	vocabulary for testing purposes
testvoc:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Itestvoc
;-
;- non-mmx  	( -- )				(X)(i)				forth
;-	non-mmx substitutes vocabulary.
nonmmx:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Inonmmx
;-
;- ans		( -- )				(X)(i)				forth
;-	{forth} sub-vocabulary,
;-	ans-4th specific supplementary words to {forth} wordlist.
ans:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ians
;-
;- fig		( -- )				(X)(i)				forth
;-	{forth} sub-vocabulary 
;-	f.i.g.-4th specific supplementary words to {forth} wordlist.
fig:GA;:ok:
	call dovoc
.d:
	dd Ifig
;-
;- editor  	( -- )				(X)(i)				fig
;-	the f.i.g-Forth screenfile editor, {fig} sub-vocabulary.
;-	some words homononymous to other, Forth words! e.g. {R}, {I}, {J}.
;-	      B         backup cursor by length of text in pad
;-	      C {text}  copy in the following text at cursor posn
;-	( n ) D         remove line no. ( n ), put it to pad
;-	( n ) E         blank line no. ( n )
;-	      F {text}  find following text
;-	( n ) H         hold line no. ( n ) in pad
;-	( n ) I         insert line no. ( n ) from pad
;-	      L         list current screen,
;-	      	L+,L-,L#    next, previous, by number.
;-	( n ) M         move cursor by ( n ) chars, show line & cursor
;-	      N         find next occurence of text ( after f )
;-	( n ) P {text}  put text until <enter> on line no. ( n )
;-	( n ) S         spread at line no. ( n ), insert an empty line
;-	( n ) R         replace line no. ( n ) with text from pad
;-	( n ) T         type line no. ( n ) and hold it in pad
;-	      X {text}  delete the following text
;-	( n ) CLEAR	clear scr no. n, append to file if appropriate
;-	( n ) DELETE	delete n chars backwards
;-	( n1 n2 ) COPY	copy screen n1 to n2
;-	      flush	save updated, modified buffers
;-	( n ) .LINE	display line no. n of current screen
;-	      TEXT {text} to pad
;-	      TILL {text} delete from cursor till {text} end
;-	      TOP	cursor to top left of current screen
;-	( p u ) using	open file to be used w/ screenfile words
;-	      where	(n.i.)
editor:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ieditor
;-
;- blkfile  	( -- )				(X)(i)				forth
;-	sub-vocabulary of {forth}
;-	block-files specific supplementary words to {forth} wordlist.
blkfile:GA;:ok:
	call dovoc
.d:
	dd Iblkfile
;-
;- vt	  	( -- )				(x)(i)				forth
;-	vocabulary, vt..-sequences for console appearance & control.
;-	ctrls can also be sent as strings { ." \33..." }, etc. -re- {[etype]}.
vt:;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ivt
;-
;- assembler	( -- )				(x)(i)				root
;-	vocabulary, native code compiling support
assembler:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Iassembler
;-
;- tools  	( -- )				(x)(i)				forth
;-	non-standard debugging (and other) utilities vocabulary
tools:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Itools
;-
;- local  	( -- )				(x)(i)				hidden
;-	local memory support vocabulary, temporary local names & memory ref-s
;-	{local} will always be searched 1st and cannot be excluded from search order.
local:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Ilocal
;-
;- compiler	( -- )				(x)(i)				root
;-	vocabulary, non-standard compiler control.
compiler:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Icompiler
;-
;- system  	( -- )				(x)(i)		'systm'		root
;-	vocabulary, highly host and system specific words, not for common access.
systm:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Isystm
;-
;- util   	( -- )				(X)(i)		'utility'	forth
;-	vocabulary, forth utility words (for app's).
utility:GA;:ok:
; c: ecx, mm0
	call dovoc
.d:
	dd Iutility
;-
;- [locdef]					(4)		'blocdefb'	hidden
;-	support for the {(locals)} definition; cpl { also local definitions }.
blocdefb:GA;:ok:
	call also
	call local	;:continues:
;-
;- definitions					(a:s)(X) 			root
;-	set current to context vocabulary, 
;-	store old in ocurrent for restauration by {prevdef}.
;-	does { current ocurrent @! context current @! }
definitions:GA;:ok:
%ifdef __MMX
; c: ecx, mm0
	movq mm0,[byte uref uv.current-4]
	movq [byte uref uv.ocurrent-4],mm0
	movq mm0,[byte uref uv.context-4]
	movq [byte uref uv.current-4],mm0
	ret
prevdef:GA;:ok:
; c: ecx, mm0
	movq mm0,[byte uref uv.ocurrent-4]
	movq [byte uref uv.current-4],mm0
	ret
rvoc:GA;:ok:
; c: ecx, mm0
	movq mm0,[byte uref uv.ocontext-4]
	movq [byte uref uv.context-4],mm0
	ret
only:GA;:ok:
; c: ecx, mm0
	push edx,eax
	mov ecx,[kref root.d-Y]
	call rvocfind
	jz .r
	movd mm0,ecx	; zero extends to 64bit
	mov ecx,[kref cnvocs-Y]
	lea edx,[byte uref uv.context]
	movq [edx],mm0	; store root to context
	lea edx,[edx-8]
	psrlq mm0,32	; clear stack
.l:
	movq [edx],mm0
	dec ecx
	lea edx,[edx-8]
	jns .l
.r:
	pop regl
	ret
also:GA;:ok:
; c: ecx,mm0
	push edx
	mov ecx,[kref cnvocs-Y]
	not ecx	; 1+ minus
	lea edx,[uref uv.context+ecx*8];top of vocstack
.l:
	movq mm0,[edx+8]
	movq [edx],mm0
	inc ecx
	lea edx,[edx+8]
	jnz .l
	pop edx
	ret
previous:GA;:ok:
; c: ecx, mm0
	push edx
	lea edx,[byte uref uv.context]
	mov ecx,[kref cnvocs-Y]
	inc ecx
.l:
	lea edx,[edx-8]
	dec ecx
	movq mm0,[edx]
	movq [edx+8],mm0
	jns .l
	pop edx
	ret
%else; __MMX undef'd 
; ->definitions:GA;:ok:
	mov ecx,[byte uref uv.current]
	mov edx,[byte uref uv.current-4]
	mov [byte uref uv.ocurrent],ecx
	mov [byte uref uv.ocurrent-4],edx
	mov ecx,[byte uref uv.context]
	mov edx,[byte uref uv.context-4]
	mov [byte uref uv.current],ecx
	mov [byte uref uv.current-4],edx
	ret
;-
;- prevdef	( -- )				(f6)(X) 			root
;-	restore previous current from {ocurrent}.
prevdef:GA;:ok:
	mov ecx,[byte uref uv.ocurrent]
	mov edx,[byte uref uv.ocurrent-4]
	mov [byte uref uv.current],ecx
	mov [byte uref uv.current-4],edx
	ret
;-
;- rvoc		( -- )				(i)(f6)(X) 			root
;-	restore previous context from {ocontext}.
rvoc:GA;:ok:
	mov ecx,[byte uref uv.ocontext]
	mov edx,[byte uref uv.ocontext-4]
	mov [byte uref uv.context],ecx
	mov [byte uref uv.context-4],edx
	ret
;-
;- only		( -- )				(i)(a:sx)(X)			root
;-	reset search order to only {root} in {context}.
only:GA;:ok:
	push edx,eax
	mov ecx,[kref root.d-Y]
	call rvocfind
	jz .r
	mov eax,ecx
	mov ecx,[kref cnvocs-Y]
	lea edx,[byte uref uv.context]
	mov [edx],eax
	xor eax,eax		; clear stack
	mov dword[edx-4],eax; store root to context
	lea edx,[edx-8]
.l:
	mov [edx],eax
	mov [edx-4],eax
	lea edx,[edx-8]
	dec ecx
	jns .l
.r:
	pop regl
	ret
;-
;- also		( -- )				(i)(a:sx)(x)			root
;-	push ocontext, duplicate context into ocontext
;-	push is move all one item down and store new to hi address.
;-	stack size is 2+{#vocs} items (dft:2+16).
also:GA;:ok:
	push edx,ebx,eax
	mov ecx,[kref cnvocs-Y]
	not ecx	; 1+ minus
	lea edx,[uref uv.context+ecx*8];top of vocstack
.l:
	mov eax,[edx+8]
	mov ebx,[edx+4]
	mov [edx],eax
	mov [edx-4],ebx
	lea edx,[edx+8]
	inc ecx
	jnz .l
	emms
	pop regl
	ret
;-
;- previous	( -- )				(i)(a:sx)(x)			root
;-	copy @{ocontext} to {context}, pop top item from vocstack to {ocontext}.
previous:GA;:ok:
	push edx,ebx,eax
	lea edx,[byte uref uv.context]
	mov ecx,[kref cnvocs-Y]
	inc ecx
.l:
	lea edx,[edx-8]
	mov eax,[edx]
	mov ebx,[edx-4]
	mov [edx+8],eax
	mov [edx+4],ebx
	dec ecx
	jns .l
	pop regl
	ret
%endif; __MMX undef'd
;-
;- get-current	( -- wid )			(a:v)		'gcurrent'	root
;-	fetch {current} wordlist identifier.
gcurrent:GA;:ok:
	spush
	mov eax,[byte uref uv.current]
	ret
;-
;- set-current	( wid -- )			(a:v)		'scurrent'	root
;-	tore {current} wordlist identifier.
scurrent:GA;:ok:
	mov [byte uref uv.current],eax
	spop
	ret
;-
;- get-vocflags	( -- wfl..wfl n )		(l4)		'gvocflags'	hidden
;-	complements {get-order} by fetching vocabulary flags
gvocflags:GA;:ok:
	lea edx,[uref uv.context-4]; vocflags
	jr gorder.n
;-
;- get-vocs	( -- wfl n wid.. n )		(l4)		'getvocs'	root
;-	fetch all wid-s and voc.-flags currently in stacked search order.
getvocs:GA;:ok:
	call gvocflags
;-
;- GET-ORDER	( -- wid .. wid n )		(a:v)		'gorder'	ans
;-	get stacked vocabularies identifiers and count. this list 
;-	includes {context} and {ocontext}, @context top wid on stack
;-	the list can be preserved w/ a bignum {ninteger}
;-		{ GET-ORDER also bignum ninteger -name- previous }
;-	for restoration w/ { -name- SET-ORDER }.
;-	ANS-4th; -re- {get-vocs}, {set-vocs} for a more complete version.
gorder:GA;:ok:
; c: ecx,edx
	lea edx,[uref uv.context]; wid-s
.n:;	;<gvocflags)
	call nvocs	; #vocs
	mov ecx,eax	;
	neg eax
	push_ibyte(0)
	lea edx,[edx+eax*8]; top of vocstack
.l:
	cmp dword[edx],byte 0
	jnz .s
	test ch,ch
	jz .z		; suppress top zeroes
.s:
	mov ch,cl
	inc dword[esp]
	mov eax,[edx]
	spush
.z:
	dec cl
	lea edx,[edx+8]
	jns .l
	pop eax
	ret
;-
;- set-order	( wid ... n -- )		(a:v)		'sorder'	root
;-	store list of n vocabulary identifiers to stacked search order.
;-	includes {context} and {ocontext}, vocflags remain unmodified.
;-	abort if false count/unsufficient data on stack
sorder:GA;:ok:
	lea edx,[uref uv.context+8]; wid-s
	jr svocflags.n
;-
;- set-vocs	( wfl.. m wid.. n -- )				'setvocs'	root
;-	restore search-order(wid,n) and voc-flags(wfl,m)
;-	abort if false count/unsufficient data on stack
setvocs:GA;:ok:
	call sorder
;-
;- set-vocflags	( -- wfl .. wfl n )				'svocflags'	hidden
;-	complements {set-order} by storing vocabulary flags
;-	abort if false count/unsufficient data on stack
svocflags:GA;:ok:
	lea edx,[uref uv.context+4]; vocflags
.n:
	test eax,eax	; dup
	jng .r		; 0> if
	lea ecx,[eax*4+4]
	call qdepthb	; check stack depth
	mov ecx,eax
.l:			;  begin
	lea PS,[PS-4]
	lea edx,[edx-8]
	mov eax,[PS]
	dec ecx		;   1- dup 0=
	mov [edx],eax
	jg .l		;  until
.r:			; endif
	spop		; drop
	ret
;-
;- order  	( -- )				(4)(a:v)			root
;-	display current, context, previous context 
;-	and stacked search order vocabularies.
order:GA;:ok:
	call pdotqtep
	db 0,'current:',0
	call current	;	current
	call .v		;	.voc
	spop		;	drop
	call pdotqtep
	db 0,'context:',0
	call context	;	context
	call .v		;	.voc
	jz .x		;	dup if(1)
	call .v		;	 .voc
	jz .x		;	 dup if(2)
	call cr		;	  cr
	call nvocs	;	  #vocs
	push eax	;	  >r
	spop
.l:			;	  begin
	dec dword[esp]	;	   r> 1- dup >r
	js .e		;	   0< 0=  while
	call .v		;	   .voc
	jnz .l		;	  dup 0= until
.e:
	lea esp,[esp+4]	;	  rdrop
.x:			;	endif(1)(2)
	spop		;	drop ;s
	ret
.v:A;			; display voc-name
	spush		; 	8- dup 8+
	sub dword TOS(1),byte 8
	mov eax,[kref eax];	@
	xor ecx,ecx	;	0 4th>abs
	cmp eax,KR	;	over-
	jz .q		;?;	if(1)
	test eax,eax	;	 -dup
	jz .q		;?;	 if(2)
	push eax	;	  >rr
	call wid2lfa	;	  voc>lfa
	spush		;	  dup
	call lfa2wid	;	  lfa>voc
	pop ecx		;	  r>
	cmp eax,ecx	;	  =/=
	spop		;	  if(3)
	mov ecx,0	;	   drop 0
	jnz .q		;?;	  else
.r:
	call iddot	;	   id.
	call space	;	   space
	mov ecx,-1
	spush		;	   dup
.q:			;	endif(1)(2)(3)
	spop		;	sdrop
	and eax,ecx
	ret
;-
;- vocs		( -- )				(4)(f6)				root
;-	display names of all vocabularies
vocs:GA;:ok:
;voc-link begin dup @ -dup while - dup wid>lfa id. repeat drop
	call voclink
.begin:
	mov ecx,[kref eax]
	test ecx,ecx
	jz .repeat
	sub eax,ecx
	jz .repeat
	spush
	call wid2lfa
	call iddot
	call space
	jr .begin
.repeat:
	spop
	ret
;-
;- order? 	( ix | lfa -- n | -1 )				'orderq'	forth
;-	ret item number wrt {current} of voc by ix in voc-stack, -1 if not found.
;-	n points to the rsp. ptr into vocstack by { ' name order? -8 * current + }.
;-	ix = -1 to finding the actual bottom of vocstack, ret -1 if no space left.
;-	-re- {WORDLIST?} for ref by wid.
orderq:GA;:ok:
	inc eax
	jz wordlq
	dec eax
	call qlfa
	mov ecx,-1
	jnz wordlq.x
	call lfa2idx
	call idx2wid
	mov ecx,-1
	test eax,eax
	jz wordlq.x	;?; not a vocabulary
;			;:continues: in <wordlq>
;-
;- WORDLIST?	( wid -- n | -1 )				'wordlq'	ans
;-	item number wrt {current} of wordlist by wid in search order, -1 not found.
;-	n is index into vocstack, by { ' name idx>wid WORDLIST? -8 * current + }.
;-	wid = 0 to finding the actual bottom of vocstack, ret -1 if no space left.
;-	-re- {order?} for ref by ix or lfa.
wordlq:GA;:ok:
	push edi
	mov ecx,[kref cnvocs-Y]; max items on vocstack
	add ecx,byte 2	; include context, ocontext
	lea edi,[byte uref uv.context]
	mov edx,ecx
.l:
	test eax,eax
	jz .d		;?; find actual top of vocstack
	cmp [edi],byte 0
	jnz .d
	xor ecx,ecx	; search order ends
.d:
	dec ecx
	js .r
	cmp eax,[edi]
	lea edi,[edi-8]
	jnz .l
	sub ecx,edx
	not ecx
.r:
	pop edi
.x:
	mov eax,ecx
	ret
%endif;__SORDER_I
;-
;-=====================================================================;
;- l4sorder.inc <eof>
