;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4sranum.inc
;-  created	: 18-jan-02
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __SRANUM_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th", sing rational numbers wordsets
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: rational numbers by ordered cells' pairs
;-  source	: Bronstein Semendjajev, Taschenbuch der Mathematik
;-		: Forth Scientific Library Algorithm #46
;-		: Htte Bd.1, Theoretische Grundlagen
;-		: and others (re asm source).
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __SRANUM_I
%define __SRANUM_I
%define sranuminc

;-
;- 0v		( -- v )					'v0v'		rational cons
;-	cons sing ranum 0
v0v:GA;:ok:
	spush
	mov eax,0
	spush
	inc eax
	ret
;-
;- 1v		( -- v )					'v1v'		rational cons
;-	cons ranum 1
v1v:GA;:ok:
	spush
	mov eax,1
	spush
	ret
;-
;- -1v		( -- v )					'vm1v'		rational cons
;-	cons ranum -1
vm1v:GA;:ok:
	spush
	mov eax,-1
	spush
	neg eax
	ret
;- e^6   	( -- v )					'cae6'		rational cons
;-	e to the power of ten, ret v := 403,42879349
cae6:GA;:ok:
	spush
	mov eax,1608856681
	spush
	mov eax,3987957
	ret
;- e		( -- num den )					'cae'		rational cons
;-	2,7182818284(5)... +/- 10^-10, e = lim (1+1/n)^n
;-					   n=1->oo
cae:GA;:ok:
	spush
	mov eax,1350212337	;260412269	;28667
	spush
	mov eax,496715360	;95800320	;10546
	ret
;- e^10		( -- v )					'cae10'		rational cons
;-	e to the power of ten, ret v := 22026,4657948
cae10:GA;:ok:
	spush
	mov eax,641697028
	spush
	mov eax,29133
	ret
;-
;-	
;--- Literals
;-

    global vliteral:function,vconstant:function,vinteger:function,vvariable:function
;-
;- vliteral	( v -- )			(i)				rational
;-	compile/interpret rational number v as a double literal.
vliteral:equ dliteral
;-
;- vconstant	( -- v )							rational
;-	-re- {2constant}
vconstant:equ twocons
;-
;- vinteger	( -- v )							rational
;-	-re- {2integer}
vinteger:equ integer2
;-
;- vvariable	( -- p )							rational
;-	-re- {2variable}
vvariable:equ twovari
;-
;- vreduce 	( v1 -- v2 )							rational
;-	reduces a ranum by greatest common divisor of numerator and denominator.
vreduce:GA;:ok:
	cmp eax,1
	mov ecx,TOS(1)
	jna .r		;?; trivial, 1(integral) or 0(ovf)
	mov edx,eax
	test eax,eax
	spush
	call gcd.n
	mov ecx,eax	; gcd
	jz .x		;?; trivial cd
	mov eax,TOS(1)
	xor edx,edx	; unsigned
	div ecx		; den/gcd
	mov TOS(1),eax
	lea edx,[kref .h-Y]
	push edx	; ret from xep-hnd
	push ebx	; save KR
	mov eax,TOS(2)	; nom
	mov edx,eax
	sar edx,31	; s->d
	lea ebx,[kref by.h-Y];(xep hnd)
	idiv ecx	; nom/gcd
	pop ebx
	add esp,byte 4	; drop xep-hnd
.h:
	mov TOS(2),eax
.x:
	spop
.r:
	ret
;-
;- ud>n?  	( ud -- +n f)					'ud2nq'		rational hidden
;-	converts unsigned double ud to non-negative single +n. 
;-	tf if conversion failed.
ud2nq:A;:ok:
; OVER highbit AND OR 0<> ;
	mov ecx,TOS(1)
	sar ecx,31
	or eax,ecx
	jz .r
	mov eax,-1
.r:
	ret
;-
;- vnormal 	( dv -- v )			(x)				rational hidden
;-	return the normalised ranum which most closely approximates the signed 
;-	double rational number passed to it. Returns 1 0 if the double was too 
;-	large and positive to be represented or -1 0 if too large and negative. 
;-	Numbers too small to be represented are rounded to zero (0 1).
;-	
;-	Works by generating approximations from successive numbers in the continued
;-	fraction expansion of dv, until no more numbers are available, or overflow
;-	occurs.
;-	The 1st number in the continued fraction expansion is the integral component 
;-	and the next the integral component of the reciprocal of the fractional 
;-	component and so on. The continued fraction of any rational number terminates 
;-	when the fractional component becomes zero.
vnormal:GA;:ok:
;  dv>udv? >R
	call dv2udvq; sg.dv, abs
	spush
	mov eax,1
	spush
;  >rr 1 1 r>
;  { sg.dv prev.numerator denominator | numerator prev.denominator } [ = setvars ]
	kloci 5
.pnu:	klocv	; prev.numerator
.dn:	klocv	; denominator
.sg:	klocv	; flag from dv>udv?
.nu:	klocz	; numerator
.pdn:	klocz	; prev.denominator
.enter:;    enter
	jr .entry
.begin:A;    begin	; ( dl dh ul uh 
;	2OVER 2DUP d0<> IF d/mod THEN			(1)
	spush 2
	mov ecx,TOS(5)	; ( dl dh ul uh dl dh
	mov eax,TOS(4)
	mov TOS(1),ecx	; 2over
	or ecx,eax	; 2dup
	jz .endif1	; d0<> if
	call dbym
.endif1:
;        2DUP num>?
	call dup2
	call .num2q
;	IF next! drop done
	jnz .if2
;	    -rot den>? 
	call mrot
	call .den2q
;	    IF next!  done
	jnz .if3
;		next! 2SWAP
	call .nexts
	call swap2
.entry:;    entry
;	2dup d0=
	mov ecx,eax
	or ecx,TOS(1)
.until:;    until
	jnz .begin
.endif23:;    endif endif
;    4drop
	spop 2
;    prev.numerator R> ?negate prev.denominator ;
	mov ecx,[lref .pnu]
	cmp dword[lref .sg],byte 0
	jns .p
	neg ecx
.p:
	mov eax,[lref .pdn]
	mov TOS(1),ecx
	ret
.if2:A
	call .nexts
	spop
	jr .endif23
.if3:A
	call .nexts
	jr .endif23
; num>? ( ud -- u f)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  numerator mu* prev.numerator 0 D+ ud>+n? ;
.num2q:A
;the next approximation to the numerator u is the current approximation
;multiplied by ud, which is the next number in the continued fraction of
;the ranum being approximated, plus the previous approximation.
;flag is true if the next approximation cannot be represented as a non-
;negative single integer (ie the approximation is as good as it can get).
; e: result
	spush
	mov eax,[lref .nu]	; numerator
	call udmmul		; udm* => treble prod, drop => mu*
	mov ecx,[lref .pnu]	; prev.numerator
.num2q.n:
	spop
	add ecx,TOS(1)
	adc eax,byte 0		; 0 d+
	mov TOS(1),ecx
	sar ecx,31
	or ecx,eax		; ud>+n? ( ud--+n f)  OVER highbit AND OR 0<> ;
	spop
	ret
;: den>? ( ud -- u f)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  denominator mu* prev.denominator 0 D+ ud>+n? ;
.den2q:A
;denominator approx. as above, numerator-
	spush
	mov eax,[lref .dn]	; denominator
	call udmmul		; udm* drop = mu*
	mov ecx,[lref .pdn]	; prev.denominator 
	jr .num2q.n
;: next! ( +v)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  denominator to prev.denominator to denominator
;  numerator   to prev.numerator   to numerator ;
.nexts:A
;make the new approximation the current approximation, and the current
;approximation the previous approxinmation.
%ifdef __MMX
	movd mm0,[lref .dn]	;c: denominator..
	movd mm1,[lref .nu]	;d: numerator..
	movd mm2,TOS(1)
	movd [lref .pdn],mm0	;c:..to prev.denominator
	mov [lref .dn],eax	; to denominator
	movd [lref .pnu],mm1	;d:..to prev.numerator
	spop 2
	movd [lref .nu],mm2	; to numerator 
%else
	mov ecx,[lref .dn]	;c: denominator..
	mov edx,[lref .nu]	;d: numerator..
	mov [lref .dn],eax	; to denominator
	mov [lref .pdn],ecx	;c:..to prev.denominator
	mov eax,TOS(1)
	mov [lref .pnu],edx	;d:..to prev.numerator
	mov [lref .nu],eax	; to numerator 
%endif
	ret
;-
;- small? 	( dv -- f )							rational hidden
;-	returns TRUE if double ranum dv can be normalised by "normal", rather than
;-	the slower "vnormal".
smallq:A;:ok:
; highbit 0 DU< >R DABS highbit 0 DU< R> AND ;
; => [ -0. swap 2dup ] dliteral du< -rot dabs dliteral du< and ;
	call smallfq
	TOSPULL(3)
	mov eax,ecx
	ret
smallfq:A; 	( dv -- dv )
; o: ecx: flag value
	test eax,eax
	jnz .f
	cmp eax,TOS(1)
	jg .f
	mov ecx,TOS(2)
	mov edx,TOS(3)
	cmp ecx,eax
        jns .p
	neg edx
	adc ecx,byte 0
	neg ecx
.p:
	test ecx,ecx
	jnz .f
	mov ecx,-1
	test ecx,edx
	jns .r
.f:
	mov ecx,0
.r:
	test ecx,ecx
	ret
;-
;-	
;--- Arithmetic
;-

vplus.zc:A
	xor eax,eax
vplus.z:A
	xor ecx,TOS(3)
	and ecx,0x80000000
	TOSPULL(2)
	mov TOS(1),ecx
	ret
;-
;- v-		( v1 v2 -- v3 )					'vsub'		rational
;-	subtract v2 from v1.
vsub:GA;:ok:
;  vnegate v+ ;
	neg dword TOS(1);:continued:
;-
;- v+		( v1 v2 -- v3 )					'vplus'		rational
;-	add two ranums using naive algorithm, 
;-	generating double length result, then round to single.
vplus:GA;:ok:
; ROT 2DUP UM* 2>R ROT M* 2SWAP M* D+ 2R> dv>v ; \ v3 := (v1n*v2d+v2n*v1d)/(v1d*v2d)
	test eax,eax	; ( n1 d1 n2 d2 --
	mov edx,TOS(2)
	mov ecx,TOS(1)
	jz vplus.z	;?; ovf
	test edx,edx
	jz vplus.zc	;?; ovf
	spush 2		; ( n1 d1 n2 d2 d2 d2 --
	imul dword TOS(5); n1*d2
	mov TOS(5),eax	; en1d2l
	mov eax,TOS(4)	;    d1
	mov TOS(4),edx	; en1d2h
	mov TOS(1),eax	; save d1
	imul dword TOS(3); n2*d1
	add TOS(5),eax	; 
	adc TOS(4),edx	; zhler1*nenner2 + zhler2*nenner1 =: num.
	mov ecx,TOS(1)	; d2 ( snl snh d2 d1 xx --
	spop 2
	mul ecx		; d1*d2	=: gem. nenner
	mov TOS(1),eax
	mov eax,edx	;	;:contiued:
;-
;- dv>v		( dv -- v )					'dv2v'		rational
;-	returns the ranum that most closely approximates the unnormalised  
;-	ranum dv. If overflow is detected (denominator is zero) action is 
;-	taken to ensure the overflow indicator 0 0 is returned.
dv2v:GA;:ok:;<toquad>
; 4dup small? IF DROP NIP normal ELSE vnormal THEN DUP 0= IF NIP 0 THEN ;
	call smallfq
	jz .e
	mov eax,TOS(1)	; drop swap drop
	TOSPULL(2)
	call vreduce
	test eax,eax
	jnz .r
.i:
	mov TOS(1),eax	; drop swap 0
	ret
.e:A
	call vnormal
	test eax,eax
	jz .i
.r:
	ret
;-
;- v/		( v1 v2 -- v3 )					'vby'		rational
;-	divide v1 by v2 giving quotient v3.
vby:GA;:ok:
; vreciprocal v* ;
	call vrecip	;:continued:
;-
;- v*		( v v -- v)					'vmul'		rational
;-	multiply two ranums using naive algorithm, generating double length 
;-	result, then round to single.
vmul:GA;:ok:		; ( n1 d1 n2 d2
	call vmmul
	jr dv2v
;-
;- vsgn		( v -- -1 | 0 | 1 )		(m,)		'vsgn'		rational
;-	ret sing signum of a ranum
vsgn:GA M;:ok:
	dec eax
	spop
	jnz .z		;?; non-zero?
	test eax,eax
	jz .r
.z:
	sar eax,31
	lea eax,[eax*2+1]
.r:
M vsgn
	ret
;-
;- v+-		( v1 n -- v2 )			(m,)		'vpm'		rational
;-	negate v1 if n < 0
vpm:GA M
	test eax,eax
	spop 
	jns vnegate.r
;-
;- vnegate	( v -- v )			(m,)				rational
;-	return { ranum -1 * }. The sign bit is in the numerator.
vnegate:GA M;:ok:
	neg dword TOS(1)
.r:
M vnegate
M vpm
	ret
;-
;- vabs		( v -- |v| )			( m, )				rational
;-	return absolute value of ranum argument.
vabs:GA M;:ok:
; SWAP ABS SWAP ;
	mov ecx,TOS(1)
	test ecx,ecx
	jns .r
	neg ecx
	jns .s
	shl ecx,1
.s:
	mov TOS(1),ecx
.r:
M vabs
	ret
;-
;- 1/v		( v -- v )			(m,)		'vrecip'	rational
;-	the reciprocal of a ranum.
vrecip:GA M;:ok:
;  SWAP DUP 0< IF  ABS vnegate  THEN ;
	mov ecx,TOS(1)
	test ecx,ecx
	jns .r
	neg eax
	neg ecx
.r:
	mov TOS(1),eax
	mov eax,ecx
M vrecip
	ret
;-
;- v1+		( v1 -- v2 )					'vonep'		rational
;-	increment v1 by one
vonep:GA;:ok:
	mov ecx,eax	; z
	add TOS(1),eax	; z := z+n
	jo .o
	ret
.o:A
	xor eax,eax
	spush 2
	mov TOS(1),ecx
	jmp dv2v
;-
;- v1-		( v1 -- v2 )					'vonem'		rational
;-	decrement v1 by one
vonem:GA;:ok:
	mov ecx,TOS(1)	; z
	sar ecx,31	; s>d
	sub TOS(1),eax	; z := z-n
	jo .o
.r:; <v2mul,v2div)
	ret
.o:A
	spush 2
	mov eax,0	; clr den.hi
	sbb ecx,eax	; C
	mov TOS(2),ecx
	jmp dv2v
;-
;- v2*		( v1 -- v2 )					'v2mul'		rational
;-	multiply v1 by 2
v2mul:GA;:ok:
	shr eax,1	; den 2/
	jc .c
.r:
	ret
.c:A
	shl dword TOS(1),1; num 2*
	lea eax,[eax*2+1]
	jno .r
	sbb ecx,ecx
	spush 2
	mov TOS(2),ecx
	xor eax,eax
	jmp dv2v
;-
;- v2/		( v1 -- v2 )					'v2div'		rational
;-	divide v1 by 2
v2div:GA;:ok:
	mov ecx,TOS(1)
	sar ecx,1
	jc .c
	mov TOS(1),ecx
	ret
.c:A
	shl eax,1
	jo .o
	ret
.o:A
	sar ecx,30
	spush 2
	mov TOS(2),ecx
	xor eax,eax
	jmp disp dv2v
;-
;-	
;--- Comparison
;-

;-
;- v0<		( v -- f)			(m,)		'vzlt'		rational
;-	flag is TRUE if v is negative.
vzlt:GA;:ok:
	spop
	sar eax,31
	ret
;-
;- v~		( v v -- f )					'vaeq'		rational
;-	tf if the two ranum arguments are approximately equal.
;-	The difference between two ranums may be less than the smallest ranum that
;-	is representable with a single cell denominator. In this instance the two
;-	ranums may be reasonably described as approximately equal.
vaeq:GA;:ok:
; v- v0= ;
	mov ecx,eax
	or ecx,TOS(1)
	or ecx,TOS(2)
	or ecx,TOS(3)
	jz near qzeq
	call vsub	;:contiued:
;-
;- v0=		( v -- f )					'vzeq'		rational
;-	flag is true if v is zero.
vzeq:GA;:ok:
; 0 1 D= ;
	dec dword TOS(1)
	jmp dzeq
;-
;- v=		( v1 v2 -- f )					'veq'		rational
;-	tf if v1, v2 exactly equal.
veq:GA;:ok:
	call mrot
	call mmul
	call swap2
	call mmul
	jmp deq
;-
;- v>		( v1 v2 -- f )					'vgt'		rational
;-	tf if v1 greater than v2.
;-	Returns FALSE if either	argument is the overflow condition.
vgt:GA;:ok:
;  2SWAP v< ;
	call swap2	;:continued:
;-
;- v<		( v1 v2 -- f )					'vlt'		rational
;-	tf if v1 less than v2. 
;-	Returns FALSE if either argument is the overflow condition.
vlt:GA;:ok:(vmax)	; n1 d1 n2 d2
; ROT ROT M* 2SWAP M* 2SWAP D< ;
	imul dword TOS(3); d2n1
	mov ecx,TOS(2)	; d1
	mov TOS(3),eax	;	dn21l
	mov TOS(2),edx	;	dn21h
	mov eax,TOS(1)	; n2
	imul ecx	; d1n2
	cmp TOS(3),eax	;	dn21l-dn12l
	mov eax,TOS(2)	; d21h
	sbb eax,edx	;	dn21h-dn12h
	TOSPULL(3)
	sar eax,31
	ret
;-
;- voverflow	( v -- f )							rational
;-	f is TRUE if v indicates overflow has occurred previously.
;-	This word should be used where overflow is possible. (propagation
;-	of overflow is not guaranteed.)
    global voverflow:function
voverflow:equ dzeq
;-
;- vmax		( v1 v2 -- v3 )							rational
;-	returns the larger of the two ranum arguments. 
;-	Will always return the overflow condition, if present.
vmax:GA;:ok:
; 2DUP voverflow >R 4DUP v< R> OR IF 2SWAP THEN 2DROP ;
	mov ecx,TOS(1)
	or ecx,eax
	push ecx	; d0=/= (= not overflow)
	call dup4
	call vlt
.n:
	pop ecx
	not eax
	and eax,ecx
	spop
	jz near sdrop2
	spop 2
	ret
;-
;- vmin		( v1 v2--v3)							rational
;-	returns the smaller of the two ranum arguments. 
;-	Will always return the overflow condition, if present.
vmin:GA;:ok:
; 2DUP voverflow >R 4DUP v> R> OR IF 2SWAP THEN 2DROP ;
	mov ecx,TOS(1)
	or ecx,eax
	push ecx	; d0= (=voverflow)
	call dup4
	call vgt
	jr vmax.n
;-
;-	
;--- Conversion
;-

;-
;- s>v		( n -- v )					's2v'		rational
;-	promote single integer to ranum equivalent. n --> n 1
s2v:GA;:ok:(vsplit)
	spush
	mov eax,1
	ret
;-
;- v>s		( v -- n )					'v2s'		rational
;-	return the floored, integral component of v as a signed integer.
v2s:GA;:ok:(vsplit)
;  >R S>D  R> FM/MOD NIP ;
	mov ecx,eax	; >r
	shl eax,1
	jz .o
	sub eax,byte 2
	mov eax,TOS(1)
	jz near drop
	sar eax,31	; s>d
	spush
	mov eax,ecx	; r>
	call fmbymod
.o:
	TOSPULL(1)
	ret
	
;-
;- vsplit 	( v1 -- v2 v3 )							rational
;-	floored integral v2 and (non-negative) fractional v3 component of v1. 
;-	w/ Floored rounding, i.e. v1 = v2 + v3 and assuming +ve v3 is always valid. 
;-	word is tolerant of the overflow condition.
vsplit:GA;:ok:
;  2DUP DUP IF v>s s>v 2SWAP THEN vfrac ;
	call dup2
	test eax,eax
	jz vfrac.r	; ovf
	call v2s
	mov edx,TOS(2)
	mov ecx,TOS(1)
	mov TOS(2),eax
	mov eax,edx
	mov dword TOS(1),1
	spush
	mov eax,ecx	;:continued:
;-
;- vfrac  	( v1 -- v2 )							rational
;-	return the fractional component of v1 as non-negative ranum v2.
;-	w/ Floored rounding. word is tolerant of the overflow condition.
vfrac:GA;:ok:
; dup 0= -exit TUCK (=swap over) >R S>D R> FM/MOD DROP SWAP ;
	test eax,eax
	mov ecx,TOS(1)
	jz .r
	TOSPUSH(1)
	mov TOS(1),ecx
	mov TOS(2),eax	; sover
	push eax	; >r
	sar ecx,31
	mov eax,ecx	; s>d
	spush
	pop eax		; r>
	call fmbymod
	mov ecx,TOS(2)
	spop
	mov TOS(1),eax	; drop swap
	mov eax,ecx
.r:
	ret	
;-
;--- Rounding
;-

;-
;- v>uv?  	( v -- uv f)					'v2uvq'		rational hidden
;-	uv is the absolute value of the v, f is TRUE if v was negative.
v2uvq:GA;:ok:
;  2DUP vabs ( = swap abs swap ) 2SWAP v0< ;
	mov ecx,TOS(1)	; lo1
	test ecx,ecx
	mov edx,ecx	; hi2
	jns .s
	neg ecx
.s:
	spush
	mov TOS(2),ecx	; swap abs swap
	mov eax,edx	; 2swap drop
	sar eax,31	; 0<
	ret
;-
;- vround 	( v1 +n -- v2 )			(x)				rational
;-	returns an approximation to the rational number v1 such that neither 
;-	the numerator nor the denominator exceeds the positive integer n.
;-	The intent is that, where high precision is not an issue, vround can 
;-	be inserted at key points in the code to ensure the quicker "normalise" 
;-	path is taken in subsequent words that call DV>V
vround:GA;:ok:
;  >R 2dup voverflow if rdrop exit endif
	mov ecx,TOS(1)
	or ecx,TOS(2)
	jz near drop
;  setvars v>uv? ( v -- uv f ) R> also ans 2>R previous
	push eax	; save limit
	spop
	call v2uvq
	spush
	mov eax,1
	spush
;  { sg.dv prev.numerator denominator | numerator prev.denominator } [ = setvars ]
	kloci 5
.pnu:	klocv		; prev.numerator
.dn:	klocv		; denominator
.sg:	klocv		; sign flag from v>uv?
.nu:	klocz		; numerator
.pdn:	klocz		; prev.denominator
.enter:;    enter
	jr .entry
.begin:;    begin	; ( d u			[ limit
;	OVER DUP IF /mod THEN		(1)
	spush
	mov eax,TOS(2)
	test eax,eax
	jz .endif1	; if
	call bymod	; /mod
.endif1:
;        DUP r@ snum>?
	spush		; dup
;snum>? ( u1 u2--u3 f)
;  >R numerator * prev.numerator + DUP R> U> ;
;the next approximation to the numerator u3, is the current approximation
;multiplied by u1, which is the next number in the continued fraction of
;the ranum being approximated, plus the previous approximation. 
;flag is true if the next approximation is greater than u2.
	imul dword[lref .nu]
	add eax,[lref .pnu]
	cmp eax,[esp]	; r@ ... >r ... r> u>
;	IF next!
	ja .if2
;	else [ 2 cs-roll 2 cs-roll ]
;	    swap r@ sden>? 
	mov ecx,TOS(1)
	mov TOS(1),eax
	mov eax,ecx	; swap
;sden>? ( u1 u2 -- u3 f)
;  locref{ prev.numerator denominator numerator prev.denominator }
;  >R denominator * prev.denominator + DUP R> U> ;
;the next approximation to the denominator u3, is the current approximation
;multiplied by u1, which is the next number in the continued fraction of
;the ranum being approximated, plus the previous approximation.
;flag is true if the next approximation is greater than u2.
	imul dword[lref vround.dn]
	add eax,[lref vround.pdn]
	cmp eax,[esp]
	ja .if3
;	    IF next! else [ 2 cs-roll 2 cs-roll ] 
;		next! SWAP
;    entry dup 0= until
	call vnormal.nexts
	mov ecx,TOS(1)
	mov TOS(1),eax
	mov eax,ecx
.entry:;    entry
;	dup 0=
	test eax,eax
.until:;    until
	jnz .begin
.endif23:;    endif endif
;    endif endif
;    2drop rdrop
	add esp,byte 4
;   prev.numerator R> +- prev.denominator ;
	mov ecx,[lref vround.pnu]
	cmp dword[lref vround.sg],byte 0
	jns .p
	neg ecx
.p:
	mov eax,[lref vround.pdn]
	mov TOS(1),ecx
	ret
.if2:A
.if3:
	call vnormal.nexts
	jr .endif23
;-
;- vsimplify	( v1 +n -- v2 )			(4)				rational
;-	returns an approximation to the rational number v1 such that neither the
;-	numerator nor the denominator of the fractional component of v1 exceeds 
;-	the positive integer n.
vsimplify:GA;:ok:
;  >R vsplit R> vround v+ ;
	push eax
	spop
	call vsplit
	spush
	pop eax
	call vround
	jmp vplus
;-
;-	
;--- Memory Ref.
;-

    global vstore:function,vfetch:function
;-
;- v@		( p -- v ) 							rational
;-	fetch v from memory at p
vfetch:equ twofetch
;-
;- v!		( v p -- ) 							rational
;-	store v to memory at p
vstore:equ twosto

;-
;-	
;--- Stacks Ref.
    global vdrop:function,vdup:function,v2dup:function,vmrot:function,vover:function,vpick:function,vrot:function,vroll:function,vswap:function
;-
;- v-rot  	( v1 v2 v3 -- v3 v1 v2 )					rational
vmrot:equ	mrot2
;- vdrop  	( v1 -- v1 v1 )							rational
vdrop:equ	drop2
;- vdup   	( v1 -- v1 v1 )							rational
vdup:equ	dup2
;- v2dup  	( v1 v2 -- v1 v2 v1 v2 )					rational
v2dup:equ	dup4
;- vover  	( v1 v2 -- v1 v2 v1 )						rational
vover:equ	over2
;- vpick  	( ... n -- ... vn )						rational
vpick:equ	pick2
;- vrot   	( v1 v2 v3 -- v2 v3 v1 )					rational
vrot:equ	rot2
;- vroll  	( ... n -- ... vn )						rational
vroll:equ	roll2
;- vswap  	( v1 v2 -- v2 v1 )						rational
vswap:equ	swap2
;-	double cells opr. apply to (current) rational numbers impl.
;-
;-
;-	
;-

;-
;- #v		( +v -- 0 0 )					'sharpv'	rational
;-	append positive ranum +v to the pictured numeric output string according to
;-	the following rules.
;-	  If +v is zero append "0".
;-	  If +v has an integral component but no fractional component, append that
;-	        number.
;-	  If +v has a fractional component but no integral component append "a/b"
;-	        where a is the numerator, and b the denominator.
;-	  If +v has both a an integral and a fractional component append "a b/c"
;-	        where a is the integral component, b the numerator and c the
;-	        denominator of the fractional component. There is a space between
;-	        the integral and fractional components.
;-	The conversion is done according to the current value of BASE.
sharpv:GA;:ok:
; 2DUP v0=
; IF[1]  "0 HOLD
; ELSE[1]  DUP 1-
;    IF[2]  DUP S>D #S 2DROP "/ HOLD 2DUP MOD S>D #S 2DROP 2DUP < 0IF[3] BL HOLD THEN[3] THEN[2]
;    2DUP < 0IF[4] v>s S>D #S THEN[4]
; THEN[1] 2DROP 0. ;
	cmp dword TOS(1),byte 0
	jz .z		; 2dup v0= if[1] "0 hold (re below) 
	cmp eax,byte 1	; else[1] dup 1- if[2]
	jz .endif2	;?; no fraction
	spush 2
	sar eax,31	; dup s->d
	call sharps	; #s
	spop
	mov eax,'/'	; 2drop "/ hold
	call hold
	mov ecx,eax
	mov edx,TOS(1)
	spush		; 2dup
	mov eax,edx
	sar edx,31
	idiv ecx	; mod (using dft xep-hnd)
	mov eax,edx
	spush
	sar eax,31	; s>d
	call sharps	; #s
	spop 2		; 2drop
	cmp eax,TOS(1)	; 2dup <
	jg .endif4	; 0= if
	call fdelim	; {#fd} hold (delimiting char between integer and fractional part)
	call hold
.endif2:; < integral part
	cmp eax,TOS(1)	; 2dup <
	jg .endif4	; 0= if
	call v2s	; v>s
	spush
	sar eax,31	; s>d
	call sharps	; #s
.endif4:
	xor eax,eax	; 2drop 0.
	mov TOS(1),eax
	ret
.z:A;	; v-zero or overflow
	mov eax,'0'	; "0 hold
	call hold
	xor eax,eax	; 2drop 0.
	spush
	ret
;-
;- #vf		( +v -- 0 0 )					'sharpvf'	rational
;-	append the positive ranum +v to the pictured numeric output string as a
;-	floating point number. That is to say, the integral component, followed by
;-	a point, followed by as many digits as specified by places. If places is zero
;-	no point is included. Returns 0 0 for consistency with other pictured numeric
;-	output conversion words. The conversion is done according to the current
;-	value of BASE.
sharpvf:GA;:ok:
;  0 SWAP >rr um/mod swap base@ u* r>	\ isolate integer
;  #prec 0> if
;      #prec 0 DO >rr UM/MOD h>a hold BASE@ UM* R> i 0< -leave LOOP
;      hld @ dup hld0 @ over- >move< 	\ sto fraction
;      #dp HOLD 
;  endif 3drop s>d #S ;			\ sto dp and integer
	mov ecx,eax
	push ecx	; 0 swap >rr	[ den
	mov eax,0
	spush
	mov eax,ecx
	call umbymod
	call swap
	mul dword[byte uref uv.base]
	push edx
	call fprec
	mov ecx,eax
	pop eax
	test ecx,ecx
	jng .e
.p:
	push ecx	; places >r	[ pl den
.do:A
	spush
	mov eax,[esp+4]
	call umbymod
	call h2a
	call hold
	mul dword[byte uref uv.base]
	spush
	mov eax,edx
.loop:
	dec dword[esp]
	jg .do
	spush
	mov ecx,[byte uref uv.hld0]
	mov eax,[byte uref uv.hld]
	spush 3
	sub ecx,eax
	mov eax,ecx
	call movex
.h:
	call vpoint
	call hold
	spop
	add esp,byte 4
.e:
	spop
	add esp,byte 4
	mov eax,TOS(1)
	sar eax,31	; s>d
	jmp sharps
;-
;- <#vf#> 	( v -- p n )					'ltvfgt'	rational hidden
;-	Returns a string representing the ranum v by stg(p,u).
;-	The string is as described in #vf, but will have a preceding minus sign if v
;-	was negative. If v is the overflow indicator, returns the string "Overflow".
ltvfgt:GA;:ok:
; 2DUP voverflow IF 2DROP S" Overflow" exit endif
; over abs swap <# #vf fig sign #> ;
	lea edx,[kref sharpvf-Y]
.n:;ovf: 0 0 or -0 0
	mov ecx,eax
	shl ecx,1
	mov ecx,TOS(1)
	jz sharpovf
	neg ecx
	jns .c
	neg ecx
.c:
	spush
	mov TOS(1),ecx
	call ltsharp
	call edx	; <sharpvf> or <sharpv>
	call sign	; (fig-version)
	jmp sharpgt
;
sharpovf:A
	mov eax,vovt-Y
	TOSPULL(1)
	jmp ccount
;-
;- <#v#>  	( v -- p u )					'ltvgt'		rational hidden
;-	Returns the string(p,u) representing the rational number v. The string 
;-	is as described in #v (above), but will have a preceding minus sign if 
;-	v was negative. If v is the overflow indicator, returns stg "Overflow".
ltvgt:GA;:ok:
;  2DUP voverflow IF 2DROP S" Overflow" exit then
;  OVER ABS SWAP <# #v ROT SIGN #> ;
	lea edx,[kref sharpv-Y]
	jmp disp ltvfgt.n
;-
;- f.j		( v n1 n2 -- )					'vdotfj'	rational
;-	display ranum v according to the rules described in <#vf#>. 
;-	The display string will be padded with leading spaces until there are n1
;-	characters before the point and padded with trailing spaces until there are
;-	n2 characters after the point. This allows numbers to be aligned about the
;-	point for display in tabular form. If there is no point in the string it will
;-	be padded with n2+1 trailing spaces for consistency. Trailing zeroes will be
;-	truncated according to the rules described in -re- {-zeroes}. If n1 is less
;-	than the number of characters to be displayed before the point no spaces will 
;-	be displayed, but the string will not be foreshortened. Therefore setting n1
;-	to zero will switch off right justification. If n2 is less than the number of
;-	characters to the right of the point no spaces will be displayed, but the
;-	string will not be truncated. If n2 is zero and places is also zero one
;-	trailing space will be displayed in lieu of the point. Therefore n2 should be
;-	-1 to switch off left justification, as this will inhibit all trailing spaces.
vdotfj:GA;:ok:
;  1+ >R >R <#vf#> truncation IF -zeroes THEN
;  2DUP point >ch R> OVER- SPACES OVER 2SWAP TYPE - r> + SPACES ;
	inc eax		; 1+
	mov ecx,TOS(1)
	push eax	; (2)
	push ecx
	spop 2
	call ltvfgt
.n:; <dvdotfj)
	call vtrunc
	test eax,eax
	spop
	jz .endif1
	call mzeroes
.endif1:
	call dup2
	call vpoint
	call gtch	; >ch
	mov ecx,eax
	spush
	pop eax
	sub eax,ecx	; over-
	call spaces
; over 2swap ; ( 1 2 3 -- 3 2 1 2
	mov ecx,TOS(2)	; (1)
	mov edx,TOS(1)	; (2)
	mov TOS(2),eax	; (3)
	mov eax,ecx	;(1)->
	spush
	mov eax,edx	;(2)->
	call type
	mov ecx,TOS(1)
	sub ecx,eax	; - r> +
	spop
	pop eax
	add eax,ecx
	jmp spaces
;-
;- f.		( v -- )					'vdotf'		rational
;-	display a rational number in floating point format as described in v.fj 
;-	but without left or right justification, and with one trailing space.
vdotf:GA;:ok:
;  0 -1 v.fj SPACE ;
	spush
	xor eax,eax
	spush
	dec eax
	call vdotfj
	jmp space
;-
;- v.j		( v n1 n2)					'vdotj'		rational
;-	display rational number v in fractioanl format, as described in <#v#> (above),
;-	padded with up to n1 leading spaces, and up to n2 trailing spaces, such that
;-	the number is justified about the space between the integral and fractional
;-	component, if present. If there is no integral component an additional
;-	leading space is displayed in lieu of the central space. If there is no
;-	fractional component an additional trailing space is displayed in lieu of the
;-	central space. No more than "digits" number of characters will be displayed
;-	in the numerator or denominator. n1 should be -1 to switch off leading spaces
;-	entirely, and likewise n2 for trailing spaces.
vdotj:GA;:ok:(scheint so...)
; >R[1] >R[2] digits> vsimplify  <#v#>
	push eax	;		[ R1
	push dword TOS(1); >r >r	[ R2 R1
	spop 2
	call vdigitsgt	; digits>
	call vsimplify
	call ltvgt	; <#v#>
.n:; <dvdotj)	; 2DUP [CHAR] / has-ch
	call dup2	; -- p u p u 
	spush
	mov eax,'/'
	call scan	; "/ hasch
	test eax,eax
	spop 2		; -- p u 
; IF 2DUP BL >ch 2DUP >R =
	jz .else1	;?; no fraction
	call dup2	; -- p u p u 
	call fdelim	; #fd
	call gtch	; -- p u u'
	push eax	; 2dup >r ; -- p u u' u	[ u' R2 R1
	cmp eax,TOS(1)	; = 	  ; -- p u f
;    IF 2R> >R 1+ SPACES TYPE 2R> - SPACES	; [2]
	jnz .else2	;?; bl not at end of stg, i.e. integer and fraction
;	; <no integral part>
	pop ecx		; Ru
	pop eax		; R2
	push ecx	;		[ Ru R1
	inc eax
	call spaces
	call type
	spush
	pop ecx
	pop eax
	sub eax,ecx
	jmp spaces
;    ELSE[2]  2R> DUP >R - SPACES TUCK TYPE 2R> + 1+ SWAP - SPACES THEN[2]
.else2:A
	pop edx		; 2R> dup >r (ans) 	:=>  2r> swap >rr	:=> r
	pop eax
	push edx
	sub eax,edx	;	{ R2 u		[ u R1
	call spaces	; spaces
	call sover	; tuck
	call type	; type
	pop edx
	pop ecx		; 2r>
	inc ecx
	add ecx,edx	; + 1+
	sub ecx,eax	; swap -
	mov eax,ecx
	jmp spaces
; ELSE[1] R> OVER - SPACES TYPE R> 1+ SPACES THEN[1] ;
.else1:A;<no fraction)	; ( p u 
	mov ecx,eax
	spush
	pop eax		; r>
	sub eax,ecx	; over-
	call spaces	; spaces
	call type	; type
	spush
	pop eax		; r> 1+
	inc eax
	jmp spaces	; spaces endif
;-
;- v.		( v -- )					'vdot'		rational
;-	display a rational number in fractional format as described in v.f 
;-	but without justification and, with one trailing space.
vdot:GA;:ok:
;  -1 -1 v.j SPACE ;
	spush
	mov eax,-1
	spush
	call vdotj
	jmp space
;-
;- --
;-	

;-
;- vsqrt  	( v1 -- v2 )							rational
;-	v2 := sgn(v1) * square root of ranum |v1|.
;-	accuracy ok w/ less than 5 decimal digits, -re- {dvsqrt} for higher praecision.
vsqrt:GA;:ok:
	call v2dv
	call dvsqrt
	jmp disp dv2v
;-
;- v^n		( v1 +n -- v2 )					'vpovn'		rational
;-	raise v1 to the power of non-negative sing n. v2 := ovf if n < 0.
;-	e.g, max e^n w/ n = 21.
vpown:GA;:ok:
; 1 s>v ROT 0 ?DO 2OVER v* LOOP 2SWAP 2DROP ;
; :HIER: wertebereich erhhen, z.b. durch {vround}, gebrochenen exp. aus quadratwurzeln zusammensetzen
	test eax,eax
	push eax
	mov eax,1
	spush	; 1 s>v rot 0 do
	js .v
	jz .z
.do:
	mov ecx,TOS(3)
	spush 2
	mov eax,TOS(4)
	mov TOS(1),ecx	; 2over
	call vmul
	dec dword[esp]
	jg .do
.z:
	add esp,byte 4
.d:
	jmp sdrop2
.v:A
	pop eax
	jr .d
;-
;- s>vfactorial	( n -- v )					's2vfac'	rational
;-	v := n!, v := overflow if not fitting into a single ranum (max n = 12).
s2vfac:GA;:ok:
;  1 SWAP 1+ 1 ?DO I * LOOP ;
	call s2nfac	; ret bignum type
	mov ecx,eax
	dec ecx
	jz .r		;?; sing
	dec ecx
	jz .v		;?;
	mov eax,ecx
	call ndrop
.v:
	xor eax,eax
	mov TOS(1),eax
.r:
	ret
; --
; -- several formulae, re Bronstein-Semendjajev, Taschenbuch der Mathematik, Teubner/Nauka 1979, pg 82 ff.
; --

;-
;- e^v		( v1 -- v2 )					'epowv'		rational
;-	returns approximation to e^x, where x is the rational number v1 | 6 > v1 > -6.
;-	Computed as 1st {#s!n} terms (dft 13) of power series; +/- 10^(-8) for |v| < 1.
;-	13 to {#n!v} is maximum for non-overflowing internal factorial calculations.
epowv:GA;:ok: - ungenau, und langsam...
;  0 S>V 13 0 DO 2OVER I v^n I s>vfactorial v/ v+ LOOP 2SWAP 2DROP ;
;			v(n-1)^n	n!	Sa( v^n/n! ), n=1..oo
	spush
	_mov eax,1
	spush
	call snfv	; #s!v
	dec eax
	push eax
	dec eax		; 1st iteration := 1
	push eax
	spop
.do:
	call over2	; v0
	spush
	mov eax,[esp+4]
	sub eax,[esp]	; n
;:HIER: wertebereich erhhen, z.b. durch {vround} rsp ber double -> sing mit {vnormal}, oder (neu) {v*/} <= :HIER:
	call vpown	; v0^n 	;	==>	v0^(n):=v0^(n-1)*v0
	spush
	mov eax,[esp+4]
	sub eax,[esp]
	call s2vfac	; n!	;	==>	n!:=(n-1)!*n
	call vby	; v/
.p:
	call vplus	; v+
	dec dword[esp]
	jns .do
	add esp,byte 8
	jmp sdrop2
;-
;-	
;--- mixed, sing and double size
;-

;-
;- vm*		( v v -- dv)					'vmmul'		rational
;-	multiply two ranums, generating double length result.
vmmul:GA;:ok:
; ROT UM* 2SWAP M* 2SWAP dv>v ; \ dv := nom1 * nom2 / den1 / den2
	mov ecx,eax	; dn2
	mov eax,TOS(1)	; nm2
	imul dword TOS(3); pnm := nm2 * nm1
	mov TOS(3),eax	; pnm.lo
	mov eax,TOS(2)	; dn1
	mov TOS(2),edx	; pnm.hi
	mul ecx		; pdn := dn1 * dn2
	mov TOS(1),eax
	mov eax,edx
	ret
;-
;- qv>v		( qv1 -- v2 )			(4)		'qv2v'		rational
;-	reduce quad, qv1 to single cells size, v2.
qv2v:GA
	call qv2dv
	jmp dv2v
%endif;__SRANUM_I
;-
;-================================================================
;- l4sranum.inc <eof>
