;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4string.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __STRING_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	

%ifndef __STRING_I
%define __STRING_I
%define stringinc
;-
;-================================================================
;--- text acquisition, strings, search & substitution
;-
;-	

;-
;-	F8 type string is either one of
;-		standard string w. leading countbyte and max 255 chars.
;-		any length string enclosed by <nul> bytes,
;-			leading flag & trailing delimiter.
;-	some words operate only up to a certain maximum of bytes (4K), 
;-		-re- {place-size} for the limitation on {place}ing words.
;-	{type} is limited to 64K bytes.
;-

;-
;- +blanks	( u1 u2 -- u1 p2 u2 )				'pblanks'	forth
;-	store count and u2:=min({pad-size}-1,max(0,u1-u2)) blanks to {here}, at 
;-	p2-1, ready to {type}. {pad-size} redefinable w/ uvari (padsz), dft 256.
;-	interpreting: { s" hallo" ] 20 +blanks type [ type } right justified.
;-	-re- {pad-size}, {(padsz)} and, {bl.r} which jumps into {type}.
pblanks:GA;:ok:
	sub eax,TOS(1)		; d:=u2-u1
	call here
	mov edx,[byte uref uv.padsz]
	mov ecx,TOS(1)		; h d u1
	cmp ecx,edx
	jc .p			; unsigned!
	lea ecx,[edx-1]
	mov TOS(1),ecx
.p:
	mov byte[kref eax],cl	; count to {here}
	mov TOS(1),eax
	mov eax,ecx		; d h u1
	call dup2		; d h  d h u1
	inc dword TOS(1)	; d h+ d h u1
	jmp blanks		; d h u1
;-
;- -trailing	( p1 u1 -- p1 u2 )		(fg)		'mtrailing'	forth
;-	cut trailing blanks and 'whitespace' off stg(p1,u1).
;-	'whitespace' in this context is <tab>, <del>, <bs>.
mtrailing:GA;:ok:
; c: ecx
	mov ecx,eax
	add eax,KR
	add eax,TOS(1)
.l:
	jecxz .r
	dec ecx
	dec eax
	cmp byte[eax],d_del	; 127 = AC_DEL
	jz .l			;?; cut <del>
	ja .d			;?; ret if code > 127
	cmp byte[eax],d_bl
	jz .l			;?; cut blank
	ja .d			;?; all 'above' is text
	cmp byte[eax],d_tab
	jz .l			;?; cut <tab>
	cmp byte[eax],d_bs
	jz .l			;?; cut <bs>
.d:
	inc ecx
.r:
	mov eax,ecx
	ret
;-
;- zcount  	( p1 -- p1 u )			(e)				forth
;-	find length u of <nul> terminated strig at p1,
;-	ret u := 0 if p1 = 0
zcount:GA;:ok:
	spush
	add eax,KR		; get s value
zcount.n:
	push edx
	jz .e			;?; zero address
.a:	test byte[eax],-1	; realign pointer on dword boundaries
	jz .e
	inc eax
	test eax,3
	jnz .a
.l:
	mov edx,[eax]		; test 4 bytes at one time
	lea ecx,[edx+0xfefefeff]
	not edx
	and edx,ecx
	add eax,byte 4
	test edx,0x80808080	; if one of the byte is null
	jz .l
.f:
	shr edx,8		; check first lo byte
    	inc eax
	jnc .f
	sub eax, byte 5		; realign pointer to null char
.e:
	sub eax,KR
	pop edx
	sub eax,TOS(1)
	ret
;-
;- count  	( p1 -- p2 u )					'ccount'	forth
;-	p2 := p1 or p1+charsize, u := string(p1) length; u := 0 if p1 = 0.
;-	{count} 1st checks whether leading byte is non-zero and returns, 
;-	else scans memory until trailing <nul> and returns asciz string's 
;-	character count. 
;-	NOT A SUBSTITUTE to { 1+ dup 1- c@ } -re- {c@+}; for instance, 
;-		{ bl word count } will fail on empty input!
;-	-re- {zcount} and ans-voc {COUNT}.
ccount:GA;:ok:
; c: ecx
	call count
	jz .r		;?; zero address
	test eax,eax
	jnz .r		;?; non-zero count
	spop
	call zcount
	mov ecx,TOS(1)	; leave ecx:=ptr for caller, cnt @tos
	cmp eax,eax	; len pts to 4th-stg:stgend, asciz:<nul>char
.r:
	ret
;-
;- COUNT  	( p1 -- p2 u )							ans
;-	p2 := p1+charsize, u := string(p1) length; u := 0 if p1 = 0.
;-	NOTE: implementation of 'characters' may change for unicode, etc, thus neither
;-	variant of {count} should be (ab)used to substitute { 1+ dup 1- c@ } - re c@+
count:GA;:ok:
; c: ecx, eflags = Z if true address = 0
	mov ecx,eax
	add ecx,KR
	jz .r		;?; ret zero count if ptr to zero adress
	inc eax
	movzx ecx,byte[ecx]
.r:
	spush
	xchg eax,ecx	; leave ecx:=ptr for caller, count @tos
	ret		; len pts to 4th-stg:stgend, asciz:<nul>char

;-
;- blw    	( -- p )			(k)		'BLWORD'	novoc
;-	does{ ans bl word }
BLWORD:GA;:ok:
	call bbl
;-
;- word		( c -- p )			(4)(a:c)(x)	'XWORD'		forth
;-	scan input source for max 255 chars of text sequence enclosed by chars c.
;-	returns ptr p to 4th-strg. a blank, not included in length, follows the string.
;-	programs may replace characters within the string. this is the ANS-style {word}
XWORD:GA;:ok:
; c: ecx
	call xword
	jmp here
;-
;- blword  	( -- )				(k)(4)(x)
;-	does{ bl word }
blword:GA
	call bbl
;-
;- word		( c -- )			(4)((x))	'xword'		fig
;-	scan input source for text sequence enclosed by chars c, max len 255.
;-	fig version, returns counted stg to {here}, where it may be modified.
xword:GA;:ok:
; >r source drop in @ + span @ in @ - 0> 0=
; IF rdrop drop 0
; ELSE dup -rot r skip dup -rot r> parsec 2swap - over + 1+ in +! 255 min 2dup here place + off ENDIF ;
; c: ecx
	call source	;		; u p c
	mov ecx,[byte uref uv.in];less parsing posn
	add TOS(1),ecx	;parse area
	sub eax,ecx	; remaining cnt -> +u
	jle near .e	; 0> IF
	push eax	;		; u p c		[ u
	mov ecx,TOS(2)	; c
	mov edx,TOS(1)	; p
	mov TOS(1),eax	; u
	mov TOS(2),edx	; p   rot
	push ecx	;		; c u p		[ c u
	mov eax,ecx	;   c 		; c u p		[ u
	call skip	; skip		; u1 p1
	mov ecx,eax
	spush
	mov eax,[esp]
	mov [esp],ecx;u1		; c  u1 p1	[ u1 u
	call parsec	;		; u2 p2		[ u1 u
	pop edx		; u1
	pop ecx		; u
	sub ecx,edx	; u-u1
	lea ecx,[ecx+eax+1];u-u1+u2+1
	add [byte uref uv.in],ecx
	spush
; truncate to max pad-size chars before(!) transfer to supplied buf.
	mov ecx,TOS(1)	; u2
	mov eax,[byte uref uv.padsz]
	sub eax,byte 4
	cmp eax,ecx
	jnc .p		;?; in range
	mov ecx,eax
	mov TOS(1),ecx
.p:
	mov eax,dataptr	; here
	lea ecx,[eax+ecx+1]
	mov byte[kref ecx],d_bl
	jmp place	; place
.e:A;			; ELSE ( parse area exhausted )
	xor ecx,ecx	; set Z
	mov eax,dataptr;here
	mov [kref eax],ecx; clr storage space
	spop 3
	ret		; ;S
;-
;- parse  	( c -- p u )			((x))				forth
;-	ret ptr p to before next occurrence of char c in current 
;-	input buf. after disp of {in@}, adjust {in} by +u.
;-	if c=<bl> find any delimiter <bl> or ctrl-code 0..31
parse:GA;:ok:
; c: ecx,edx
	call tor	; save c
	call source	; a +u
	mov ecx,[byte uref uv.in];less parsing posn
	add TOS(1),ecx	;parse area
	sub eax,ecx	; remaining cnt -> +u
	jle parse.e	;?; ready
	call rfrom
	call parsec
	add [byte uref uv.in],eax
	inc dword[byte uref uv.in]; discard delimiter ::-?-::
	ret
.e:A; parse area exhausted
	lea esp,[esp+4]
	xor eax,eax
	ret
;-
;- enclose	( p c -- p n1 n2 n3 )						fig
;-	separate by char c enclosed text in max. {#tib} chars at ptr p+
;-	ret n1 disp to 1st char =/= c, n2 next char = c, n3 next char
;-	ans stg then { drop over - rot rot + swap }
enclose:GA
	push dword[uref uv.uintib]; #tib
	jr enclose4.n
;-
;- enclose	( p u c -- p n1 n2 n3 )				'enclose4'	forth
;-	separate by char c enclosed text in max. u chars at ptr p+
;-	ret n1 disp to 1st char =/= c, n2 next char = c, n3 next char
;-	ans stg then { drop over - rot rot + swap }; w/ char = byte.
enclose4:GA;:ok:
	TOSPULL(1)
	push dword[PS]	;		[ u
.n:;		; ( p c -- ...
	mov edx,TOS(1)	; p
	cmp al,[kref edx]
	spush 2
	mov TOS(1),edx	; c p c p
	jnz .s		;?; 1st char =/= dlm
	mov ecx,eax	; c
	mov eax,[esp]	; u
	spush
	mov eax,ecx	; c u p c p
	call skip
	mov [esp],eax	;;u';
	mov edx,TOS(1)
	mov eax,TOS(2)	; c p' c p
.s:
	sub edx,TOS(3)	; disp to 1st =/= dlm
	mov TOS(2),edx	; c p' n1 p
	mov ecx,eax
	pop eax		; u'
	spush
	mov eax,ecx
	call scan	; u" p" n1 p
	mov eax,TOS(1)
	sub eax,TOS(3)	; n2
	mov TOS(1),eax
	inc eax		; n3 n2 n1 p
	ret
;-
;- zblcount	( a1 -- a1 u )			(k)(X)
;-	cnt string length until ctrl code or <bl>
;-	does{ -1 bl parsec }, for {[number]}
zblcount:A;:ok:
	call mone
	call bbl
;-
;- parsec  	( p1 u1 c -- p1 u2 )		(X)				forth
;-	push length of char c or <nul> terminated string,
;-	up to +u chars parsed from address p1, @in not modified
;-	if c=<bl> find any delimiter <bl> or ctrl-code 0..31
parsec:GA;:ok:
; c: ecx,edx
%ifdef __MMX;
	movq mm0,TOS(2)	; mm0 := p u	{ c u p
	spush
	movzx eax,al	; c->s rot dup 2swap
	movq TOS(2),mm0	;		{ c u p p
%else
	mov ecx,TOS(1)
	mov edx,TOS(2)
	spush
	movzx eax,al
	mov TOS(1),ecx
	mov TOS(2),edx
%endif
	mov ah,al
	call scan	;		{ u' p' p
	spop		;		{ p' p
	sub eax,TOS(1)	;		{ u1 p
	ret
;-
;- skip		( p u c -- p' u' )		(83)				forth
;-	skip leading chars = c of length u string at ptr p,
;-	ret posn p' of 1st char =/= c and remaining cnt u'.
;-	scan string length u string at ptr p for 1st occurrence of char =/= c.
;-	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
skip:GA;:ok:
; c: ecx,edx
	mov edx,eax	; char
	spop
	mov ecx,eax	; cnt
	test ecx,ecx
	jz .r
	push edi
	mov edi,TOS(1)
	add edi,KR
	mov eax,edx
	cmp eax,d_bl<<8|d_bl
	jz .b
	repz scasb
	jz .n
	inc ecx
.x:
	dec ecx
	lea eax,[ecx+1]
	lea ecx,[edi-1]
	pop edi
	sub ecx,KR
	mov TOS(1),ecx; 1st =/= char
.r:
	ret
.b:A;	;cmp al,[edi]+
	scasb		; skip any ctrl and <bl>
	jc .x		;?; higher code
	dec ecx		; bl or lower
	jnz .b
.n:
	inc edi
	jr .x		;>; no other chars
;-
;- scan		( p u c -- p' u' )		(83)				forth
;-	scan string length u string at ptr p for 1st occurrence of char = c,
;-	ret posn p' of 1st char = c and remaining cnt u'.
;-	ret u' = 0 and ptr to after stg if dlm not found.
;-	skip leading chars =/= c of length u string at ptr p
;-	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
;-	(e: flags reflect value of u)
scan:GA;:ok:
; c: ecx,edx
	push edx
	mov edx,eax	; char
	spop
	mov ecx,eax	; cnt
	test ecx,ecx
	jz .r
	push edi
	mov edi,TOS(1)
	add edi,KR
	mov eax,edx
	cmp eax,d_bl<<8|d_bl
	jz .b
	repnz scasb
	jz .q
.n:
	inc edi		; zero if char not found
.x:
	dec ecx
.q:
	lea eax,[ecx+1]
	lea ecx,[edi-1]
	pop edi
	sub ecx,KR
	mov TOS(1),ecx	; ptr to 1st =/= char
.r:
	pop edx
	test eax,eax
	ret
.b:A
	scasb		; skip any ctrl and <bl>
	jnc .x		;?; char above <bl>
	dec ecx		; bl or lower code
	jnz .b
	jr .n
;-
;- csearch 	( p1 u1 p2 u2 c -- p3 u3 flg )	(4)				forth
;-	{search} in mem(p1,u1) by char c terminated trailing string(p2,u2).
csearch:GA;:ok:
	push eax	; char >r
	spop
	call swap2	;		{ u1 p1 u2 p2
.l:
	call over2	;		{ u2 p2 u1 p1 u2 p2
	call search
	test eax,eax
	mov ecx,TOS(2)	; ptr
	jz .r		;?; not found	{ f u1' p1' u2 p2
	add ecx,TOS(3)
	movzx ecx,byte[kref ecx]
	cmp ecx,[esp]
	jnz .n		;?; stg doesn't fit
.x:
	mov ecx,TOS(2)
.r:
	mov edx,TOS(1)
	mov TOS(4),ecx	; >r 2sdrop r>
	mov TOS(3),edx
	_add esp,4	; rdrop
	TOSPULL(2)
	ret
.n:A;		; retry next
	inc ecx		; ptr to after beginning
	spop		; drop flag
	dec eax		; reduce count
	mov TOS(1),ecx	; store ptr
	jg .l		;?; continue searching
	_mov eax,0	; flg
	jr .x		; not found
;-
;- search  	( p1 u1 p2 u2 -- p3 u3 flg )	(a:s)				forth
;-	search u1 bytes beginning at p1 for string(p2,u2).
;-	ret p3 where string found and u3 remaining chars cnt w. true flag,
;-	else p3=p1, u3=u1 and false flag.
;-	-re- {csearch}, additionally checking terminating char.
search:GA;:ok:
	xor ecx,ecx
	inc dword TOS(1);	{ u2s p2s' u1m p1m
	test eax,eax
	jg .g		;?;
	dec ecx		; empty stg is 'found'
.r:
	spop
	mov eax,ecx
	ret
.g:A
	cmp TOS(2),eax	;	{ u2s p2s' u1m' p1m
	js .r		;?;not found if too small a memory range
	mov ecx,TOS(2)	; u1
	sub ecx,eax	; u1-u2
	inc ecx		; d(u) for char after end of shorter stg -?-
	pushad
	mov edi,TOS(3)	; edi := p1m
	add edi,KR	; a1m mem beginning
.l:
	mov edx,edi	; a1m'
	readad eax,edi	; PS
	readad ecx,ecx	; d(u)
	mov eax,[eax-4]	; p2s'
	movzx eax,byte[kref eax-1]; 1st char
	repnz scasb
	jz .d
	mov ecx,[byte uref uv.bits]
	and ecx,mcaps
	jz .n
	readad ecx,ecx	; d(u)
	mov edi,edx	; a1'
	add eax,[uref uv.uhi2lo]
	movzx eax,byte[kref eax]; lo.char
	repnz scasb
	jz .d
	readad ecx,ecx	; d(u)
	mov edi,edx	; a1'
	add eax,[uref uv.ulo2hi]
	movzx eax,byte[kref eax]; hi.char
	repnz scasb
	jnz .n		; not found
.d:
	writead ecx,ecx	;-> d(u)'
	readad eax,eax	; u2
	add ecx,eax
	dec ecx
	jz .f		;?; ready and found
	push edi	; a1r memadr
	mov PS,[byte esp s_edi+4]
	call dup2	;	{ u2s p2s' ..
	dec eax		;	{ u2s' p2s' ..
	mov ecx,eax
	spush
	mov eax,[esp]	; a1r
	sub eax,KR	; p1r	{ p1m u1m' p2'
	spush
	mov eax,ecx	;	{ u2' p1' u2' p2' ..
	call compare	;	{ f ..
	test eax,eax
	pop edi		; a1m
	jz .e		;	{ f u1 p1' u2 p2 [ 
	cmp dword[byte esp s_ecx],0
	jnz .l
.n:;	; not found
	popad		;	{ u2s p2s' u1m p1m
	spop
	xor eax,eax	;	{ 0 u2 p2
	ret
.f:A;	; found
.e:
	writead edx,edi	; a3m'
	popad		;	{ u2s p2s' u1m p1m
	add ecx,eax	; u3m'+stg len
	dec edx		; back to 1st char
	spop 3
	sub edx,KR	; p3m stg
	mov eax,edx	; p3
	spush
	mov eax,ecx	; u3
	jmp true
;-
;- s=		( p1 u1 p2 u2 -- n )				'seq'		forth
;-      compare case dependently, regardless of uc-.. setting.
seq:GA;:ok:
	push dword[byte uref uv.bits]
	and dword[byte uref uv.bits],~(mcaps)
	call compare
	pop dword[byte uref uv.bits]
	ret
;-
;- compare	( p1 u1 p2 u2 -- n )		(a:s)				forth
;-	compare strings (p1,u1) and (p2,u2) by subtraction of character code
;-	in stg2 from char. code in stg1 at the rsp. position. assume stgs 
;-	terminated by a <nul> byte, while comparing, for deriving the result.
;-	letter case dependency obeyed as set by lc-depend/ignore.
;-	return n:
;-		1: s1>s2, -1: s1<s2, 0: s1=g2
;-	i.e. difference of 1st different chars determine flag, else length.
compare:GA;:ok:
; c: ecx,edx
	mov ecx,TOS(2)	; u1
	mov edx,ecx	; l1
	sub edx,eax	; l1-l2
	test eax,eax
	jz .l		;?; 2nd empty, s1 > s2
	test ecx,ecx
	jz .l		;?; 1st empty, s2 > s1
; <search) stg-s not empty	{ u2 p2 u1 p1
	push esi,edi,edx
	mov esi,TOS(3)	; s1
	mov edi,TOS(1)	; s2
	add esi,KR	; a1
	add edi,KR	; a2
	cmp ecx,eax	;l1-l2
	jna .c
	mov ecx,eax
.t:
	jecxz .s
.c:
	repz cmpsb
	jz .p		; common part is equal, d/length is result
	movzx eax,byte[edi-1];c2
	movzx edx,byte[esi-1];c1
	test dword[byte uref uv.bits],mcaps
	jz .s
	add edx,[uref uv.uhi2lo]
	movzx edx,byte[kref edx]
	add eax,[uref uv.uhi2lo]
	movzx eax,byte[kref eax]
	cmp edx,eax
	jz .t		;?; more
.s:
	sub edx,eax	; c1-c2
	mov [esp],edx
.p:
	pop esi,edi,edx
.l:
	spop 3
	test edx,edx
	mov eax,edx
	jz .o
	sar eax,31
	lea eax,[eax*2+1]
.o:
	ret		; ret len1-len2
;-	

;-
;- rskip  	( p u c -- p' u' )		(83)				forth
;-	skip trailing chars = c of length u string at ptr p,
;-	ret posn p' of 1st char =/= c and remaining cnt u'.
;-	scan string length u string at ptr p for 1st occurrence of char =/= c.
;-	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
rskip:GA;-:ok:
; c: ecx,edx
	mov edx,eax	; char
	spop
	mov ecx,eax	; cnt
	test ecx,ecx
	jz .r
	push edi
	std
	mov edi,TOS(1)
	add edi,KR
	lea edi,[edi+ecx-1]
	mov eax,edx
	cmp eax,d_bl<<8|d_bl
	mov edx,ecx
	jz .b
	repz scasb
	inc ecx
.x:
	dec ecx
.q:
	sub edx,ecx
	lea eax,[edx-1]
	lea ecx,[edi+1]
	pop edi
	sub ecx,KR
	mov TOS(1),ecx; 1st =/= char
	cld
.r:
	ret
.b:A
	scasb		; skip any ctrl and <bl>
	jc .x		;?; higher code
	dec ecx		; bl or lower
	jnz .b
	dec edi
	jr .q		;>; no other chars
;-
;- rscan  	( p u c -- p' u' )		(83)				forth
;-	reverse scan stg(p,u) for last occurrence of char c,
;-	ret posn p' of 1st found char = c and remaining cnt u'.
;-	ret u' = 0 and ptr to stg p' = p if dlm not found.
;-	skip trailing chars =/= c off stg(p,u).
;-	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
rscan:GA;:ok:
; c: ecx,edx
	push edx
	mov edx,eax	; char
	spop
	mov ecx,eax	; cnt
	test ecx,ecx
	jz .r
	push edi
	std
	mov edi,TOS(1)	; ptr
	add edi,KR
	lea edi,[edi+eax-1]; stg end
	mov eax,edx	; char
	cmp eax,d_bl<<8|d_bl
	mov edx,ecx	; copy count
	jz .b
	repnz scasb
	jz .q
.n:;	: stg not found:
	mov ecx,edx	; zero count
	dec edx
.x:
	dec ecx
.q:
	sub edx,ecx	; remg count
	mov eax,edx
	lea ecx,[edi+1]	;
	sub ecx,KR
	pop edi
	mov TOS(1),ecx	; ptr to 1st =/= char
	cld
.r:
	pop edx
	ret
.b:A
	scasb		; skip any ctrl and <bl>
	jnc .x		;?; char above <bl>
	dec ecx		; bl or lower code
	jnz .b		;?; more
	jr .n
;-
;- /string	( p u n -- p1 u1 )		(a:s)(e)(4)	'sstring'	forth
;-	adjust string at p by +n and len u by -u,
;-	subtracts n from u @tos and adds n to p @nos, unless @tos < 0.
;-	+ve n truncates string(p,u) from beginning towards end, min. u1=0,
;-	-ve n extends string ptr p and length u beyond its beginning.
;-	(e: flags reflect length result)
sstring:GA;:ok:; { over min rot over+ -rot - }
; c: ecx, eflags
	mov ecx,eax	; n
	spop
	cmp ecx,eax
	jle .m		;?; u > n
	mov ecx,eax	; m
.m:
	add TOS(1),ecx	; m u p+m
	sub eax,ecx	; u-m p+m
	ret
;-
;- 1/string	( p u -- p1 u1 )		(e)		's1string'	forth
;-	cut 1st char (byte) off stg(p,u); short for { 1 /string }.
;-	(e: flags reflect length result)
s1string:GA;:ok:<cslice,e\stg,replace,substitute)
; c: eflags
	test eax,eax
	jng .r
	inc dword TOS(1)
	dec eax
.r:
	ret
;-
;- cslice  	( p1 u1 c1 c2 -- p2 u2 p3 u3 )	(4)				forth
;-	stg slice by delimiting chars, return remaining stg(p2,u2) and slice(p3,u3).
;-	neither ret-stg includes the delimiters, remainder begins after end-delimiter,
;-	-ve character code may be passed for the rsp. stg bound, start or end.
;-	adjust slice w/ { -1 /string 1+ } to including the delimiting chars.
cslice:GA;:ok:
; 2over 2swap -3 roll scan 1/string over swap 3 roll scan
; drop over- 2swap 2over + rot over- rot + 1/string 2swap ;
	call oswap2
	spush
	mov eax,-3
	call roll	; hide c2
	test eax,eax
	js .a		;?; slice from beginning
	call scan	; find c1
	call s1string	; next char
	jz .d		;?;c1 not present
.b:; <from start
	call oswap	; over swap
	spush
	_mov eax,3
	call roll	; re-fetch c2
	test eax,eax
	js .e		;?; slice till end
	call scan	; find end dlm
	jz .d		;?;c2 not present
.f:
	spop		; drop rmg count
	sub eax,TOS(1)	; over-
	call sover2	; put slice to lower item
	add eax,TOS(1)
	TOSPULL(1)	; 2swap 2over +
	call rot
	sub eax,TOS(1)	; rot over -
	call rot
	add eax,TOS(1)	; rot +
	TOSPULL(1)
	call s1string	; 1/string 2swap
	jmp swap2
.a:A
	spop
	jr .b
.d:A
	mov ecx,TOS(1)	; rot drop
	TOSPULL(1)
	mov TOS(1),ecx
	ret
.e:A;	; p' u' c2 
	mov ecx,TOS(1)	; drop + 0
	xor eax,eax
	add TOS(2),ecx
	TOSPULL(1)
	jr .f
;-
;- sslice  	( p1 u1 p2 u2 p3 u3 -- p4 u4 p5 u5 )				forth
;-	stg slice(p5,u5) enclosed by stg(p2,u2) and stg(p3,u3), remainder(p5,u5).
;-	sliced stg does not include the delimiters, remainder begins after end-delimiter,
;-	zero length stg-s may be passed for the rsp. stg bound, start or end.
;-	adjust slice w/ { u2 negate /string u3 + } to including the delimiting stg-s.
sslice:GA;:ok:
; 2>r >rr search if r /string endif rdrop
; 2dup 2r> >rr search >r 2dup r> if r /string endif rdrop
; 2-rot drop sdrop over- 
	push dword[byte uref uv.s0]; -?-
	kloci 6
	mov ecx,[esp]
	add ecx,KR
	mov [esp],LP
	mov LP,ecx
.u3:	klocv		; 2nd
.p3:	klocv		;	dlm
.u2:	klocv		; 1st
.p2:	klocv		;	dlm
	call dup2
	call dup2
.u1:	klocv		; src
.p1:	klocv		;	stg
	spush
	mov eax,[lref .p2]
	spush
	mov eax,[lref .u2];	{ u2 p2 u1 p1 u1 p1
	call search	;	{ f u1' p1' u1 p1
	test eax,eax
	jz .z		;?; 1st dlm not found
	mov eax,[lref .u2]
	call sstring
	mov ecx,[lref .u3]
	test ecx,ecx
	jz .e		;?; zero end-dlm
	call dup2	;	{ u1' p1' u1' p1' u1 p1
	spush
	mov eax,[lref .p3]
	spush
	mov eax,[lref .u3]
	call search
	test eax,eax	;	{ f u1" p1" u1' p1' u1 p1
	jz .n		;?; 2nd dlm not found
	mov eax,[lref .u3]
	spop
	mov ecx,TOS(1)
	sub TOS(2),eax	;	{         u5  p5
	mov TOS(4),eax	;	{		  u4
	mov TOS(5),ecx	;	{		     p4
	spop 2
.r:
	pop LP
	ret		; de-alloc locals, ret
;{ u1' p1' u1 p1
.e:A;	; zero 2nd dlm, ret slice till end of stg, zero rmd
	mov eax,TOS(3)
	add eax,TOS(2)
	mov TOS(2),ecx	; u4 := 0
	sub eax,TOS(1)	; u5 := p1 - p1' + u1
	jr .r
;{ f u1' p1' u1 p1
.z:A;	; 1st dlm not found, ret zero slice, rmd := src
	TOSPULL(1)
	jr .r
.n:A;	; 2nd dlm not found, ret zero slice, rmd := src
	TOSPULL(3)
	jr .r
;-
;- bslice  	( p1 u1 n1 n2 -- p4 u4 p5 u5 )					forth
;-	get stg slice(p5,u5) and remainder(p4,u4) by char(byte) positions n1,n2.
;-	sliced stg includes both, the byte positions n1 and n2.
bslice:GA;:ok:
; -	-ve position number may be passed for the rsp. stg bound, start or end.
	sub eax,TOS(1)	; size of slice
	jng .z		;?; ret empty stg
	push eax
	spop
	dec eax
	jns .s
	xor eax,eax
.s:
	call sstring
	call dup2
	spush
	pop eax
	call sstring
	jz near swap2	;?;out of bounds
	call swap2
	mov eax,TOS(3)
	sub eax,TOS(1)
	ret
.z:A
	TOSPULL(2)
	xor eax,eax
	ret
;-
;- substitute	( p1 u1 c1 c2 l1 l2 p2 u2 -- p3 u3 n | abort )(4)(x)		forth
;-	Recursive, multiply nestable string substitution/text translation.
;-	in:	(p1,u1)	source stg
;-		c1, c2	start-/end-dlm
;-		l1, l2	ptr to reference-/substitution-list
;-		(p2,u2)	workspace, for result stg
;-	out:	(p3,u3)	stg result, in buffer(p2,u2)
;-		n	count of substitutions
;-		abort	if resolved stg doesn't fit in {place}, {+place} limits.
;-	Recursively extract stg within innermost delimiters(c1,c2) in stg(p1,u1),
;-	search list(l1) for extracted unique counted or <nul> terminated string 
;-	and substitute with <nul>-enclosed stg at corresponding posn in list(l2).
;-	The empty stg can be simulated w/ "<bl><del>", in substitution list.
;-	store modified string to buf. (p2,u2),
;-	ret counted stg(p3,u3) and flag n := count of substitutions done
;-	or, n=0 and p3:=p1, u3:=u1 if no refs found.
;-	Abort if {place-size} or storage space u2 exceded.
substitute:GA;:ok:
; s0 @ $1000 dup >loc 0. 0. 0. 0
	spush
	mov eax,[byte uref uv.s0]; s0 @		: p4
	spush
	mov eax,0x2000	; max for {place} etc.
	spush		; dup #locals to remove	: 2 * u4
	call toloc
	shr eax,1	; u4
	call dup2
	add TOS(1),eax	; u5 p5
;  locals| flgx flag tpos tlen pos1 cnt1 pos3 cnt3 u5 p5 u4 p4 u2 p2 l2 l1 c2 c1 u1 p1 |
	kloci 20
; -zero self-initiated- ;
.flgx:	klocz		; flag to exit, 0 if no ref substituted
.flag:	klocz		; replacements count
.tpos:	klocz		; start scanning for end-dlm
.tlen:	klocz		; remaining stg length
.pos1:	klocz		; ptr   \ leading
.cnt1:	klocz		; count /	part, before ref
.pos3:	klocz		; ptr	\ trailing
.cnt3:	klocz		; count /	part, after ref
; -generated locals   - ;
.u5:	klocv		; 2ndary buf
.p5:	klocv		;   size & ptr
.u4:	klocv		; primary buf
.p4:	klocv		;   size & ptr
; -supplied parameters- ;
.u2:	klocv		; result stg
.p2:	klocv		;   size & ptr
.l2:	klocv		; substitution stgs list
.l1:	klocv		; reference stgs list
.c2:	klocv		; trailing dlm
.c1:	klocv		; leading dlm
.u1:	klocv		; input stg
.p1:	klocv		;   size & ptr
;  p1 u1 2dup p2 place p4 place		\ copy to stg(p3,u4) for case no dlm found
	spush 
	mov eax,[lref .p1]
	spush 
	mov eax,[lref .u1]
	call dup2
	spush 
	mov eax,[lref .p2]
	call place
	spush 
	mov eax,[lref .p4]
	call place
;  p4 count to tlen to tpos		\ start ptrs
.loop:;					\ entry to replacement pass
	spush 
	mov eax,[lref .p4]
	call ccount
	mov [lref .tlen],eax
	spop
	mov [lref .tpos],eax
	spop
	jr .entry
.enter:
	spush
;  begin
.begin:
	spop 2
.entry:
;   tpos tlen c2 scan 	 		\ ready if no end-dlm found
	spush 
	mov eax,[lref .tpos]
	spush 
	mov eax,[lref .tlen]
	spush 
	mov eax,[lref .c2]
	call scan
; dup if				\ (F1)
	jz near .endif1	;?; ready
;    2dup 1 /string to cnt3 to pos3	\ trailing stg
	call dup2
	call s1string
	mov [lref .cnt3],eax 
	spop
	mov [lref .pos3],eax 
	spop
;    sdrop tpos tlen rot - c1 rscan
	TOSPULL(1)
	mov ecx,eax
	mov eax,[lref .tpos]
	spush
	mov eax,[lref .tlen]
	sub eax,ecx
	spush 
	mov eax,[lref .c1]
	call rscan
;    cnt3 to tlen pos3 to tpos		\ remainder to scan
	mov ecx,[lref .pos3]
	mov edx,[lref .cnt3]
	mov [lref .tpos],ecx
	mov [lref .tlen],edx
; dup if				\ (F2) continue w/ rmd if no start-dlm found
	test eax,eax
	jz disp .begin
;     over p4 1+ swap over- to cnt1 to pos1 \ leading stg
	mov edx,[lref .p4]
	mov ecx,TOS(1)	; over
	inc edx		; p4 1+
	sub ecx,edx	; swap over-
	mov [dword lref .pos1],edx; (alignment, > 20% of runtime!)
	mov [lref .cnt1],ecx
;     1 /string l1 enum# 1+
	call s1string
	spush 
	mov eax,[lref .l1]
	call enumn
	inc eax
; dup 0> if				\ (E3)
	jng disp .enter
;      1- l2 enum@
	dec eax
	spush 
	mov eax,[lref .l2]
	call enumf
; dup 0> if				\ (E4)
	test eax,eax
	jng disp .begin
; handle the empty string:
	mov ecx,TOS(1)
	cmp dword[kref ecx-1],(d_del<<8|d_bl)<<8; <nul><bl><del><nul>
	jz disp .begin
;       pos1 cnt1 p2 place 		\ assemble intermediate result
	spush 
	mov eax,[lref .pos1]
	spush 
	mov eax,[lref .cnt1]
	spush 
	mov eax,[lref .p5]
	call place
;       p2 +place pos3 cnt3 p2 +place
	spush 
	mov eax,[lref .p5]
	call pplace
	spush 
	mov eax,[lref .pos3]
	spush 
	mov eax,[lref .cnt3]
	spush 
	mov eax,[lref .p5]
	call pplace
;       p2 count 2dup p4 place 		\ copy new stg to workspace
	spush 
	movd mm0,[lref .p5]
	movd mm1,[lref .p4]
	mov eax,[lref .p5]
	call ccount
;       dup tlen - p4 + 1+ to tpos	\ adjust next scan posn
	movd [lref .p4],mm0
	lea ecx,[eax+1]
	sub ecx,[lref .tlen]
	add ecx,[lref .p4]
	movd [lref .p5],mm1
	mov [lref .tpos],ecx
;       true to flag	=> 1 +to flag	1 +to flgx
	inc dword[lref .flgx]
	inc dword[lref .flag]
	test eax,eax
	jnz disp .begin
;      else 0= endif			\ (N4) no subst. stg, try next
;     else  1 endif			\ (N3) ref not found, try next
;    endif				\ (N2) no start-dlm, ready
.endif2:
;   endif				\ (N1) no end-dlm, ready
.endif1:
;  sdrop 0= until
	spop 2
;  p2 count flag			\ ret stg(p3,u3)
	test dword[lref .flgx],-1
	mov dword[lref .flgx],0
	jnz disp .loop	;?; until no subst done
	spush 
	mov eax,[lref .p4]
	call ccount
	kloc@(ecx,.u2)
; check buffer bounds
	cmp eax,ecx
	jnc near m.oa
	spush 
	mov eax,[lref .p2]
	call place
	spush 
	mov eax,[lref .p2]
	call ccount
	spush 
	mov eax,[lref .flag]
;  u4 negate >loc ;
	spush 
	mov eax,[lref .u4]
	add eax,[lref .u5]
	neg eax
	call toloc	; remove local buffer
	ret		; and local vari, revert to plain <ret> code
;-
;- replace	( p1 u1 p2 u2 p3 u3 n p4 u4 -- p5 u5 f)				forth
;-	replace one text in a string by another text.
;-	letter case dependency as previously set, temporarily being forced to 
;-	case-dependency if source & replacement text would otherwise be equal.
;-	
;-	in:	(p1,u1)	source string;
;-		(p2,u2)	text to replace by..
;-		(p3,u3)	  ..replacement text;
;-		n	operational mode flag (re below);
;-		(p4,u4)	ptr p4 to buffer of size u4, stg and buffer space may be identical.
;-	out:	(p5,u5)	stg result - NOTE: not a "counted" stg!
;-			convert replaced to counted stg e.g. { drop over place };
;-		f	result flag (re below).
;-	
;-	n >  0	replace n-th occurrence of text(p2,u2) in stg(p1,u1) by text(p3,u3).
;-	n =  0	count text(p2,u2) occurrences, w/o any actual replacement.
;-	n = -ve replace any occurrence of text(p2,u2).
;-	f := n	ret flg & counted stg(p5,u5), stored to buf(p4,u4). p5:=p4 and
;-		 u5:=u4,  f:=-1 if initially unsufficient buffer size;
;-		 u5:=len, f:=-ve no. of replacements done, if next wouldn't fit;
;-		 u5:=0,   f:= 0 if stg(p1,u1) is empty (u1=0);
;-		 u5:=u1,  f:= 0 if stg(p2,u2) not found in stg(p1,u1);
;-		 u5:=len, f:= n if replacement done.
replace:GA;:ok:
	test eax,eax	; u4		( p1 u1 p2 u2 p3 u3 n p4 u4 
	mov ecx,TOS(7)	; u1
	jng .o		; no/false buffer args
	test ecx,ecx
	js .o		;?; no stg/erroneous data
	jz .q		;?; empty stg
; sing. subst. new stg len = u1+u3-u2 = leading:u1-u1' + subst:u3 + trailing:(u1'-u2) = u1-u1'+u3+u1'-u2
; -> test u4-u1 >/ u3-u2
	mov edx,TOS(3)
	sub edx,TOS(5)	; u3-u2
	add ecx,edx
	cmp eax,ecx
	jc .o		;?; unsufficient buffer space
; new stg len = leading:u1-u1' + subst:u3 + trailing:(u1'-u2) = u1-u1'+u3+u1'-u2 = u1+u3-u2
	cmp dword TOS(2),byte 0
	js .m		;?; loop through multiple repl's
	jnz near .c	;?; replace n-th occurrence ( p1 u1 p2 u2 p3 u3 n p4 u4 
	spush 
	mov eax,TOS(9)	; p1	( p1 u1 p2 u2 p3 u3 n p4 u4 p1
	spush
	mov eax,TOS(3)	; p4	( p1 u1 p2 u2 p3 u3 n p4 u4 p1 p4
	spush
	mov eax,TOS(10)	; u1	( p1 u1 p2 u2 p3 u3 n p4 u4 p1 p4 u1
	call move	; store stg to destn
; - just count - 	;	( p1 u1 p2 u2 p3 u3 n p4 u4
	mov ecx,TOS(1)	; p4
	TOSPULL(3)
	mov TOS(1),ecx	;	{ u4 p4 u2 p2 u1 p1
	mov eax,TOS(4)
	call srot2	;	{ u2 p2 u1 p1 u1 p5
	push_ibyte(0)
	jr .pn		; count refs
; empty input stg, before locals def'd
.q:;	( p1 u1 p2 u2 p3 u3 n p4 u4 -- p4 0 0 )
	xor eax,eax
	call .o
	inc eax		; Z
	ret
.pl:A;<loop entry) counting ref occurrences 
	spop
	call s1string
.pn:; <count refs)
	call over2	;	{ u2 p2 u1 p1 u2 p2 u1 p5
	call search
	sub [esp],eax	; inc count
	jc .pl		;	{ u1' p1' u2 p2 u1 p5
	TOSPULL(4)
	pop eax		;	{ n u1 p5
	ret
; short buffer, before locals def'd
.o:;	( p1 u1 p2 u2 p3 u3 n p4 u4 -- p4 u4 -1 )
	mov ecx,TOS(1)	; p4
	mov TOS(7),eax	; sto u4
	mov TOS(8),ecx	; sto p4
	TOSPULL(6)
	mov eax,-1	; ( .. -- p4 u4 -1 )
	ret
;( p1 u1 p2 u2 p3 u3 n p4 u4 -- p5 u5 f
;   8  7  6  5  4  3 2  1  eax
.m:A; <multiple replacements)
	mov dword TOS(2),1; will be replacing all 1st refs
	spush		; ( p1 u1 p2 u2 p3 u3 1 p4 u4 @bits 
	call .p
	mov eax,[byte uref uv.bits]; caps-flag
	jnz .i		;?; not equal
	mov ecx,eax
	and ecx,mcaps	; caps flag
	jz near .nc	;?; stgs exactly equal, ret unmodified stg and flg:=0
	xor dword[byte uref uv.bits],ecx; set to case dependency
.i:
	push eax	; @bits
	push_ibyte(0)	; repl. count
; ( p1 u1 p2 u2 p3 u3 n p4 u4 xx
	jr .mn
; - p1 u1 p2 u2 p3 u3  n p4 u4   p5 u5 f
; - 11 10 9  8  7  6  5  4  3    2  1  eax
.ml:A; -- ; make result the new source, rst buffer size
	mov ecx,TOS(2)
	mov edx,TOS(1)
	mov TOS(11),ecx
	mov TOS(10),edx	; p5,u5 to source
	TOSPULL(2)
.mn:;( p1 u1 p2 u2 p3 u3 n p4 u4 xx -- p1 u1 p2 u2 p3 u3 n p4 u4 p5 u5 f
    ;   8  7  6  5  4  3  2  1 eax
	mov eax,9	;	drop 9 ndup
	call ndup	;		( p1 u1 p2 u2 p3 u3 n p4 u4 p1 u1 p2 u2 p3 u3 n p4 u4 -- ..
	call .c		; single pass	( .. -- p1 u1 p2 u2 p3 u3 n p4 u4 p5 u5 f
	add [esp],eax; rpl. count
	test eax,eax
	jnz .ml		;?; more
; -- ;; - p1 u1 p2 u2 p3 u3  n p4 u4   p5 u5 f
	mov ecx,TOS(2)
	mov edx,TOS(1)
	mov TOS(11),ecx
	mov TOS(10),edx	; p5,u5 to source
	pop eax		; flag
	pop dword[byte uref uv.bits]; rst 'caps' state
	TOSPULL(9)	; ( .. --  p5 u5 f
	ret
;( p1 u1 p2 u2 p3 u3 n p4 u4
;   8  7  6  5  4  3 2  1  eax
.c:A; single replacement:
;kommt ohne localen speicher aus:
;stg zum ziel copieren, dort suchen, haengenden teil an entspr. stelle copieren, zuletzt ersatztext.
	mov edx,TOS(3)
	sub edx,TOS(5)	; u3-u2
	add ecx,edx
	cmp eax,ecx
	jc disp .o		;?; unsufficient buffer space
	call rot	; ( p1 u1 p2 u2 p3 u3 p4 u4 n
	push eax
	push eax	; n								[: n n
	mov eax,3
	call pick2	; p1,u1			( p1 u1 p2 u2 p3 u3 p4 u4 p1 u1
	mov ecx,eax	; u1
	mov edx,eax
	mov eax,TOS(3)	; p4			( p1 u1 p2 u2 p3 u3 p4 u4 p1 p4
	sub edx,TOS(6)
	add edx,TOS(4)
	mov TOS(9),eax	; p5:=p4
	mov TOS(2),edx	; u5:=u1-u2+u3, for the rsult stg
	mov TOS(8),ecx	; save u1
	spush
	mov eax,ecx	;			( p5 u1 p2 u2 p3 u3 p4 u5 p1 p4 u1
	call move	; src stg to buffer	( p5 u1 p2 u2 p3 u3 p4 u5		[: n n
	call over4	;			( p5 u1 p2 u2 p3 u3 p4 u5 p5 u1 p2 u2
	jr .cn
.cl:A;			;			( p5 u1 p2 u2 p3 u3 p4 u5 p5' u1'
	call s1string	; advance ptr to next char or count := 0
	spush
	mov eax,3
	call pick2
.cn:;			;			( p5 u1 p2 u2 p3 u3 p4 u5 p5 u1 p2 u2
	call search	;			( p5 u1 p2 u2 p3 u3 p4 u5 p5' u1' f
	test eax,eax
	jz .cx
	add [esp],eax	; dec src counter						[: n' n
.cx:
	spop
	jnz .cl		;?; next occurence
	cmp dword[esp],byte 0
	lea esp,[esp+4]	;								[: n
	jnz .nf		;?; requested occurrence not found, ret unmodified stg(p4,u1)
; whether subst. fitting into buf was already tested on entry at <.c>
; leading part already copied; 		( p5 u1 p2 u2 p3 u3 p4 u5 p5' u1'
	mov ecx,TOS(1)	; p5'
	mov edx,TOS(6)	; u2
	mov TOS(6),ecx	; p5', save for subst dst
	add ecx,TOS(4)	; p5'+u3, trailer dst
	mov TOS(7),ecx	; save
	spush
	mov eax,edx	; u2		( p5 u1 [p5'+u3] p5' p3 u3 p4 u5 p5' u1' u2
; determine trailing part
	call sstring	; trailer	( p5 u1 [p5'+u3] p5' p3 u3 p4 u5 [p5'+u2] [u1'-u2]
	mov ecx,TOS(7)	; p5'+u3 trailer posn
	spush
	mov TOS(1),ecx	;
; copy trailer len (u1'-u2) from (p5'+u2) to (p5'+u3)
	call move	; sto trailer	( p5 u1 [p5'+u3] p5' p3 u3 p4 u5
; setup result args
	mov ecx,TOS(4)	; p5'
	mov edx,TOS(3)	; p3
	mov TOS(6),eax	; u5, sto final length
	mov TOS(3),ecx	; p5'
	mov TOS(4),edx	; p3
	pop dword TOS(5); flg:=n							[: -/-
; copy substitute stg into gap
	spop 2		;		( p5 u5 flg p3 p5' u3
	jmp xcmove	; substitute	( .. -- p5 u5 flg )
; ret unmodified and flg := 0 if ref-stg not found
.nf:A;	( p5 u1 p2 u2 p3 u3 p4 u5 p5' u1'
	lea esp,[esp+4]	; drop subst. no.						[: -/-
	xor eax,eax
	TOSPULL(7)	; ( .. -- p5 u1 0 )
	ret
; ret unmodified & flg := n if ref and subs stgs equal
.nc:A;	( p1 u1 p2 u2 p3 u3 n p4 u4 @bits -- p5 u5 n
	mov eax,TOS(3)	; n
	mov edx,TOS(2)	; p4
	mov ecx,TOS(9)	; p1
	mov TOS(7),eax	; flg:=n
	mov eax,TOS(8)	; u5:=u1
	mov TOS(9),edx	; p5:=p4
	mov TOS(5),edx	; p4
	mov TOS(6),ecx	; p1
	TOSPULL(4)	; ( 9:p5 8:u1 7:flg 6:p1 5:p4 4:xx 3:xx 2:xx 1:xx 0:u1 -- 8:p5 7:u1 6:flg 5:p1 4:p4 3:u1
	jmp move	; ( .. -- p5 u5 flg
; multiple replacements, test whether ref and subs case-(in)dependently equal
.p:A; ( p1 u1 p2 u2 p3 u3 n p4 u4 @bits -- .., no change )
	call .pp	; p2,u2
	call .pp	; p3,u3
	call compare	; -- flg
	test eax,eax
	spop		; ( -- )
	ret
.pp:A
	spush
	mov eax,3
	jmp pick2
;-
;- c>lower	( c1 -- c2 )					'c2lower'	forth
;-	convert char c1 to lower case character c2, char = byte, -re- {s>lower}.
c2lower:GA;:ok:
	movzx eax,al
	add eax,[uref uv.uhi2lo]
	movzx eax,byte[kref eax]
	ret
;-
;- c>upper	( c1 -- c2 )					'c2upper'	forth
;-	convert char c1 to upper case character c2, char = byte, -re- {s>upper}.
c2upper:GA;:ok:
	movzx eax,al
	add eax,[uref uv.ulo2hi]
	movzx eax,byte[kref eax]
	ret
;-
;- s>lower	( p1 u1 p2 u2 -- p3 u3 )			's2lower'	forth
;-	convert stg(p1,u1) to all lower case characters (per byte),
;-	buf-size expected to provide for chars and terminating <nul>.
;-	using conversion table { (hi>lo) @ }, no conversion if zero entry.
;-	ret stg(p3,u3) in buf(p2,u2). u3 = 0 if u2 not > u1.
;-	if (p1,u1) writable, source and buf may be same area w/ p2:=p1, u2:=u1+1.
;-	char conversion tables by dft from iso-8559-1 charset data arrangement, i.e.
;-	upper case ranges [65,90] and [192,222], lower case in [97,122] and [224,254],
s2lower:GA;:ok:
	mov edx,[uref uv.uhi2lo]; conversion table
	jr s2upper.n
;-
;- s>upper	( p1 u1 p2 u2 -- p3 u3 )			's2upper'	forth
;-	convert stg(p1,u1) to all upper case characters (per byte),
;-	buf-size expected to provide for chars and terminating <nul>.
;-	using conversion table { (lo>hi) @ }, no conversion if zero entry.
;-	ret <nul>-terminated, un-counted stg(p3,u3) in buf(p2,u2). 
;-	u3 = 0 if u2 < u1.
;-	if (p1,u1) writable, source and buf may be at same address.
;-	char conversion tables by dft from iso-8559-1 charset data arrangement, i.e.
;-	upper case ranges [65,90] and [192,222], lower case in [97,122] and [224,254],
s2upper:GA;:ok:
	mov edx,[uref uv.ulo2hi]; conversion table
.n:
	mov ecx,TOS(2)
	cmp eax,ecx
	mov eax,0
	jle .r		;?; ret empty if unsufficient size of buf
	test ecx,ecx
	jng .r		;?; empty src or -ve count
	test edx,edx
	jnz .p		;?; w/ conversion
	mov edx,tcharchar-Y
.p:
	push edi,esi,ecx; save PS, LP, u1
	xor eax,eax	; clr workspace
	mov esi,TOS(3)	; stg
	mov edi,TOS(1)	; buf
	add esi,KR
	add edi,KR
	add edx,KR
.l:
	lodsb
	mov al,[edx+eax]
	stosb
	dec ecx
	jnz .l
	mov ecx,edi
	pop edi,esi,eax
	mov byte[ecx],0	; <nul>-terminated
.r:;	; unsufficient workspace, ret(p2,0)
	jmp sdrop2
;-
;- printing?	( p u -- f )					'printingq'	forth
;-	tf if stg(p,u) contains no other than printing chars, incl. <nl> or <tab>.
;-	u=0 returns true flg (nothing to print is considerd always 'printable').
;-	using -re- {ekey>char} per stg bytes.
printingq:GA
	mov ecx,eax
	spop
	lea eax,[kref eax]
.l:
	dec ecx
	js .r		;?; all done
	push ecx	; cnt
	mov ecx,eax
	inc eax
	spush
	movzx eax,byte[ecx]
	call ekey2char
	spop 2
	pop ecx
	jnz .l		;?; more
	cmp byte[eax-1],d_ccr; <nl> is 'printable'
	jz .l		;?; more
.q:
	xor ecx,ecx
.r:
	mov eax,ecx
	ret
;-
;- e\stg  	( p1 u1 p2 -- p3 u3 )				'estg'		forth
;-	convert stg(p,u) w/ escaped tokens to stg(p3,u3) at buffer (p2).
;-	converted stg is never longer than source, thus buffer expected 
;-	at least u1 chars of size, and not checked. p2 not in stg memory!
;-	the escape symbol is re-definable in uvari {(escc)}, dft "\", 
;-	other chars may follow which receive special treatment:
;-	numeric:
;-		\nnn	character code, octal numeric sequence as valid for a forth 
;-			single integer, any other radix w/ the appropriate praefix.
;-			-re- {[number]}, for instance,
;-				s" hick\46"	displays	hick!
;-				s" hick\#33"	does the same, w/ decimal code
;-	special symbols, letter case dependent:
;-		\a	<bel>
;-		\b	<bs>
;-		\c	w/o trailing <lf> - suppressed, not applicable
;-		\e	<esc>
;-		\f	<ff>
;-		\i	<csi>
;-		\n	<new line>
;-		\r	<cr>
;-		\t	<ht>
;-		\v	<vt>
;-		\\	double escape symbol, sending the escape symbol as a char.
estg:GA;:ok:
	mov dword[kref eax],0
	spush
	call swap2		; init p3	( p2 p3' p1 u1
	push dword[byte uref uv.base]	; save base
	push dword[byte uref uv.bits]	; i/o flags
	mov dword[byte uref uv.base],8	; default to octal base
	and dword[byte uref uv.bits],~(mfloat|mqranu|mranu|mquad|mbign); clr numeric input settings
	jr .l
.e:A;<[etype], no escape seq. found)
	spop 2
.q:A;<[etype], false escape seq at stg end)	; ( p2 p3' p' u'
	pop dword[byte uref uv.bits]
	pop dword[byte uref uv.base]
	mov ecx,TOS(2)		; p2'
	add TOS(2),eax
	spush
	mov TOS(1),ecx		;		( p2 p3' p' p2 u'
	call move		;		( .. -- p2 p3'
	sub eax,TOS(1)		;		( p3 u3
	ret
.c:; <emit escaped code)
	not ecx
	mov eax,[kref ecx*4+((etypet.e-etypet)*4)+etypet.t-Y]
.n:; ;<after \smb)
	mov ecx,TOS(3)		;		( p2 p3" p' u' c
	inc dword TOS(3)
	mov [kref ecx],al
	spop
	call s1string		; advance over char
	jz .q			;?; stg ends
.l:; <loop entry)		;		( p2 p3' p u -- ..
	mov ecx,TOS(1)
	spush 3
	mov TOS(2),ecx
	mov eax,[uref uv.uescc]	; get escape char, dft "\"
	call scan		;		( p2 p3' p u p' u'
	mov edx,TOS(1)		; p'
	mov ecx,TOS(3)		; p
	jz .e			;?; no escaped chars, copy till end
	mov TOS(2),eax		; u:=u' rmg len
	mov TOS(3),edx		; p:=p' ptr
	mov eax,edx
	sub eax,ecx		; u" = p'-p, till escape char
	mov edx,TOS(4)		; p3
	mov TOS(1),ecx		; p" = p
	add TOS(4),eax		; next p3'
	spush
	mov TOS(1),edx		; destn p3'
	call move		;		( p2 p3' p' u' p" p3'u" -- p2 p3" p' u'
	call s1string		; 1/string, discard escape smb
	mov ecx,TOS(1)
	jz .q			;?; stg ends
	spush
	movzx eax,byte[kref ecx]; next char
	cmp eax,[uref uv.uescc]	; @(escc)
	jz .n			;?; inst the escape symbol
; test/get escaped special symbol;		( p2 p3' p' u' nl nh f=u"
	mov edx,edi
	lea edi,[kref etypet-Y]
	mov ecx,etypet.e-etypet
	repnz scasb
	mov edi,edx		; PS
	jz disp estg.c		;?; inst symbol	( p2 p3' p1 u1 p1 u1
	mov eax,TOS(2)
	spush
	mov eax,TOS(2)
	call abnumberb		; [number]	( p2 p3' p' u' nl nh f=u"
; flag u" = u' if no numeric chars converted
	cmp eax,TOS(3)		;		( p2 p3' p' u' nl nh f=u"
	mov ecx,TOS(2)		; nl
	mov edx,TOS(5)		; p3
	jz .u			;?; non-numeric, discard num & flg
	inc dword TOS(5)	; p3'
	mov [kref edx],cl
	mov ecx,TOS(3)		; u', before "\"
	mov edx,TOS(4)		; p'
	sub ecx,eax		; ud:=u'-um 
	add edx,ecx		; pm:=p'+ud
; test whether bl after num which would have been discarded by [number]
	cmp byte[kref edx-1],d_bl
	jnz .p			;?; not <bl> after \num
	dec edx
	inc eax			; include blank into rmg stg
.p:
	mov TOS(4),edx
.u:
	TOSPULL(3)		;		( p2 p3' pm um
	jmp disp estg.l
;-	
;-
;- ----------------------------------------------------------------
;--- {blanks}, {cmove}, {cmove>}, {erase}, {fill}, {move}, {place}, {spaces}
;-	character is size and cnt is number of 8bit bytes.

;-
;- place-size	( -- n )			(c),m,6)	'placesize'	forth
;-	biggest string size which {place}, {+place} can generate; cpl'd dft 4K.
placesize:GA M;:ok:
	spush
	mov eax,pthfsize
M placesize
	ret
;-
;- place  	( p1 n p2 -- )			(a:?)				forth
;-	copy stg(p1,n) to byte-counted stg at p2. n limited to -re- {place-size}.
;-	store <nul>+asciz stg if count > 255, i.e. leading and trailing <nul>.
;-	buffer at p2 unchanged if n < 1.
place:GA;:ok:
; o: edx := p, ecx := u
	push eax	; save ptr to countbyte in case of overlap/asciz
	mov ecx,TOS(1)	; u
	cmp ecx,pthfsize
	jnc near r.oa	;?; too many 
	push ecx
	inc eax
	mov TOS(1),eax	; 1st char
	mov eax,[esp]	; count
	call move
	pop ecx		; count
	pop edx		; buf
	test dword[byte uref uv.state],masciz
	jnz .a		;?; forced asciz
	cmp ecx,256
	jc .s		;?; counted stg
.a:
	lea ecx,[ecx+edx+1]
	mov byte[kref ecx],0; terminate
	xor ecx,ecx
.s:
	mov [kref edx],cl
	ret
;
pplace.m:A;:ok:	; prepend to counted/asciz stg
	neg edx		; real len
	mov TOS(1),edx	; patched
	call ccount	;	{ n2 p2+ n1 p1
	mov ecx,TOS(2)	; n1
	add ecx,eax	; sum(n1,n2)
	cmp ecx,pthfsize
	ja pplace.r	;?; not within max path len
	spush
	test ch,ch
	mov ecx,TOS(2)	; p2+
	push eax	; len				[ n2
	push ecx	; dst				[ p2+ n2
	push eax	; flg/count			[ n2 p2+ n2
	jz .a		;?; counted stg
	xor [esp],eax	; asciz mode
.a:
	add ecx,TOS(3)	; p2+n1 new posn of stg2
	mov TOS(1),ecx	;	{ n2 p2e p2+ n1 p1	[ n2 p2+
	inc eax		; include terminating <nul>
	call move	;	{ n1 p1
	add [esp+8],eax	;				[ n2+n1 p2+
	spush
	mov ecx,[esp+4]	; p2+
	mov TOS(1),ecx	;	{ n1 p2+ p1
	call move
	pop ecx		; n21
	test dword[byte uref uv.state],masciz
	jnz .z		;?; forced asciz
	test ecx,ecx
	jnz .l		;?; type of counted stg
.z:
	mov dword[esp+4],ecx; asciz dummy 'count'
.l:
	pop edx
	pop ecx
	lea edx,[edx-1]
	mov [kref edx],cl; sto countbyte
	ret
;
pplace.e:A;:ok:;<append: len > MAX_PATH)
	lea esp,[esp+4]	; drop ptr to chars
	mov [kref eax],cl; rst countbyte
	spush 2
pplace.r:;:ok:;<prepend: len > MAX_PATH)
	TOSPULL(4+1)
m.oa:
	spush
r.oa:
	mov eax,23	; 'off working area'
	jmp errorf
;-
;- +place 	( p1 n1 p2 -- | abort )		(f6)		'pplace'	forth
;-	+n1: append stg(p1,n1) to byte-counted stg at p2 or,
;-	-n1: prepend byte-counted stg at p2 by stg(p1,-n1).
;-	counted stg if len < 256, else stored w. leading <nul> byte, for asciz.
;-	forcedly storing a <nul>-enclosed string if state-flag <asciz> is set.
;-	in either case, dstn stg will be <nul>-terminated.
;-	abort if accumulated length > -re- {place-size}, old len recovered in +n case. 
;-	( for safety, stg result length limited to max(PAGE_SIZE,PATH_MAX), which can be
;-	( changed by re-compiling w/ a modified value of <pthfsize> -re- "constants.inc".
pplace:GA;:ok:
; c: ecx,edx
; o: edx := p, ecx := u
	mov edx,TOS(1)
	test edx,edx
	js near .m	;?; prepend
	jz near drop3	;?; empty stg
	push eax	;@tor := ptr to dst stg
	call ccount	; fetch len & ptr to chars, whether counted or asciz stg
	mov ecx,eax	;ecx := present cnt
	spop
	lea edx,[edx+ecx];edx := accumulated len
	_xchg eax,[esp]	;@tos := ptr to stg, @tor := ptr to chars
	cmp edx,pthfsize
	ja .e		;?; abort if len > max
	test dword[byte uref uv.state],masciz
	push edx	; acc. length
	jnz .a		;?; forced asciz
	test dh,dh
	mov byte[kref eax],dl;new dst countbyte
	jz .c		;?; < 256
.a:
	mov byte[kref eax],0; asciz
.c:
	pop edx		; acc. len
	mov eax,[esp]	; ptr to 1st dst char
	add edx,eax	; end of stg
	mov [esp],edx	; save for <zero> termination
	mov edx,TOS(1)	; n1
	add eax,ecx	; end of present stg
	mov TOS(1),eax	; p2+n1+countbyte
	mov eax,edx
	call move
	pop edx
	mov byte[kref edx],0; zero terminate (:ok:)
	ret
;-
;- move		( p1 p2 +n -- )							forth
;-	non-overwriting copy +n bytes from p1 to p2, noop if n < 0.
move:GA;:ok:
; c: ecx,edx
; o: edx := p2, ecx := n
	mov ecx,TOS(1)	; dst a2
	cmp ecx,TOS(2)	; src a1
	jc xcmove	;?; if a1 < a2 move bottom up
;-
;- cmove>  	( p1 p2 +n -- )					'cmovegt'	forth
;-	copy +n bytes from p1 to p2, beginning at high address
;-	no action if n = zero or -ve; data copying by bytes.
cmovegt:GA;:ok:
; c: ecx,edx
; o: edx := p2, ecx := n
	mov ecx,eax	; cnt
	mov edx,TOS(1)	; destn
	spop 2		; eax:=src
	test ecx,ecx
	jle .r
	push esi,edi,edx,ecx; save LS,PS,ptr,cnt
%xdefine aregl regl
%xdefine bregl regl
	add eax,KR
	add edx,KR
	lea esi,[eax+ecx-1]; src
	lea edi,[edx+ecx-1]; dst
; -	suid programs, only! %define __lockthread in "constants.inc" to activate.
%ifdef __lockthread
	lea eax,[kref srmovsb-Y]
	push eax
	call rlock	; exec routine @tor w. locked int's
%else
	std
	rep movsb
	cld
%endif
	pop aregl
.r:
	spop
	ret
;-
;- cmove  	( p1 p2 +n -- )			(fg)		'xcmove'	forth
;-	copy +n bytes from p1 to p2, beginning at low address.
;-	faster than {cmove>}, copying by 32bit items if appropriate.
xcmove:GA;:ok:
; c: ecx,edx
; o: edx := p2, ecx := n
	mov ecx,eax	; cnt
	mov edx,TOS(1)	; destn
	spop 2		; eax:=src
	test ecx,ecx
	jle .r
	push esi,edi,edx,ecx; save LS,PS,ptr,cnt
%xdefine aregl regl
%xdefine bregl regl
	lea esi,[kref eax]; src
	lea edi,[kref edx]; dst
; -	suid programs, only! %define __lockthread in "constants.inc" to activate.
%ifdef __lockthread
    %if 1
	lea eax,[kref mtunlock-Y]
	push eax
	call mtlock
    %else
	lea eax,[kref smovsb-Y]
	push eax
	call rlock	; wait until memory opr unlocked, lock scheduler
	pop bregl
.r:
	spop
	ret
    %endif
%else
    %if __OPTIMIZE__=__O_SPEED__
	cmp ecx,12	; dword copying always, and up to 4 times faster!
	mov edx,ecx
	jc .1		;?; 1..11 bytes
	mov ecx,esi	; src
	and ecx,3
	jz .4		;?; aligned
	sub edx,ecx	; sub unaligned bytes
	rep movsb	; mov bytes till ptr aligned
.4:
	mov ecx,edx
	and edx,3
	shr ecx,2	; count by 4-bytes
	push edx	; rmg unaligned byte-count
	test esi,7
	jz .8		;?; 8-aligned
	movsd
	dec ecx		; min 4 bytes left
.8:
	rep movsd
.b:
	pop ecx
	jecxz .p
	rep movsb
.p:
	pop aregl
.r:
	spop
	ret
.1:A
	rep movsb
	jr .p
    %else; __OPTIMIZE__=__O_SIZE__
	rep movsb
	pop aregl
.r:
	spop
	ret
    %endif
%endif
;-
;- erase  	( p +u -- )							forth
;-	store +u <zero> bytes to memory beginning at ptr p
erase:GA;:ok:
; c: ecx,edx
	spush
	xor eax,eax	; char
	jr fill
;-
;- blanks  	( p +u -- )			(fg)				fig
;- blank  	( p +u -- )			(fg)(a:stg)	'blanks'	forth
;-	store +u <bl> characters (8bit-bytes) to memory beginning at ptr p.
blanks:GA;:ok:
; c: ecx,edx
	call bbl
;-
;- fill		( p +u c -- )			(fg)				forth
;-	store +u (8bit-)bytes c to memory beginning at ptr p
;-	optimizing to 32bit items if count and destn at cell bounds.
fill:GA;:ok:
; c: ecx,edx
	movzx edx,al	; char
	mov ecx,TOS(1)	; cnt
	spop 2		; eax=destn
.e:
	test ecx,ecx
	jle .r
	push edi
	lea edi,[kref eax]
	mov eax,edx	; char
; -	suid programs, only! %define __lockthread in "constants.inc" to activate.
%ifdef __lockthread
    %if 1
	lea eax,[kref mtunlock-Y]
	push eax
	call mtlock
    %else
	lea edx,[kref sstosb-Y]
	push edx
	call rlock	; store bytes
    %endif
%endif
    %if __OPTIMIZE__=__O_SPEED__
	mov edx,ecx
	and ecx,3
	jz .t
	rep stosb
.t:
	shr edx,2
	jz .p		;?; ready
	mov ecx,edx
	mov ah,al
	shl eax,8
	jz .i
	mov al,ah
	shl eax,8
	mov al,ah
.i:
	test edi,7
	jz .s
	stosd
	dec ecx
	jz .p		;?; ready
.s:
	rep stosd
    %else
	rep stosb
    %endif
.p:
	pop edi
.r:
	spop
	ret
;-
;- >move< 	( p1 p2 +n -- )					'movex'		forth
;-	copy n>0 bytes from p1 in reversed order to p2, w/ unrestricted range p1,p2.
movex:GA;:ok:
	test eax,eax
	jng near drop3	;?; nothing to copy
	spush
	shr eax,2
	add eax,byte 2
	push eax	; locals
	push dword[byte uref uv.s0]
	call toloc
	pop edx		; @s0
	push edi,esi
	mov esi,TOS(2)	; p1
; align such that w/ swapped dwords src of final move will be aligned
	add edx,byte 4	; safety margin for aligned transfer
	lea ecx,[eax+3]	; count
	shr ecx,2	; copying by dwords
	mov TOS(2),edx	; p1 <- @s0 src for back-transfer
; provide for unaligned end of stg
	add edx,byte 3	;dst end
	and edx,byte -4
	add edx,KR	; dst
	add esi,KR	; src beginning, true address
	lea edi,[eax+edx];dst end
.l:;A
	mov edx,[esi]
	lea edi,[edi-4]
	bswap edx
	mov [edi],edx
	lea esi,[esi+4]
	dec ecx
	jns .l
	pop edi,esi
	call xcmove	; copy reversed stg to p2+
	spush
	pop eax
	neg eax
	jmp toloc
;-
;-	
;---------------------------------------------------------------------------------
;- 

;
; <nxpfile>	( p1 u1 -- p2 u2 pf uf )	(k)(4)
;	next filename in a PATH stg
nxpfile:A
	inc dword TOS(1)	; 1 -1 d+	{ u- p+
	dec eax
	jng .z
	call dup2		; 2dup		{ u- p+ u- p+
	spush
	mov eax,0+':'		; ":
	call scan		; scan		{ u' p2 u- p+
; ( 1 2 3 4 -- 3 4 1 2-4 ) rot over- -rot 2swap
	push eax		;4
	mov edx,TOS(1)		;3
	neg eax
	add eax,TOS(2)		;2-4
	mov ecx,TOS(3)		;1
	pop dword TOS(2)	;4		{ uf pf u2 p2
	mov TOS(1),ecx		;1
	mov TOS(3),edx		;3
	ret
.z:
	xor eax,eax
	spush 2
	ret
; remove local dataspace after file-names which refer to cwd
ppfrloc:A
	spush
	pop eax			; no. of allocated local cells
	jmp toloc
;-
;- path+file	( pe ue pf uf pb ub -- pp up | xx 0 )		'pthfile'	linux
;-	find file w. name (pf,uf) in PATH supplied by string(pe,ue).
;-	search cwd, only if file-name(pf,uf) begins w/ "./".
;-	search HOME-dir, only, if file-name(pf,uf) begins w/ "~/".
;-	put found absolute path and name stg(pp,up) into buffer(pb,ub),
;-	ret empty stg, up = 0, if no such file found.
;-	stg result length limited to size of Linux PATH_MAX (4K).
;-	-re- {f"}, {fname"}, {[fname]} which search {f8dir} and {path}.
;-	examples:
;-	  s" /devfs:/dev" s" tty" pad 200 path+file
;-		path to current kbd console, whether "devfs" active or not.
;-	  path s" util.f8" pad 400 path+file
;-		path to 'nearest' f8-utils files (or directory).
;-	NOTE:
;-	  path not taken from env-vari, for security reasons but,
;-	  HOME env-vari read if passed filename begins with '~/'.
pthfile:GA;:ok:
	call dup2
	call erase		; safe asciz
	sub eax,TOS(2)		; buf. less filename
	sub eax,byte 2		;   less count.w
	jng near r.oa		;?; 'off working area'
	call swap2		;	     tos{ uf pf ub pb ue pe
	mov ecx,TOS(1)
	mov ecx,[kref ecx]
	cmp cl,'/'
	jz near .c		;?; absolute path
	cmp cx,'~/'
	jz near .h		;?; home path
	cmp cx,'./'
	jnz .w			;?; not the cwd
	sub eax,byte 2
	add dword TOS(1),byte 2
	call rotd2		; discard supplied path
	spush
	mov eax,[byte uref uv.s0]
	lea ecx,[kref ppfrloc-Y]; prepare removing local space
	call pathmax		; safety limit
	spush
	shr eax,2
	inc eax
	push eax
	push ecx
	call toloc		; >loc, path-max+4 local mem
	call cwdf		; cwd@
	call mrot2		; put cwd in place of supplied path
.w:
	call tor2		; 2>r		{ ub pb ue pe	[ uf pf
	call swap2		; path & buf	{ ue pe ub pb	[ uf pf
	dec dword TOS(1)	; prepare
	inc eax			;  directory slices
	spush 2
.l:
	spop 2
	call nxpfile		;		{ ud pd ue'pe'ub pb
	xor edx,edx		; prepare zero count return
	test eax,eax
	jng .n		;?; end of path	{ ue'pe'ub pb
	cmp eax,TOS(4)		; buf-size, less file-name count
	jnc near r.oa		;?; buf. space ovf
	spush
	mov eax,TOS(6)		; pb		{ pb ud pd ue'pe'ub pb
	call place		;		{ ue'pe'ub pb
	call r2			; uf pf		{ uf pf ue'pe'ub pb [uf pf
;		-; append "/" to  path
	spush
	mov eax,TOS(6)		; pb		{ pb uf pf ue'pe'ub pb
	call ccount
	mov ecx,TOS(1)
	lea ecx,[kref ecx]
	mov byte[ecx+eax],'/'
	cmp byte[ecx-1],0
	jz .z			;?; asciz
	inc byte[ecx-1]
.z:
	spop
	mov eax,TOS(6)		; buf
	call pplace		; append file	{ ue' pe' ub pb
	spush
	mov eax,TOS(4)		; pb
	call .f
	jnz .l			; 		{ ur pr ue' pe' ub pb
	mov edx,eax
	call mrot2
.n:;;< not found	( ... -- ptr 0 )
	lea esp,[esp+8]	; 2rdrop
	spop 4
	mov eax,edx
	ret
.d:; { pb ub pf uf ph uh; HOME env-vari unset
	call mrot2	; revert to input args arrangement
; ( pe ue pb ub pf uf -- pp up | xx 0 )	;< cwd
.c:A;<abs path		;	     tos{ pb ub pf uf
	call rotd2	; drop supplied path
	call rot	;		{ pb pf uf ub
	cmp TOS(1),eax
	jnc .b		;?; buf. space ovf
	mov eax,TOS(3)	
	call place	;		{ pb
	jr .x
; ( pe ue pb ub pf uf -- pp up | xx 0 )	;< HOME path
;   pth   buf   fil
.h:A;<$home by ~/...	;		{ pb ub pf uf
	call rotd2	; drop pth(pe,ue)
	dec eax
	inc dword TOS(1); advance past leading "~"
.e:
	call ehome	; s" HOME" getenv { pb ub pf uf ph uh
	test eax,eax
	jz .d		;?; HOME not set
	spush
	mov eax,TOS(6)	; buf		{ pb ub pf uf ph uh pb
	call place	; path at buf	{ pb ub pf uf
	spush
	mov eax,TOS(4)	; pb		{ pb ub pf uf pb
	call ccount	; (if u'>255)	{ pb ub pf uf b' u'
	mov ecx,TOS(5)	;c: pb
	mov edx,TOS(4)	;d: ub
	add eax,TOS(2)	;2: uf
	cmp eax,edx	; accum. len
.b:;;<ovf w/ cwd)
	jnc near r.oa	;?; buf. space ovf
	mov eax,ecx
	TOSPULL(1)	;		{ pb ub pf uf pb
	call pplace	; 		{ pb ub
	spop
.x:;;<ret from cwd)
	call .f
	jz .r
	xor eax,eax
.r:
	ret
.f:A
	call ccount
	call paccessf	; p,u -> p,u,f
	test eax,eax
	spop
	ret
;-
;-	
;- -- strings
;-

;-
;- stg		ccc( p u n -- )(X: -- p u )	(4)		'stg4'		forth
;-	define a string value w/ at most n characters of storage space.
stg4:GA;:ok:
	mov ecx,eax
	cmp ecx,TOS(1)
	jc near r.oa	; 2dup > abort" stg storage space?"
	call create
	call rrbrac	; cpl persistently
	spush
	call comma	; dup ,
	mov ecx,dataptr
	add eax,byte 2
	push ecx
	call allot	; here swap 2+ allot place
	spush
	pop eax
	call place
	call llbrac
	call pscodep	; does>
dostg4:
	add eax,byte 4	; cell+
	jmp ccount	; count ;S
;
stgs.c:
	mov eax,stgs.d-Y
	jmp cplicall
;-
;- to-stg	ccc( p1 u1 -- )			(i)		'stgs'		forth
;-	store stg(p1,u1) into string variable ccc.
stgs:GA
	call .f			; find ccc
	jnz .c			;?; cpl
.d:
	mov dword[kref eax+4],0	; clr stg
	call .s			; get+check stg param
	jns near place		; sto stg
.r:
	jmp near r.oa		;?; 'off working area'
.f:A
	call blwf
	spop
	jz near r.ni		;?; ccc not found
	call tobody
	call ccplq
	jz .x
	spush
	jmp literal
.s:A
	push dword[kref eax]	; limit
	lea eax,[eax+4]
	push eax		; ptr to stg
	call ccount
	mov edx,TOS(1)
	add edx,eax		; used space
	add edx,TOS(2)		; +stg(u1)
	TOSPULL(1)
	pop eax
	pop ecx			; ( .. -- p1 u1 p2' ) ecx:=u.max, edx=u.used
	sub edx,eax
	sub ecx,edx		; ecx:=rmg space
.x:
	ret
stgps.c:
	mov eax,stgps.d-Y
	jmp cplicall
;-
;- +to-stg  	ccc( p1 u1 -- )			(i)		'stgps'		forth
;-	append stg(p1,u1) to string variable ccc.
stgps:GA
	call stgs.f		; find ccc
	jnz .c			;?; cpl
.d:
	call stgs.s		; get+check stg param
	jns near pplace		; append
	jr stgs.r		; abort
;-
;- :slist 	ccc( -- )(X: -- p )		(4)		'colslist'	forth
;-	initiate a reference/substitution list for {enum@}, {enum#}, {substitute}.
;-	constitute a list of stg-s w/ -re- {slist"}, {slist}, {slist,}, {slist;}.
colslist:GA;:ok:
;create here 0 , here 1+ swap ! does> @ ;
	call create
	call rrbrac
	mov ecx,dataptr
	push ecx
	call zcomma
	mov ecx,dataptr
	inc ecx
	pop edx
	mov [kref edx],ecx
	call llbrac
	call pscodep
docolslist:
	mov eax,[kref eax]
	ret
;-
;- slist" 	ccc( -- )			(4)		'slistqte'	forth
;-	fetch by char {"} enclosed stg ccc and cpl to data-space for ref/subst-list
;-	overwrites {pad2} if executed in interpreting mode. -re- {:slist}.
slistqte:GA;:ok:
;[_] s" slist, ;
	call sqte
	jr slistc
;-
;- slist  	c ccc( -- )			(4)		'slists'	forth
;-	fetch by char c enclosed stg ccc and cpl to data-space for ref/subst-list.
;-	overwrites {pad2} if executed in interpreting mode. -re- {:slist}.
slists:GA;:ok:
;[_] string slist, ;
	call stng	;:continues: in <slistc>
;-
;- slist,  	(C: p u -- )			(4)		'slistc'	forth
;-	cpl <nul>-"counted" stg w/o trailing <nul>, add stg(p,u) to ref/subst 
;-	list. overwrites {pad2} if executed in interpreting mode. -re- {:slist}.
slistc:GA;:ok:
;[_] szlit, -1 allot ;
	call szlitc
	call mone
	jmp allot
;-
;- slist; 	( -- )				(4)		'slistsemi'	forth
;-	terminate addressable list of stg-s. -re- {:slist}.
slistsemi:GA;:ok:
	spush
	_mov eax,0
	jmp wcomma
;-
;- szlit, 	( p u -- )					'szlitc'	forth
;-	allot u+2 bytes and store <nul> delimited string(p,u) to here.
szlitc:GA;:ok:
	call szliteral
	or dword[byte uref uv.state],masciz
	jr slitc.c
;-
;- slit,  	( p u -- )					'slitc'		forth
;-	allot u+2 bytes and store counted string(p,u) to here.
;-	always <nul> terminated, whether counted or asciz stg.
slitc:GA;:ok:
	call sliteral
.c:
	call ccplq
	jnz .r		;?; {sliteral} if compiling
	spush
	push dword dataptr; {here}
	add eax,byte 2
	call allot
	spush
	mov eax,[esp]	; destn
	call place
	spush
	pop eax
	call ccount	; length and stg beginning, counted or asciz
	add eax,TOS(1)
	mov byte[kref eax],0
.x:
	spop 2
.r:
	and dword[byte uref uv.state],~(masciz)
	ret
; interpreting
sliteral.i:;:ok:
	push eax	; count
	call pad2	;	{ p2 c h
	push eax
	call place
	spush
	pop eax
	pop ecx
	add ecx,KR
	mov byte[ecx+eax+1],0
	and dword[byte uref uv.state],~(masciz)
	jmp ccount
; sliteral asciz-stg
sliteral.a:
	push ecx
	call sliteral.z
	spush
	mov eax,[esp]	; ptr to stg
	call ccount
	pop edx		; stg
	mov ecx,TOS(1)	; 1st char
	mov byte[kref edx],0; zero count
	lea ecx,[eax+ecx+1]; stg end
	mov word[kref ecx-1],x_ret<<8; <nul><ret>
	mov codeptr,ecx	; advance to after stg
	spop 2
	and dword[byte uref uv.state],~(masciz)
	ret
;-
;- sliteral	(C: p u -- )(X: -- a u)		(i)(a:s)			forth
;-	compile a string passed into a colon definition
sliteral:GA;:ok:
	movzx edx,word[kref Lpsqtep-Y+whead.i-whead.l]; (s")
	push edx
.n:; ;<szliteral,dotquote.s)
	call ccplq
	pop edx		; runtime entry
	jz .i		;?; interpreting
	call rrbrac
	spush
	mov eax,edx	; runtime
	call compilec
	spush		; dup stg count
	mov ecx,codeptr
	test dword[byte uref uv.state],masciz
	jnz .a
.z:
	push ecx
	inc eax
	call callot
	call llbrac
	spush
	mov ecx,codeptr
	mov byte[kref ecx-1],x_ret
	pop eax		; beginning of stg storage
.p:
	jmp place	; store stg to codespace

;-
;- szliteral	(C: p u -- )(X: -- a u)		(i)(a:s)			forth
;-	compile an asciz unlimited length string into a colon definition
szliteral:GA;:ok:
	movzx edx,word[kref Lpsqtep-Y+whead.i-whead.l]; (z")
	or dword[byte uref uv.state],masciz
	push edx
	jr sliteral.n
;-
;- ."		ccc( -- )			(i)(4)		'dotquote'	forth
;-	display counted string literal
;-	compiling:	stores 4th-string to {c-here}
;-	execution:	displays 4th-string from [IP]+
;-	interpreting:	copies 4th-string to {pad2}+
dotquote:GA;:ok:;<abort")
	movzx edx,word[kref Lpdotqtep-Y+whead.i-whead.l]
	call .n
	call ccplq
	jz near type
	ret
.n:; ;<zqte,sqte,dotquote,cqte) 
	push_ibyte('"')
.i:
	spush
	mov eax,[esp]	; char
	mov [esp],edx	; ptr to runtime code
.s:; ;<stng)
	call XWORD	; word		{ here    
	call ccount	; count		{ cnt stg
	jmp sliteral.n
;-
;- s"		ccc( -- p u )			(i)		'sqte'		forth
;-	counted string literal, p := {pad2} when interpreting
sqte:GA;:ok:
	movzx edx,word[kref Lpsqtep-Y+whead.i-whead.l]
	jr dotquote.n
;-
;- string	c ccc( -- p u )			(i)		'stng'		forth
;-	by char c delimited counted string ccc literal, p:={pad2} when interpreting.
;-	use{ string <char> <text><char> } e.g. { string ! hallo! type } displays <hallo>.
stng:GA;:ok:
	call BLWORD
	movzx eax,byte[kref eax+1]
.n:; ;<name)
	movzx edx,word[kref Lpsqtep-Y+whead.i-whead.l]
	push edx
	jr dotquote.s
;-
;- z"		ccc( -- p u )			(i)		'zqte'		forth
;-	unlimited length asciz string literal, p := {pad2} when interpreting
;-	stg length actually limited by {word} which truncates after 255 bytes,
;-	use {szliteral} to compiling longer asciz string literals:
;-	{ [ stg1 count pad place stg2 count pad +place pad count ] szliteral ... }
zqte:GA;:ok:
	movzx edx,word[kref Lpsqtep-Y+whead.i-whead.l]; (z")
	or dword[byte uref uv.state],masciz
	jr dotquote.n
;-
;- name		ccc( -- p u )			(i)				hidden
;-	fetch a <bl>-enclosed name from input stream to {here}
;-	cpl only: immediately overwritten while in interpreting mode
name:GA;:ok:
	call bbl
	jr stng.n
;-
;- "		ccc( -- p )			(i)		'qte'		forth
;-	counted string literal, p := {pad2} when interpreting; synonym {c"}
qte:GA;:ok:; == cqte
;-
;- c"		ccc( -- p )			(i)(a:ce)	'qte'		forth
;-	counted string literal, p := {pad2} when interpreting
cqte:GA;:ok:
	movzx edx,word[kref Lpqtep-Y+whead.i-whead.l]; (")
	call dotquote.n
	call ccplq
	jz .p
	ret
.p:
	spop 2
	jmp pad2
%endif;__STRING_I
;-
;-=====================================================================;
;- l4string.inc <eof>
