;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4time.inc
;-  created	: 14-jul-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __TIME_I
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: (me, hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
%ifndef __TIME_I
%define __TIME_I
%define timeinc

;-	
;-
;-================================================================
;- -- system time
;-

;-
;-	date & time
;-

;-
;- time		( -- ts tm| er -1 )		(s)		'time4'		forth
;-	ret ts := time in seconds, tm := remaining microsec (on tos)
;-	or ernum and -1. syscall gettimeofday, re man(2), linux/time.h
time4:GA C;:ok:
	spush
	mov ecx,PS	; timeval
	xor eax,eax
	push ebx
	spush 2
	mov ebx,ecx
	spush 2
	lea ecx,[ecx+8]	; dummy timezone
	sys_gettimeofday
	lea PS,[ebx+8]
	mov ecx,eax
	pop ebx
	spop		; drop ernum
	test ecx,ecx
	jns .r
	mov TOS(1),ecx
	mov eax,ecx
	sar eax,31
.r:
C time4
	ret
;-
;- date		( -- n )			(s)		'date4'		forth
;-	ret n := time in seconds sice 1.1.1970 or, error
;-	syscall time, re man(2), linux/time.h
date4:GA C;:ok:
	spush
	push ebx
	sys_time NULL	; time_t = long
	pop ebx
	test eax,eax
C date4
	ret
;-
;- day		( -- n )					'day4'		forth
;-	no. of day per week, 0 for Sunday.
day4:GA;:ok:
	call ndated
	cmp eax,byte -1
	jz .r
	xor edx,edx
	mov eax,TOS(1)
	lea ecx,[edx+7]
	lea eax,[eax+4]	; 1.1.1970 was thursday = 4
	div ecx
	mov eax,edx
.r:
	TOSPULL(4)
	ret
;-
;- time&date	( -- n1...n6 )			(4)(a:yx)	'timedate'	forth
;-	ret time & date n1..n6, second, minute, hour, day, month, year.
;-	NOTE:
;-	POSIX.1 defines seconds since the Epoch as a value to be interpreted 
;-	as the number of seconds between a specified time and the Epoch, 
;-	according to a formula for conversion from UTC equivalent, leap 
;-	seconds ignoring and considering all years divisible by 4 being leap years. 
timedate:GA;:ok:
	call ndated
	cmp eax,byte -1
	jnz ndatey
	spush		; 6th cell
	ret
; resolve date to seconds/min, min/hour, hours/day, days
ndated:A
	call date4
	jns totida.p
	cmp eax,EFAULT
	jz .r
	cmp eax,byte -1
	jnz totida.p
.r:
	spush 4
	ret
;-
;- >time&date	( n1 -- n2..n7 )		(4)		'totida'	forth
;-	convert n1 seconds since "the Epoch" to time&date format
;-	n7:year n6:month n5:day n4:hour n3:minute n2:second
totida:GA;:ok:
	lea ecx,[kref ndatey-Y]
	push ecx
.p:A
	spush
	xor eax,eax	; time 0
	spush
	mov eax,60	;	{ #s dd
	push eax
	call umby	; sec	{ dd1 s 
	spush
	pop eax		;	{ #h d1 s
	call umby	; min	{ d2 m s
	spush
	mov eax,24	;	{ #d d2 m s
	jmp umby	;>;hours{ d3 h m s
;-
;- <ndatey>	( s m h d3 -- s m h d m y	(k)				novoc
;-	resolve date from seconds/min, min/hour, hours/day, days
;-	to further returning month & year, incl. leap years wrt 1970
ndatey:A
	call dup2	; ddays	{ d3 d3 d m s
	sub dword TOS(1),730
	sbb eax,byte 0	; 1968 war schaltjahr
	spush
	mov eax,1461	; 4y+1d	{ #l d3 d3 d m s
	call umby	;	{ dl l d3 d m s
	call rot	;	{ dl d3 d m s
	push eax	; leap days  >r
	spop
	call dsub	;	{ d5 d m s
	spush
	mov eax,365	;	{ #y d5 d m s 
	call umby	; years	{ d6 y d m s
	add dword TOS(1),1970
	mov eax,TOS(2)	;days
	mov ecx,-1
	lea edx,[kref days-Y-2]
.l:
	lea edx,[edx+2]
	dec ecx		; ~(month)
	sub ax,[edx]
	jg .l
	add ax,[edx]
	not ecx
	xor dword[esp],-1
	lea esp,[esp+4]	; r>
	jnz .d
; leap year handling, incomplete w.r.t 
; centuries, though POSIX-ly correct...
	cmp ecx,byte 2
	jc .d		; jan
	lea eax,[eax+1]	; d:=d+1
	cmp eax,[edx+daysl-days]
	jna .d
	lea eax,[eax-2]	; d:=d-1
	inc ecx		; m:=m+1
.d:
	mov TOS(2),eax	; d
	mov eax,TOS(1)	; y
	mov TOS(1),ecx	; m
	ret
;-
;- "day		( [cc] -- p u )			(4)		'sday4'		forth
;-	ret stg(p,u) w/ name-stg of system time day
;-	use <en> or <de> decoding if [cc] present and = "en" or "de"
sday4:GA;:ok:
	call day4
	mov ecx,TOS(1)
	cmp ecx,0+'de'
	jz .s
	cmp ecx,0+'en'
	jnz .d
.s:
	call swap
.d:
	jmp daystg
;-
;- "time  	( -- p u )			(4)		'stime4'	forth
;-	return system time stg(p,u)
;-	stg in {hld}-buffer below {pad2}, using {here} for intermediate storage.
stime4:GA;:ok:;<stimedate)
	call timedate	; second, minute, hour, day, month, year }@tos
	spop 3		; drop y,m,d ( -- s m h )
	_mov ecx,0+':'
.n:; ;<sdate4)
	push dword[byte uref uv.base]
	push ecx
	mov dword[byte uref uv.base],10
	call ltsharp
	call rot
	spush
	xor eax,eax
	call sharp	; min 2 digits
	call sharps
	mov eax,[esp]
	call hold
	call rots
	call sharp	; min 2 digits
	call sharps
	pop eax
	call hold
	call sharp	; min 2 digits
	call sharps
	call sharpgt
	pop dword[byte uref uv.base]
	ret
;-
;- "date  	( [cc] -- p u )			(4)		'sdate4'	forth
;-	return system date stg(p,u)
;-	stg in {hld}-buffer below {pad2}, using {here} for intermediate storage.
;-	use <en> or <de> decoding if [cc] present and = "en" or "de",
;-	dft & "en": year-month-day, "de": day-month-year.
sdate4:GA;:ok:;<stimedate)
	cmp eax,0+'de'
	push eax
	jz .p
	cmp eax,0+'en'
	jz .p
	spush
.p:
	call timedate	; second, minute, hour, day, month, year }@tos
	pop ecx
	cmp ecx,0+'de'
	jnz .d
	call srot
.d:
	_mov ecx,'-'
	call stime4.n
	call mrot2
	spop 4
	ret
;-
;- "time&date	( [cc] -- p u )			(4)		'stimedate'	forth
;-	full size system date & time stg, in stg buffer at {pad2}.
stimedate:GA;:ok:
;"day type space "date type space "time type ;
	spush
	mov ecx,[byte uref uv.s0]
	push ecx	; stg ptr
	_mov eax,64
	call toloc
	push eax	; save "en, "de
	call sday4
	call .p
	mov ecx,[esp]
	cmp ecx,0+'de'
	jz .d
	cmp ecx,0+'en'
	jnz .c
.d:
	spush
	mov eax,ecx
.c:
	call sdate4
	call .p
	call stime4
	call .p
	spush
	pop eax		; drop flag
	pop eax		; local stg
	call ccount
	call dup2
	call pad2
	sub eax,byte 64	; above 'here', in <hld> buffer
	mov TOS(4),eax
	inc dword TOS(4)
	call place
	spush
	_mov eax,-64
	jmp toloc
.p:A; ptr to local stg 2nd on tor
	spush
	mov eax,[esp+8]
	call pplace
	mov ecx,[esp+8]
	add ecx,KR
	inc byte[ecx]
	movzx edx,byte[ecx]
	mov byte[ecx+edx],d_bl
	ret
;-
;- .time&date	( [cc] -- )			(4)		'dotimedate'	forth
;-	display full size system date & time stg
;-	using {hld} buffer below {pad2} and {here} for intermediate storage.
dotimedate:GA
	call stimedate
	jmp type
;-
%endif
;-
;-=====================================================================;
;- l4time.inc <eof>
