;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4uvari.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __UVARI_I
;-  description	: user-variables, dto access, kernel vari & "value"s
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __UVARI_I
%define __UVARI_I
%define uvariinc
;-
;-================================================================
;--- user vari/cons
;-

;-
;- up+		( -n -- p )			(m,7)		'upplus'	forth
;-	get ptr to user vari by disp
;-	(consecutive disp's wrt uv.link := 0, .aux:=124, .vocs:=-128)
upplus:GA M;:ok:
; c: -/-
	lea eax,[uref eax+(uv.link)]
	sub eax,KR
M upplus
	ret
;-
;- uv@		( -n -- p )			(m,7)		'uvfetch'	forth
;-	get content of user vari by -ve(!) disp, -n < 0
;-	(consecutive disp's wrt uv.link := 0, .aux:=124, .vocs:=-128)
uvfetch:GA M
; c: -/-
	mov eax,[uref eax+(uv.link)]
M uvfetch
	ret
;-
;- bl		( -- c )			(m,6)		'bbl'		forth
;-	dft. character code of blank space
bbl:GA M;:ok:
; c: -/-
	spush
	mov eax,d_bl
M bbl
	ret
;-
;- dpl@		( -- n ) 			(f6)(m,4)	'dplfetch'	forth
;-	does{ dpl @ }, fetch decimal point location after numeric input.
dplfetch:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.dpl]
M dplfetch
	ret
;-
;- edpl@	( -- n ) 			(m,5)		'edplfetch'	forth
;-	signed, (optional) base-exponent from numeric input.
edplfetch:GA M;:ok:
; c: -/-
	spush
	movsx eax,byte[byte uref uv.bits+edpl]
M edplfetch
	ret
;-
;- in@		( -- n ) 			(f6)(m,4)	'infetch'	forth
;-	does{ in @ }
infetch:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.in]
M infetch
	ret
;-
;- blk@		( -- n ) 			(f6)(m,4)	'blkfetch'	forth
;-	does{ blk @ }
blkfetch:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.blk]
M blkfetch
	ret
;-
;- span@  	( -- n )			(m,4)		'spanf'		forth
;-	number of chars fetched with latest expect, -re- {#tib}
spanf:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.span]
M spanf
	ret
;-
;- #tib		( -- u )			(83)(m,7)	'ntib'		forth
;-	received no. of bytes in tib storage - f83 eq. to {span}
ntib:GA M;:ok:
; c: -/-
	spush
	mov eax,[uref uv.uintib]
M ntib
	ret
;-
;- tib		( -- p ) 			(m,4)				forth
;-	push ptr to terminal input buf.
;-	ANS-4th standard: tib is tib and is tIB wich is an Input Buffer, 
;-	thus any question whether it might be written to is just stupid!
tib:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.tib]
M tib
	ret
;-
;- tib-max	( -- u )			(f6)(m,7)	'tibmax'	hidden
;-	max bytes of tib storage
tibmax:GA M;:ok:
	spush
	mov eax,[uref uv.utibmax]
M tibmax
	ret
;-
;- tib-dft	( -- p )			(m,8)		'tibdft'	hidden
;-	default terminal input buffer for stdin input. below user area.
tibdft:GA M;:ok:
	spush
	lea eax,[uref uv.utibot]
	sub eax,KR
M tibdft
	ret
;-
;- tib>dft	( -- )				(m,10)		'tib2dft'	root
;-	set terminal input buffer to default location -re- {tib-dft}.
tib2dft:GA M;:ok:
	lea ecx,[uref uv.utibot]
	sub ecx,KR
	mov [byte uref uv.tib],ecx
M tib2dft
	ret
;-
;- pad-size	( -- n )			(m,4)		'pad-size'	forth
;-	size of {pad}, {pad2}, implied size of numeric stg conversion buf. at {hld}.
;-	should not be set to less than 136 bytes, i.e. for quads' number conversion.
;-	system default 256 bytes, 4096 (Page Size) recommended for long strings, etc.
padsize:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.padsz]
M padsize
	ret
;-
;- pad		( -- p )			(a:c)				forth
;-	double cell aligned auxilary memory {pad-size} bytes on top of current 
;-	dictionary ptr, "above"  {pad2}, stored data may extend until below {sp@}.
pad:GA;:ok:
; c: -/-
	call pad2
.n:
	add eax,[byte uref uv.padsz]
	jmp faligned
;-
;- pad2		( -- ptr )							forth
;-	2ndary double-cell aligned auxilary memory {pad-size} bytes on top of current 
;-	dictionary ptr used by literal string storage opr's while interpreting, {s"} etc.
pad2:GA;:ok:
; c: -/-
	call here
	jr pad.n
;-
;- cc4cc@  	( c -- p )			(83)		'cc4thf'	forth
;-	user vector, zero ctrl action for char c in jumptable
cc4thf:GA;:ok: -?-
	mov ecx,eax
	mov eax,[byte uref uv.icc4th]
	test eax,eax
	jz .r			;?; no table
	add eax,KR		; ptr -> addr
	jz .r			;?; no table
	mov eax,[eax+4*ecx]
	test eax,eax
	jz .r			;?; no entry
	add eax,KR
	jz .r
	sub eax,KR		; ret ptr
.r:
	ret
;-
;- cc4cc!  	( c p -- )					'cc4ths'	forth
;-	store execution ptr p for char c to (cc4th) table.
;-	2drop if @(cc4th) = 0; p = 0 or kref(p) = 0 de-acivates entry
cc4ths:GA;:ok: -?-
; c: ecx,edx
	mov ecx,eax
	mov edx,[byte uref uv.icc4th]; ptr to table
	add edx,KR
	spop			; get char code
	jz .r
	test ecx,ecx		; execution ptr
	jz .z			;?; de-activate
	add ecx,KR
.z:
	mov [edx+4*eax],ecx	; x-ptr to table at char offset
.r:
	spop			; drop char
	ret
;-
;- here		( -- p )			(m,4)				forth
;-	ptr to bot of unused dataspace
;-	current dataspace dictionary ptr, between @{d0} and @{d-top}.
here:GA M;:ok:
; c: -/-
	spush
	mov eax,dataptr
M here
	ret
;-
;- unused  	( -- n )			(a:cx)				forth
;-	available data-space cells,
;-	{ d-top @ here - 4/ }, may be extended w. {allocate}
unused:GA;:ok:
; c: -/-
	call dtop
	mov eax,[kref eax]
	sub eax,dataptr
	sar eax,2
	ret
;-
;- c-here  	( -- p ) 			(m,4)		'chere'		forth
;-	current executeable code dictionary ptr, between @{c0} and @{c-top}.
chere:GA M;:ok:
; c: -/-
	spush
	mov eax,codeptr
M chere
	ret
;-
;- state@  	( -- n ) 			(f6)(m,4)	'statef'	forth
;-	fetch compile/interpret state flag(s)
;-	zero if actively interpreting, any non-zero otherwise
statef:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.state]
M statef
	ret
;-
;- (r)@		( -- p ) 			(f6)(m,7)	'prpf'		forth
;-	ptr to fill-up char for formatted string output
prpf:GA M;:ok:
; c: -/-
	spush
	mov eax,[uref uv.uprp]
M prpf
	ret
;-
;- source-id	( -- n ) 			(a:cx)(m,7)	'sourceid'	forth
;-	current input source
;-		-2: initial arguments present
;-		-1: input by {evaluate}d string
;-		 0: user input device = <stdin>, file if -ve blk@
;-		>0: channel number of file input, 0 and -ve @blk@ for stdin
;-		    file pointer in channels chp data
sourceid:GA M;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.sourceid]
M sourceid
	ret
;-
;- 4th-id  	( -- n ) 			(f6)(m,7)	'fthid'		forth
;-	basic task's process id, set on startup.
;-	re {pid@} for currently running job's id
fthid:GA M;:ok:
; c: -/-
	spush
	mov eax,[uref uv.upid]
M fthid
	ret
;-
;- ti-cal  	( -- dn )			(f6)(m,7)	'tical'		forth
;-	fetch cpu clocks per second, from (tick).
;-	non-macro variant checks and inititates calibration, if required
tical:GA M;:ok:
; c: ecx (initial call)
	cmp dword[uref uv.utick],0	; fetch timing value
	jnz .t
	call setcal
.t:
M,tical
; c: -/-
	spush
	mov eax,[uref uv.utick]		; fetch timing value
M tical
	ret
;-
;- cmc-off	( -- )				(i)		'cmcoff'	root 
;-	disable "optimization", i.e. end-of-word <call> to <jmp> conversion 
;-	until next new entry into compile state. 
;-	use { ... [ cmc off ] ... } to disabling for next word, only, and 
;-	to re-enabling common optimization features after disabling. 
cmcoff:GA;:ok:
; c: -/-, @tos=eax preserved
	mov dword[uref uv.ucmc],0xffff0000	; words counter
	ret
;-
;- cmc?		( -- )				(i)		'cmcq'		compiler
;-	immediately(!) sedecimal display {cmc} value.
cmcq:GA;:ok:
	spush
	mov eax,[uref uv.ucmc]
	jmp hexdot
;-
;- !fence 	( -- | abort )					'stofence'	forth
;-	store current top values to {fence},{c-fence},{i-fence} variables
;-	backing up code-, compile- and auxilary compile-ptrs to c-here,
;-	thus restricted to be used while (non-intermediately!) executing.
;-	NOTE: {i-fence} stores 1st un-used index, following last unsed one.
stofence:GA;:ok:
; c: ecx,edx
	cmp dword[byte uref uv.state],0
	jnz near rxo
	call hpts
	mov edx,dataptr
	mov [byte uref uv.cfence],ecx
	mov [byte uref uv.fence],edx
	mov ecx,[byte uref uv.lkindex]
	mov [byte uref uv.ifence],ecx
	ret
;-
;- fence! 	( n1 n2 n3 -- )					'fences'	forth
;-	store values from tos, n1:=c-fence, n2:=fence, n3=i-fence to {..fence} ptrs
;-	NOTE: n3, @i-fence, is 1st un-fenced index, following fenced one.
fences:GA;:ok:
; c: ecx,edx
	mov ecx,TOS(1)
	mov edx,TOS(2)
	mov [byte uref uv.ifence],eax
	mov [byte uref uv.fence],ecx
	mov [byte uref uv.cfence],edx
	spop 3
	ret
;-
;- fence@ 	( -- n1 n2 n3 )					'fencef'	forth
;-	fetch latest {fence!}-ed values to tos, n1:=c-fence, n2:=fence, n3=i-fence
;-	NOTE: n3, @i-fence, is 1st un-fenced index, following fenced one.
fencef:GA;:ok:
; c: -/-
	spush
	mov eax,[byte uref uv.cfence]
	spush
	mov eax,[byte uref uv.fence]
	spush
	mov eax,[byte uref uv.ifence]
	ret
;-
;- argp   	( -- p )			(m,7)		'argpp'		linux
;-	get ptr to array of ptr-s to program's asciz environment strings.
argpp:GA M;:ok:
	spush
	mov eax,[uref uv.uargp]
M argp
	ret
;-
;- envp   	( -- p )			(m,21)		'envpp'		linux
;-	get ptr to array of ptr-s to program's asciz environment strings.
envpp:GA M;:ok:
	spush
	mov eax,[uref uv.uargc]
	lea eax,[eax*4+8]
	add eax,[uref uv.uargp]
M envp
	ret
;-
;----------------------------------------------------------------
;--- user vari ptrs (re glossary)
;-	NOTE: those w. no stack notation not (yet) available by name.
;-

;-	
;:ok:(all user vari)
;-
;- <top>  	( -- p )					'utop'		hidden
;-	p *is* ptr to top of user-vari area - neither read nor store!
utop:	uvuv
;-
;- (eax)  	( -- p )					'ueax'		hidden
;-	@tos, initial/task switching register value
ueax:	uvuv	
;-
;- (ecx)  	( -- p )					'uecx'		hidden
;-	W-reg., initial/task switching register value
uecx:	uvuv	
;-
;- (edx)  	( -- p )					'uedx'		hidden
;-	T-reg., initial/task switching register value
uedx:	uvuv	
;-
;- (ebx)  	( -- p )					'uebx'		hidden
;-	'KR', ref into kernel memory (ROM)
;-	initial/task switching register value
uebx:	uvuv	
;-
;- (esp)  	( -- p )					'uesp'		hidden
;-	'RP', esp of initiating program
uesp:	uvuv
;-
;- (ebp)  	( -- p )					'uebp'		hidden
;-	'UR', ref into user vari
;-	initial/task switching register value
uebp:	uvuv	
;-
;- (esi)  	( -- p )					'uesi'		hidden
;-	'LS', locals ptr
;-	initial/task switching register value
uesi:	uvuv	
;-
;- (edi)  	( -- p )					'uedi'		hidden
;-	'PS', datastack running ptr
;-	initial/task switching register value
uedi:	uvuv	
;-
;- (aux)  	( -- p )					'pauxp'		hidden
;-	i/o system flags
pauxp:	uvuv uv.aux
;-
;- (sstop)  	( -- p )					'sstop'		hidden
;-	true address of top of .ss section (lowest address)
;-	section can be extended by pushing any item to requested new address, e.g:
;-	{ ]] (sstop) @ 10000 - abs>4th r0 @ swap r0 ! rp! 0 >r rdrop r0 ! rp! [[ }
sstop:	uvuv
;-
;- (bot)  	( -- p )					'mbot'		hidden
;-	true address of bottom of user task memory
mbot:	uvuv
;-
;- (top)  	( -- p )					'mtop'		hidden
;-	true address of global RAM/.bss top, end of user area
mtop:	uvuv
;-
;- d-top  	( -- p ) 					'dtop'		forth
;-	top of data-space (kref'd ptr!), -re- {d0}, {here}, {unused}
dtop:	uvuv
;-
;- c-top  	( -- p ) 					'ctop'		forth
;-	top of code-space (kref'd ptr!), -re- {c0}, {c-here}
ctop:	uvuv
;-
;- s0		( -- p )							forth
;-	datastack, bottom up; -re- {sp@}, {depth}, {-depth}
s0:	uvuv
;-
;- lp		( -- p )							local
;-	ptr into local data-space,
;-	content temporarily exchanged w/ {dp} while compiling a local word header.
lp:	uvuv
;-
;- r0		( -- p )							forth
;-	initial returnstackptr, -re- {rp@}
r0:	uvuv
;-
;- up		( -- p )							forth
;-	memtop uservari ptr, top down, after last used item
up:	uvuv
;-
;- d0		( -- p )							forth
;-	bottom of dataspace, in .bss section
d0:	uvuv
;-
;- dp		( -- p )							forth
;-	ptr into dataspace, -re- {here}, {unused}, {d0}
;-	content temporarily exchanged w/ {lp} while compiling a local word header.
dp:	uvuv
;-
;- c0		( -- p )							forth
;-	base address of codespace
c0:	uvuv
;-
;- cp		( -- p )							forth
;-	ptr into codespace, in .bss section; -re- {c-here}, {c-top}
cp:	uvuv
;-
;- state  	( -- p ) 			(fg)(a)				forth
;-	ptr to compile/interpret state flag(s).
;-	state@ zero if actively interpreting, any non-zero otherwise.
;-	lib4th {state} complies to standard, though non-zero values 
;-	have specific control functions different to common useage!
state:	uvuv
;-
;- (head)  	( -- p ) 					'pheadp'	hidden
;-	xec entry ptr while compiling
pheadp:	uvuv uv.head
;-
;- (tail)  	( -- p ) 					'ptailp'	hidden
;-	backup ptr while compiling
ptailp:	uvuv uv.tail
;-
;- (scanwl)	( -- p )					'uscanwl'	hidden
;-	wordlists searching 2nd-ary error handler
uscanwl:uvuv
;-
;- csp    	( -- p ) 							forth
;-	saved compile-stack ptr
csp:	uvuv uv.ucsp
;-
;- (source-id)	( -- p ) 					'psourceidp'	hidden
;-	root of input source redirection linklist
psourceidp:uvuv uv.sourceid
;-
;- (tib)  	( -- p ) 					'ptibp'		hidden
;-	ptr to ptr to terminal input buf.
ptibp:	uvuv uv.tib
;-
;- span	  	( -- p )			(79)				forth
;-	ptr to number of chars fetched with latest expect
span:	uvuv
;-
;- blk 	  	( -- p ) 			(a:b)				forth
;-	ptr to block number storeage
;-	double will fetch {blk} to l.s. cell, and {in}
blk:	uvuv
;-
;- in		( -- p ) 			(fg)		'xin'		fig
;- >in		( -- p ) 			(a:c)		'xin'		forth
;-	disp from start to parse area in input buf.
xin:	uvuv uv.in
;		received string len in/of current parse buf.
in0:	uvuv
;-
;- (sighnd) 	( -- p )					'sigh'		linux
;-	base address of linux signal handlers, -re- {sighnd-size}
sigh:	uvuv
;-
;- scr		( -- p )							forth
;-	fig-4th screenfile line editor, screen number.
scr:	uvuv uv.uscr
;-
;- (argn)  	( -- p ) 	arguments count(er)		'argn'		hidden
argn:	uvuv uv.uargn
;-
;- (bits)  	( -- p ) 					'bbitsb'	hidden
;-	internally used different purpose auxilary flagbits
bbitsb:	uvuv uv.bits
;-
;- out		( -- p ) 			(fg)		'nout'		forth
;-	editor, output chars cnt
nout:	uvuv
;-
;- rows		( -- p ) 							forth
;-	editor, output lines cnt
rows:	uvuv
;-
;- r#		( -- p ) 			(fg)		'unr'		editor
;-	editor, cursor posn in screenfile
unr:	uvuv
;-
;- (pchr)  	( -- p )					'pchr'		hidden
;-	dft: [32,126] & [128,255] ranges of printing chars
pchr:	uvuv uv.upchr
;-
;- (escc)  	( -- p )					'uescc'		hidden
;-	escape symbol for -re- {etype}
uescc:	uvuv
;-
;- (vtim)  	( -- p ) 					'uiovtim'	hidden
;-	lo.b: VTIME, hi.b VMAX byte values for stdin ioctl
uiovtim:uvuv
;-
;- (iolink)	( -- p ) 					'iolink'	hidden
;-	i/o redirection link pointer, re "lib4th.mac", <pushiosrc> & <popiosrc>.
iolink:uvuv
;-
;- (ierr)  	( -- p )	input error code		'ierr'		hidden
ierr:	uvuv
;-
;- a-heap 	( -- p ) 					'aheap'		hidden
;-	allocated, used memory heap linklist
aheap:	uvuv
;-
;- f-heap 	( -- p )					'fheap'		hidden
;-	linked list to free allocated heap blocks
fheap:	uvuv
;-
;- (ulink)	( -- p )					'plinkp'	hidden
;- 	circular linklist to user vari & task reference.
;-	this is where <uref> points, jobs' UP = ebp.
plinkp:	uvuv uv.link
;-
;- (lkindex)	( -- p )					'plkindexp'	hidden
;-	running ptr to word execution address indices (preserved in tasks!)
plkindexp:uvuv uv.lkindex
;-
;- (m/mod)  	( -- p ) 	xec vec				'upmbymp'	hidden
upmbymp:uvuv
;-
;- (eof)  	( -- p ) 	eof character			'peofp'		hidden
peofp:	uvuv uv.ueof
;-
;- (ioacc)	( -- p ) 					'uioacc'	hidden
;-	up to 4 accumulated input chars (for vt emu)
uioacc:	uvuv
;-
;- fld		( -- p )	uvari, free for user aplication	'ufld'		forth
ufld:	uvuv uv.fld
;-
;- dpl		( -- p )			(fg)				forth
;-	decimal 'point' location, i.e. posn of last numeric fractions marker from 
;-	(dpl). type flag settings from numeric stg conversion -re- {[number]}:
dpl:	uvuv
;-
;- base		( -- p )							forth
;-	ptr to number conversion radix
base:	uvuv
;-
;- hld		( -- p )							forth
;-	ptr to current posn in number to string conversion buf.
hld:	uvuv
;-
;- hld0		( -- p )							hidden
;-	initial {hld} after start of number conversion
hld0:	uvuv
;-
;- (tick)  	( -- p )					'ptickp'	hidden
;-	cpu timing cons, cps.
ptickp:	uvuv uv.utick
;-
;- (locb) 	( -- p )					'locb'		hidden
;-	used local bytes counter
locb:	uvuv
;-
;- (locn) 	( -- p )					'locn'		hidden
;-	local cells count in currently being defined word
locn:	uvuv
;-
;- (padsz) 	( -- p )					'padsz'		hidden
;-	disp from here to pad2, pad2 to pad, max+4 stg size for {word}
;-	{word} fetching long stg-s, <nul>-enclosed, if @(padsz) set to > 256.
padsz:	uvuv
;-
;- fence  	( -- p )							forth
;-	upper bound of non-alterable dictionary, 1st writable position.
fence:	uvuv
;-
;- c-fence	( -- p )					'cfence'	forth
;-	upper bound of non-alterable compile-space, 1st writable position.
cfence:	uvuv
;-
;- i-fence	( -- p )					'ifence'	forth
;-	upper bound of non-alterable uot-index, no. of 1st writable ix.
ifence:	uvuv
;-
;- (argc)  	( -- p ) 							hidden
;-	program arguments cnt
argc:	uvuv uv.uargc
;-
;- (argp)  	( -- p )							hidden
;-	ptr to current argument
argp:	uvuv uv.uargp
;-
;- (bootp)  	( -- p )					'ubootp'	hidden
;-	ptr to executeable's data area, supplied sourcefile/data, etc
ubootp:	uvuv
;-
;- (path)	( -- p )					'upath'		hidden
;-	ptr to {path} string
upath:	uvuv
;-
;- (cmc)  	( -- p )					'ucmc'		hidden
;-	compiled words counter, for optimization support
ucmc:	uvuv
;-
;- (oerr)  	( -- p )	output error on write		'oerr'		hidden
oerr:	uvuv uv.uoerr
;-
;- (chans)	( -- p ) 	i/o channels redirection table	'pchansp'	hidden
pchansp:uvuv  uv.chans
;-
;- (inchp)		( -- p ) 				'pinchpp'	hidden
;-	stores latest used channels chp, of _any_ i/o call
pinchpp:uvuv  uv.inchp
;-
;- (outchp)		( -- p ) 				'poutchpp'	hidden
;-	stores latest used channels chp, of _any_ i/o call
poutchpp:uvuv  uv.outchp
;-
;- (cc4th)	( -- p ) 			(83)		'icc4th'	hidden
;-	user vector, kref(ptr) to base of ctrl actions jump-table, else zero
icc4th:uvuv 
;-
;- voc-link	( -- p )					'voclink'	forth
;-	vocabulary link pointer, linkage by disp to respective ptrs,
;-	chained { voc-link begin dup @ -dup while - dup voc>lfa id. repeat drop }
voclink:uvuv
;-
;- last		( -- p )			(79)		'lastu'		forth
;-	last accessible word in system. -re- {latest} wrt {current} voc.
lastu:	uvuv uv.last
;-
;- current	( -- p )							forth
;-	ptr to double voc descriptor of defining vocabulary, i.e. vocabulary 
;-	the wordlist of which currently will be extended by new definitions.
current:uvuv
;-
;- ocurrent	( -- p )							hidden
;-	previous current voc, after {definitions}. 
;-	{prevdef} restores {current} from {ocurrent}.
ocurrent:uvuv
;-
;- context	( -- p )							forth
;-	ptr to double voc descriptor of 1st voc in search order
context:uvuv
;-
;- ocontext	( -- p )							hidden
;-	old context, 2nd volatile item on stack, ALSO moves into voc-stack
ocontext:uvuv
;-
;- (tattr)	( -- a )					'utattr'	hidden
;-	program's terminal settings, true address of data structure
utattr:	uvuv
;-
;- (tattr-s)	( -- a )					'ustattr'	hidden
;-	initial terminal settings, true address of data structure
ustattr:uvuv
;-
;- (vpal) 	( -- p )					'uvpalette'	hidden
;-	program's video terminal mode, ptr to data structure
uvpalette:uvuv
;-
;- width  	( -- p )							forth
;-	max. length of a 4th name moved into header data (currently un-used).
width:	uvuv uv.uwidth
;-
;- (dpl)  	( -- p )					'pdplp'		forth
;-	decimal point location ptr
pdplp:	uvuv uv.udplm
;-
;- (#tab) 	( -- p)						'pntabp'	hidden
;-	<tab> spacing. (1 if set to zero)
pntabp:	uvuv uv.untab
;-
;- (tib-max)	( -- u )					'ptibmaxp'	hidden
;-	max bytes of tib storage
ptibmaxp:uvuv uv.utibmax
;-
;- (r)		( -- p )			(f6)		'uprp'		hidden
;-	ptr to max 4 fill-up chars for formatted string output, for .r, u.r etc
prp:	uvuv uv.uprp
;-
;- (eol)  	( -- p ) 					'ueol'		hidden
;- 	up to 4 bytes for eol sequence, CR/LF, <nl> char(s)
ueol:	uvuv
;-
;- (lo>hi)	( -- p )					'ulo2hi'	hidden
;-	i/o chars, bytewise, lo mapped into hi, 0:no translation
ulo2hi:	uvuv
;-
;- (hi>lo)	( -- p ) 					'uhi2lo'	hidden
;-	i/o chars, bytewise, hi mapped into lo, 0:no translation
uhi2lo:	uvuv
;-
;- (fperm)	( -- p )					'ufperm'	hidden
;-	file permission bits for {open-file} etc, dft 600o
ufperm:	uvuv
;-
;- (can)  	( -- p ) 					'ucan'		hidden
;-	lo-w:d_can exit code, hi-w:no. of sequential chars=d_can
ucan:	uvuv
;-
;- (quit) 	( -- p )					'uquit'		hidden
;-	execution entry
uquit:	uvuv
;-
;- (ekey?)	( -- p )					'uekeyq'	hidden
;-	basic character input
uekeyq:	uvuv
;-
;- (emit) 	( -- p )					'uemit'		hidden
;-	{emit} xec vec.
uemit:	uvuv
vemit:	uiui emit	; ptr is in words' pfa
;-
;- (type) 	( -- p )					'utype'		hidden
;-	the basic operation for text output to stdout.
utype:	uvuv
;-
;- (accept)	( -- p )					'uaccept'	hidden
;-	ptr to true address of {accept}.
uaccept:uvuv
;-
;- (stimo)  	( -- p ) 					'ustimo'	hidden
;-	timout when waiting for input. nanoseconds setting at (p-4).
ustimo:uvuv
;-
;- (echo)					(n.i.)
;-	single chars output (stdout)
uecho:	uvuv
;-
;- (ok)		( -- p )					'upokb'		hidden
;-	ptr to true execution ptr to kbd input prompt
;-	to re-define, for instance, enter { ' .s idx>xec 4th>abs (ok) ! }
upokp:uvuv
;-
;- (error)	( -- p )					'uerror'	hidden
;-	ptr to true address of error handler
uerror:	uvuv
;-
;- (?error)	( -- p )					'uqerror'	hidden
;-	ptr to true address of conditional error handler
uqerror:uvuv
;-
;- (eval)  	( -- p )					'ueval'		hidden
;-	ptr to true address of {evaluate}
ueval:	uvuv
;-
;- (compile)	( -- p )	xec vec				'ucompile'	hidden
ucompile:uvuv
;-
;- (interpret)	( -- p )	xec vec				'uinterpret'	hidden
uinterpret:uvuv
;-
;- (]interpret)   ( -- p )	dft unset			'ubinterpret'	hidden
ubinterpret:uvuv
;-
;- (interpret[)	( -- p )	dft unset			'uinterpretb'	hidden
uinterpretb:uvuv
;-
;- ([debug])	( -- p )	dft (compile)			'updebugp'	hidden
updebugp:uvuv
;-
;- (sp!)  	( -- p )	xec vec				'ustosp'	hidden
ustosp:	uvuv
;-
;- (>number)	( -- p )					'utonum'	hidden
;-	convert, default xec vec
utonum:uvuv
;-
;- (number)	( -- p )	xec vec				'unumber'	hidden
;-	dft {f-number}/{a-number} according to preferred mode, re Makefile
unumber:uvuv
;-
;- (um/mod)	( -- p )					'usmbym'	hidden
;-	default UM/MOD variant (floored/near zero), init {sm/rem}
usmbym:	uvuv
;-
;- (?terminal)	( -- p )	xec vec				'uqterminal'	hidden
uqterminal:uvuv
;-
;- (pause)	( -- p )	xec vec				'upause'	hidden
;-	0 or xec after timout to {wait4..} words elapsed, by true address.
upause:	uvuv
;-
;- (buffer)	( -- P )	xec vec				'ubuffer'	hidden
ubuffer:uvuv
;-
;- (block)	( -- p )	xec vec				'ublock'	hidden
ublock:	uvuv
;-
;- (source)	( -- p )	xec vec				'usource'	hidden
usource:uvuv
;-
;- (vbot) 	( -- p )							hidden
;-	end of pre-defined user area, up := disp to uv.link in uv.up,
;-	{ (vbot) tib tib-max + - 4/ } cells free for application defined user vari
vbot:	uvuv
;-
;- (#tib)  	( -- p )					'pntibp'	hidden
;-	no. of chars in tib
pntibp:	uvuv uv.uintib
;-
;- (uot)  	( -- p )					'ultab'		hidden
;-	bottom of executeables pointers, all true addresses
ultab:	uvuv

;-

;-
;-================================================================
;--- kernel vari/"value"s
;-	

;-
;- wdw-lk	( -- p )			n.i.		'wdwlk'		hidden
;-	console windows linkage, top down ptr, bottom up is next cell
wdwlk:GA
	mov ecx,wdwlkv-Y
	jmp pdovar
;-
;- (alarm)	( -- )						'palarmp'	linux
;-	kernel {defer}red, executed by redirected SIGALRM handler
;-	if @p =/= 0 xec kernel-vari (alarm), else dft { [s] }
palarmp:GA
	mov ecx,palarmpv-Y
	jmp pdodefer
;-	

%endif;__UVARI_I
;-
;-=====================================================================;
;- l4uvari.inc <eof>
