;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4vt.inc
;-  created	: 04-aug-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __VT_I
;-  assembler	: nasm 0.98-3e
;-  description	: "lib4th" vt control, termcaps
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  source	: man-pages
;-		: console_codes(4), console_ioctls(4), term(5)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-
;-	
;-	

%ifndef __VT_I
%define __VT_I
%define vtinc
;-
;-----------------------------------------------------------------
;- -- vt... and ctrl sequences
;-	different terminal emulations emulate "standards" differently and
;-	not always correct. following was tested in console, xterm, rxvt.
;-	all handle CSI sequences differently, false or, simply don't...
;-
;-	vt words is a minimal set of word for console control until final
;-	implementation, which is planned to utilize terminfo settings.
;-	words' names might change, according to terminfo.
;-
;- -- console.c:
;-  r	0x00,0xaa,0x00,0xaa,0x00,0xaa,0x00,0xaa,0x55,0xff,0x55,0xff,0x55,0xff,0x55,0xff
;-  g	0x00,0x00,0xaa,0x55,0x00,0x00,0xaa,0xaa,0x55,0x55,0xff,0xff,0x55,0x55,0xff,0xff
;-  b	0x00,0x00,0x00,0x00,0xaa,0xaa,0xaa,0xaa,0x55,0x55,0x55,0x55,0xff,0xff,0xff,0xff
;- -- attribute bits:
;-  0,1	intensity (0..2)
;-  2	underline
;-  3	reverse
;-  7	blink
;-	
;-
;-----------------------------------------------------------------
;- -- ascii ctrl constants, in {vt} voc.
;-

;- 
;- <bel>  	( -- n )					'lbelg'		vt
;-	cons, BEL (0x07, ^G) beeps;
lbelg:GA;:ok:
	call dobcons
	db AC_BEL
;-
;- <bs>		( -- n )					'lbsg'		vt
;-	cons, BS  (0x08, ^H) cursor back  one  column, not clearly
;-	def'd: backing up or additionally erasing char before curpos.
;-	(but not past the beginning of the line)
lbsg:GA;:ok:
	call dobcons
	db AC_BS
;-
;- <ht>		( -- n )					'lhtg'		vt
;-	cons, HT  (0x09, ^I) goes to the next tab stop or to 
;-	the end of the line if there is no earlier tab stop
lhtg:GA;:ok:
	call dobcons
	db AC_HT
;-
;- <lf>		( -- n )					'llfg'		vt
;-	cons, LF  (0x0A, ^J), linefeed; Linux: also console & files <eol> code.
llfg:GA;:ok:
	call dobcons
	db AC_LF
;-
;- <vt>		( -- n )					'lvtg'		vt
;-	cons, VT (0x0B, ^K) vertical tabulator, linefeed to console channels
lvtg:GA;:ok:
	call dobcons
	db AC_VT
;-
;- <ff>		( -- n )					'lffg'		vt
;-	cons, FF (0x0C, ^L), linefeed to console channels
lffg:GA;:ok:
	call dobcons
	db AC_FF
;-
;- <cr>		( -- n )					'lcrg'		vt
;-	cons, CR  (0x0D, ^M) gives a carriage return
lcrg:GA;:ok:
	call dobcons
	db AC_CR
;-
;- <so>		( -- n )					'lsog'		vt
;-	cons, SO  (0x0E, ^N) activates G1 character set (graphic chars), 
;-	and if LF/NL (new line mode) is set also a carriage return
lsog:GA;:ok:
	call dobcons
	db AC_SO
;-
;- <si>		( -- n )					'lsig'		vt
;-	cons, SI  (0x0F, ^O) activates the G0 character set
lsig:GA;:ok:
	call dobcons
	db AC_SI
;-
;- <can>   	( -- n )					'lcang'		vt
;-	cons, CAN (0x18, ^X) interrupt escape sequence
lcang:GA;:ok:
	call dobcons
	db AC_CAN
;-
;- <sub>  	( -- n )					'lsubg'		vt
;-	cons, SUB  (0x1A, ^Z)  interrupt escape sequence
lsubg:GA;:ok:
	call dobcons
	db AC_SUB
;-
;- <esc>  	( -- n )					'lescg'		vt
;-	cons, ESC (0x1B, ^[) starts an escape sequence
lescg:GA;:ok:
	call dobcons
	db AC_ESC
;-
;- <del>  	( -- n )					'ldelg'		vt
;-	cons, DEL (0x7F) is ignored (?, linux console)
ldelg:GA;:ok:
	call dobcons
	db AC_DEL
;-
;- <csi>  	( -- n )					'lcsig'		vt
;-	cons, CSI (0x9B) is equivalent to ESC [.
lcsig:GA;:ok:
	call dobcons
	db AC_CSI
;-
;-	
;-----------------------------------------------------------------

;
%if 1
%xdefine vt_key   kbdkey	; vtfp,qvt
%xdefine vt_pkey  pqkey		; wcsi=wesc(,vtfp)
%xdefine vt_qkey  kbdqkey	; pqkey
%xdefine vt_type  kbdtype	; vtsp,setpalette
%xdefine vt_wait4 wait4kbd	; vtfp,xy
%xdefine vt_emit  kbdemit	; rstpalette,qvt,xy,csi2p=csr+atxy,vts,vtf,curs,curf
%else
%xdefine vt_key   ekey
%xdefine vt_pkey  ekey
%xdefine vt_qkey  qekey
%xdefine vt_type  betypeb
%xdefine vt_wait4 wait4in
%xdefine vt_emit  emit
%endif
;

; non-blocking read a char from input & ret eflags Z if none received
; out:	eax:=char, eflags := NZ invalid
pqkey:A;:ok:
	call kbdqkey	; ( -- char flg )
	inc eax		; Z if char received
	spop		; drop flag
	ret
;-
;- !vtp		( n c -- )					'vtsp'		vt
;-	send a vt.. numeric parameter w/ trailing char c, unless c<0.
;-	NOTE: numeric stg output buffer @hld will be overwritten.
vtsp:GA;:ok:
	push eax
	mov eax,0
	call ltsharp
	pop ecx
	test ecx,ecx
	js .d		;?; digits only
	spush
	mov eax,ecx
	call hold
.d:
	call sharps
	call sharpgt
	jmp kbdtype
;-
;- @vtp		( -- n )					'vtfp'		vt
;-	fetch a vt.. decimal parameter, regardless of @base setting.
vtfp:GA;:ok:
; in:	-/-
; out:	eax:=edx:=vt-parameter, ecx:=dlm char
	xor edx,edx
	jr .n
.l:
	cmp eax,byte'9'
	ja .r		;?; not a decimal digit
	lea eax,[eax+edx*8-'0']
	lea edx,[eax+edx*2]
	spop
.n:
;;	call wait4kbd;
	call kbdkey
	cmp eax,byte'0';;test eax,eax
	jnc .l;;jnz .l;
;;	test eax,eax;
;;	jz .n		;?; ignore <nul>
.r:
	mov ecx,eax	; pass 1st non-numeric char
	mov eax,edx
	ret
;-
;- vt?esc 	( -- flg )					'wesc'		vt
;-	initial response to vt request:
;-	wait for <esc>, timeout after kbd?key failed for 1s, 
;-	ret ff if non-<esc> char received w/ <ecx>-reg = char-code.
;-	(sec of reading timeout modifiable in asm source, at <wesc> label)
wesc:GA;:ok:
; ret:	eflags	Z if <esc> received, else ret NZ
;	eax	char, else zero
; c: ecx, edx
	push_ibyte(AC_ESC)
%if 0;;;
	spush
.n:
	xor eax,eax
	call wait4in	; wait for response
	spop
	call pqkey
%else
	spush
.n:
	spop
	call kbdkey
%endif
	mov ecx,eax
	cmp eax,[esp]
.c:
	jz .r		;?; <esc> or <csi> or "[ received
.e:
	xor eax,eax	; prepare 'false' return-flag
	cmp eax,ebx	; force NZ
.r:
	lea esp,[esp+4]
	ret
;-
;- vt?csi 	( -- flg )					'wcsi'		vt
;-	initial response to vt request:
;-	wait for <esc>[ or <csi>, timeout after kbd?key failed for 1s,
;-	ret ff if non-<esc> char received w/ <ecx>-reg = char-code.
;-	(sec of reading timeout modifiable in asm source, at <wesc> label)
wcsi:GA
	call wesc
	push_ibyte("[")
	jz wesc.n	;?; <esc> received
	cmp cl,AC_CSI
	jnz wesc.e
	mov eax,ecx
	jr wesc.r
;
ibyocons:A; ret NZ if stdin or stdout is a console channel
	call getstdout
        test dword[kref ecx+cdt.iofl-cdt],mcons
	jz .r
	call getstdin
        test dword[kref ecx+cdt.iofl-cdt],mcons
.r:
	ret
;-
;-----------------------------------------------------------------
;- -- ctrl and vt.. sequences/actions
;-	orderly output might require an initial {max-xy} if stdout connected to a pipe
;-
;-	

; ( -- ) ecx = value to {hold}
setpalette.c:A;<colour)
	spush
	mov eax,ecx
	jmp hold
;-
;- set-palette	( r g b n -- )					'setpalette'	vt
;-	set a selected colour by number -1 < n < 16 to the composition 
;-	of three 8bit colour values for (r)ed, (g)reen, (b)lue.
;-	n: 0 black, 1 red, 2 green, 3 yellow, 4 blue, 5 magenta, 6 cyan, 7 white
;-	n: -1 to resetting to default values, colour items discarded.
;-	NOTE1: restricted to the Linux console, won't work in an xterm, etc.
;-	NOTE2: use Linux <reset> to recover from an unexpectedly aborted program.
setpalette:GA;:ok:
	cmp eax,byte -1
	jz .d
	push dword[byte uref uv.base]
	mov dword[byte uref uv.base],16
	spush
	_mov eax,-3		; n to last arg
	call roll		; -3 roll
	spush
	call ltsharp		; begin numeric conversion
	call .s			; r
	call .s			; g
	call .s			; b
	_mov eax,0
	call sharp		; n
	_mov ecx,'P'		; prepend command chars
	call .c
	_mov ecx,']'
	call .c
.t:;	<colour)
	_mov ecx,AC_ESC
	call .c
	call sharpgt		; finish stg
	pop dword[byte uref uv.base]
	jmp kbdtype		; send stg to keyboard channel
.s:A;	; put two digits into {hld} buf
	_mov eax,0		; drop 0
	call sharp		; #
.o:;	<colour)	; put one digit into {hld} buf
	cmp dword TOS(1),byte -1
	jz .u			;?; a noop argument
	call sharp		; #
.u:
	TOSPULL(1)		; sdrop
	ret
.d:A
	TOSPULL(4)
;-
;- reset-palette	( -- )						'rstpalette'	vt
;-	restore terminal colours from inital (default) values. 
;-	this will restore to defaults from the Linux KERNEL! neither of ioctl or 
;-	vt sequences, offer any means to reading the #actual# console settings! 
;-	NOTE: use Linux <reset> to restore from an unexpectedly aborted program. 
rstpalette:GA;:ok:
	spush
	mov eax,']R'<<8|AC_ESC
	jmp kbdemit
;-
;- [i]		( -- co )			(kv)		'bibv'		vt
;-	latest {ink} colour
bib:KV
;-
;- ink		(co -- )					'vtink'		vt
;-	set foreground colour/vt-mode, update {[i]} if co is a colour value.
vtink:GA
	mov ecx,[kref fgv-Y]	; {fg}
	jr vtpaper.n
;-
;- [p]		( -- co )			(kv)		'bpbv'		vt
;-	latest {paper} colour
bpb:KV
;-
;- paper  	( co -- )					'vtpaper'	vt
;-	set background colour/vt-mode, update {[p]} if co is a colour value.
vtpaper:GA
	mov ecx,[kref bgv-Y]	; {bg}
.n:
	add ecx,databot
	cmp eax,byte 8
	jnc .c			;?; not a coulour
	mov [kref ecx+bpbvv-bgvv],eax; sto [p] or [i], rsp
.c:
	spush
	mov eax,[kref ecx]	; :continued: in vtcolour
;-
;- colour 	( co bf -- )					'vtcolour'	vt
;-	set console colour from colour- and back- or foreground values,
;-	re colour and bg/fg names in {vt}-wordlist - {black}.., {bg}..
;-	NOTE:
;-	these settings can also control the xterm, rxvt, etc. display but, use 
;-	"<esc>[" for <csi> because some terminals won't interpret the latter.
;-	{colour} applicable to all 'ECMA-48 Graphic Rendition' cmd-s, "<esc>[<num>m"
;-	other parameter values -re- man 4 console_codes:
;-	{me}	( -1 0 -- )	reset all attributes to their defaults
;-	{md}	( -1 1 -- )	bold
;-	{so}	( -1 2 -- )	half-bright
;-	{us}	( -1 4 -- )	underscore
;-	{mb}	( -1 5 -- )	blink (light colour background)
;-	{mr}	( -1 7 -- )	reverse video
;-		( 0  1 -- )	reset toggle-meta flg, selected mapping, display ctrl flg.
;-		( 1  1 -- )	reset toggle-meta flg, select null mapping, set display ctrl flg.
;-		( 2  1 -- )	set toggle-meta flg, select null mapping, set display ctrl flg.
;-		( 1  2 -- )	normal intensity (not ECMA-48)
;-	{se}	( 2  2 -- )	normal intensity
;-	{ue}	( 4  2 -- )	underline off
;-		( 5  2 -- )	blink off
;-		( 7  2 -- )	reverse video off
;-		( 8 fg -- )	underscore on, set default foreground colour
;-		( 9 fg -- )	underscore off, set default foreground colour
;-		( 9 bg -- )	default background colour
vtcolour:GA;:ok:
	push dword[byte uref uv.base]
	mov dword[byte uref uv.base],10
	mov ecx,eax
	mov eax,TOS(1)		; swap
	mov TOS(1),ecx
	spush
	call ltsharp		; co co bf
	_mov ecx,'m'
	call setpalette.c	; {hold} char <ecx>
	_mov eax,0		; 0 co bf
	call setpalette.o	; put one digit into {hld} buf
	call setpalette.o
	_mov ecx,'['
	call setpalette.c
	spush
	jmp disp setpalette.t	; finish stg, rst base, type to console channel
;-
;- ?vt		( -- flg )					'qvt'		forth
;-	true flag if stdout connected to a terminal of "vt102" type.
;-	test for '[?6' vt102, then '[?1;Pc', P=dec.digit, vt100+ of xterm/rxvt.
;-	determines whether i/o is file, returns false if in or out not a console.
;-	reading response to stdin w. zero timeout,
;-	1s timeout for leading <esc>, modify source at <qvt> if not sufficent.
qvt:GA;:ok:(w. console, xterm, rxvt)
; c: ecx,edx
;:incons: oder :outcons:, eins davon ist wurscht - welches, das ist die frage...
;			, denn: incons fr systemeigenes echo, out fr emit, etc.
	spush
	call ibyocons
	jz .n		;?; file
	call .y
	mov eax,(("c"<<8)|"[")<<8|AC_ESC;("c"<<8)|AC_CSI <- not with rxvt
	call kbdemit
	call wesc
	jnz .q
	spop
	push_ibyte(3)
.l:
	call kbdkey
	shl edx,8
	mov ecx,-1
	mov dl,al
	cmp al,'c'
	jz .e
	spop
	dec dword[esp]
	jns .l		;?; more expected
	spush
.e:
	inc ecx
.v:
	cmp edx,reg_vt102;	'c6?[' <- '[?6c'
	jz .x		;?; linux console, vt102
	cmp edx,reg_vt100x;	';1?[' <- '[?1;2c'
	jnz .r		;?; not vt100
	spop
	call vtfp
	mov eax,ecx
	mov ecx,-1;;xor ecx,ecx;
		    ;;dec ecx;
	cmp al,'c'	; '[?1;2c'
	jnz .b		;?; not in an xterm/rxvt
	inc ecx
.x:
	dec ecx		; Z
.r:
	lea esp,[esp+4]
	mov eax,ecx
.q:
	ret
.b:
	lea esp,[esp+4]
.n:
	xor eax,eax
	ret
.y:A
	call curs
	spush
	mov eax,"[H"<<8|AC_ESC
	call kbdemit	; 'home'
	call xy
	call curf
	mov ecx,databot
	mov edx,TOS(1)
	add ecx,[kref vt0v-Y]
	mov [kref ecx+4],eax
	mov [kref ecx],edx
	spop 2
	ret
;-
;- xy		( -- u1 u2 )							forth
;-	fetch current cursor posn, col u1, row u2, = -1 if posn not available.
;-	NOTE:	false values if sending a char to the last column, i.e. posn 
;-		fetched before and after return identical co-ordinates!
;-		(console should solve 'pending newline problems', internally!)
xy:GA;:ok:
; c: ecx,edx
	spush		;;
	mov eax,-1	;;call mone
	call ibyocons
;;	jz near mone;
	spush
	jz .r	;;;
	mov eax,"[6n"<<8|AC_ESC
	call kbdemit
%ifdef __xwait
	push edx
	mov edx,1	; timo
	lea ecx,[kref wait4kbd-Y]
	call wait4x.n
	pop edx
%else
;;	call wait4kbd;;=> waiting in <wcsi>
%endif
	call wcsi	; expect "\e[" or "\i"
	not eax		; .xx. | -1
	jnz .r
	spop 2
;;.l:
;;	jnz near true;
;;	spop
	call vtfp
	call vtfp
	jmp swap
;-
;- max-x  	 ( -- n )			(vt)				vt
;-	max. cursor position to the right of terminal window
maxx:GA;:ok:
	call maxxy
	spop
xy.r:
	ret
;-
;- max-y   	 ( -- n )			(vt)				vt
;- 	max. cursor position toward bottom of terminal window
maxy:GA;:ok:
	call maxxy 
	TOSPULL(1)
	ret
;-
;- row		( -- n )			(vt)				vt
;-	current cursor posn in row n
row:GA;:ok:
	call xy
	TOSPULL(1)
	ret
;-
;- to-row  	( n -- )			(vt)				vt
;- 	set cursor posn to row n at current column
torow:GA;:ok:
	push eax
	spop
	call xy
	pop eax
	jr atxy
;-
;- col		( -- n )			(vt)				vt
;-	current cursor posn in column n
col:GA;:ok:
	call xy
	spop
	ret
;-
;- tab-to	( n -- )					'tabto'		forth
;-	advance cursor to column max(current+1,n) at current line
tabto:GA;:ok:
	call xy
	spop
	inc eax
	call max	;	:continued: in <tocol>
;-
;-
;- to-col  	( n -- )			(vt)				vt
;- 	set cursor posn to column n at current line
tocol:GA;:ok:
	call xy
	TOSPULL(1)
	jr atxy
;-
;- home		( -- )								vt
;-	^^ (ascii 30), cursor to top left of screen, non-terminal receives <ff>.
;-	emulated via { 0 0 at-xy } because only linux console does the standard 
;-	vt command emulation correctly - may be different with other linux setup...
home:GA;:ok:
	spush
	xor eax,eax
	spush
;-
;- at-xy  	( u1 u2 - )			(vt)(a:f)	'atxy'		forth
;-	place cursor at posn u1 chars from the left and u2 rows below 
;-	top, store u1 to user vari {out}, u2 to {rows}.
;-	linux console esc[6n and esc[p1;p2H are in-consistent (Lx 2.2.14):
;-	reading refers to home posn = 1;1 while storing begins at 0;0.
;-	Lx 2.2.19 was ok. thus kvalue provided, for 'calibration' by {?vt}  
;-	such, that {xy} adjusts parameters to keeping same posn w/ { xy at-xy }.
;-	re DPANS 10.4.1.2: no-op if i/o is not a terminal.
atxy:GA;:ok:
	call swap
	spush
	mov eax,'H'
	jr csi2p
;-
;- csr		( n1 n2 -- )							vt
;-	set scrolling region to n1=top, n2=bottom row.
;-	termcap 'cs' supported by linux console, rxvt, xterm.
csr:GA;:ok:
	spush
	mov eax,'r'		; terminating char
; -
; - csi2p  	( n1 n2 c3 -- )							vt
; -	send 2 decimal parameters' csi sequence to stdout, terminated by char c3.
csi2p:;;GA;:ok:
	call ibyocons
	jz near drop3	;?; not a console
	call ltsharp
	push dword[byte uref uv.base]
	mov dword[byte uref uv.base],10
	call hold	; terminator
	spush
	xor eax,eax
	call sharps	; "y" or bot
	mov eax,';'
	call hold
	call sharps	; "x" or top
	mov eax,'['
	call hold
	mov eax,AC_ESC
	call hold
	spush 2
	call sharpgt
	pop dword[byte uref uv.base]
	jmp kbdtype
;-
;- vt!state	( -- )						'vts'		vt
;-	save console state:
;-      cursor, intensity, underline, blink, reverse, charset, color, G0, G1
;-      (note: xterm-4 doesn't save the colours)
vts:GA;:ok:
	spush
	mov eax,"7"<<8|AC_ESC
	jmp kbdemit
;-
;- vt@state	( -- )						'vtf'		vt
;-	restore most recently saved console state. -re- {vt!state}
vtf:GA;:ok:
	spush
	mov eax,"8"<<8|AC_ESC
	jmp kbdemit
;
escp:A;:ok:<g0,g1)
;eax:num ch:charset cl:default
	test eax,eax
	jz .n		;?; default mode, force "B"
	jns .e		;?; numeric arg
	movzx eax,al	; code by character
	jr .c
.n:
	inc eax
.e:
	and eax,byte 7
	jz .r		;?; use the default code
	movzx eax,byte[kref eax+escpt-Y]
; 0,'B0UK',0,0,0 => A:UK B:US 0:DEC U:?? K:DE Z:ES R:FR H:SE' 
	test eax,eax
	jz .r
.c:
	shl eax,24
	mov cl,AC_ESC
	mov ah,ch
.r:
	mov al,cl
	ret
;-
;- tab		( -- )								forth
;-	blanks till next horizontal tabulator wrt @out,
;-	using the tab spacing value from uvari (#tab).
tab:GA;:ok:
	spush
	xor edx,edx
	mov ecx,[uref uv.untab]
	mov eax,[byte uref uv.nout]
	test ecx,ecx
	jg .d
	lea ecx,[ecx+8]
.d:
	div ecx
	sub ecx,edx
	mov eax,ecx
	jmp spaces
;-
;- rev-off	( -- )				(vt)		'revoff'	vt
;-	reverse video off
revoff:GA;:ok:
	call zero
	jr rev
;-
;- rev-on 	( -- )				(vt)		'revon'		vt
;-	reverse video on
revon:GA;:ok:
	call pone
;-
;- rev		( flg -- )			(vt)				vt
;-	reverse video on (tf) or off (ff)
rev:GA;:ok:
	mov ecx,[kref brbv-Y]
	add ecx,databot
	test eax,eax
	mov [kref ecx],eax
	mov ecx,4
	mov eax,vt.revvf-Y	;off
	jz .j
	dec ecx
	mov eax,vt.revvn-Y	;on
.j:
	sub [byte uref uv.nout],ecx
	jmp zprint
;-
;- cure		( -- )				(vt)				vt
;-	enable cursor display.
;-	
;-	console (only) cursor appearance can be modified with, e.g.
;-		normal blinking underline : '\e[?2c'
;-		blinking block            : '\e[?6c'
;-		red non-blinking block    : '\e[?17;0;64c'
;-	1st arg:	0:dft, 1:invis, 2:underline, 4:(?), 8:block
;-			+16: soft curs, +32:change bg, +64: bg=/=fg
;-	2nd:		char attr to change, 'toggle mask' (XOR):
;-			bits 0..2: fg colour, 3:fg highlight/blink
;-			bits 4..6: bg colour, 7:fg highlight/blink
;-	3rd:		char attr to set, before applying 2nd arg,
;-			i.e. clearing a bit by passing as 2nd and 3rd.
;-	un-specified args default to zero.
cure:GA
	mov edx,'25'
	jr decpm.h
;- curd		( -- )				(vt)				vt
;-	disable cursor display.
curd:GA
; , terminfo Linux console 'vi' capability
	mov edx,'25'
	jr decpm.l
;-
;- abscur	( -- )								vt
;-	cursor addressing wrt top left of terminal screen (console dft).
;-	NOTE: immediately changes cursor posn if prev. state was relative.
abscur:GA;:ok:
	mov edx,'6'
decpm.l:
	mov ecx,'l'
	jr decpm
;-
;- relcur	( -- )								vt
;-	cursor addressing wrt top left of scrolling region.
;-	NOTE: immediately changes curpos and, is inconsistent wrt {xy} and {at-xy}.
relcur:GA;:ok:
	mov edx,'6'
decpm.h:
	mov ecx,'h'		; set'
; ( -- ) ecx: "h"/"l", edx: ascii-#num arg
decpm:
	spush 3
	mov eax,'[?'<<8|AC_ESC	; introduce DECSET/DECRST
	mov TOS(1),edx		; ascii # arg
	mov TOS(2),ecx		; terminal char
	call kbdemit
	call kbdemit		; num
	jmp kbdemit		; 'h'/'l'
;-
;- cup		( -- )						'cup'		vt
;-	cursor one line up
;-	(vt sequence, ok w. console, xterm, rxvt)
cup:GA;:ok:
	spush
	dec dword[byte uref uv.rows]
	mov eax,"[A"<<8|AC_ESC
	jr bs.c
;-
;- cdn		( -- )						'cdn'		vt
;-	cursor one posn to the left
;-	(vt sequence, ok w. console, xterm, rxvt)
cdn:GA;:ok:
	spush
	inc dword[byte uref uv.rows]
	mov eax,"M"<<8|AC_ESC
	jr bs.c
;-
;- cur-		( -- )						'clft'		vt
;-	cursor one posn to the left
;-	(vt sequence, ok w. console, xterm, rxvt)
clft:GA;:ok:
	spush
	mov eax,"[D"<<8|AC_ESC
	jr bs.d
;-
;- cur+		( -- )						'crgt'		vt
;-	cursor one posn to the right, cc4th action for '^_', ascii 31
;-	(vt sequence, ok w. console, xterm, rxvt)
crgt:GA;:ok:(c,x,r)
	spush
	inc dword[byte uref uv.nout]
	mov eax,"[C"<<8|AC_ESC
	jr bs.c
;-
;- bs		( -- )								forth
;-	send <bs> to @{stdout}
;-	NOTE: {bs} is back-_SPACE_, which is not back-_BLANK_ - re {del}!
bs:GA;:ok:
	call lbsg
.d:
	dec dword[byte uref uv.nout]
.c:
	push dword[byte uref uv.nout]
	call emit
	pop dword[byte uref uv.nout]
	ret
;-
;- del		( -- )								vt
;-	move cursor one column back and delete char at cursor posn
del:GA;:ok:
	spush
	mov eax,((AC_BS<<8)|' ')<<8|AC_BS
	jr bs.d
;-
;- cur!		( -- )						'curs'		vt
;-	save console cursor posn
curs:GA;:ok:
	spush
	mov eax,"[s"<<8|AC_ESC
	jmp kbdemit
;-
;- max-xy 	( -- x y )					'maxxy'		vt
;-	max. cursor position x to the right and y bottom of terminal window.
;-	NOTE: changed sizes orderly received in rxvt or xterm but, 
;-	      segfault in console e.g, after TIOCSWINSZ ioctl ($5414).
maxxy:GA;:ok:
	mov ecx,[byte uref uv.sigh]
	mov ecx,[kref ecx+sa_winch.w-sighndblk]; ptr to sizes
	test ecx,ecx
	jg .m		;?; sizes already known
	call winsz	; ioctl for window outline by chars
.m:A
	spush
	mov eax,ecx
	shr eax,16  
	spush
	movzx eax,cx
maxlw.r:
	ret
;-
;- cur@		( -- )						'curf'		vt
;-	rst console saved cursor posn
curf:GA;:ok:
	spush
	mov eax,"[u"<<8|AC_ESC
	jmp kbdemit
;-
;- max-lw 	( -- n )					'maxlw'		editor
;-	max stdout display line width, whether terminal window or file.
;-	line-width for non-terminal channels from (kernel-)value {#col}.
maxlw:GA;:ok:
	call getstdout
	test dword[kref ecx+cdt.iofl-cdt],mcons|micon; flags
	jz .f
	call maxxy
	inc eax
	spop
	jg .r		;?; console
	spop
.f:; fetch kernel value {#col}
	mov ecx,hcolv-Y
	jmp pdoval
g0.g:A
	cmp eax,byte 6
	jnc .r
	add ecx,databot
	mov [kref ecx],eax
.r:
	ret
;-
;- cls		( -- )								forth
;-	clear screen & {home} if stdin or stdout is console.
cls:GA
	call home
;-
;- clsc		( -- )								forth
;-	cc4th action for <ff>, clear screen & <cr> if stdin or stdout is console.
;-	does not affect the cursor position - use vt {home} for cursor to top left.
clsc:GA;:ok:(w. console, xterm, rxvt)
	call ibyocons
	jz ff
	spush
	mov eax,((("J")<<8|"2")<<8|"[")<<8|AC_ESC
	jr ff.n	; clr {out}, {rows}, and emit
;-
;- rs		( -- )								root
;-	reset terminal, colours palette, char founts, G0/0 character set.
rs:GA;:ok:
	call rstpalette
	spush 2
	movzx eax,word[kref ttyrst.b-Y];0+'c'<<8|AC_ESC
	call emit
	call getstdout
	push ebx
	xor edx,edx
	mov ebx,[kref ecx+cdt.chid]
	mov ecx,0x4b6d	; 'PIO_FONTRESET'
	sys_ioctl
	pop ebx
	xor eax,eax
;-
;- g0		( n -- )							vt
;-	activate G0, n-th charcter set; { 0 g0 } for the default charset.
;-	n may be passed by character when ORed with -0, { char -0 or }
;-	the rsp, linux kernel defined <esc>-sequences:
;-	 1 : LAT1_MAP	Latin-1 (ISO 8859-1)		ESC ( B
;-	 2 : GRAF_MAP	DEC VT100 pseudographics	ESC ( 0
;-	 3 : IBMPC_MAP	IBM code page 437		ESC ( U
;-	 4 : USER_MAP	User defined			ESC ( K
g0:GA;:ok:
; 0,'B0UK',0,0,0 => A:UK B:US 0:DEC U:?? K:DE Z:ES R:FR H:SE' 
	mov ecx,[kref bg0bv-Y]
	call .g
	mov ecx,'('<<8|AC_SI; ch:charset cl:default
	call escp
	jmp j04 emit
;-
;- g1		( n -- )							vt
;-	activate G1, n-th charcter set, 0 for default, -re- {g0}
g1:GA;:ok:
	mov ecx,[kref bg1bv-Y]
	call g0.g
	mov ecx,')'<<8|AC_SO
	call escp
	jmp j16 emit
;-
;- ff		( -- )								vt
;- page		( -- )								forth
;-	form feed & <cr>, <nl> if out-fd is console, clr {out} and {rows}
;-	"clear screen" is -re- {clsc}, in forth vocabulary.
ff:GA;:ok:
	call lffg	; fetch cons
	or eax,AC_CR<<8
.n:
	mov dword[byte uref uv.rows],0
	mov dword[byte uref uv.nout],0
	jmp j32 emit
;-
;- bel		( -- )								forth
;-	pc-speaker's noise from kbd/console
bel:GA;:ok:
	call lbelg
	jmp j32 emit
;-
;-----------------------------------------------------------------
;- -- termcap -re- man term(5), terminfo(5), curs_terminfo(3), curs_termcap(3)
;-	fixed capabilities' order -re- <inc/term.h.inc>, from /usr/include/term.h
;-	compiled data structure ('int' is 2 bytes):
;-		1)	00	'magic number' &432 = $11a
;-			02	bytes size of names' section (4th)
;-			04	no. of bytes in boolean section (3rd)
;-			06	no. of 'int's in numbers' section (5th)
;-			08	no. of offsets by 'int's in stgs' section (6th)
;-			0a	bytes size of string table
;-		2) terminal names, <nul>-terminated
;-		3) boolean flags bytes, even aligned at end of table
;-		4) numbers per 'int's, -1 for an undefined capability, no other -ve
;-		5) strings section, capabilites by 'int's,
;-				w/ special chars in interpreted form, e.g. "^X", "\c"
;-				padding information by "$<nn>"
;-				parameter information in un-interpretted form, e.g "%x"
;-		6) string table, stgs <nul> terminated, refs may also point into stgs.
;-		new capabilites may follow
;-	



%endif;__VT_I
;-
;-=====================================================================;
;- l4vt.inc <eof>
