;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-  file  	: l4whead.inc
;-  created	: 12-sep-01
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __WHEAD_I
;-  assembler	: nasm 0.98-3e
;-  description	: word header acces, ptr conversion
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: doc re "glossary.txt" and "l4...html" in ./doc dir,
;-  source	: (hp)
;-  requires	: the "lib4th" package
;-=====================================================================;
;-

%ifndef __WHEAD_I
%define __WHEAD_I
%define wheadinc
;-
;-================================================================
;--- interpreter, word headers structure
;-	conversion words all but {idx>lfa} & {>body} in {hidden}.
;-	"(c)" marks system defined constants, where reverse operation
;-	can be derived, e.g. for hypothetical mfa>lfa: { 0 lfa>mfa - }
;-	"(p)" depend on parameter*cons1+cons2.
;-

;-
;-	NOTE:
;-	ANS terminology "execution token" is index ix into uot.
;-	Basic words reference is by uot index.
;-	All linkage by displacement wrt lfa, towards lower addresses.
;-	re word header structure <whead> and data/flags in "lib4th.mac".
;-	valid range of ix in the actual system setup is 0 to @(lkindex)-1,
;-	absolute max is { (#tib) (ltab) - 4/ }, and less than 65535/4.
;-	a(ny) job can implement up to 16196 words.
;-
;-	

;-
;- header pointers, reference conversion
;-

;-
;- lfa>nfa	( n1 -- n2 )			(c)(m,3)	'lfa2nfa'	hidden
;-	convert wordheader linkfield ptr to namefield ptr
lfa2nfa:GA M;:ok:
; c: -/-
	lea eax,[eax-whead.l+whead.n]
M lfa2nfa
	ret
;-
;- nfa>lfa	( n1 -- n2 )			(c)(m,3)	'nfa2lfa'	hidden
;-	convert wordheader namefield ptr to linkfield ptr
nfa2lfa:GA M;:ok:
; c: -/-
	lea eax,[eax+whead.l-whead.n]
M nfa2lfa
	ret
;-
;- lfa>ifa	( n1 -- n2 )			(c)(m,3)	'lfa2ifa'	hidden
;-	convert wordheader linkfield ptr to indexfield ptr
lfa2ifa:GA M;:ok:
; c: -/-
	lea eax,[eax+whead.i-whead.l]
M lfa2ifa
	ret
;-
;- lfa>head	( n1 -- n2 )			(c)(m,3)	'lfa2head'	forth
;-	convert wordheader linkfield ptr to indexfield ptr
lfa2head:GA {equ lfa2ifa}
; c: -/-

;-
;- ifa>lfa	( n1 -- n2 )			(c)(m,3)	'ifa2lfa'	hidden
;-      convert wordheader linkfield ptr to indexfield ptr
ifa2lfa:GA M;:ok:
; c: -/-
	lea eax,[eax-whead.i+whead.l]
M ifa2lfa
	ret
;-
;- idx>wid	( ix -- wid | 0 )				'idx2wid'	hidden
;-	convert vocabulary name's index to wordlist identifier

;- 
;- wid>idx	( wid -- ix | -1 )				'wid2idx'	hidden
;-	convert wordlist identifier to vocabulary name's index
;-	check { ix (lkindex) @ u< } whether ix is valid.
wid2idx:GA;:ok:
; c: ecx,ef
	mov ecx,[kref eax]
	test ecx,ecx
.r:
	jz near rtrue
	cmp ecx,KR
	jz .r
	call wid2lfa
;-
;- lfa>idx	( n1 -- n2 )			(m,3)		'lfa2idx'	forth
;-	convert wordheader linkfield ptr to uot index
lfa2idx:GA M;:ok:
; c: -/-
	movsx eax,word[kref eax+whead.i-whead.l]
M lfa2idx
	ret
;-
;- xec>lfa	( x -- p | 0 )					'xec2lfa'	hidden
;-	search uot for x, return lfa ptr p, corresponding to uot index.
;-	scan all vocabularies, independently of actual search order.
xec2lfa:GA;:ok:
; c: ecx
	call xec2idx
	jns idx2lfa
	inc eax
	ret
;-
;- idx>lfa	( ix -- lfa | 0 )		(cm,37)		'idx2lfa'	forth
;-	convert word index to linkfield ptr
;-	scan all {voc-link}ed wordlists, independently of actual search order.
idx2lfa:GA C;:ok:
; c: ef
	cmp eax,[byte uref uv.lkindex]
	push ecx
	push edx
	push KR
	jnc .r		;?; not within index range
	lea edx,[byte uref uv.voclink]; voc-link
.v:
	sub edx,[edx]	; voc-link dup @ -
	jz .r		;?; end of voc linkage
	cmp edx,KR
	jz .r		;?; end of voc linkage
	lea ecx,[edx-vhead.m+vhead.p]; lfa>wid
.l:
	cmp dword[ecx],byte 0
	jz .v
	sub ecx,[ecx]	; wid>top or next lfa, rsp
	jz .v
	cmp [ecx+whead.i-whead.l],ax;index
	jnz .l
	mov [esp],ecx
.r:
	pop eax
	sub eax,KR	; 0 or lfa
	pop edx
	pop ecx
C idx2lfa
	ret
;-
;- lfa>ffa	( n1 -- n2 )			(c)(m,3)	'lfa2ffa'	hidden
;-	convert wordheader linkfield ptr to compile-flags field ptr
lfa2ffa:GA M;:ok:
; c: -/-
	lea eax,[eax-whead.l+whead.f]
M lfa2ffa
	ret
;-
;- lfa>mfa	( n1 -- n2 )			(c)(m,3)	'lfa2mfa'	hidden
;-	convert wordheader linkfield ptr to macrofield ptr
lfa2mfa:GA M;:ok:
; c: -/-
	lea eax,[eax-whead.l+whead.m]
M lfa2mfa
	ret
;-
;- lfa>wid	( p1 -- p2 | 0 )				'lfa2wid'	hidden
;-	convert p1 lfa to ptr p2 to voclink posn
lfa2wid:GA;:ok:
; c: ecx, ef
	call lfa2pfa
	lea eax,[eax+4+vhead.i+vhead.m]	; voclink posn
	ret
;-
;- wid>lfa	( p1 -- p2 )			(m,6)		'wid2lfa'	hidden
;-	find primary lfa of vocabulary wordlist
wid2lfa:GA M;:ok:
; c: ef
	lea eax,[eax-vhead.l+vhead.m]	; voclink->disp
	sub eax,[kref eax]		; ref to lfa
M wid2lfa
	ret
;-
;- idx>pfa	( ix -- p | 0 )					'idx2pfa'	hidden
;-	convert uot index ix (execution token) to pfa p
idx2pfa:GA;:ok:
; c: ecx
	call idx2lfa
	jz lfa2pfa.r
;-
;- lfa>pfa	( n1 -- n2 )			(m,14)		'lfa2pfa'	forth
;-	convert wordheader linkfield ptr to parameterfield ptr
lfa2pfa:GA M;:ok:
; c: ecx
	lea eax,[eax+whead.i-whead.l+whead.n]
	movzx ecx,byte[kref eax]
	lea eax,[eax+ecx+4]
	and eax,byte -4
.r:
M lfa2pfa
	ret
;-
;- pfa>lfa	( p1 -- p2 | 0 )		(cm,47)		'pfa2lfa'	hidden
;-	convert wordheader parameterfield ptr to linkfield ptr
;-	ret 0 if no corresponding lfa found.
pfa2lfa:GA C;:ok:
; c: eflags
	push ecx,edx
.d:
	dec eax
	cmp byte[kref eax],0
	jz .d
	mov ecx,[uref uv.uwidth]
	xor edx,edx	;cnt
	inc eax
	dec edx
.w:
	dec eax
	inc edx
	cmp byte[kref eax],dl
	loopnz .w
	jz .r
.n:
	mov eax,whead.n-whead.l
.r:
%if 1
	test eax,3	;;;
%else
	mov edx,eax
	and dl,-4
	cmp eax,edx	; check for valid ptr, i.e. 4-aligned
%endif
	jnz .n
	lea eax,[eax-(whead.n-whead.l)]
	pop regl
C pfa2lfa
	ret
;-
;- >data  	( ix -- p )			(l4)		'todata'	root
;-	ref into re-located data-field in RAM of kernel-supplied variables,
;-	etc; aequivalent to -re- {>body}, applicable to kernel words, only.
;-	NOTE: kernel words store disp into data-space wrt @{d0} while new, 
;-	{create}d words store the disp to "body" directly as a kref'd ptr.
todata:GA;:ok:
; c: ecx
	call tobody
	jz .r
.p:
	mov eax,[kref eax]
	add eax,databot
.r:
	ret
;-
;- lfa>body	( lfa -- p )					'lfa2body'	hidden
;-	convert lfa ptr to ptr to words' data space
lfa2body:GA;:ok:
; c: ecx,ef
	call lfa2idx
;-
;- >body  	( ix -- p )			(a:c)		'tobody'	root
;-	ANS defines this to fetch the reference where "," etc. would store wrt a 
;-	-re- {create}d word which, with lib4th is data-memory and does not apply
;-	to kernel defined words' relocated alterable memory address -re- {>data}.
tobody:GA;:ok:
; c: ecx,ef
	call idx2pfa
	jz .r
.p:
	lea eax,[eax+4]
.r:
	ret
;-
;- latest  	( -- p )			(m,10)				forth
;-	return lfa of latest word in {current} vocabulary, which
;-	is the wordlist currently extended by new definitions.
latest:GA M;:ok:
; c: ef
	spush
	mov eax,[byte uref uv.current]
;-
;- wid>top	( p1 -- p2 )			(m,6)		'wid2top'	hidden
;-	fetch top lfa p2 of vocabulary wordlist from @voclink p1
wid2top:GA M;:ok:
; c: ef
	lea eax,[eax-vhead.m+vhead.p]; disp to top lfa
	sub eax,[kref eax]
M wid2top
M latest
	ret
;-
;- uot>idx	( p -- ix )			(p)(m,13)	'uot2idx'	hidden
;-	get word index ix by uot execution ptr p.
;-	actually valid range of ix is 0 to @(lkindex)-1,
;-	absolute max is { (#tib) (ltab) - 4/ }, always less than 65535.
uot2idx:GA M;:ok:
; c: ecx, ef
	add eax,KR
	lea ecx,[uref uv.ultab]
	sub eax,ecx
	sar eax,2
M uot2idx
	ret
;-
;- idx>uds	( ix -- p )			(p)(m,7)	'idx2uds'	hidden
;-	return disp to base of uot for indirect call [uref uv.ultab+disp].
idx2uds:GA M;:ok:
; c: -/-
	lea eax,[eax*4+uv.ultab-(uv.link)]
M idx2uds
	ret
;-
;- uds>idx	( n -- ix )			(p)(m,9)	'uds2idx'	hidden
;-	return ix index into uot from compiled disp n wrt uot base ptr.
uds2idx:GA M;:ok:
; c: ef
	lea eax,[eax+(uv.link)-(uv.ultab)]
	shr eax,2
M uds2idx
	ret
;-
;- xec>uot  	( x -- p )					'xec2uot'	hidden
;-	return uot ptr p to execution ptr x, ret 0 if x not in uot
xec2uot:GA;:ok:
; c: ecx,ef
	call xec2idx
;-
;- idx>uot	( ix -- p | 0 )			(p)(m,18)	'idx2uot'	hidden
;-	return ptr p to execution ptr by index ix, ret 0 if ix not in uot
idx2uot:GA M;:ok:
; c: ecx,ef
	cmp eax,[byte uref uv.lkindex]
.e:
	mov ecx,KR
	jnc .r
	lea ecx,[eax*4+uref uv.ultab]
.r:
	mov eax,ecx
	sub eax,KR
M idx2uot
	ret
;-
;- lfa>xec	( p -- x | 0 )					'lfa2xec'	hidden
;-	return execution ptr x from uot pointed to by wordheader via lfa p
lfa2xec:GA;:ok:
; c: ef
	movsx eax,word[kref eax+whead.i-whead.l]
;-
;- idx>xec	( ix -- x | 0 )					'idx2xec'	hidden
;-	return executionptr from uot by word index ix.
;-	(debugging: x is equal to code-ptr in assembly listing)
idx2xec:GA;:ok:
; c: ef
	cmp eax,[byte uref uv.lkindex]
	jnc near rfalse		;?; prevent segfault after false index value
	mov eax,[eax*4+uref uv.ultab]
	sub eax,KR
	ret
;-
;- idx>do	( ix -- x | 0 )					'idx2do'	hidden
;-	DOES>-entry of word by ix or, zero if not applicable.
idx2do:GA;:ok:
	spush
	call idx2xec
	cmp eax,byte 1
	jc .r			;?; 0, false ix, ret NZ
	push eax		; save xec entry, does>-part
	mov eax,TOS(1)
	call idx2pfa
	pop ecx			; xec-ptr by uot
	mov edx,[kref eax]	; by word header
	xor eax,eax
	cmp edx,ecx		; should equal
	jnz .r			;?; not a word w/ DOES>
	lea eax,[ecx+2+4+1+4]	; <lit><num><jmp><disp>, does-entry is not cpl'd by uot index
	add eax,[kref eax-4]	; ptr to does> entry
.r:
	TOSPULL(1)
	ret
;-
;- dsp>xec	( p1 -- p2 )			(p)(m,9)	'dsp2xec'	hidden
;-	return xec ptr p2 from dword disp at ptr p1.
dsp2xec:GA M
; c: ef
	add eax,[kref eax]
	lea eax,[eax+4]
M dsp2xec
	ret
;-
;- xec>idx	( x -- ix | -1 )		(cm,36)		'xec2idx'	hidden
;-	return uot index of execution ptr x, else -1
;-	(exec ptr is to library code, not to got of user program!
xec2idx:GA C;:ok:
; c: ef
	push edi,edx,ecx
	add eax,KR		; uot ref is true address
	lea edi,[uref uv.ultab]
	lea ecx,[uref uv.vbot-4]
	sub ecx,edi
	shr ecx,2
	mov edx,ecx
	repnz scasd
	jz .r
.n:
	mov ecx,edx
.r:
	inc ecx
	mov eax,edx		; top index
	sub eax,ecx		; last
	pop edi,edx,ecx
C xec2idx
	ret
%endif; __WHEAD_I
;-
;-=====================================================================;
;- l4whead.inc <eof>
