;- Copyright (C) 2000 H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-=====================================================================;
;-
;-  file  	: lib4th.asm
;-  created	: 10-jun-00
;-  modified	: -> re "version.inc"
;-  version	: 0.01(+)
;-  switches	: __LIB4TH
;-		: for proper alignment: -D__OPTIMIZE__=__O_SPEED__
;-  assembler	: nasm 0.98-3e (patched & bugs fixed 0.98)
;-  description	: "lib4th" shared library Forth kernel, assembly control
;-  author	: H-Peter Recktenwald, Berlin <phpr@snafu.de>
;-  comment	: a minimal forth system, further doc re "glossary.txt",
;-		: "lib4words.inc" for which words belong to which vocabulary
;-  source	: (me)
;-  requires	: running an application:
;-		:	no prerequisites other than those implied by the 
;-		:	program concerned, plus ld-linux.so (no "libc"!)
;-		: lo-level programming:
;-		:	with lib4th is just greatly simplified assembly,
;-		:	requiring knowledge of 4th & "lib4th" glossary
;-		: compiling prerequisites:
;-		:	asmutils 0.09, nasm-0.98-3e (patched 0.98)
;-=====================================================================;
;-
;-
%ifndef __LIB4TH
%define __LIB4TH

%ifidn __NASM,9820
%elifidn __NASM,9821
%elif
%macro l 0-*.nolist
;[nolist]
%endm
%endif
;[list]
; NOTE:
; Due to nasm specific issues (deficiencies) DO NOT CHANGE the general structures
; SETUP AND LOCATIONS! Current arrangement found feasible after many (weeks of)
; trials and errors and apparently is on a very narrow path of possible solutions!
; When adding further code watch to never put anything else on a labeled line, 
; "nasm" might compile faulty binaries in the shared lib case, otherwise!
;

;-
;- %define __ANS	; assemble ANS-std defaults & initiation, else (fig-)forth
;- %undef  __ANS	; preferrably, set switch in Makefile - ANS is default
%ifdef __fig
%define __FIG
%elifdef __FIG
%define __fig
%else
%define __ANS	; default to ANS mode
%endif
%ifdef __ans
%define __ANS
%elifdef __ANS
%define __ans
%endif

%undef __singsig
;
%define quit_sig
%undef  quit_sig

;-
%ifndef __STATIC
%define __DBG
;%undef __DBG	; assembly w. empty debugging routines, or undef'd, below
%else
%undef __DBG
%endif

%define __tosi	; display stack flow & system vari
%undef  __tosi	; in-active

%define __NO_PAGE_ALIGN; disable page alignment

;-
;[nolist]
%ifndef l4l4inc
%define l4l4inc
%include "l4-lib4th.inc"	; l4 includefiles
%endif
%ifidn __NASM,9821
[list +]
%endif
;[list]
;-
;- %include "lib4words.asm"
;- to programming with "lib4th" is what "..h" headerfiles inclusion
;- is to "C"-ish source programming. It should be included prior to
;- any code writen in an application program, prior to .text section.
;
;%include "lib4words.asm" included near <eof>, after all words defn's
;-
;%ifndef __DBG		; inner switch
;%undef __MK_DBG		; compile-swith
;%else
;%ifndef debuginc
;%include "dbg/l4-debug.inc"	; debugging macros
;%endif
;%ifndef __MK_DBG
;%undef __DBG
;%endif
;%endif
%ifidn __NASM,9821
[list +]
%endif
;[list]
;-
;-================================================================
; org = 08048000
;-  kernel code
;-  basic, initial references
;-

;-
;- Y						(k)
;-	1st address in .text, <kref> local to lib4th
;-	this address stores lib4th kernel+headers size, <Y> until <YE>.
; -	address occupied by dword <zero>, 1st valid 4th ptr is kref(4)
Y:GA
	dd YE-Y
; default initial entry from application - re ini4th
kvquit:A
	dd quit-Y
;-
;-================================================================
;- -- debugging
;-
%include "l4debug.inc"

;- -- debugging ends
;-================================================================
;- -- text, cc4th, see decoding
;-
%include "l4text.inc"

;-
;-================================================================================
;- -- host definitions source file names :
;-
%include "l4lxsrc.inc"

;-
;-================================================================================
;- -- ctrlchars, ctrlbits, stty sane :
;-
%include "l4tty.inc"

;-
;-================================================================
;- -- syscall names
;-	src/sys_{os-name}.inc
;- -- vt ctrl stg-s
;-	src/l4vtseq.inc
;-	! do not change names and sequence labels/data !
sysc:A
 dw   0	; start marker
.exit:
%include "scdata.inc"
	align 2,db 0
sysc.enum:;A
 dw (control-$)/2
 dw sysc.exit-$
 dw sysc.enum-$
; do not change names and sequence of non-local labels
%include "l4vtseq.inc"
;:TERMINFO: (re ncurses)
;Libraries reading terminfo entries are expected to check for a TERMINFO directory, 
;at $HOME/.terminfo if TERMINFO is not set and, finally in /usr/share/terminfo.

;-
;-
;-================================================================
;- -- signals constants and error messages, jumptables
;-    by macros <ertstg>, <ertdsp>, <sigstg>, <sigdsp>
;-	src/l4ersigt.inc
%include "l4ersigt.inc"

;-
;-================================================================
;- -- signal handlers & setup
;-	src/l4sighnd.inc
%include "l4sighnd.inc"

;-
;-================================================================
;- -- initial memory setup, yet all preliminary and highly experimental!
;- -- program initialization, jobs setup & control
;-
%include "l4init.inc"

	PAGE_ALIGN
;-
;- <nimp>
;	common entry for (yet?) in-active words
nimp:GA;:ok:
	spush
	mov eax,-38
	call kerror
	jmp quit
;-
;- getrefb	( -- )				(k)				system
;-	restore kernel library reference register (KR=ebx)
;-	subroutines, 4th-words, all data refer to this address;
;-	for direct 4th/asm programming, single one non-relative reference.
getrefb:GA;:ok:(for kernel/asm access)
; c: -/- & #must# preserve eflags!
	call .r
.n:
	not ebx
	lea ebx,[ebx+.n-Y]
	not ebx
	ret
.r:A
	mov ebx,[esp]
	ret	
;-
;- -- end of initial lo-level routines
;-================================================================
;- -- basic memory allocation in .ss and .bss
;-
%include "l4sys.inc"

;-
;-================================================================
;-
;- -- error handling, messages.
;-
%include "l4errhd.inc"

;-
;-================================================================
;- -- memory
;-
%include "l4smem.inc"

;-
;-================================================================
;- -- m-tasking
;-
%include "l4jobs.inc"

;-
;-================================================================
;- -- i/o port access
;-
%include "l4ioport.inc"

;-
;-================================================================
;- -- system time
;-
%include "l4time.inc"

;-
;-================================================================
;- -- constants
;-
%include "l4constants.inc"

;-
;-----------------------------------------------------------------
;- -- save/restore input
;- -- file/devices i/o support, operations refer to {work} channel
;- -- i/o channels & redirection, tests
;-
;-	re man(2) fcntl, ioctl, open, close
;-	(Vorsicht: einige "man de"-Seiten enthalten fehler!)
;-	error code is any -ve value, non-error results dependent on opr.
;-	
%include "l4file.inc"

;-
;-----------------------------------------------------------------
;- -- i/o source, block, screenfile support
;- -- the f.i.g. line editor (n.i.)
;-
%include "l4block.inc"

;-
;-----------------------------------------------------------------
;- -- vt... and ctrl sequences
;-	
%include "l4vt.inc"

;-
;-----------------------------------------------------------------
;- -- standard i/o (more or less...)
;-
%include "l4consio.inc"

;-
;-================================================================
;--- basic data-stack opr
;-
%include "l4dstack.inc"

;-
;---------------------------------------------------------------------
;---  returnstack
;-	push/pop to returnstack adds/subtracts <kref> value
;-	2>r, r> r> cannot be used to swap cells! re {rswap}.
;-
%include "l4rstack.inc"

;-
;-================================================================
;---  memory access
;-	false data may result if F8 stacks accessed directly!
;-	explicit addresses shall be passed as ptr's wrt kref.
;-	fetched data items not sign extended, unless mentioned.
;-
%include "l4memfsto.inc"

;-
;-================================================================
;---  arithmetic logic opr
;-
%include "l4alogic.inc"

;-
;-================================================================
;--- arithmetic opr
;-
%include "l4dsarith.inc"

;-	
;-
;-----------------------------------------------------------------
;--- mixed quad and double
;-	
%include "l4qdarith.inc"

;-	
;-
;-================================================================
;--- multi praecision opr
;-

; forth in "l4bignum.inc":
;	ndup, n>r, nr>, s>m

%include "l4bignum.inc"

;-
;-
;-----------------------------------------------------------------
;--- number conversion
;-
%include "l4numconv.inc"

;-
;----------------------------------------------------------------------------------
;--- assembler cons
;-
%include "l4assembler.inc"

;-
;-================================================================
;--- kernel and user vari/cons
;-
%include "l4uvari.inc"

;-
;-================================================================
;--- search order
;-	unless mentioned all vocabularies belong to {root},
;-	{root} vocabulary is sub-vocabulary of itself.
;-	lib4th supplied vocabularies are set "immediate".
;-	
%include "l4sorder.inc"

;-
;-----------------------------------
;---  tools
;-	
%include "l4tools.inc"

;-
;-================================================================
;--- text acquisition, strings, search & substitution
;-
%include "l4string.inc"

;-	
;-
;-================================================================
;--- interpreter, word headers structure
;-	conversion words all but {idx>lfa} & {>body} in {hidden}.
%include "l4whead.inc"

;-
;-------------------------------------------------------------------------
;--- interpreter, forget, find
;-	
%include "l4interpret.inc"

;-
;-------------------------------------------------------------------------
;-	
; -- Compiler support
%include "l4cpl.inc"

;-
;-----------------------------------------------------------------
;- -- local memory
;-	<top> @s0[ data-stack ... tos]			- no locals
;-	<top> ..locals.. @s0[ data-stack ... tos]	- locals allocated
;-	most recent local data allocated @s0, stack content moved "up".
;-	ptr to <top> is absolute low limit for @s0 and locals and, can be
;-	used to control error situations if stack-ptrs became un-certain.
;-	
%include "l4locals.inc"

;-
;-----------------------------------------------------------------
;- -- word header, defining words ---
;-	
%include "l4defining.inc"

;-
;-----------------------------------------------------------------
;- -- control structures
;-	
%include "l4cstruct.inc"

;-----------------------------------------------------------------
%ifdef  __RATIONAL
%include "l4rational.inc"
%ifdef __SRANUM
%include "l4sranum.inc"
%endif
%endif

;-
; -----------------------------------------------------------------
%include "l4envq.inc"

;-
;-================================================================
;-
;--- initial wordlist
;-

%include "l4words.inc"
%ifidn __NASM,9821
[list +]
%endif
;[list]
;-
;-================================================================
;-	library system memory, not accessible, i.e. no .data, no .bss! 
;-	256-alignment required, <YE> is final label 

; page bounds to simplify permissions setting
	align PAGE_SIZE,db -1
YE:
    %pop; check sourcefiles consistency, corresponding %push in "asmutils/system.inc"
;-
;-
%endif; __LIB4TH
;-================================================================
;- lib4th.asm <eof>
