#!/bin/sh
# for lib4th, 6/2000 by h-peter recktenwald <phpr@snafu.de>

# hp'27-04-00, v0.03, GNU GPL licensed, h-peter recktenwald, <phpr@snafu.de>
# ...extra fr Christian Fowelin <christian@fowelin.de> und seine libASM :)
# generate glossary of *.s and *.h files from ";;;*" leading comments
# execute from directory where the final glossary file(s) should reside.
# no --help provided, comments should suffice, else pse, mailme... (hp)

# MK.glossary [ args ]
# up to two, each optional, arguments accepted:
#	a directory argument for top directory of sources path,
#		use "../src" if arg explicitely passed as "..".
#	a(ny) non-directory argument used to leave intermediate ..gls files in place.

# avoid preset names' conflicts
unset a c d i o k l m md t tm p r s w x CUT EGREP FIND GREP HEAD RM TAIL

# ----------------------------------------------------------------------------------------- #
# vari, to speeding up PATH searching

CUT=$(type -p cut)
EGREP=$(type -p egrep)
FIND=$(type -p find)
GREP=$(type -p grep)
HEAD=$(type -p head)
RM=$(type -p /bin/rm||type -p rm)
TAIL=$(type -p tail)

# vari end
# ----------------------------------------------------------------------------------------- #
# configurable (watch the different quotation modes!)

# output path/filename
o=./glossary

# default source directories may also be set to several, space delimited path's
# "quoted" string for path argument(s) to <${FIND}>:
# default if no directory type arg received:
s=$(dirname $(pwd -P||pwd))/src
# this will be taken if directory arg was "..":
p=$(dirname $(pwd))/src

# filenames selection to scan for comment lines
# 'quoted' string for -iname argument to <${FIND}>:
c='*.[sSh]';		#1#2#
c='*.[aim]*[mc]';	#3#
# filter filenames ea. to include in index-file
a='*.*[chmsS]'		#1#2#
a='*.*[mc]'		#3#

# leading character sequence of what should be considered a comment
# 'quoted' string for searchstring argument to <${EGREP}>:
w="^;;;\*";		#1# libASM
w="^;;;";		#2# old(?) libASM
w='^;?;-';		#3# hp asm-files comments
#lis2lis="$(dirname ${0})/lis2lis.as"
#[ -x "${lis2lis}" ]||lis2lis=

# per line what <${EGREP}> should exclude when above string found
# 'quoted' string for -v string argument to <${EGREP}>:
x='Hey\ '

# field marker prepending 1st char of glossary lines, e.g. last char of above ${w}
# single character for -d option of <${CUT}>:
k="$(echo ${w}|${CUT} -b$(($(echo ${w}|wc -c)-1)))"

# ${k} specified fields count where to begin comment for glossary
# numeric value for -f option of <${CUT}>:
f=10;			#1#
f=4;			#2#
f=2;			#3#
# f=2: use lis2lis utility, if present, to extracting lines beginning w. ";-"
[ "${f}" == "2" ]&&
{
    lis2lis="$(dirname ${0})/lis2lis.as"
    [ -x "${lis2lis}" ]||lis2lis=
}
f=$((${f}&7))

#function to
# returning 1st line no. of gls-files content supposed to copying to glossary.
# the entire file will be omitted if this function returns 0 or -ve
# which can be used to further limit selection to additional criteria, for instance:
#	if 2nd occurrence of string "==========================" found
#	excluding those lines with a ":" in their 1st colum,
#	copy this gls-files content, beginning with the next line.
#3#
glstop(){
    echo $(($(${GREP} -n "==========================" ${1}|${CUT} -d: -f1|${HEAD} -n2|${TAIL} -n1)))
}
#1#2# start after 1st line where "Released" and in 1st column no ":" was found,
#    echo $(($(${GREP} -n "Released" ${1}|${CUT} -d: -f1|${HEAD} -n1|${TAIL} -n1)))
#		        text^; where^; xclude^char;  text ^occurence
# or, count only those lines where "Released" with trailing and no leading ":" found:
#    echo $(($(${CUT} -d: -f1 ${1}|${GREP} -n "Released" ${1}|${HEAD} -n1|${TAIL} -n1)))

# send debugging text to ${0}.dbg, cannot be used in free running mode - "&"
debug=
#"tja..."

# configurable ends
# ----------------------------------------------------------------------------------------- #
# shell fu-s

# tempfiles storage
mktmp(){
    echo "/tmp"
}

# extract glossary lines from file ${1} to stdout, suppress multiple empty lines
mkgla(){
    local o
#dbg "mkg A:${a} 1:${1} "#c:$(echo ${1}|${CUT} -c$(($(($(echo ${1}|wc -c)))-1)))"
    o=${1}.gls
#    a=${a} $(echo ${1}|${CUT} -c$(($(($(echo ${1}|wc -c)))-1)))
    ${EGREP} -l "${w}" ${1} >/dev/null&&
    {
	echo >${o}
	echo -ne " <$(echo ${1}|${CUT} -d/ -f2-)>\n ">>${o}
	date >>${o};echo >>${o}
	[ -n "${lis2lis}" ]&&${lis2lis} + <${1}|${GREP} -v "${x}">>${o}||${EGREP} "${w}" ${1}|cat -s|${GREP} -v "${x}"|${CUT} -d${k} -f${f}->>${o}
    }
}

# ${CUT} (hopefully) informative content out of ..gls files to stdout
glscut(){
    local e l
    l=$(($(basename $(wc -l ${1}))-2))
    e=$(glstop ${1})
    [ ${e} = 0 ]&&return
    l=$((${l}-1-${e}))||return
    [ $((${l}>1)) -lt $((2)) ]||return
    ${TAIL} -n1 ${1}|${HEAD} -n2 ${1}
    echo -e "--------------------------------------------------------------------------\n"
    ${TAIL} -n${l} ${1}|cat -s
    echo -e "\n+ -- +"
}

# debug...
dbg(){
    [ -z "${debug}" ]||echo ${@}>>${0}.dbg
}

# shell fu-s end
# ----------------------------------------------------------------------------------------- #
# evaluate...

# fetch optional directory and flag (i.e. any non-directory) arguments
# flag used as a signal to NOT removeing newly generated ..gls files
r="${1}";shift
[ -z "${r}" ]||{ [ -d $(echo ${r}|${CUT} -d\  -f1) ]&&{ d="${r}";r=""; }; }
[ -z "${1}" ]||{ [ -n "${r}" ]&&{ [ -d $(echo ${1}|${CUT} -d\  -f1) ];d="${1}"; }||r="${1}"; }
[ -d "${r}" ]&&r=""

# ".." special if called from one dir 'below' main, for instance in "docs"
[ "${d}" == ".." ]&&d=${p}
# use set default if no directory arg passed
[ -z "${d}" ]&&d="${s}"
# reduce multiple path recursive search to single level
[ $(($(echo ${d}|wc -w)>1)) == 1 ]&&md="-maxdepth 1"||md="-depth"

# temps
m=$(mktmp)
[ -n "${m}" ]||m=/tmp
t=${m}/G.${$}
g=${m}/G.${$}g

[ -z "${debug}" ]||${RM} -f ${0}.dbg
dbg "CUT=$CUT EGREP=$EGREP FIND=$FIND GREP=$GREP HEAD=$HEAD RM=$RM TAIL=$TAIL"

#clear temp storage
${RM} -f ${m}/G.*

dbg "1 d:$d c:$c d:$md"

# remove old ..gls files
# this has also the effect of cacheing CWD, thus speeding up what follows
${FIND} ${d} ! -type d -name "*.gls" -exec ${RM} -f \{\} \;

# generate specific gls files
for i in $(${FIND} ${d} ${md} ! -type d -iname "${c}");do dbg "I:$i";mkgla ${i};done

dbg "2 d:$d x:$x"

# generate sorted gls files reference
${FIND} ${d} ${md} ! -type d -name "*.gls" -exec basename \{\}>>${t} \;
sort -f -d +0.0 -1.0 ${t} -o ${t}s

# document header# m=$(echo $(./relse.sh));echo $m
m=$(dirname $(dirname ${0}))
m=$(cd ${m};./relse.sh)
dbg m:${m}
echo -ne "\tGlossary of \"`basename ${d}`\", ver ${m}  -  ">${g}
date>>${g}
[ -r "./LEGAL" ]&&
    echo -e "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n">>${g}&&
    {
	[ -z "${lis2lis}" ]&&
	${FIND} ${d} ${md} ! -type d -name "LEGAL" -exec cat \{\} >>${g} \;||
	{ i=$(${FIND} ${d} ${md} ! -type d -name "LEGAL");${lis2lis} + <${i} >>${g}; }
    }&&
    echo -e "\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n">>${g}||
    echo -e "--------------------------------------------------------------------------\n">>${g}
dbg "3 d:$d"

# copy gls-files to glossary by sorted filenames
for i in $(cat ${t}s)
do
    for l in $(${FIND} ${d} ${md} ! -type d -name ${i});do glscut "${l}" >>${g};done;#2>/dev/null
done
echo "<eof> $(basename ${d})-${m} .hpr'$(date +%Y)">>${g}

dbg "4 d:$d g:$g o:$o"

# final docs, clean up
[ "${?}" == "0" ]&&
{
## weidergeier, warum damit kein text ankommt:
##    cp ${g} > ${o}.txt
# compress glossary
    gzip -9cf ${g}>${o}.gz
# delete intermediates"
    [ -n "${r}" ]||
    {
	${FIND} ${d} ${md} ! -type d -name "*.gls" -exec ${RM} -rf \{\} \;&
    }
# generate (simplistic) indexfile: files names per glossary line no.
    echo -e "\t<${o}> v${m} filenames' index by lines\n--------------------------------------------------------------------------\n">${o}.idx
    ${GREP} -n "^\ <\<*/${a}\>>$" ${g}|grep "/">>${o}.idx
## mit und ohne gzip in ordnung (s.o.)
#    gzip -9f ${o}.idx &
# delete temps
    ${RM} -f ${g}* ${t}* &
}
## ok mit diesem umweg (s.o.)
[ -r "${o}.gz" ]&&gzip -df ${o}.gz &

# MK.glossary <eof>
